/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.applications.morphing;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.border.*;
import javax.swing.table.*;

import com.frojasg1.applications.libmorphing.view.mainwindow.evt.MainWindowEvent;
import com.frojasg1.applications.libmorphing.view.mainwindow.evt.enums.MainWindowEventType;
import com.frojasg1.general.exceptions.ConfigurationException;
import com.frojasg1.applications.morphing.configuration.ApplicationConfiguration;
import com.frojasg1.applications.morphing.context.ApplicationContextImp;
import com.frojasg1.general.view.internationalization.window.InternException;
import com.frojasg1.applications.common.components.internationalization.window.InternationalizationInitializationEndCallback;
import com.frojasg1.applications.common.components.internationalization.window.InternationalizedJFrame;
import com.frojasg1.applications.common.components.resizecomp.MapResizeRelocateComponentItem;
import com.frojasg1.applications.common.components.resizecomp.ResizeRelocateItem;
import com.frojasg1.applications.common.configuration.listener.ConfigurationParameterListener;
import com.frojasg1.applications.common.configuration.listener.ConfigurationParameterObserved;
import com.frojasg1.applications.libmorphing.frames.MorphingFrameSequencer;
import com.frojasg1.applications.libmorphing.model.images.image.MorphingImageModel;
import com.frojasg1.applications.morphing.imagelisttable.MorphingImageModelListTableWrapper;
import com.frojasg1.applications.libmorphing.view.timers.SwingImageSequenceShower;
import com.frojasg1.applications.morphing.language.MorphingApplicationInternationalTexts;
import com.frojasg1.libgeometry.types.GeometricRectangle;
import com.frojasg1.libgeometry.types.factory.GeometryCommonFactory;
import com.frojasg1.general.CollectionFunctions;
import com.frojasg1.general.DateFunctions;
import com.frojasg1.general.collection.functions.ListStateContextBase;
import com.frojasg1.general.desktop.view.about.GenericAboutJDialog;
import com.frojasg1.general.desktop.view.color.ColorInversor;
import com.frojasg1.general.desktop.view.icons.IconResources;
import com.frojasg1.general.desktop.view.panels.impl.file.FileHandlingPanelEvent;
import com.frojasg1.general.desktop.view.panels.impl.file.FileHandlingPanelEventType;
import com.frojasg1.general.desktop.view.panels.impl.file.FileHandllingJPanel;
import com.frojasg1.general.desktop.view.pdf.ImageJPanel;
import com.frojasg1.general.desktop.view.pdf.ImageJPanelSizeParams;
import com.frojasg1.general.desktop.view.zoom.mapper.ComponentMapper;
import com.frojasg1.general.dialogs.DialogsWrapper;
import com.frojasg1.general.number.IntegerFunctions;
import com.frojasg1.general.view.evt.wrapper.GenericWrapperEvent;
import com.frojasg1.general.view.evt.wrapper.impl.table.impl.GenTableWrapperRecordsModifiedEventBase;
import com.frojasg1.generic.GenericFunctions;
import java.awt.Dimension;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.awt.image.BufferedImage;
import java.util.List;
import java.util.Properties;
import java.util.Vector;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicReference;
import java.util.function.Consumer;
import javax.swing.JPopupMenu;
import javax.swing.JScrollBar;
import javax.swing.SwingUtilities;
import javax.swing.event.DocumentEvent;
import javax.swing.event.DocumentListener;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;



// Preferred size: [955, 720]

/**
 *
 * @author Fran
 */
public class MainWindow extends InternationalizedJFrame< ApplicationContextImp >
		implements ConfigurationParameterListener//, MorphingView
{
	private static final Logger LOGGER = LoggerFactory.getLogger(MainWindow.class);

    public static final String CONF_DEMOS_OPENING_PROJECT_OF_VERSION_V1_0__URL = "DEMOS_OPENING_PROJECT_OF_VERSION_V1_0__URL";
    public static final String CONF_DEMOS_FACE_DETECTION_TRANSITION_PROJECT_1_URL = "DEMOS_FACE_DETECTION_TRANSITION_PROJECT_1_URL";
	public static final String CONF_DEMOS_NORMAL_TRANSITION_PROJECT_2_URL = "DEMOS_NORMAL_TRANSITION_PROJECT_2_URL";

	public static final String CONF_DEMOS_IMPORTED_MORPHING_PROJECT_VIDEO_URL = "DEMOS_IMPORTED_MORPHING_PROJECT_VIDEO_URL";
	public static final String CONF_DEMOS_EXAMPLE_1_MORPHING_VIDEO_URL = "DEMOS_EXAMPLE_1_MORPHING_VIDEO_URL";
    public static final String CONF_DEMOS_EXAMPLE_2_MORPHING_VIDEO_URL = "DEMOS_EXAMPLE_2_MORPHING_VIDEO_URL";
	public static final String CONF_DOWNLOAD_IMPORTED_PROJECT_FROM_V1_0_URL = "DOWNLOAD_IMPORTED_PROJECT_FROM_V1_0_URL";
	public static final String CONF_DOWNLOAD_EXAMPLE_1_PROJECT_URL = "DOWNLOAD_EXAMPLE_1_PROJECT_URL";
    public static final String CONF_DOWNLOAD_EXAMPLE_2_PROJECT_URL = "DOWNLOAD_EXAMPLE_2_PROJECT_URL";

	protected static final String CONF_USER_HANDBOOK_FILE_NAME = "USER_HANDBOOK_FILE_NAME";
	protected static final String CONF_ERROR_OPENING_FILE = "ERROR_OPENING_FILE";
	protected static final String CONF_ERROR_OPENING_HELP = "ERROR_OPENING_HELP";

	protected static final String CONF_ERROR_CREATING_MORPHING_FRAME = "ERROR_CREATING_MORPHING_FRAME";
	protected static final String CONF_IGNORE = "IGNORE";
	protected static final String CONF_CANCEL = "CANCEL";

	public static final String sa_configurationBaseFileName = "MainWindow";

	protected Vector<JPopupMenu> a_vectorJpopupMenus = null;

	protected GenericAboutJDialog _aboutDial = null;

	protected MorphingImageModelListTableWrapper _tableWrapper;
//	protected DelaunayTriangulationJPanel _delaunayPanel;

//	protected MorphingController _controller;

	protected FileHandllingJPanel _fileHandlingJPanel;

//	protected MorphingVerticesMapManager _morphingVerticesManager;
	protected AtomicBoolean _cancelMorphingSimulation = new AtomicBoolean();

	protected SwingImageSequenceShower _imageShower;
	protected ImageJPanel _imageJPanel;
	protected AtomicReference<ImageJPanelSizeParams> _imageJPanelSizeReference = new AtomicReference<>();

	protected DocumentListener _projectNameDocumentListener;

	protected Consumer<MainWindowEvent> mainWindowListener;

	/** Creates new form NewJFrame */
	public MainWindow( String windowName, ApplicationContextImp appCtx ) throws ConfigurationException
	{
		super( ApplicationConfiguration.instance(), appCtx, true );
		setName( windowName );

		initComponents();

		initOwnComponents();

		setWindowConfiguration( );
	}

	protected void addListeners()
	{
		if( _projectNameDocumentListener == null )
		{
			_projectNameDocumentListener = new DocumentListener() {
				@Override
				public void insertUpdate(DocumentEvent e) {
					updateTitle();
				}

				@Override
				public void removeUpdate(DocumentEvent e) {
					updateTitle();
				}

				@Override
				public void changedUpdate(DocumentEvent e) {
					updateTitle();
				}
			};
		}
		jTF_projectName.getDocument().addDocumentListener( _projectNameDocumentListener );
	}

	public void setMainWindowListener(Consumer<MainWindowEvent> listener) {
		mainWindowListener = listener;
	}

	protected Consumer<MainWindowEvent> getMainWindowListener() {
		return mainWindowListener;
	}

	@Override
	public void internationalizationInitializationEndCallback()
	{
		super.internationalizationInitializationEndCallback();

		SwingUtilities.invokeLater( () -> { setListeners(); } );
		setAfterWindowConfiguration();

		addListeners();

		postInitValues();
		M_applyNewApplicationConfiguration();
	}

	protected void postInitValues()
	{
		jTF_projectName.setText(
			MorphingApplicationInternationalTexts.instance().getInternationalString(
				MorphingApplicationInternationalTexts.CONF_DEFAULT_PROJECT_NAME )
		);
	}

	public void addDemoMenuItems()
	{
		addDemoMenuItem(jMI_demos_morphingOpeningProjectOfPreviousVersion, CONF_DEMOS_OPENING_PROJECT_OF_VERSION_V1_0__URL);
		addDemoMenuItem(jMI_demos_morphingNormalTransitionsProject1, CONF_DEMOS_FACE_DETECTION_TRANSITION_PROJECT_1_URL);
		addDemoMenuItem(jMI_demos_morphingProject2, CONF_DEMOS_NORMAL_TRANSITION_PROJECT_2_URL);
		addDemoMenuItem(jMI_demos_importedMorphingProjectExample, CONF_DEMOS_IMPORTED_MORPHING_PROJECT_VIDEO_URL);
		addDemoMenuItem(jMI_demos_example1, CONF_DEMOS_EXAMPLE_1_MORPHING_VIDEO_URL);
		addDemoMenuItem(jMI_demos_example2, CONF_DEMOS_EXAMPLE_2_MORPHING_VIDEO_URL);
		addDemoMenuItem(jMI_downloadImportedProjectFromv1_0, CONF_DOWNLOAD_IMPORTED_PROJECT_FROM_V1_0_URL);
		addDemoMenuItem(jMI_downloadExampleProject1, CONF_DOWNLOAD_EXAMPLE_1_PROJECT_URL);
		addDemoMenuItem(jMI_downloadExampleProject2, CONF_DOWNLOAD_EXAMPLE_2_PROJECT_URL);
	}

/*
	protected void setViewToController()
	{
		if( isInitialized() && ( _controller != null ) )
			_controller.setView(this);
		else if( _controller == null )
			LOGGER.info( "Controller was null" );
		else
			LOGGER.info( "Not initializaed yet" );
	}
*/
	@Override
	protected void invokeLaterAfterInitialized()
	{
//		setViewToController();
	}

	protected void setAfterWindowConfiguration()
	{
	}
/*
	public void setController(MorphingController _controller) {
		this._controller = _controller;
		setViewToController();
	}

	public MorphingController getController() {
		return _controller;
	}
*/
	protected void initOwnComponents()
	{
		assignTableWrapper();
//		_delaunayPanel = createDelaunayTriangulationJPanel();
//		addComponentToCompletelyFillParent( jPanel1, _delaunayPanel );
//		_delaunayPanel.init();
		_fileHandlingJPanel = createFileHandlerJPanel();
		addComponentToCompletelyFillParent(jPanel9, _fileHandlingJPanel );

		_imageJPanel = createImageJPanel();
		_imageJPanel.addComponentListener( new ComponentAdapter() {
			@Override
			public void componentResized(ComponentEvent e) {
				updateImageJPanelSize();
			}
		});

		jSplitPane1.setBottomComponent(_imageJPanel);
		SwingUtilities.invokeLater( () -> updateImageJPanelSize() );

		initValues();
	}

	protected ImageJPanelSizeParams createImageSizeParams()
	{
		ImageJPanelSizeParams result = new ImageJPanelSizeParams();
		result.setSize( _imageJPanel.getRealSize() );
		result.setZoomedWidth( _imageJPanel.getZoomedFrameWidth() );

		return( result );
	}

	protected void updateImageJPanelSize()
	{
		_imageJPanelSizeReference.set( createImageSizeParams() );
	}

	protected ImageJPanel createImageJPanel()
	{
		ImageJPanel result = new ImageJPanel( null );
		result.setCanInvertImageColors(false);
		result.setAllowShiftForMouseWheel(false);

		return( result );
	}

	public void addFileHandlingListener( Consumer<FileHandlingPanelEvent> listener )
	{
		_fileHandlingJPanel.addListener(listener);
	}

	public void removeFileHandlingListener( Consumer<FileHandlingPanelEvent> listener )
	{
		_fileHandlingJPanel.removeListener(listener);
	}

	protected FileHandllingJPanel createFileHandlerJPanel()
	{
		FileHandllingJPanel result = new FileHandllingJPanel();
		result.init();

		result.addAssociatedButton(FileHandlingPanelEventType.OPEN, jMI_open);
		result.addAssociatedButton(FileHandlingPanelEventType.SAVE, jMI_save);
		result.addAssociatedButton(FileHandlingPanelEventType.SAVE_AS, jMI_saveAs);

		return( result );
	}

	public void setFileName( String fileName )
	{
		jTF_morphingProjectFileName.setText(fileName);
	}

	public String getFileName()
	{
		return( jTF_morphingProjectFileName.getText() );
	}

	public String getProjectName()
	{
		return( jTF_projectName.getText() );
	}

	public void setProjectName( String projectName )
	{
		jTF_projectName.setText(projectName);
		updateTitle();
	}

	protected void updateTitle()
	{
		SwingUtilities.invokeLater( () -> setTitle( getProjectName() ) );
	}

	protected void assignTableWrapper()
	{
		_tableWrapper = createTableWrapper();
	}
/*
	protected DelaunayTriangulationJPanel createDelaunayTriangulationJPanel()
	{
		DelaunayTriangulationJPanel result = new DelaunayTriangulationJPanel(
						getApplicationContext().getAppliConf(),
						getApplicationContext().getDelaunayBuilder() );

		return( result );
	}
*/
	protected void initValues()
	{
		getTableWrapper().setTableColumnDataString( getAppliConf().getTableColumnsString() );
	}

	protected void setListeners()
	{
	}

	public void newImageChosen( MorphingImageModel mim )
	{
		getTableWrapper().setRecordLight(mim);
	}

	public void setNewImageListInternal(List<MorphingImageModel> imageList,
										Integer selectedIndex, int[] rowSelection,
										boolean mandatoryUpdate) {
		MorphingImageModel current = CollectionFunctions.instance().get( imageList, selectedIndex );
		getTableWrapper().setRecordList(imageList);
		if( mandatoryUpdate )
			SwingUtilities.invokeLater( getTableWrapper()::updateData );
//		selectNewTrack( current );
//		ExecutionFunctions.instance().nestedInvokeLater( 4, () ->
//						getTableWrapper().setModelRowSelection( rowSelection, null ) );
	}

	protected void setWindowConfiguration( )
	{
		MapResizeRelocateComponentItem mapRRCI = new MapResizeRelocateComponentItem();
		try
		{
			mapRRCI.putResizeRelocateComponentItem( jPanel1, ResizeRelocateItem.FILL_WHOLE_PARENT );
			mapRRCI.putResizeRelocateComponentItem( jSplitPane1, ResizeRelocateItem.RESIZE_TO_RIGHT +
																ResizeRelocateItem.RESIZE_TO_BOTTOM);
//			mapRRCI.putResizeRelocateComponentItem( _delaunayPanel, ResizeRelocateItem.FILL_WHOLE_PARENT );
			mapRRCI.putResizeRelocateComponentItem( jTF_morphingProjectFileName,
													ResizeRelocateItem.RESIZE_TO_RIGHT);

			boolean postponeInitialization = true;
			mapRRCI.putResizeRelocateComponentItem(jPanel9, 0);
			mapRRCI.putResizeRelocateComponentItem(_fileHandlingJPanel,
				ResizeRelocateItem.FILL_WHOLE_PARENT, postponeInitialization);
		}
		catch( Throwable th )
		{
			th.printStackTrace();
		}

		setLanguageMenu( jM_languageSubmenu );
		setZoomMenu( jM_zoom );

		setMenuItemIcons();

		createInternationalization(	getAppliConf().getConfigurationMainFolder(),
									getAppliConf().getApplicationNameFolder(),
									getAppliConf().getApplicationGroup(),
									getAppliConf().getInternationalPropertiesPathInJar(),
									sa_configurationBaseFileName,
									this,
									null,
									a_vectorJpopupMenus,
									true,
									mapRRCI );

		registerInternationalString(CONF_DEMOS_OPENING_PROJECT_OF_VERSION_V1_0__URL, "https://frojasg1.com/demos/aplicaciones/morphing/v1.1_morphingProjectEditor.OpeningVersionv1_0Project.mp4" );
		registerInternationalString(CONF_DEMOS_FACE_DETECTION_TRANSITION_PROJECT_1_URL, "https://frojasg1.com/demos/aplicaciones/morphing/v1.1_morphingProjectEditor.CreatingANormalTransitionProject1FromScratch.mp4" );
		registerInternationalString(CONF_DEMOS_NORMAL_TRANSITION_PROJECT_2_URL, "https://frojasg1.com/demos/aplicaciones/morphing/v1.1_morphingProjectEditor.CreatingFaceDetectionTransitionProject2FromScratch.mp4" );
		registerInternationalString(CONF_DEMOS_IMPORTED_MORPHING_PROJECT_VIDEO_URL, "https://frojasg1.com/demos/aplicaciones/morphing/v1.1_morphingProjectEditor.ImportedProjectExampleMorphingVideo1.mp4" );
		registerInternationalString(CONF_DEMOS_EXAMPLE_1_MORPHING_VIDEO_URL, "https://frojasg1.com/demos/aplicaciones/morphing/v1.1_morphingProjectEditor.ExampleMorphingVideo1.mp4" );
		registerInternationalString(CONF_DEMOS_EXAMPLE_2_MORPHING_VIDEO_URL, "https://frojasg1.com/demos/aplicaciones/morphing/v1.1_morphingProjectEditor.ExampleMorphingVideo2.mp4" );
		registerInternationalString(CONF_DOWNLOAD_IMPORTED_PROJECT_FROM_V1_0_URL, "https://frojasg1.com/demos/aplicaciones/morphing/20220414.v1.1.imported.Morphing.example.zip" );
		registerInternationalString(CONF_DOWNLOAD_EXAMPLE_1_PROJECT_URL, "https://frojasg1.com/demos/aplicaciones/morphing/20230618.Modelos-1.zip" );
		registerInternationalString(CONF_DOWNLOAD_EXAMPLE_2_PROJECT_URL, "https://frojasg1.com/demos/aplicaciones/morphing/20230618.Modelos-2.zip" );

		registerInternationalString(CONF_USER_HANDBOOK_FILE_NAME, "../_documents/_manual/Espanyol/Editor de proyectos de Morphing.v1.1.Manual de usuario.docx" );
		registerInternationalString(CONF_ERROR_OPENING_FILE, "Error opening file: $1" );
		registerInternationalString(CONF_ERROR_OPENING_HELP, "Error opening help" );

		registerInternationalString(CONF_ERROR_CREATING_MORPHING_FRAME, "Error creating morphing frame: $1\nYou can ignore or cancel.\nYou can inform the developer about this issue: frojasg1@hotmail.com" );
		registerInternationalString(CONF_IGNORE, "Ignore" );
		registerInternationalString(CONF_CANCEL, "Cancel" );

		boolean isClosure = false;
	}

	protected void setMenuItemIcons()
	{
		setDarkModeMenuItemIcon( jMI_darkMode );

		setIconForMenuItem( jMI_open, IconResources.OPEN_FILE_ICON_RESOURCE );
		setIconForMenuItem( jMI_save, IconResources.SAVE_FILE_ICON_RESOURCE );
		setIconForMenuItem( jMI_saveAs, IconResources.SAVE_FILE_AS_ICON_RESOURCE );
		setIconForMenuItem( jMI_importFaces, IconResources.IMPORT_ICON_RESOURCE );
		setIconForMenuItem( jMI_exportFaces, IconResources.EXPORT_ICON_RESOURCE );

		setIconForMenuItem( jMI_applicationConfiguration, IconResources.CONFIGURATION_ICON_RESOURCE );
		setIconForMenuItem( jMI_simulateMorphing, IconResources.PLAY_ICON_RESOURCE );
		setIconForMenuItem( jMI_cancelMorphingSimulation, IconResources.STOP_ICON_RESOURCE );

		setIconForMenuItem(jMI_downloadImportedProjectFromv1_0, IconResources.CLOUD_DOWNLOAD_RESOURCE );
		setIconForMenuItem(jMI_downloadExampleProject1, IconResources.CLOUD_DOWNLOAD_RESOURCE );
		setIconForMenuItem(jMI_downloadExampleProject2, IconResources.CLOUD_DOWNLOAD_RESOURCE );

		setIconForMenuItem(jMI_demos_morphingOpeningProjectOfPreviousVersion, IconResources.VIDEO_ICON_RESOURCE );
		setIconForMenuItem(jMI_demos_morphingNormalTransitionsProject1, IconResources.VIDEO_ICON_RESOURCE );
		setIconForMenuItem(jMI_demos_morphingProject2, IconResources.VIDEO_ICON_RESOURCE );
		setIconForMenuItem(jMI_demos_importedMorphingProjectExample, IconResources.VIDEO_ICON_RESOURCE );
		setIconForMenuItem(jMI_demos_example1, IconResources.VIDEO_ICON_RESOURCE );
		setIconForMenuItem(jMI_demos_example2, IconResources.VIDEO_ICON_RESOURCE );

		setIconForMenuItem( jMI_exit, IconResources.EXIT_ICON_RESOURCE );
		setIconForMenuItem( jM_demos, IconResources.VIDEO_ICON_RESOURCE );
		setIconForMenuItem( jMI_userHandbook, IconResources.USER_HANDBOOK_ICON_RESOURCE );
		setIconForMenuItem( jMI_whatIsNew, IconResources.WHAT_IS_NEW_ICON_RESOURCE );
		setIconForMenuItem( jMI_license, IconResources.LICENSE_ICON_RESOURCE );
		setIconForMenuItem( jMI_about, IconResources.ABOUT_ICON_RESOURCE );
	}

    private void jMI_importFaces(ActionEvent e) {
        // TODO add your code here
		notifyEvt(createMainWindowEvt(MainWindowEventType.IMPORT_FACES));
    }

    private void jMI_exportFaces(ActionEvent e) {
        // TODO add your code here
		notifyEvt(createMainWindowEvt(MainWindowEventType.EXPORT_FACES));
    }

	protected void M_applyNewApplicationConfiguration()
	{
		setLanguageInMenu( getAppliConf().getLanguage() );
	}

	/** This method is called from within the constructor to
	 * initialize the form.
	 * WARNING: Do NOT modify this code. The content of this method is
	 * always regenerated by the Form Editor.
	 */
	@SuppressWarnings("unchecked")
    // <editor-fold defaultstate="collapsed" desc="Generated Code">//GEN-BEGIN:initComponents
    // Generated using JFormDesigner non-commercial license
    private void initComponents() {
        jMenuBar1 = new JMenuBar();
        jM_File = new JMenu();
        jMI_open = new JMenuItem();
        jMI_save = new JMenuItem();
        jMI_saveAs = new JMenuItem();
        jMI_importFaces = new JMenuItem();
        jMI_exportFaces = new JMenuItem();
        jMI_downloadImportedProjectFromv1_0 = new JMenuItem();
        jMI_downloadExampleProject1 = new JMenuItem();
        jMI_downloadExampleProject2 = new JMenuItem();
        jMI_exit = new JMenuItem();
        jM_view = new JMenu();
        jM_zoom = new JMenu();
        jMI_darkMode = new JMenuItem();
        jM_tools = new JMenu();
        jM_languageSubmenu = new JMenu();
        jRadioButtonMenuItemIdiomaEn = new JRadioButtonMenuItem();
        jRadioButtonMenuItemIdiomaEs = new JRadioButtonMenuItem();
        jRadioButtonMenuItemIdiomaOtro = new JRadioButtonMenuItem();
        jMI_applicationConfiguration = new JMenuItem();
        jMI_simulateMorphing = new JMenuItem();
        jMI_cancelMorphingSimulation = new JMenuItem();
        jM_help = new JMenu();
        jM_demos = new JMenu();
        jMI_demos_morphingOpeningProjectOfPreviousVersion = new JMenuItem();
        jMI_demos_morphingNormalTransitionsProject1 = new JMenuItem();
        jMI_demos_morphingProject2 = new JMenuItem();
        jMI_demos_importedMorphingProjectExample = new JMenuItem();
        jMI_demos_example1 = new JMenuItem();
        jMI_demos_example2 = new JMenuItem();
        jMI_userHandbook = new JMenuItem();
        jMI_lookForNewVersion = new JMenuItem();
        jMI_whatIsNew = new JMenuItem();
        jMI_license = new JMenuItem();
        jMI_about = new JMenuItem();
        jPanel1 = new JPanel();
        jSplitPane1 = new JSplitPane();
        jScrollPane1 = new JScrollPane();
        jT_imageModelList = new JTable();
        jTF_morphingProjectFileName = new JTextField();
        jPanel9 = new JPanel();
        jTF_projectName = new JTextField();

        //======== this ========
        setDefaultCloseOperation(WindowConstants.DO_NOTHING_ON_CLOSE);
        setAutoRequestFocus(false);
        setMinimumSize(new Dimension(700, 500));
        setName("MainWindow");
        Container contentPane = getContentPane();
        contentPane.setLayout(null);

        //======== jMenuBar1 ========
        {
            jMenuBar1.setName("jMenuBar1");
            jMenuBar1.setPreferredSize(new Dimension(940, 21));

            //======== jM_File ========
            {
                jM_File.setText("File");
                jM_File.setName("jM_File");

                //---- jMI_open ----
                jMI_open.setText("Open Morphing project");
                jMI_open.setName("jMI_open");
                jM_File.add(jMI_open);

                //---- jMI_save ----
                jMI_save.setText("Save Morphing project");
                jMI_save.setName("jMI_save");
                jM_File.add(jMI_save);

                //---- jMI_saveAs ----
                jMI_saveAs.setText("Save Morphing project as");
                jMI_saveAs.setName("jMI_saveAs");
                jM_File.add(jMI_saveAs);
                jM_File.addSeparator();

                //---- jMI_importFaces ----
                jMI_importFaces.setText("Import face detections");
                jMI_importFaces.setName("jMI_importFaces");
                jMI_importFaces.addActionListener(e -> jMI_importFaces(e));
                jM_File.add(jMI_importFaces);

                //---- jMI_exportFaces ----
                jMI_exportFaces.setText("Export face detections");
                jMI_exportFaces.setName("jMI_exportFaces");
                jMI_exportFaces.addActionListener(e -> jMI_exportFaces(e));
                jM_File.add(jMI_exportFaces);
                jM_File.addSeparator();

                //---- jMI_downloadImportedProjectFromv1_0 ----
                jMI_downloadImportedProjectFromv1_0.setText("Download imported project from v1.0");
                jMI_downloadImportedProjectFromv1_0.setName("jMI_downloadImportedProjectFromv1_0");
                jM_File.add(jMI_downloadImportedProjectFromv1_0);

                //---- jMI_downloadExampleProject1 ----
                jMI_downloadExampleProject1.setText("Download example Morphing project 1");
                jMI_downloadExampleProject1.setName("jMI_downloadExampleProject1");
                jM_File.add(jMI_downloadExampleProject1);

                //---- jMI_downloadExampleProject2 ----
                jMI_downloadExampleProject2.setText("Download example Morphing project 2");
                jMI_downloadExampleProject2.setName("jMI_downloadExampleProject2");
                jM_File.add(jMI_downloadExampleProject2);
                jM_File.addSeparator();

                //---- jMI_exit ----
                jMI_exit.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_Q, KeyEvent.CTRL_DOWN_MASK));
                jMI_exit.setText("Exit");
                jMI_exit.setName("jMI_exit");
                jMI_exit.addActionListener(e -> jMI_exitActionPerformed(e));
                jM_File.add(jMI_exit);
            }
            jMenuBar1.add(jM_File);

            //======== jM_view ========
            {
                jM_view.setText("View");
                jM_view.setName("jM_view");

                //======== jM_zoom ========
                {
                    jM_zoom.setText("Zoom");
                    jM_zoom.setName("jM_zoom");
                }
                jM_view.add(jM_zoom);
                jM_view.addSeparator();

                //---- jMI_darkMode ----
                jMI_darkMode.setText("Dark mode");
                jMI_darkMode.setName("jMI_darkMode");
                jMI_darkMode.addActionListener(e -> jMI_darkModeActionPerformed(e));
                jM_view.add(jMI_darkMode);
            }
            jMenuBar1.add(jM_view);

            //======== jM_tools ========
            {
                jM_tools.setText("Tools");
                jM_tools.setName("jM_tools");

                //======== jM_languageSubmenu ========
                {
                    jM_languageSubmenu.setText("Language");
                    jM_languageSubmenu.setName("jM_languageSubmenu");

                    //---- jRadioButtonMenuItemIdiomaEn ----
                    jRadioButtonMenuItemIdiomaEn.setText("EN");
                    jRadioButtonMenuItemIdiomaEn.setName("");
                    jM_languageSubmenu.add(jRadioButtonMenuItemIdiomaEn);

                    //---- jRadioButtonMenuItemIdiomaEs ----
                    jRadioButtonMenuItemIdiomaEs.setText("ES");
                    jRadioButtonMenuItemIdiomaEs.setName("");
                    jM_languageSubmenu.add(jRadioButtonMenuItemIdiomaEs);

                    //---- jRadioButtonMenuItemIdiomaOtro ----
                    jRadioButtonMenuItemIdiomaOtro.setText("CAT");
                    jRadioButtonMenuItemIdiomaOtro.setName("");
                    jM_languageSubmenu.add(jRadioButtonMenuItemIdiomaOtro);
                }
                jM_tools.add(jM_languageSubmenu);
                jM_tools.addSeparator();

                //---- jMI_applicationConfiguration ----
                jMI_applicationConfiguration.setText("Application Configuration");
                jMI_applicationConfiguration.setName("jMI_applicationConfiguration");
                jMI_applicationConfiguration.addActionListener(e -> jMI_applicationConfigurationActionPerformed(e));
                jM_tools.add(jMI_applicationConfiguration);
                jM_tools.addSeparator();

                //---- jMI_simulateMorphing ----
                jMI_simulateMorphing.setText("Simulate morphing");
                jMI_simulateMorphing.setName("jMI_simulateMorphing");
                jMI_simulateMorphing.addActionListener(e -> jMI_simulateMorphingActionPerformed(e));
                jM_tools.add(jMI_simulateMorphing);

                //---- jMI_cancelMorphingSimulation ----
                jMI_cancelMorphingSimulation.setText("Cancel simluation");
                jMI_cancelMorphingSimulation.setName("jMI_cancelMorphingSimulation");
                jMI_cancelMorphingSimulation.addActionListener(e -> jMI_cancelMorphingSimulationActionPerformed(e));
                jM_tools.add(jMI_cancelMorphingSimulation);
            }
            jMenuBar1.add(jM_tools);

            //======== jM_help ========
            {
                jM_help.setText("Help");
                jM_help.setName("jM_help");

                //======== jM_demos ========
                {
                    jM_demos.setText("Demos");
                    jM_demos.setName("jM_demos");

                    //---- jMI_demos_morphingOpeningProjectOfPreviousVersion ----
                    jMI_demos_morphingOpeningProjectOfPreviousVersion.setText("Opening a project created with version v1.0");
                    jMI_demos_morphingOpeningProjectOfPreviousVersion.setName("jMI_demos_morphingOpeningProjectOfPreviousVersion");
                    jM_demos.add(jMI_demos_morphingOpeningProjectOfPreviousVersion);
                    jM_demos.addSeparator();

                    //---- jMI_demos_morphingNormalTransitionsProject1 ----
                    jMI_demos_morphingNormalTransitionsProject1.setText("Creation of project 1 from scratch with normal transitions");
                    jMI_demos_morphingNormalTransitionsProject1.setName("jMI_demos_morphingFaceDetectionProject1");
                    jM_demos.add(jMI_demos_morphingNormalTransitionsProject1);

                    //---- jMI_demos_morphingProject2 ----
                    jMI_demos_morphingProject2.setText("Creation of project 2 from scratch with FACE_DETECTION transitions");
                    jMI_demos_morphingProject2.setName("jMI_demos_morphingProject2");
                    jM_demos.add(jMI_demos_morphingProject2);
                    jM_demos.addSeparator();

                    //---- jMI_demos_importedMorphingProjectExample ----
                    jMI_demos_importedMorphingProjectExample.setText("Imported morphing project video example");
                    jMI_demos_importedMorphingProjectExample.setName("jMI_demos_importedMorphingProjectExample");
                    jM_demos.add(jMI_demos_importedMorphingProjectExample);

                    //---- jMI_demos_example1 ----
                    jMI_demos_example1.setText("Morphing video example 1");
                    jMI_demos_example1.setName("jMI_demos_example1");
                    jM_demos.add(jMI_demos_example1);

                    //---- jMI_demos_example2 ----
                    jMI_demos_example2.setText("Morphing video example 2");
                    jMI_demos_example2.setName("jMI_demos_example2");
                    jM_demos.add(jMI_demos_example2);
                }
                jM_help.add(jM_demos);
                jM_help.addSeparator();

                //---- jMI_userHandbook ----
                jMI_userHandbook.setText("User handbook");
                jMI_userHandbook.setName("jMI_userHandbook");
                jMI_userHandbook.addActionListener(e -> jMI_userHandbookActionPerformed(e));
                jM_help.add(jMI_userHandbook);
                jM_help.addSeparator();

                //---- jMI_lookForNewVersion ----
                jMI_lookForNewVersion.setText("Look for a new version");
                jMI_lookForNewVersion.setName("jMI_lookForNewVersion");
                jMI_lookForNewVersion.addActionListener(e -> jMI_lookForNewVersionActionPerformed(e));
                jM_help.add(jMI_lookForNewVersion);
                jM_help.addSeparator();

                //---- jMI_whatIsNew ----
                jMI_whatIsNew.setText("What is new");
                jMI_whatIsNew.setName("jMI_whatIsNew");
                jMI_whatIsNew.addActionListener(e -> jMI_whatIsNewActionPerformed(e));
                jM_help.add(jMI_whatIsNew);
                jM_help.addSeparator();

                //---- jMI_license ----
                jMI_license.setText("License");
                jMI_license.setName("jMI_license");
                jMI_license.addActionListener(e -> jMI_licenseActionPerformed(e));
                jM_help.add(jMI_license);
                jM_help.addSeparator();

                //---- jMI_about ----
                jMI_about.setText("About");
                jMI_about.setName("jMI_about");
                jMI_about.addActionListener(e -> jMI_aboutActionPerformed(e));
                jM_help.add(jMI_about);
            }
            jMenuBar1.add(jM_help);
        }
        setJMenuBar(jMenuBar1);

        //======== jPanel1 ========
        {
            jPanel1.setName("jPanel1");
            jPanel1.setLayout(null);

            //======== jSplitPane1 ========
            {
                jSplitPane1.setDividerLocation(150);
                jSplitPane1.setOrientation(JSplitPane.VERTICAL_SPLIT);

                //======== jScrollPane1 ========
                {

                    //---- jT_imageModelList ----
                    jT_imageModelList.setModel(new DefaultTableModel(
                        new Object[][] {
                            {null, null, null, null},
                            {null, null, null, null},
                            {null, null, null, null},
                            {null, null, null, null},
                        },
                        new String[] {
                            "Title 1", "Title 2", "Title 3", "Title 4"
                        }
                    ));
                    jT_imageModelList.setAutoResizeMode(JTable.AUTO_RESIZE_OFF);
                    jScrollPane1.setViewportView(jT_imageModelList);
                }
                jSplitPane1.add(jScrollPane1, "top");
            }
            jPanel1.add(jSplitPane1);
            jSplitPane1.setBounds(0, 35, 755, 405);

            //---- jTF_morphingProjectFileName ----
            jTF_morphingProjectFileName.setFont(new Font(Font.DIALOG, Font.BOLD, 12));
            jPanel1.add(jTF_morphingProjectFileName);
            jTF_morphingProjectFileName.setBounds(230, 5, 515, jTF_morphingProjectFileName.getPreferredSize().height);

            //======== jPanel9 ========
            {
                jPanel9.setBorder(LineBorder.createBlackLineBorder());
                jPanel9.setMinimumSize(new Dimension(80, 30));
                jPanel9.setLayout(null);

                {
                    // compute preferred size
                    Dimension preferredSize = new Dimension();
                    for(int i = 0; i < jPanel9.getComponentCount(); i++) {
                        Rectangle bounds = jPanel9.getComponent(i).getBounds();
                        preferredSize.width = Math.max(bounds.x + bounds.width, preferredSize.width);
                        preferredSize.height = Math.max(bounds.y + bounds.height, preferredSize.height);
                    }
                    Insets insets = jPanel9.getInsets();
                    preferredSize.width += insets.right;
                    preferredSize.height += insets.bottom;
                    jPanel9.setMinimumSize(preferredSize);
                    jPanel9.setPreferredSize(preferredSize);
                }
            }
            jPanel1.add(jPanel9);
            jPanel9.setBounds(145, 5, 80, 30);

            //---- jTF_projectName ----
            jTF_projectName.setText("Project Name");
            jPanel1.add(jTF_projectName);
            jTF_projectName.setBounds(10, 5, 120, jTF_projectName.getPreferredSize().height);

            {
                // compute preferred size
                Dimension preferredSize = new Dimension();
                for(int i = 0; i < jPanel1.getComponentCount(); i++) {
                    Rectangle bounds = jPanel1.getComponent(i).getBounds();
                    preferredSize.width = Math.max(bounds.x + bounds.width, preferredSize.width);
                    preferredSize.height = Math.max(bounds.y + bounds.height, preferredSize.height);
                }
                Insets insets = jPanel1.getInsets();
                preferredSize.width += insets.right;
                preferredSize.height += insets.bottom;
                jPanel1.setMinimumSize(preferredSize);
                jPanel1.setPreferredSize(preferredSize);
            }
        }
        contentPane.add(jPanel1);
        jPanel1.setBounds(0, 0, 755, 440);

        {
            // compute preferred size
            Dimension preferredSize = new Dimension();
            for(int i = 0; i < contentPane.getComponentCount(); i++) {
                Rectangle bounds = contentPane.getComponent(i).getBounds();
                preferredSize.width = Math.max(bounds.x + bounds.width, preferredSize.width);
                preferredSize.height = Math.max(bounds.y + bounds.height, preferredSize.height);
            }
            Insets insets = contentPane.getInsets();
            preferredSize.width += insets.right;
            preferredSize.height += insets.bottom;
            contentPane.setMinimumSize(preferredSize);
            contentPane.setPreferredSize(preferredSize);
        }
        pack();
        setLocationRelativeTo(getOwner());
    }// </editor-fold>//GEN-END:initComponents

    private void jMI_exitActionPerformed(java.awt.event.ActionEvent evt)//GEN-FIRST:event_jMI_exitActionPerformed
    {//GEN-HEADEREND:event_jMI_exitActionPerformed
	  // TODO add your handling code here:
		boolean closeWindow = true;
		formWindowClosing( closeWindow );
    }//GEN-LAST:event_jMI_exitActionPerformed



    private void jMI_applicationConfigurationActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_jMI_applicationConfigurationActionPerformed
        // TODO add your handling code here:

        JDial_applicationConfiguration dialog = new JDial_applicationConfiguration( this, true, this::showConfigurationWindow );
    }//GEN-LAST:event_jMI_applicationConfigurationActionPerformed

    private void jMI_lookForNewVersionActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_jMI_lookForNewVersionActionPerformed
        // TODO add your handling code here:

        checkForNewVersion();
    }//GEN-LAST:event_jMI_lookForNewVersionActionPerformed

    private void jMI_whatIsNewActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_jMI_whatIsNewActionPerformed
        // TODO add your handling code here:

        showWhatIsNew();
    }//GEN-LAST:event_jMI_whatIsNewActionPerformed

    private void jMI_licenseActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_jMI_licenseActionPerformed
        // TODO add your handling code here:

        showLicense();
    }//GEN-LAST:event_jMI_licenseActionPerformed

    private void jMI_aboutActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_jMI_aboutActionPerformed
        // TODO add your handling code here:

		_aboutDial = new AboutJDialog( this, true, (iec)->setVisibleAbout(),
			getApplicationContext().getAnimationForAboutFactory() );
    }//GEN-LAST:event_jMI_aboutActionPerformed

    private void jMI_darkModeActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_jMI_darkModeActionPerformed
        // TODO add your handling code here:

        toggleDarkMode();
    }//GEN-LAST:event_jMI_darkModeActionPerformed

    private void jMI_simulateMorphingActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_jMI_simulateMorphingActionPerformed
        // TODO add your handling code here:

		showMorphing();
    }//GEN-LAST:event_jMI_simulateMorphingActionPerformed

    private void jMI_cancelMorphingSimulationActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_jMI_cancelMorphingSimulationActionPerformed
        // TODO add your handling code here:

		_cancelMorphingSimulation.set(true);
    }//GEN-LAST:event_jMI_cancelMorphingSimulationActionPerformed

    private void jMI_userHandbookActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_jMI_userHandbookActionPerformed
        // TODO add your handling code here:

        String fileName = getInternationalString(CONF_USER_HANDBOOK_FILE_NAME);
        try
        {
            GenericFunctions.instance().getSystem().openDocument( fileName );
        }
        catch( Exception ex )
        {
            GenericFunctions.instance().getDialogsWrapper().showMessageDialog( this,
                createCustomInternationalString( CONF_ERROR_OPENING_FILE, fileName ),
                getInternationalString( CONF_ERROR_OPENING_HELP ),
                DialogsWrapper.ERROR_MESSAGE );
        }

    }//GEN-LAST:event_jMI_userHandbookActionPerformed

    // Variables declaration - do not modify//GEN-BEGIN:variables
    // Generated using JFormDesigner non-commercial license
    private JMenuBar jMenuBar1;
    private JMenu jM_File;
    private JMenuItem jMI_open;
    private JMenuItem jMI_save;
    private JMenuItem jMI_saveAs;
    private JMenuItem jMI_importFaces;
    private JMenuItem jMI_exportFaces;
    private JMenuItem jMI_downloadImportedProjectFromv1_0;
    private JMenuItem jMI_downloadExampleProject1;
    private JMenuItem jMI_downloadExampleProject2;
    private JMenuItem jMI_exit;
    private JMenu jM_view;
    private JMenu jM_zoom;
    private JMenuItem jMI_darkMode;
    private JMenu jM_tools;
    private JMenu jM_languageSubmenu;
    private JRadioButtonMenuItem jRadioButtonMenuItemIdiomaEn;
    private JRadioButtonMenuItem jRadioButtonMenuItemIdiomaEs;
    private JRadioButtonMenuItem jRadioButtonMenuItemIdiomaOtro;
    private JMenuItem jMI_applicationConfiguration;
    private JMenuItem jMI_simulateMorphing;
    private JMenuItem jMI_cancelMorphingSimulation;
    private JMenu jM_help;
    private JMenu jM_demos;
    private JMenuItem jMI_demos_morphingOpeningProjectOfPreviousVersion;
    private JMenuItem jMI_demos_morphingNormalTransitionsProject1;
    private JMenuItem jMI_demos_morphingProject2;
    private JMenuItem jMI_demos_importedMorphingProjectExample;
    private JMenuItem jMI_demos_example1;
    private JMenuItem jMI_demos_example2;
    private JMenuItem jMI_userHandbook;
    private JMenuItem jMI_lookForNewVersion;
    private JMenuItem jMI_whatIsNew;
    private JMenuItem jMI_license;
    private JMenuItem jMI_about;
    private JPanel jPanel1;
    private JSplitPane jSplitPane1;
    private JScrollPane jScrollPane1;
    private JTable jT_imageModelList;
    private JTextField jTF_morphingProjectFileName;
    private JPanel jPanel9;
    private JTextField jTF_projectName;
    // End of variables declaration//GEN-END:variables

	protected LicenseJDialog createLicenseJDialog( Consumer<InternationalizationInitializationEndCallback> initializationEndCallBack )
	{
		return( new LicenseJDialog( this, getAppliConf(), initializationEndCallBack, false ) );
	}

	protected void showLicense()
	{
        createLicenseJDialog( this::showLicenseCallback );
	}

	protected void showLicenseCallback(InternationalizationInitializationEndCallback iiec)
	{
        LicenseJDialog ljd = (LicenseJDialog) iiec;
        ljd.setVisible(true);
        boolean closeWindow = true;
        ljd.formWindowClosing(closeWindow);
	}

	protected void showConfigurationWindow( InternationalizationInitializationEndCallback window )
	{
		JDial_applicationConfiguration dialog = (JDial_applicationConfiguration) window;

		dialog.setVisibleWithLock(true);

        if( dialog.M_getUserHasPressedOK() )
        {
            M_applyNewApplicationConfiguration();
        }
        dialog.M_releaseResources();
        dialog.dispose();
        dialog = null;
	}

	protected void setVisibleAbout()
	{
		SwingUtilities.invokeLater( () -> _aboutDial.setVisibleWithLock(true) );
	}

	protected void setConfigurationChanges()
	{
		getAppliConf().setTableColumnsString( getTableWrapper().getTableColumnDataString() );
	}

	@Override
	public void formWindowClosing( boolean closeWindow )
	{
		try
		{
			setConfigurationChanges();

			safeMethodExecution(() -> getApplicationContext().saveFaceDetections());

			try
			{
				ApplicationConfiguration.instance().M_saveConfiguration();
			}
			catch( ConfigurationException ce )
			{
				ce.printStackTrace();
			}
/*
			try
			{
//				StringsConfiguration.instance().M_saveConfiguration();
			}
			catch( Exception ce )
			{
				ce.printStackTrace();
			}

			if( a_intern != null )
			{
				try
				{
					a_intern.saveConfiguration();
				}
				catch( InternException ex )
				{
					ex.printStackTrace();
				}
			}
*/
			if( closeWindow )
			{
				removeListeners();

//				getAppliConf().removeConfigurationParameterListener( ApplicationConfiguration.CONF_PRECISION_FOR_CALCULATIONS, this );
			}

			super.formWindowClosing(closeWindow);
		}
		catch( Throwable th )
		{
			th.printStackTrace();
			System.exit(1);
		}

		if( closeWindow )
			System.exit(0);

//		System.exit(0);
	}

	protected void removeListeners()
	{
//		ComponentFunctions.instance().browseComponentHierarchy( this, (comp) -> { comp.removeFocusListener(_exitFocusListener); return( null ); } );
		jTF_projectName.getDocument().removeDocumentListener( _projectNameDocumentListener );
	}

	protected void changeZoomFactor( JScrollBar sb, double relativeZoom, Integer pos )
	{
		if( ( sb != null ) && ( pos != null ) );
		{
			pos = IntegerFunctions.zoomValueInt( pos, relativeZoom );

			sb.setValue(pos);
		}
	}

	protected void exit()
	{
		jMI_exit.doClick();
	}

	@Override
	protected void onIconify()
	{
		super.onIconify();
	}

	protected void addProperty( Properties prop, String label, String value )
	{
		prop.setProperty( label, value);
	}

	@Override
	public void changeLanguage( String language ) throws ConfigurationException, InternException
	{
		super.changeLanguage( language );
	}

	@Override
	public void changeZoomFactor( double zoomFactor )
	{
		super.changeZoomFactor( zoomFactor );
	}

	@Override
	public <CC> void configurationParameterChanged(ConfigurationParameterObserved observed, String label, CC oldValue, CC newValue)
	{

	}

	@Override
	public ApplicationConfiguration getAppliConf()
	{
		return( (ApplicationConfiguration) _appliConf );
	}

	@Override
	protected void translateMappedComponents(ComponentMapper compMapper)
	{
		jMI_about = compMapper.mapComponent( jMI_about );
		jMI_applicationConfiguration = compMapper.mapComponent( jMI_applicationConfiguration );
		jMI_darkMode = compMapper.mapComponent( jMI_darkMode );
		jMI_exit = compMapper.mapComponent( jMI_exit );
		jMI_license = compMapper.mapComponent( jMI_license );
		jMI_lookForNewVersion = compMapper.mapComponent( jMI_lookForNewVersion );
		jMI_open = compMapper.mapComponent( jMI_open );
		jMI_save = compMapper.mapComponent( jMI_save );
		jMI_saveAs = compMapper.mapComponent( jMI_saveAs );
		jMI_whatIsNew = compMapper.mapComponent( jMI_whatIsNew );
		jM_File = compMapper.mapComponent( jM_File );
		jM_help = compMapper.mapComponent( jM_help );
		jM_languageSubmenu = compMapper.mapComponent( jM_languageSubmenu );
		jM_tools = compMapper.mapComponent( jM_tools );
		jM_view = compMapper.mapComponent( jM_view );
		jM_zoom = compMapper.mapComponent( jM_zoom );
		jMenuBar1 = compMapper.mapComponent( jMenuBar1 );
		jPanel1 = compMapper.mapComponent( jPanel1 );
		jPanel9 = compMapper.mapComponent( jPanel9 );
		jRadioButtonMenuItemIdiomaEn = compMapper.mapComponent( jRadioButtonMenuItemIdiomaEn );
		jRadioButtonMenuItemIdiomaEs = compMapper.mapComponent( jRadioButtonMenuItemIdiomaEs );
		jRadioButtonMenuItemIdiomaOtro = compMapper.mapComponent( jRadioButtonMenuItemIdiomaOtro );
		jScrollPane1 = compMapper.mapComponent( jScrollPane1 );
//		jSeparator1 = compMapper.mapComponent( jSeparator1 );
//		jSeparator11 = compMapper.mapComponent( jSeparator11 );
//		jSeparator12 = compMapper.mapComponent( jSeparator12 );
//		jSeparator2 = compMapper.mapComponent( jSeparator2 );
//		jSeparator25 = compMapper.mapComponent( jSeparator25 );
//		jSeparator5 = compMapper.mapComponent( jSeparator5 );
		jSplitPane1 = compMapper.mapComponent( jSplitPane1 );
		jTF_morphingProjectFileName = compMapper.mapComponent( jTF_morphingProjectFileName );
		jTF_projectName = compMapper.mapComponent( jTF_projectName );
		jT_imageModelList = compMapper.mapComponent( jT_imageModelList );

		jM_demos = compMapper.mapComponent( jM_demos );
		jMI_demos_morphingOpeningProjectOfPreviousVersion = compMapper.mapComponent(jMI_demos_morphingOpeningProjectOfPreviousVersion);
		jMI_demos_morphingNormalTransitionsProject1 = compMapper.mapComponent(jMI_demos_morphingNormalTransitionsProject1);
		jMI_demos_morphingProject2 = compMapper.mapComponent(jMI_demos_morphingProject2);
		jMI_demos_importedMorphingProjectExample = compMapper.mapComponent(jMI_demos_importedMorphingProjectExample);
		jMI_demos_example1 = compMapper.mapComponent(jMI_demos_example1);
		jMI_demos_example2 = compMapper.mapComponent(jMI_demos_example2);
		jMI_downloadImportedProjectFromv1_0 = compMapper.mapComponent(jMI_downloadImportedProjectFromv1_0);
		jMI_downloadExampleProject1 = compMapper.mapComponent(jMI_downloadExampleProject1);
		jMI_downloadExampleProject2 = compMapper.mapComponent(jMI_downloadExampleProject2);

//		_delaunayPanel = compMapper.mapComponent( _delaunayPanel );
	}

	@Override
	public void setDarkMode(boolean value, ColorInversor colorInversor)
	{
		super.setDarkMode( value, colorInversor );
	}

	public void addImageListTableListener(Consumer<? super GenericWrapperEvent> listener,
										Long ... eventTypes) {
		getTableWrapper().addWrapperListener(listener, eventTypes);
	}

	public void removeImageListTableListener(Consumer<? super GenericWrapperEvent> listener) {
		getTableWrapper().removeWrapperListener(listener);
	}

	public MorphingImageModelListTableWrapper getTableWrapper() {
		return _tableWrapper;
	}

	protected MorphingImageModelListTableWrapper createTableWrapper()
	{
		MorphingImageModelListTableWrapper result = new MorphingImageModelListTableWrapper( jT_imageModelList );
/*
		result.addWrapperListener( this::processTableWrapperEvent, JTableWrapperEventType.ALL );
		result.addWrapperListener(
			evt -> {
				if( evt instanceof JTableWrapperDropFilesEvent)
				{
					JTableWrapperDropFilesEvent<PlayListEntryBase, JTable> dadEvt = (JTableWrapperDropFilesEvent) evt;
					getController().droppedFiles( dadEvt.getEventData().getFileList(),
												dadEvt.getEventData().getIndexForInsertion() );
				}
			}, JTableWrapperEventType.FILES_DROPPED );
*/
		return( result );
	}

	public void updateTriangulation(String imageName)//,
//								MorphingVerticesMapManager morphingVerticesManager)
	{
		// intentionally left blank
//		_morphingVerticesManager = morphingVerticesManager;
	}

	public void updateCurrentTableList( ListStateContextBase<MorphingImageModel> modifiedListState )
	{
		getTableWrapper().updateCurrentList( modifiedListState );
	}

	protected void showImage( BufferedImage image )
	{
//		safeMethodExecution( () -> ImageIO.write(image, "png", new File( "J:\\image.png" ) ) );
		LOGGER.info( "New image: {}", DateFunctions.instance().formatDurationLong( System.currentTimeMillis() ) );
		_imageJPanel.setImageCenteringIt(image);
	}

//	protected void showImage( BufferedImage image )
//	{
//		SwingUtilities.invokeLater( () -> showImageInternal( image ) );
//	}

	protected ImageJPanelSizeParams getSizeParams()
	{
		return( _imageJPanelSizeReference.get() );
	}

	public GeometricRectangle createRectangle(double x, double y, double width, double height )
	{
		return( GeometryCommonFactory.instance().createDelaunayRectangle( x, y, width, height ) );
	}

	public GeometricRectangle getFittingImageSizeForJPanel()
	{
		ImageJPanelSizeParams sizeParams = getSizeParams();
//		Dimension dimen = _imageJPanel.getSizeToFit( sizeParams.getSize(), sizeParams.getZoomedWidth() );
		Dimension dimen = _imageJPanel.getSizeToFit( sizeParams.getSize(), 0 );
		return( createRectangle( 0, 0, dimen.width, dimen.height ) );
	}

	protected void showMorphing()
	{
		new Thread( this::showMorphingInternal ).start();
	}

	protected void showMorphingInternal()
	{
//		MorphingFrameSequencer seq = new MorphingFrameSequencer( _morphingVerticesManager );
		MorphingFrameSequencer seq = new MorphingFrameSequencer( getApplicationContext() );

		_cancelMorphingSimulation.set(true);
		_cancelMorphingSimulation = new AtomicBoolean(false);
		AtomicReference<Exception> exceptionReference = new AtomicReference<>();

		SwingImageSequenceShower imageShower = new SwingImageSequenceShower( this::showImage, 250 );
		_imageShower = imageShower;

		int delayBetweenFrames = 100;
		seq.doMorphing(imageShower::addImage,
				ex -> processExceptionProducedDuringMorphing(_cancelMorphingSimulation, exceptionReference, ex),
				_cancelMorphingSimulation,
				imageShower::hasToStop,
				this::getFittingImageSizeForJPanel,
				delayBetweenFrames);
	}

	protected synchronized void processExceptionProducedDuringMorphing(AtomicBoolean cancelMorphingSimulation,
																	   AtomicReference<Exception> exceptionReference,
																	   Exception ex) {
		if(exceptionReference.get() == null) {
			exceptionReference.set(ex);
			SwingUtilities.invokeLater(() -> processExceptionProducedDuringMorphingInternal(cancelMorphingSimulation, ex));
		}
	}

	protected void processExceptionProducedDuringMorphingInternal(AtomicBoolean cancelMorphingSimulation,
																  Exception ex) {
		int result = GenericFunctions.instance().getDialogsWrapper().showOptionDialog(this,
				createCustomInternationalString(CONF_ERROR_CREATING_MORPHING_FRAME, ex.getMessage()),
				"ERROR", DialogsWrapper.OK_CANCEL_OPTION, DialogsWrapper.QUESTION_MESSAGE,
				0,
				getInternationalString(CONF_IGNORE),
				getInternationalString(CONF_CANCEL));

		if(result == 1) {
			cancelMorphingSimulation.set(true);
		}
	}

	public boolean confirmErasureOfSelectedRecords(
		GenTableWrapperRecordsModifiedEventBase<MorphingImageModel,
                                                    ?,
                                                    ListStateContextBase<MorphingImageModel>> evt )
	{
		return( getTableWrapper().confirmErasureOfSelectedRecords(evt) );
	}

	protected void notifyEvt(MainWindowEvent evt) {
		Consumer<MainWindowEvent> listener = getMainWindowListener();
		if(listener != null) {
			listener.accept(evt);
		}
	}

	protected MainWindowEvent createMainWindowEvt(MainWindowEventType type) {
		return new MainWindowEvent(type);
	}
}
