/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.applications.morphing.controller;

import com.frojasg1.applications.libmorphing.api.facedetector.FaceDetectionQueryFactory;
import com.frojasg1.applications.libmorphing.contexts.MorphingEditorApplicationContext;
import com.frojasg1.applications.libmorphing.api.facedetector.helpers.FaceDetectionMeshDataBuilderAndUpdater;
import com.frojasg1.applications.libmorphing.api.facedetector.pojo.FaceDetection;
import com.frojasg1.applications.libmorphing.api.facedetector.pojo.FaceDetectionMeshData;
import com.frojasg1.applications.libmorphing.api.facedetector.pojo.ImageFaceDetection;
import com.frojasg1.applications.libmorphing.api.facedetector.runnable.FaceDetectionQueryExecution;
import com.frojasg1.applications.libmorphing.configuration.MorphingApplicationConfiguration;
import com.frojasg1.applications.libmorphing.controller.MorphingVerticesMapManager;
import com.frojasg1.applications.libmorphing.view.filefilters.GenericFileFilterChooserGeneralMorphing;
import com.frojasg1.applications.common.configuration.imp.InternationalizedStringConfImp;
import com.frojasg1.applications.libmorphing.model.images.color.types.MultiTriangle;
import com.frojasg1.applications.libmorphing.model.images.configuration.MorphingProjectConfigurationModel;
import com.frojasg1.applications.libmorphing.model.images.image.MorphingImageModel;
import com.frojasg1.applications.libmorphing.view.mainwindow.evt.MainWindowEvent;
import com.frojasg1.applications.libmorphing.view.mainwindow.evt.enums.MainWindowEventType;
import com.frojasg1.applications.libmorphing.view.painters.TriangulationPainterLayerTypeEnum;
import com.frojasg1.applications.libmorphing.view.panels.evt.*;
import com.frojasg1.applications.libmorphing.view.panels.evt.enums.FaceDetectionSelectionType;
import com.frojasg1.applications.libmorphing.view.panels.evt.enums.RecalculateTriangulationType;
import com.frojasg1.applications.libmorphing.view.panels.evt.enums.VertexEventType;
import com.frojasg1.applications.libmorphing.xml.facedetector.FaceDetectionWholeFileModel;
import com.frojasg1.applications.libmorphing.xml.whole.MorphingProjectWholeFileModel;
import com.frojasg1.applications.morphing.imagelisttable.evt.JTableWrapperRecordsClonedEvent;
import com.frojasg1.applications.morphing.language.MorphingApplicationInternationalTexts;
import com.frojasg1.applications.morphing.mvc.MorphingController;
import com.frojasg1.applications.morphing.mvc.MorphingView;
import com.frojasg1.libgeometry.types.GeometricRectangle;
import com.frojasg1.libgeometry.types.Point2d;
import com.frojasg1.general.CollectionFunctions;
import com.frojasg1.general.ExecutionFunctions;
import com.frojasg1.general.FileFunctions;
import com.frojasg1.general.NullFunctions;
import com.frojasg1.general.collection.functions.ListStateContextBase;
import com.frojasg1.general.desktop.generic.dialogs.impl.DesktopFileChooserParameters;
import com.frojasg1.general.desktop.view.panels.impl.file.FileHandlingPanelEvent;
import com.frojasg1.general.dialogs.FileChooserParameters;
import com.frojasg1.general.view.evt.wrapper.impl.table.impl.GenTableWrapperRecordsModifiedEventBase;
import com.frojasg1.general.dialogs.DialogsWrapper;
import com.frojasg1.general.dialogs.filefilter.GenericFileFilter;
import com.frojasg1.general.dialogs.filefilter.GenericFileFilterChooser;
import com.frojasg1.general.number.DoubleFunctions;
import com.frojasg1.general.string.StringFunctions;
import com.frojasg1.general.view.ViewComponent;
import com.frojasg1.general.view.evt.wrapper.GenericWrapperEvent;
import com.frojasg1.general.view.evt.wrapper.impl.GenComponentWrapperEventType;
import com.frojasg1.general.view.evt.wrapper.impl.table.impl.GenTableWrapperDropFilesEvent;
import com.frojasg1.general.view.evt.wrapper.impl.table.impl.GenTableWrapperNewCurrentEvent;
import com.frojasg1.general.view.evt.wrapper.impl.table.impl.deleterecords.GenTableWrapperRecordsDeletedEvent;
import com.frojasg1.general.view.evt.wrapper.impl.table.impl.moverecords.GenTableWrapperRecordsMovedEvent;
import com.frojasg1.general.view.evt.wrapper.impl.table.impl.sortedrecords.GenTableWrapperRecordsSortedEvent;
import com.frojasg1.generic.GenericFunctions;
import java.io.File;
import java.io.IOException;
import java.util.List;
import java.util.Objects;
import java.util.function.Function;
import javax.swing.*;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.TransformerException;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.xml.sax.SAXException;

/**
 *
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 */
public class MorphingControllerImpl implements MorphingController
{
	private static final Logger LOGGER = LoggerFactory.getLogger(MorphingControllerImpl.class);

	public static final String CONF_ERROR_WHEN_LOADING_MORPHING_PROJECT = "ERROR_WHEN_LOADING_MORPHING_PROJECT";
	public static final String CONF_ERROR_WHEN_SAVING_MORPHING_PROJECT = "ERROR_WHEN_SAVING_MORPHING_PROJECT";
	public static final String CONF_SUCCESS_EXPORTING_FILE = "SUCCESS_EXPORTING_FILE";
	public static final String CONF_SUCCESS_IMPORTING_FILE = "SUCCESS_IMPORTING_FILE";
	public static final String CONF_ERROR_EXPORTING_FILE = "ERROR_EXPORTING_FILE";
	public static final String CONF_ERROR_IMPORTING_FILE = "ERROR_IMPORTING_FILE";

	public static final String CONF_ERROR_WHEN_ERASING_IMAGES = "ERROR_WHEN_ERASING_IMAGES";
	public static final String CONF_DEFAULT_PROJECT_NAME = "DEFAULT_PROJECT_NAME";
	public static final String CONF_PROJECT_SAVED_SUCCESSFULLY = "PROJECT_SAVED_SUCCESSFULLY";
//	public static final String CONF_ARE_YOU_SURE_YOU_WANT_TO_ERASE_THE_SELECTED_ROWS = "ARE_YOU_SURE_YOU_WANT_TO_ERASE_THE_SELECTED_ROWS";
	public static final String CONF_CONFIRM_RECALCULATION_OF_ALL_MESHES = "CONFIRM_RECALCULATION_OF_ALL_MESHES";

	protected MorphingView _view;

//	protected MorphingApplicationConfiguration _appliConf;
//
//	protected MorphingProjectWholeFileModelManager _projectModelManager;

	protected InternationalizedStringConfImp _internationalizedStringConf = null;

	protected MorphingEditorApplicationContext appliCtx;

	public MorphingControllerImpl(MorphingEditorApplicationContext appliCtx)
	{
		this( appliCtx, MorphingApplicationInternationalTexts.instance() );
	}

//	public MorphingControllerImpl(MorphingEditorApplicationContext appliCtx,
//								  InternationalizedStringConfImp internationalizedStringConf)
//	{
//		this( appliCtx, internationalizedStringConf, true );
//	}

	protected MorphingControllerImpl( MorphingEditorApplicationContext appliCtx,
									InternationalizedStringConfImp internationalizedStringConf)
//			,
//									boolean init )
	{
//		_appliConf = appliConf;
		this.appliCtx = appliCtx;
		_internationalizedStringConf = internationalizedStringConf;
//
//		if( init )
//			init();
	}
//
//	protected void init()
//	{
//		_projectModelManager = createMorphingProjectWholeFileModelManager();
//	}
//
//	public MorphingProjectWholeFileModelManager getProjectModelManager() {
//		return getAppliCtx().getProjectModelManager();
//	}

//	public void setProjectModelManager(MorphingProjectWholeFileModelManager _projectModelManager) {
//		this._projectModelManager = _projectModelManager;
//	}
//
//	protected MorphingProjectWholeFileModelManager createMorphingProjectWholeFileModelManagerEmpty()
//	{
//		return( new MorphingProjectWholeFileModelManager(ApplicationConfiguration.instance()) );
//	}
//
//	protected MorphingProjectWholeFileModelManager createMorphingProjectWholeFileModelManager()
//	{
//		MorphingProjectWholeFileModelManager result = createMorphingProjectWholeFileModelManagerEmpty();
////		MorphingProjectWholeFileModelBase model = safeFunctionExecution( this::loadOrGetMorphingProjectModel );
////		if( model != null )
////			result.setMorphingProjectModel( model );
//
//		return( result );
//	}

//	protected MorphingProjectWholeFileModelBase loadMorphingProjectModel(String fileName ) throws SAXException, IOException, ParserConfigurationException
//	{
//		MorphingProjectWholeFileModelBase projectModel = getProjectModelManager().loadMorphingProjectModel(fileName);
//
//		return( projectModel );
//	}

//	protected MorphingProjectWholeFileModelBase loadOrGetMorphingProjectModel() throws SAXException, IOException, ParserConfigurationException
//	{
//		return( loadOrGetMorphingProjectModel( getAppliConf().getLatestProjectXmlFileName() ) );
//	}

//	protected MorphingProjectWholeFileModelBase loadOrGetMorphingProjectModel(String fileName)
//		throws SAXException, IOException, ParserConfigurationException
//	{
//		MorphingProjectWholeFileModelBase result = null;
//		if( !StringFunctions.instance().isEmpty( fileName ) )
//			result = loadMorphingProjectModel( fileName );
//		else if( getProjectModelManager() != null )
//			result = getProjectModelManager().getMorphingProjectModel();
//
//		return( result );
//	}

	public MorphingEditorApplicationContext getAppliCtx() {
		return appliCtx;
	}

	public MorphingApplicationConfiguration getAppliConf() {
		return getAppliCtx().getAppliConf();
	}

	@Override
	public void setView(MorphingView view) {
		removeViewListeners();

		_view = view;

		if( _view != null )
		{
			_view.addDelaunayTriangulatorViewListener( this::processDelaunayTriangulatorViewEvent, GenComponentWrapperEventType.ALL );
			_view.addImageListTableListener( this::processImageListTableEvent, GenComponentWrapperEventType.ALL );
			_view.addFileHandlingListener( this::processFileHandlingPanelEvent );
			_view.setMainWindowListener(this::processMainWindowEvent);
		}
	}

	protected void processMainWindowEvent(MainWindowEvent evt)
	{
		MainWindowEventType type = evt.getType();
		if(type == MainWindowEventType.IMPORT_FACES) {
			importFaceDetections();
		} else if(type == MainWindowEventType.EXPORT_FACES) {
			exportFaceDetections();
		} else {
			LOGGER.warn("Main window event not recognized");
		}
	}


	protected void processFileHandlingPanelEvent( FileHandlingPanelEvent evt )
	{
		String fileName = null;
		switch( evt.getType() )
		{
			case OPEN:
				fileName = loadProject( getViewFileName() );
			break;

			case SAVE:
				fileName = saveProject( getViewFileName() );
			break;

			case SAVE_AS:
				fileName = saveProjectAs();
			break;

		}
		if( fileName != null )
//			updateModelAndView(getProjectModelManager().getMorphingProjectModel());
			updateModelAndView();
	}

	protected String saveProjectAs()
	{
		return( saveProjectAs( getViewFileName() ) );
	}

	protected String getViewFileName()
	{
		return( getView().getFileName() );
	}

	protected GenericFileFilterChooser getGenericFileFilterChooser()
	{
		return( GenericFileFilterChooserGeneralMorphing.instance() );
	}

	protected GenericFileFilter getGenericFileFilter( int index )
	{
		return( getGenericFileFilterChooser().getGenericFileFilter( index ) );
	}

	protected GenericFileFilter getMorphingProjectGenericFileFilter()
	{
		return( getGenericFileFilter( GenericFileFilterChooserGeneralMorphing.MORPHING_PROJECT_FILE_FILTER ) );
	}

	protected List<GenericFileFilter> createMorphingProjectGenericFileFilter()
	{
		GenericFileFilter fnef1 = getMorphingProjectGenericFileFilter();
		List<GenericFileFilter> result = CollectionFunctions.instance().createList( fnef1 );

		return( result );
	}

	protected List<GenericFileFilter> createXmlProjectGenericFileFilter()
	{
		GenericFileFilter fnef1 = getGenericFileFilter(GenericFileFilterChooserGeneralMorphing.XML_FILE_FILTER);
		List<GenericFileFilter> result = CollectionFunctions.instance().createList( fnef1 );

		return( result );
	}

	public DialogsWrapper getDialogsWrapper()
	{
		return( GenericFunctions.instance().getDialogsWrapper() );
	}

	protected String useFileChooser(String initialFileName, int open_save_dialog)
	{
		List<GenericFileFilter> ffl = createMorphingProjectGenericFileFilter();
		String fileName = showFileChooserDialog(open_save_dialog, ffl, initialFileName);

		return( fileName );
	}

	protected String useXmlFileChooser(int open_save_dialog)
	{
		List<GenericFileFilter> ffl = createXmlProjectGenericFileFilter();

		FileChooserParameters fcp = new FileChooserParameters();
		fcp.setGenericFileFilterList(ffl);
		fcp.setMode( DesktopFileChooserParameters.FILES_ONLY );
		fcp.setOpenOrSaveDialog(open_save_dialog);
		fcp.setChosenGenericFileFilter(ffl.get(0));

		String fileName = showFileChooserDialog(fcp);

		return( forceExtension(fileName, "xml") );
	}

	protected FileFunctions getFileFunctions() {
		return FileFunctions.instance();
	}

	public String forceExtension(String fileName, String extension) {
		return getFileFunctions().forceExtension(fileName, extension);
	}

	protected String showFileChooserDialog(FileChooserParameters pffc) {
		return getDialogsWrapper().showFileChooserDialog(getMainWindow(), pffc);
	}

	protected String showFileChooserDialog(int open_save_dialog, List<GenericFileFilter> fileFilterList, String defaultFileName) {
		return getDialogsWrapper().showFileChooserDialog(getMainWindow(), open_save_dialog, fileFilterList, defaultFileName);
	}

	public String getProjectFileNameAs(String initialFileName)
	{
		String fileName = useFileChooser(initialFileName, DialogsWrapper.SAVE);
		String result = FileFunctions.instance().forceExtension( fileName, "morph" );

		return( result );
	}

	protected ExecutionFunctions getExecutionFunctions() {
		return ExecutionFunctions.instance();
	}

	public <CC> CC functionWithDialog(ExecutionFunctions.UnsafeFunction<CC> run, int dialogType,
									  Function<Throwable, String> errorMessageBuilder) {
		return getExecutionFunctions().functionWithDialog(run, getMainWindow(), dialogType, errorMessageBuilder);
	}

	public Exception methodWithDialog(ExecutionFunctions.UnsafeMethod run,
									  int dialogType, Function<Throwable, String> errorMessageBuilder) {
		return getExecutionFunctions().methodWithDialog(run, getMainWindow(), dialogType, errorMessageBuilder);
	}

	protected Exception methodWithDialog(ExecutionFunctions.UnsafeMethod run,
										 int dialogType, String exceptionMessagePattern )
	{
		return( getExecutionFunctions().methodWithDialog(run, getMainWindow(),
																dialogType, exceptionMessagePattern) );
	}

	protected  <CC> CC functionWithDialog( ExecutionFunctions.UnsafeFunction<CC> run,
										   int dialogType, String exceptionMessagePattern )
	{
		return( getExecutionFunctions().functionWithDialog(run, getMainWindow(),
																dialogType, exceptionMessagePattern) );
	}

	protected boolean isEmpty( String str )
	{
		return( StringFunctions.instance().isEmpty(str) );
	}

	public String saveProjectAs(String initialFileName)
	{
		String fileName = getProjectFileNameAs(initialFileName);
		if( isEmpty(fileName) )
			fileName = null;
		else
			fileName = saveProject(fileName);

		return( fileName );
	}

	protected void updateProjectModel( MorphingProjectWholeFileModel projectModel )
	{
		projectModel.setName( getView().getProjectName() );
	}

	public MorphingProjectWholeFileModel getMorphingProjectModel() {
		return getAppliCtx().getMorphingProjectModel();
	}

	public FaceDetectionWholeFileModel getFaceDetectionWholeFileModel() {
		return getAppliCtx().getFaceDetectionWholeFileModel();
	}

	public void saveMorphingProjectModelAs(String fileName) throws IOException, TransformerException {
		getAppliCtx().saveMorphingProjectModelAs(fileName);
	}

	public MorphingProjectWholeFileModel loadMorphingProjectModel(String fileName) throws IOException, SAXException, ParserConfigurationException {
		return getAppliCtx().loadMorphingProjectModel(fileName);
	}

	public void saveFaceDetections() {
		getAppliCtx().saveFaceDetections();
	}

	public String saveProject(String fileName)
	{
		String result = null;
		if( ! isEmpty(fileName) )
		{
			updateProjectModel( getMorphingProjectModel() );

			Exception ex = methodWithDialog( () ->
							saveMorphingProjectModelAs(fileName),
					DialogsWrapper.ERROR_MESSAGE,
					getPatternWithFile( getInternationalString( CONF_ERROR_WHEN_SAVING_MORPHING_PROJECT ), fileName ) );

			if( ex == null )
			{
				showInfoMessage( getInternationalString( CONF_PROJECT_SAVED_SUCCESSFULLY ) );
				result = fileName;
			}
		}
		else
			result = saveProjectAs();

		return( result );
	}

	protected String createErrorString(String label, Throwable th) {
		return createCustomInternationalString(label, th.getMessage());
	}

	public void exportFaceDetections()
	{
		String fileName = functionWithDialog( this::exportFaceDetectionsInternal,
				DialogsWrapper.ERROR_MESSAGE,
				th -> createErrorString(CONF_ERROR_EXPORTING_FILE, th) );

		if(fileName != null) {
			showInfoMessage(createCustomInternationalString(CONF_SUCCESS_EXPORTING_FILE, fileName));
		}
	}

	public void importFaceDetections()
	{
		Integer numImportedFaceImages = functionWithDialog(this::importFaceDetectionsInternal,
				DialogsWrapper.ERROR_MESSAGE,
				th -> createErrorString(CONF_ERROR_IMPORTING_FILE, th));

		if(numImportedFaceImages != null) {
			showInfoMessage(createCustomInternationalString(CONF_SUCCESS_IMPORTING_FILE, numImportedFaceImages));
		}
	}

	protected String exportFaceDetectionsInternal() {
		String fileName = useXmlFileChooser(DialogsWrapper.SAVE);
		if(fileName != null) {
			getAppliCtx().exportFaceDetections(fileName);
		}
		return fileName;
	}

	protected Integer importFaceDetectionsInternal() {
		Integer numImportedImages = null;
		String fileName = useXmlFileChooser(DialogsWrapper.OPEN);
		if(fileName != null) {
			numImportedImages = getAppliCtx().importFaceDetections(fileName);
		}
		return numImportedImages;
	}

	protected void showInfoMessage( String message )
	{
		getDialogsWrapper().showInformation( getMainWindow(), message);
	}

	public String getProjectFileNameForLoading(String initialFileName)
	{
		return( useFileChooser(initialFileName, DialogsWrapper.OPEN) );
	}

	public String loadProject(String initialFileName)
	{
//		MorphingProjectWholeFileModel result = null;
		String fileName = getProjectFileNameForLoading(initialFileName);
		if( fileName != null )
		{
//			result = functionWithDialog( () ->
			functionWithDialog( () ->
//							getProjectModelManager().loadMorphingProjectModel(fileName),
					getAppliCtx().loadMorphingProjectModel(fileName),
				DialogsWrapper.ERROR_MESSAGE,
				getPatternWithFile( getInternationalString( CONF_ERROR_WHEN_LOADING_MORPHING_PROJECT ), fileName ) );
		}

		return( fileName );
	}

	protected String getDefaultProjectName()
	{
		return( getInternationalString( CONF_DEFAULT_PROJECT_NAME ) );
	}

	protected String getProjectName( MorphingProjectWholeFileModel projectModel )
	{
		String result = projectModel.getName();
		if( StringFunctions.instance().isEmpty(result) )
			result = getDefaultProjectName();

		return( result );
	}
//
//	protected void updateModel( MorphingProjectWholeFileModel projectModel )
//	{
//		getProjectModelManager().setMorphingProjectModel(projectModel);
//	}

	protected void updateModelAndView()//MorphingProjectWholeFileModel projectModel )
	{
////		updateModel( projectModel );
//		updateView( projectModel, true );
		updateView( true );
	}

	protected void updateView( //MorphingProjectWholeFileModel projectModel,
								boolean madatoryUpdate)
	{
		MorphingProjectWholeFileModel projectModel = getMorphingProjectModel();
		Point2d normamlizedPointCenter = getNormalizedPointCenter();
		getView().setProjectName( getProjectName( projectModel ) );
		getView().setFileName( projectModel.getFileName() );
		setCurrentTriangulationToView(normamlizedPointCenter, madatoryUpdate);
	}

	protected String getPatternWithFile( String pattern, String fileName )
	{
		return( String.format( "%s (%s)", pattern, fileName ) );
	}

	protected void removeViewListeners()
	{
		if( _view != null )
		{
			_view.removeDelaunayTriangulatorViewListener( this::processDelaunayTriangulatorViewEvent );
			_view.removeImageListTableListener( this::processImageListTableEvent );
			_view.removeFileHandlingListener( this::processFileHandlingPanelEvent );
			_view.setMainWindowListener(null);
		}
	}

	public MorphingView getView() {
		return _view;
	}

	protected MorphingVerticesMapManager getVerticesManager()
	{
		return( getAppliCtx().getVerticesManager() );
	}

	public void processImageListTableEvent( GenericWrapperEvent evt )
	{
		if( evt instanceof JTableWrapperRecordsClonedEvent)
			processClonedRecordsEvent( (JTableWrapperRecordsClonedEvent) evt );
		if( evt instanceof GenTableWrapperDropFilesEvent)
			processDropFilesEvent( (GenTableWrapperDropFilesEvent) evt );
		else if( evt instanceof GenTableWrapperNewCurrentEvent)
			processNewImageSelectedEvent( (GenTableWrapperNewCurrentEvent) evt );
		else if( evt instanceof GenTableWrapperRecordsDeletedEvent)
			deleteImages((GenTableWrapperRecordsDeletedEvent) evt);
		else if( evt instanceof GenTableWrapperRecordsMovedEvent)
			relocateImages((GenTableWrapperRecordsMovedEvent) evt);
		else if( evt instanceof GenTableWrapperRecordsSortedEvent)
			relocateImages((GenTableWrapperRecordsSortedEvent) evt);
	}

	protected void processNewImageSelectedEvent( GenTableWrapperNewCurrentEvent<MorphingImageModel, ?> newImageSelectedEvt )
	{
		Point2d normamlizedPointCenter = getNormalizedPointCenter();
		GeometricRectangle oldBounds = getVerticesManager().getBounds( getVerticesManager().getCurrentImageModelName() );
		getVerticesManager().setCurrentImageIndex( newImageSelectedEvt.getEventData().getNewCurrentRow() );
		GeometricRectangle newBounds = getVerticesManager().getBounds( getVerticesManager().getCurrentImageModelName() );
		double newZoomFactorForImage = calculateNewZoomFactor( oldBounds, newBounds );

		setCurrentTriangulationToView(normamlizedPointCenter, true, newZoomFactorForImage);
	}

	protected double calculateNewZoomFactor( GeometricRectangle oldBounds,
											GeometricRectangle newBounds )
	{
		return( getView().getImageZoomFactor() *
				calculateFactorOfFactors( oldBounds, newBounds ) );
	}

	protected double aritmeticMean( double ... values )
	{
		return( DoubleFunctions.instance().aritmeticMean(values));
	}

	protected double calculateFactorOfFactors( GeometricRectangle oldBounds,
											GeometricRectangle newBounds )
	{
		double result = 1.0d;
		if( ( oldBounds != null ) && ( newBounds != null ) )
			result = aritmeticMean( oldBounds.width / newBounds.width,
									oldBounds.height / newBounds.height );

		return( result );
	}

	protected Point2d getNormalizedPointCenter()
	{
		return( getView().getNormalizedPointImageCenterShown() );
	}

	protected void setNotMandatoryTriangulationToView()
	{
		Point2d normamlizedPointCenter = null;
		boolean mandatoryUpdate = false;
		Double newImageZoomFactor = null;
		setTriangulationToView( normamlizedPointCenter, mandatoryUpdate,
								newImageZoomFactor);
	}

	protected void processClonedRecordsEvent( JTableWrapperRecordsClonedEvent<MorphingImageModel, JTable> evt )
	{
		cloneRecords( evt.getEventData().getPrevious() );

//		setCurrentTriangulationToView(null, true);

		setNotMandatoryTriangulationToView();

		getView().repaint();
	}

	protected void cloneRecords( ListStateContextBase<MorphingImageModel> previousListState )
	{
		getVerticesManager().cloneImages( previousListState );
//		setCurrentTriangulationToView(null, true);
	}

	protected void processDropFilesEvent( GenTableWrapperDropFilesEvent<MorphingImageModel, JTable> dadEvt )
	{
		Point2d normamlizedPointCenter = getNormalizedPointCenter();
		droppedFiles( dadEvt.getEventData().getFileList(),
					dadEvt.getEventData().getIndexForInsertion() );
		setCurrentTriangulationToView(normamlizedPointCenter, true);
	}

	protected void setCurrentTriangulationToView(Point2d normamlizedPointCenter,
												boolean madatoryUpdate )
	{
		setCurrentTriangulationToView( normamlizedPointCenter, madatoryUpdate, null );
	}

	public void setTriangulationToView(Point2d normamlizedPointCenter,
										boolean madatoryUpdate,
										Double newImageZoomFactor)
	{
		getView().updateTriangulation( getVerticesManager().getCurrentImageModelName(),
//									getVerticesManager(),
									normamlizedPointCenter, madatoryUpdate,
									newImageZoomFactor);
	}

	protected void setCurrentTriangulationToView(Point2d normamlizedPointCenter,
												boolean madatoryUpdate,
												Double newImageZoomFactor )
	{
		setTriangulationToView(normamlizedPointCenter, madatoryUpdate,
								newImageZoomFactor);

		getView().newImageChosen(getVerticesManager().getCurrentImageModel());
		getView().setLayerType(getImageName(getVerticesManager().getCurrentImageModel()),
				TriangulationPainterLayerTypeEnum.DELAUNAY_TRIANGULATION);
		getView().repaint();
	}

	protected String getImageName(MorphingImageModel imageModel) {
		return getIfNotNull(imageModel, MorphingImageModel::getName);
	}

	protected NullFunctions getNullFunctions() {
		return NullFunctions.instance();
	}

	protected <CC, RR> RR getIfNotNull(CC obj, Function<CC, RR> getter) {
		return getNullFunctions().getIfNotNull(obj, getter);
	}

	protected void droppedFiles(List<File> droppedFileList, int indexWhereToInsert)
	{
		getVerticesManager().droppedFiles( droppedFileList, indexWhereToInsert );
	}

	public void processDelaunayTriangulatorViewEvent( GenericWrapperEvent evt )
	{
		if( evt instanceof VertexEvent )
			processVertexEvent( (VertexEvent)evt );
		else if( evt instanceof NavigatorEvt )
			processNavigatorEvt( (NavigatorEvt) evt );
		else if( evt instanceof NewBoundaryButtonPressedEvt )
			processNewBoundaryAsked((NewBoundaryButtonPressedEvt) evt );
		else if( evt instanceof FadedMasterImageButtonPressedEvt )
			processFadedMasterImageButtonPressedEvt((FadedMasterImageButtonPressedEvt) evt );
		else if( evt instanceof MakeMasterImageEvt )
			processMakeMasterImageEvt((MakeMasterImageEvt) evt );
		else if( evt instanceof NewOnTheFlyBoundsEvt )
			processNewOnTheFlyBoundsEvt((NewOnTheFlyBoundsEvt) evt );
		else if( evt instanceof NewConfirmedBoundsEvt )
			processNewConfirmedBoundsEvt((NewConfirmedBoundsEvt) evt );
		else if( evt instanceof ClearTriangulationEvent) {
			clearTriangulation();
		} else if( evt instanceof TriangleSelectedEvent ) {
			processTriangleSelectedEvent( (TriangleSelectedEvent) evt );
		} else if(evt instanceof RunFaceDetectionEvent) {
			processRunFaceDetectionEvent((RunFaceDetectionEvent) evt);
		} else if(evt instanceof FaceDetectionSelectionEvent) {
			processFaceDetectionSelectionEvent((FaceDetectionSelectionEvent) evt);
		} else if(evt instanceof LayerSelectionEvent) {
			processLayerSelectionEvent((LayerSelectionEvent) evt);
		} else if(evt instanceof RecalculateTriangulationEvent) {
			processRecalculateTriangulationEvent((RecalculateTriangulationEvent) evt);
		} else if(evt instanceof ResetModifiedDetectedFaceMeshEvent) {
			processResetModifiedDetectedFaceMeshEvent((ResetModifiedDetectedFaceMeshEvent) evt);
		} else if(evt instanceof ResetDetectedFaceMeshFrameEvent) {
			processResetDetectedFaceMeshFrameEvent((ResetDetectedFaceMeshFrameEvent) evt);
		} else if(evt instanceof ChangeFollowingTransitionEvt) {
			processChangeFollowingTransitionEvt((ChangeFollowingTransitionEvt) evt);
		} else if(evt instanceof Insert68DetectedPointsEvt) {
			processInsert68DetectedPointsEvt((Insert68DetectedPointsEvt) evt);
		}
	}

	protected void processResetModifiedDetectedFaceMeshEvent(ResetModifiedDetectedFaceMeshEvent<?> evt) {
		getVerticesManager().resetModifiedDetectedFaceMesh(evt.getEventData().getImageName());
		getView().repaint();
	}

	protected void processInsert68DetectedPointsEvt(Insert68DetectedPointsEvt<?> evt) {
		Insert68DetectedPointsData evtData = evt.getEventData();
		String imageName = evtData.getImageName();

		getVerticesManager().insertControlVerticesToTriangulation(imageName);
	}

	protected void processChangeFollowingTransitionEvt(ChangeFollowingTransitionEvt<?> evt) {
		ChangeFollowingTransitionEvtData evtData = evt.getEventData();
		MorphingImageModel imageModel = getImageModel(evtData.getImageName());
		if(imageModel != null) {
			imageModel.setFollowingTransition(evtData.getTransition());
		}
	}

	protected void processResetDetectedFaceMeshFrameEvent(ResetDetectedFaceMeshFrameEvent<?> evt) {
		getVerticesManager().resetDetectedFaceMeshFrame(evt.getEventData().getImageName());
		getView().repaint();
	}

	protected boolean confirmDialog(ViewComponent viewComponent, String message, String title) {
		return GenericFunctions.instance().getDialogsWrapper().showYesNoDialog(
				viewComponent, message, title, DialogsWrapper.NO) == DialogsWrapper.YES;
	}

	protected boolean confirmationOfRecalculationOfAllMeshes() {
		return confirmDialog(getView().getDelaunayTriangulatorView(),
				getInternationalString(CONF_CONFIRM_RECALCULATION_OF_ALL_MESHES), null);
	}

	protected void processRecalculateTriangulationEvent(RecalculateTriangulationEvent<?> evt) {
		RecalculateTriangulationData data = evt.getEventData();
		RecalculateTriangulationType type = data.getType();

		if(type == RecalculateTriangulationType.CURRENT) {
			getVerticesManager().reAutoCalculateVertices(data.getImageName());
		} else if(type == RecalculateTriangulationType.ALL) {
			if(confirmationOfRecalculationOfAllMeshes()) {
				getVerticesManager().reCalculateVerticesForAllImagesDifferentFromMaster();
			}
		} else {
			LOGGER.warn("RecalculateTriangulationType not valid: {}", type);
		}

		getView().repaint();
	}

	protected void processLayerSelectionEvent(LayerSelectionEvent<?> evt) {
		LayerSelectionData data = evt.getEventData();

		getView().setLayerType(data.getImageName(), data.getSelectedLayer());
		getView().repaint();
	}

	protected MorphingImageModel getImageModel(String imageName) {
		return getMorphingProjectModel().getImageModel(imageName);
	}

	protected void processFaceDetectionSelectionEvent(FaceDetectionSelectionEvent<?> evt) {
		FaceDetectionSelectionData data = evt.getEventData();
		String imageName = data.getImageName();
		FaceDetectionMeshData faceDetectionMeshData = data.getFaceDetectionMeshData();
		FaceDetectionSelectionType type = data.getType();
		getView().setFaceDetectionMeshData(imageName, faceDetectionMeshData);

		if(type == FaceDetectionSelectionType.SELECTED) {
			FaceDetection modifiedFaceDetection = faceDetectionMeshData == null ? null : faceDetectionMeshData.getModifiedFaceDetection();
			getVerticesManager().setImageFaceDetection(getImageModel(imageName), modifiedFaceDetection);

			TriangulationPainterLayerTypeEnum layer;
			if(modifiedFaceDetection == null) {
				layer = TriangulationPainterLayerTypeEnum.DELAUNAY_TRIANGULATION;
			} else {
				layer = TriangulationPainterLayerTypeEnum.FACE_DETECTION_VERTEX_MODIFICATION;
			}
			getView().updateTriangulation(imageName);//, getProjectModelManager()
			getView().setLayerType(imageName, layer);
		}
		getView().repaint();
	}

	protected void processRunFaceDetectionEvent(RunFaceDetectionEvent<?> evt) {
		String imageName = evt.getEventData().getImageName();
		FaceDetectionQueryExecution faceDetectionQueryExecution = new FaceDetectionQueryExecution(
				getView().getDelaunayTriangulatorView(),
				getVerticesManager().getImageFileName(imageName),
				getAppliCtx().getImageFaceDetectionPersistenceModelMap(),
				getAppliConf(), FaceDetectionQueryFactory.instance(),
				imageFaceDetection -> processFaceDetectionResult(imageName, imageFaceDetection));

		// blocking the EDT, preventing from the user doing actions
		faceDetectionQueryExecution.run();
	}

	protected void processFaceDetectionResult(String imageName, ImageFaceDetection imageFaceDetection) {
		SwingUtilities.invokeLater(() -> processFaceDetectionResultInternal(imageName, imageFaceDetection));
	}

	protected void processFaceDetectionResultInternal(String imageName, ImageFaceDetection imageFaceDetection) {
		getView().setAllFaces(imageName, createFaceDetectionMeshDataList(imageFaceDetection));
		getView().setLayerType(imageName, TriangulationPainterLayerTypeEnum.FACE_DETECTION_SELECTION);
		getView().repaint();
	}

	public List<FaceDetectionMeshData> createFaceDetectionMeshDataList(ImageFaceDetection imageFaceDetection) {
		return getFaceDetectionMeshDataBuilder().createFaceDetectionMeshDataList(imageFaceDetection);
	}

	protected FaceDetectionMeshDataBuilderAndUpdater getFaceDetectionMeshDataBuilder() {
		return FaceDetectionMeshDataBuilderAndUpdater.instance();
	}

	protected void clearTriangulation()
	{
		Point2d normamlizedPointCenter = getNormalizedPointCenter();
		getVerticesManager().clearTriangulation();
		setCurrentTriangulationToView(normamlizedPointCenter, true);
		getView().resetSelectedMultiTriangle();
	}

	protected void processNewOnTheFlyBoundsEvt( NewOnTheFlyBoundsEvt<?> evt )
	{
		NewBoundsEventData data = evt.getEventData();
		GeometricRectangle newBounds = data.getNewBounds();
		getView().setNewOnTheFlyBounds(data.getImageName(), newBounds );
		getView().repaint();
	}

	protected void processTriangleSelectedEvent( TriangleSelectedEvent<?> evt )
	{
		TriangleSelectedEventData data = evt.getEventData();
		MultiTriangle multiTriangle = getVerticesManager().getMultiTriangle( data.getVertex(), data.getImageName() );
		if( multiTriangle != null )
			getView().setSelectedMultiTriangle(multiTriangle);
	}

	protected void processNewConfirmedBoundsEvt( NewConfirmedBoundsEvt<?> evt )
	{
		NewBoundsEventData data = evt.getEventData();
		MorphingImageModel imageModel = getVerticesManager().getImageModel(data.getImageName());
		processNewBounds( imageModel, data.getNewBounds() );
	}

	protected void processFadedMasterImageButtonPressedEvt( FadedMasterImageButtonPressedEvt<?> evt )
	{
		getView().fadeToMaster( evt.getEventData().getFadingPercentage() );
	}

	protected void processMakeMasterImageEvt( MakeMasterImageEvt<?> evt )
	{
		String imageName = evt.getEventData().getImageName();
		MorphingImageModel im = this.getVerticesManager().getImageModel(imageName);
		if( ! im.isMaster() )
			getVerticesManager().setMasterImageName(imageName);

		setCurrentTriangulationToView(null, true);
	}

	protected void processNewBoundaryAsked( NewBoundaryButtonPressedEvt evt )
	{
		MorphingImageModel imageModel = getVerticesManager().getCurrentImageModel();
		MorphingImageModel masterImageModel = getVerticesManager().getMasterImageModel();
		getView().askNewForANewBoundary( imageModel,
										masterImageModel.getBounds(),
										bounds -> processNewBounds( imageModel, bounds ) );
	}

	protected void processNewBounds( MorphingImageModel imageModel,
									GeometricRectangle newBounds )
	{
		getVerticesManager().setNewBounds(imageModel, newBounds);

		setNotMandatoryTriangulationToView();
		getView().repaint();
//		setCurrentTriangulationToView(null, true);
	}

	protected void processVertexEvent( VertexEvent<?> evt )
	{
		if( evt.getEventData().getType() == VertexEventType.DELETE )
			additionalDeletionTasks( evt.getEventData().getImageName(),
									evt.getEventData().getVertex() );

		getVerticesManager().processNewPoint(getView(), evt);
	}

	protected void additionalDeletionTasks( String imageName, Point2d pt )
	{
		if( contains( getView().getSelectedMultiTriangle(), imageName, pt ) )
			getView().resetSelectedMultiTriangle();
	}

	protected boolean contains( MultiTriangle selectedTriangle, String imageName, Point2d pt )
	{
		boolean result = false;
		if( selectedTriangle != null )
			result = Objects.equals( selectedTriangle.getVertex1(imageName), pt ) ||
				Objects.equals( selectedTriangle.getVertex2(imageName), pt ) ||
				Objects.equals( selectedTriangle.getVertex3(imageName), pt );

		return( result );
	}		

	protected void processNavigatorEvt( NavigatorEvt<?> evt )
	{
		Point2d normamlizedPointCenter = getNormalizedPointCenter();
		GeometricRectangle oldBounds = getVerticesManager().getBounds( getVerticesManager().getCurrentImageModelName() );
		NavigatorEvtType type = evt.getEventData().getNavigatorEvtType();
		switch( type )
		{
			case START:
				switchToTriangulationLayer();
				getVerticesManager().imageStart();
				break;
			case PREVIOUS:
				switchToTriangulationLayer();
				getVerticesManager().imagePrevious();
				break;
			case NEXT:
				switchToTriangulationLayer();
				getVerticesManager().imageNext();
				break;
			case END:
				switchToTriangulationLayer();
				getVerticesManager().imageEnd();
				break;
		}
		GeometricRectangle newBounds = getVerticesManager().getBounds( getVerticesManager().getCurrentImageModelName() );
		double newZoomFactorForImage = calculateNewZoomFactor( oldBounds, newBounds );
		setCurrentTriangulationToView(normamlizedPointCenter, true, newZoomFactorForImage);
	}

	protected void switchToTriangulationLayer() {
		getView().setLayerType(getVerticesManager().getCurrentImageModelName(),
				TriangulationPainterLayerTypeEnum.DELAUNAY_TRIANGULATION);
	}


	public MorphingImageModel addImage( String imageName, String imageFileName,
										GeometricRectangle bounds,
										MorphingProjectConfigurationModel configurationModel,
										boolean isMaster,
										int index)
	{
		return( getAppliCtx().addImage(imageName, imageFileName,
			bounds, configurationModel, isMaster, index) );
	}
/*
	public boolean removeImageModel( int index )
	{
		return( getProjectModelManager().removeImageModel( index ) );
	}

	public boolean removeImageModel( String imageName )
	{
		return( getProjectModelManager().removeImageModel( imageName ) );
	}

	public void relocateImage( String imageName, int toIndex )
	{
		getProjectModelManager().relocateImage(imageName, toIndex);
	}

	public void relocateImage( int fromIndex, int toIndex )
	{
		getProjectModelManager().relocateImage(toIndex, toIndex);
	}
*/
	public void relocateImages( GenTableWrapperRecordsModifiedEventBase<MorphingImageModel, ?, ListStateContextBase<MorphingImageModel>> evt )
	{
		ListStateContextBase<MorphingImageModel> modifiedListState = evt.getEventData().getAfter();

		getAppliCtx().relocateImages( modifiedListState.getList() );

		updateCurrentImageList( modifiedListState );
//		setCurrentTriangulationToView(null, true);
	}

	protected void updateCurrentImageList( ListStateContextBase<MorphingImageModel> modifiedListState )
	{
		getView().updateCurrentTableList(modifiedListState);

		if( modifiedListState.getSelectedItem() != getVerticesManager().getCurrentImageModel() )
			setCurrentTriangulationToView(null, true);
	}

	protected <CC> CC safeFunctionExecution( ExecutionFunctions.UnsafeFunction<CC> run )
	{
		return( ExecutionFunctions.instance().safeFunctionExecution(run) );
	}

	public Exception safeMethodExecution( ExecutionFunctions.UnsafeMethod run )
	{
		return( ExecutionFunctions.instance().safeMethodExecution(run) );
	}

	protected ViewComponent getMainWindow()
	{
		return( getView().getMainWindow() );
	}

	public void deleteImages( GenTableWrapperRecordsModifiedEventBase<MorphingImageModel, ?, ListStateContextBase<MorphingImageModel>> evt )
	{
		if( confirmErasureOfImages(evt) )
			deleteImagesInternal( evt );
	}

	protected boolean confirmErasureOfImages(GenTableWrapperRecordsModifiedEventBase<MorphingImageModel, ?, ListStateContextBase<MorphingImageModel>> evt)
	{
		return( getView().confirmErasureOfSelectedRecords(evt) );
	}

	protected boolean askToProceed( String text )
	{
		return( getDialogsWrapper().showYesNoDialog( getView().getMainWindow(), text, null, 1 ) == 0 );
	}

	public void deleteImagesInternal( GenTableWrapperRecordsModifiedEventBase<MorphingImageModel, ?, ListStateContextBase<MorphingImageModel>> evt )
	{
		ListStateContextBase<MorphingImageModel> modifiedListState = evt.getEventData().getAfter();
		Point2d normamlizedPointCenter = getNormalizedPointCenter();
		Exception ex = ExecutionFunctions.instance().methodWithDialog(
				() -> getAppliCtx().deleteImages( evt.getEventData().getPrevious().getSelection(),
												modifiedListState.getList() ),
				getMainWindow(), DialogsWrapper.ERROR_MESSAGE,
				getInternationalString( CONF_ERROR_WHEN_ERASING_IMAGES ) );

		if( ex == null )
			updateCurrentImageList( modifiedListState );

//			setCurrentTriangulationToView(normamlizedPointCenter, true);
	}

	public String getInternationalString(String label)
	{
		return( _internationalizedStringConf.getInternationalString( label ) );
	}

	public String createCustomInternationalString(String label, Object... args) {
		return( _internationalizedStringConf.createCustomInternationalString( label, args ) );
	}
}
