/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.applications.morphing.imagelisttable;

import com.frojasg1.applications.common.configuration.InternationalizedStringConf;
import com.frojasg1.applications.libmorphing.model.images.image.MorphingImageModel;
import com.frojasg1.applications.morphing.imagelisttable.evt.JTableWrapperRecordsClonedEvent;
import com.frojasg1.applications.morphing.imagelisttable.evt.MorphingTableEventBuilder;
import com.frojasg1.applications.morphing.language.MorphingApplicationInternationalTexts;
import com.frojasg1.general.FileFunctions;
import com.frojasg1.general.desktop.view.colors.Colors;
import com.frojasg1.general.desktop.view.componentwrapper.jtable.column.RecordToTableCellStringsBase;
import com.frojasg1.general.desktop.view.componentwrapper.jtable.files.BaseFileListJTableUI;
import com.frojasg1.general.desktop.view.componentwrapper.jtable.files.InternationalColumnArrangeableDragAndDropFilesJTableWrapper;
import com.frojasg1.general.desktop.view.icons.IconResources;
import com.frojasg1.general.desktop.view.menus.BaseJPopupMenu;
import com.frojasg1.general.desktop.view.paint.PaintTextContext;
import java.awt.Font;
import java.util.List;
import javax.swing.JMenuItem;
import javax.swing.JSeparator;
import javax.swing.JTable;
import javax.swing.table.TableCellRenderer;

/**
 *
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 */
public class MorphingImageModelListTableWrapper
	extends InternationalColumnArrangeableDragAndDropFilesJTableWrapper<MorphingImageModel, BaseFileListJTableUI> {

	public static final String CONF_DRAG_AND_DROP_FILES = "DRAG_AND_DROP_FILES";

	public static final String CONF_CLONE_SELECTED_ROWS = "CLONE_SELECTED_ROWS";


	public MorphingImageModelListTableWrapper( JTable table )
	{
		this( table, MorphingApplicationInternationalTexts.instance(),
			new MorphingImageModelToTableCellStrings(table, MorphingApplicationInternationalTexts.instance()) );
	}

	public MorphingImageModelListTableWrapper( JTable table, InternationalizedStringConf internationalStrings,
								RecordToTableCellStringsBase<MorphingImageModel> recordToCellStrings )
	{
		super( table, new MorphingTableEventBuilder( table, MorphingImageModel.class ),
				internationalStrings, recordToCellStrings );

		init();
	}

	@Override
	protected PaintTextContext createPaintTextContext()
	{
		PaintTextContext result = new PaintTextContext();
		
		result.setHundredPercentSize( 18 );
		result.setNotInvertedForegroundColor( Colors.SEMITRANSPARENT_SEA_BLUE );
		result.setTextGetter( this::getTableText );

		return( result );
	}

	protected String getTableText()
	{
		String result = null;
		if( isEmpty() )
			result = getInternationalString( CONF_DRAG_AND_DROP_FILES );

		return( result );
	}

	public void updateImageList( List<MorphingImageModel> images )
	{
		setRecordList(images);
	}

	@Override
	protected MorphingImageModel createEmptyRecord() {
		return( null );
	}

	public MorphingImageModel getCurrentImage()
	{
		return( getCurrentRecord() );
	}

	@Override
	protected TableCellRenderer createSimpleCellRenderer()
	{
		return( new ModifiedCellRendererPlayList( getFont() ) );
	}

	@Override
	protected BaseFileListJTableUI createEmptyUI() {
		BaseFileListJTableUI result = new BaseFileListJTableUI();
		
		return( result );
	}

	protected boolean isFile(String fileName)
	{
		return( FileFunctions.instance().isFile(fileName) );
	}

	@Override
	protected BaseJPopupMenu createBasePopupMenu() {
		TablePopupMenuWithClonation result = new TablePopupMenuWithClonation();
		result.init();

		return( result );
	}

	protected void cloneSelectedRows()
	{
		notifyRecordsCloned();
	}

	protected JTableWrapperRecordsClonedEvent<MorphingImageModel, ?> createRecordsClonedEvt()
	{
		return( getDragAndDropJTableWrapperContext()
			.createRecordsModifiedGenEvt(JTableWrapperRecordsClonedEvent.class, null) );
	}

	public void notifyRecordsCloned()
	{
		if( ! isEmpty( getTable().getSelectedRows() ) )
			notifyEvt( createRecordsClonedEvt() );
	}

	protected class TablePopupMenuWithClonation extends TablePopupMenu
	{
		protected void createFixedPart()
		{
			super.createFixedPart();

			addMenuComponent( new JSeparator() );
			addMenuComponent( CONF_CLONE_SELECTED_ROWS, new JMenuItem(),
				() -> getInternationalStrings().getInternationalString(CONF_CLONE_SELECTED_ROWS),
				() -> cloneSelectedRows(),
				IconResources.COPY_ICON_RESOURCE );
		}

		protected void preparePopupMenuItemsFixedPart() {
			super.preparePopupMenuItemsFixedPart();
			boolean enable = !isSelectionEmpty();
			getMenuItem(CONF_CLONE_SELECTED_ROWS).setEnabled( enable );
		}
	}

	protected class ModifiedCellRendererPlayList extends ModifiedCellRenderer
	{
		public ModifiedCellRendererPlayList( Font font )
		{
			super(font);
		}

		@Override
		protected boolean isAnErroneusCell( int row, int column )
		{
			MorphingImageModel image = getRecordAtModelRowIndex(row);
			return( ( image == null ) || !isFile( image.getFileName() ) );
		}
	}
}
