/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.applications.morphing.view;

import com.frojasg1.applications.libmorphing.api.facedetector.pojo.FaceDetectionMeshData;
import com.frojasg1.applications.libmorphing.model.images.color.types.MultiTriangle;
import com.frojasg1.applications.libmorphing.model.images.image.MorphingImageModel;
import com.frojasg1.applications.libmorphing.mvc.DelaunayTriangulatorWindowView;
import com.frojasg1.applications.libmorphing.view.DelaunayTriangulationViewerWindowBase;
import com.frojasg1.applications.libmorphing.view.mainwindow.evt.MainWindowEvent;
import com.frojasg1.applications.libmorphing.view.newboundschooser.NewBoundsChooserLauncher;
import com.frojasg1.applications.libmorphing.view.painters.TriangulationPainterLayerTypeEnum;
import com.frojasg1.applications.libmorphing.xml.whole.MorphingProjectWholeFileModel;
import com.frojasg1.applications.morphing.MainWindow;
import com.frojasg1.applications.morphing.context.ApplicationContextImp;
import com.frojasg1.applications.morphing.mvc.MorphingView;
import com.frojasg1.libgeometry.types.GeometricRectangle;
import com.frojasg1.libgeometry.types.Point2d;
import com.frojasg1.general.ExecutionFunctions;
import com.frojasg1.general.ExecutionFunctions.UnsafeFunction;
import com.frojasg1.general.collection.functions.ListStateContextBase;
import com.frojasg1.general.desktop.view.panels.impl.file.FileHandlingPanelEvent;
import com.frojasg1.general.exceptions.ConfigurationException;
import com.frojasg1.general.view.ViewComponent;
import com.frojasg1.general.view.evt.wrapper.GenericWrapperEvent;
import com.frojasg1.general.view.evt.wrapper.impl.table.impl.GenTableWrapperRecordsModifiedEventBase;
import java.util.List;
import java.util.Objects;
import java.util.function.Consumer;

/**
 *
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 */
public class MorphingViewManager implements MorphingView
{
	protected MainWindow _mainWindow;
	protected DelaunayTriangulatorWindowView _delaunayTriangulatorView;


	protected ApplicationContextImp appliCtx;

	protected String _imageName;

	public MorphingViewManager ( ApplicationContextImp ac )
	{
		this(ac, true);
	}

	protected MorphingViewManager ( ApplicationContextImp ac, boolean init )
	{
		this.appliCtx = ac;

		if( init )
			init();
	}

	public <CC> CC safeFunctionExecution( UnsafeFunction<CC> run )
	{
		return( ExecutionFunctions.instance().safeFunctionExecution(run) );
	}

	protected void init()
	{
		_mainWindow = safeFunctionExecution( this::createMainWindow );
		_delaunayTriangulatorView = createDelaunayTriangulatorView();
	}

	public ApplicationContextImp getAppliCtx() {
		return appliCtx;
	}

	protected DelaunayTriangulationViewerWindowBase createDelaunayTriangulatorView()
	{
		return( new DelaunayTriangulationViewerWindowBase( getAppliCtx() ) );
	}

	protected MainWindow createMainWindow() throws ConfigurationException
	{
		MainWindow result = new MainWindow("Morphing effect creator application", getAppliCtx() );

		return( result );
	}

	@Override
	public MainWindow getMainWindow() {
		return _mainWindow;
	}

	public void setMainWindow(MainWindow _mainWindow) {
		this._mainWindow = _mainWindow;
	}

	@Override
	public DelaunayTriangulatorWindowView getDelaunayTriangulatorView() {
		return _delaunayTriangulatorView;
	}

	public void setDelaunayTriangulatorView(DelaunayTriangulatorWindowView _delaunayTriangulatorView) {
		this._delaunayTriangulatorView = _delaunayTriangulatorView;
	}

	@Override
	public void addDelaunayTriangulatorViewListener(Consumer<? super GenericWrapperEvent> listener,
													Long ... eventTypes) {
		getDelaunayTriangulatorView().addDelaunayTriangulatorViewListener(listener, eventTypes);
	}

	@Override
	public void removeDelaunayTriangulatorViewListener(Consumer<? super GenericWrapperEvent> listener) {
		getDelaunayTriangulatorView().removeDelaunayTriangulatorViewListener(listener);
	}

	@Override
	public void addImageListTableListener(Consumer<? super GenericWrapperEvent> listener,
										Long ... eventTypes) {
		getMainWindow().addImageListTableListener(listener, eventTypes);
	}

	@Override
	public void removeImageListTableListener(Consumer<? super GenericWrapperEvent> listener) {
		getMainWindow().removeImageListTableListener(listener);
	}

	@Override
	public void setPointInProgress(String imageName, Point2d pt) {
		getDelaunayTriangulatorView().setPointInProgress(imageName, pt);
	}

	@Override
	public void setClosestPoint(String imageName, Point2d pt) {
		getDelaunayTriangulatorView().setClosestPoint(imageName, pt);
	}

	@Override
	public void setControlMeshClosestPoint(String imageName, Point2d pt) {
		getDelaunayTriangulatorView().setControlMeshClosestPoint(imageName, pt);
	}

	@Override
	public void updateTriangulation(String imageName,
//									MorphingVerticesMapManager morphingVerticesManager,
									Point2d normalizedCenterPoint,
									boolean mandatoryUpdate,
									Double newImageZoomFactor) {
		if( ! Objects.equals( imageName, _imageName ) || mandatoryUpdate )
		{
			getDelaunayTriangulatorView().updateTriangulation(imageName,
//															morphingVerticesManager,
															normalizedCenterPoint,
															mandatoryUpdate,
															newImageZoomFactor);
			_imageName = imageName;
		}

//		String currentImageName = morphingVerticesManager.getCurrentImageModelName();
		List<MorphingImageModel> imageList = getAppliCtx().getImageModelList();
		getMainWindow().setNewImageListInternal( imageList,
////												imageList.indexOf( morphingVerticesManager.getImageModel(currentImageName)),
//				imageList.indexOf( morphingVerticesManager.getImageModel(imageName)),
				imageList.indexOf( getAppliCtx().getImageModel(imageName)),
												null, mandatoryUpdate);
		getMainWindow().updateTriangulation(imageName);//, morphingVerticesManager);
	}

	public void updateCurrentTableList( ListStateContextBase<MorphingImageModel> modifiedListState )
	{
		getMainWindow().updateCurrentTableList( modifiedListState );
	}

	@Override
	public void updateTriangulation(String imageName)
	{
		updateTriangulation(imageName,
						null, true, null);
		getMainWindow().setFileName( getCurrentProjectFileName());
		getMainWindow().updateTriangulation(imageName);
	}

	protected MorphingProjectWholeFileModel getMorphingProjectModel() {
		return getAppliCtx().getMorphingProjectModel();
	}

	protected String getCurrentProjectFileName()
	{
		String result = null;
		MorphingProjectWholeFileModel morphingProjectModel = getMorphingProjectModel();
//		if( morphingProjectManager != null )
//			morphingProjectModel = morphingProjectManager.getMorphingProjectModel();

		if( morphingProjectModel != null )
			result = morphingProjectModel.getFileName();

		if( result == null )
			result = "";

		return( result );
	}

	@Override
	public void setVisible(boolean value) {
		getMainWindow().setVisible(value);
		getDelaunayTriangulatorView().setVisible(true);
	}

	@Override
	public void requestFocus() {
		throw new UnsupportedOperationException("Not supported yet."); //To change body of generated methods, choose Tools | Templates.
	}

	@Override
	public boolean isFocusable() {
		throw new UnsupportedOperationException("Not supported yet."); //To change body of generated methods, choose Tools | Templates.
	}

	@Override
	public boolean hasFocus() {
		throw new UnsupportedOperationException("Not supported yet."); //To change body of generated methods, choose Tools | Templates.
	}

	@Override
	public ViewComponent getParentViewComponent() {
		throw new UnsupportedOperationException("Not supported yet."); //To change body of generated methods, choose Tools | Templates.
	}

	@Override
	public void repaint() {
		getDelaunayTriangulatorView().repaint();
		getMainWindow().repaint();
	}

	@Override
	public void releaseResources() {
		throw new UnsupportedOperationException("Not supported yet."); //To change body of generated methods, choose Tools | Templates.
	}

	@Override
	public void addFileHandlingListener(Consumer<FileHandlingPanelEvent> listener)
	{
		getMainWindow().addFileHandlingListener(listener);
	}

	@Override
	public void removeFileHandlingListener(Consumer<FileHandlingPanelEvent> listener)
	{
		getMainWindow().removeFileHandlingListener(listener);
	}

	@Override
	public void setMainWindowListener(Consumer<MainWindowEvent> listener) {
		getMainWindow().setMainWindowListener(listener);
	}

	@Override
	public String getFileName() {
		return( getMainWindow().getFileName() );
	}

	@Override
	public void setFileName(String fileName) {
		getMainWindow().setFileName(fileName);
	}

	@Override
	public String getProjectName()
	{
		return( getMainWindow().getProjectName() );
	}

	@Override
	public void setProjectName( String projectName )
	{
		getMainWindow().setProjectName(projectName);
	}

	@Override
	public void newImageChosen( MorphingImageModel mim )
	{
		getMainWindow().newImageChosen(mim);
		getDelaunayTriangulatorView().setFaceDetectionMeshData(mim.getName(), null);
	}

	@Override
	public void askNewForANewBoundary( MorphingImageModel mim,
										GeometricRectangle masterBoundary,
										Consumer<GeometricRectangle> callback )
	{
		NewBoundsChooserLauncher.instance().askForANewBoundary( getMainWindow(), mim,
																masterBoundary, callback);
	}

	@Override
	public void setNewOnTheFlyBounds( String imageName, GeometricRectangle bounds )
	{
		getDelaunayTriangulatorView().setNewOnTheFlyBounds( imageName, bounds );
	}

	@Override
	public GeometricRectangle getFittingImageSizeForJPanel()
	{
		return( getMainWindow().getFittingImageSizeForJPanel() );
	}

	public boolean confirmErasureOfSelectedRecords(
		GenTableWrapperRecordsModifiedEventBase<MorphingImageModel,
                                                    ?,
                                                    ListStateContextBase<MorphingImageModel>> evt )
	{
		return( getMainWindow().confirmErasureOfSelectedRecords(evt) );
	}

	@Override
	public double getFadingFraction()
	{
		return( getDelaunayTriangulatorView().getFadingFraction() );
	}

	@Override
	public void fadeToMaster(double fadingFraction)
	{
		getDelaunayTriangulatorView().fadeToMaster(fadingFraction);
	}

	@Override
	public void resetFadedToMasterImage()
	{
		getDelaunayTriangulatorView().resetFadedToMasterImage();
	}

	@Override
	public Point2d getNormalizedPointImageCenterShown()
	{
		return( getDelaunayTriangulatorView().getNormalizedPointImageCenterShown() );
	}	

	@Override
	public void setSelectedMultiTriangle( MultiTriangle multiTriangle )
	{
		getDelaunayTriangulatorView().setSelectedMultiTriangle(multiTriangle);
	}

	@Override
	public MultiTriangle getSelectedMultiTriangle()
	{
		return( getDelaunayTriangulatorView().getSelectedMultiTriangle() );
	}

	@Override
	public void resetSelectedMultiTriangle()
	{
		getDelaunayTriangulatorView().resetSelectedMultiTriangle();
	}

	@Override
	public double getImageZoomFactor()
	{
		return( getDelaunayTriangulatorView().getImageZoomFactor() );
	}

	@Override
	public void setAllFaces(String imageName, List<FaceDetectionMeshData> allFaces) {
		getDelaunayTriangulatorView().setAllFaces(imageName, allFaces);
	}

	@Override
	public void setLayerType(String imageName, TriangulationPainterLayerTypeEnum layerType) {
		getDelaunayTriangulatorView().setLayerType(imageName, layerType);
	}

	@Override
	public void setFaceDetectionMeshData(String imageName, FaceDetectionMeshData faceDetectionMeshData) {
		getDelaunayTriangulatorView().setFaceDetectionMeshData(imageName, faceDetectionMeshData);
	}
}
