/*
 * Copyright (C) 2022 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/gpl-3.0.txt
 *
 */

package com.frojasg1.applications.morphing.videobuilder;

import com.frojasg1.applications.common.components.internationalization.window.InternationalizationInitializationEndCallback;
import com.frojasg1.applications.common.components.internationalization.window.InternationalizedJFrame;
import com.frojasg1.applications.common.components.resizecomp.MapResizeRelocateComponentItem;
import com.frojasg1.applications.common.components.resizecomp.ResizeRelocateItem;
import com.frojasg1.applications.common.configuration.listener.ConfigurationParameterListener;
import com.frojasg1.applications.common.configuration.listener.ConfigurationParameterObserved;
import com.frojasg1.applications.libmorphing.frames.MorphingFrameSequencer;
import com.frojasg1.applications.libmorphing.controller.MorphingVerticesMapManager;
import com.frojasg1.applications.libmorphing.view.filefilters.GenericFileFilterChooserGeneralMorphing;
import com.frojasg1.applications.libmorphing.view.timers.SwingImageSequenceShower;
import com.frojasg1.applications.libmorphing.xml.whole.MorphingProjectWholeFileModel;
import com.frojasg1.applications.morphing.videobuilder.configuration.ApplicationConfiguration;
import com.frojasg1.applications.morphing.videobuilder.context.ApplicationContextImp;
import com.frojasg1.applications.morphing.videobuilder.video.MorphingVideoCreator;
import com.frojasg1.libgeometry.types.GeometricRectangle;
import com.frojasg1.libgeometry.types.factory.GeometryCommonFactory;
import com.frojasg1.general.CollectionFunctions;
import com.frojasg1.general.DateFunctions;
import com.frojasg1.general.ExecutionFunctions;
import com.frojasg1.general.FileFunctions;
import com.frojasg1.general.desktop.view.about.GenericAboutJDialog;
import com.frojasg1.general.desktop.view.color.ColorInversor;
import com.frojasg1.general.desktop.view.icons.IconResources;
import com.frojasg1.general.desktop.view.panels.impl.file.FileHandlingPanelEventType;
import com.frojasg1.general.desktop.view.panels.impl.file.FileHandllingJPanel;
import com.frojasg1.general.desktop.view.pdf.ImageJPanel;
import com.frojasg1.general.desktop.view.pdf.ImageJPanelSizeParams;
import com.frojasg1.general.desktop.view.zoom.mapper.ComponentMapper;
import com.frojasg1.general.dialogs.DialogsWrapper;
import com.frojasg1.general.dialogs.filefilter.GenericFileFilter;
import com.frojasg1.general.dialogs.filefilter.GenericFileFilterChooser;
import com.frojasg1.general.dialogs.filefilter.impl.gen.GenericFileFilterChooserGeneral;
import com.frojasg1.general.exceptions.ConfigurationException;
import com.frojasg1.general.number.IntegerFunctions;
import com.frojasg1.general.process.ProcExecResult;
import com.frojasg1.general.process.ProcessFunctions;
import com.frojasg1.general.string.StringFunctions;
import com.frojasg1.general.view.internationalization.window.InternException;
import com.frojasg1.generic.GenericFunctions;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.swing.*;
import java.awt.*;
import java.awt.event.ComponentAdapter;
import java.awt.event.ComponentEvent;
import java.awt.event.KeyEvent;
import java.awt.image.BufferedImage;
import java.util.Collections;
import java.util.List;
import java.util.Properties;
import java.util.Vector;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicReference;
import java.util.function.Consumer;


// Preferred size: [955, 720]

/**
 * @author Fran
 */
public class MainWindow extends InternationalizedJFrame<ApplicationContextImp>
        implements ConfigurationParameterListener//, MorphingView
{
    private static final Logger LOGGER = LoggerFactory.getLogger(MainWindow.class);

    public static final String CONF_ERROR_WHEN_LOADING_MORPHING_PROJECT = "ERROR_WHEN_LOADING_MORPHING_PROJECT";
    public static final String CONF_ERROR_WHEN_SAVING_MORPHING_VIDEO = "ERROR_WHEN_SAVING_MORPHING_VIDEO";
    public static final String CONF_MORPHING_VIDEO_SUCCESSFULLY_SAVED = "MORPHING_VIDEO_SUCCESSFULLY_SAVED";
    public static final String CONF_FFMPEG_BINARY_FILENAME_DOES_NOT_SEEM_TO_BE_VALID = "FFMPEG_BINARY_FILENAME_DOES_NOT_SEEM_TO_BE_VALID";
    public static final String CONF_FFMPEG_BINARY_FILENAME_NOT_SET = "FFMPEG_BINARY_FILENAME_NOT_SET";
    public static final String CONF_YOU_MUST_DOWNLOAD_IT_UNPACK_IT_AND_SELECT_IT_FROM_YOUR_FILE_SYSTEM = "YOU_MUST_DOWNLOAD_IT_UNPACK_IT_AND_SELECT_IT_FROM_YOUR_FILE_SYSTEM";

    protected static final String CONF_ERROR_CREATING_MORPHING_FRAME = "ERROR_CREATING_MORPHING_FRAME";
    protected static final String CONF_IGNORE = "IGNORE";
    protected static final String CONF_CANCEL = "CANCEL";

    protected static final String CONF_DEMOS_VIDEO_CREATION_URL = "DEMOS_VIDEO_CREATION_URL";
    protected static final String CONF_DEMOS_DEMO_VIDEO_URL = "DEMOS_DEMO_VIDEO_URL";

    public static final String sa_configurationBaseFileName = "MainWindow";

    protected Vector<JPopupMenu> a_vectorJpopupMenus = null;

    protected GenericAboutJDialog _aboutDial = null;

    //	protected MorphingVerticesMapManager _morphingVerticesManager;
    protected AtomicBoolean _cancelMorphingSimulation = new AtomicBoolean();

    protected SwingImageSequenceShower _imageShower;
    protected ImageJPanel _imageJPanel;
    protected AtomicReference<ImageJPanelSizeParams> _imageJPanelSizeReference = new AtomicReference<>();
//
//	protected MorphingProjectWholeFileModelManager _projectModelManager;

    /**
     * Creates new form NewJFrame
     */
    public MainWindow(String windowName, ApplicationContextImp appCtx) throws ConfigurationException {
        super(ApplicationConfiguration.instance(), appCtx, true);
        setName(windowName);

        initComponents();

        initOwnComponents();

        setWindowConfiguration();
    }

    protected void setInitialValues() {
        setFfmpegBinaryFileName(getAppliConf().getFfmpegBinaryFileName());
    }

    protected boolean isEmpty(String str) {
        return (StringFunctions.instance().isEmpty(str));
    }

    protected void setFfmpegBinaryFileName(String fileName) {
        if (checkFfmpegBinaryFileName(fileName))
            setFfmpegBinaryFileNameInternal(fileName);
        else {
            if (isEmpty(getFfmpegBinaryFileName()))
                setFfmpegBinaryFileNameInternal(fileName);
            else
                showFfmpegBinaryCheckWarning(fileName);
        }
    }

    protected ProcExecResult execute(String... commandAndParams) {
        return (ExecutionFunctions.instance().safeFunctionExecution(() ->
                ProcessFunctions.instance().execute(commandAndParams))
        );
    }

    protected boolean checkFfmpegBinaryFileName(String fileName) {
        ProcExecResult procExecResult = execute(fileName, "-version");

        return ((procExecResult != null) && procExecResult.wasSuccessful() &&
                startsWith(getFirst(procExecResult.getStandardOutput()), "ffmpeg version")
        );
    }

    protected boolean startsWith(String str, String start) {
        return (StringFunctions.instance().stringStartsWith(str, start));
    }

    protected <V> V getFirst(List<V> list) {
        return (CollectionFunctions.instance().getFirstOf(list));
    }

    protected void showFfmpegBinaryCheckWarning(String fileName) {
        getDialogsWrapper().showError(this, createCustomInternationalString(CONF_FFMPEG_BINARY_FILENAME_DOES_NOT_SEEM_TO_BE_VALID, fileName));
    }

    protected void showFfmpegBinaryNotSetWarning(String fileName) {
        getDialogsWrapper().showError(this,
                getInternationalString(CONF_FFMPEG_BINARY_FILENAME_NOT_SET) + "\n" +
                        getInternationalString(CONF_YOU_MUST_DOWNLOAD_IT_UNPACK_IT_AND_SELECT_IT_FROM_YOUR_FILE_SYSTEM));
    }

    protected void setFfmpegBinaryFileNameInternal(String fileName) {
        jTF_ffmpegBinary.setText(fileName);
    }

    protected String getFfmpegBinaryFileName() {
        return (jTF_ffmpegBinary.getText());
    }

    @Override
    public void internationalizationInitializationEndCallback() {
        super.internationalizationInitializationEndCallback();

        SwingUtilities.invokeLater(() -> {
            setListeners();
        });
        setAfterWindowConfiguration();

        M_applyNewApplicationConfiguration();
    }

    public void addDemoMenuItems() {
        addDemoMenuItem(jMI_demos_videoCreation, CONF_DEMOS_VIDEO_CREATION_URL);
        addDemoMenuItem(jMI_demos_videoCreatedWithThisApplication, CONF_DEMOS_DEMO_VIDEO_URL);
    }

    /*
        protected void setViewToController()
        {
            if( isInitialized() && ( _controller != null ) )
                _controller.setView(this);
            else if( _controller == null )
                LOGGER.info( "Controller was null" );
            else
                LOGGER.info( "Not initializaed yet" );
        }
    */
    @Override
    protected void invokeLaterAfterInitialized() {
//		setViewToController();
    }

    protected void setAfterWindowConfiguration() {
    }

    /*
        public void setController(MorphingController _controller) {
            this._controller = _controller;
            setViewToController();
        }

        public MorphingController getController() {
            return _controller;
        }
    */
    protected void initOwnComponents() {
//		_projectModelManager = createMorphingProjectWholeFileModelManager();

        _imageJPanel = createImageJPanel();
        _imageJPanel.addComponentListener(new ComponentAdapter() {
            @Override
            public void componentResized(ComponentEvent e) {
                updateImageJPanelSize();
            }
        });
        addComponentToCompletelyFillParent(jPanel3, _imageJPanel);

        SwingUtilities.invokeLater(() -> updateImageJPanelSize());

        initialValues();
    }
//
//    protected MorphingProjectWholeFileModelManagerBase createMorphingProjectWholeFileModelManagerEmpty() {
//        return (new MorphingProjectWholeFileModelManagerBase(getApplicationContext()));
//    }
//
//    protected MorphingProjectWholeFileModel loadMorphingProjectModel(String fileName) throws SAXException, IOException, ParserConfigurationException {
//        MorphingProjectWholeFileModel projectModel = getApplicationContext().loadMorphingProjectModel(fileName);
//
//        return (projectModel);
//    }

//	public MorphingProjectWholeFileModelManager getProjectModelManager() {
//		return getApplicationContext().getProjectModelManager();
//	}
//
//	protected MorphingProjectWholeFileModelManager createMorphingProjectWholeFileModelManager()
//	{
//		MorphingProjectWholeFileModelManager result = createMorphingProjectWholeFileModelManagerEmpty();
////		MorphingProjectWholeFileModel model = safeFunctionExecution( this::loadOrGetMorphingProjectModel );
////		if( model != null )
////			result.setMorphingProjectModel( model );
//
//		return( result );
//	}

    protected ImageJPanelSizeParams createImageSizeParams() {
        ImageJPanelSizeParams result = new ImageJPanelSizeParams();
        result.setSize(_imageJPanel.getRealSize());
        result.setZoomedWidth(_imageJPanel.getZoomedFrameWidth());

        return (result);
    }

    protected void updateImageJPanelSize() {
        _imageJPanelSizeReference.set(createImageSizeParams());
    }

    protected ImageJPanel createImageJPanel() {
        ImageJPanel result = new ImageJPanel(null);
        result.setCanInvertImageColors(false);
        result.setAllowShiftForMouseWheel(false);

        return (result);
    }

    protected FileHandllingJPanel createFileHandlerJPanel() {
        FileHandllingJPanel result = new FileHandllingJPanel();
        result.init();

        result.addAssociatedButton(FileHandlingPanelEventType.OPEN, jMI_open);
        result.addAssociatedButton(FileHandlingPanelEventType.SAVE, jMI_saveVideo);

        return (result);
    }

    /*
        public String getProjectName()
        {
            return( jTF_projectName.getText() );
        }

        public void setProjectName( String projectName )
        {
            jTF_projectName.setText(projectName);
        }
    */
    protected void initialValues() {
    }

    protected void setListeners() {
    }

    protected void setWindowConfiguration() {
        MapResizeRelocateComponentItem mapRRCI = new MapResizeRelocateComponentItem();
        try {
            mapRRCI.putResizeRelocateComponentItem(jPanel1, ResizeRelocateItem.FILL_WHOLE_PARENT);
            mapRRCI.putResizeRelocateComponentItem(jPanel2, ResizeRelocateItem.RESIZE_TO_RIGHT);
            mapRRCI.putResizeRelocateComponentItem(jPanel3, ResizeRelocateItem.RESIZE_TO_RIGHT +
                    ResizeRelocateItem.RESIZE_TO_BOTTOM);
            mapRRCI.putResizeRelocateComponentItem(jTF_ffmpegBinary, ResizeRelocateItem.RESIZE_TO_RIGHT);
            mapRRCI.putResizeRelocateComponentItem(jB_ffmpegBinaryFileName, ResizeRelocateItem.MOVE_TO_RIGHT);
            mapRRCI.putResizeRelocateComponentItem(_imageJPanel, ResizeRelocateItem.FILL_WHOLE_PARENT);
        } catch (Throwable th) {
            th.printStackTrace();
        }

        setLanguageMenu(jM_languageSubmenu);
        setZoomMenu(jM_zoom);

        setMenuItemIcons();

        createInternationalization(getAppliConf().getConfigurationMainFolder(),
                getAppliConf().getApplicationNameFolder(),
                getAppliConf().getApplicationGroup(),
                getAppliConf().getInternationalPropertiesPathInJar(),
                sa_configurationBaseFileName,
                this,
                null,
                a_vectorJpopupMenus,
                true,
                mapRRCI);

        registerInternationalString(CONF_ERROR_WHEN_LOADING_MORPHING_PROJECT, "Error when loading morphing project");
        registerInternationalString(CONF_ERROR_WHEN_SAVING_MORPHING_VIDEO, "Error when saving video");
        registerInternationalString(CONF_MORPHING_VIDEO_SUCCESSFULLY_SAVED, "Morphing video successfully saved");
        registerInternationalString(CONF_FFMPEG_BINARY_FILENAME_DOES_NOT_SEEM_TO_BE_VALID, "Ffmeg binary does not seem to be valid: \\n $1");
        registerInternationalString(CONF_FFMPEG_BINARY_FILENAME_NOT_SET, "Ffmeg binary is not set");
        registerInternationalString(CONF_YOU_MUST_DOWNLOAD_IT_UNPACK_IT_AND_SELECT_IT_FROM_YOUR_FILE_SYSTEM, "You must download it, unpack it and select it from your file system.");
        registerInternationalString(CONF_DEMOS_VIDEO_CREATION_URL, "https://frojasg1.com/demos/aplicaciones/morphing/v1.0_morphingVideoBuilder.VideoCreation.mp4");
        registerInternationalString(CONF_DEMOS_DEMO_VIDEO_URL, "https://frojasg1.com/demos/aplicaciones/morphing/solo.Modelos.mp4");

        registerInternationalString(CONF_ERROR_CREATING_MORPHING_FRAME, "Error creating morphing frame: $1\nYou can ignore or cancel.\nYou can inform the developer about this issue: frojasg1@hotmail.com" );
        registerInternationalString(CONF_IGNORE, "Ignore" );
        registerInternationalString(CONF_CANCEL, "Cancel" );

        boolean isClosure = false;
    }

    protected void setMenuItemIcons() {
        setDarkModeMenuItemIcon(jMI_darkMode);

        setIconForMenuItem(jMI_open, IconResources.OPEN_FILE_ICON_RESOURCE);
//		setIconForMenuItem(jMI_saveVideo, IconResources.SAVE_FILE_ICON_RESOURCE );
        setIconForMenuItem(jMI_exit, IconResources.EXIT_ICON_RESOURCE);

        setIconForMenuItem(jMI_saveVideo, IconResources.VIDEO_ICON_RESOURCE);
        setIconForMenuItem(jM_demos, IconResources.VIDEO_ICON_RESOURCE);
        setIconForMenuItem(jMI_demos_videoCreation, IconResources.VIDEO_ICON_RESOURCE);
        setIconForMenuItem(jMI_demos_videoCreatedWithThisApplication, IconResources.VIDEO_ICON_RESOURCE);

        setIconForMenuItem(jMI_whatIsNew, IconResources.WHAT_IS_NEW_ICON_RESOURCE);
        setIconForMenuItem(jMI_license, IconResources.LICENSE_ICON_RESOURCE);
        setIconForMenuItem(jMI_about, IconResources.ABOUT_ICON_RESOURCE);
        setIconForMenuItem(jM_demos, IconResources.VIDEO_ICON_RESOURCE);
    }

    protected void M_applyNewApplicationConfiguration() {
        setLanguageInMenu(getAppliConf().getLanguage());

        setInitialValues();
    }

    /**
     * This method is called from within the constructor to
     * initialize the form.
     * WARNING: Do NOT modify this code. The content of this method is
     * always regenerated by the Form Editor.
     */
    @SuppressWarnings("unchecked")
    // <editor-fold defaultstate="collapsed" desc="Generated Code">//GEN-BEGIN:initComponents
    // Generated using JFormDesigner non-commercial license
    private void initComponents() {
        jMenuBar1 = new JMenuBar();
        jM_File = new JMenu();
        jMI_open = new JMenuItem();
        jMI_saveVideo = new JMenuItem();
        jMI_exit = new JMenuItem();
        jM_view = new JMenu();
        jM_zoom = new JMenu();
        jMI_darkMode = new JMenuItem();
        jM_tools = new JMenu();
        jM_languageSubmenu = new JMenu();
        jRadioButtonMenuItemIdiomaEn = new JRadioButtonMenuItem();
        jRadioButtonMenuItemIdiomaEs = new JRadioButtonMenuItem();
        jRadioButtonMenuItemIdiomaOtro = new JRadioButtonMenuItem();
        jMI_simulateMorphing = new JMenuItem();
        jMI_cancelMorphingSimulation = new JMenuItem();
        jM_help = new JMenu();
        jM_demos = new JMenu();
        jMI_demos_videoCreation = new JMenuItem();
        jMI_demos_videoCreatedWithThisApplication = new JMenuItem();
        jMI_lookForNewVersion = new JMenuItem();
        jMI_whatIsNew = new JMenuItem();
        jMI_license = new JMenuItem();
        jMI_about = new JMenuItem();
        jPanel1 = new JPanel();
        jPanel2 = new JPanel();
        jL_Ffmpeg_download = new JLabel();
        jTF_ffmpegBinary = new JTextField();
        jL_ffmpeg = new JLabel();
        jB_ffmpegBinaryFileName = new JButton();
        jL_downloadItAt = new JLabel();
        jPanel3 = new JPanel();

        //======== this ========
        setDefaultCloseOperation(WindowConstants.DO_NOTHING_ON_CLOSE);
        setAutoRequestFocus(false);
        setMinimumSize(new Dimension(700, 500));
        setName("MainWindow");
        Container contentPane = getContentPane();
        contentPane.setLayout(null);

        //======== jMenuBar1 ========
        {
            jMenuBar1.setName("jMenuBar1");
            jMenuBar1.setPreferredSize(new Dimension(940, 21));

            //======== jM_File ========
            {
                jM_File.setText("File");
                jM_File.setName("jM_File");

                //---- jMI_open ----
                jMI_open.setText("Open Morphing project");
                jMI_open.setName("jMI_open");
                jMI_open.addActionListener(e -> jMI_openActionPerformed(e));
                jM_File.add(jMI_open);

                //---- jMI_saveVideo ----
                jMI_saveVideo.setText("Save Video");
                jMI_saveVideo.setEnabled(false);
                jMI_saveVideo.setName("jMI_saveVideo");
                jMI_saveVideo.addActionListener(e -> jMI_saveVideoActionPerformed(e));
                jM_File.add(jMI_saveVideo);
                jM_File.addSeparator();

                //---- jMI_exit ----
                jMI_exit.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_Q, KeyEvent.CTRL_DOWN_MASK));
                jMI_exit.setText("Exit");
                jMI_exit.setName("jMI_exit");
                jMI_exit.addActionListener(e -> jMI_exitActionPerformed(e));
                jM_File.add(jMI_exit);
            }
            jMenuBar1.add(jM_File);

            //======== jM_view ========
            {
                jM_view.setText("View");
                jM_view.setName("jM_view");

                //======== jM_zoom ========
                {
                    jM_zoom.setText("Zoom");
                    jM_zoom.setName("jM_zoom");
                }
                jM_view.add(jM_zoom);
                jM_view.addSeparator();

                //---- jMI_darkMode ----
                jMI_darkMode.setText("Dark mode");
                jMI_darkMode.setName("jMI_darkMode");
                jMI_darkMode.addActionListener(e -> jMI_darkModeActionPerformed(e));
                jM_view.add(jMI_darkMode);
            }
            jMenuBar1.add(jM_view);

            //======== jM_tools ========
            {
                jM_tools.setText("Tools");
                jM_tools.setName("jM_tools");

                //======== jM_languageSubmenu ========
                {
                    jM_languageSubmenu.setText("Language");
                    jM_languageSubmenu.setName("jM_languageSubmenu");

                    //---- jRadioButtonMenuItemIdiomaEn ----
                    jRadioButtonMenuItemIdiomaEn.setText("EN");
                    jRadioButtonMenuItemIdiomaEn.setName("");
                    jM_languageSubmenu.add(jRadioButtonMenuItemIdiomaEn);

                    //---- jRadioButtonMenuItemIdiomaEs ----
                    jRadioButtonMenuItemIdiomaEs.setText("ES");
                    jRadioButtonMenuItemIdiomaEs.setName("");
                    jM_languageSubmenu.add(jRadioButtonMenuItemIdiomaEs);

                    //---- jRadioButtonMenuItemIdiomaOtro ----
                    jRadioButtonMenuItemIdiomaOtro.setText("CAT");
                    jRadioButtonMenuItemIdiomaOtro.setName("");
                    jM_languageSubmenu.add(jRadioButtonMenuItemIdiomaOtro);
                }
                jM_tools.add(jM_languageSubmenu);
                jM_tools.addSeparator();

                //---- jMI_simulateMorphing ----
                jMI_simulateMorphing.setText("Simulate morphing");
                jMI_simulateMorphing.setName("jMI_simulateMorphing");
                jMI_simulateMorphing.addActionListener(e -> jMI_simulateMorphingActionPerformed(e));
                jM_tools.add(jMI_simulateMorphing);

                //---- jMI_cancelMorphingSimulation ----
                jMI_cancelMorphingSimulation.setText("Cancel simluation");
                jMI_cancelMorphingSimulation.setName("jMI_cancelMorphingSimulation");
                jMI_cancelMorphingSimulation.addActionListener(e -> jMI_cancelMorphingSimulationActionPerformed(e));
                jM_tools.add(jMI_cancelMorphingSimulation);
            }
            jMenuBar1.add(jM_tools);

            //======== jM_help ========
            {
                jM_help.setText("Help");
                jM_help.setName("jM_help");

                //======== jM_demos ========
                {
                    jM_demos.setText("Demos");
                    jM_demos.setName("jM_demos");

                    //---- jMI_demos_videoCreation ----
                    jMI_demos_videoCreation.setText("Video creation");
                    jMI_demos_videoCreation.setName("jMI_demos_videoCreation");
                    jM_demos.add(jMI_demos_videoCreation);

                    //---- jMI_demos_videoCreatedWithThisApplication ----
                    jMI_demos_videoCreatedWithThisApplication.setText("Video created with this application");
                    jMI_demos_videoCreatedWithThisApplication.setName("jMI_demos_videoCreatedWithThisApplication");
                    jM_demos.add(jMI_demos_videoCreatedWithThisApplication);
                }
                jM_help.add(jM_demos);
                jM_help.addSeparator();

                //---- jMI_lookForNewVersion ----
                jMI_lookForNewVersion.setText("Look for a new version");
                jMI_lookForNewVersion.setName("jMI_lookForNewVersion");
                jMI_lookForNewVersion.addActionListener(e -> jMI_lookForNewVersionActionPerformed(e));
                jM_help.add(jMI_lookForNewVersion);
                jM_help.addSeparator();

                //---- jMI_whatIsNew ----
                jMI_whatIsNew.setText("What is new");
                jMI_whatIsNew.setName("jMI_whatIsNew");
                jMI_whatIsNew.addActionListener(e -> jMI_whatIsNewActionPerformed(e));
                jM_help.add(jMI_whatIsNew);
                jM_help.addSeparator();

                //---- jMI_license ----
                jMI_license.setText("License");
                jMI_license.setName("jMI_license");
                jMI_license.addActionListener(e -> jMI_licenseActionPerformed(e));
                jM_help.add(jMI_license);
                jM_help.addSeparator();

                //---- jMI_about ----
                jMI_about.setText("About");
                jMI_about.setName("jMI_about");
                jMI_about.addActionListener(e -> jMI_aboutActionPerformed(e));
                jM_help.add(jMI_about);
            }
            jMenuBar1.add(jM_help);
        }
        setJMenuBar(jMenuBar1);

        //======== jPanel1 ========
        {
            jPanel1.setName("jPanel1");
            jPanel1.setLayout(null);

            //======== jPanel2 ========
            {
                jPanel2.setLayout(null);

                //---- jL_Ffmpeg_download ----
                jL_Ffmpeg_download.setForeground(new Color(0x279b00));
                jL_Ffmpeg_download.setText("https://github.com/BtbN/FFmpeg-Builds/releases");
                jL_Ffmpeg_download.setName("name=jL_Ffmpeg_download,url=https://github.com/BtbN/FFmpeg-Builds/releases");
                jPanel2.add(jL_Ffmpeg_download);
                jL_Ffmpeg_download.setBounds(145, 5, 615, jL_Ffmpeg_download.getPreferredSize().height);

                //---- jTF_ffmpegBinary ----
                jTF_ffmpegBinary.setMinimumSize(new Dimension(14, 15));
                jTF_ffmpegBinary.setPreferredSize(new Dimension(73, 15));
                jPanel2.add(jTF_ffmpegBinary);
                jTF_ffmpegBinary.setBounds(145, 25, 590, 20);

                //---- jL_ffmpeg ----
                jL_ffmpeg.setHorizontalAlignment(SwingConstants.TRAILING);
                jL_ffmpeg.setText("ffmpeg binary:");
                jL_ffmpeg.setName("jL_ffmpeg");
                jPanel2.add(jL_ffmpeg);
                jL_ffmpeg.setBounds(15, 25, 125, jL_ffmpeg.getPreferredSize().height);

                //---- jB_ffmpegBinaryFileName ----
                jB_ffmpegBinaryFileName.setText("...");
                jB_ffmpegBinaryFileName.setMaximumSize(new Dimension(1000, 1000));
                jB_ffmpegBinaryFileName.setMinimumSize(new Dimension(15, 15));
                jB_ffmpegBinaryFileName.setPreferredSize(new Dimension(77, 15));
                jB_ffmpegBinaryFileName.addActionListener(e -> jB_ffmpegBinaryFileNameActionPerformed(e));
                jPanel2.add(jB_ffmpegBinaryFileName);
                jB_ffmpegBinaryFileName.setBounds(740, 23, 20, 20);

                //---- jL_downloadItAt ----
                jL_downloadItAt.setHorizontalAlignment(SwingConstants.TRAILING);
                jL_downloadItAt.setText("Download it at:");
                jL_downloadItAt.setName("jL_downloadItAt");
                jPanel2.add(jL_downloadItAt);
                jL_downloadItAt.setBounds(15, 5, 125, jL_downloadItAt.getPreferredSize().height);

                {
                    // compute preferred size
                    Dimension preferredSize = new Dimension();
                    for(int i = 0; i < jPanel2.getComponentCount(); i++) {
                        Rectangle bounds = jPanel2.getComponent(i).getBounds();
                        preferredSize.width = Math.max(bounds.x + bounds.width, preferredSize.width);
                        preferredSize.height = Math.max(bounds.y + bounds.height, preferredSize.height);
                    }
                    Insets insets = jPanel2.getInsets();
                    preferredSize.width += insets.right;
                    preferredSize.height += insets.bottom;
                    jPanel2.setMinimumSize(preferredSize);
                    jPanel2.setPreferredSize(preferredSize);
                }
            }
            jPanel1.add(jPanel2);
            jPanel2.setBounds(0, 0, 770, 55);

            //======== jPanel3 ========
            {
                jPanel3.setLayout(null);

                {
                    // compute preferred size
                    Dimension preferredSize = new Dimension();
                    for(int i = 0; i < jPanel3.getComponentCount(); i++) {
                        Rectangle bounds = jPanel3.getComponent(i).getBounds();
                        preferredSize.width = Math.max(bounds.x + bounds.width, preferredSize.width);
                        preferredSize.height = Math.max(bounds.y + bounds.height, preferredSize.height);
                    }
                    Insets insets = jPanel3.getInsets();
                    preferredSize.width += insets.right;
                    preferredSize.height += insets.bottom;
                    jPanel3.setMinimumSize(preferredSize);
                    jPanel3.setPreferredSize(preferredSize);
                }
            }
            jPanel1.add(jPanel3);
            jPanel3.setBounds(0, 55, 770, 385);

            {
                // compute preferred size
                Dimension preferredSize = new Dimension();
                for(int i = 0; i < jPanel1.getComponentCount(); i++) {
                    Rectangle bounds = jPanel1.getComponent(i).getBounds();
                    preferredSize.width = Math.max(bounds.x + bounds.width, preferredSize.width);
                    preferredSize.height = Math.max(bounds.y + bounds.height, preferredSize.height);
                }
                Insets insets = jPanel1.getInsets();
                preferredSize.width += insets.right;
                preferredSize.height += insets.bottom;
                jPanel1.setMinimumSize(preferredSize);
                jPanel1.setPreferredSize(preferredSize);
            }
        }
        contentPane.add(jPanel1);
        jPanel1.setBounds(0, 0, 770, 440);

        {
            // compute preferred size
            Dimension preferredSize = new Dimension();
            for(int i = 0; i < contentPane.getComponentCount(); i++) {
                Rectangle bounds = contentPane.getComponent(i).getBounds();
                preferredSize.width = Math.max(bounds.x + bounds.width, preferredSize.width);
                preferredSize.height = Math.max(bounds.y + bounds.height, preferredSize.height);
            }
            Insets insets = contentPane.getInsets();
            preferredSize.width += insets.right;
            preferredSize.height += insets.bottom;
            contentPane.setMinimumSize(preferredSize);
            contentPane.setPreferredSize(preferredSize);
        }
        pack();
        setLocationRelativeTo(getOwner());
    }// </editor-fold>//GEN-END:initComponents

    private void jMI_exitActionPerformed(java.awt.event.ActionEvent evt)//GEN-FIRST:event_jMI_exitActionPerformed
    {//GEN-HEADEREND:event_jMI_exitActionPerformed
        // TODO add your handling code here:
        boolean closeWindow = true;
        formWindowClosing(closeWindow);
    }//GEN-LAST:event_jMI_exitActionPerformed


    private void jMI_lookForNewVersionActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_jMI_lookForNewVersionActionPerformed
        // TODO add your handling code here:

        checkForNewVersion();
    }//GEN-LAST:event_jMI_lookForNewVersionActionPerformed

    private void jMI_whatIsNewActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_jMI_whatIsNewActionPerformed
        // TODO add your handling code here:

        showWhatIsNew();
    }//GEN-LAST:event_jMI_whatIsNewActionPerformed

    private void jMI_licenseActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_jMI_licenseActionPerformed
        // TODO add your handling code here:

        showLicense();
    }//GEN-LAST:event_jMI_licenseActionPerformed

    private void jMI_aboutActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_jMI_aboutActionPerformed
        // TODO add your handling code here:

        _aboutDial = new AboutJDialog(this, true, (iec) -> setVisibleAbout(),
                getApplicationContext().getAnimationForAboutFactory());
    }//GEN-LAST:event_jMI_aboutActionPerformed

    private void jMI_darkModeActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_jMI_darkModeActionPerformed
        // TODO add your handling code here:

        toggleDarkMode();
    }//GEN-LAST:event_jMI_darkModeActionPerformed

    private void jMI_simulateMorphingActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_jMI_simulateMorphingActionPerformed
        // TODO add your handling code here:

        showMorphing();
    }//GEN-LAST:event_jMI_simulateMorphingActionPerformed

    private void jMI_cancelMorphingSimulationActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_jMI_cancelMorphingSimulationActionPerformed
        // TODO add your handling code here:

        _cancelMorphingSimulation.set(true);
    }//GEN-LAST:event_jMI_cancelMorphingSimulationActionPerformed

    private void jMI_saveVideoActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_jMI_saveVideoActionPerformed
        // TODO add your handling code here:

        saveVideo();
    }//GEN-LAST:event_jMI_saveVideoActionPerformed

    private void jMI_openActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_jMI_openActionPerformed
        // TODO add your handling code here:

        loadProject();

    }//GEN-LAST:event_jMI_openActionPerformed

    private void jB_ffmpegBinaryFileNameActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_jB_ffmpegBinaryFileNameActionPerformed
        // TODO add your handling code here:

        assignFfmpegFileName();
    }//GEN-LAST:event_jB_ffmpegBinaryFileNameActionPerformed

    // Variables declaration - do not modify//GEN-BEGIN:variables
    // Generated using JFormDesigner non-commercial license
    private JMenuBar jMenuBar1;
    private JMenu jM_File;
    private JMenuItem jMI_open;
    private JMenuItem jMI_saveVideo;
    private JMenuItem jMI_exit;
    private JMenu jM_view;
    private JMenu jM_zoom;
    private JMenuItem jMI_darkMode;
    private JMenu jM_tools;
    private JMenu jM_languageSubmenu;
    private JRadioButtonMenuItem jRadioButtonMenuItemIdiomaEn;
    private JRadioButtonMenuItem jRadioButtonMenuItemIdiomaEs;
    private JRadioButtonMenuItem jRadioButtonMenuItemIdiomaOtro;
    private JMenuItem jMI_simulateMorphing;
    private JMenuItem jMI_cancelMorphingSimulation;
    private JMenu jM_help;
    private JMenu jM_demos;
    private JMenuItem jMI_demos_videoCreation;
    private JMenuItem jMI_demos_videoCreatedWithThisApplication;
    private JMenuItem jMI_lookForNewVersion;
    private JMenuItem jMI_whatIsNew;
    private JMenuItem jMI_license;
    private JMenuItem jMI_about;
    private JPanel jPanel1;
    private JPanel jPanel2;
    private JLabel jL_Ffmpeg_download;
    private JTextField jTF_ffmpegBinary;
    private JLabel jL_ffmpeg;
    private JButton jB_ffmpegBinaryFileName;
    private JLabel jL_downloadItAt;
    private JPanel jPanel3;
    // End of variables declaration//GEN-END:variables

    protected void assignFfmpegFileName() {
        try {
            String newFileName = useFileChooserForOpeningExecBinary(getFfmpegBinaryFileName());

            if (newFileName != null)
                setFfmpegBinaryFileName(newFileName);
        } catch (Exception ex) {
            LOGGER.warn("Error when loading project", ex);
        }
    }

    public String getProjectFileNameForLoading(String initialFileName) {
        return (useFileChooserForProject(initialFileName, DialogsWrapper.OPEN, null));
    }

    public String getProjectFileNameForVideo() {
        return (useFileChooserForVideo(null, DialogsWrapper.SAVE, "mp4"));
    }

    protected String useFileChooserForVideo(String initialFileName, int open_save_dialog,
                                            String extension) {
        return (useFileChooser(initialFileName, open_save_dialog, Collections.emptyList(),
                extension));
    }

    protected String useFileChooserForProject(String initialFileName, int open_save_dialog,
                                              String extension) {
        return (useFileChooser(initialFileName, open_save_dialog,
                createMorphingProjectGenericFileFilter(), extension));
    }

    protected String useFileChooserForOpeningExecBinary(String initialFileName) {
        return (useFileChooser(initialFileName, DialogsWrapper.OPEN,
                createExecBinaryGenericFileFilter(), null));
    }

    protected String useFileChooser(String initialFileName, int open_save_dialog,
                                    List<GenericFileFilter> ffl, String extension) {
        String fileName = getDialogsWrapper().showFileChooserDialog(this,
                open_save_dialog,
                ffl, initialFileName);

        fileName = FileFunctions.instance().addExtension(fileName, extension);

        return (fileName);
    }

    public DialogsWrapper getDialogsWrapper() {
        return (GenericFunctions.instance().getDialogsWrapper());
    }

    protected List<GenericFileFilter> createMorphingProjectGenericFileFilter() {
        return (createList(getMorphingProjectGenericFileFilter()));
    }

    protected List<GenericFileFilter> createExecBinaryGenericFileFilter() {
        return (createList(getExecBinaryGenericFileFilter()));
    }

    public <CC> List createList(CC... elements) {
        return (CollectionFunctions.instance().createList(elements));
    }

    protected GenericFileFilterChooser getGenericFileFilterChooser() {
        return (GenericFileFilterChooserGeneralMorphing.instance());
    }

    protected GenericFileFilter getExecBinaryGenericFileFilter() {
        return (getGenericFileFilter(GenericFileFilterChooserGeneral.GENERIC_FILE_FILTER_FOR_EXECUTABLE_BY_ATTRIBUTES));
    }

    protected GenericFileFilter getMorphingProjectGenericFileFilter() {
        return (getGenericFileFilter(GenericFileFilterChooserGeneralMorphing.MORPHING_PROJECT_FILE_FILTER));
    }

    protected String getPatternWithFile(String pattern, String fileName) {
        return (String.format("%s (%s)", pattern, fileName));
    }

    public <CC> CC functionWithDialog(ExecutionFunctions.UnsafeFunction<CC> run,
                                      int dialogType, String exceptionMessagePattern) {
        return (ExecutionFunctions.instance().functionWithDialog(run, this,
                dialogType, exceptionMessagePattern));
    }
//
//	protected void updateModel( MorphingProjectWholeFileModel projectModel )
//	{
//		getProjectModelManager().setMorphingProjectModel(projectModel);
//	}

//	protected void updateModelAndView( MorphingProjectWholeFileModel projectModel )
//	{
////		updateModel( projectModel );
////		updateView( projectModel );
//		updateView();
//	}

    protected String getProjectName(MorphingProjectWholeFileModel projectModel) {
        String result = projectModel.getName();

        return (result);
    }

    //	protected void updateView( MorphingProjectWholeFileModel projectModel )
    protected void updateView() {
        MorphingProjectWholeFileModel projectModel = getApplicationContext().getMorphingProjectModel();
        String projectName = getProjectName(projectModel);
        setTitle(projectName);
        setTriangulation(null);//, getProjectModelManager().getVerticesManager() );

        jMI_saveVideo.setEnabled(projectModel != null);
    }

    protected MorphingProjectWholeFileModel loadProject() {
        MorphingProjectWholeFileModel result = null;
        String fileName = getProjectFileNameForLoading(null);
        if (fileName != null) {
            result = functionWithDialog(() ->
                            getApplicationContext().loadMorphingProjectModel(fileName),
                    DialogsWrapper.ERROR_MESSAGE,
                    getPatternWithFile(getInternationalString(CONF_ERROR_WHEN_LOADING_MORPHING_PROJECT), fileName));

            if (result != null)
                updateView();
//				updateModelAndView(result);
        }

        return (result);
    }

    protected void saveVideo() {
        String ffmpegBinaryFileName = getFfmpegBinaryFileName();
        if (!this.checkFfmpegBinaryFileName(ffmpegBinaryFileName)) {
            if (isEmpty(ffmpegBinaryFileName)) {
                showFfmpegBinaryNotSetWarning(ffmpegBinaryFileName);
                SwingUtilities.invokeLater(() -> focusAndHighlightComponent(jTF_ffmpegBinary));
            } else
                showFfmpegBinaryCheckWarning(ffmpegBinaryFileName);
        } else {
            String fileName = getProjectFileNameForVideo();

            if (fileName != null)
                saveVideo(fileName);
        }
    }

    protected LicenseJDialog createLicenseJDialog(Consumer<InternationalizationInitializationEndCallback> initializationEndCallBack) {
        return (new LicenseJDialog(this, getAppliConf(), initializationEndCallBack, false));
    }

    protected void showLicense() {
        createLicenseJDialog(this::showLicenseCallback);
    }

    protected void showLicenseCallback(InternationalizationInitializationEndCallback iiec) {
        LicenseJDialog ljd = (LicenseJDialog) iiec;
        ljd.setVisible(true);
        boolean closeWindow = true;
        ljd.formWindowClosing(closeWindow);
    }

    protected void setVisibleAbout() {
        SwingUtilities.invokeLater(() -> _aboutDial.setVisibleWithLock(true));
    }

    protected void setConfigurationChanges() {
        getAppliConf().setFfmpegBinaryFileName(getFfmpegBinaryFileName());
    }

    @Override
    public void formWindowClosing(boolean closeWindow) {
        try {
            setConfigurationChanges();

            try {
                ApplicationConfiguration.instance().M_saveConfiguration();
            } catch (ConfigurationException ce) {
                ce.printStackTrace();
            }
/*
			try
			{
//				StringsConfiguration.instance().M_saveConfiguration();
			}
			catch( Exception ce )
			{
				ce.printStackTrace();
			}

			if( a_intern != null )
			{
				try
				{
					a_intern.saveConfiguration();
				}
				catch( InternException ex )
				{
					ex.printStackTrace();
				}
			}
*/
            if (closeWindow) {
                removeListeners();

//				getAppliConf().removeConfigurationParameterListener( ApplicationConfiguration.CONF_PRECISION_FOR_CALCULATIONS, this );
            }

            super.formWindowClosing(closeWindow);
        } catch (Throwable th) {
            th.printStackTrace();
            System.exit(1);
        }

        if (closeWindow)
            System.exit(0);

//		System.exit(0);
    }

    protected void removeListeners() {
//		ComponentFunctions.instance().browseComponentHierarchy( this, (comp) -> { comp.removeFocusListener(_exitFocusListener); return( null ); } );
    }

    protected void changeZoomFactor(JScrollBar sb, double relativeZoom, Integer pos) {
        if ((sb != null) && (pos != null)) ;
        {
            pos = IntegerFunctions.zoomValueInt(pos, relativeZoom);

            sb.setValue(pos);
        }
    }

    protected void exit() {
        jMI_exit.doClick();
    }

    @Override
    protected void onIconify() {
        super.onIconify();
    }

    protected void addProperty(Properties prop, String label, String value) {
        prop.setProperty(label, value);
    }

    @Override
    public void changeLanguage(String language) throws ConfigurationException, InternException {
        super.changeLanguage(language);
    }

    @Override
    public void changeZoomFactor(double zoomFactor) {
        super.changeZoomFactor(zoomFactor);
    }

    @Override
    public <CC> void configurationParameterChanged(ConfigurationParameterObserved observed, String label, CC oldValue, CC newValue) {

    }

    @Override
    public ApplicationConfiguration getAppliConf() {
        return ((ApplicationConfiguration) _appliConf);
    }

    @Override
    protected void translateMappedComponents(ComponentMapper compMapper) {
        jMI_about = compMapper.mapComponent(jMI_about);
        jMI_darkMode = compMapper.mapComponent(jMI_darkMode);
        jMI_exit = compMapper.mapComponent(jMI_exit);
        jMI_license = compMapper.mapComponent(jMI_license);
        jMI_lookForNewVersion = compMapper.mapComponent(jMI_lookForNewVersion);
        jMI_open = compMapper.mapComponent(jMI_open);
        jMI_saveVideo = compMapper.mapComponent(jMI_saveVideo);
        jMI_whatIsNew = compMapper.mapComponent(jMI_whatIsNew);
        jM_File = compMapper.mapComponent(jM_File);
        jM_help = compMapper.mapComponent(jM_help);
        jM_languageSubmenu = compMapper.mapComponent(jM_languageSubmenu);
        jM_tools = compMapper.mapComponent(jM_tools);
        jM_view = compMapper.mapComponent(jM_view);
        jM_zoom = compMapper.mapComponent(jM_zoom);
        jMenuBar1 = compMapper.mapComponent(jMenuBar1);
        jPanel1 = compMapper.mapComponent(jPanel1);
        jRadioButtonMenuItemIdiomaEn = compMapper.mapComponent(jRadioButtonMenuItemIdiomaEn);
        jRadioButtonMenuItemIdiomaEs = compMapper.mapComponent(jRadioButtonMenuItemIdiomaEs);
        jRadioButtonMenuItemIdiomaOtro = compMapper.mapComponent(jRadioButtonMenuItemIdiomaOtro);
//		jSeparator1 = compMapper.mapComponent( jSeparator1 );
//		jSeparator11 = compMapper.mapComponent( jSeparator11 );
//		jSeparator12 = compMapper.mapComponent( jSeparator12 );
//		jSeparator2 = compMapper.mapComponent( jSeparator2 );
//		jSeparator25 = compMapper.mapComponent( jSeparator25 );
//		jSeparator5 = compMapper.mapComponent( jSeparator5 );

        jM_demos = compMapper.mapComponent(jM_demos);
        jMI_demos_videoCreation = compMapper.mapComponent(jMI_demos_videoCreation);
        jMI_demos_videoCreatedWithThisApplication = compMapper.mapComponent(jMI_demos_videoCreatedWithThisApplication);
    }

    @Override
    public void setDarkMode(boolean value, ColorInversor colorInversor) {
        super.setDarkMode(value, colorInversor);
    }

    public void setTriangulation(String imageName)
//			,
//								MorphingVerticesMapManager morphingVerticesManager)
    {
//		_morphingVerticesManager = morphingVerticesManager;
    }

    protected void showImage(BufferedImage image) {
        SwingUtilities.invokeLater( () -> showImageInternal(image));
    }

    protected void showImageInternal(BufferedImage image) {
//		safeMethodExecution( () -> ImageIO.write(image, "png", new File( "J:\\image.png" ) ) );
        LOGGER.info("New image: {}", DateFunctions.instance().formatDurationLong(System.currentTimeMillis()));
        _imageJPanel.setImageCenteringIt(image);
    }

//	protected void showImage( BufferedImage image )
//	{
//		SwingUtilities.invokeLater( () -> showImageInternal( image ) );
//	}

    protected ImageJPanelSizeParams getSizeParams() {
        return (_imageJPanelSizeReference.get());
    }

    public GeometricRectangle createRectangle(double x, double y, double width, double height) {
        return (GeometryCommonFactory.instance().createDelaunayRectangle(x, y, width, height));
    }

    public GeometricRectangle getFittingImageSizeForJPanel() {
        ImageJPanelSizeParams sizeParams = getSizeParams();
        Dimension dimen = _imageJPanel.getSizeToFit(sizeParams.getSize(), sizeParams.getZoomedWidth());
        return (createRectangle(0, 0, dimen.width, dimen.height));
    }

    protected void showMorphing() {
//		MorphingFrameSequencer seq = new MorphingFrameSequencer( _morphingVerticesManager );
        MorphingFrameSequencer seq = new MorphingFrameSequencer(getApplicationContext());

        _cancelMorphingSimulation.set(true);
        _cancelMorphingSimulation = new AtomicBoolean(false);
        AtomicReference<Exception> exceptionReference = new AtomicReference<>();

        SwingImageSequenceShower imageShower = new SwingImageSequenceShower(this::showImage, 250);
        _imageShower = imageShower;

        int delayBetweenFrames = 100;
        seq.doMorphing(imageShower::addImage,
                ex -> processExceptionProducedDuringMorphing(_cancelMorphingSimulation, exceptionReference, ex),
                _cancelMorphingSimulation,
                imageShower::hasToStop,
                this::getFittingImageSizeForJPanel,
                delayBetweenFrames);
    }
//
//    protected void createMorphingVideo() {
////		MorphingFrameSequencer seq = new MorphingFrameSequencer( _morphingVerticesManager );
//        MorphingFrameSequencer seq = new MorphingFrameSequencer(getApplicationContext());
//
//        _cancelMorphingSimulation.set(true);
//        _cancelMorphingSimulation = new AtomicBoolean(false);
//        AtomicReference<Exception> exceptionReference = new AtomicReference<>();
//
//        SwingImageSequenceShower imageShower = new SwingImageSequenceShower(this::showImage, 250);
//        _imageShower = imageShower;
//
//        int delayBetweenFrames = 0;
//        seq.doMorphing(imageShower::addImage,
//                ex -> processExceptionProducedDuringMorphing(_cancelMorphingSimulation, exceptionReference, ex),
//                _cancelMorphingSimulation,
//                imageShower::hasToStop,
//                this::getFittingImageSizeForJPanel,
//                delayBetweenFrames);
//    }

    protected void saveVideo(String fileName) {
        new Thread(() -> saveVideoInternal(fileName)).start();
    }

    protected MorphingVerticesMapManager getVerticesManager() {
        return getApplicationContext().getVerticesManager();
    }

    protected void saveVideoInternal(String fileName) {
//		MorphingVerticesMapManager verticesMapManager = getVerticesManager();
        MorphingFrameSequencer seq = new MorphingFrameSequencer(getApplicationContext());

        _cancelMorphingSimulation.set(true);
        _cancelMorphingSimulation = new AtomicBoolean(false);
        AtomicReference<Exception> exceptionReference = new AtomicReference<>();

        GeometricRectangle frameSize = getVideoFrameSize();
        MorphingVideoCreator videoCreator = new MorphingVideoCreator(getFfmpegBinaryFileName(),
                fileName,
//				verticesMapManager,
                seq.getNumTotalFrames(),
                _cancelMorphingSimulation);

        addPendingNonStopableExecutor(videoCreator);

        int delayBetweenFrames = 0;
        seq.doMorphing(im -> addImageToVideoCreator(videoCreator, im),
                ex -> processExceptionProducedDuringMorphing(_cancelMorphingSimulation, exceptionReference, ex),
                _cancelMorphingSimulation,
                null,
                () -> frameSize,
                delayBetweenFrames);
    }

    protected synchronized void processExceptionProducedDuringMorphing(AtomicBoolean cancelMorphingSimulation,
                                                                       AtomicReference<Exception> exceptionReference,
                                                                       Exception ex) {
        if(exceptionReference.get() == null) {
            exceptionReference.set(ex);
            SwingUtilities.invokeLater(() -> processExceptionProducedDuringMorphingInternal(cancelMorphingSimulation, ex));
        }
    }

    protected void processExceptionProducedDuringMorphingInternal(AtomicBoolean cancelMorphingSimulation,
                                                                  Exception ex) {
        int result = GenericFunctions.instance().getDialogsWrapper().showOptionDialog(this,
                createCustomInternationalString(CONF_ERROR_CREATING_MORPHING_FRAME, ex.getMessage()),
                "ERROR", DialogsWrapper.OK_CANCEL_OPTION, DialogsWrapper.QUESTION_MESSAGE,
                0,
                getInternationalString(CONF_IGNORE),
                getInternationalString(CONF_CANCEL));

        if(result == 1) {
            cancelMorphingSimulation.set(true);
        }
    }

    protected void addImageToVideoCreator(MorphingVideoCreator videoCreator,
                                          BufferedImage image) {
        videoCreator.addImage(image);
        SwingUtilities.invokeLater(() -> showImageFittingIt(image));
    }

    public GeometricRectangle createDelaunayRectangle(double xx, double yy, double width, double height) {
        return (GeometryCommonFactory.instance().createDelaunayRectangle(xx, yy, width, height));
    }

    protected GeometricRectangle getVideoFrameSize() {
        return (createDelaunayRectangle(0, 0, 1920, 1080));
    }

    protected void showImageFittingIt(BufferedImage image) {
        SwingUtilities.invokeLater(() -> showImageFittingItInternal(image));
    }

    protected void showImageFittingItInternal(BufferedImage image) {
//		safeMethodExecution( () -> ImageIO.write(image, "png", new File( "J:\\image.png" ) ) );
        LOGGER.info("New image: {}", DateFunctions.instance().formatDurationLong(System.currentTimeMillis()));
        _imageJPanel.setImageFittingIt(image);
    }
}
