/* 
 * Copyright (C) 2022 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/gpl-3.0.txt
 *
 */

package com.frojasg1.applications.morphing.videobuilder.video;

import com.frojasg1.applications.libmorphing.controller.MorphingVerticesMapManagerBase;
import com.frojasg1.general.ExecutionFunctions;
import com.frojasg1.general.FileFunctions;
import com.frojasg1.general.process.ProcExecResult;
import com.frojasg1.general.process.ProcessFunctions;
import com.frojasg1.general.threads.LockAndConditionOwner;
import java.awt.image.BufferedImage;
import java.io.File;
import java.util.LinkedList;
import java.util.concurrent.atomic.AtomicBoolean;
import javax.imageio.ImageIO;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 *
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 */
public class MorphingVideoCreator extends LockAndConditionOwner implements Runnable {
	private static final Logger LOGGER = LoggerFactory.getLogger(MorphingVideoCreator.class);
	protected static final int MAX_NUM_OF_STORED_ELEMS = 4;

	protected static final int NUM_CONDITIONS = 2;
	protected static final int WAITING_FOR_IMAGE = 0;
	protected static final int WAITING_FOR_ROOM = 1;

	protected LinkedList<BufferedImage> _list;

	protected AtomicBoolean _cancelMorphingVideo = null;

//	protected MorphingVerticesMapManagerBase _morphingVerticesManager;
	protected String _videoFileName;
	protected String _ffmpegBinaryFileName;
	protected String _imagesFolderName;

	protected int _numTotalFrames;
	protected int _currentFrame;

	protected boolean _hasToStop = false;

	public MorphingVideoCreator(String ffmpegBinaryFileName, String fileName,
//									MorphingVerticesMapManagerBase morphingVerticesManager,
									int numTotalFrames, AtomicBoolean cancelMorphingVideo)
	{
		_list = new LinkedList<>();
		_videoFileName = fileName;
		_ffmpegBinaryFileName = ffmpegBinaryFileName;
		_numTotalFrames = numTotalFrames;
		_cancelMorphingVideo = cancelMorphingVideo;

		init();
	}

	protected String getAbsoluteDirName( String longFileName )
	{
		return( FileFunctions.instance().getAbsoluteDirName( longFileName ) );
	}

	protected String convertFileNameToFfmpeg( String fileName )
	{
		return( fileName.replace( "\\", "/" ) );
	}

	protected String getImagesFolderName()
	{
		if( _imagesFolderName == null )
			_imagesFolderName = getAbsoluteDirName( getVideoFileName() ) +
				"/folderToBeErased";

		return( _imagesFolderName );
	}

	protected String getFrameFileName()
	{
		return( String.format( "%s/frame.%06d.png", getImagesFolderName(), _currentFrame ) );
	}

	protected void eraseRecursiveFileRegex( String folder, String fileNamePatternToErase )
	{
		ExecutionFunctions.instance().safeMethodExecution( () -> FileFunctions.instance().eraseRecursiveFileRegex(folder, fileNamePatternToErase ) );
	}

	protected void createImagesFolder()
	{
		String folderName = getImagesFolderName();
		if( FileFunctions.instance().isDirectory( folderName ) )
			eraseRecursiveFileRegex( folderName, ".*\\.png" );

		new File( folderName ).mkdirs();
	}

	protected void startRunning()
	{
		_currentFrame = 0;

		createImagesFolder();
	}

	public String getFfmpegBinaryFileName() {
		return _ffmpegBinaryFileName;
	}

	public String getFfplayBinaryFileName() {
		String ffDirName = getAbsoluteDirName( getFfmpegBinaryFileName() );
		return( String.format( "%s/ffplay", ffDirName ) );
	}

	public String getVideoFileName() {
		return _videoFileName;
	}

	@Override
	public void run()
	{
		startRunning();

		try
		{
			BufferedImage image;
			while( ! isHasToStop() )
			{
				image = getImage();
				processImage( image );
			}

			if( ! hasToCancel() )
			{
				if( wasASuccess( createVideo() ) )
					playVideo();
			}
		}
		finally
		{
			releaseResources();
		}
	}

	protected boolean wasASuccess( ProcExecResult procExecResult )
	{
		return( ProcessFunctions.instance().wasASuccess( procExecResult ) );
	}

//	protected long getFrameTime()
//	{
//		return( _framePeriod * _currentFrame );
//	}

	protected void processImage( BufferedImage frame )
	{
//		_writer.encodeVideo(_videoStreamIndex, frame, getFrameTime(), 
//					DEFAULT_TIME_UNIT);

		LOGGER.info( "Saving frame: {}", _currentFrame );
		ExecutionFunctions.instance().safeMethodExecution( () ->
			ImageIO.write(frame, "png", new File( getFrameFileName() ) )
		);
		_currentFrame++;
	}

	protected ProcExecResult execute( String ... commandAndParams )
	{
		return( ExecutionFunctions.instance().safeFunctionExecution( () -> 
			ProcessFunctions.instance().execute(commandAndParams) )
		);
	}

	protected String getFrameBlob()
	{
		return( String.format( "%s/frame.%%06d.png", getImagesFolderName() ) );
	}

	protected ProcExecResult createVideo()
	{
		// "ffmpeg -framerate 30 -pattern_type glob -i '*.png' -c:v libx264 -pix_fmt yuv420p out.mp4"
		ProcExecResult procExecResult = execute( getFfmpegBinaryFileName(),
			"-y",
			"-start_number", "0",
			"-i", convertFileNameToFfmpeg( getFrameBlob() ),
			"-c:v", "libx264",
			"-vf", "fps=25",
			"-frames:v", String.valueOf(_numTotalFrames),
			"-pix_fmt", "yuv420p",
			convertFileNameToFfmpeg( getVideoFileName() )
			);

		logProcessResult( procExecResult );

		return( procExecResult );
	}

	protected ProcExecResult playVideo()
	{
		// "ffmpeg -framerate 30 -pattern_type glob -i '*.png' -c:v libx264 -pix_fmt yuv420p out.mp4"
		ProcExecResult procExecResult = execute( getFfplayBinaryFileName(),
			getVideoFileName()
			);

		logProcessResult( procExecResult );

		return( procExecResult );
	}

	protected void logProcessResult( ProcExecResult procExecResult )
	{
		ProcessFunctions.instance().logProcessResult( procExecResult );
	}

	protected void releaseResources()
	{
		signalAll();
		
//		_writer.close();
	}

	protected boolean hasToCancel()
	{
		boolean result = false;

		if( _cancelMorphingVideo != null )
			result = _cancelMorphingVideo.get();

		return( result );
	}

	protected boolean isHasToStop()
	{
		return( ( _currentFrame >= _numTotalFrames ) || _hasToStop || hasToCancel() );
	}

	protected BufferedImage getImage()
	{
		return( lockFunction( this::getImageInternal ) );
	}

	protected BufferedImage getImageInternal()
	{
		BufferedImage result = null;
		while( !isHasToStop() && _list.isEmpty() && !isInterrupted() )
			awaitInterrupted(WAITING_FOR_IMAGE, 1000);

		if( !_list.isEmpty() )
		{
			result = _list.removeFirst();

			ifHasWaitersSignal(WAITING_FOR_ROOM);
		}
		else
			LOGGER.info( "Frame list empty" );

		return( result );
	}

	protected boolean canAdd()
	{
		return( _list.size() < MAX_NUM_OF_STORED_ELEMS );
//		return( true );
	}

	public void addImage( BufferedImage image )
	{
		this.lockProcedure( () -> addImageInternal( image ) );
	}

	protected void addImageInternal( BufferedImage image )
	{
		while( ! _hasToStop && ! canAdd() && !isInterrupted() )
			awaitInterrupted(WAITING_FOR_ROOM, null);
//			awaitInterrupted(WAITING_FOR_ROOM, 1000);

//		if( ! _hasToStop && canAdd() )
		if( ! _hasToStop && !isInterrupted() )
		{
			_list.add( image );
			signalCondition(WAITING_FOR_IMAGE);
		}
	}

	public void hasToStop()
	{
		_hasToStop = true;
		signalAll();
	}

	@Override
	protected int getNumConditionsForLock() {
		return( NUM_CONDITIONS );
	}
}
