/* 
 * Copyright (C) 2021 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/gpl-3.0.txt
 *
 */
package com.frojasg1.applications.libmorphing.api.facedetector.jersey;

import com.frojasg1.applications.libmorphing.api.facedetector.pojo.FaceDetection;
import com.frojasg1.applications.libmorphing.api.facedetector.FaceDetectionQuery;
import com.frojasg1.applications.libmorphing.api.facedetector.FaceDetectionQueryResult;
import com.frojasg1.applications.libmorphing.api.facedetector.pojo.ImageFaceDetection;
import com.frojasg1.libgeometry.types.GeometricRectangle;
import com.frojasg1.libgeometry.types.Point2d;
import com.frojasg1.general.desktop.queries.InetQueryException;
import com.frojasg1.general.exceptions.IllegalCallException;
import com.frojasg1.general.jersey.queries.InetQueryBaseWithResult;
import org.glassfish.jersey.media.multipart.FormDataContentDisposition;
import org.glassfish.jersey.media.multipart.FormDataMultiPart;
import org.glassfish.jersey.media.multipart.MultiPart;
import org.glassfish.jersey.media.multipart.MultiPartFeature;
import org.glassfish.jersey.media.multipart.file.FileDataBodyPart;
import org.json.JSONArray;
import org.json.JSONObject;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.ws.rs.client.Client;
import javax.ws.rs.client.Entity;
import javax.ws.rs.client.WebTarget;
import javax.ws.rs.core.Response;
import java.io.File;

/**
 *
 * @author fjavier.rojas
 */
public abstract class JerseyFaceDetectionInetQuery extends InetQueryBaseWithResult<ImageFaceDetection, FaceDetectionQueryResult>
								implements FaceDetectionQuery
{
	private static final Logger LOGGER = LoggerFactory.getLogger(JerseyFaceDetectionInetQuery.class);
	protected static final String RELATIVE_RESTFUL_URL = "restful2/faceDetector/";

	// usage:
	// curl -i -X POST https://frojasg1.com:8443/downloads_web/restful2/faceDetector/20221204.TextCrypt.v1.6.7z -F "file=@./fotos/Fran.202106.143x143.png" -o out.json

	protected static final String IMAGE_FILENAME = "file";

	@Override
	public void init( String uri, String path )
	{
		super.init( uri, path);
	}

	@Override
	protected FaceDetectionQueryResult createInetQueryResult()
	{
		return( new FaceDetectionQueryResult(null) );
	}

	@Override
	public abstract FaceDetectionQueryResult queryForApplication(String imageFileName) throws InetQueryException;

	@Override
	protected ImageFaceDetection convertInternalResultInternal(JSONObject jsonObject) {
		throw new IllegalCallException("Function call not allowed");
	}

	protected ImageFaceDetection createImageFaceDetection() {
		return new ImageFaceDetection();
	}

	@Override
	protected ImageFaceDetection convertInternalResultInternal(JSONArray jsonArray) {
		ImageFaceDetection result = createImageFaceDetection();
		if(jsonArray != null) {
			for(int ii=0; ii<jsonArray.length(); ii++) {
				result.addFaceDetection(createFaceDetection(jsonArray.getJSONObject(ii)));
			}
		}
		return result;
	}

	protected FaceDetection createFaceDetection(JSONObject jsonObject) {
		FaceDetection result = new FaceDetection();
		result.setBoundingBox(createDelaunayRectangle(jsonObject.getJSONArray("bbox")));
		JSONArray array = jsonObject.getJSONArray("points");
		for(int ii=0; ii<array.length(); ii++) {
			result.addPoint(createPoint2d(array.getJSONArray(ii)));
		}
		return result;
	}

	protected GeometricRectangle createDelaunayRectangle(JSONArray jsonArray) {
		return new GeometricRectangle(jsonArray.getInt(0), jsonArray.getInt(1), jsonArray.getInt(2),
				jsonArray.getInt(3));
	}

	protected Point2d createPoint2d(JSONArray jsonArray) {
		return new Point2d(jsonArray.getInt(0), jsonArray.getInt(1));
	}

	@Override
	public FaceDetectionQueryResult query(String imageFileName) throws InetQueryException
	{
		Client client = createClientBuilder().newClient().register(MultiPartFeature.class);
		WebTarget server = client.target(getUri()).path(getPath());

		File myFile = new File(imageFileName);
		FileDataBodyPart filePart = new FileDataBodyPart("file", myFile);
		filePart.setContentDisposition(FormDataContentDisposition.name("file").fileName(myFile.getName()).build());

		MultiPart multipartEntity = new FormDataMultiPart().bodyPart(filePart);

		Response result = server.request().post(Entity.entity(multipartEntity, multipartEntity.getMediaType()));

		LOGGER.info("detect_face Status: {}", result.getStatus());

		String responseJson = result.readEntity(String.class);
		LOGGER.info("result: {}", responseJson);

		return convert(responseJson);
	}
}
