/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.delaunay.triangulation.quadedge;

import com.frojasg1.libgeometry.algebra.helpers.GeometryHelper;
import com.frojasg1.libgeometry.types.Point2d;

/**
 *
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 */
// http://karlchenofhell.org/cppswp/lischinski.pdf
public class DelaunayQuadEdgeHelpers {
	protected static DelaunayQuadEdgeHelpers INSTANCE = new DelaunayQuadEdgeHelpers();

//	protected static final double EPS = 1.0d-13;
//	protected static final double EPS = 1.0e-6d;

	public static DelaunayQuadEdgeHelpers instance()
	{
		return( INSTANCE );
	}

	/*************** Geometric Predicates for Delaunay Diagrams *****************/
	public double TriArea(Point2d a, Point2d b, Point2d c)
	// Returns twice the area of the oriented triangle (a, b, c), i.e., the
	// area is positive if the triangle is oriented counterclockwise.
	{
		return (b.x - a.x)*(c.y - a.y) - (b.y - a.y)*(c.x - a.x);
	}

	public boolean InCircle(Point2d a, Point2d b, Point2d c, Point2d d)
	// Returns TRUE if the point d is inside the circle defined by the
	// points a, b, c. See Guibas and Stolfi (1985) p.107.
	{
	return (a.x*a.x + a.y*a.y) * TriArea(b, c, d) -
			(b.x*b.x + b.y*b.y) * TriArea(a, c, d) +
			(c.x*c.x + c.y*c.y) * TriArea(a, b, d) -
			(d.x*d.x + d.y*d.y) * TriArea(a, b, c) > 0;
	}

	public boolean ccw(Point2d a, Point2d b, Point2d c)
	// Returns TRUE if the points a, b, c are in a counterclockwise order
	{
		return (TriArea(a, b, c) > 0);
	}

	public boolean RightOf(Point2d x, Edge e)
	{
		return ccw(x, e.Dest2d(), e.Org2d());
	}

	public boolean LeftOf(Point2d x, Edge e)
	{
		return ccw(x, e.Org2d(), e.Dest2d());
	}

	public boolean OnEdge(Point2d x, Edge e)
	// A predicate that determines if the point x is on the edge e.
	// The point is considered on if it is in the EPS-neighborhood
	// of the edge.
	{
		boolean result = false;
		double t1, t2, t3;
//		t1 = x.subtract(e.Org2d()).norm();
//		t2 = x.subtract(e.Dest2d()).norm();
		t1 = x.squaredDistance(e.Org2d());
		t2 = x.squaredDistance(e.Dest2d());
		if ( isNearZero(t1) || isNearZero(t2) )
			result = true;
		else
		{
//			t3 = e.Org2d().subtract(e.Dest2d()).norm();
			t3 = e.Org2d().squaredDistance(e.Dest2d());
			if (t1 > t3 || t2 > t3)
				result = false;
			else
				result = isNearZero(
						getDistanceBetweenSegmentAndPoint( e.Org2d(),
														e.Dest2d(), x ) );
		}

		return( result );
	}

	protected boolean isNearZero( double value )
	{
		return getGeometryHelper().isNearZero(value);
	}
//
//	protected boolean areNear(Point2d pt1, Point2d pt2) {
//		return isNearZero(pt1.x - pt2.x) && isNearZero(pt1.y - pt2.y);
//	}

	protected GeometryHelper getGeometryHelper() {
		return GeometryHelper.instance();
	}

	public double getDistanceBetweenSegmentAndPoint(Point2d pt1, Point2d pt2, Point2d pt )
	{
		return getGeometryHelper().getSquaredDistanceBetweenLineAndPoint(pt1, pt2, pt);
	}

//	public double getDistanceBetweenLineAndPoint( double A, double B, double C, Point2d pt )
//	{
//		return( ( A * pt.x + B * pt.y + C ) / ( A * A + B * B ) );
//	}

	public boolean isInsideSector( Point2d vertex, Point2d firstClockWise,
									Point2d secondClockWise, Point2d pt )
	{
		return ( TriArea(vertex, firstClockWise, pt) <= 0 ) &&
				( TriArea(vertex, pt, secondClockWise) <= 0 );// ||
//				isNearZero(getDistanceBetweenLineAndPoint( vertex, firstClockWise, pt ) ) ||
//				isNearZero(getDistanceBetweenLineAndPoint( vertex, secondClockWise, pt ) ) ||
//				isNearZero(vertex.squaredDistance(pt)) ||
//				isNearZero(firstClockWise.squaredDistance(pt)) ||
//				isNearZero(secondClockWise.squaredDistance(pt));
	}
}
