/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.delaunay.triangulation.quadedge;

import com.frojasg1.libgeometry.algebra.helpers.GeometryHelper;
import com.frojasg1.libgeometry.algebra.helpers.VectorFunctions;
import com.frojasg1.libgeometry.types.GeometricRectangle;
import com.frojasg1.libgeometry.types.Point2d;
import com.frojasg1.delaunay.triangulation.quadedge.builder.QuadEdgeDelaunayBuilder;
import com.frojasg1.general.changestate.ChangeableState;
import com.frojasg1.general.changestate.staterevertible.AtomicReversionList;
import java.util.ArrayList;
import java.util.List;
import java.util.Objects;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 *
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 */
// http://karlchenofhell.org/cppswp/lischinski.pdf
public class DelaunaySubdivision implements ChangeableState<DelaunaySubdivision>
{
	private static final Logger LOGGER = LoggerFactory.getLogger(DelaunaySubdivision.class);

	protected static final double BOUNDARY_DELTA = 0.05d;

	protected Edge _startingEdge;

	protected QuadEdgeDelaunayBuilder _builder;

	protected AtomicReversionList _lastInsertionReversionList;

	protected GeometricRectangle bounds;

	protected Point2d deltaForBoundary;

	/************* Topological Operations for Delaunay Diagrams *****************/
	// constructor only for state copy
	public DelaunaySubdivision()
	{
	}

	public DelaunaySubdivision(GeometricRectangle bounds,
                               QuadEdgeDelaunayBuilder builder)
	// Initialize a subdivision to the triangle defined by the points a, b, c.
	{
		this.bounds = bounds;
		_lastInsertionReversionList = new AtomicReversionList();
		_builder = builder;
		deltaForBoundary = _builder.createPoint(0, 0);

		double xx = bounds.x;
		double maxX = xx + bounds.width - 1;
		double yy = bounds.y;
		double maxY = yy + bounds.height - 1;

		// cw
//		Point2d da = builder.createPoint(xx, yy);
//		Point2d db = builder.createPoint(xx, maxY);
//		Point2d dc = builder.createPoint(maxX, maxY);
//		Point2d dd = builder.createPoint(maxX, yy);

		///////////////////////////////////////////////
		// quadedge_topology_for_initial_rectangle.jpg
		///////////////////////////////////////////////
		// ccw
		Point2d da = builder.createPoint(xx, yy);
		Point2d db = builder.createPoint(maxX, yy);
		Point2d dc = builder.createPoint(maxX, maxY);
		Point2d dd = builder.createPoint(xx, maxY);

		Edge ea = MakeEdge();
		ea.EndPoints(da, db);

		Edge eb = MakeEdge();
		eb.EndPoints(db, dc);

		Edge ec = MakeEdge();
		ec.EndPoints(dc, dd);

		Edge ed = MakeEdge();
		ed.EndPoints(dd, da);

		Edge ee = MakeEdge();
		ee.EndPoints(da, dc);

		// vertices
		// ea
		ea.setNext(ee);
		ea.Sym().setNext(eb);

		// eb
		eb.setNext(ea.Sym());
		eb.Sym().setNext(ec);

		// ec
		ec.setNext(ee.Sym());
		ec.Sym().setNext(ed);

		// ed
		ed.setNext(ec.Sym());
		ed.Sym().setNext(ea);

		// ee
		ee.setNext(ed.Sym());
		ee.Sym().setNext(eb.Sym());


		// edges
		// ea
		ea.Rot().setNext(ed.Rot());
		ea.invRot().setNext(eb.invRot());

		// eb
		eb.Rot().setNext(ea.Rot());
		eb.invRot().setNext(ee.Rot());

		// ec
		ec.Rot().setNext(eb.Rot());
		ec.invRot().setNext(ed.invRot());

		// ed
		ed.Rot().setNext(ec.Rot());
		ed.invRot().setNext(ee.invRot());

		// ee
		ee.Rot().setNext(ea.invRot());
		ee.invRot().setNext(ec.invRot());

//
//		// edges
//		// ea
//		ea.invRot().setNext(eb.invRot());
//		ea.Rot().setNext(eb.Rot());
//
//		// eb
//		eb.invRot().setNext(ee.Rot());
//		eb.Rot().setNext(ec.Rot());
//
//		// ec
//		ec.invRot().setNext(ed.invRot());
//		ec.Rot().setNext(ed.Rot());
//
//		// ed
//		ed.invRot().setNext(ee.invRot());
//		ed.Rot().setNext(ea.Rot());
//
//		// ee
//		ee.invRot().setNext(ec.invRot());
//		ee.Rot().setNext(ea.invRot());

		_startingEdge = ea;
	}

	public DelaunaySubdivision(GeometricRectangle bounds,
                               QuadEdgeDelaunayBuilder builder, Point2d a, Point2d b, Point2d c)
	// Initialize a subdivision to the triangle defined by the points a, b, c.
	{
		this.bounds = bounds;
		_lastInsertionReversionList = new AtomicReversionList();
		_builder = builder;
		deltaForBoundary = _builder.createPoint(0, 0);

		Point2d da = builder.createPoint(a);
		Point2d db = builder.createPoint(b);
		Point2d dc = builder.createPoint(c);

		Edge ea = MakeEdge();
		ea.EndPoints(da, db);

		Edge eb = MakeEdge();
		Splice(ea.Sym(), eb);
		eb.EndPoints(db, dc);
		Edge ec = MakeEdge();
		Splice(eb.Sym(), ec);
		ec.EndPoints(dc, da);
		Splice(ec.Sym(), ea);
		_startingEdge = ea;
	}

	public GeometricRectangle getBounds() {
		return bounds;
	}

	protected AtomicReversionList getLastInsertionReversionList() {
		return _lastInsertionReversionList;
	}

	public void clearReversion()
	{
		getLastInsertionReversionList().clear();
	}

	protected <C extends ChangeableState<C>> void addElemToRevert( C elem )
	{
		getLastInsertionReversionList().add(elem);
	}

	public boolean canRevert()
	{
		return( !getLastInsertionReversionList().isEmpty() );
	}

	public void revertLastInsertion()
	{
		getLastInsertionReversionList().revert();
		clearReversion();
	}

	public Edge getStartingEdge()
	{
		return( _startingEdge );
	}

	protected Edge Connect(Edge a, Edge b)
	// Add a new edge e connecting the destination of a to the
	// origin of b, in such a way that all three have the same
	// left face after the connection is complete.
	// Additionally, the data pointers of the new edge are set.
	{
		Edge e = MakeEdge();
		Splice(e, a.Lnext());
		Splice(e.Sym(), b);
		e.EndPoints(a.Dest(), b.Org());

		return e;
	}

	protected Edge InsertRectangleDiagonal(Edge a, Edge b, Edge c, Edge d)
	// Add a new edge e connecting the destination of a to the
	// origin of b, in such a way that all three have the same
	// left face after the connection is complete.
	// Additionally, the data pointers of the new edge are set.
	{
		Edge e = MakeEdge();
		e.EndPoints(b.Dest(), a.Org());

		Splice(e, b.Lnext());
		Splice(e.Sym(), a);

		Splice(e.Sym(), d.Lnext());
		Splice(e, c);

		return e;
	}

	protected void Swap(Edge e)
	// Essentially turns edge e counterclockwise inside its enclosing
	// quadrilateral. The data pointers are modified accordingly.
	{
		Edge a = e.Oprev();
		Edge b = e.Sym().Oprev();
		Splice(e, a);
		Splice(e.Sym(), b);
		Splice(e, a.Lnext());
		Splice(e.Sym(), b.Lnext());

		getLastInsertionReversionList().add(e);
		getLastInsertionReversionList().add(e.Sym());
		e.EndPoints(a.Dest(), b.Dest());
	}

	public boolean RightOf(Point2d x, Edge e)
	{
		return( DelaunayQuadEdgeHelpers.instance().RightOf(x, e) );
	}

	// in order to avoid TriArea be zero and the system to collapse under points on the boundary.
	protected Point2d addBoundaryDelta(Point2d x) {
		double deltax = boundaryDelta(x.x, bounds.x, bounds.width);
		double deltay = boundaryDelta(x.y, bounds.y, bounds.height);
		// corners do not apply deltas
		if((deltax != 0) && (deltay != 0)) {
			deltax = 0;
			deltay = 0;
		}
		deltaForBoundary.x = x.x + deltax;
		deltaForBoundary.y = x.y + deltay;

		return deltaForBoundary;
//		return x;
	}

	// we assume that coordinates are ints, although they are double
	protected double boundaryDelta(double value, double lowerBound, double length) {
		double result = 0;
		if(value == lowerBound) {	// lower boundary
			result = BOUNDARY_DELTA;
		} else if(value == lowerBound + length - 1) {	// upper boundary
			result = -BOUNDARY_DELTA;
		}
		return result;
	}

	protected VectorFunctions getVectorFunctions() {
		return VectorFunctions.instance();
	}

	protected GeometryHelper getGeometryHelper() {
		return GeometryHelper.instance();
	}

	protected Point2d limit(Point2d pt, GeometricRectangle rectangle) {
		return getGeometryHelper().limit(pt, rectangle);
	}

	protected boolean anyVertexMatch(Point2d x, Edge e) {
		return x.equals( e.Org2d() ) || x.equals( e.Dest2d() );
	}

	protected double getAppliedDelta(double value, double start, double length) {
		double result = 0d;
		if(value == (start + BOUNDARY_DELTA)) {
			result = BOUNDARY_DELTA;
		} else if(value == (start + length - 1 - BOUNDARY_DELTA)) {
			result = -BOUNDARY_DELTA;
		}
		return result;
	}

	public Point2d removeBoundaryDelta(Point2d x) {
		Point2d result = x;
		if((x != null) && !x.isInt()) {
			GeometricRectangle bounds = getBounds();
			result = x.createCopy();
			result.x -= getAppliedDelta(x.x, bounds.x, bounds.width);
			result.y -= getAppliedDelta(x.y, bounds.y, bounds.height);
		}
		return result;
	}


	/************* An Incremental Algorithm for the Construction of *************/
	/************************ Delaunay Diagrams *********************************/
	protected Edge Locate(Point2d rawX)
	// Returns an edge e, s.t. either x is on e, or e is an edge of
	// a triangle containing x. The search starts from startingEdge
	// and proceeds in the general direction of x. Based on the
	// pseudocode in Guibas and Stolfi (1985) p.121.
	{
		Point2d x = limit(rawX, getBounds());

		Edge e = _startingEdge;
		Edge oNextEdge;
		Edge dPrevEdge;
		Point2d xx = addBoundaryDelta(x);
		while (true) {

			oNextEdge = e.Onext();
			dPrevEdge = e.Dprev();
			if (anyVertexMatch(x, e))
				return e;
			else if (anyVertexMatch(x, oNextEdge))
				return oNextEdge;
			else if (anyVertexMatch(x, dPrevEdge))
				return dPrevEdge;
			else if (RightOf(xx, e))
				e = e.Sym();
			else if (!RightOf(xx, oNextEdge))
				e = oNextEdge;
			else if (!RightOf(xx, dPrevEdge))
				e = dPrevEdge;
			else {
				if(OnEdge(x, oNextEdge))
					return whichSymsInside(x, oNextEdge);
				else if(OnEdge(x, dPrevEdge))
					return whichSymsInside(x, dPrevEdge);
				else
					return e;
			}
		}
	}

	protected Edge whichSymsInside(Point2d x, Edge e) {
		Edge result = null;
		if(vertexCanBeInside(x, e) || !vertexCanBeInside(x, e.Sym())) {
			result = e;
		} else {
			result = e.Sym();
		}
		return result;
	}

	protected boolean vertexCanBeInside(Point2d x, Edge e) {
		Point2d xx = addBoundaryDelta(x);
		Edge oNextEdge = e.Onext();
		Edge dPrevEdge = e.Dprev();
		return anyVertexMatch(x, e) || anyVertexMatch(x, oNextEdge) || anyVertexMatch(x, dPrevEdge)
				|| !RightOf(xx, e) && RightOf(xx, oNextEdge) && RightOf(xx, dPrevEdge);
	}

	protected Point2d addRawX(List<Point2d> list, Point2d x) {
		Point2d rawX = removeBoundaryDelta(x);
		list.add(rawX);

		return rawX;
	}

	// it returns the nearest point first, and then the umbrella of vertices clockwise
	public List<Point2d> getPossibleTriangles( Point2d rawX )
	{
		List<Point2d> result = new ArrayList<>();

		Point2d x = createPointWithBounderyDelta(rawX);

		Edge e;
		Edge center = LocateOrigin(x);

		addRawX(result, center.Org2d());
		addRawX(result, center.Dest2d() );
		e = center.Onext();
		while( e != center )
		{
			addRawX(result, e.Dest2d());
			e = e.Onext();
		} 

		return( result );
	}

	protected Edge LocateOrigin( Point2d x )
	{
		Edge center = Locate(x);
		double t1 = x.squaredDistance(center.Org2d());
		double t2 = x.squaredDistance(center.Dest2d());
		if( t1 > t2 )
			center = center.Sym();

		return( center );
	}

	// it returns the nearest point first, and then the two other vertices of the triangle clockwise
	protected List<Point2d> LocateTriangle( Point2d rawX )
	{
		List<Point2d> result = new ArrayList<>();

		Point2d x = createPointWithBounderyDelta(rawX);

		Edge center = LocateOrigin(x);

		addRawX(result, center.Org2d() );
		Point2d second = center.Dest();
		Point2d third = center.Onext().Dest();
		Point2d fourth = center.Oprev().Dest();
		if( isInsideSector( result.get(0), second, third, x ) )
		{
			addRawX(result, second );
			addRawX(result, third );
		}
		else if( isInsideSector( result.get(0), fourth, third, x ) )
		{
			addRawX(result, fourth );
			addRawX(result, third );
		}
		else
			throw new RuntimeException( "Triangle not found for point: " + x );

		return( result );
	}

	public boolean isInsideSector( Point2d vertex, Point2d firstClockWise,
									Point2d secondClockWise, Point2d pt )
	{
		return( DelaunayQuadEdgeHelpers.instance().isInsideSector(vertex, firstClockWise,
																secondClockWise, pt) );
	}

	protected boolean equals( Point2d pt1, Point2d pt2 )
	{
		return( Objects.equals( pt1, pt2 ) );
	}

	public boolean OnEdge(Point2d x, Edge e)
	{
		return( DelaunayQuadEdgeHelpers.instance().OnEdge(x, e) );
	}

	public boolean InCircle(Point2d a, Point2d b, Point2d c, Point2d d)
	{
		return( DelaunayQuadEdgeHelpers.instance().InCircle(a, b, c, d) );
	}

	public QuadEdgeDelaunayBuilder getBuilder() {
		return _builder;
	}

	protected Point2d createPoint2d( Point2d that )
	{
		return( getBuilder().createPoint(that) );
	}

	public boolean InsertSite(Point2d x)
	// Inserts a new point into a subdivision representing a Delaunay
	// triangulation, and fixes the affected edges so that the result
	// is still a Delaunay triangulation. This is based on the
	// pseudocode from Guibas and Stolfi (1985) p.120, with slight
	// modifications and a bug fix.
	{
		clearReversion();
		return( InsertSiteNotClearingReversion(x) );
	}

	protected Point2d createPointWithBounderyDelta(Point2d rawX) {
		Point2d result = addBoundaryDelta(limit(rawX, getBounds()));
		if(result.equals(rawX)) {
			result = rawX;
		} else {
			result = result.createCopy();
		}
		return result;
	}

	public boolean InsertSiteNotClearingReversion(Point2d rawX)
	// Inserts a new point into a subdivision representing a Delaunay
	// triangulation, and fixes the affected edges so that the result
	// is still a Delaunay triangulation. This is based on the
	// pseudocode from Guibas and Stolfi (1985) p.120, with slight
	// modifications and a bug fix.
	{
//		clearReversion();
		getLastInsertionReversionList().add(this);

		Point2d x = createPointWithBounderyDelta(rawX);

		Edge e = Locate(x);
		if (equals(x, e.Org2d()) || equals(x, e.Dest2d())) // point is already in
			return false;
		else if (OnEdge(x, e)) {
			e = e.Oprev();
			DeleteEdge(e.Onext());
		}
		// Connect the new point to the vertices of the containing
		// triangle (or quadrilateral, if the new point fell on an
		// existing edge.)
		Edge base = MakeEdge();
		base.EndPoints(e.Org(), createPoint2d(x));

		return InsertSiteKnowigEdge(x, base, e);
	}

	protected boolean InsertSiteKnowigEdge(Point2d x, Edge newEdge, Edge existingEdge) {
		Edge e = existingEdge;
		Edge base = newEdge;
		Splice(base, e);
		_startingEdge = base;
		do {
			base = Connect(e, base.Sym());
			e = base.Oprev();
		} while (e.Lnext() != _startingEdge);

		// Examine suspect edges to ensure that the Delaunay condition
		// is satisfied.
		do {
			Edge t = e.Oprev();
			if (RightOf(t.Dest2d(), e) &&
				InCircle(e.Org2d(), t.Dest2d(), e.Dest2d(), x)) {
				Swap(e);
				e = e.Oprev();
			}
			else if (e.Onext() == _startingEdge) // no more suspect edges
				return true;
			else // pop a suspect edge
				e = e.Onext().Lprev();
		} while (true);
	}

	// this functions is obsolete with version v1.1,
	// as now the triangulation is inside a rectangle
	// not inside a triangle, as it was in v1.0
	// as the triangulation is inside the exact rectangle, no new edge will intersect the boundaries
	@Deprecated
	public void formalizeBoundsOfLatestInsertion(GeometricRectangle bounds, List<Point2d> result )
	{
		Edge oldStartingEdge = _startingEdge;

		// we place at the just inserted vertex, and browse all edges round.
		Edge current = oldStartingEdge.Sym();
		Edge to = current;
		do
		{
			if( this.getBuilder().intersectsWithBoundary( current, bounds ) )
			{
				Point2d intx = getBuilder().getIntersectionWithBoundary(bounds, current.Org(), current.Dest());
				getBuilder().round(intx);
				if( ( intx != null ) && !intx.equals(current.Org()) && !intx.equals(current.Dest()))
				{
					if( InsertSiteNotClearingReversion(intx) )
					{
						if( result != null )
							addRawX(result, intx);

						_startingEdge = oldStartingEdge;
						formalizeBoundsOfLatestInsertion( bounds, result );
						break;
					}
				}

				LOGGER.error( "Error inserting intersection with boundary: {}", intx );
			}
			current = current.Onext();
		}
		while( current != to );

//		clearReversion();
	}

/*****************************************************************************/

	/*********************** Basic Topological Operators ************************/
	public Edge MakeEdge()
	{
		QuadEdge ql = new QuadEdge();
		return ql.get(0);
	}

	public void Splice(Edge a, Edge b)
	// This operator affects the two edge rings around the origins of a and b,
	// and, independently, the two edge rings around the left faces of a and b.
	// In each case, (i) if the two rings are distinct, Splice will combine
	// them into one; (ii) if the two are the same ring, Splice will break it
	// into two separate pieces.
	// Thus, Splice can be used both to attach the two edges together, and
	// to break them apart. See Guibas and Stolfi (1985) p.96 for more details
	// and illustrations.
	{
		Edge alpha = a.Onext().Rot();
		Edge beta = b.Onext().Rot();

		getLastInsertionReversionList().add(a);
		getLastInsertionReversionList().add(b);
		getLastInsertionReversionList().add(alpha);
		getLastInsertionReversionList().add(beta);

		Edge t1 = b.Onext();
		Edge t2 = a.Onext();
		Edge t3 = beta.Onext();
		Edge t4 = alpha.Onext();
		a.setNext( t1 );
		b.setNext( t2 );
		alpha.setNext( t3 );
		beta.setNext( t4 );
	}

	public void DeleteEdge(Edge e)
	{
		Splice(e, e.Oprev());
		Splice(e.Sym(), e.Sym().Oprev());

		getLastInsertionReversionList().add(e.Qedge());
//		e.Qedge().release();
	}

	@Override
	public void changeState(DelaunaySubdivision that) {
		_startingEdge = that._startingEdge;
	}

	@Override
	public DelaunaySubdivision copyState() {
		DelaunaySubdivision result = new DelaunaySubdivision();
		result._startingEdge = _startingEdge;

		return( result );
	}
}
