/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.delaunay.triangulation.quadedge;

import com.frojasg1.libgeometry.types.Point2d;
import com.frojasg1.general.changestate.ChangeableState;

/**
 *
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 */
// http://karlchenofhell.org/cppswp/lischinski.pdf
public class Edge implements ChangeableState<Edge>
{
//friend QuadEdge;
//friend void Splice(Edge*, Edge*);
	protected int _num;
	protected Edge _next;
	protected Point2d _data;

	protected EdgeArray _array;

	protected QuadEdge _quadEdge;

	public Edge(QuadEdge quadEdge)
	{
		_quadEdge = quadEdge;
		_data = null;
	}

	protected Edge getRelativeElem( int relativeIndex )
	{
		return( _array.get( _num + relativeIndex ) );
	}

	// Return the dual of the current edge, directed from its right to its left.
	public Edge Rot()
	{
		return( getRelativeElem( ( _num < 3 ) ? 1 : -3 ) );
	}

	// Return the dual of the current edge, directed from its left to its right.
	public Edge invRot()
	{
		return( getRelativeElem( ( _num > 0 ) ? -1 : 3 ) );
	}

	// Return the edge from the destination to the origin of the current edge.
	public Edge Sym()
	{
		return( getRelativeElem( ( _num < 2 ) ? 2 : -2 ) );
	}

	// Return the next ccw edge around (from) the origin of the current edge.
	public Edge Onext()
	{
		return( _next );
	}

	// Return the next cw edge around (from) the origin of the current edge.
	public Edge Oprev()
	{
		return( Rot().Onext().Rot() );
	}

	// Return the next ccw edge around (into) the destination of the current edge.
	public Edge Dnext()
	{
		return( Sym().Onext().Sym() );
	}

	// Return the next cw edge around (into) the destination of the current edge.
	public Edge Dprev()
	{
		return( invRot().Onext().invRot() );
	}

	// Return the ccw edge around the left face following the current edge.
	public Edge Lnext()
	{
		return( invRot().Onext().Rot() );
	}

	// Return the ccw edge around the left face before the current edge.
	public Edge Lprev()
	{
		return( Onext().Sym() );
	}

	// Return the edge around the right face ccw following the current edge.
	public Edge Rnext()
	{
		return( Rot().Onext().invRot() );
	}

	// Return the edge around the right face ccw before the current edge.
	public Edge Rprev()
	{
		return( Sym().Onext() );
	}

	public Point2d Org()
	{
		return( _data );
	}

	public Point2d Dest()
	{
		return( Sym()._data );
	}

	public Point2d Org2d()
	{
		return( _data );
	}

	public Point2d Dest2d()
	{
		return( Sym()._data );
	}

	public void EndPoints(Point2d or, Point2d de)
	{
		_data = or;
		Sym()._data = de;
	}

	public QuadEdge Qedge()
	{
		return( _quadEdge );
	}

	public void release()
	{
		_quadEdge = null;
		_array = null;
	}

	public int getNum() {
		return _num;
	}

	public void setNum(int _num) {
		this._num = _num;
	}

	public Edge getNext() {
		return _next;
	}

	public void setNext(Edge _next) {
		this._next = _next;
	}

	public EdgeArray getArray() {
		return _array;
	}

	public void setArray(EdgeArray _array) {
		this._array = _array;
	}

	protected void setEmptyArray()
	{
		setArray( new EdgeArray() );
	}

	@Override
	public void changeState(Edge that)
	{
		copyInternal(that, this);
	}

	protected Edge createEmptyCopy()
	{
		Edge result = new Edge(null);
		result.setEmptyArray();

		return( result );
	}

	protected void copyInternal( Edge from, Edge to )
	{
		to._num = from._num;
		to._next = from._next;
		to._data = from._data;
		to._array.changeState( from._array );
		to._quadEdge = from._quadEdge;
	}

	@Override
	public Edge copyState()
	{
		Edge result = createEmptyCopy();
		result.changeState(this);

		return( result );
	}

	@Override
	public String toString() {
		return String.format("Edge{ %s, %s }", _data, Sym()._data);
	}

}
