/*
 * Copyright (c) 2023. Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 3.0 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You may obtain a copy of the License at
 *
 *       http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */

/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package com.frojasg1.delaunay.triangulation.quadedge.builder.impl;

import com.frojasg1.libgeometry.types.GeometricRectangle;
import com.frojasg1.delaunay.triangulation.quadedge.DelaunaySubdivision;
import com.frojasg1.delaunay.triangulation.quadedge.Edge;
import com.frojasg1.delaunay.triangulation.quadedge.EdgeView;
import com.frojasg1.libgeometry.types.Point2d;
import com.frojasg1.delaunay.triangulation.quadedge.builder.QuadEdgeDelaunayBuilder;
import com.frojasg1.libgeometry.types.factory.GeometryCommonFactory;
import java.util.HashSet;
import java.util.Set;
import java.util.function.BiConsumer;
import java.util.function.BiFunction;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 *
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 */
public class QuadEdgeDelaunayBuilderImpl implements QuadEdgeDelaunayBuilder
{
	private static final Logger LOGGER = LoggerFactory.getLogger(QuadEdgeDelaunayBuilderImpl.class);

	protected static final QuadEdgeDelaunayBuilderImpl INSTANCE = new QuadEdgeDelaunayBuilderImpl();

	public static QuadEdgeDelaunayBuilderImpl instance()
	{
		return( INSTANCE );
	}

	public GeometricRectangle createRectangle(double x, double y, double width, double height )
	{
		return( GeometryCommonFactory.instance().createDelaunayRectangle( x, y, width, height ) );
	}

	public DelaunaySubdivision createDelaunaySubdivision(GeometricRectangle bounds)
	{
		return createDelaunaySubdivisionBetter(bounds);
	}

	@Deprecated
	protected DelaunaySubdivision createDelaunaySubdivisionInternal(GeometricRectangle bounds)
	{
		double topx = bounds.width;
		double topy = bounds.height;
		return createDelaunaySubdivision(bounds, topx, topy);
	}

	public DelaunaySubdivision createDelaunaySubdivisionBetter(GeometricRectangle bounds)
	{
		return new DelaunaySubdivision( bounds, this);
	}

	public DelaunaySubdivision createDelaunaySubdivision(GeometricRectangle bounds, double topx, double topy)
	{
		DelaunaySubdivision result = null;

		// Vertices of the rectangle
		Point2d ul = createPoint( bounds.x, bounds.y );
		Point2d dl = createPoint( bounds.x, bounds.y + topy );
		Point2d ur = createPoint( bounds.x + topx, bounds.y );
		Point2d dr = createPoint( bounds.x + topx, bounds.y + topy );

		// we calculate the triangle that contains the rectangle ( 0, 0, width, height )
		double x1 = bounds.x + topx / 2d;
		double y1 = bounds.y - topy;
		Point2d pt1 = createPoint( x1, y1 );

		double y2 = bounds.y + topy * 3 / 2d;
//		double x2 = calculateX( pt1, ul, y2 );
		double x2 = bounds.x - topx * 3 / 2d;
		Point2d pt2 = createPoint( x2, y2 );

		double y3 = bounds.y + topy * 3 / 2d;
//		double x3 = calculateX( pt1, ur, y3 );
		double x3 = bounds.x + topx * 5 / 2d;
		Point2d pt3 = createPoint( x3, y3 );

		result = new DelaunaySubdivision( bounds, this, pt1, pt2, pt3 );
/*
		LOGGER.info( "pt1: {}", pt1 );
		LOGGER.info( "pt2: {}", pt2 );
		LOGGER.info( "pt3: {}", pt3 );
		LOGGER.info( "pt1: {}", pt1 );

		LOGGER.info( "ul: {}", ul );
		LOGGER.info( "dl: {}", dl );
		LOGGER.info( "ur: {}", ur );
		LOGGER.info( "dr: {}", dr );
		LOGGER.info( "ul: {}", ul );
*/
		// we insert the vertices
		result.InsertSite(ul);
		result.InsertSite(dl);
		result.InsertSite(ur);
		result.InsertSite(dr);

		result.clearReversion();

		return( result );
	}

	public void browseEdges(DelaunaySubdivision delaunaySubdivision,
							GeometricRectangle bounds,
							BiConsumer<Point2d, Point2d> edgeConsumer )
	{
		Edge edge = delaunaySubdivision.getStartingEdge();
		Set processed = new HashSet<>();

		browseEdgesInternal(delaunaySubdivision, bounds, processed, edge, edgeConsumer );
	}

	protected void browseEdgesInternal(DelaunaySubdivision delaunaySubdivision,
									   GeometricRectangle bounds,
									   Set processed,
									   Edge edge,
									   BiConsumer<Point2d, Point2d> edgeConsumer )
	{
		if( !processed.contains( edge ) )
		{
			processed.add( edge );

			EdgeView ev = new EdgeView(edge.Org(), edge.Dest());
			if( !processed.contains(ev) && isInside( edge, bounds ) )
			{
				edgeConsumer.accept( delaunaySubdivision.removeBoundaryDelta(edge.Org()),
						delaunaySubdivision.removeBoundaryDelta(edge.Dest()) );
				processed.add( ev );
			}

			browseEdgesInternal(delaunaySubdivision, bounds, processed, edge.Onext(), edgeConsumer);
			browseEdgesInternal(delaunaySubdivision, bounds, processed, edge.Sym(), edgeConsumer);
		}
	}

	protected boolean isInside( Edge edge, GeometricRectangle bounds )
	{
		return( bounds.isStrictlyInside( edge.Org() ) || bounds.isStrictlyInside( edge.Dest() ) ||
				bounds.isInside( edge.Org() ) && bounds.isInside( edge.Dest() ) );
	}

	@Override
	public boolean intersectsWithBoundary( Edge edge, GeometricRectangle bounds )
	{
		return( ( getOutsidePt( bounds, edge.Org(), edge.Dest() ) != null ) && 
				( getStrictlyInsidePt( bounds, edge.Org(), edge.Dest() ) != null ) );
	}

	// equation of the straight line:   A * x + B * y = C
	// the line: (x1, y1) - (x2, y2)
	// we calculate the x0 that corresponds to y0
	protected double calculateX( Point2d pt1, Point2d pt2, double y0 )
	{
		double A = pt2.y - pt1.y;
		double B = pt1.x - pt2.x;
		double C = A * pt1.x + B * pt1.y;

		double x0 = ( C - B * y0 ) / A;

		return( x0 );
	}

	// equation of the straight line:   A * x + B * y = C
	// the line: (x1, y1) - (x2, y2)
	// we calculate the x0 that corresponds to y0
	protected double calculateY( Point2d pt1, Point2d pt2, double x0 )
	{
		double A = pt2.y - pt1.y;
		double B = pt1.x - pt2.x;
		double C = A * pt1.x + B * pt1.y;

		double y0 = ( C - A * x0 ) / B;

		return( y0 );
	}

	@Override
	public Point2d createPoint(double xx, double yy) {
		return( GeometryCommonFactory.instance().createPoint2d( xx, yy ) );
	}

	@Override
	public Point2d createPoint( Point2d that ) {
		return( GeometryCommonFactory.instance().createPoint2d( that ) );
	}

	protected Point2d getIfOutside( GeometricRectangle bounds,
									Point2d pt )
	{
		return( ( bounds.isInside(pt) ? null : pt ) );
	}

	protected Point2d getIfStrictlyInside( GeometricRectangle bounds,
											Point2d pt )
	{
		return( ( bounds.isStrictlyInside(pt) ? pt : null ) );
	}

	protected Point2d getIfPtWithCriterium( GeometricRectangle bounds,
											BiFunction<GeometricRectangle, Point2d, Point2d> selector,
											Point2d pt1,
											Point2d pt2 )
	{
		Point2d result = selector.apply( bounds, pt1 );
		if( result == null )
			result = selector.apply( bounds, pt2 );

		return( result );
	}

	protected Point2d getOutsidePt( GeometricRectangle bounds,
									Point2d pt1,
									Point2d pt2 )
	{
		return( getIfPtWithCriterium( bounds, this::getIfOutside, pt1, pt2 ) );
	}

	protected Point2d getStrictlyInsidePt( GeometricRectangle bounds,
											Point2d pt1,
											Point2d pt2 )
	{
		return( getIfPtWithCriterium( bounds, this::getIfStrictlyInside, pt1, pt2 ) );
	}

	protected boolean isNorth( GeometricRectangle bounds,
								Point2d pt )
	{
		return( pt.y < bounds.y );
	}

	protected boolean isWest( GeometricRectangle bounds,
								Point2d pt )
	{
		return( pt.x < bounds.x );
	}

	protected Point2d getHorizontalIntersection( GeometricRectangle bounds,
												Point2d outsidePt,
												Point2d strictlyInsidePt)
	{
		Point2d result = createPoint(0, 0);

		boolean isNorth = isNorth( bounds, outsidePt );

		if( isNorth )
			result.y = bounds.y;
		else
			result.y = bounds.y + bounds.height;

		result.x = this.calculateX(outsidePt, strictlyInsidePt, result.y);

		if( ! bounds.isInside(result) )
			result = null;

		return( result );
	}

	protected Point2d getVerticalIntersection( GeometricRectangle bounds,
												Point2d outsidePt,
												Point2d strictlyInsidePt)
	{
		Point2d result = createPoint(0, 0);

		boolean isWest = isWest( bounds, outsidePt );

		if( isWest )
			result.x = bounds.x;
		else
			result.x = bounds.x + bounds.width;

		result.y = this.calculateY(outsidePt, strictlyInsidePt, result.x);

		if( ! bounds.isInside(result) )
			result = null;

		return( result );
	}

	@Override
	public void round( Point2d pt )
	{
		pt.x = Math.round(pt.x);
		pt.y = Math.round(pt.y);
	}

	@Override
	public Point2d getIntersectionWithBoundary( GeometricRectangle bounds,
												Point2d pt1,
												Point2d pt2 )
	{
		Point2d result = null;
		Point2d outsidePt = getOutsidePt( bounds, pt1, pt2 );

		if( outsidePt != null )
		{
			Point2d strictlyInsidePt = getStrictlyInsidePt( bounds, pt1, pt2 );

			if( strictlyInsidePt != null )
			{
				result = getHorizontalIntersection( bounds, outsidePt, strictlyInsidePt );
				if( result == null )
				{
					boolean isWest = isWest( bounds, outsidePt );
					result = getVerticalIntersection( bounds, outsidePt, strictlyInsidePt );

					if( result == null )
						LOGGER.info( "Intersection not found!! for: {}, ({})-({})", bounds, pt1, pt2 );
				}
			}
		}

		return( result );
	}
}
