/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.delaunay.triangulation.quadedge.helpers;

import com.frojasg1.delaunay.triangulation.quadedge.DelaunayQuadEdgeHelpers;
import com.frojasg1.libgeometry.types.GeometricRectangle;
import com.frojasg1.delaunay.triangulation.quadedge.DelaunaySubdivision;
import com.frojasg1.delaunay.triangulation.quadedge.Edge;
import com.frojasg1.libgeometry.types.Point2d;
import com.frojasg1.libgeometry.types.GeometricTriangle;
import com.frojasg1.libgeometry.types.factory.GeometryCommonFactory;
import com.frojasg1.general.CollectionFunctions;

import java.util.*;
import java.util.function.Consumer;
import java.util.function.Predicate;
import java.util.function.Supplier;
import java.util.function.UnaryOperator;

import com.frojasg1.general.ExecutionFunctions;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 *
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 */
public class DelaunayTriangleBrowser
{
	private static final Logger LOGGER = LoggerFactory.getLogger(DelaunayTriangleBrowser.class);

	protected static final DelaunayTriangleBrowser INSTANCE = new DelaunayTriangleBrowser();

	public static DelaunayTriangleBrowser instance()
	{
		return( INSTANCE );
	}

	public List<Point2d> getVertices(DelaunaySubdivision delaunaySubdivision,
									 GeometricRectangle bounds) {
		Set<Point2d> set = browseVertices(delaunaySubdivision, bounds, null);

		List<Point2d> result = new ArrayList<>(set);
		Collections.sort(result);

		return result;
	}

	public Set<Point2d> browseVertices(DelaunaySubdivision delaunaySubdivision,
									   GeometricRectangle bounds,
									   Consumer<Point2d> vertexConsumer) {
		Set<Point2d> set = new HashSet<>();

		browseTriangles(delaunaySubdivision, bounds, trian -> {
			browseVertex(trian.getVertex1(), bounds, set, vertexConsumer);
			browseVertex(trian.getVertex2(), bounds, set, vertexConsumer);
			browseVertex(trian.getVertex3(), bounds, set, vertexConsumer);
			return true;
		});

		return set;
	}

	protected void browseVertex(Point2d vertex, GeometricRectangle bounds, Set<Point2d> set,
                                Consumer<Point2d> vertexConsumer) {
		if(!set.contains(vertex) && ((bounds == null) || (bounds.isInside(vertex)))) {
			set.add(vertex);
			if(vertexConsumer != null) {
				vertexConsumer.accept(vertex);
			}
		}
	}

	public List<GeometricTriangle> getTriangles(DelaunaySubdivision delaunaySubdivision,
												GeometricRectangle bounds) {
		List<GeometricTriangle> result = new ArrayList<>();
		browseTriangles(delaunaySubdivision, bounds, trian -> {
			result.add(trian);
			return true;
		});
		Collections.sort(result);

		return result;
	}

	public void browseTriangles( DelaunaySubdivision delaunaySubdivision,
								GeometricRectangle bounds,
								Predicate<GeometricTriangle> trianglePredicate )
	{
		browseTriangles( delaunaySubdivision, bounds, trianglePredicate, null );
	}

	public void browseTriangles( DelaunaySubdivision delaunaySubdivision,
								GeometricRectangle bounds,
								Predicate<GeometricTriangle> trianglePredicate,
								Supplier<Boolean> hasToCancelFunction )
	{
		Context context = new Context();
		context.setDelaunaySubdivision(delaunaySubdivision);
		context.setBounds(bounds);
		context.setTriangleConsumer(trianglePredicate);
		context.setHasToCancelFunction(hasToCancelFunction);

		Edge edge = delaunaySubdivision.getStartingEdge();

		ExecutionFunctions.instance().runtimeExceptionMethodExecution(() ->
				browseTrianglesInternal( context, edge ));
	}

	protected GeometricTriangle createTriangle(Context context,
											   Point2d v1, Point2d v2, Point2d v3) {
		UnaryOperator<Point2d> preprocess = context.getDelaunaySubdivision()::removeBoundaryDelta;
		return createTriangle(preprocess.apply(v1),
				preprocess.apply(v2),
				preprocess.apply(v3));
	}

	protected GeometricTriangle createTriangle(Point2d v1, Point2d v2, Point2d v3)
	{
		return( GeometryCommonFactory.instance().createDelaunayTriangle( v1, v2, v3, true ) );
	}

	protected DelaunayQuadEdgeHelpers getDelaunayQuadEdgeHelpers() {
		return DelaunayQuadEdgeHelpers.instance();
	}

	protected boolean isConvexOrLine(Point2d v1, Point2d v2, Point2d v3) {
		return getDelaunayQuadEdgeHelpers().TriArea(v1, v2, v3) >= 0;
	}

	protected boolean processTriangle(Context context,
									Point2d v1, Point2d v2, Point2d v3)
	{
		boolean result = true;
		// the condition is to avoid the "triangles" that are outside the convex boundaries
		// of the triangulation
		if(isConvexOrLine(v1, v2, v3)) {
			GeometricTriangle triangle = createTriangle(context, v1, v2, v3);
			if (context.isInside(triangle) && !context.hasBeenProcessed(triangle)) {
				context.addProcessed(triangle);
				result = context.process(triangle);
			}
		}
		return result;
	}

	protected void processPointTriangles( Context context, Point2d pt )
	{
		if( !context.hasToCancel() && !context.hasBeenProcessed( pt ) ) {
			context.addProcessed( pt );

			List<Point2d> list = context.getDelaunaySubdivision().getPossibleTriangles(pt);

			Iterator<Point2d> it = list.iterator();
			Point2d v1 = it.next();
			Point2d v2 = CollectionFunctions.instance().getLastOf(list);
			Point2d v3 = it.next();
			while (true) {
				if(!processTriangle(context, v1, v2, v3) || !it.hasNext()) {
					break;
				}

				v2 = v3;
				v3 = it.next();
			}
		}
	}

	protected void browseTrianglesInternal( Context context,
											Edge edge )
	{
		context.addEdgeToPending(edge);

		Edge edge3;
		Edge edge2 = context.popPending();
		while(!context.hasToCancel() && (edge2 != null)) {
			if(!context.hasBeenProcessed(edge2)) {
				context.addProcessed(edge2);
				processPointTriangles(context, edge2.Org());
//				processPointTriangles(context, edge2.Dest());

				context.addEdgeToPending(edge2.Onext());
				context.addEdgeToPending(edge2.Sym());
			}
			edge2 = context.popPending();
		}
	}

	public void browseTrianglesOfVertex(DelaunaySubdivision triangulation, Point2d vertex, GeometricRectangle imageBounds,
										Consumer<GeometricTriangle> visitor) {
		List<Point2d> masterTrianglesUmbrella = triangulation.getPossibleTriangles(vertex);
		Point2d ma = masterTrianglesUmbrella.get(0);
		Point2d mb = masterTrianglesUmbrella.get(masterTrianglesUmbrella.size() - 1);
		Point2d mc = masterTrianglesUmbrella.get(1);

		Iterator<Point2d> it = masterTrianglesUmbrella.iterator();
		it.next();
		it.next();
		do
		{
			if(isConvexOrLine(ma, mb, mc) &&
					((imageBounds == null) ||
					imageBounds.isInside(ma) && imageBounds.isInside(mb) && imageBounds.isInside(mc))) {
				visitor.accept(createTriangle(ma, mb, mc));
			}
			if( !it.hasNext() )
				break;

			mb = mc;
			mc = it.next();
		}
		while( true );
	}

	protected static class Context
	{
		protected Supplier<Boolean> _hasToCancelFunction;
		protected DelaunaySubdivision _delaunaySubdivision;
		protected GeometricRectangle _bounds;
		protected Set _processed = new HashSet();
		protected Predicate<GeometricTriangle> trianglePredicate;

		public DelaunaySubdivision getDelaunaySubdivision() {
			return _delaunaySubdivision;
		}

		LinkedList<Edge> pending = new LinkedList<Edge>();
		Set<Edge> addedToPending = new HashSet<>();

		public void setDelaunaySubdivision(DelaunaySubdivision _delaunaySubdivision) {
			this._delaunaySubdivision = _delaunaySubdivision;
		}

		public GeometricRectangle getBounds() {
			return _bounds;
		}

		public void setBounds(GeometricRectangle _bounds) {
			this._bounds = _bounds;
		}

		public boolean hasBeenProcessed( Object obj )
		{
			return( _processed.contains(obj) );
		}

		public void addProcessed( Object obj )
		{
			_processed.add( obj );
		}

		public Predicate<GeometricTriangle> getTriangleConsumer() {
			return trianglePredicate;
		}

		public void setTriangleConsumer(Predicate<GeometricTriangle> _triangleConsumer) {
			this.trianglePredicate = _triangleConsumer;
		}

		public boolean isInside( GeometricTriangle triangle )
		{
			return( (_bounds == null) ||
					_bounds.isInside( triangle.getVertex1() ) &&
							_bounds.isInside( triangle.getVertex2() ) &&
							_bounds.isInside( triangle.getVertex3() )
			);
		}

		public boolean process( GeometricTriangle triangle )
		{
			return getTriangleConsumer().test(triangle);
		}

		public Supplier<Boolean> getHasToCancelFunction() {
			return _hasToCancelFunction;
		}

		public void setHasToCancelFunction(Supplier<Boolean> _hasToCancelRef) {
			this._hasToCancelFunction = _hasToCancelRef;
		}

		public boolean hasToCancel()
		{
			Supplier<Boolean> hasToCancelFunction = getHasToCancelFunction();
			boolean result = false;
			if( hasToCancelFunction != null )
				result = hasToCancelFunction.get();

			return( result );
		}

		public void addEdgeToPending(Edge edge) {
			if(!addedToPending.contains(edge)) {
				pending.addLast(edge);
				addedToPending.add(edge);
			}
		}

		public Edge popPending() {
			Edge result = null;
			if(!pending.isEmpty()) {
				result = pending.removeFirst();
			}
			return result;
		}
	}
}
