/*
 * Copyright (c) 2023. Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 3.0 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You may obtain a copy of the License at
 *
 *       http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */

package com.frojasg1.applications.libmorphing.api.facedetector.helpers;

import com.frojasg1.applications.libmorphing.api.facedetector.constants.Landmark68constants;
import com.frojasg1.applications.libmorphing.api.facedetector.helpers.facedetectionframe.FaceDetectionFrameAngleCorrector;
import com.frojasg1.applications.libmorphing.api.facedetector.helpers.facedetectionframe.FaceDetectionFrameCalculatorContext;
import com.frojasg1.applications.libmorphing.api.facedetector.helpers.facedetectionframe.FaceDetectionFrameShrinker;
import com.frojasg1.applications.libmorphing.api.facedetector.pojo.FaceDetection;
import com.frojasg1.applications.libmorphing.api.facedetector.pojo.FaceDetectionFrame;
import com.frojasg1.applications.libmorphing.model.images.vertex.VertexModel;
import com.frojasg1.applications.libmorphing.model.images.vertex.map.impl.control.DynamicMultimapVertexWithIndex;
import com.frojasg1.applications.libmorphing.model.images.vertex.map.impl.control.IndexedVerticesMap;
import com.frojasg1.delaunay.triangulation.quadedge.DelaunaySubdivision;
import com.frojasg1.libgeometry.types.GeometricRectangle;
import com.frojasg1.libgeometry.types.GeometricTriangle;
import com.frojasg1.libgeometry.types.Point2d;
import com.frojasg1.general.number.DoubleFunctions;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.function.BinaryOperator;
import java.util.function.Function;
import java.util.function.Predicate;
import java.util.function.UnaryOperator;

public class Control88meshCreator extends FaceDetectionBasicHelper {
    private static final Logger LOGGER = LoggerFactory.getLogger(Control88meshCreator.class);

    protected static Control88meshCreator INSTANCE = new Control88meshCreator();

    public static Control88meshCreator instance() {
        return INSTANCE;
    }

    // see 88-control.point.mesh.with.68-landmark.jpg
    protected static final Set<Integer> BOUNDS_VERTEX_INDICES = new HashSet<>(Arrays.asList(68, 71, 74, 77));


    protected FaceDetectionFrameAngleCorrector getFaceDetectionFrameAngleCorrector() {
        return FaceDetectionFrameAngleCorrector.instance();
    }

    protected FaceDetectionFrameShrinker getFaceDetectionFrameShrinker() {
        return FaceDetectionFrameShrinker.instance();
    }

    protected void createIdealNormalizedIndexedVerticesMap(IndexedVerticesMap idealControl80MeshVertices, String idealName,
                                                           String normalizedName, int[][] trianglesIndicesResult,
                                                           IndexedVerticesMap result, GeometricRectangle bounds) {
        getControlMeshMapper().createIdealNormalizedIndexedVerticesMap(idealControl80MeshVertices, idealName,
                normalizedName, trianglesIndicesResult, result, bounds);
    }

    protected FaceDetectionFrameCalculatorContext createFrameCalculatorContext() {
        return new FaceDetectionFrameCalculatorContext();
    }

    public FaceDetectionFrame getFaceDetectionFrameWithAngleCorrection(FaceDetection faceDetection,
                                                                       GeometricRectangle bounds,
                                                                       double maxFactor,
                                                                       double maxAngleCorrectionInDegrees) {
        double tolerance = 0.01;
        FaceDetectionFrameCalculatorContext context = createFrameCalculatorContext();
        context.setFaceDetection(faceDetection);
        context.setResult(getFaceDetectionFrame(faceDetection));
        context.setMaxAngleCorrectionInDegrees(maxAngleCorrectionInDegrees);

        return map(context, null, Function.identity(),
                ctx ->
                        getFaceDetectionFrameShrinker()
                                .shrink(ctx, null, 1.05d, maxFactor, tolerance,
                                        ShrinkerTarget.GREATEST),
                ctx -> getFaceDetectionFrameAngleCorrector().apply(ctx),
                ctx -> getFaceDetectionFrameShrinker().shrink(ctx, bounds,
                        1.0d / ctx.getAppliedFactor(),maxFactor / ctx.getAppliedFactor(), tolerance,
                        ShrinkerTarget.AVERAGE))
                .getResult()
                .applyTransformation(Point2d::createIntCopyIfNotInt);
    }

    public void applyNewFrame(IndexedVerticesMap vertices, String imageName, GeometricRectangle bounds,
                              FaceDetectionFrame frame) {
        UnaryOperator<Point2d> pointLimiter = pt -> getNearestSuitablePoint(pt, bounds, imageName, vertices);
        applyMeshFrameInternal(frame, imageName, null, vertices, pointLimiter, bounds);
    }

    public Point2d getNearestEmptyPointRespectingThePreviousOne(Point2d pt, Predicate<Integer> indexValidator,
                                                                GeometricRectangle bounds,
                                                                String name, IndexedVerticesMap vertices) {
        Point2d result = pt;
        Integer index = getIndexInternal(pt, vertices, name);
        if((indexValidator == null) || !indexValidator.test(index)) {
            result = getNearestSuitablePoint(pt, bounds, name, vertices);
        }
        return result;
    }

    public void recalculateAllControlFramePoints(String name, IndexedVerticesMap result) {
        AtomicInteger ind = new AtomicInteger(68);

        // see 88-control.point.mesh.with.68-landmark.jpg
        Point2d ul = getVertex(result, name, 68);
        Point2d dl = getVertex(result, name, 71);
        Point2d dr = getVertex(result, name, 74);
        Point2d ur = getVertex(result, name, 77);

        GeometricRectangle bounds = getImageBounds(result, name);

        UnaryOperator<Point2d> pointMapper = pt -> getNearestEmptyPointRespectingThePreviousOne(
                pt,
                index -> (index == null) || (ind.get() == index),
                bounds, name, result);
        addPolygonFrameWithStepsDirectMapping(3, name, ind, result,
                (pt, totalVector) -> roundRecalculatedControlFramePoints(pt, totalVector, true),
                pointMapper, bounds, ul, dl, dr, ur, ul);
    }

    protected DoubleFunctions getDoubleFunctions() {
        return DoubleFunctions.instance();
    }

    public int sgn(double dd) {
        return getDoubleFunctions().sgn(dd);
    }

    protected double round(double value, int sign) {
        double result;
        if(sign <= 0) {
            result = Math.floor(value);
        } else {
            result = Math.ceil(value);
        }
        return result;
    }

    // to right with y axis increasing to below direction
    protected Point2d roundRecalculatedControlFramePoints(Point2d point2d, Point2d totalVector, boolean roundToRight) {
        // see Rounding.direction.for.intermediate.control.frame.points.png
        int sign = roundToRight ? 1 : -1;

        int rightXSign = -sgn(totalVector.y);
        int rightYSign = sgn(totalVector.x);

        Point2d result = point2d.createCopy();
        result.x = round(result.x, sign * rightXSign);
        result.y = round(result.y, sign * rightYSign);

        return result;
    }

    public AtomicInteger createControl80mesh(FaceDetection faceDetection, String name, IndexedVerticesMap result,
                                             UnaryOperator<Point2d> pointLimiter,
                                             Function<FaceDetection, FaceDetectionFrame> faceDetectorToFrameFunction,
                                             GeometricRectangle bounds) {
        List<Point2d> list = faceDetection.getPoints();
        checkNumPoints(Landmark68constants.NB_POINTS_LANDMARK_68, list.size());

        if(result.mapExists(name)) {
            result.removeEntryByName(name);
        }
        result.addMapByName(name);

        AtomicInteger ind = new AtomicInteger(0);
        for(Point2d pt: list) {
            set(result, name, ind.getAndIncrement(), pt.createIntCopyIfNotInt(), pointLimiter, bounds);
        }

        FaceDetectionFrame frame = faceDetectorToFrameFunction.apply(faceDetection);

        LOGGER.info("face frame: {}", frame);

        applyMeshFrameInternal(frame, name, ind, result, pointLimiter, bounds);

        return ind;
    }

    protected void applyMeshFrameInternal(FaceDetectionFrame frame, String name, AtomicInteger ind,
                                          IndexedVerticesMap result, UnaryOperator<Point2d> pointLimiter,
                                          GeometricRectangle bounds) {
        if(ind == null) {
            ind = new AtomicInteger(Landmark68constants.NB_POINTS_LANDMARK_68);
        }

        addPolygonFrameWithSteps(3, name, ind, result, pointLimiter, bounds,
                frame.getLeftUpperCorner(),
                frame.getLeftLowerCorner(),
                frame.getRightLowerCorner(),
                frame.getRightUpperCorner(),
                frame.getLeftUpperCorner());

        checkNumPoints(Landmark68constants.NB_POINTS_LANDMARK_80, ind.get());
    }

    protected void addLandmark80Frame(String name, AtomicInteger ind, IndexedVerticesMap result,
                                      UnaryOperator<Point2d> pointLimiter,
                                      GeometricRectangle bounds,
                                      Point2d ... points) {
        if(ind == null) {
            ind = new AtomicInteger(Landmark68constants.NB_POINTS_LANDMARK_68);
        }
        addPolygonFrameWithSteps(3, name, ind, result, pointLimiter, bounds, points);
    }

    public void createIdealControl80mesh(FaceDetection faceDetection, String name, IndexedVerticesMap result,
                                         GeometricRectangle bounds) {
        createControl80mesh(faceDetection, name, result, null,
                faceDetection1 -> getFaceDetectionFrameWithAngleCorrection(faceDetection1, null, 1.33d, 0),
                bounds);
    }
//
//    public void createControl80mesh(FaceDetection faceDetection, String name, DelaunayRectangle bounds,
//                                    IndexedVerticesMap result) {
//        createControl80mesh(faceDetection, name, result,
//                pt -> getNearestSuitablePoint(pt, bounds),
//                faceDetection1 -> getFaceDetectionFrameWithAngleCorrection(faceDetection1, null));
//    }

    protected void addPolygonFrameWithSteps(int nstepsPerEdge, String imageName, AtomicInteger ind,
                                            IndexedVerticesMap result,
                                            UnaryOperator<Point2d> pointLimiter,
                                            GeometricRectangle bounds, Point2d ... polygonVertices) {
        for(int ii=0; ii<(polygonVertices.length - 1); ii++) {
            addSteps(polygonVertices[ii], polygonVertices[ii+1], nstepsPerEdge, imageName, ind, result, pointLimiter, bounds);
        }
    }

    protected void addSteps(Point2d origin, Point2d dest, int steps, String name, AtomicInteger ind,
                            IndexedVerticesMap result, UnaryOperator<Point2d> pointLimiter,
                            GeometricRectangle bounds) {
        Point2d step = dest.subtract(origin);
        multiplyByScalar(step, 1.0d/steps, step);
        Point2d current;
        for(int ii=0; ii<steps; ii++) {
            current = linearCombination(1.0d, origin, ii, step).createIntCopyIfNotInt();
            set(result, name, ind.getAndIncrement(), current, pointLimiter, bounds);
        }
    }

    // Direct mapping because the pointMapper has already included the pointLimiter and the search for the nearest empty point
    protected void addPolygonFrameWithStepsDirectMapping(int nstepsPerEdge, String imageName, AtomicInteger ind,
                                                         IndexedVerticesMap result, BinaryOperator<Point2d> rounder,
                                                         UnaryOperator<Point2d> pointMapper, GeometricRectangle bounds,
                                                         Point2d ... polygonVertices) {
        for(int ii=0; ii<(polygonVertices.length - 1); ii++) {
            addStepsDirectMapping(polygonVertices[ii], polygonVertices[ii+1], nstepsPerEdge, imageName, ind, result,
                    rounder, pointMapper, bounds);
        }
    }

    // Direct mapping because the pointMapper has already included the pointLimiter and the search for the nearest empty point
    protected void addStepsDirectMapping(Point2d origin, Point2d dest, int steps, String name, AtomicInteger ind,
                                         IndexedVerticesMap result, BinaryOperator<Point2d> rounder,
                                         UnaryOperator<Point2d> pointMapper,
                                         GeometricRectangle bounds) {
        Point2d totalVector = dest.subtract(origin);
        Point2d step = dest.subtract(origin);
        multiplyByScalar(step, 1.0d/steps, step);
        Point2d current;
        for(int ii=0; ii<steps; ii++) {
            current = linearCombination(1.0d, origin, ii, step);
            current = rounder.apply(current, totalVector);
            // if the rounded point is already occupied, there can be problems when mapping.
            // but ... what can we do ?
            current = pointMapper.apply(current);
            result.set(name, ind.getAndIncrement(), buildVertexModel(current, bounds));
        }
    }

    protected VertexModel buildVertexModel(Point2d pt, GeometricRectangle bounds) {
        return VertexModel.build(pt, bounds);
    }

    public void createParticularControl88mesh(FaceDetection faceDetection, GeometricRectangle bounds,
                                              String name, IndexedVerticesMap result) {
        createControl88mesh(faceDetection, bounds, name, result,
                faceDetection1 -> getFaceDetectionFrameWithAngleCorrection(faceDetection1, bounds, 2.5d, 30));
    }

    protected void createControl88mesh(FaceDetection faceDetection, GeometricRectangle bounds,
                                       String name, IndexedVerticesMap result,
                                       Function<FaceDetection, FaceDetectionFrame> faceDetectorToFrameFunction) {
//        UnaryOperator<Point2d> pointLimiter = pt -> limit(pt, bounds);
        UnaryOperator<Point2d> pointLimiter = pt -> getNearestSuitablePoint(pt, bounds, name, result);
        AtomicInteger ind = createControl80mesh(faceDetection, name, result, pointLimiter, faceDetectorToFrameFunction,
                bounds);

        double x1 = bounds.x;
        double x2 = bounds.x + bounds.width - 1;
        double y1 = bounds.y;
        double y2 = bounds.y + bounds.height - 1;

        Point2d ul = createPoint2d(x1, y1);
        Point2d dl = createPoint2d(x1, y2);
        Point2d dr = createPoint2d(x2, y2);
        Point2d ur = createPoint2d(x2, y1);

        addPolygonFrameWithSteps(2, name, ind, result, pointLimiter, bounds, ul, dl, dr, ur, ul);

        checkNumPoints(Landmark68constants.NB_POINTS_LANDMARK_88, ind.get());
    }

//    protected Point2d set(IndexedVerticesMap vertices, String mapName, int index, Point2d pt,
//                          UnaryOperator<Point2d> pointLimiter, int maxDistance) {
//        Point2d result = setInternal(vertices, mapName, index, pt);
//        if(result == null) {
//            Point2d tmp = pt.createCopy();
//            for (int dist = 0; (dist <= maxDistance) && (result == null); dist++) {
//                result = setTotalSlice(vertices, mapName, index, pt, pointLimiter, dist, tmp);
//            }
//        }
//        if(result == null) {
//            throw new IllegalStateException(String.format("Could not add point: %s, with maxDistance: %d", pt, maxDistance));
//        }
//        return result;
//    }
//
//    protected Point2d setTotalSlice(IndexedVerticesMap vertices, String mapName, int index, Point2d pt,
//                                    UnaryOperator<Point2d> pointLimiter, int sliceDistance, Point2d tmpResult) {
//        Point2d result = null;
//        if(tmpResult == null) {
//            tmpResult = pt.createCopy();
//        }
//        for(int iterDist = 0; (iterDist <= sliceDistance) && (result == null); iterDist++) {
//            result = setSliceIteration(vertices, mapName, index, pt, pointLimiter, sliceDistance, tmpResult, iterDist);
//        }
//        return result;
//    }
//
//    protected Point2d setSliceIterationParticularization(IndexedVerticesMap vertices, String mapName, int index, Point2d pt,
//                                                         UnaryOperator<Point2d> pointLimiter, Point2d tmpResult,
//                                                         int deltax, int deltay) {
//        tmpResult.x = pt.x + deltax;
//        tmpResult.y = pt.y + deltay;
//        return setInternal(vertices, mapName, index, apply(pointLimiter, tmpResult));
//    }
//
//    protected Point2d setSliceIteration(IndexedVerticesMap vertices, String mapName, int index, Point2d pt,
//                                        UnaryOperator<Point2d> pointLimiter, int sliceDistance,
//                                        Point2d tmpResult, int iterDist) {
//        Point2d result = null;
//        BiFunction<Integer, Integer, Point2d> myFunction =
//                (dx, dy) -> setSliceIterationParticularization(vertices, mapName, index, pt,
//                        pointLimiter, tmpResult, dx, dy);
//
//        if((iterDist == 0) && (sliceDistance == 0)) {
//            result = myFunction.apply(0, 0);
//        } else if(iterDist == 0) {
//            result = getFirst(
//                    () -> myFunction.apply(-sliceDistance, 0),
//                    () -> myFunction.apply(sliceDistance, 0),
//                    () -> myFunction.apply(0, -sliceDistance),
//                    () -> myFunction.apply(0, sliceDistance)
//            );
//        } else if(iterDist == sliceDistance) {
//            result = getFirst(
//                    () -> myFunction.apply(-sliceDistance, -sliceDistance),
//                    () -> myFunction.apply(-sliceDistance, sliceDistance),
//                    () -> myFunction.apply(sliceDistance, -sliceDistance),
//                    () -> myFunction.apply(sliceDistance, sliceDistance)
//            );
//        } else {
//            result = getFirst(
//                    () -> myFunction.apply(-sliceDistance, -iterDist),
//                    () -> myFunction.apply(-sliceDistance, iterDist),
//                    () -> myFunction.apply(sliceDistance, -iterDist),
//                    () -> myFunction.apply(sliceDistance, iterDist),
//                    () -> myFunction.apply(-iterDist, -sliceDistance),
//                    () -> myFunction.apply(iterDist, -sliceDistance),
//                    () -> myFunction.apply(-iterDist, +sliceDistance),
//                    () -> myFunction.apply(iterDist, +sliceDistance)
//            );
//        }
//        return result;
//    }

    protected IndexedVerticesMap createIdealAndNormalizedControlMeshVerticesMap(FaceDetection faceDetection,
                                                                                String idealName,
                                                                                String normalizedName) {
        IndexedVerticesMap result = null;
        if(faceDetection != null) {
            List<Point2d> list = faceDetection.getPoints();
            checkNumPoints(Landmark68constants.NB_POINTS_LANDMARK_68, list.size());

            GeometricRectangle bounds = null;
            result = createControlMeshVerticesMap();
            createIdealControl80mesh(faceDetection, idealName, result, bounds);

            int[][] trianglesIndicesResult = new int[2][];
            IndexedVerticesMap verticesForMapper = createControlMeshVerticesMap();

            createIdealNormalizedIndexedVerticesMap(result,
                    idealName, normalizedName, trianglesIndicesResult, verticesForMapper, bounds);

            result.addMapByName(normalizedName);

//            int luInd = 0;
//            int rdInd = 2;
//            DelaunayRectangle bounds = getBounds(verticesForMapper, idealName, luInd, rdInd);
//            LOGGER.info("ideal bounds: {}", bounds);

            Point2d[] tmpArray = new Point2d[3];
            Point2d idealPt;
            for(int ii = 0; ii < Landmark68constants.NB_POINTS_LANDMARK_68; ii++) {
                idealPt = getVertex(result, idealName, ii);
//                LOGGER.info("index: {}", ii);
                Point2d pt = map(idealPt, null,
                        verticesForMapper, idealName, normalizedName, trianglesIndicesResult,
                        tmpArray).createIntCopyIfNotInt();

                result.set(normalizedName, ii, buildVertexModel(pt, null));
            }

            addLandmark80Frame(normalizedName, null, result, null, null,
                    getPoint2d(verticesForMapper, 0, normalizedName),
                    getPoint2d(verticesForMapper, 1, normalizedName),
                    getPoint2d(verticesForMapper, 2, normalizedName),
                    getPoint2d(verticesForMapper, 3, normalizedName),
                    getPoint2d(verticesForMapper, 0, normalizedName));
        }

        return result;
    }

    protected GeometricRectangle getNormalizedBounds(IndexedVerticesMap vertices, String normalizedName) {
        // see 88-control.point.mesh.with.68-landmark.jpg
        return getBounds(vertices, normalizedName, 68, 74);
    }

    protected GeometricRectangle getImageBounds(IndexedVerticesMap vertices, String name) {
        // see 88-control.point.mesh.with.68-landmark.jpg
        return getBounds(vertices, name, 80, 84);
    }

    public boolean isBoundsVertex(int ind) {
        // see 88-control.point.mesh.with.68-landmark.jpg
        return BOUNDS_VERTEX_INDICES.contains(ind);
    }

    // internal triangles from the 80 points (68-landmark + 12 frame points)
    protected List<GeometricTriangle> getNormalizedTriangles(IndexedVerticesMap vertices, String normalizedName) {
        DelaunaySubdivision triangulation = createTriangulation(getNormalizedBounds(vertices, normalizedName));

        DynamicMultimapVertexWithIndex multiVertex = null;
        int ind = -1;
        do {
            multiVertex = vertices.get(++ind);
            if((multiVertex != null) && (!isBoundsVertex(ind))) {
                triangulation.InsertSite(multiVertex.get(normalizedName).getPoint());
            }
        } while(multiVertex != null);

        return getTriangles(triangulation, null);
    }

    protected void checkNumPoints(int numExpectedPoints, int numPoints) {
        if(numPoints != numExpectedPoints) {
            throw new IllegalArgumentException(
                    String.format("Expected: %d points, found %d", numExpectedPoints, numPoints));
        }
    }

    public int[][] createControlMeshTriangles(FaceDetection faceDetection) {
        int[][] result = null;
        String normalizedName = "normalized";
        String idealName = "ideal";
        IndexedVerticesMap vertices = createIdealAndNormalizedControlMeshVerticesMap(faceDetection, idealName, normalizedName);

        result = createMeshTriangles(vertices, normalizedName);
        if(result != null) {
            result = joinArrays(result, createSurroundingMeshTriangles());
        }

        return result;
    }

    protected int[][] createSurroundingMeshTriangles() {
        int[][] result = new int[12][];

        // see 88-control.point.mesh.with.68-landmark.jpg
        result[0] = new int[]{80, 68, 81};
        result[1] = new int[]{71, 68, 81};
        result[2] = new int[]{71, 82, 81};
        result[3] = new int[]{71, 82, 83};
        result[4] = new int[]{71, 83, 74};
        result[5] = new int[]{83, 74, 84};
        result[6] = new int[]{84, 74, 85};
        result[7] = new int[]{74, 85, 77};
        result[8] = new int[]{77, 85, 86};
        result[9] = new int[]{77, 86, 87};
        result[10] = new int[]{77, 68, 87};
        result[11] = new int[]{80, 68, 87};

        return result;
    }

    protected int[][] createMeshTriangles(IndexedVerticesMap vertices, String normalizedName) {
        int[][] result = null;
        if(vertices != null) {
            List<GeometricTriangle> normalizedTriangles = getNormalizedTriangles(vertices, normalizedName);
            result = new int[normalizedTriangles.size()][];

            int ind = 0;
            for (GeometricTriangle triangle : normalizedTriangles) {
                result[ind++] = createVertexIndexTriangle(triangle, vertices, normalizedName);
            }
        }

        return result;
    }

    protected int[] createVertexIndexTriangle(GeometricTriangle triangle, IndexedVerticesMap vertices, String normalizedName) {
        int[] result = new int[3];
        result[0] = getIndex(triangle.getVertex1(), vertices, normalizedName);
        result[1] = getIndex(triangle.getVertex2(), vertices, normalizedName);
        result[2] = getIndex(triangle.getVertex3(), vertices, normalizedName);

        return result;
    }
}
