/*
 * Copyright (c) 2023. Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 3.0 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You may obtain a copy of the License at
 *
 *       http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */

package com.frojasg1.applications.libmorphing.api.facedetector.helpers;

import com.frojasg1.libgeometry.algebra.helpers.Point2dDiscarder;
import com.frojasg1.applications.libmorphing.model.images.vertex.VertexModel;
import com.frojasg1.applications.libmorphing.model.images.vertex.map.impl.control.IndexedVerticesMap;
import com.frojasg1.libgeometry.types.GeometricRectangle;
import com.frojasg1.libgeometry.types.Point2d;
import com.frojasg1.general.NullFunctions;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.function.BiConsumer;
import java.util.function.Function;

public class ControlMeshBrowser {
    private static final Logger LOGGER = LoggerFactory.getLogger(ControlMeshBrowser.class);


    protected static ControlMeshBrowser INSTANCE = new ControlMeshBrowser();

    public static ControlMeshBrowser instance() {
        return INSTANCE;
    }

    public void browseEdges(int[][] triangles, IndexedVerticesMap controlMeshVerticesMap, String imageName,
                            GeometricRectangle bounds, BiConsumer<Point2d, Point2d> edgeConsumer) {
        if(triangles != null) {
            boolean mapExists = controlMeshVerticesMap.mapExists(imageName);
            if ((triangles != null) && mapExists) {
                browseEdges(triangles, index -> getPoint2d(controlMeshVerticesMap, imageName, index),
                        bounds, edgeConsumer);
            } else {
                LOGGER.warn("could not browse edges. Triangles: {}, mapExists({}): {}", triangles, imageName, mapExists);
            }
        }
    }

    public void browseEdges(int[][] triangles, Function<Integer, Point2d> pointGetter,
                            GeometricRectangle bounds, BiConsumer<Point2d, Point2d> edgeConsumer) {
        if(pointGetter != null) {
            for (int[] triangleIndices : triangles) {
                browseTriangle(triangleIndices, pointGetter, bounds, edgeConsumer);
            }
        }
    }

    protected void browseTriangle(int[] triangleIndices, Function<Integer, Point2d> pointGetter,
                                  GeometricRectangle bounds, BiConsumer<Point2d, Point2d> edgeConsumer) {
        Point2d v0 = pointGetter.apply(triangleIndices[0]);
        Point2d v1 = pointGetter.apply(triangleIndices[1]);
        Point2d v2 = pointGetter.apply(triangleIndices[2]);
        if (!canDiscard(bounds, v0, v1, v2)) {
            edgeConsumer.accept(v0, v1);
            edgeConsumer.accept(v1, v2);
            edgeConsumer.accept(v2, v0);
        }
    }

    protected Point2dDiscarder getDiscarder() {
        return Point2dDiscarder.instance();
    }

    protected boolean canDiscard(GeometricRectangle bounds, Point2d... point2ds) {
        return getDiscarder().canDiscard(bounds, point2ds);
    }

    protected NullFunctions getNullFunctions() {
        return NullFunctions.instance();
    }

    protected <CC, RR> RR getIfNotNull(CC obj, Function<CC, RR> getter) {
        return getNullFunctions().getIfNotNull(obj, getter);
    }

    protected Point2d getPoint2d(IndexedVerticesMap controlMeshVerticesMap, String imageName, int index) {
        return getIfNotNull(
                getIfNotNull(
                        getIfNotNull(controlMeshVerticesMap, map -> map.get(index)),
                        multivertex -> multivertex.get(imageName)),
                VertexModel::getPoint);
    }
}
