/*
 * Copyright (c) 2023. Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 3.0 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You may obtain a copy of the License at
 *
 *       http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */

package com.frojasg1.applications.libmorphing.api.facedetector.helpers;

import com.frojasg1.libgeometry.algebra.helpers.GeometryHelper;
import com.frojasg1.libgeometry.algebra.helpers.VectorFunctions;
import com.frojasg1.applications.libmorphing.api.facedetector.helpers.facedetectionframe.FaceDetectionFrameCalculator;
import com.frojasg1.applications.libmorphing.api.facedetector.pojo.FaceDetection;
import com.frojasg1.applications.libmorphing.api.facedetector.pojo.FaceDetectionFrame;
import com.frojasg1.applications.libmorphing.api.facedetector.pojo.FaceDetectionMeshData;
import com.frojasg1.applications.libmorphing.mapper.ControlMeshMapper;
import com.frojasg1.applications.libmorphing.model.images.vertex.VertexModel;
import com.frojasg1.applications.libmorphing.model.images.vertex.map.impl.control.DynamicMultimapVertexWithIndex;
import com.frojasg1.applications.libmorphing.model.images.vertex.map.impl.control.IndexedVerticesMap;
import com.frojasg1.delaunay.triangulation.quadedge.DelaunaySubdivision;
import com.frojasg1.delaunay.triangulation.quadedge.builder.impl.QuadEdgeDelaunayBuilderImpl;
import com.frojasg1.delaunay.triangulation.quadedge.helpers.DelaunayTriangleBrowser;
import com.frojasg1.libgeometry.types.GeometricRectangle;
import com.frojasg1.libgeometry.types.GeometricTriangle;
import com.frojasg1.libgeometry.types.Point2d;
import com.frojasg1.libgeometry.types.factory.GeometryCommonFactory;
import com.frojasg1.general.ArrayFunctions;
import com.frojasg1.general.NullFunctions;
import com.frojasg1.general.functional.FunctionalInterfaceFunctions;
import com.frojasg1.general.functional.interfaces.TriFunction;
import com.frojasg1.general.selectors.Nearest2DSelector;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.List;
import java.util.Objects;
import java.util.function.Function;
import java.util.function.Supplier;
import java.util.function.UnaryOperator;

public class FaceDetectionBasicHelper {
    private static final Logger LOGGER = LoggerFactory.getLogger(FaceDetectionBasicHelper.class);

    protected static FaceDetectionBasicHelper INSTANCE = new FaceDetectionBasicHelper();

    public static FaceDetectionBasicHelper instance() {
        return INSTANCE;
    }

    protected DelaunayTriangleBrowser getDelaunayTriangleBrowser() {
        return DelaunayTriangleBrowser.instance();
    }

    public List<GeometricTriangle> getTriangles(DelaunaySubdivision delaunaySubdivision, GeometricRectangle bounds) {
        return getDelaunayTriangleBrowser().getTriangles(delaunaySubdivision, bounds);
    }

    protected GeometryCommonFactory getFactory() {
        return GeometryCommonFactory.instance();
    }

    public Point2d createPoint2d(double xx, double yy) {
        return getFactory().createPoint2d(xx, yy);
    }

    protected GeometricRectangle createDelaunayRectangle(double xx, double yy, double width, double height) {
        return getFactory().createDelaunayRectangle(xx, yy, width, height);
    }

    protected DelaunaySubdivision createTriangulation(GeometricRectangle bounds) {
        return QuadEdgeDelaunayBuilderImpl.instance().createDelaunaySubdivisionBetter(bounds);
    }

    public Point2d getVertex(IndexedVerticesMap vertices, String name, int index) {
        return vertices.get(index).get(name).getPoint();
    }

    protected VectorFunctions getVectorFunctions() {
        return VectorFunctions.instance();
    }

    protected FaceDetectionFrameCalculator getFaceDetectionFrameCalculator() {
        return FaceDetectionFrameCalculator.instance();
    }

    public Point2d linearCombination(double factor1, Point2d vector1, double factor2, Point2d vector2) {
        return getVectorFunctions().linearCombination(factor1, vector1, factor2, vector2);
    }

    protected Point2d addVectors(Point2d... points) {
        return getVectorFunctions().addVectors(points);
    }

    protected GeometryHelper getGeometryHelper() {
        return GeometryHelper.instance();
    }

    public Point2d limit(Point2d pt, GeometricRectangle rectangle) {
        return getGeometryHelper().limit(pt, rectangle);
    }

    protected Point2d multiplyByScalar(Point2d pt, double factor, Point2d result) {
        return getVectorFunctions().multiplyByScalar(pt, factor, result);
    }

    protected ControlMeshMapper getControlMeshMapper() {
        return ControlMeshMapper.instance();
    }

    protected FunctionalInterfaceFunctions getFunctionalInterfaceFunctions() {
        return FunctionalInterfaceFunctions.instance();
    }

    protected <C> C getFirst(Supplier<C>... suppliers) {
        return getFunctionalInterfaceFunctions().getFirst(suppliers);
    }

    protected <C> C apply(UnaryOperator<C> operator, C value) {
        return getFunctionalInterfaceFunctions().apply(operator, value);
    }

    protected <C> List<C> asList(C... elems) {
        return getFunctionalInterfaceFunctions().asList(elems);
    }

    protected <I, O> O map(I value, List<UnaryOperator<I>> preprocessors, Function<I, O> mapper, UnaryOperator<O>... postprocessors) {
        return getFunctionalInterfaceFunctions().map(value, preprocessors, mapper, postprocessors);
    }

    public Point2d map(Point2d originPt, UnaryOperator<Point2d> pointLimiter,
                       IndexedVerticesMap vertices, String originName, String destinationName,
                       int[][] trianglesIndices, Point2d[] tmpOriginTriangle) {
        return getControlMeshMapper().map(originPt, pointLimiter, vertices, originName, destinationName,
                trianglesIndices, tmpOriginTriangle);
    }

    protected FaceDetectionFrame getFaceDetectionFrame(FaceDetection faceDetection) {
        return getFaceDetectionFrameCalculator().apply(faceDetection);
    }

    protected IndexedVerticesMap createControlMeshVerticesMap() {
        return new IndexedVerticesMap();
    }

    protected ArrayFunctions getArrayFunctions() {
        return ArrayFunctions.instance();
    }

    public <C> C[] getArray(C... array) {
        return getArrayFunctions().getArray(array);
    }

    public <C> C[] joinArrays(C[]... arrays) {
        return getArrayFunctions().joinArrays(arrays);
    }

    public int[][] joinArrays(int[][] array1, int[][] array2) {
        return getArrayFunctions().joinArrays(array1, array2);
    }

    protected Nearest2DSelector getNearest2DSelector() {
        return Nearest2DSelector.instance();
    }

    protected boolean isPresent(Point2d pt, String name, IndexedVerticesMap result) {
        return result.get(name, pt) != null;
    }

    public Point2d getNearestSuitablePoint(Point2d pt, GeometricRectangle bounds, String name, IndexedVerticesMap result) {
        Point2d tmp = pt.createCopy();
        UnaryOperator<Point2d> selector = pt1 -> isPresent(pt1, name, result) ? null : pt1;
        return getNearestSuitable(pt, selector, pt1 -> limit(pt1, bounds),
                (pt1, dx, dy) -> applyDeltaToPoint(pt1, dx, dy, tmp), 5);
    }

    protected <C> C getNearestSuitable(C elem, UnaryOperator<C> selector, UnaryOperator<C> limiter,
                                       TriFunction<C, Integer, Integer, C> deltaXdeltaYfunction, int maxDelta) {
        return getNearest2DSelector().getNearestSuitable(elem, selector, limiter, deltaXdeltaYfunction, maxDelta);
    }

    protected Point2d setInternal(IndexedVerticesMap vertices, String mapName, int index, Point2d pt,
                                  GeometricRectangle bounds) {
        Point2d result = null;
        if(vertices.get(mapName, pt) == null) {
            result = pt;
            vertices.set(mapName, index, buildVertexModel(pt, bounds));
        }
        return result;
    }

    protected VertexModel buildVertexModel(Point2d pt, GeometricRectangle bounds) {
        return VertexModel.build(pt, bounds);
    }

    protected Point2d set(IndexedVerticesMap vertices, String mapName, int index, Point2d pt,
                          UnaryOperator<Point2d> pointLimiter,
                          GeometricRectangle bounds) {
        pt = apply(pointLimiter, pt);
        return set(vertices, mapName, index, pt, pointLimiter, 5, bounds);
    }

//    protected Point2d set(IndexedVerticesMap vertices, String mapName, int index, Point2d pt,
//                          UnaryOperator<Point2d> pointLimiter) {
//        pt = apply(pointLimiter, pt);
//        return set(vertices, mapName, index, pt, pointLimiter, 5);
//    }

    protected Point2d applyDeltaToPoint(Point2d pt, int deltax, int deltay, Point2d result) {
        if(pt == null) {
            result = null;
        } else {
            result.x = pt.x + deltax;
            result.y = pt.y + deltay;
        }
        return result;
    }

    protected Point2d set(IndexedVerticesMap vertices, String mapName, int index, Point2d pt,
                          UnaryOperator<Point2d> pointLimiter, int maxDelta,
                          GeometricRectangle bounds) {
        Point2d tmp = pt.createCopy();
        return getNearestSuitable(pt, pt2 -> setInternal(vertices, mapName, index, pt2, bounds), pointLimiter,
                (pt3, deltax, deltay) -> applyDeltaToPoint(pt3, deltax, deltay, tmp), maxDelta);
    }

    protected Point2d getPoint2d(IndexedVerticesMap verticesForMapper, int index, String name) {
        return verticesForMapper.get(index).get(name).getPoint();
    }

    protected GeometricRectangle getBounds(IndexedVerticesMap vertices, String normalizedName, int luIndex, int rdIndex) {
        Point2d lu = getVertex(vertices, normalizedName, luIndex);
        Point2d size = getVertex(vertices, normalizedName, rdIndex).subtract(lu);

        return createDelaunayRectangle(lu.x, lu.y, size.x + 1, size.y + 1);
    }

    public int getIndex(Point2d pt, IndexedVerticesMap vertices, String normalizedName) {
        Integer result = getIndexInternal(pt, vertices, normalizedName);
        if (result == null) {
            throw new IllegalArgumentException(String.format("Normalized point not found in map: %s", pt));
        }
        return result;
    }

    protected Integer getIndexInternal(Point2d pt, IndexedVerticesMap vertices, String normalizedName) {
        Integer result = null;
        DynamicMultimapVertexWithIndex multiVertex = vertices.get(normalizedName, pt);
        if(multiVertex != null) {
            result = multiVertex.getIndex();
        }
        return result;
    }

    protected NullFunctions getNullFunctions() {
        return NullFunctions.instance();
    }

    protected <CC, RR> RR getIfNotNull(CC obj, Function<CC,RR> getter )
    {
        return( getNullFunctions().getIfNotNull(obj, getter) );
    }

    protected Point2d getCenter(FaceDetectionMeshData faceDetectionMeshData) {
        return getIfNotNull(
                getIfNotNull(faceDetectionMeshData, FaceDetectionMeshData::getFaceDetectionFrame),
                FaceDetectionFrame::getCenter);
    }

    protected boolean isCenter(FaceDetectionMeshData faceDetectionMeshData, Point2d faceDetectionFrameCenter) {
        return Objects.equals(faceDetectionFrameCenter, getCenter(faceDetectionMeshData));
    }

    public FaceDetectionMeshData getFaceDetectionMeshData(Point2d faceDetectionFrameCenter,
                                                          List<FaceDetectionMeshData> faceDetectionMeshes) {
        FaceDetectionMeshData result = null;
        if((faceDetectionFrameCenter != null) && (faceDetectionMeshes != null)) {
            result = faceDetectionMeshes.stream()
                    .filter(faceDetectionMeshData1 -> isCenter(faceDetectionMeshData1, faceDetectionFrameCenter))
                    .findFirst().orElse(null);
        }
        return result;
    }

    protected boolean isInside(Point2d pt, Point2d a, Point2d b, Point2d c) {
        return getVectorFunctions().isInside(pt, a, b, c);
    }

    protected boolean isStrictlyInside(Point2d pt, FaceDetectionFrame frame) {
        boolean result = false;
        if(frame != null) {
            result = frame.isStrictlyInside(pt);
        }
        return result;
    }

    public boolean isValid(FaceDetection faceDetection, FaceDetectionFrame frame) {
        return faceDetection.getPoints().stream().allMatch(pt -> isStrictlyInside(pt, frame));
    }

    public boolean isInside(FaceDetectionFrame faceDetectionFrame, GeometricRectangle bounds) {
        boolean result = false;
        if(faceDetectionFrame != null) {
            result = true;
            if(bounds != null) {
                result = bounds.isInside(faceDetectionFrame.getLeftUpperCorner()) &&
                        bounds.isInside(faceDetectionFrame.getRightUpperCorner()) &&
                        bounds.isInside(faceDetectionFrame.getLeftLowerCorner()) &&
                        bounds.isInside(faceDetectionFrame.getRightLowerCorner());
            }
        }
        return result;
    }
}
