/*
 * Copyright (c) 2023. Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 3.0 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You may obtain a copy of the License at
 *
 *       http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */

package com.frojasg1.applications.libmorphing.api.facedetector.helpers.facedetectionframe;

import com.frojasg1.libgeometry.algebra.helpers.VectorFunctions;
import com.frojasg1.applications.libmorphing.api.facedetector.helpers.FaceDetectionBasicHelper;
import com.frojasg1.applications.libmorphing.api.facedetector.pojo.FaceDetection;
import com.frojasg1.applications.libmorphing.api.facedetector.pojo.FaceDetectionFrame;
import com.frojasg1.libgeometry.types.Point2d;
import com.frojasg1.general.number.DoubleFunctions;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.function.BiFunction;
import java.util.function.UnaryOperator;

public class FaceDetectionFrameAngleCorrector implements UnaryOperator<FaceDetectionFrameCalculatorContext> {
    private static final Logger LOGGER = LoggerFactory.getLogger(FaceDetectionFrameAngleCorrector.class);

    private static final double ANGLE_TOLERANCE = Math.PI / 720d;
    protected static double MAX_CORRECTION_ANGLE = Math.PI / 6d;

    protected static FaceDetectionFrameAngleCorrector INSTANCE = new FaceDetectionFrameAngleCorrector();

    public static FaceDetectionFrameAngleCorrector instance() {
        return INSTANCE;
    }

    protected BiFunction<FaceDetectionFrame, Double, FaceDetectionFrame> rotator;

    public FaceDetectionFrameAngleCorrector() {
        rotator = this::particularApplyAngle2;
    }

    protected double calculateVerticalAngle(Point2d up, Point2d down) {
        Point2d vector = down.subtract(up);

        return Math.atan2(vector.x, vector.y);
    }

    protected double calculateHorizontalAngle(Point2d left, Point2d right) {
        Point2d vector = right.subtract(left);

        return Math.atan2(-vector.y, vector.x);
    }

    protected double calculateVerticalAngle(FaceDetectionFrame frame) {
        return calculateVerticalAngle(frame.getLeftUpperCorner(), frame.getLeftLowerCorner());
    }

    protected double calculateHorizontalAngle(FaceDetectionFrame frame) {
        return calculateHorizontalAngle(frame.getLeftUpperCorner(), frame.getRightUpperCorner());
    }

    protected VectorFunctions getVectorFunctions() {
        return VectorFunctions.instance();
    }

    protected Point2d arithmeticMean(Point2d ... points) {
        return getVectorFunctions().arithmeticMean(points);
    }

    public Point2d applyAngle(Point2d vector, double angle) {
        return getVectorFunctions().rotate(vector, angle);
    }

    protected Point2d intersection(Point2d point1, Point2d directorVector1, Point2d point2, Point2d directorVector2) {
        return getVectorFunctions().intersection(point1, directorVector1, point2, directorVector2);
    }

    protected FaceDetectionBasicHelper getFaceDetectionBasicHelper() {
        return FaceDetectionBasicHelper.instance();
    }

    public boolean isValid(FaceDetection faceDetection, FaceDetectionFrame frame) {
        return getFaceDetectionBasicHelper().isValid(faceDetection, frame);
    }

    protected boolean isValid(FaceDetection faceDetection, FaceDetectionFrame frame, double angle) {
        return isValid(faceDetection, applyAngleInternal(frame, angle));
    }

    protected double calculateClosestApplicableAngle(FaceDetection faceDetection, FaceDetectionFrame frame, double angle) {
        double result = angle;

        double validLimit = 0d;
        double notValidLimit = angle;

        // if angle equals to zero of angle is valid, that is the correct result
        if(angle != 0 && !isValid(faceDetection, frame, notValidLimit)) {
            if(!isValid(faceDetection, frame, validLimit)) { // if well programmed, this condition would not be necessary, but let's calculate it just in case
                throw new IllegalArgumentException("Premise not fulfilled");
            }
            double mid;
            boolean isValid;
            while(!isClose(validLimit, notValidLimit)) {
                mid = (validLimit + notValidLimit) / 2.0d;
                isValid = isValid(faceDetection, frame, mid);
                if(isValid) {
                    validLimit = mid;
                } else {
                    notValidLimit = mid;
                }
            }
            result = validLimit;
        }

        return result;
    }

    public FaceDetectionFrame applyAngle(FaceDetection faceDetection, FaceDetectionFrame frame, double angle) {
        double closestApplicableAngle = angle;
        FaceDetectionFrame result = applyAngleInternal(frame, angle);
        if(!isValid(faceDetection, result)){
            closestApplicableAngle = calculateClosestApplicableAngle(faceDetection, frame, angle);

            result = applyAngleInternal(frame, closestApplicableAngle);
        }

        LOGGER.info("Asked for correction angle: {} degrees   ----->   applied angle: {} degrees",
                Math.toDegrees(angle), Math.toDegrees(closestApplicableAngle));
        return result;
    }

    protected Point2d getOrthogonalVector(Point2d vector) {
        return getVectorFunctions().getOrthogonalVector(vector);
    }

    protected FaceDetectionFrame applyAngleInternal(FaceDetectionFrame frame, double angle) {
        return rotator.apply(frame, angle);
    }

    protected FaceDetectionFrame particularApplyAngle1(FaceDetectionFrame frame, double angle) {
        FaceDetectionFrame result = frame.zoom(1.0d);

        Point2d horizontalVector = applyAngle(frame.getRightUpperCorner().subtract(frame.getLeftUpperCorner()), angle);
        Point2d verticalVector = getOrthogonalVector(horizontalVector);

        Point2d midUp = arithmeticMean(frame.getLeftUpperCorner(), frame.getRightUpperCorner());
        Point2d midDown = arithmeticMean(frame.getLeftLowerCorner(), frame.getRightLowerCorner());
        Point2d midLeft = arithmeticMean(frame.getLeftUpperCorner(), frame.getLeftLowerCorner());
        Point2d midRight = arithmeticMean(frame.getRightUpperCorner(), frame.getRightLowerCorner());

        result.setLeftUpperCorner(intersection(midLeft, verticalVector, midUp, horizontalVector));
        result.setLeftLowerCorner(intersection(midLeft, verticalVector, midDown, horizontalVector));
        result.setRightUpperCorner(intersection(midRight, verticalVector, midUp, horizontalVector));
        result.setRightLowerCorner(intersection(midRight, verticalVector, midDown, horizontalVector));

        return result;
    }

    protected FaceDetectionFrame particularApplyAngle2(FaceDetectionFrame frame, double angle) {
        return frame.rotate(angle);
    }

    protected DoubleFunctions getDoubleFunctions() {
        return DoubleFunctions.instance();
    }

    protected Double min(Double d1, Double d2) {
        return getDoubleFunctions().min(d1, d2);
    }

    protected Double max(Double d1, Double d2) {
        return getDoubleFunctions().max(d1, d2);
    }

    protected int sgn(double dd) {
        return getDoubleFunctions().sgn(dd);
    }

    protected boolean isClose(double value1, double value2) {
        return getDoubleFunctions().isClose(value1, value2, ANGLE_TOLERANCE);
    }

    protected double calculateCorrectionAngle(double angle1, double angle2) {
        double result = 0d;

        int sgn = sgn(angle1);
        if((sgn != sgn(angle2)) && !isClose(angle1, angle2)) { // a different sgn is acceptable if the angles are close to 0
            throw new IllegalArgumentException(String.format("Something wrong is happening. The sign of the angles differ (%f, %f)", angle1, angle2));
        }

        if(sgn == 1) {
            result = min(min(angle1, angle2), MAX_CORRECTION_ANGLE);
        } else if(sgn == -1) {
            result = max(max(angle1, angle2), -MAX_CORRECTION_ANGLE);
        }

        return result;
    }

    @Override
    public FaceDetectionFrameCalculatorContext apply(FaceDetectionFrameCalculatorContext context) {
        FaceDetectionFrame result = null;
        FaceDetection faceDetection = context.getFaceDetection();
        FaceDetectionFrame frame = context.getResult();
        if(frame != null) {
            double verticalAngle = calculateVerticalAngle(frame);
            double horizontalAngle = calculateHorizontalAngle(frame);

            double correctionAngle = calculateCorrectionAngle(verticalAngle, horizontalAngle);
            double maxCorrectionAngle = context.getMaxAngleCorrectionInRadians();
            correctionAngle = getDoubleFunctions().limit(correctionAngle, -maxCorrectionAngle, maxCorrectionAngle);
            result = applyAngle(faceDetection, frame, correctionAngle);

            context.addAngle(correctionAngle);
            context.setResult(result);
        }

        return context;
    }
}
