/*
 * Copyright (c) 2023. Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 3.0 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You may obtain a copy of the License at
 *
 *       http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */

package com.frojasg1.applications.libmorphing.api.facedetector.helpers.facedetectionframe;

import com.frojasg1.libgeometry.algebra.helpers.VectorFunctions;
import com.frojasg1.applications.libmorphing.api.facedetector.pojo.FaceDetection;
import com.frojasg1.applications.libmorphing.api.facedetector.pojo.FaceDetectionFrame;
import com.frojasg1.libgeometry.types.Point2d;

import java.util.List;
import java.util.function.Function;

public class FaceDetectionFrameCalculator implements Function<FaceDetection, FaceDetectionFrame> {

    protected static FaceDetectionFrameCalculator INSTANCE = new FaceDetectionFrameCalculator();

    public static FaceDetectionFrameCalculator instance() {
        return INSTANCE;
    }

    protected VectorFunctions getVectorFunctions() {
        return VectorFunctions.instance();
    }

    protected Point2d arithmeticMean(Point2d ... points) {
        return getVectorFunctions().arithmeticMean(points);
    }

    protected Point2d getOrthogonalVector(Point2d originalDirectionVector, Point2d secondVector) {
        return getVectorFunctions().getOrthogonalProjection(originalDirectionVector, secondVector);
    }

    protected Point2d addVectors(Point2d ... points) {
        return getVectorFunctions().addVectors(points);
    }

    @Override
    public FaceDetectionFrame apply(FaceDetection faceDetection) {
        List<Point2d> points = faceDetection.getPoints();
        Point2d rightEar = points.get(0);
        Point2d leftEar = points.get(16);
        Point2d midChin = points.get(8);
        Point2d rightEyebrown = points.get(19);
        Point2d leftEyebrown = points.get(24);

        Point2d midEyebrown = arithmeticMean(rightEyebrown, leftEyebrown);

        Point2d midVerticalVector = midChin.subtract(midEyebrown);
        Point2d vectorToRight = getOrthogonalVector(midVerticalVector, rightEar.subtract(midEyebrown));
        Point2d vectorToLeft = getOrthogonalVector(midVerticalVector, leftEar.subtract(midEyebrown));

        // left side of the face, is right side of the frame, seen from the viewer's point of view.
        FaceDetectionFrame result = new FaceDetectionFrame();
        result.setRightUpperCorner(addVectors(midEyebrown, vectorToLeft));
        result.setLeftUpperCorner(addVectors(midEyebrown, vectorToRight));
        result.setLeftLowerCorner(addVectors(midChin, vectorToRight));
        result.setRightLowerCorner(addVectors(midChin, vectorToLeft));
        result.setCenter(arithmeticMean(midChin, midEyebrown));

        return result;
    }
}
