/*
 * Copyright (c) 2023. Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 3.0 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You may obtain a copy of the License at
 *
 *       http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */

package com.frojasg1.applications.libmorphing.api.facedetector.helpers.facedetectionframe;

import com.frojasg1.applications.libmorphing.api.facedetector.helpers.FaceDetectionBasicHelper;
import com.frojasg1.applications.libmorphing.api.facedetector.helpers.ShrinkerTarget;
import com.frojasg1.applications.libmorphing.api.facedetector.pojo.FaceDetection;
import com.frojasg1.applications.libmorphing.api.facedetector.pojo.FaceDetectionFrame;
import com.frojasg1.delaunay.triangulation.quadedge.builder.impl.QuadEdgeDelaunayBuilderImpl;
import com.frojasg1.libgeometry.types.GeometricRectangle;
import com.frojasg1.libgeometry.types.Point2d;
import com.frojasg1.general.number.DoubleFunctions;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.function.Predicate;

public class FaceDetectionFrameShrinker {
    private static final Logger LOGGER = LoggerFactory.getLogger(FaceDetectionFrameShrinker.class);

    protected static FaceDetectionFrameShrinker INSTANCE = new FaceDetectionFrameShrinker();

    public static FaceDetectionFrameShrinker instance() {
        return INSTANCE;
    }

    public FaceDetectionFrameCalculatorContext shrink(FaceDetectionFrameCalculatorContext context,
                                                      GeometricRectangle bounds,
                                                      double minFactor, double maxFactor, double tolerance,
                                                      ShrinkerTarget target) {
        FaceDetection faceDetection = context.getFaceDetection();
        FaceDetectionFrame faceDetectionFrame = context.getResult();

        Double restrictive = minFactor;
        double relaxed = maxFactor;
        restrictive = getMostRestrictiveAcceptableFactor(faceDetectionFrame, restrictive, relaxed, tolerance,
                frame -> isValid(faceDetection, frame));
        if(restrictive == null) {
            throw new IllegalArgumentException("This should not happen if maxFactor is at least slightly greater than 1.0d");
        }

        GeometricRectangle boundsWithMargin = zoom(bounds, 0.95d);

        double min1 = restrictive;
        Double max1 = relaxed;

        double restrictive1 = max1;
        double relaxed1 = min1;

        max1 = getMostRestrictiveAcceptableFactor(faceDetectionFrame, restrictive1, relaxed1, tolerance,
                frame -> isInside(frame, boundsWithMargin));

        double factorResult;
        if(max1 == null) {
            factorResult = min1;
        } else if(target == ShrinkerTarget.SMALLEST) {
            factorResult = min1;
        } else if(target == ShrinkerTarget.GREATEST) {
            factorResult = max1;
        } else if(target == ShrinkerTarget.AVERAGE) {
            factorResult = (max1 + min1) / 2;
        } else {
            throw new IllegalArgumentException("target not recognized");
        }

        FaceDetectionFrame result = faceDetectionFrame.zoom(factorResult);

        LOGGER.info("Applied factor to the face detection frame: {}, with coordinates: {}", factorResult, result);

        context.setResult(result);
        context.applyFactor(factorResult);
        return context;
    }

    // let's suppose that if restrictiveFactor is suitable, then relaxedFactor is too.
    protected Double getMostRestrictiveAcceptableFactor(FaceDetectionFrame faceDetectionFrame,
                                                        double restrictiveFactor, double relaxedFactor, double tolerance,
                                                        Predicate<FaceDetectionFrame> isSuitableFunction) {
//        double maxFactor = 1.33d;
//        double minFactor = 1.05d;
//        double tolerance = 0.02d;

        Double result = restrictiveFactor;
        Double restrictiveResult = restrictiveFactor;
        Double relaxedResult = relaxedFactor;

        FaceDetectionFrame frameResult = faceDetectionFrame.zoom(restrictiveResult);
        if(!isSuitableFunction.test(frameResult)) {
            frameResult = faceDetectionFrame.zoom(relaxedResult);
            if(isSuitableFunction.test(frameResult)) {
                double midFactor;
                FaceDetectionFrame tmpFrame;
                while(Math.abs(restrictiveResult - relaxedResult) > tolerance) {
                    midFactor = (restrictiveResult + relaxedResult) / 2d;
                    tmpFrame = faceDetectionFrame.zoom(midFactor);
                    if(isSuitableFunction.test(tmpFrame)) {
                        relaxedResult = midFactor;
                    } else {
                        restrictiveResult = midFactor;
                    }
                }
                result = relaxedResult;
            } else {
                result = null;
            }
        }

        LOGGER.info("Most restrictive suitable factor to the face detection frame: {}, starting with: [{}, {}]",
                result, restrictiveFactor, relaxedFactor);

        return result;
    }

    protected <C> C[] toArray(C ... elems) {
        return elems;
    }

    protected FaceDetectionBasicHelper getFaceDetectionBasicHelper() {
        return FaceDetectionBasicHelper.instance();
    }

    public boolean isValid(FaceDetection faceDetection, FaceDetectionFrame frame) {
        return getFaceDetectionBasicHelper().isValid(faceDetection, frame);
    }

    public boolean isInside(FaceDetectionFrame faceDetectionFrame, GeometricRectangle bounds) {
        return getFaceDetectionBasicHelper().isInside(faceDetectionFrame, bounds);
    }

    protected QuadEdgeDelaunayBuilderImpl getQuadEdgeDelaunayBuilder() {
        return QuadEdgeDelaunayBuilderImpl.instance();
    }

    public Point2d createPoint(double xx, double yy) {
        return getQuadEdgeDelaunayBuilder().createPoint(xx, yy);
    }

    protected GeometricRectangle zoom(GeometricRectangle rectangle, double factor) {
        GeometricRectangle result = null;
        if(rectangle != null) {
            Point2d center = rectangle.center();

            result = rectangle.createCopy();
            result.width = round(rectangle.width * factor);
            result.height = round(rectangle.height * factor);
            result.x = (int) (center.x - result.width / 2);
            result.y = (int) (center.y - result.height / 2);
        }

        return result;
    }

    protected DoubleFunctions getDoubleFunctions() {
        return DoubleFunctions.instance();
    }

    public int round(double value) {
        return getDoubleFunctions().round(value);
    }
}
