/*
 * Copyright (c) 2023. Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 3.0 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You may obtain a copy of the License at
 *
 *       http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */

package com.frojasg1.applications.libmorphing.api.facedetector.pojo;

import com.frojasg1.libgeometry.algebra.helpers.VectorFunctions;
import com.frojasg1.libgeometry.types.Point2d;

import java.util.function.UnaryOperator;

public class FaceDetectionFrame {
    protected Point2d leftUpperCorner;
    protected Point2d rightUpperCorner;
    protected Point2d rightLowerCorner;
    protected Point2d leftLowerCorner;

    protected Point2d center;

    public Point2d getLeftUpperCorner() {
        return leftUpperCorner;
    }

    public void setLeftUpperCorner(Point2d leftUpperCorner) {
        this.leftUpperCorner = leftUpperCorner;
    }

    public Point2d getRightUpperCorner() {
        return rightUpperCorner;
    }

    public void setRightUpperCorner(Point2d rightUpperCorner) {
        this.rightUpperCorner = rightUpperCorner;
    }

    public Point2d getRightLowerCorner() {
        return rightLowerCorner;
    }

    public void setRightLowerCorner(Point2d rightLowerCorner) {
        this.rightLowerCorner = rightLowerCorner;
    }

    public Point2d getLeftLowerCorner() {
        return leftLowerCorner;
    }

    public void setLeftLowerCorner(Point2d leftLowerCorner) {
        this.leftLowerCorner = leftLowerCorner;
    }

    protected VectorFunctions getVectorFunctions() {
        return VectorFunctions.instance();
    }

    public double scalarProduct(Point2d vector1, Point2d vector2) {
        return getVectorFunctions().scalarProduct(vector1, vector2);
    }

    public Point2d rotate(Point2d originPt, Point2d destPt, double angle) {
        return getVectorFunctions().rotate(originPt, destPt, angle);
    }

    protected Point2d zoom(Point2d origin, Point2d destination, double factor) {
        return getVectorFunctions().zoom(origin, destination, factor);
    }
//
//    public Point2d getCenter() {
//        return GeometryHelper.instance().arithmeticMean(getLeftLowerCorner(), getLeftUpperCorner(),
//                getRightLowerCorner(), getRightUpperCorner());
//    }

    public Point2d getCenter() {
        return center;
    }

    public void setCenter(Point2d center) {
        this.center = center;
    }

    public double getWidth() {
        return getLeftUpperCorner().distance(getRightUpperCorner());
    }

    public double getHeight() {
        return getLeftUpperCorner().distance(getLeftLowerCorner());
    }

    public FaceDetectionFrame applyTransformation(UnaryOperator<Point2d> pointTransformation) {
        FaceDetectionFrame result = new FaceDetectionFrame();

        result.setLeftUpperCorner(pointTransformation.apply(getLeftUpperCorner()));
        result.setLeftLowerCorner(pointTransformation.apply(getLeftLowerCorner()));
        result.setRightUpperCorner(pointTransformation.apply(getRightUpperCorner()));
        result.setRightLowerCorner(pointTransformation.apply(getRightLowerCorner()));
        result.setCenter(pointTransformation.apply(getCenter()));

        return result;
    }

    public FaceDetectionFrame zoom(double factor) {
        return applyTransformation(pt -> zoom(getCenter(), pt, factor));
    }

    public FaceDetectionFrame rotate(double angle) {
        return applyTransformation(pt -> rotate(getCenter(), pt, angle));
    }

    protected boolean isStrictlyInsideHalfSpace(Point2d pt, Point2d a, Point2d b, Point2d c) {
        return getVectorFunctions().isStrictlyInsideHalfSpace(pt, a, b, c);
    }

    protected boolean isApproximatelyInsideHalfSpace(Point2d pt, Point2d a, Point2d b, Point2d c) {
        double tolerance = 1d;
        return getVectorFunctions().isApproximatelyInsideHalfSpace(pt, a, b, c, tolerance);
    }

    public boolean isStrictlyInside(Point2d pt) {
        return isStrictlyInsideHalfSpace(pt, getLeftUpperCorner(), getRightUpperCorner(), getLeftLowerCorner()) &&
                isStrictlyInsideHalfSpace(pt, getRightLowerCorner(), getRightUpperCorner(), getLeftLowerCorner());
    }

    public boolean isApproximatelyInside(Point2d pt) {
        return isApproximatelyInsideHalfSpace(pt, getLeftUpperCorner(), getRightUpperCorner(), getLeftLowerCorner()) &&
                isApproximatelyInsideHalfSpace(pt, getRightLowerCorner(), getRightUpperCorner(), getLeftLowerCorner());
    }

    @Override
    public String toString() {
        return "FaceDetectionFrame{" +
                "leftUpperCorner=" + leftUpperCorner +
                ", rightUpperCorner=" + rightUpperCorner +
                ", rightLowerCorner=" + rightLowerCorner +
                ", leftLowerCorner=" + leftLowerCorner +
                ", center=" + center +
                '}';
    }
}
