/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.applications.libmorphing.api.facedetector.runnable;

import com.frojasg1.applications.common.components.internationalization.window.InternationalizedWindow;
import com.frojasg1.applications.common.configuration.InternationalizedStringConf;
import com.frojasg1.applications.libmorphing.api.facedetector.FaceDetectionQuery;
import com.frojasg1.applications.libmorphing.api.facedetector.FaceDetectionQueryFactory;
import com.frojasg1.applications.libmorphing.api.facedetector.FaceDetectionQueryResult;
import com.frojasg1.applications.libmorphing.api.facedetector.pojo.ImageFaceDetection;
import com.frojasg1.applications.libmorphing.configuration.MorphingApplicationConfiguration;
import com.frojasg1.applications.libmorphing.language.LibMorphingInternationalTexts;
import com.frojasg1.applications.libmorphing.model.facedetector.ImageFaceDetectionPersistenceModel;
import com.frojasg1.applications.libmorphing.model.facedetector.ImageFaceDetectionPersistenceModelKey;
import com.frojasg1.applications.libmorphing.model.facedetector.ImageFaceDetectionPersistenceModelMap;
import com.frojasg1.general.desktop.queries.api.execution.GenericQueryWithResultExecution;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.function.Consumer;

/**
 *
 * @author fjavier.rojas
 */
public class FaceDetectionQueryExecution
		extends GenericQueryWithResultExecution<ImageFaceDetectionPersistenceModelKey, ImageFaceDetection, FaceDetectionQueryResult, FaceDetectionQuery>
{
	private static final Logger LOGGER = LoggerFactory.getLogger(FaceDetectionQueryExecution.class);

	public static final String FACE_DETECTION_QUERY_SERVICE_NAME = "FACE_DETECTION_QUERY_SERVICE_NAME";

	protected String imageFileName;

	protected ImageFaceDetectionPersistenceModelMap cache;

	public FaceDetectionQueryExecution(InternationalizedWindow parentJFrame,
									   String imageFileName, ImageFaceDetectionPersistenceModelMap cache,
									   MorphingApplicationConfiguration applicationConfiguration,
									   FaceDetectionQueryFactory queryFactory,
									   Consumer<ImageFaceDetection> callback ) {
		this(parentJFrame, imageFileName, cache, applicationConfiguration, queryFactory,
				LibMorphingInternationalTexts.instance(), callback);
	}

	public FaceDetectionQueryExecution(InternationalizedWindow parentJFrame,
									   String imageFileName, ImageFaceDetectionPersistenceModelMap cache,
									   MorphingApplicationConfiguration applicationConfiguration,
									   FaceDetectionQueryFactory queryFactory,
									   InternationalizedStringConf internationalStringsConf,
									   Consumer<ImageFaceDetection> callback ) {
		super(parentJFrame, applicationConfiguration::getUrlForFaceDetectionQuery, queryFactory::createFaceDetectionQuery,
				internationalStringsConf, callback);
		this.imageFileName = imageFileName;
		this.cache = cache;
	}

	public String getImageFileName() {
		return imageFileName;
	}

	protected ImageFaceDetectionPersistenceModelMap getCache() {
		return cache;
	}

	@Override
	protected ImageFaceDetectionPersistenceModelKey getKey() {
		return getCache().createKey(getImageFileName());
	}

	@Override
	protected ImageFaceDetection getCacheResult(ImageFaceDetectionPersistenceModelKey key) {
		return getIfNotNull(getCache().get(ImageFaceDetectionPersistenceModelMap.MAIN_MAP, key),
				ImageFaceDetectionPersistenceModel::getImageFaceDetection);
	}

	@Override
	public void run()
	{
		super.run();
	}

	protected ImageFaceDetectionPersistenceModel createImageFaceDetectionPersistenceModel(ImageFaceDetection imageFaceDetection) {
		ImageFaceDetectionPersistenceModel result = new ImageFaceDetectionPersistenceModel(getKey());
		result.setImageFaceDetection(imageFaceDetection);

		return result;
	}

	@Override
	protected void putInCache(ImageFaceDetection imageFaceDetection) {
		getCache().put(createImageFaceDetectionPersistenceModel(imageFaceDetection));
	}

	@Override
	protected String getServiceName() {
		return getInternationalString(FACE_DETECTION_QUERY_SERVICE_NAME);
	}

	@Override
	protected FaceDetectionQueryResult getQueryResult(FaceDetectionQuery query) {
		return runtimeExceptionFunctionExecution(() -> query.queryForApplication(getImageFileName()));
	}
}
