/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.applications.libmorphing.contexts;

import com.frojasg1.applications.libmorphing.configuration.MorphingApplicationConfiguration;
import com.frojasg1.applications.libmorphing.controller.MorphingVerticesMapManager;
import com.frojasg1.applications.libmorphing.model.facedetector.ImageFaceDetectionPersistenceModelMap;
import com.frojasg1.applications.libmorphing.model.images.configuration.MorphingProjectConfigurationModel;
import com.frojasg1.applications.libmorphing.model.images.image.MorphingImageModel;
import com.frojasg1.applications.libmorphing.xml.facedetector.FaceDetectionWholeFileModel;
import com.frojasg1.applications.libmorphing.xml.facedetector.manager.FaceDetectionWholeFileModelManagerBase;
import com.frojasg1.applications.libmorphing.xml.whole.MorphingProjectWholeFileModel;
import com.frojasg1.applications.libmorphing.xml.whole.manager.MorphingProjectWholeFileModelManager;
import com.frojasg1.applications.libmorphing.xml.whole.manager.MorphingProjectWholeFileModelManagerBase;
import com.frojasg1.delaunay.triangulation.quadedge.builder.QuadEdgeDelaunayBuilder;
import com.frojasg1.libgeometry.types.GeometricRectangle;
import com.frojasg1.general.ExecutionFunctions;
import com.frojasg1.general.FileFunctions;
import com.frojasg1.general.NullFunctions;
import com.frojasg1.general.desktop.startapp.impl.GenericDesktopInitContextImpl;
import org.xml.sax.SAXException;

import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.TransformerException;
import java.io.IOException;
import java.util.List;
import java.util.function.Function;

/**
 *
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 */
public abstract class MorphingEditorApplicationContextBase
		extends GenericDesktopInitContextImpl
	implements MorphingEditorApplicationContext
{
	protected static final String MY_FACE_DETECTION_XML_RELATIVE_FILENAME = "../xml/faceDetections.xml";

//	protected MorphingProjectWholeFileModel projectWholeModel;
//	protected FaceDetectionWholeFileModel faceDetectionWholeFileModel;

	protected QuadEdgeDelaunayBuilder _quadEdgeDelaunayBuilder;

	protected MorphingProjectWholeFileModelManager morphingProjectWholeFileModelManager;
	protected FaceDetectionWholeFileModelManagerBase faceDetectionWholeFileModelManager;


	public MorphingEditorApplicationContextBase init()
	{
		morphingProjectWholeFileModelManager = createMorphingProjectWholeFileModelManager();
		faceDetectionWholeFileModelManager = createFaceDetectionWholeFileModelManager();

		postInit();

		return this;
	}

	protected void postInit() {
		loadFaceDetections();
	}

	protected FileFunctions getFileFunctions() {
		return FileFunctions.instance();
	}

	protected String convertFolderSeparator(String compoundFolder) {
		return getFileFunctions().convertFolderSeparator(compoundFolder);
	}

	protected boolean isFile(String fileName) {
		return getFileFunctions().isFile(fileName);
	}

	protected String getFaceDetectionsXmlFilename() {
		return convertFolderSeparator(getAppliConf().getDefaultLanguageBaseConfigurationFolder() + "/" +
				MY_FACE_DETECTION_XML_RELATIVE_FILENAME);
	}

	protected FaceDetectionWholeFileModelManagerBase getFaceDetectionWholeFileModelManager() {
		return faceDetectionWholeFileModelManager;
	}

	protected void loadFaceDetections() {
		String fileName = getFaceDetectionsXmlFilename();
		if(isFile(fileName)) {
			getFaceDetectionWholeFileModelManager().loadFaceDetectionWholeModel(fileName);
		}
	}

	@Override
	public abstract MorphingApplicationConfiguration getAppliConf();

	@Override
	public QuadEdgeDelaunayBuilder getDelaunayBuilder() {
		return _quadEdgeDelaunayBuilder;
	}

	public void setDelaunayBuilder(QuadEdgeDelaunayBuilder _quadEdgeDelaunayBuilder) {
		this._quadEdgeDelaunayBuilder = _quadEdgeDelaunayBuilder;
	}

	@Override
	public MorphingProjectWholeFileModel getMorphingProjectModel() {
		return getProjectModelManager().getMorphingProjectModel();
	}

	@Override
	public FaceDetectionWholeFileModel getFaceDetectionWholeFileModel() {
		return getFaceDetectionWholeFileModelManager().getFaceDetectionWholeModel();
	}

	protected NullFunctions getNullFunctions() {
		return NullFunctions.instance();
	}

	public <CC, RR> RR getIfNotNull(CC obj, Function<CC, RR> getter) {
		return getNullFunctions().getIfNotNull(obj, getter);
	}

	@Override
	public ImageFaceDetectionPersistenceModelMap getImageFaceDetectionPersistenceModelMap() {
		return getIfNotNull(
						getFaceDetectionWholeFileModel(),
				FaceDetectionWholeFileModel::getImageFaceDetectionPersistenceModelMap);
	}

	protected MorphingProjectWholeFileModelManager getProjectModelManager() {
		return morphingProjectWholeFileModelManager;
	}

	protected MorphingProjectWholeFileModelManager createMorphingProjectWholeFileModelManager()
	{
		return(new MorphingProjectWholeFileModelManagerBase(this));
	}

	protected FaceDetectionWholeFileModelManagerBase createFaceDetectionWholeFileModelManager() {
		return new FaceDetectionWholeFileModelManagerBase();
	}

	@Override
	public void saveMorphingProjectModelAs(String fileName) throws IOException, TransformerException {
		runtimeExceptionMethodExecution( () ->
				getProjectModelManager().saveMorphingProjectModelAs(fileName));
	}

	@Override
	public MorphingProjectWholeFileModel loadMorphingProjectModel(String fileName)
			throws IOException, SAXException, ParserConfigurationException {
		return getProjectModelManager().loadProject(fileName);
	}

	@Override
	public MorphingVerticesMapManager getVerticesManager() {
		return getProjectModelManager().getVerticesManager();
	}

	protected ExecutionFunctions getExecutionFunctions() {
		return ExecutionFunctions.instance();
	}

	public void runtimeExceptionMethodExecution(ExecutionFunctions.UnsafeMethod run) {
		getExecutionFunctions().runtimeExceptionMethodExecution(run);
	}

	@Override
	public void saveFaceDetections() {
		runtimeExceptionMethodExecution( () ->
				getFaceDetectionWholeFileModelManager().saveFaceDetectionWholeModelAs(getFaceDetectionsXmlFilename()));
	}
	@Override
	public void exportFaceDetections(String fileName) {
		runtimeExceptionMethodExecution( () ->
				getFaceDetectionWholeFileModelManager().saveFaceDetectionWholeModelAs(fileName));
	}

	@Override
	public int importFaceDetections(String fileName) {
		return getFaceDetectionWholeFileModelManager().importFaceDetectionWholeModel(fileName);
	}

	@Override
	public boolean allHaveFaceDetection() {
		return getMorphingProjectModel().getImagesListWrapper().getList().stream()
				.allMatch(this::hasFaceDetection);
	}
	protected boolean hasFaceDetection(MorphingImageModel imageModel) {
		return getVerticesManager().getControlMeshManager().hasControlMesh(imageModel.getName());
	}

	@Override
	public List<MorphingImageModel> getImageModelList() {
		return getVerticesManager().getImageModelList().getList();
	}

	@Override
	public List<MorphingImageModel> getImageModelListCopy() {
		return getVerticesManager().getImageModelList().getListCopy();
	}

	@Override
	public MorphingImageModel getImageModel(String imageName) {
		return getVerticesManager().getImageModel(imageName);
	}

	@Override
	public MorphingImageModel addImage(String imageName, String imageFileName,
									   GeometricRectangle bounds,
                                       MorphingProjectConfigurationModel configurationModel,
									   boolean isMaster,
									   int index) {
		return getProjectModelManager().addImage(imageName, imageFileName, bounds, configurationModel, isMaster, index);
	}

	@Override
	public void relocateImages(List<MorphingImageModel> resultingList) {
		getProjectModelManager().relocateImages(resultingList);
	}

	@Override
	public void deleteImages(int[] indicesToBeDeleted, List<MorphingImageModel> resultingList) {
		getProjectModelManager().deleteImages(indicesToBeDeleted, resultingList);
	}
}
