/*
 * Copyright (c) 2023. Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 3.0 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You may obtain a copy of the License at
 *
 *       http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */

package com.frojasg1.applications.libmorphing.controller;

import com.frojasg1.libgeometry.algebra.helpers.GeometryHelper;
import com.frojasg1.libgeometry.algebra.helpers.VectorFunctions;
import com.frojasg1.applications.libmorphing.api.facedetector.constants.Landmark68constants;
import com.frojasg1.applications.libmorphing.contexts.MorphingEditorApplicationContext;
import com.frojasg1.applications.libmorphing.api.facedetector.helpers.Control88meshCreator;
import com.frojasg1.applications.libmorphing.api.facedetector.helpers.FaceDetectionMeshDataBuilderAndUpdater;
import com.frojasg1.applications.libmorphing.api.facedetector.pojo.FaceDetection;
import com.frojasg1.applications.libmorphing.api.facedetector.pojo.FaceDetectionFrame;
import com.frojasg1.applications.libmorphing.api.facedetector.pojo.FaceDetectionMeshData;
import com.frojasg1.applications.libmorphing.configuration.MorphingApplicationConfiguration;
import com.frojasg1.applications.libmorphing.mapper.ControlMeshMapper;
import com.frojasg1.applications.libmorphing.model.facedetector.ImageFaceDetectionPersistenceModel;
import com.frojasg1.applications.libmorphing.model.images.image.MorphingImageModel;
import com.frojasg1.applications.libmorphing.model.images.image.MorphingImageModelList;
import com.frojasg1.applications.libmorphing.model.images.vertex.VertexModel;
import com.frojasg1.applications.libmorphing.model.images.vertex.map.impl.control.DynamicMultimapVertexWithIndex;
import com.frojasg1.applications.libmorphing.model.images.vertex.map.impl.control.IndexedVerticesMap;
import com.frojasg1.applications.libmorphing.mvc.DelaunayTriangulatorView;
import com.frojasg1.applications.libmorphing.view.panels.evt.VertexEventData;
import com.frojasg1.applications.libmorphing.xml.facedetector.FaceDetectionWholeFileModel;
import com.frojasg1.applications.libmorphing.xml.whole.MorphingProjectWholeFileModel;
import com.frojasg1.applications.libmorphing.xml.whole.MorphingProjectWholeFileModelBase;
import com.frojasg1.libgeometry.types.GeometricRectangle;
import com.frojasg1.libgeometry.types.Point2d;
import com.frojasg1.general.NullFunctions;
import com.frojasg1.general.functional.interfaces.TriFunction;
import com.frojasg1.general.selectors.Nearest2DSelector;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.function.Function;
import java.util.function.UnaryOperator;

public class ControlMeshManager {
    private static final Logger LOGGER = LoggerFactory.getLogger(ControlMeshManager.class);

    // it is the possibly modified 68-landmark (+ 12 points + 8 points) mapped into the subimage domain
//    protected MorphingApplicationConfiguration appliConf;
    protected IndexedVerticesMap controlMeshVerticesMap;


    // it is the map with the original 68-landmark (+ 12 frame points) control detection of the original image.
    protected IndexedVerticesMap idealControlMeshVerticesMap;
    protected String masterImageName;

    protected int[][] controlMeshTriangles;

    protected MorphingProjectWholeFileModelBase morphingProjectWholeFileModel;

    protected Point2d _latestControlMovingPoint;

    protected MorphingEditorApplicationContext appliCtx;

    public ControlMeshManager(MorphingEditorApplicationContext appliCtx) {
        this.appliCtx = appliCtx;

        init();
    }

    protected void init() {
        controlMeshVerticesMap = createControlMeshVerticesMap();
        idealControlMeshVerticesMap = createControlMeshVerticesMap();
    }

    protected MorphingEditorApplicationContext getAppliCtx() {
        return appliCtx;
    }

    public MorphingApplicationConfiguration getAppliConf() {
        return getAppliCtx().getAppliConf();
    }

    protected Control88meshCreator getControl88meshCreator() {
        return Control88meshCreator.instance();
    }

    protected FaceDetectionMeshDataBuilderAndUpdater getFaceDetectionMeshDataBuilder() {
        return FaceDetectionMeshDataBuilderAndUpdater.instance();
    }

    protected IndexedVerticesMap createControlMeshVerticesMap() {
        return new IndexedVerticesMap();
    }

    public IndexedVerticesMap getControlMeshVerticesMap() {
        return controlMeshVerticesMap;
    }

    protected boolean exists( Point2d pt, String imageName )
    {
        return( getControlMeshVerticesMap().get(imageName, pt) != null );
    }

    protected Point2d emptyPointSelector(Point2d pt, String imageName) {
        return exists(pt, imageName) ? null : pt;
    }

    protected Nearest2DSelector getNearest2DSelector() {
        return Nearest2DSelector.instance();
    }

    protected <C> C getNearestSuitable(C elem, UnaryOperator<C> selector, UnaryOperator<C> limiter,
                                       TriFunction<C, Integer, Integer, C> deltaXdeltaYfunction, int maxDelta) {
        return getNearest2DSelector().getNearestSuitable(elem, selector, limiter, deltaXdeltaYfunction, maxDelta);
    }

    protected Point2d chooseTheNearestEmpty( Point2d initialValue,
                                             String imageName,
                                             UnaryOperator<Point2d> limiter)
    {
        Point2d tmp = getControl88meshCreator().createPoint2d(0, 0);
        return getNearestSuitable(initialValue.createIntCopyIfNotInt(),
                pt -> emptyPointSelector(pt, imageName), limiter,
                (pt, dx, dy) -> pt.applyDeltas(dx, dy, tmp), 5);
    }

    protected VectorFunctions getVectorFunctions() {
        return VectorFunctions.instance();
    }

    protected GeometryHelper getGeometryHelper() {
        return GeometryHelper.instance();
    }

    public Point2d limit(Point2d pt, GeometricRectangle rectangle) {
        return getGeometryHelper().limit(pt, rectangle);
    }

    public Point2d controlMeshMovingPoint( DelaunayTriangulatorView view, VertexEventData evtData )
    {
        String imageName = evtData.getImageName();

        Point2d result = moveControlMeshImageVertex(view, evtData);

        _latestControlMovingPoint = result;
        return result;
    }

    public void controlMeshApproachingPoint( DelaunayTriangulatorView view, String imageName,
                                                Point2d pt )
    {
        view.setControlMeshClosestPoint(imageName, pt);
    }

    protected Point2d getOriginalVertex( VertexEventData evtData, Point2d myDefault )
    {
        Point2d result = evtData.getOriginalVertex();
        if( result == null )
            result = myDefault;

        return( result );
    }

    protected Point2d getControlMeshOriginalVertex( VertexEventData evtData )
    {
        return getOriginalVertex(evtData, _latestControlMovingPoint);
    }


    protected Point2d moveControlMeshImageVertex(DelaunayTriangulatorView view,
                                                 VertexEventData evtData )
    {
        Point2d originalVertex = getControlMeshOriginalVertex(evtData);
        boolean confirmed = evtData.isConfirmed();
        Point2d pt = confirmed ? originalVertex : evtData.getVertex();
        String imageName = evtData.getImageName();

        Point2d result = null;
        DynamicMultimapVertexWithIndex multimapVertex =
                getControlMeshVerticesMap().get(imageName, originalVertex);
        if( multimapVertex != null )
        {
//            getControlMeshVerticesMap().getMapByName(imageName, VertexModel.class).remove(multimapVertex.get(imageName));
////            VertexModel vm = multimapVertex.get(imageName);
////            result = fixMappedPointToBeUniqueAndInsideBounds( pt.createCopy(), imageName, getBounds(imageName) );
////
////            vm.setPoint( result );
////
////            getControlMeshVerticesMap().put(multimapVertex);
            result = moveControlPoint(imageName, originalVertex, pt);

            if( !confirmed )
                view.setControlMeshClosestPoint(imageName, result);
            else
                view.setControlMeshClosestPoint(imageName, null);
        }
        else
        {
            LOGGER.warn( "Vertex ({}) attempted to erase not found for image: {}",
                    pt, imageName );
        }

        return( result );
    }

    protected Point2d fixMappedPointToBeUniqueAndInsideBounds(Point2d initialValue,
                                                              String imageName,
                                                              GeometricRectangle bounds)
    {
        UnaryOperator<Point2d> limiter = pt -> limit(pt, bounds).createIntCopyIfNotInt();
        Point2d result = limiter.apply(initialValue);

        if( exists( initialValue, imageName ) )
            result = chooseTheNearestEmpty(initialValue, imageName, limiter);

        return( result );
    }

    public IndexedVerticesMap getIdealControlMeshVerticesMap() {
        return idealControlMeshVerticesMap;
    }

    public String getMasterImageName() {
        return masterImageName;
    }

    public void setMasterImageName(String masterImageName) {
        if(!masterImageName.equals(getMasterImageName())) {
            this.masterImageName = masterImageName;

            updateMasterImage();
        }
    }

    protected void updateMasterImage() {
        FaceDetection faceDetection = getFaceDetection(getMasterImageName());
        int[][] controlMeshTriangles = createControlMeshTriangles(faceDetection);
        setControlMeshTriangles(controlMeshTriangles);
    }

    protected ControlMeshMapper getMapper() {
        return ControlMeshMapper.instance();
    }

    protected NullFunctions getNullFunctions() {
        return NullFunctions.instance();
    }

    public <CC, RR> RR getIfNotNull(CC obj, Function<CC, RR> getter) {
        return getNullFunctions().getIfNotNull(obj, getter);
    }

    protected MorphingImageModel getImageModel( String imageName ) {
        return getIfNotNull(getMorphingProjectWholeFileModel(), model -> model.getImageModel(imageName));
    }

    protected FaceDetection getFaceDetection(String imageName) {
        return getIfNotNull(getImageModel(imageName), MorphingImageModel::getFaceDetection);
    }

    protected GeometricRectangle getBounds(String imageName) {
        return getBounds(getImageModel(imageName));
    }

    protected GeometricRectangle getBounds(MorphingImageModel imageModel) {
        return getIfNotNull(imageModel, MorphingImageModel::getBounds);
    }

    public int[][] getControlMeshTriangles() {
        return controlMeshTriangles;
    }

    protected void setControlMeshTriangles(int[][] controlMeshTriangles) {
        this.controlMeshTriangles = controlMeshTriangles;
    }
//
//    public void createControl80mesh(FaceDetection faceDetection, String name, DelaunayRectangle bounds, IndexedVerticesMap result) {
//        getControl88meshCreator().createControl80mesh(faceDetection, name, bounds, result);
//    }

    protected void createIdealControl80mesh(FaceDetection faceDetection, String name, IndexedVerticesMap result,
                                            GeometricRectangle bounds) {
        getControl88meshCreator().createIdealControl80mesh(faceDetection, name, result, bounds);
    }

    protected void createParticularControl88mesh(FaceDetection faceDetection, GeometricRectangle bounds, String name, IndexedVerticesMap result) {
        getControl88meshCreator().createParticularControl88mesh(faceDetection, bounds, name, result);
    }

    protected int[][] createControlMeshTriangles(FaceDetection faceDetection) {
        return getControl88meshCreator().createControlMeshTriangles(faceDetection);
    }

    public void setBounds(String imageName, GeometricRectangle bounds) {
        updateFaceDetection(getImageModel(imageName), getFaceDetection(imageName), bounds);
    }

    public void setImageFaceDetection(MorphingImageModel imageModel, FaceDetection faceDetection) {
        updateFaceDetection(imageModel, faceDetection, getBounds(imageModel));
    }

    // not thread safe.
    // we should invokeLater, in order to assure sequentiality
    protected void updateFaceDetection(MorphingImageModel imageModel, FaceDetection faceDetection, GeometricRectangle bounds) {
//        MorphingImageModel imageModel = getMorphingProjectWholeFileModel().getImageModel(imageName);
        if(imageModel != null) {
//            &&
//                ((faceDetection != imageModel.getFaceDetection()) || !Objects.equals(bounds, imageModel.getBounds()))) {
            String imageName = imageModel.getName();

            getControlMeshVerticesMap().removeEntryByName(imageName);
            getIdealControlMeshVerticesMap().removeEntryByName(imageName);

            imageModel.setFaceDetection(faceDetection);
            imageModel.setBounds(bounds);
            if((faceDetection != null) && (bounds != null)) {
                createIdealControl80mesh(faceDetection, imageName, getIdealControlMeshVerticesMap(), bounds);
                createParticularControl88mesh(faceDetection, bounds, imageName, getControlMeshVerticesMap());
            }

            if(imageModel.isMaster()) {
                updateMasterImage();
            }
        } else if(imageModel == null) {
            LOGGER.warn("No imageModel with name {} was found", imageModel);
        } else {
            LOGGER.info("Nothing to change to image: {}", imageModel);
        }
    }

    public void removeImage(String imageName) {
        getIdealControlMeshVerticesMap().removeEntryByName(imageName);
        getControlMeshVerticesMap().removeEntryByName(imageName);
    }

    public MorphingProjectWholeFileModel getMorphingProjectWholeFileModel() {
        return getAppliCtx().getMorphingProjectModel();
    }
//
//    public void setMorphingProjectWholeFileModel(MorphingProjectWholeFileModelBase morphingProjectWholeFileModel) {
//        kk
//        this.morphingProjectWholeFileModel = morphingProjectWholeFileModel;
//    }

    public boolean canMap(String imageName) {
        return thereIsControlTriangulation() && hasControlMesh(imageName) && hasControlMesh(getMasterImageName());
    }

    protected boolean thereIsControlTriangulation() {
        return getControlMeshTriangles() != null;
    }

    public boolean hasControlMesh(String imageName) {
        return (getFaceDetection(imageName) != null) && (getControlMeshVerticesMap().mapExists(imageName));// &&
//                (getIdealControlMeshVerticesMap().mapExists(imageName));
    }

    public Point2d initialControlMap(Point2d masterPt, String imageName) {
        Point2d result = null;
        if(canMap(imageName)) {
            result = getMapper().map(masterPt, null, getControlMeshVerticesMap(), getMasterImageName(), imageName,
                    getControlMeshTriangles());
        }
        return result;
    }

    public Integer getControlVertexIndex(Point2d pt, String imageName) {
        return getIfNotNull( getControlMultiVertex(pt, imageName),
                DynamicMultimapVertexWithIndex::getIndex);
    }

    public DynamicMultimapVertexWithIndex getControlMultiVertex(Point2d pt, String imageName) {
        return getIfNotNull(getControlMeshVerticesMap(), multiMap -> multiMap.get(imageName, pt));
    }

    public boolean isAnyOfTheControlFrameCorners(Integer index) {
        // see 88-control.point.mesh.with.68-landmark.jpg
        return (index != null) && (index == 68) || (index == 71) || (index == 74) || (index == 77);
    }

    public boolean isAnyOfTheImageCorners(Integer index) {
        // see 88-control.point.mesh.with.68-landmark.jpg
        return (index != null) && (index == 80) || (index == 82) || (index == 84) || (index == 86);
    }

    public boolean canMoveControlPoint(Point2d pt, String imageName) {
        boolean result = false;
        Integer myIndex = getControlVertexIndex(pt, imageName);
        if(myIndex != null) {
            // see 88-control.point.mesh.with.68-landmark.jpg
            // not at bounds
            result = (myIndex < Landmark68constants.NB_POINTS_LANDMARK_80);
            if(result) {
                result = (myIndex < Landmark68constants.NB_POINTS_LANDMARK_68) // is any of 68-Landmark
                        || isAnyOfTheControlFrameCorners(myIndex); // is any of the controlFrame corners
            }
        }
        return result;
    }

    protected Point2d moveControlPointInternal(String imageName, Integer index, Point2d newPoint) {

        UnaryOperator<Point2d> limiter =
                pt -> getControl88meshCreator()
                        .getNearestEmptyPointRespectingThePreviousOne(
                                pt,
                                ind -> (index == null) || (ind == null) || (index == ind),
                                getBounds(imageName), imageName, getControlMeshVerticesMap());

        Point2d result = limiter.apply(newPoint);
        getControlMeshVerticesMap().set(imageName, index, buildVertexModel(result, getBounds(imageName)));

        return result;
    }

    protected VertexModel buildVertexModel(Point2d pt, GeometricRectangle bounds) {
        return VertexModel.build(pt, bounds);
    }

    protected boolean hasToRecalculateControlPoints(int index) {
        return isAnyOfTheControlFrameCorners(index);
    }

    protected void updateAutoCalculableControlPointsIfNecessary(String imageName, int index) {
        if(hasToRecalculateControlPoints(index)) {
            getControl88meshCreator().recalculateAllControlFramePoints(imageName, getControlMeshVerticesMap());
        }
    }

    public boolean isAnyOfTheControlFrameCorners(String imageName, Point2d originalPt) {
        boolean result = false;
        Integer index = getControlVertexIndex(originalPt, imageName);
        if(index != null) {
            result = isAnyOfTheControlFrameCorners(index);
        }
        return result;
    }

    public Point2d moveControlPoint(String imageName, Point2d originalPt, Point2d newPt) {
        Point2d result = null;
        Integer index = getControlVertexIndex(originalPt, imageName);
        if(index != null) {
            result = moveControlPointInternal(imageName, index, newPt);
            updateAutoCalculableControlPointsIfNecessary(imageName, index);
        }
        return result;
    }

    public List<VertexModel> createVertexList(String imageName) {
        List<VertexModel> result = null;

        if(getControlMeshVerticesMap().mapExists(imageName)) {
            Map<Integer, DynamicMultimapVertexWithIndex> myMap = getControlMeshVerticesMap().getMapByName(IndexedVerticesMap.MAP_BY_INDEX, Integer.class);
            result = new ArrayList<>();
            DynamicMultimapVertexWithIndex multiVertex;
            VertexModel vertex;
            int index = 0;
            while((multiVertex = myMap.get(index++)) != null) {
                vertex = multiVertex.get(imageName);
                if(vertex == null) {
                    break;
                }
                result.add(vertex);
            }
        }

        return result;
    }

    public List<FaceDetectionMeshData> getAllFacesFromFileName(String imageFileName) {
        return getIfNotNull(
                getImageFaceDetectionPersistenceModel(imageFileName),
                ImageFaceDetectionPersistenceModel::getDetectedFaceMeshes);
    }

    public ImageFaceDetectionPersistenceModel getImageFaceDetectionPersistenceModel(String imageFileName) {
        return getIfNotNull(
                        getIfNotNull(
                                getIfNotNull(getAppliCtx(),
                                        MorphingEditorApplicationContext::getFaceDetectionWholeFileModel),
                                FaceDetectionWholeFileModel::getImageFaceDetectionPersistenceModelMap),
                        map -> map.get(imageFileName));
    }

    public int getNumImages() {
        return getIfNotNull(
                getIfNotNull(getMorphingProjectWholeFileModel(), MorphingProjectWholeFileModel::getImagesListWrapper),
                MorphingImageModelList::size);
    }

//    protected FaceDetection createModifiedParticularFaceDetection(List<Point2d> points) {
//        FaceDetection result = new FaceDetection();
//        Iterator<Point2d> it = points.iterator();
//        Point2d pt;
//        for(int ii=0; ii < Landmark68constants.NB_POINTS_LANDMARK_68; ii++) {
//            pt = it.next();
//            result.addPoint(pt.createCopy());
//        }
//        return result;
//    }

    protected FaceDetection createModifiedParticularFaceDetection(String imageName) {
        FaceDetection result = new FaceDetection();
        Point2d pt;
        for(int ii=0; ii < Landmark68constants.NB_POINTS_LANDMARK_68; ii++) {
            pt = getControlMeshVerticesMap().get(ii).get(imageName).getPoint();
            result.addPoint(pt.createCopy());
        }
        return result;
    }

    public void resetDetectedFaceMeshFrame(String imageName, double maxFactor) {
        MorphingImageModel imageModel = getImageModel(imageName);
//        FaceDetection modifiedParticularFaceDetection = createModifiedParticularFaceDetection(imageModel.getFaceDetection88Points());
        FaceDetection modifiedParticularFaceDetection = createModifiedParticularFaceDetection(imageName);
        FaceDetectionFrame frame = getControl88meshCreator().getFaceDetectionFrameWithAngleCorrection(
                modifiedParticularFaceDetection, imageModel.getBounds(), maxFactor, 30);

        getControl88meshCreator().applyNewFrame(getControlMeshVerticesMap(), imageName, imageModel.getBounds(), frame);
    }
}
