/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.applications.libmorphing.controller;

import com.frojasg1.libgeometry.algebra.helpers.GeometryHelper;
import com.frojasg1.applications.libmorphing.model.images.image.MorphingImageModel;
import com.frojasg1.applications.libmorphing.model.images.vertex.VertexModel;
import com.frojasg1.libgeometry.types.GeometricRectangle;
import com.frojasg1.delaunay.triangulation.quadedge.DelaunaySubdivision;
import com.frojasg1.delaunay.triangulation.quadedge.builder.QuadEdgeDelaunayBuilder;
import com.frojasg1.delaunay.triangulation.quadedge.builder.impl.QuadEdgeDelaunayBuilderImpl;
import com.frojasg1.libgeometry.types.Point2d;
import com.frojasg1.general.view.ReleaseResourcesable;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

/**
 *
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 */
public class DelaunayTriangulatorController implements ReleaseResourcesable
{
	protected List<Point2d> _triangulationPoints;

	protected DelaunaySubdivision _triangulation;

	protected MorphingImageModel _masterImageModel;

	protected void revertLastInsertion()
	{
		if( (_triangulation != null ) && _triangulation.canRevert() )
			_triangulation.revertLastInsertion();
	}

	public List<Point2d> getTriangulationPoints() {
		return _triangulationPoints;
	}

	public DelaunaySubdivision getTriangulation() {
		return _triangulation;
	}

	public DelaunaySubdivision clearTriangulation()
	{
		_triangulation = createTriangulation();

		return( _triangulation );
	}

	public DelaunaySubdivision clearEverything()
	{
		getTriangulationPoints().clear();

		return( createTriangulation() );
	}

	public List<Point2d> getPossibleTriangles( Point2d masterVertex )
	{
		List<Point2d> result = null;
		DelaunaySubdivision triangulation = getTriangulation();
		if( triangulation != null )
			result = triangulation.getPossibleTriangles(masterVertex);

		return( result );
	}

	public DelaunaySubdivision processNewPoint( Point2d pt, boolean confirmed,
												List<Point2d> listResult )
	{
		DelaunaySubdivision result = processNewPointNotClearingReversion( pt, confirmed, listResult );

		clearReversion();

		return( result );
	}

	public DelaunaySubdivision processNewPointNotClearingReversion( Point2d rawPt, boolean confirmed,
																	List<Point2d> result )
	{
		if( //!confirmed &&
			( rawPt == null ) )
		{
			// erase vertex
			revertLastInsertion();
		}
		else
		{
//			if( ! confirmed )
			revertLastInsertion();

			Point2d pt = GeometryHelper.instance().limit(rawPt, getBounds());

			if(_triangulation.InsertSite(pt)) {
				if (result != null)
					result.add(pt);

				if (confirmed) {
					_triangulation.formalizeBoundsOfLatestInsertion(getBounds(), result);
					//				_triangulation.clearReversion();
				}
			}
		}

		return( _triangulation );
	}

	public void clearReversion()
	{
		_triangulation.clearReversion();
	}

	protected QuadEdgeDelaunayBuilder getBuilder()
	{
		return( QuadEdgeDelaunayBuilderImpl.instance() );
	}

	public MorphingImageModel getMasterImageModel() {
		return _masterImageModel;
	}

	public boolean setMasterImageModel(MorphingImageModel masterImageModel) {
		boolean result = ( (masterImageModel == null ) || ( masterImageModel != _masterImageModel ) );
		if( result )
			this._masterImageModel = masterImageModel;

		return( result );
	}

	protected List<VertexModel> getVertices( MorphingImageModel mim )
	{
		return( ( mim != null ) ? mim.getVertices() : null );
	}

	protected Set<Point2d> getVertices( DelaunaySubdivision triangulation )
	{
		Set<Point2d> result = new HashSet<>();
		getBuilder().browseEdges(triangulation, getBounds(),
			(pt1, pt2) -> { addPt( result, pt1 ); addPt( result, pt2 ); } );

		return( result );
	}

	protected void addPt( Set<Point2d> result, Point2d pt )
	{
		GeometricRectangle bounds = getBounds();
		if( ( bounds != null ) && ( bounds.isInside(pt) ) )
			result.add(pt);
	}

	public List<Point2d> updateTriangles()
	{
		return( addVertices( _triangulation, getVertices(getMasterImageModel()) ) );
	}

	// a list with the frame vertices is returned
	public Set<Point2d> createAndSetTriangulation()
	{
		DelaunaySubdivision triangulation = createTriangulation();
		Set<Point2d> frameVertices = getVertices(triangulation);

		_triangulation = triangulation;

		return( frameVertices );
	}

	protected List<Point2d> addVertices( DelaunaySubdivision triangulation,
										List<VertexModel> masterVertices )
	{
		List<Point2d> result = new ArrayList<>();

		if( ( triangulation != null ) && ( masterVertices != null ) )
		{
			Point2d pt;
			for( VertexModel vm: masterVertices )
			{
				pt = vm.getPoint();
//				triangulation.InsertSite( pt );

				processNewPoint( pt, true, result );
			}
		}

		_triangulationPoints = result;
		return( result );
	}

	protected GeometricRectangle getBounds()
	{
		return( getMasterImageModel().getBounds() );
	}

	protected DelaunaySubdivision createTriangulation()
	{
		DelaunaySubdivision result = null;
		if( getMasterImageModel() != null )
			result = getBuilder().createDelaunaySubdivision( getBounds() );

		return( result );
	}

	@Override
	public void releaseResources()
	{
	}
}
