/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.applications.libmorphing.controller;

import com.frojasg1.libgeometry.algebra.helpers.VectorFunctions;
import com.frojasg1.applications.common.configuration.imp.InternationalizedStringConfImp;
import com.frojasg1.libgeometry.algebra.helpers.GeometryHelper;
import com.frojasg1.applications.libmorphing.api.facedetector.constants.Landmark68constants;
import com.frojasg1.applications.libmorphing.contexts.MorphingEditorApplicationContext;
import com.frojasg1.applications.libmorphing.api.facedetector.pojo.FaceDetection;
import com.frojasg1.applications.libmorphing.api.facedetector.pojo.FaceDetectionMeshData;
import com.frojasg1.applications.libmorphing.helpers.MorphingHelper;
import com.frojasg1.applications.libmorphing.language.LibMorphingInternationalTexts;
import com.frojasg1.applications.libmorphing.model.facedetector.ImageFaceDetectionPersistenceModel;
import com.frojasg1.applications.libmorphing.model.images.color.ColorModel;
import com.frojasg1.applications.libmorphing.model.images.color.types.MultiTriangle;
import com.frojasg1.applications.libmorphing.model.images.image.MorphingImageModel;
import com.frojasg1.applications.libmorphing.model.images.image.MorphingImageModelList;
import com.frojasg1.applications.libmorphing.model.images.vertex.VertexModel;
import com.frojasg1.applications.libmorphing.model.images.vertex.map.impl.control.DynamicMultimapVertexWithIndex;
import com.frojasg1.applications.libmorphing.model.images.vertex.map.impl.control.IndexedVerticesMap;
import com.frojasg1.applications.libmorphing.model.images.vertex.map.impl.normal.MorphingVerticesMap;
import com.frojasg1.applications.libmorphing.model.images.vertex.map.MorphingVerticesMapMapEntryContext;
import com.frojasg1.applications.libmorphing.mvc.DelaunayTriangulatorView;
import com.frojasg1.applications.libmorphing.view.panels.evt.enums.VertexEventType;
import com.frojasg1.applications.libmorphing.view.panels.evt.VertexEvent;
import com.frojasg1.applications.libmorphing.view.panels.evt.VertexEventData;
import com.frojasg1.delaunay.triangulation.quadedge.DelaunayQuadEdgeHelpers;
import com.frojasg1.delaunay.triangulation.quadedge.DelaunaySubdivision;
import com.frojasg1.libgeometry.types.GeometricRectangle;
import com.frojasg1.libgeometry.types.Point2d;
import com.frojasg1.libgeometry.types.factory.GeometryCommonFactory;
import com.frojasg1.general.ExecutionFunctions;
import com.frojasg1.general.FileFunctions;
import com.frojasg1.general.NullFunctions;
import com.frojasg1.general.collection.functions.ListOperationFunctions;
import com.frojasg1.general.collection.functions.ListStateContextBase;
import com.frojasg1.general.dialogs.DialogsWrapper;
import com.frojasg1.general.functional.interfaces.TriFunction;
import com.frojasg1.general.list.ListWrapper;
import com.frojasg1.general.map.multimap.dynamic.deductible.DeductibleDynamicMultiMapElementBase;
import com.frojasg1.general.number.DoubleFunctions;
import com.frojasg1.general.number.IntegerFunctions;
import com.frojasg1.general.selectors.Nearest2DSelector;
import com.frojasg1.generic.GenericFunctions;
import java.awt.image.BufferedImage;
import java.io.File;
import java.lang.ref.Reference;
import java.lang.ref.WeakReference;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.Objects;
import java.util.Set;
import java.util.concurrent.atomic.AtomicReference;
import java.util.function.Function;
import java.util.function.UnaryOperator;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 *
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 */
public class MorphingVerticesMapManagerBase<PP> implements MorphingVerticesMapManager<PP> {
	private static final Logger LOGGER = LoggerFactory.getLogger(MorphingVerticesMapManagerBase.class);

	public static final String CONF_COULD_NOT_CREATE_IMAGES = "COULD_NOT_CREATE_IMAGES";

//	protected MorphingApplicationConfiguration applyConf;
	protected MorphingEditorApplicationContext appliCtx;

	protected MorphingImageModelList _imageModelList;

	protected DelaunayTriangulatorController _delaunayController;
	protected String _masterImageName;
	protected String _currentImageName;
	protected MorphingVerticesMap _verticesMap;

	protected MorphingHelper _morphingHelper;
	protected GeometryHelper _geometryHelper;

	protected InternationalizedStringConfImp _internationalizedStringConf = null;

	protected Point2d _latestMovingPoint;

	protected DeductibleDynamicMultiMapElementBase<VertexModel> _originalMasterMovedMultiVertex;

	protected ControlMeshManager controlMeshManager;

	protected boolean initialized;

	public MorphingVerticesMapManagerBase(MorphingEditorApplicationContext appliCtx,
										  String masterImageName)
	{
		this( appliCtx, masterImageName, LibMorphingInternationalTexts.instance(),
			MorphingHelper.instance(), GeometryHelper.instance(), true );
	}

	protected MorphingVerticesMapManagerBase(MorphingEditorApplicationContext appliCtx,
											 String masterImageName,
											 InternationalizedStringConfImp internationalizedStringConf,
											 MorphingHelper morphingHelper,
											 GeometryHelper geometryHelper,
											 boolean init )
	{
		this.appliCtx = appliCtx;
		_masterImageName = masterImageName;
		_internationalizedStringConf = internationalizedStringConf;
		_morphingHelper = morphingHelper;
		_geometryHelper = geometryHelper;

		if( init )
			init();
	}

	protected void init()
	{
		_delaunayController = createDelaunayController();
		_verticesMap = createVerticesMap();
		controlMeshManager = createControlMeshManager();
	}

//	@Override
//	public void setMorphingProjectWholeFileModel(MorphingProjectWholeFileModel morphingProjectWholeFileModel) {
//		getControlMeshManager().setMorphingProjectWholeFileModel(morphingProjectWholeFileModel);
//	}
//
//	protected MorphingApplicationConfiguration getApplyConf() {
//		return applyConf;
//	}


	protected boolean isInitialized() {
		return initialized;
	}

	protected void setInitialized(boolean initialized) {
		this.initialized = initialized;
	}

	protected MorphingEditorApplicationContext getAppliCtx() {
		return appliCtx;
	}

	protected ControlMeshManager createControlMeshManager() {
		return new ControlMeshManager(getAppliCtx());
	}

	@Override
	public DelaunayTriangulatorController getDelaunayController() {
		return _delaunayController;
	}

	@Override
	public MorphingVerticesMap getVerticesMap() {
		return _verticesMap;
	}

	@Override
	public String getMasterImageName() {
		return _masterImageName;
	}

	@Override
	public MorphingImageModel getMasterImageModel() {
		return( getImageModel( getMasterImageName() ) );
	}

	protected void updateMasterIsMaster( boolean isMaster )
	{
		MorphingImageModel masterImageModel = getMasterImageModel();
		if( masterImageModel != null )
			masterImageModel.setIsMaster( isMaster );
	}

	@Override
	public void setMasterImageName(String masterImageName) {
		if( !Objects.equals( masterImageName, this._masterImageName ) )
		{
			updateMasterIsMaster(false);
			_masterImageName = masterImageName;
			updateMasterIsMaster(true);

			MorphingImageModelList imageModelList = getImageModelList();
			if( getVerticesMap().getMapByName(masterImageName, VertexModel.class ) == null )
				updateVerticesMap( imageModelList, masterImageName, _verticesMap );

			getVerticesMap().setMasterName(masterImageName);

			getControlMeshManager().setMasterImageName(masterImageName);

			boolean masterHasChanged = getDelaunayController().setMasterImageModel( getMasterImageModel() );
			if( masterHasChanged )
				updateMaster();
		}
	}

	protected GeometricRectangle getTriangulationBounds() {
		return getIfNotNull(
				getIfNotNull(
						getDelaunayController(), DelaunayTriangulatorController::getTriangulation),
				DelaunaySubdivision::getBounds);
	}

	protected void updateMaster()
	{
		List<Point2d> allVertices = recreateTriangulation();

		boolean existed;
		String masterName = getMasterImageName();
		GeometricRectangle bounds = getBounds(masterName);
		DeductibleDynamicMultiMapElementBase<VertexModel> multiVer;
		for( Point2d pt: allVertices )
		{
			existed = ( getVerticesMap().get(masterName, pt) != null );
			if( ! existed )
			{
				multiVer = getVerticesMap().addNewElement(masterName, createVertexModel(pt, bounds, null));
				addMultiVertexToImageModels( multiVer );
			}
		}

		recreateTriangulation();
	}

	protected List<Point2d> recreateTriangulation()
	{
		List<Point2d> vertices = new ArrayList<>();
		Set<Point2d> frameVertices = getDelaunayController().createAndSetTriangulation();
		vertices.addAll(frameVertices);

		List<Point2d> allNonFrameVertices = getDelaunayController().updateTriangles();
		vertices.addAll( allNonFrameVertices );

		return( vertices );
	}

	protected MorphingVerticesMap createVerticesMap()
	{
		return( new MorphingVerticesMap( getMasterImageName() ) );
	}

	protected DelaunayTriangulatorController createDelaunayController()
	{
		return( new DelaunayTriangulatorController() );
	}

	@Override
	public void clearTriangulation()
	{
		DelaunaySubdivision triangulation = _delaunayController.clearEverything();

		for(MorphingImageModel imageModel: getImageModelList().getList() )
			imageModel.getVertices().clear();

		getVerticesMap().clear();
		updateMaster();
	}

	@Override
	public GeometricRectangle getBounds(String imageName)
	{
		MorphingImageModel im = getImageModel(imageName);

		return( (im == null) ? null : im.getBounds() );
	}
/*
	public void relocateImage( String imageName, int toIndex )
	{
		relocateImage( getImageIndex( imageName ), toIndex );
	}

	public void relocateImage( int fromIndex, int toIndex )
	{
		MorphingImageModel fromImageModel = getImageModelList().getList().remove(fromIndex);
		addImageModel(toIndex, fromImageModel);
	}
*/
	@Override
	public synchronized void relocateImages(List<MorphingImageModel> resultingList)
	{
		getImageModelList().replace( resultingList );
	}

	@Override
	public synchronized void deleteImages(int[] indicesToBeDeleted, List<MorphingImageModel> resultingList)
	{
		List<MorphingImageModel> imageListCopy = getImageModelList().getListCopy();
		List<MorphingImageModel> deletedImages =
			ListOperationFunctions.instance().getListOfSelectedRecordsRemovingThem( imageListCopy,
										indicesToBeDeleted );

		if( containsMaster( deletedImages ) )
			throw( new RuntimeException( "Cannot erase master image: " + getMasterImageName() +
										" change or unselect master image before asking to erase selection." ) );

		checkContentsEquals( imageListCopy, resultingList );

		getImageModelList().replace(imageListCopy);

		for( MorphingImageModel mim: deletedImages )
		{
			String imageName = mim.getName();
			removeImageModel(imageName);
		}
	}

	protected void removeImageModel(String imageName) {
		if( Objects.equals( imageName, getCurrentImageModelName() ) )
			setCurrentImageName( getMasterImageName() );

		getVerticesMap().removeEntryByName( imageName );
		getControlMeshManager().removeImage(imageName);
		getImageModelList().remove(getImageModel(imageName));
	}

	@Override
	public ControlMeshManager getControlMeshManager() {
		return controlMeshManager;
	}

	protected boolean containsMaster(List<MorphingImageModel> imageList )
	{
		return( imageList.stream().anyMatch( this::isMaster ) );
	}

	protected boolean isMaster( MorphingImageModel mim )
	{
		return( mim.getName().equals( getMasterImageName() ) );
	}

	protected <RR> void checkContentsEquals( List<RR> list1, List<RR> list2 )
	{
		if( ! ListOperationFunctions.instance().checkContentsEquals(list1, list2) )
			throw( new RuntimeException( "Concurrent modification" ) );
	}

	@Override
	public int getImageIndex(String imageName)
	{
		return( getImageIndex( getImageModel(imageName) ) );
	}

	@Override
	public int getImageIndex(MorphingImageModel imageModel)
	{
		return( getImageModelList().getList().indexOf( imageModel ) );
	}

	@Override
	public MorphingImageModel getImageModel(int index)
	{
		return( getImageModelList().getList().get(index) );
	}
/*
	public boolean removeImageModel( int index )
	{
		return( removeImageModel( getMasterImageModel( index ) ) );
	}

	public boolean removeImageModel( String imageName )
	{
		return( removeImageModel( getMasterImageModel(imageName) ) );
	}

	protected boolean removeImageModel( MorphingImageModel imageModel )
	{
		getVerticesMap().removeEntryByName( imageModel.getName() );

		return( getImageModelList().remove(imageModel) );
	}
*/
	@Override
	public synchronized void addImageModel(int index, MorphingImageModel imageModel)
	{
		MorphingVerticesMap verticesMap = getVerticesMap();
		addMapByName(imageModel, verticesMap);

		String masterImageName = getMasterImageName();
		if( masterImageName != null )
			for( VertexModel vm: getImageModel(masterImageName).getVertices() )
				imageModel.addVertex( verticesMap.get(masterImageName , vm).get( imageModel.getName() ) );

		getImageModelList().add(index, imageModel);
	}

	@Override
	public MorphingImageModel getImageModel(String imageName)
	{
		return( getImageModelList().getList().stream()
			.filter( im1 -> Objects.equals( im1.getName(), imageName) )
			.findFirst().orElse(null) );
	}

	@Override
	public MultiTriangle getMultiTriangle(Point2d pt, String imageName)
	{
		return( getMultiTriangle(pt, getImageModel(imageName) ));
	}

	protected Point2d moveImageVertexNoMaster( DelaunayTriangulatorView view,
											   VertexEventData evtData )
	{
		Point2d originalVertex = getOriginalVertex(evtData);
		boolean confirmed = evtData.isConfirmed();
		Point2d pt = confirmed ? originalVertex : evtData.getVertex();
		String imageName = evtData.getImageName();

		Point2d result = null;
		DeductibleDynamicMultiMapElementBase<VertexModel> multimapVertex =
				getVerticesMap().remove(imageName, originalVertex);
		if( multimapVertex != null )
		{
			VertexModel vm = multimapVertex.get(imageName);
			result = fixMappedPointToBeUniqueAndInsideBounds( pt.createCopy(), imageName );

			vm.setPoint( result );
			vm.setNormalizedPoint( toNormalized( result, getBounds( imageName ) ) );

			getVerticesMap().put(multimapVertex);

			if( !confirmed )
				view.setPointInProgress(imageName, result);
			else
				view.setPointInProgress(imageName, null);
		}
		else
		{
			LOGGER.warn( "Vertex ({}) attempted to erase not found for image: {}",
					pt, imageName );
		}

		return( result );
	}

	protected Point2d moveImageVertexMaster( DelaunayTriangulatorView view,
								VertexEventData evtData )
	{
		boolean isStart = (evtData.getOriginalVertex() != null);
		boolean confirmed = evtData.isConfirmed();
		Point2d originalVertex = getOriginalVertex(evtData);
		Point2d pt = confirmed ? originalVertex : evtData.getVertex();

		String imageName = getMasterImageName();
		Point2d result = null;

		if(isStart) {
			_originalMasterMovedMultiVertex = getVerticesMap().remove(imageName, originalVertex);
		}

		if( _originalMasterMovedMultiVertex != null ) {
			if(isStart) {
				removeMultiVertex(_originalMasterMovedMultiVertex);
				updateMaster();
			}

			MorphingImageModel masterImageModel = getMasterImageModel();

			result = fixMappedPointToBeUniqueAndInsideBounds( pt, getMasterImageName() );
			List<Point2d> addedVertices = ( confirmed ? new ArrayList<>() : null );

			// we have not to clear the reversion, as we will revert the last insertion
			// before inserting the "new" point in the map, in order that
			// the same point is not found in the triangulation, and it can be inserted
			// in the other maps.
			getDelaunayController().processNewPointNotClearingReversion(result, confirmed, addedVertices);

			if( confirmed )
			{
				// reverting last insertion
				getDelaunayController().revertLastInsertion();

				if(addedVertices.size() != 1) {
					LOGGER.warn("added {} vertices. Expected to add just one", addedVertices.size());
				}
				DeductibleDynamicMultiMapElementBase<VertexModel> multiVertex;
				for( Point2d masterPt: addedVertices )
				{
					if( Objects.equals( masterPt, result ) )
//					if( masterPt == result )
					{
						multiVertex = _originalMasterMovedMultiVertex;
//						multiVertex = getVerticesMap().remove(_originalMasterMovedMultiVertex);
//						if(multiVertex != _originalMasterMovedMultiVertex) {
//							LOGGER.warn("multivertex to modify not expected");
//						}

						VertexModel vm = multiVertex.get(imageName);
						vm.setPoint(result);
						vm.setNormalizedPoint(toNormalized(result, getBounds(imageName)));
						getVerticesMap().put(multiVertex);

						addMultiVertexToImageModels( multiVertex );
					}
//					else
//						multiVertex =
//							getVerticesMap().addNewElement(getMasterImageName(),
//								createVertexModel( masterPt, masterImageModel.getBounds(), null ) );

//					addMultiVertexToImageModels( multiVertex );
				}

				// inserting the point again.
				getDelaunayController().processNewPoint(result, confirmed, null);

				view.setPointInProgress(imageName, null);
				view.setClosestPoint(imageName, null);
				_originalMasterMovedMultiVertex = null;
			}
			else
				view.setPointInProgress(imageName, result);
		}
		else
		{
			LOGGER.warn( "Vertex ({}) attempted to erase not found for master image: {}",
				pt, imageName );
		}

		return( result );
	}

//	protected Point2d moveImageVertexMaster( DelaunayTriangulatorView view,
//											 VertexEventData evtData )
//	{
//		boolean isStart = (evtData.getOriginalVertex() != null);
//		boolean confirmed = evtData.isConfirmed();
//		Point2d originalVertex = getOriginalVertex(evtData);
//		Point2d pt = confirmed ? originalVertex : evtData.getVertex();
//
//		String imageName = getMasterImageName();
//		Point2d result = null;
//
//		if( ( _originalMasterMovedMultiVertex == null ) || isStart )
//		{
//			_originalMasterMovedMultiVertex = getVerticesMap().remove(imageName, originalVertex);
//		}
//
//		if( _originalMasterMovedMultiVertex != null )
//		{
//			if(isStart)
//				removeMultiVertex(_originalMasterMovedMultiVertex);
//
//			MorphingImageModel masterImageModel = getMasterImageModel();
//
//			result = fixMappedPointToBeUniqueAndInsideBounds( pt, getMasterImageName() );
//			List<Point2d> addedVertices = ( confirmed ? new ArrayList<>() : null );
//
//			// we have not to clear the reversion, as we will revert the last insertion
//			// before inserting the "new" point in the map, in order that
//			// the same point is not found in the triangulation, and it can be inserted
//			// in the other maps.
//			getDelaunayController().processNewPointNotClearingReversion(result, confirmed, addedVertices);
//
//			if( confirmed )
//			{
//				// reverting last insertion
//				getDelaunayController().revertLastInsertion();
//
//				DeductibleDynamicMultiMapElementBase<VertexModel> multiVertex;
//				for( Point2d masterPt: addedVertices )
//				{
//					if( Objects.equals( masterPt, result ) )
//					{
////						multiVertex = _originalMasterMovedMultiVertex;
//						multiVertex = getVerticesMap().remove(_originalMasterMovedMultiVertex);
//						if(multiVertex != _originalMasterMovedMultiVertex) {
//							LOGGER.warn("multivertex to modify not expected");
//						}
//
//						VertexModel vm = multiVertex.get(imageName);
//						vm.setPoint(result);
//						vm.setNormalizedPoint(toNormalized(result, getBounds(imageName)));
//						getVerticesMap().put(multiVertex);
//					}
//					else
//						multiVertex =
//								getVerticesMap().addNewElement(getMasterImageName(),
//										createVertexModel( masterPt, masterImageModel.getBounds(), null ) );
//
//					addMultiVertexToImageModels( multiVertex );
//				}
//
//				// inserting the point again.
//				getDelaunayController().processNewPoint(result, confirmed, null);
//
//				view.setPointInProgress(imageName, null);
//				view.setClosestPoint(imageName, null);
//				_originalMasterMovedMultiVertex = null;
//			}
//			else
//				view.setPointInProgress(imageName, result);
//		}
//		else
//		{
//			LOGGER.warn( "Vertex ({}) attempted to erase not found for master image: {}",
//					pt, imageName );
//		}
//
//		return( result );
//	}

	protected void approachingPoint( DelaunayTriangulatorView view, String imageName,
									 Point2d pt )
	{
		view.setClosestPoint(imageName, pt);
	}

	protected void pointInProgress( DelaunayTriangulatorView view, String imageName,
									Point2d pt )
	{
		view.setPointInProgress(imageName, pt);
	}

	protected Point2d getOriginalVertex( VertexEventData evtData, Point2d myDefault )
	{
		Point2d result = evtData.getOriginalVertex();
		if( result == null )
			result = myDefault;

		return( result );
	}

	protected Point2d getOriginalVertex( VertexEventData evtData )
	{
		return getOriginalVertex(evtData, _latestMovingPoint);
	}

	protected void movingPoint( DelaunayTriangulatorView view, VertexEventData evtData )
	{
		String imageName = evtData.getImageName();

		Point2d pt;
		if( getImageModel(imageName).isMaster() )
			pt = moveImageVertexMaster( view, evtData );
		else
			pt = moveImageVertexNoMaster( view, evtData );

		_latestMovingPoint = pt;
	}

	protected void deletePoint( VertexEventData evtData )
	{
		String imageName = evtData.getImageName();
		Point2d pt = evtData.getVertex();
		DeductibleDynamicMultiMapElementBase<VertexModel> multiVertex =
			getVerticesMap().get(imageName, pt);

		if( multiVertex == null )
		{
			boolean confirmed = evtData.isConfirmed();
			if( ! confirmed )
				getDelaunayController().processNewPoint(null, confirmed, null);
		}
		else {
			removeMultiVertex(multiVertex);
			updateMaster();
		}
	}

	protected boolean canRevert()
	{
		return( getIfNotNullOrElse( 
			getIfNotNull( getDelaunayController(),
							DelaunayTriangulatorController::getTriangulation ),
			DelaunaySubdivision::canRevert, false )
		);
	}

	protected void insertPoint( DelaunayTriangulatorView view,
								VertexEventData evtData )
	{
		MorphingImageModel masterImageModel = getMasterImageModel();
		Point2d pt = evtData.getVertex();
		boolean confirmed = evtData.isConfirmed();
		String imageName = evtData.getImageName();

//		if( !confirmed )
		if( canRevert() )
			getDelaunayController().revertLastInsertion();

		Point2d masterPoint = this.toMaster(pt, getImageModel(imageName) );

		insertMasterPointToTriangulation(masterImageModel, confirmed, masterPoint);

		if(confirmed) {
			view.setPointInProgress(imageName, null);
		}
	}

	private void insertMasterPointToTriangulation(MorphingImageModel masterImageModel, boolean confirmed,
												  Point2d masterPoint) {
		masterPoint = fixMappedPointToBeUniqueAndInsideBounds(masterPoint, getMasterImageName());

		List<Point2d> addedVertices = ( confirmed ? new ArrayList<>() : null );

		// to add the new vertex, without clearing reversion, so as it can be reverted
		getDelaunayController().processNewPointNotClearingReversion(masterPoint, confirmed, addedVertices);

		if(confirmed)
		{
			// so as the just inserted points be reverted,
			// as the new point is not inserted in multimap and if we do not do the reversion
			// then that point will not have a mapping to the other images.
			getDelaunayController().revertLastInsertion();

			DeductibleDynamicMultiMapElementBase<VertexModel> multiVertex;
			for( Point2d masterPt: addedVertices )
			{
				multiVertex =
					getVerticesMap().addNewElement(getMasterImageName(),
						createVertexModel( masterPt, masterImageModel.getBounds(), null ) );

				addMultiVertexToImageModels( multiVertex );
			}

			// to add again the new vertex, and clearing reversion
			getDelaunayController().processNewPoint(masterPoint, confirmed, null);
		}
	}

	@Override
	public void processNewPoint(DelaunayTriangulatorView view,
								VertexEvent<PP> vertexEvent)
	{
		MorphingImageModel masterImageModel = getMasterImageModel();
		VertexEventData evtData = vertexEvent.getEventData();
		Point2d pt = evtData.getVertex();
		boolean confirmed = evtData.isConfirmed();
		String imageName = evtData.getImageName();
		VertexEventType type = evtData.getType();

		if( type == VertexEventType.APPROACHING )
			approachingPoint( view, imageName, pt);
		else if( type == VertexEventType.DELETE )
			deletePoint( evtData );
		else if( type == VertexEventType.MOVING )
			movingPoint(view, evtData);
		else if(type == VertexEventType.CONTROL_MESH_VERTEX_APPROACHING) {
			getControlMeshManager().controlMeshApproachingPoint(view, imageName, pt);
		} else if(type == VertexEventType.CONTROL_MESH_VERTEX_MOVING) {
			Point2d movedPoint = getControlMeshManager().controlMeshMovingPoint(view, evtData);
			if(!evtData.isConfirmed()) {
				getControlMeshManager().controlMeshApproachingPoint(view, imageName, movedPoint);
			} else {
				modifyAllControlPoints(imageName, movedPoint);
			}
		}
		else
		{
			DeductibleDynamicMultiMapElementBase<VertexModel> multiVertex =
				getVerticesMap().get(imageName, pt);
			pt = round(pt);

			if( multiVertex == null )
			{
				if( !confirmed )
					pointInProgress( view, imageName, pt );

				if( type == VertexEventType.INSERT )
					insertPoint( view, evtData );

				view.updateTriangulation(imageName, null, false, null );//, getBounds(imageName));
			}
		}

		view.repaint();
		setCurrentImageName( imageName );
	}

//	protected boolean hasToUpdateMovedControlPoint(MorphingImageModel imageModel, Point2d movedPoint) {
//		String imageName = imageModel.getName();
//		Integer index = getControlMeshManager().getControlVertexIndex(movedPoint, imageName);
//		return (index < Landmark68constants.NB_POINTS_LANDMARK_68) &&
//				imageModel.getBounds().isInside(
//						getControlMeshManager().getIdealControlMeshVerticesMap().get(index).get(imageName).getPoint());
//	}

	protected boolean hasToUpdateMovedControlPoint(MorphingImageModel imageModel, int index) {
		boolean result = false;
		if(index < Landmark68constants.NB_POINTS_LANDMARK_68) {
			FaceDetectionMeshData faceDetectionMeshData = getFaceDetectionMeshData(imageModel);
			Point2d idealMeshPoint = null;
			if (faceDetectionMeshData != null) {
				idealMeshPoint = faceDetectionMeshData.getOriginalFaceDetection().getPoints().get(index);
			} else {
				LOGGER.warn("faceDetectionMeshData not found");
			}
			result = imageModel.getBounds().isInside(idealMeshPoint);
		}
		return result;
	}

	protected void modifyAllControlPoints(String imageName, Point2d movedPoint) {
		MorphingImageModel imageModel = getImageModel(imageName);
		DynamicMultimapVertexWithIndex multiVertex = getControlMeshManager().getControlMultiVertex(movedPoint, imageName);
		Integer index = multiVertex.getIndex();
		if(hasToUpdateMovedControlPoint(imageModel, index)) {
			updateIdealModifiedControlMeshPoint(imageModel, index, movedPoint);
			updateAllImagesControlPoint(imageModel, multiVertex, movedPoint);
		}
	}

	protected FaceDetectionMeshData getFaceDetectionMeshData(MorphingImageModel imageModel) {
		FaceDetectionMeshData result = null;
		ImageFaceDetectionPersistenceModel imageFaceDetectionPersistenceModel =
				getImageFaceDetectionPersistenceModel(imageModel);
		int imageFaceDetectionIndex;
		if((imageFaceDetectionPersistenceModel != null) &&
				((imageFaceDetectionIndex = imageModel.getImageFaceDetectionIndex()) > -1)) {
			result = getIfNotNull(
							getIfNotNull(
									imageFaceDetectionPersistenceModel,
									ImageFaceDetectionPersistenceModel::getDetectedFaceMeshes),
					faceDetectionMeshDataList -> faceDetectionMeshDataList.get(imageFaceDetectionIndex));
		}

		return result;
	}

	public ImageFaceDetectionPersistenceModel getImageFaceDetectionPersistenceModel(MorphingImageModel imageModel) {
		ImageFaceDetectionPersistenceModel result = null;
		if(imageModel != null) {
			result = getIfNotNull(
					getAppliCtx().getImageFaceDetectionPersistenceModelMap(),
					imageFaceDetectionPersistenceModelMap -> imageFaceDetectionPersistenceModelMap.get(imageModel.getFileName()));
		}

		return result;
	}

	protected void updateIdealModifiedControlMeshPoint(MorphingImageModel imageModel, int index, Point2d movedPoint) {
		FaceDetectionMeshData faceDetectionMeshData = getFaceDetectionMeshData(imageModel);
		if(faceDetectionMeshData != null) {
			faceDetectionMeshData.getModifiedFaceDetection().getPoints().set(index, movedPoint.createCopy());
		} else {
			LOGGER.warn("faceDetectionMeshData was null");
		}
	}

	protected FileFunctions getFileFunctions() {
		return FileFunctions.instance();
	}

	protected boolean areTheSame(String fileName1, String fileName2) {
		return getFileFunctions().areTheSame(fileName1, fileName2);
	}

	protected boolean areTheSameFace(MorphingImageModel imageModel1, MorphingImageModel imageModel2) {
		return getFileFunctions().areTheSame(imageModel1.getFileName(), imageModel2.getFileName()) &&
				Objects.equals(imageModel1.getImageFaceDetectionIndex(), imageModel2.getImageFaceDetectionIndex());
	}

	protected void updateAllImagesControlPoint(MorphingImageModel imageModel, DynamicMultimapVertexWithIndex multiVertex,
											   Point2d movedPoint) {
		for(String imageName1: multiVertex.keySet()) {
			MorphingImageModel imageModel1 = getImageModel(imageName1);
			if((imageModel1 != imageModel) &&
					areTheSameFace(imageModel, imageModel1) &&
					hasToUpdateMovedControlPoint(imageModel1, multiVertex.getIndex())) {
				getControlMeshManager().getControlMeshVerticesMap().set(imageModel1.getName(), multiVertex.getIndex(),
						buildVertexModel(movedPoint, imageModel1.getBounds()));
			}
		}
	}

	protected VertexModel buildVertexModel(Point2d pt, GeometricRectangle bounds) {
		return VertexModel.build(pt, bounds);
	}

	// at this function even imageModel is updated
	protected void updateAllImagesControlPoint(MorphingImageModel imageModel, int index, Point2d movedPoint) {
		for(String imageName1: getControlMeshManager().getControlMeshVerticesMap().getNames()) {
			MorphingImageModel imageModel1 = getImageModel(imageName1);
			if((imageModel1 != null) &&
					areTheSameFace(imageModel, imageModel1) &&
					hasToUpdateMovedControlPoint(imageModel1, index)) {
				getControlMeshManager().getControlMeshVerticesMap().set(imageModel1.getName(), index,
						buildVertexModel(movedPoint, imageModel1.getBounds()));
			}
		}
	}

	protected void removeMultiVertex( DeductibleDynamicMultiMapElementBase<VertexModel> multiVertex )
	{
		if( multiVertex == null )
			LOGGER.warn("Trying to remove null multivertex");
		else
		{
			MorphingVerticesMap verticesMap = getVerticesMap();
			VertexModel vm;
			String imageName;
			verticesMap.remove(multiVertex);
			for( MorphingImageModel imageModel: getImageModelList().getList() )
			{
				imageName = imageModel.getName();
				vm = multiVertex.get( imageName );
				if( ! imageModel.removeVertexModel( vm ) )
					LOGGER.warn( "Could not erase vertex: {} from image: {}", vm, imageName );
			}

			// to take into account the boundary points that appear automatically when inserting other points.
//			DeductibleDynamicMultiMapElementBase<VertexModel> multiVer;
//			List<Point2d> newVertexList = recreateTriangulation();
//			for(Point2d pt: newVertexList)
//			{
//				multiVer = verticesMap.get( getMasterImageName(), pt );
//				if( multiVer == null )
//				{
//					multiVer = verticesMap.addNewElement(getMasterImageName(),
//											createVertexModel(pt, getBounds(getMasterImageName()), null) );
//					addMultiVertexToImageModels( multiVer );
//				}
//			}
		}
	}

	protected void addMultiVertexToImageModels( DeductibleDynamicMultiMapElementBase<VertexModel> multiVertex )
	{
		if( multiVertex == null )
			LOGGER.warn("Trying to insert null multivertex");
		else
		{
			for( MorphingImageModel imageModel: getImageModelList().getList() )
				imageModel.addVertex( multiVertex.get( imageModel.getName() ) );
		}
	}

	protected Point2d round( Point2d pt )
	{
		return( getMorphingHelper().toIntCoordinates(pt) );
	}

	@Override
	public void setCurrentImageName(String _currentImageName) {
		this._currentImageName = _currentImageName;
	}

	protected MorphingImageModel getImageByIndex(int index)
	{
		return( getImageModelList().get(index) );
	}

	protected String getImageName( MorphingImageModel mim )
	{
		return( (mim != null ) ? mim.getName() : null );
	}

	@Override
	public void setCurrentImageIndex(int index) {
		setCurrentImageName( getImageName( getImageByIndex( index ) ) );
	}

	protected MorphingImageModel createImageModel()
	{
		return( new MorphingImageModel() );
	}

	protected MorphingImageModel createImageModel(String name,
												String imageFileName,
												GeometricRectangle bounds)
	{
		MorphingImageModel imageModel = createImageModel();

		imageModel.setName(name);
		imageModel.setBounds(bounds);
		imageModel.setFileName(imageFileName);

		return( imageModel );
	}

	protected VertexModel createVertexModel( Point2d pt, GeometricRectangle bounds,
											AtomicReference<ListWrapper<ColorModel>> colors )
	{
		return( getMorphingHelper().createVertexModel( pt, bounds, colors ) );
	}

	protected MorphingImageModel newImageModel( String name,
												String imageFileName,
												GeometricRectangle bounds,
												int index)
	{
		MorphingImageModel result = createImageModel(name,
													imageFileName,
													bounds);

		addImageModel(index, result);

		return( result );
	}

	@Override
	public MorphingImageModelList getImageModelList() {
		return _imageModelList;
	}

	@Override
	public void setImageModelList(MorphingImageModelList _imageModelList,
								  String masterImageName) {
		this._imageModelList = _imageModelList;

		initializeControlMeshMap();
		setMasterImageName(masterImageName);
	}

	protected void initializeControlMeshMap() {
		if(!isInitialized()) {
			IndexedVerticesMap controlMeshVerticesMap = getControlMeshManager().getControlMeshVerticesMap();
			if(controlMeshVerticesMap.getNames().size() > 1) {
				throw new RuntimeException("initializing, bug control mesh vertices map is not empty");
			}
			for (MorphingImageModel imageModel : getImageModelList().getList()) {
				List<Point2d> controlPoints = imageModel.getFaceDetection88Points();
				if(!controlPoints.isEmpty()) {
					String imageName = imageModel.getName();
					controlMeshVerticesMap.addMapByName(imageName);
					FaceDetectionMeshData faceDetectionMeshData = getFaceDetectionMeshData(imageModel);
					imageModel.setFaceDetection(faceDetectionMeshData.getOriginalFaceDetection());

					int index = 0;
					for(Point2d pt: controlPoints) {
						// just in case the modified control mesh was modified by another project.
						// then we should update all images face detection meshes.
						if(hasToUpdateMovedControlPoint(imageModel, index)) {
							Point2d pt1 = faceDetectionMeshData.getModifiedFaceDetection().getPoints().get(index);
							controlMeshVerticesMap.set(imageName, index,
									buildVertexModel(pt1, imageModel.getBounds()));
						} else {
							controlMeshVerticesMap.set(imageName, index, buildVertexModel(pt, imageModel.getBounds()));
						}

						index++;
					}
				}
			}
			setInitialized(true);
		}
	}

	protected void updateVerticesMap( MorphingImageModelList imageModelList,
										String masterImageName,
										MorphingVerticesMap verticesMap)
	{
		List<ListIterator<VertexModel>> iterators = new ArrayList<>();

		verticesMap.clearAll();
		for( MorphingImageModel mim: imageModelList.getList() )
		{
			iterators.add( mim.getVertices().listIterator() );
			addMapByName( mim, verticesMap );
		}

		DeductibleDynamicMultiMapElementBase<VertexModel> element;
		verticesMap.setMasterName(masterImageName);

		do
		{
			element = createVerticesMultimapElement(imageModelList, iterators);
			if( element != null )
				insertMultiVertex(imageModelList, iterators, verticesMap, element);

		} while( element != null );
	}

	protected void insertMultiVertex(MorphingImageModelList imageModelList,
									 List<ListIterator<VertexModel>> iterators,
									 MorphingVerticesMap verticesMap,
									 DeductibleDynamicMultiMapElementBase<VertexModel> multiVertex) {
		String imageName;
		VertexModel vm;
		Point2d suitablePt;
		Iterator<ListIterator<VertexModel>> itit = iterators.iterator();
		ListIterator<VertexModel> it;
		for(MorphingImageModel imageModel: imageModelList.getList()) {
			it = itit.next();
			imageName = imageModel.getName();
			vm = multiVertex.get(imageName);
			suitablePt = getDefaultNearestSuitable(verticesMap, imageModel, vm.getPoint());
			if(!Objects.equals(vm.getPoint(), suitablePt)) {
				vm = createVertexModel(suitablePt, imageModel.getBounds(), null);
				it.set(vm);
				multiVertex.put(imageName, vm);
			}
//			verticesMap.put(multiVertex);
		}

		verticesMap.put(multiVertex);

	}




	protected void addMapByName( MorphingImageModel mim, MorphingVerticesMap verticesMap )
	{
		verticesMap.addMapByName( mim.getName(), createVerticesMapElementContext(verticesMap, mim) );
	}

	protected MorphingVerticesMapMapEntryContext createVerticesMapElementContext(MorphingVerticesMap verticesMap,
																				 MorphingImageModel mim)
	{
		MorphingVerticesMapMapEntryContext result = new MorphingVerticesMapMapEntryContext( mim.getName() );

		WeakReference<MorphingVerticesMap> verticesMapRef = new WeakReference<>(verticesMap);
		result.setFromMasterDirectTranslator( mvm -> translateVertexModelDirectlyFromMaster( mvm, mim, verticesMapRef ) );
		result.setFromMasterTranslator( mvm -> translateVertexModelFromMaster( mvm, mim ) );
		result.setToMasterTranslator( vm -> translateVertexModelDirectlyToMaster( vm, mim ) );

		return( result );
	}

	protected MorphingHelper getMorphingHelper() {
		return _morphingHelper;
	}

	protected GeometryHelper getGeometryHelper() {
		return _geometryHelper;
	}

	protected VectorFunctions getVectorFunctions() {
		return VectorFunctions.instance();
	}

	protected VertexModel createVertexModel( AtomicReference<ListWrapper<ColorModel>> colorModelListWrapper )
	{
		return( new VertexModel(colorModelListWrapper) );
	}

	@Override
	public Point2d initialControlMap(Point2d masterPt, String imageName) {
		return getControlMeshManager().initialControlMap(masterPt, imageName);
	}

	protected VertexModel translateVertexModelDirectlyFromMaster(VertexModel masterVertexModel,
																 MorphingImageModel mim,
																 Reference<MorphingVerticesMap> verticesMapReference)
	{
		VertexModel result = createVertexModel(masterVertexModel.getColorsRef());

		Point2d myMappedPoint = initialControlMap(masterVertexModel.getPoint(), mim.getName());
		if(myMappedPoint == null) {
			myMappedPoint = fromNormalizedInt(masterVertexModel.getNormalizedPoint(), mim.getBounds() );
		}

		Point2d nearestSuitablePoint = getDefaultNearestSuitable(verticesMapReference.get(), mim, myMappedPoint);

		result.setPoint( nearestSuitablePoint );
		result.setNormalizedPoint( toNormalized( result.getPoint(), mim.getBounds() ) );

		return( result );
	}

	protected Point2d getDefaultNearestSuitable(MorphingVerticesMap verticesMap, MorphingImageModel mim,
												Point2d pt) {
		Point2d tmp = pt.createCopy();
		UnaryOperator<Point2d> selector = pt1 -> isPresent(pt1, mim.getName(), verticesMap) ? null : pt1;
		Point2d nearestSuitablePoint = getNearestSuitable(pt, selector, pt1 -> limit(pt1, mim.getBounds()),
				(pt1, dx, dy) -> (pt1 == null) ? null : pt1.applyDeltas(dx, dy, tmp), 5);

		return nearestSuitablePoint;
	}

	protected Point2d limit(Point2d pt, GeometricRectangle rectangle) {
		return getGeometryHelper().limit(pt, rectangle);
	}

	protected boolean isPresent(Point2d pt, String name, MorphingVerticesMap result) {
		return result.get(name, pt) != null;
	}

	protected VertexModel translateVertexModelDirectlyToMaster( VertexModel vertexModel,
																MorphingImageModel mim )
	{
		VertexModel result = createVertexModel(vertexModel.getColorsRef());

		result.setPoint( toMaster( vertexModel.getPoint(), mim ) );
		result.setNormalizedPoint( toNormalized( result.getPoint(), mim.getBounds() ) );

		return( result );
	}

	protected VertexModel translateVertexModelFromMaster( VertexModel masterVertexModel,
														MorphingImageModel mim )
	{
		VertexModel result = createVertexModel(masterVertexModel.getColorsRef());

		result.setPoint( fromMaster( masterVertexModel.getPoint(), mim ) );
		result.setNormalizedPoint( toNormalized( result.getPoint(), mim.getBounds() ) );

		return( result );
	}

	protected Point2d fromMaster( Point2d masterPt, MorphingImageModel mim )
	{
		List<Point2d> masterTrianglesUmbrella = getDelaunayController().getPossibleTriangles(masterPt);
		Point2d ma = masterTrianglesUmbrella.get(0);
		Point2d mb = masterTrianglesUmbrella.get(1);
		Point2d mc = masterTrianglesUmbrella.get(2);
		if( !isInside( masterPt, ma, mb, mc ) )
		{
			mc = masterTrianglesUmbrella.get(masterTrianglesUmbrella.size() - 1);
			if( ! isInside( masterPt, ma, mb, mc ) )
				LOGGER.warn( "Triangle that contains masterPoint not found. master point: {}, Umbrella: {}",
					masterPt, masterTrianglesUmbrella);
		}
/*
		LOGGER.info( "===" );
		LOGGER.info( "ma: {}", ma );
		LOGGER.info( "mb: {}", mb );
		LOGGER.info( "mc: {}", mc );
		LOGGER.info( "===" );
*/
		Point2d result = mapTriangle( masterPt, ma, mb, mc, getMasterImageName(), mim.getName() );
		result = fixMappedPointToBeUniqueAndInsideBounds( result, mim.getName() );

		return( result );
	}
//
//	protected Point2d toIntCoordinates( Point2d pt )
//	{
//		Point2d result = pt;
//		if( !coordinatesAreInt(pt) )
//			result = MorphingHelper.instance().toIntCoordinates( pt.createCopy() );
//
//		return( result );
//	}

	protected VertexModel getNearest( Point2d pt, MorphingImageModel mim )
	{
		AtomicReference<Double> min = new AtomicReference(Double.MAX_VALUE);
		return( mim.getVertices().stream()
					.reduce( null, ( vm1, vm2 ) -> getNearest( vm1, vm2, pt, min ) ) );
	}

	protected VertexModel getNearest( VertexModel vm1, VertexModel vm2, Point2d pt,
										AtomicReference<Double> min )
	{
		double squaredNorm = vm2.getPoint().squaredDistance(pt);
		VertexModel result = vm1;
		if( squaredNorm < min.get() )
		{
			min.set(squaredNorm);
			result = vm2;
		}
		return( result );
	}

	protected DeductibleDynamicMultiMapElementBase<VertexModel> getMultiVertexFromMaster( Point2d masterPt )
	{
		return( getVerticesMap().get( getMasterImageName(), masterPt ) );
	}

	protected MultiTriangle getMultiTriangle( Point2d pt, MorphingImageModel mim )
	{
		List<Point2d> masterTriangleAndImageTriangle = 
						getMasterAndImageTriangles( pt, mim );
		MultiTriangle result = null;

		if( masterTriangleAndImageTriangle != null )
		{
			result = new MultiTriangle();

			Point2d ma = masterTriangleAndImageTriangle.get(0);
			Point2d mb = masterTriangleAndImageTriangle.get(1);
			Point2d mc = masterTriangleAndImageTriangle.get(2);

			result.setMultiVertex1( getMultiVertexFromMaster(ma) );
			result.setMultiVertex2( getMultiVertexFromMaster(mb) );
			result.setMultiVertex3( getMultiVertexFromMaster(mc) );
		}

		return( result );
	}

	protected List<Point2d> getMasterAndImageTriangles( Point2d pt, MorphingImageModel mim )
	{
		List<Point2d> result = null;

		VertexModel nearest = getNearest( pt, mim );
		if( nearest != null )
		{
			Point2d masterNearest = vertexToMaster( nearest, mim.getName() ).getPoint();

			result = getMasterAndImageTriangles( masterNearest,
												nearest.getPoint(),
												pt, mim.getName() );
		}

		return( result );
	}

	protected Point2d toMaster( Point2d pt, MorphingImageModel mim )
	{
		Point2d result = null;
		if( mim != null )
		{
			if( mim.isMaster() )
				result = pt;
			else
			{
				List<Point2d> masterTriangleAndImageTriangle = 
								getMasterAndImageTriangles( pt, mim );

				if( masterTriangleAndImageTriangle != null )
				{
					Point2d ma = masterTriangleAndImageTriangle.get(0);
					Point2d mb = masterTriangleAndImageTriangle.get(1);
					Point2d mc = masterTriangleAndImageTriangle.get(2);
					Point2d ia = masterTriangleAndImageTriangle.get(3);
					Point2d ib = masterTriangleAndImageTriangle.get(4);
					Point2d ic = masterTriangleAndImageTriangle.get(5);

//					result = mapTrianglePt( nearest.getPoint(), ia, ib, ic, ma, mb, mc );
					result = mapTrianglePt( pt, ia, ib, ic, ma, mb, mc );
					result = fixMappedPointToBeUniqueAndInsideBounds( result, getMasterImageName() );
/*
					LOGGER.info( "ia: {}", ia);
					LOGGER.info( "ib: {}", ib);
					LOGGER.info( "iv: {}", ic);
					LOGGER.info( "pt: {}", pt);

					LOGGER.info( "ma: {}", ma);
					LOGGER.info( "mb: {}", mb);
					LOGGER.info( "mv: {}", mc);
					LOGGER.info( "result: {}", result);
*/
				}
				else
				{
					MorphingImageModel masterImageModel = getMasterImageModel();
					if( masterImageModel != null )
					{
						result = getMorphingHelper().toNormalized(pt, mim.getBounds() );
						result = getMorphingHelper().fromNormalized(result, masterImageModel.getBounds());
					}
				}
			}
		}

		return( round(result) );
	}

	protected Point2d fixMappedPointToBeUniqueAndInsideBounds( Point2d initialValue,
																String imageName )
	{
		return( fixMappedPointToBeUniqueAndInsideBounds( initialValue, imageName,
														getBounds( imageName ) ) );
	}

	protected Point2d fixMappedPointToBeUniqueAndInsideBounds( Point2d initialValue,
																String imageName,
																GeometricRectangle bounds)
	{
		Point2d result = round( bound( initialValue, bounds ) );

		if( exists( initialValue, imageName ) )
			result = chooseTheNearestEmpty( initialValue, imageName, bounds );

		return( result );
	}

	protected Point2d bound( Point2d initialValue, GeometricRectangle bounds )
	{
		return( getGeometryHelper().intAndBound(initialValue, bounds) );
	}

	protected double limit( double value, double lower, double upper )
	{
		return( DoubleFunctions.instance().limit(value, lower, upper) );
	}

	protected Nearest2DSelector getNearest2DSelector() {
		return Nearest2DSelector.instance();
	}

	protected <C> C getNearestSuitable(C elem, UnaryOperator<C> selector, UnaryOperator<C> limiter,
									   TriFunction<C, Integer, Integer, C> deltaXdeltaYfunction, int maxDelta) {
		return getNearest2DSelector().getNearestSuitable(elem, selector, limiter, deltaXdeltaYfunction, maxDelta);
	}

	protected boolean exists( Point2d pt, String imageName )
	{
		return( getVerticesMap().get(imageName, pt) != null );
	}
//
//	protected boolean coordinatesAreInt( Point2d pt )
//	{
//		return( MorphingHelper.instance().coordinatesAreInt(pt) );
//	}

	protected Point2d emptyPointSelector(Point2d pt, String imageName) {
		return exists(pt, imageName) ? null : pt;
	}

	protected Point2d chooseTheNearestEmpty( Point2d initialValue,
											 String imageName,
											 GeometricRectangle bounds)
	{
		return getNearestSuitable(initialValue.createIntCopyIfNotInt(),
				pt -> emptyPointSelector(pt, imageName), pt -> boundSame(pt, bounds),
				(pt, dx, dy) -> pt.applyDeltas(dx, dy, pt), 4);
	}

//	protected Point2d chooseTheNearestEmpty( Point2d initialValue,
//											 String imageName,
//											 DelaunayRectangle bounds)
//	{
//		initialValue = toIntCoordinates(initialValue);
//		Point2d res = initialValue.createCopy();
//		Point2d result = null;
//		for(int maxDelta = 1; ( result == null ) && ( maxDelta <= 4 ); maxDelta ++ )
//			result = chooseEmptyPoint(initialValue, imageName, bounds, res, maxDelta);
//
//		return( result );
//	}
//
//	protected Point2d chooseEmptyPoint( Point2d initialValue,
//										String imageName,
//										DelaunayRectangle bounds,
//										Point2d resultToReturn,
//										int maxDelta)
//	{
//		Point2d result = null;
//
//		resultToReturn.x = initialValue.x - maxDelta;
//		resultToReturn.y = initialValue.y - maxDelta;
//		boundSame( resultToReturn, bounds );
//		double fromX = resultToReturn.x;
//		double fromY = resultToReturn.y;
//
//		resultToReturn.x = initialValue.x + maxDelta;
//		resultToReturn.y = initialValue.y + maxDelta;
//		boundSame( resultToReturn, bounds );
//		double toX = resultToReturn.x;
//		double toY = resultToReturn.y;
//
//		// upper and lower lines
//		for( double xx = fromX; xx <= toX; xx += 1 )
//		{
//			resultToReturn.y = fromY;
//			resultToReturn.x = xx;
//			if( !exists(resultToReturn, imageName) )
//			{
//				result = resultToReturn;
//				break;
//			}
//
//			resultToReturn.y = toY;
//			if( !exists(resultToReturn, imageName) )
//				result = resultToReturn;
//		}
//
//		if( result == null )
//		{
//			fromY += 1;
//			toY -= 1;
//			// left and right lines
//			for( double yy = fromY; yy <= toY; yy += 1 )
//			{
//				resultToReturn.x = fromX;
//				resultToReturn.y = yy;
//				if( !exists(resultToReturn, imageName) )
//				{
//					result = resultToReturn;
//					break;
//				}
//
//				resultToReturn.x = toX;
//				if( !exists(resultToReturn, imageName) )
//					result = resultToReturn;
//			}
//		}
//		return( result );
//	}

	protected Point2d boundSame( Point2d result, GeometricRectangle bounds )
	{
		return( getGeometryHelper().boundSame(result, bounds) );
	}

	protected List<Point2d> getMasterAndImageTriangles( Point2d masterNearest,
														Point2d imageNearest,
														Point2d imagePt,
														String imageName )
	{
		Set<Point2d> processed = new HashSet<>();
		List<Point2d> result = getMasterAndImageTrianglesInternal( masterNearest, imageNearest,
													imagePt, imageName, 7, 0,
													processed );
		if( result == null )
			throw( new RuntimeException( "Triangle containing Point not found" ) );

		return( result );
	}

	protected List<Point2d> getMasterAndImageTrianglesBrowseLevelsInternal( Point2d imageNearest,
																			Point2d imagePt,
																			String imageName,
																			int maxLevel,
																			int targetLevel,
																			List<Point2d> masterTrianglesUmbrella,
																			Set<Point2d> processed)
	{
		List<Point2d> result = null;
		Iterator<Point2d> it = masterTrianglesUmbrella.iterator();
		it.next();
		while( it.hasNext() && ( result == null ) )
			result = getMasterAndImageTrianglesInternal( it.next(),
														imageNearest,
														imagePt,
														imageName,
														maxLevel,
														targetLevel,
														processed);

		return( result );
	}

	protected List<Point2d> getMasterAndImageTrianglesInternal( Point2d masterNearest,
																Point2d imageNearest,
																Point2d imagePt,
																String imageName,
																int maxLevel,
																int targetLevel,
																Set<Point2d> processed)
	{
		if( processed.contains(masterNearest) )
			return( null );

		processed.add(masterNearest);

		List<Point2d> masterTrianglesUmbrella = getDelaunayController().getPossibleTriangles(masterNearest);
		Point2d ma = masterTrianglesUmbrella.get(0);
		Point2d mb = masterTrianglesUmbrella.get(masterTrianglesUmbrella.size() - 1);
		Point2d mc = masterTrianglesUmbrella.get(1);
//		Point2d ia = imageNearest;
		Point2d ia = pointFromMaster( ma, imageName );
		Point2d ib = pointFromMaster( mb, imageName );
		Point2d ic = pointFromMaster( mc, imageName );

		List<Point2d> result = null;
		boolean ok = false;
		if( targetLevel == 0 )
		{
			Iterator<Point2d> it = masterTrianglesUmbrella.iterator();
			it.next();
			it.next();
//		LOGGER.info( "Point: {}", imagePt );
//		LOGGER.info( "Center: {}", ia );
//		LOGGER.info( "ib: {}", ib );
//		LOGGER.info( "ic: {}", ic );
			do
			{
				ok = isInside( imagePt, ia, ib, ic );
				if( ok || !it.hasNext() )
					break;

				mb = mc;
				mc = it.next();
				ib = ic;
				ic = pointFromMaster( mc, imageName );
//			LOGGER.info( "ic: {}", ic );
			}
			while( true );

			if( !ok && ( maxLevel <= 0 ) )
				result = null;
			else if( ! ok )
			{
				maxLevel--;
				while( ( result == null ) && ( targetLevel <= maxLevel ) )
					result = getMasterAndImageTrianglesBrowseLevelsInternal( imageNearest,
																			imagePt,
																			imageName,
																			maxLevel,
																			targetLevel++,
																			masterTrianglesUmbrella,
																			processed);
			}
			else // ok
			{
				result = masterTrianglesUmbrella;
				result.clear();
				// first, the master triangle
				result.add(ma);
				result.add(mb);
				result.add(mc);

				// after, the image triangle
				result.add(ia);
				result.add(ib);
				result.add(ic);
			}
		}
		else if( targetLevel > 0 )
		{
			result = getMasterAndImageTrianglesBrowseLevelsInternal( imageNearest,
																	imagePt,
																	imageName,
																	maxLevel--,
																	targetLevel--,
																	masterTrianglesUmbrella,
																	processed);
		}

		return( result );
	}



	protected Point2d triangleFromMaster( Point2d fromPt, Point2d fa, Point2d fb, Point2d fc,
											String toImageName )
	{
		return( mapTriangle( fromPt, fa, fb, fc, getMasterImageName(), toImageName ) );
	}

	protected Point2d triangleToMaster( Point2d fromPt, Point2d fa, Point2d fb, Point2d fc,
										String fromImageName )
	{
		return( mapTriangle( fromPt, fa, fb, fc, fromImageName, getMasterImageName() ) );
	}

	protected Point2d mapTriangle( Point2d fromPt, Point2d fa, Point2d fb, Point2d fc,
									String fromImageName, String toImageName )
	{
		Point2d tA = mapPoint( fa, fromImageName, toImageName );
		Point2d tB = mapPoint( fb, fromImageName, toImageName );
		Point2d tC = mapPoint( fc, fromImageName, toImageName );

		Point2d toPt;
		if( ( tA == null ) || ( tB == null ) || ( tC == null ) )
			toPt = directMap( fromPt, fromImageName, toImageName );
		else
			toPt = mapTrianglePt(fromPt, fa, fb, fc, tA, tB, tC);

		return( toPt );
	}

	protected Point2d toNormalizedPt( Point2d pt, String imageName )
	{
		return( getMorphingHelper().toNormalized(pt, getBounds(imageName) ) );
	}

	protected Point2d fromNormalizedPt( Point2d pt, String imageName )
	{
		return( getMorphingHelper().fromNormalized(pt, getBounds(imageName) ) );
	}

	protected Point2d directMap( Point2d fromPt, String fromImage, String toImage )
	{
		Point2d result = null;
		if( fromPt != null )
			result = fromNormalizedPt( toNormalizedPt( fromPt, fromImage ), toImage );

		return( result );
	}

	protected Point2d pointFromMaster( Point2d masterPt, String toImageName )
	{
		return( mapPoint( masterPt, getMasterImageName(), toImageName ) );
	}

	protected Point2d pointToMaster( Point2d pt, String fromImageName )
	{
		return( mapPoint( pt, fromImageName, getMasterImageName() ) );
	}

	protected VertexModel vertexFromMaster( VertexModel masterVert, String toImageName )
	{
		return( mapVertex( masterVert, getMasterImageName(), toImageName ) );
	}

	protected VertexModel vertexToMaster( VertexModel vert, String fromImageName )
	{
		return( mapVertex( vert, fromImageName, getMasterImageName() ) );
	}

	protected Point2d mapPoint( Point2d fromPt, String fromImageName, String toImageName )
	{
		Point2d result = null;
		DeductibleDynamicMultiMapElementBase<VertexModel> vv = getVerticesMap().get(fromImageName, fromPt);
		if( vv != null )
			result = vv.get(toImageName).getPoint();

		return( result );
	}

	protected VertexModel mapVertex( VertexModel fromVert, String fromImageName, String toImageName )
	{
		VertexModel result = null;
		DeductibleDynamicMultiMapElementBase<VertexModel> vv = getVerticesMap().get(fromImageName, fromVert);
		if( vv != null )
			result = vv.get(toImageName);

		return( result );
	}

	protected Point2d getNearest(Point2d pt, String name) {
		double min = Double.MAX_VALUE;
		Point2d result = null;
		double squaredDist;
		for(VertexModel vm: getVerticesMap().getMapByName(name, VertexModel.class).keySet()) {
			squaredDist = pt.squaredDistance(vm.getPoint());
			if(squaredDist < min) {
				result = vm.getPoint();
			}
		}
		return result;
	}

	protected Point2d mapTrianglePt( Point2d pt,
		Point2d pt11, Point2d pt12, Point2d pt13,
		Point2d pt21, Point2d pt22, Point2d pt23 )
	{
		return( getGeometryHelper().mapTrianglePt(pt, pt11, pt12, pt13, pt21, pt22, pt23) );
	}

	protected DelaunayQuadEdgeHelpers getDelaunayQuadEdgeHelpers() {
		return DelaunayQuadEdgeHelpers.instance();
	}

	// returns if the point pt is inside the triangle a, b, c
	protected boolean isInside( Point2d pt, Point2d a, Point2d b, Point2d c )
	{
		return( ( pt != null ) && ( a != null ) && ( b != null ) && ( c != null )&&
				(getDelaunayQuadEdgeHelpers().TriArea(a, b, c) >= 0) && // for discarding the case in which the angle is greater than 180 degrees
				getVectorFunctions().isInside(pt, a, b, c) );
	}

	protected Point2d fromNormalizedInt( Point2d percentagesPoint, GeometricRectangle rect )
	{
		return( getMorphingHelper().fromNormalizedInt(percentagesPoint, rect) );
	}

	protected Point2d toNormalized( Point2d pt, GeometricRectangle rect )
	{
		return( getMorphingHelper().toNormalized(pt, rect) );
	}

	protected DeductibleDynamicMultiMapElementBase<VertexModel> createVerticesMultimapElement( MorphingImageModelList imageModelList,
																								List<? extends Iterator<VertexModel>> iterators )
	{
		DeductibleDynamicMultiMapElementBase<VertexModel> result = new DeductibleDynamicMultiMapElementBase<>();
		Iterator<? extends Iterator<VertexModel>> it = iterators.iterator();

		Iterator<VertexModel> itElem;
		int index = 0;
		for( MorphingImageModel mim: imageModelList.getList() )
		{
			itElem = it.next();
			if(!itElem.hasNext())
			{
				if(index > 0)
					throw( new RuntimeException( "Different number of elements in Multimap maps" ) );
				else {
					while(it.hasNext()) {
						itElem = it.next();
						if(itElem.hasNext()) {
							throw( new RuntimeException("Different number of elements in Multimap maps"));
						}
					}
				}

				result = null;
				break;
			}
			index++;
			result.put(mim.getName(), itElem.next());
		}

		return( result );
	}

	@Override
	public void cloneImages(ListStateContextBase<MorphingImageModel> previousListState)
	{
		List<MorphingImageModel> list = previousListState.getList();
		for( int index: previousListState.getSelection() )
			cloneImageModel( list.get(index) );
	}

	protected MorphingImageModel cloneImageModel( MorphingImageModel imageModel )
	{
		MorphingImageModel newImageModel = imageModel.createCopy();
		newImageModel.setName( calculateImageModelName( newImageModel.getFileName() ) );
		newImageModel.setIsMaster(false);

		updateImageFaceDetection(newImageModel);

		addClonedImageModel( imageModel, newImageModel);

		return newImageModel;
	}

	protected synchronized void addClonedImageModel( MorphingImageModel imageModel,
												MorphingImageModel newImageModel)
	{
		MorphingVerticesMap verticesMap = getVerticesMap();
		addMapByName(newImageModel, verticesMap);

		String newImageName = newImageModel.getName();
		Map<VertexModel, DeductibleDynamicMultiMapElementBase<VertexModel>> map =
			getVerticesMap().getMapByName( newImageName, VertexModel.class );

		map.clear();

		String modelImageName = imageModel.getName();
		for( VertexModel vm: getImageModel(modelImageName).getVertices() )
		{
			DeductibleDynamicMultiMapElementBase<VertexModel> multiVertex = verticesMap.get(modelImageName , vm);
			VertexModel vertexModel = multiVertex.get( modelImageName );
			VertexModel clonedVertexModel = vertexModel.createCopy();
			multiVertex.put(newImageName, clonedVertexModel);

			map.put( clonedVertexModel, multiVertex );

//			newImageModel.addVertex( clonedVertexModel );
		}

		getImageModelList().add( getNumImages(), newImageModel );
	}

	@Override
	public void droppedFiles(List<File> droppedFileList, int indexWhereToInsert)
	{
		int numFailed = 0;
		for( File file: droppedFileList )
			if( !addImageFile( indexWhereToInsert++, file ) )
				numFailed++;

		showMessageNumImagesFailed( numFailed );
	}

	protected void showMessageNumImagesFailed( int numFailed )
	{
		if( numFailed > 0 )
		{
			String message = createCustomInternationalString(CONF_COULD_NOT_CREATE_IMAGES, numFailed);
			showError(message);
		}
	}

	protected DialogsWrapper getDialogsWrapper()
	{
		return( GenericFunctions.instance().getDialogsWrapper() );
	}

	protected void showError(String errorText )
	{
		getDialogsWrapper().showError( null, errorText );
	}

	protected String calculateImageModelName( String fileName )
	{
		String baseName = FileFunctions.instance().getBaseName(fileName);
		String nakedName = FileFunctions.instance().cutOffExtension(baseName);
		String result = nakedName;
		int ii = 1;
		while( existsImageName(result) )
			result = String.format( "%s%d", nakedName, ii++ );

		return( result );
	}

	protected boolean existsImageName( String name )
	{
		return( getImageModelList().get(name) != null );
	}

	protected MorphingImageModel createMorphingImageModel( File file )
	{
		MorphingImageModel result = new MorphingImageModel();
		String fileName = file.getAbsolutePath();
		result.setFileName( fileName );
		result.setName( calculateImageModelName( fileName ));
		BufferedImage image = result.getImage();
		result.setBounds( createBounds(image.getWidth(), image.getHeight()) );

		return( result );
	}

	protected GeometricRectangle createRectangle(double x, double y, double width, double height )
	{
		return( GeometryCommonFactory.instance().createDelaunayRectangle( x, y, width, height ) );
	}

	protected GeometricRectangle createBounds(int width, int height )
	{
		return( createRectangle(0, 0, width, height) );
	}

	protected boolean addImageFile( int indexWhereToInsert, File file )
	{
		MorphingImageModel imageModel = ExecutionFunctions.instance().safeFunctionExecution(
			() -> createMorphingImageModel( file ) );
		boolean result = ( imageModel != null );
		if(result)
		{
			addImageModel(indexWhereToInsert, imageModel);
//			getImageModelList().add(imageModel);

			if( getMasterImageModel() == null )
			{
				imageModel.setIsMaster(true);
				setMasterImageName( imageModel.getName() );
			}
		}

		return( result );
	}

	protected <CC, RR> RR getIfNotNull( CC obj, Function<CC,RR> getter )
	{
		return( NullFunctions.instance().getIfNotNull(obj, getter) );
	}

	protected <CC, RR> RR getIfNotNullOrElse( CC obj, Function<CC,RR> getter, RR defaultValue )
	{
		return( NullFunctions.instance().getIfNotNull(obj, getter, defaultValue) );
	}

	@Override
	public String getCurrentImageModelName()
	{
//		return( getIfNotNullOrElse( getDelaunayController().getMasterImageModel(),
//			MorphingImageModel::getName ) );
		if( _currentImageName == null )
			_currentImageName = firstImageName();

		return( _currentImageName );
	}

	protected MorphingImageModel getFirstImageModel()
	{
		MorphingImageModel result = null;
		if( _imageModelList != null )
		{
			List<MorphingImageModel> list = _imageModelList.getList();
			if( ! list.isEmpty() )
				result = list.get(0);
		}

		return( result );
	}

	protected String firstImageName()
	{
		String result = null;
		MorphingImageModel firstImageModel = getFirstImageModel();
		if( firstImageModel != null )
			result = firstImageModel.getName();
		
		return( result );
	}

	protected String getInternationalString(String label)
	{
		return( _internationalizedStringConf.getInternationalString(label ) );
	}

	protected String createCustomInternationalString(String label, Object... args) {
		return( _internationalizedStringConf.createCustomInternationalString( label, args ) );
	}

	@Override
	public void imageStart()
	{
		setCurrentImageIndexLimiting(0);
	}

	@Override
	public void imageEnd()
	{
		setCurrentImageIndexLimiting( getNumImages() - 1 );
	}

	@Override
	public void imagePrevious()
	{
		setCurrentImageIndexLimiting( getImageIndex() - 1 );
	}

	@Override
	public void imageNext()
	{
		setCurrentImageIndexLimiting( getImageIndex() + 1 );
	}

	protected int getImageIndex()
	{
		int result = -1;
		MorphingImageModel mim = getImageModelList().get( getCurrentImageModelName() );
		if( mim != null )
			result = getImageModelList().indexOf(mim);

		return( result );
	}

	protected int getNumImages()
	{
		return( getImageModelList().size() );
	}

	protected int limitImageIndex( int newIndex )
	{
		return( IntegerFunctions.limit(newIndex, -1, getNumImages() - 1) );
	}

	protected void setCurrentImageIndexLimiting( int newIndex )
	{
		int limitedIndex = limitImageIndex( newIndex );
		if( limitedIndex == -1 )
			setCurrentImageName( null );
		else
			setCurrentImageIndex( limitedIndex );
	}

	@Override
	public MorphingImageModel getCurrentImageModel()
	{
		return( getImageModel( getCurrentImageModelName() ) );
	}

	protected void updateImageFaceDetection(MorphingImageModel imageModel) {
		FaceDetection imageModelFaceDetection = imageModel.getFaceDetection();
		if(imageModelFaceDetection != null) {
			setImageFaceDetection(imageModel, imageModelFaceDetection);
		}
	}

	@Override
	public void setNewBounds(MorphingImageModel imageModel,
							 GeometricRectangle newBounds)
	{
		String imageName = imageModel.getName();

//		// update of vertices
//		List<VertexModel> list = imageModel.getVertices();
//
//		Map<VertexModel, DeductibleDynamicMultiMapElementBase<VertexModel>> map =
//			getVerticesMap().getMapByName( imageName, VertexModel.class );
//
//		Map<VertexModel, DeductibleDynamicMultiMapElementBase<VertexModel>> copyMap =
//			new HashMap<>(map);
//
//		map.clear();
//
//		Point2d oldPt;
//		Point2d newPt;
//		DeductibleDynamicMultiMapElementBase<VertexModel> multiVertice;
//		for( VertexModel vm: list )
//		{
//			oldPt = vm.getPoint();
//			multiVertice = copyMap.get(vm);
//			newPt = fromNormalizedInt( vm.getNormalizedPoint(), newBounds );
//			newPt = fixMappedPointToBeUniqueAndInsideBounds(newPt, imageName, newBounds);
//
//			vm.setPoint(newPt);
//			if( multiVertice == null )
//				LOGGER.warn( "Vertex ({}) not found when setting new bounds for image: {}",
//					oldPt, imageName );
//			else
//				map.put(vm, multiVertice);
//		}

		if(imageModel.isMaster()) {
			// if we have changed master bounds, we will do the trick to clone master to try to save vertex locations.
			MorphingImageModel masterClonation = cloneImageModel(imageModel);
			setMasterImageName(masterClonation.getName());

			updateMaster();

			// changing bounds and updating face detection
			imageModel.setBounds(newBounds);
			updateImageFaceDetection(imageModel);

			// updating the triangulation
			reAutoCalculateVerticesInternal(imageName);

			// setting back master to the original one
			setMasterImageName(imageName);
			removeImageModel(masterClonation.getName());

			updateMaster();
		} else {
			imageModel.setBounds(newBounds);
			updateImageFaceDetection(imageModel);
			reAutoCalculateVerticesInternal(imageName);
		}

		updateVerticesModel();
	}

	@Override
	public List<FaceDetectionMeshData> getAllFaces(String imageName) {
		MorphingImageModel imageModel = getImageModel(imageName);
		List<FaceDetectionMeshData> result = null;
		if(imageModel != null) {
			ImageFaceDetectionPersistenceModel imageFaceDetectionPersistenceModel =
					imageModel.getImageFaceDetectionPersistenceModel();
			if(imageFaceDetectionPersistenceModel == null) {
				imageFaceDetectionPersistenceModel = getIfNotNull(getControlMeshManager(),
						manager -> manager.getImageFaceDetectionPersistenceModel(imageModel.getFileName()));
				imageModel.setImageFaceDetectionPersistenceModel(imageFaceDetectionPersistenceModel);
			}
			if(imageFaceDetectionPersistenceModel != null) {
				result = imageFaceDetectionPersistenceModel.getDetectedFaceMeshes();
			}
		}
		return result;
	}

	public List<FaceDetectionMeshData> getAllFacesInternal(String imageName) {
		String imageFileName = getImageFileName(imageName);
		return getIfNotNull(getControlMeshManager(), manager -> manager.getAllFacesFromFileName(imageFileName));
	}

	@Override
	public String getImageFileName(String imageName) {
		return getIfNotNull(getImageModel(imageName), MorphingImageModel::getFileName);
	}

	@Override
	public void reAutoCalculateVertices(String imageName) {
		reAutoCalculateVerticesInternal(imageName);
		updateVerticesModel();
	}

	public void reAutoCalculateVerticesInternal(String imageName) {
		if(getVerticesMap().removeEntryByName(imageName)) {
			addMapByName(getImageModel(imageName), getVerticesMap());
		}
	}

	@Override
	public void reCalculateVerticesForAllImagesDifferentFromMaster() {
		MorphingImageModel morphingImageModel;
		Set<String> set = new HashSet<>(getVerticesMap().getNames());
		for(String imageName: set) {
			morphingImageModel = getImageModel(imageName);
			if((morphingImageModel != null) && !isMaster(morphingImageModel)) {
				reAutoCalculateVerticesInternal(imageName);
			}
		}
		updateVerticesModel();
	}

	@Override
	public void updateImageModelFaceDetectionMeshes() {
		IndexedVerticesMap controlVerticesMap = getControlMeshManager().getControlMeshVerticesMap();
		for(MorphingImageModel imageModel: getImageModelList().getList()) {
			imageModel.clearFaceDetectionPoints();
		}
		int index = 0;
		DynamicMultimapVertexWithIndex multiVertex;
		while((multiVertex = controlVerticesMap.get(index++)) != null) {
			for(Map.Entry<String, VertexModel> entry: multiVertex.entrySet()) {
				String imageName = entry.getKey();
				VertexModel vertexModel = entry.getValue();
				getImageModelList().get(imageName).addFaceDetectionPoint(vertexModel.getPoint().createCopy());
			}
		}
	}

	@Override
	public void setImageFaceDetection(MorphingImageModel imageModel, FaceDetection modifiedFaceDetection) {
		getControlMeshManager().setImageFaceDetection(imageModel, modifiedFaceDetection);

		updateImageModelFaceDetectionIndex(imageModel, modifiedFaceDetection);
	}

	@Override
	public void updateVerticesModel()
	{
		List<MorphingImageModel> list = getImageModelList().getListCopy();

		Map<String, DeductibleDynamicMultiMapElementBase<VertexModel>> map =
				getVerticesMap().getMapByName( getMasterImageName(), String.class );

		for( MorphingImageModel mim: list )
			mim.getVertices().clear();

		for( DeductibleDynamicMultiMapElementBase<VertexModel> valueMap: map.values() )
			for( MorphingImageModel mim: list )
				mim.getVertices().add( valueMap.get( mim.getName() ) );
	}

	protected FaceDetection getOriginalFaceDetection(MorphingImageModel imageModel) {
		return getIfNotNull(getFaceDetectionMeshData(imageModel), FaceDetectionMeshData::getOriginalFaceDetection);
	}

	public void resetModifiedDetectedFaceMesh(String imageName) {
		MorphingImageModel imageModel = getImageModel(imageName);
		FaceDetection originalFaceDetection = getOriginalFaceDetection(imageModel);

		int index = 0;
		for(Point2d pt: originalFaceDetection.getPoints()) {
			updateIdealModifiedControlMeshPoint(imageModel, index++, pt);
			updateAllImagesControlPoint(imageModel, index, pt);
		}
	}

	@Override
	public void resetDetectedFaceMeshFrame(String imageName) {
		resetDetectedFaceMeshFrame(imageName, 2.75d);
	}

	@Override
	public void resetDetectedFaceMeshFrame(String imageName, double maxFactor) {
		getControlMeshManager().resetDetectedFaceMeshFrame(imageName, maxFactor);
	}

	@Override
	public boolean isDelaunayTriangulationEmpty() {
		return getDelaunayController().getVertices(getMasterImageModel()).size() == 4;
	}

	@Override
	public boolean insertControlVerticesToTriangulation(String imageName) {
		boolean result = false;
		if( isDelaunayTriangulationEmpty() && getControlMeshManager().hasControlMesh(imageName)) {
			MorphingImageModel imageModel = getImageModel(imageName);
			if(imageModel.isMaster()) {
				boolean confirmed = true;
				Point2d masterPt;
				IndexedVerticesMap controlVerticesMap = getControlMeshManager().getControlMeshVerticesMap();
				for (int index = 0; index < Landmark68constants.NB_POINTS_LANDMARK_68; index++) {
//					if(!getControlMeshManager().isAnyOfTheImageCorners(index)) {
						masterPt = controlVerticesMap.get(index).get(getMasterImageName()).getPoint();
						insertMasterPointToTriangulation(imageModel, confirmed, masterPt);
//					}
				}
				reCalculateVerticesForAllImagesDifferentFromMaster();
				result = true;
			}
		}
		return result;
	}

	protected Integer getFaceDetectionIndex(MorphingImageModel imageModel, FaceDetection modifiedFaceDetection) {
		Integer result = null;
		ImageFaceDetectionPersistenceModel imageFaceDetectionPersistenceModel =
				getImageFaceDetectionPersistenceModel(imageModel);

		int index = 0;
		for(FaceDetectionMeshData faceDetectionMeshData: imageFaceDetectionPersistenceModel.getDetectedFaceMeshes()) {
			if(faceDetectionMeshData.getModifiedFaceDetection() == modifiedFaceDetection) {
				result = index;
				break;
			}
			index++;
		}

		return result;
	}

	protected void updateImageModelFaceDetectionIndex(MorphingImageModel imageModel, FaceDetection modifiedFaceDetection) {
//		MorphingImageModel imageModel = getImageModel(imageName);

		Integer faceDetectionIndex = getFaceDetectionIndex(imageModel, modifiedFaceDetection);
		if(faceDetectionIndex != null) {
			imageModel.setImageFaceDetectionIndex(faceDetectionIndex);
		} else {
			LOGGER.warn("Could not set face detection index. modifiedFaceDetection not found");
		}
	}
}
