/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.applications.libmorphing.frames;

import com.frojasg1.applications.libmorphing.contexts.MorphingEditorApplicationContext;
import com.frojasg1.applications.libmorphing.contexts.holder.ApplicationCtxHolder;
import com.frojasg1.applications.libmorphing.frames.context.FrameSequenceContext;
import com.frojasg1.applications.libmorphing.frames.framebuilder.MorphingFrameCreatorBase;
import com.frojasg1.applications.libmorphing.frames.framebuilder.impl.FaceDetectionTransitionMorphingFrameCreator;
import com.frojasg1.applications.libmorphing.frames.framebuilder.impl.FadingToBlack;
import com.frojasg1.applications.libmorphing.frames.framebuilder.impl.MorphingTransition;
import com.frojasg1.applications.libmorphing.frames.framebuilder.impl.NormalTransitionMorphingFrameCreator;
import com.frojasg1.applications.libmorphing.frames.runnable.MorphingFramesConsumtionPace;
import com.frojasg1.applications.libmorphing.model.images.image.MorphingImageModel;
import com.frojasg1.delaunay.triangulation.quadedge.DelaunaySubdivision;
import com.frojasg1.delaunay.triangulation.quadedge.helpers.DelaunayTriangleBrowser;
import com.frojasg1.general.getters.LazyHolder;
import com.frojasg1.libgeometry.types.GeometricRectangle;
import com.frojasg1.general.ExecutionFunctions;
import com.frojasg1.general.NullFunctions;
import com.frojasg1.general.executor.ExecutorInterface;
import com.frojasg1.general.executor.worker.PullOfExecutorWorkers;
import com.frojasg1.general.number.IntegerFunctions;
import com.frojasg1.libgeometry.types.GeometricTriangle;
import java.awt.image.BufferedImage;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.function.Consumer;
import java.util.function.Function;
import java.util.function.Predicate;
import java.util.function.Supplier;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 *
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 */
public class MorphingFrameSequencer extends ApplicationCtxHolder {
	private static final Logger LOGGER = LoggerFactory.getLogger(MorphingFrameSequencer.class);

	protected static final int NUM_STEPS_PER_TRANSITION = 120;

//	protected static final int NUM_MAX_PENDING_TASKS = 4;

//	protected MorphingVerticesMapManager _morphingVerticesManager;
//	protected NormalTransitionMorphingFrameCreator _frameCreator;

	protected LazyHolder<List<GeometricTriangle>> masterTrianglesGetter;

	protected PullOfExecutorWorkers _pullOfWorkers;

//	public MorphingFrameSequencer( MorphingVerticesMapManager morphingVerticesManager)
	public MorphingFrameSequencer( MorphingEditorApplicationContext appliCtx)
	{
		super(appliCtx);
//		_morphingVerticesManager = morphingVerticesManager;
//		_frameCreator = new MorphingFrameCreator( morphingVerticesManager );
//		_frameCreator = new NormalTransitionMorphingFrameCreator( appliCtx );
		_pullOfWorkers = createPullOfWorkers();
		masterTrianglesGetter = new LazyHolder<>(this::getMasterTriangles);
	}

	protected int max( int v1, int v2 )
	{
		return( IntegerFunctions.max(v1, v2) );
	}

	protected int limit( int value, int min, int max )
	{
		return( IntegerFunctions.limit(value, min, max) );
	}

	protected int getNumThreads()
	{
//		int numMaxThreads = ThreadFunctions.instance().getNumMaxThreads();
//		return( max( 2, numMaxThreads - 1 ) );
		return getAppliCtx().getAppliConf().getNumThreads();
	}

	protected PullOfExecutorWorkers createPullOfWorkers()
	{
		int numMaxThreads = getNumThreads();
		PullOfExecutorWorkers result = new PullOfExecutorWorkers( "Morphing simulation" );
		result.init(numMaxThreads);
		result.setMaxNumberOfPendingExecutors(numMaxThreads);
		result.start();

		return( result );
	}

	public PullOfExecutorWorkers getPullOfWorkers() {
		return _pullOfWorkers;
	}

	public <RR> void addPendingExecutor( ExecutionFunctions.UnsafeFunction<RR> function, Consumer<RR> callback )
	{
		getPullOfWorkers().addPendingExecutor(function, callback);
	}

	public void addPendingExecutor( ExecutorInterface executor )
	{
		getPullOfWorkers().addPendingExecutor(executor);
	}

	protected void addPendingNonStopableExecutor(Runnable executor) {
		getPullOfWorkers().addPendingNonStopableExecutor(executor);
	}

	protected void stopExecutors()
	{
		getPullOfWorkers().stopExecutors();
	}

//	protected MorphingVerticesMapManager getVerticesManager() {
//		return _morphingVerticesManager;
//	}

	protected MorphingEditorApplicationContext getAppliCtx() {
		return appliCtx;
	}

	public List<MorphingImageModel> getImageModelListCopy() {
		return getAppliCtx().getImageModelListCopy();
	}

	protected FrameCreatorKey createFrameCreatorKey(String image1, String image2) {
		return new FrameCreatorKey(image1, image2);
	}

	protected NullFunctions getNullFunctions() {
		return NullFunctions.instance();
	}

	protected <CC, RR> RR getIfNotNull(CC obj, Function<CC, RR> getter) {
		return getNullFunctions().getIfNotNull(obj, getter);
	}

	protected <CC, RR> RR getIfNotNull(CC obj, Function<CC, RR> getter, RR defaultValue) {
		return getNullFunctions().getIfNotNull(obj, getter, defaultValue);
	}

	protected boolean isParticularTransition(MorphingTransition morphingTransition,
											 Function<MorphingTransition, Boolean> getter) {
		return getIfNotNull(morphingTransition, getter, false);
	}

	protected MorphingImageModel getImageModel(String imageName) {
		return getAppliCtx().getImageModel(imageName);
	}

	protected String getMasterImageName() {
		return getAppliCtx().getMorphingProjectModel().getMasterImageName();
	}

	protected MorphingTransition getMorphingTransition(String imageName) {
		return getIfNotNull(getImageModel(imageName),
				MorphingImageModel::getFollowingTransition);
	}

	protected boolean isParticularTransition(String imageName, Function<MorphingTransition, Boolean> getter) {
		return isParticularTransition(getMorphingTransition(imageName), getter);
	}

	protected boolean isNormalTransition(String imageName1) {
		return isParticularTransition(imageName1, MorphingTransition::isNormal);
	}

	protected boolean isFaceDetectionTransition(String imageName1) {
		return isParticularTransition(imageName1, MorphingTransition::isFaceDetection);
	}

	protected MorphingFrameCreatorBase buildFrameCreator(String previousImageName,
														 String imageName1, String imageName2) {
		MorphingFrameCreatorBase result = null;
		if(allHaveFaceDetection()) {
			boolean previousIsFaceDetectionTransition = isFaceDetectionTransition(previousImageName);
			boolean firstIsFaceDetectionTransition = isFaceDetectionTransition(imageName1);
			if (!previousIsFaceDetectionTransition && !firstIsFaceDetectionTransition) {
				result = createNormalTransitionFrameCreator(imageName1, imageName2);
			} else {
				result = createFaceDetectionTransition(previousImageName, imageName1, imageName2);
			}
		} else {
			LOGGER.info("Not all images have face detections. We cannot make use of Face detection transition");
			result = createNormalTransitionFrameCreator(imageName1, imageName2);
		}
		return result;
	}

	protected boolean allHaveFaceDetection() {
		return getAppliCtx().allHaveFaceDetection();
	}

	protected FadingToBlack getFadingToBlack(String previousImageName,
											 String imageName1) {
		boolean previousIsFaceDetectionTransition = isFaceDetectionTransition(previousImageName);
		boolean firstIsFaceDetectionTransition = isFaceDetectionTransition(imageName1);

		FadingToBlack fadingToBlack;
		if(previousIsFaceDetectionTransition && firstIsFaceDetectionTransition) {
			fadingToBlack = FadingToBlack.ONLY_BLACK;
		} else if(previousIsFaceDetectionTransition && !firstIsFaceDetectionTransition) {
			fadingToBlack = FadingToBlack.FADE_FROM_BLACK;
		} else if(!previousIsFaceDetectionTransition && firstIsFaceDetectionTransition) {
			fadingToBlack = FadingToBlack.FADE_TO_BLACK;
		} else {
			throw new IllegalArgumentException("Cannot fade from or to black");
		}
		return fadingToBlack;
	}

	protected GeometricRectangle getMasterBounds() {
		return getImageModel(getMasterImageName()).getBounds();
	}

	protected FaceDetectionTransitionMorphingFrameCreator createFaceDetectionTransition(String previousImageName,
																						String imageName1, String imageName2) {
		FadingToBlack fadingToBlack = getFadingToBlack(previousImageName, imageName1);

		return createFaceDetectionTransition(imageName1, imageName2, fadingToBlack);
	}

	protected FaceDetectionTransitionMorphingFrameCreator createFaceDetectionTransition(String imageName1, String imageName2,
																						FadingToBlack fadingToBlack) {
		return new FaceDetectionTransitionMorphingFrameCreator(
				getImageModel(imageName1), getImageModel(imageName2), getMasterBounds(),
				getAppliCtx().getVerticesManager().getControlMeshManager().getControlMeshVerticesMap(),
				getAppliCtx().getVerticesManager().getControlMeshManager().getControlMeshTriangles(),
				fadingToBlack);
	}

	public LazyHolder<List<GeometricTriangle>> getMasterTrianglesGetter() {
		return masterTrianglesGetter;
	}

	protected NormalTransitionMorphingFrameCreator createNormalTransitionFrameCreator(String imageName1, String imageName2) {
		MorphingImageModel masterImageModel = getImageModel(getMasterImageName());
		NormalTransitionMorphingFrameCreator result = new NormalTransitionMorphingFrameCreator(
				getMasterTrianglesGetter()::get,
				getImageModel(imageName1), getImageModel(imageName2),
				masterImageModel.getBounds(),
				getAppliCtx().getVerticesManager().getVerticesMap(),
				masterImageModel,
				getDelaunaySubdivision());

		return result;
	}

	protected DelaunaySubdivision getDelaunaySubdivision() {
		return getAppliCtx().getVerticesManager().getDelaunayController().getTriangulation();
	}

	protected List<GeometricTriangle> getMasterTriangles() {
		List<GeometricTriangle> result = new ArrayList<>();
		DelaunayTriangleBrowser.instance().browseTriangles(getDelaunaySubdivision(),
				getMasterBounds(),
				triangle -> addTriangle(result, triangle) );

		return result;
	}

	protected boolean addTriangle(List<GeometricTriangle> result, GeometricTriangle masterTriangle) {
		result.add(masterTriangle);
		return true;
	}

	protected MorphingFrameCreatorBase<?, ?> getFrameCreator(String previousImageName,
													   String imageName1, String imageName2,
													   Map<FrameCreatorKey, MorphingFrameCreatorBase> frameCreatorMap,
													   Supplier<MorphingFrameCreatorBase> frameCreatorBuilder) {
		MorphingFrameCreatorBase result;
		if(frameCreatorBuilder == null) {
			result = frameCreatorMap.computeIfAbsent(createFrameCreatorKey(imageName1, imageName2),
					key -> buildFrameCreator(previousImageName, imageName1, imageName2));
		} else {
			result = frameCreatorMap.computeIfAbsent(createFrameCreatorKey(imageName1, imageName2),
					key -> frameCreatorBuilder.get());
		}
		return result;
	}

	protected MorphingFramesConsumtionPace createMorphingFramesConsumtionPace(
			FrameSequenceContext frameSequenceContext, int numTotalFrames)
	{
		return( new MorphingFramesConsumtionPace(frameSequenceContext, numTotalFrames, getNumThreads()) );
	}

	protected int getNumStepsPerTransition()
	{
		return( NUM_STEPS_PER_TRANSITION );
	}

	public int getNumTotalFrames()
	{
//		List<MorphingImageModel> list = getVerticesManager().getImageModelList().getListCopy();
		List<MorphingImageModel> list = getImageModelListCopy();

		// if the last transition was a Face detection one (faded to black), we add another transition with
		// the two images the same, but fading from black to color
		int increment = 0;
		if((list.size() > 1) && list.get(list.size() - 2).getFollowingTransition().isFaceDetection()) {
			increment++;
		}

		return( getNumStepsPerTransition() * ( list.size() - 1 + increment) + 1 );
	}

	public FrameSequenceContext createFrameSequenceContext(Consumer<BufferedImage> frameConsumer,
														   Consumer<Exception> exceptionConsumer,
														   AtomicBoolean hasToCancel,
														   Runnable releaseResourcesFunction,
														   Supplier<GeometricRectangle> frameSizeSupplier,
														   int delayMsBetweenFrames) {
		return new FrameSequenceContext(frameConsumer, exceptionConsumer, hasToCancel, releaseResourcesFunction,
				frameSizeSupplier, delayMsBetweenFrames);
	}

	public FrameSequenceContext doMorphing(Consumer<BufferedImage> frameConsumer,
										   Consumer<Exception> exceptionConsumer,
										   AtomicBoolean hasToCancel,
										   Runnable releaseResourcesFunction,
										   Supplier<GeometricRectangle> frameSizeSupplier,
										   int delayMsBetweenFrames) {
		FrameSequenceContext context = createFrameSequenceContext(frameConsumer, exceptionConsumer, hasToCancel,
				releaseResourcesFunction, frameSizeSupplier, delayMsBetweenFrames);

		doMorphing(context);

		return context;
	}

	public void doMorphing(FrameSequenceContext context) {
		stopExecutors();

		addPendingNonStopableExecutor( () ->
				doMorphingInternal(context));
	}

	public void doMorphingInternal(FrameSequenceContext frameSequenceContext)
	{
		int stepsPerTransition = getNumStepsPerTransition();

//		List<MorphingImageModel> list = getVerticesManager().getImageModelList().getListCopy();
		List<MorphingImageModel> list = getImageModelListCopy();
		Iterator<MorphingImageModel> it = list.iterator();
		MorphingImageModel firstImageModel = null;
		MorphingImageModel secondImageModel = null;
		if( it.hasNext() )
			firstImageModel = it.next();
		boolean executed = false;
		String previousImageName = null;
		String imageName1 = null;
		String imageName2 = null;

		Map<FrameCreatorKey, MorphingFrameCreatorBase> frameCreatorMap = new ConcurrentHashMap<>();
		int numTotalFrames = getNumTotalFrames();
		MorphingFramesConsumtionPace framePaceWorker =
			createMorphingFramesConsumtionPace(frameSequenceContext, numTotalFrames);
		addPendingExecutor( framePaceWorker );

		int index = 0;
		while( it.hasNext() )
		{
			if( frameSequenceContext.hasToCancel() )
				break;

			secondImageModel = it.next();
			previousImageName = imageName1;
			imageName1 = firstImageModel.getName();
			imageName2 = secondImageModel.getName();

			index = doTransition(previousImageName, imageName1, imageName2,
					index, stepsPerTransition,
					framePaceWorker, frameSequenceContext, frameCreatorMap, null);

			executed = true;

			firstImageModel = secondImageModel;
		}

		// if the last transition was a Face detection one (faded to black), we add another transition with
		// the two images the same, but fading from black to color
		if((imageName1 != null) && getImageModel(imageName1).getFollowingTransition().isFaceDetection()) {
			previousImageName = null;
			imageName1 = imageName2;

			String constImageName1 = imageName1;
			String constImageName2 = imageName2;
			index = doTransition(previousImageName, imageName1, imageName2,
					index, stepsPerTransition,
					framePaceWorker, frameSequenceContext,
					frameCreatorMap,
					() -> createFaceDetectionTransition(constImageName1, constImageName2, FadingToBlack.FADE_FROM_BLACK));
		}

		if( executed && !frameSequenceContext.hasToCancel() )
			addMorphingFrameCreatorExecutor(previousImageName, imageName1, imageName2, 1d,
				index, framePaceWorker, frameSequenceContext, frameCreatorMap, null );
	}

	protected int doTransition(String previousImageName, String imageName1, String imageName2,
							   int index, int stepsPerTransition,
							   MorphingFramesConsumtionPace framePaceWorker,
							   FrameSequenceContext frameSequenceContext,
							   Map<FrameCreatorKey, MorphingFrameCreatorBase> frameCreatorMap,
							   Supplier<MorphingFrameCreatorBase> frameCreatorBuilder) {
		for( int ii=0; ii<stepsPerTransition; ii++ )
		{
			if( frameSequenceContext.hasToCancel() )
				break;

			addMorphingFrameCreatorExecutor(previousImageName, imageName1, imageName2,
					( 1d * ii ) / stepsPerTransition,
					index, framePaceWorker, frameSequenceContext,
					frameCreatorMap, frameCreatorBuilder);
			index++;
		}
		return index;
	}



	protected void addMorphingFrameCreatorExecutor(String previousImageName, String imageName1, String imageName2,
												   double fraction, int index,
												   MorphingFramesConsumtionPace framePaceWorker,
												   FrameSequenceContext frameSequenceContext,
												   Map<FrameCreatorKey, MorphingFrameCreatorBase> frameCreatorMap,
												   Supplier<MorphingFrameCreatorBase> frameCreatorBuilder)
	{
		LOGGER.info( "added Morphing Frame productor {}", index );
		addPendingExecutor(
			() -> produceMorphingFrame( previousImageName, imageName1, imageName2, frameCreatorMap,
					fraction,
					index, frameSequenceContext, frameCreatorBuilder ),
			morphingFrame -> framePaceWorker.addItemToBeConsumed(morphingFrame) );
	}

	protected MorphingFrame produceMorphingFrame(String previousImageName,
												 String imageName1, String imageName2,
												 Map<FrameCreatorKey, MorphingFrameCreatorBase> frameCreatorMap,
												 double fraction, int index,
												 FrameSequenceContext frameSequenceContext,
												 Supplier<MorphingFrameCreatorBase> frameCreatorBuilder)
	{
		MorphingFrame result = new MorphingFrame();
		try {
			result.setIndex(index);

			long start = System.currentTimeMillis();
			LOGGER.info("Producing frame with index: {}", index);

//			if(index % 5 == 0) {
//				throw new RuntimeException("Error creating morphing frame test");
//			}

			BufferedImage frame = getFrameCreator(previousImageName, imageName1, imageName2, frameCreatorMap, frameCreatorBuilder)
					.createFrame(fraction, fraction, frameSequenceContext.getHasToCancelRef(),
							frameSequenceContext::getFrameSize, frameSequenceContext::processException, index);

			result.setImage(frame);

			long elapsed = System.currentTimeMillis() - start;
			LOGGER.info("Produced frame with index: {}   -  in {} ms", index, elapsed);
		} catch(Exception ex) {
			LOGGER.error("Error creating frame: {}", index, ex);
			frameSequenceContext.processException(ex);
		}

		return( result );
	}

	protected static class FrameCreatorKey implements Comparable<FrameCreatorKey> {
		protected String imageName1;
		protected String imageName2;

		public FrameCreatorKey(String imageName1, String imageName2) {
			this.imageName1 = imageName1;
			this.imageName2 = imageName2;
		}

		public String getImageName1() {
			return imageName1;
		}

		public String getImageName2() {
			return imageName2;
		}

		@Override
		public boolean equals(Object o) {
			if (this == o) return true;
			if (o == null || getClass() != o.getClass()) return false;
			FrameCreatorKey that = (FrameCreatorKey) o;
			return Objects.equals(imageName1, that.imageName1) && Objects.equals(imageName2, that.imageName2);
		}

		@Override
		public int hashCode() {
			return Objects.hash(imageName1, imageName2);
		}

		@Override
		public String toString() {
			return "FrameCreatorKey{" +
					"imageName1='" + imageName1 + '\'' +
					", imageName2='" + imageName2 + '\'' +
					'}';
		}

		@Override
		public int compareTo(FrameCreatorKey o) {
			int result = imageName1.compareTo(o.imageName1);
			if(result == 0) {
				result = imageName2.compareTo(o.imageName2);
			}
			return result;
		}
	}
}
