/*
 * Copyright (c) 2023. Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 3.0 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You may obtain a copy of the License at
 *
 *       http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */

package com.frojasg1.applications.libmorphing.frames.context;

import com.frojasg1.libgeometry.types.GeometricRectangle;

import java.awt.image.BufferedImage;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.function.Consumer;
import java.util.function.Supplier;

public class FrameSequenceContext {

    protected Consumer<BufferedImage> bufferedImageConsumer;
    protected Consumer<Exception> exceptionConsumer;
    protected AtomicBoolean hasToCancel;
    protected Runnable releaseResourcesFunction;
    protected Supplier<GeometricRectangle> frameSizeSupplier;
    protected int delayMsBetweenFrames;

    public FrameSequenceContext(Consumer<BufferedImage> bufferedImageConsumer, Consumer<Exception> exceptionConsumer,
                                AtomicBoolean hasToCancel, Runnable releaseResourcesFunction,
                                Supplier<GeometricRectangle> frameSizeSupplier, int delayMsBetweenFrames) {
        this.bufferedImageConsumer = bufferedImageConsumer;
        this.exceptionConsumer = exceptionConsumer;
        this.hasToCancel = hasToCancel;
        this.releaseResourcesFunction = releaseResourcesFunction;
        this.frameSizeSupplier = frameSizeSupplier;
        this.delayMsBetweenFrames = delayMsBetweenFrames;
    }

    public void setExceptionConsumer(Consumer<Exception> exceptionConsumer) {
        this.exceptionConsumer = exceptionConsumer;
    }

    public void setHasToCancel(boolean value) {
        if(hasToCancel != null) {
            hasToCancel.set(value);
        }
    }

    public AtomicBoolean getHasToCancelRef() {
        return hasToCancel;
    }

    public boolean hasToCancel() {
        boolean result = false;
        if(hasToCancel != null) {
            result = hasToCancel.get();
        }
        return result;
    }

    public int getDelayMsBetweenFrames() {
        return delayMsBetweenFrames;
    }

    public GeometricRectangle getFrameSize() {
        return frameSizeSupplier.get();
    }

    public void releaseResources() {
        if(releaseResourcesFunction != null) {
            releaseResourcesFunction.run();
        }
    }

    public void processBufferedImage(BufferedImage image) {
        if(bufferedImageConsumer != null) {
            bufferedImageConsumer.accept(image);
        }
    }

    public void processException(Exception ex) {
        if(exceptionConsumer != null) {
            exceptionConsumer.accept(ex);
        }
    }
}
