/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.applications.libmorphing.frames.framebuilder;

import com.frojasg1.libgeometry.algebra.helpers.GeometryHelper;
import com.frojasg1.libgeometry.algebra.helpers.Point2dDiscarder;
import com.frojasg1.libgeometry.algebra.helpers.VectorFunctions;
import com.frojasg1.applications.libmorphing.api.facedetector.helpers.Control88meshCreator;
import com.frojasg1.applications.libmorphing.helpers.MorphingHelper;
import com.frojasg1.applications.libmorphing.model.images.image.MorphingImageModel;
import com.frojasg1.applications.libmorphing.model.images.vertex.VertexModel;
import com.frojasg1.delaunay.triangulation.quadedge.DelaunayQuadEdgeHelpers;
import com.frojasg1.libgeometry.types.GeometricRectangle;
import com.frojasg1.libgeometry.types.GeometricTriangle;
import com.frojasg1.libgeometry.types.Point2d;
import com.frojasg1.libgeometry.types.factory.GeometryCommonFactory;
import com.frojasg1.general.desktop.image.ImageFunctions;
import com.frojasg1.general.desktop.view.ViewFunctions;
import com.frojasg1.general.map.multimap.dynamic.DynamicMultiMapElementBase;
import com.frojasg1.general.number.DoubleFunctions;
import java.util.concurrent.atomic.AtomicInteger;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.awt.*;
import java.awt.image.BufferedImage;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.function.Consumer;
import java.util.function.Predicate;
import java.util.function.Supplier;


/**
 *
 * @param <MT> - Master Triangle
 * @param <MP> - Master point
 */
public abstract class MorphingFrameCreatorBase<MT, MP> { //extends ApplicationCtxHolder {
	private static final Logger LOGGER = LoggerFactory.getLogger(MorphingFrameCreatorBase.class);

//	protected MorphingVerticesMapManager _morphingVerticesManager;

	protected MorphingImageModel imageModel1;
	protected MorphingImageModel imageModel2;
	protected GeometricRectangle masterBounds;

	public MorphingFrameCreatorBase(MorphingImageModel imageModel1, MorphingImageModel imageModel2,
									GeometricRectangle masterBounds) {
		this.imageModel1 = imageModel1;
		this.imageModel2 = imageModel2;
		this.masterBounds = masterBounds;
	}

	protected MorphingImageModel getImageModel1() {
		return imageModel1;
	}

	protected MorphingImageModel getImageModel2() {
		return imageModel2;
	}

	protected GeometricRectangle getMasterBounds() {
		return masterBounds;
	}

	protected BufferedImage createEmptyImage(GeometricRectangle imageSize)
	{
		return( createEmptyImage( (int) imageSize.width, (int) imageSize.height ) );
	}

	protected BufferedImage createEmptyImage(int width, int height )
	{
		return( ImageFunctions.instance().createImage(width, height) );
	}

	protected MorphingHelper getMorphingHelper()
	{
		return( MorphingHelper.instance() );
	}

	protected GeometryHelper getGeometryHelper() {
		return GeometryHelper.instance();
	}

	protected DoubleFunctions getDoubleFunctions() {
		return DoubleFunctions.instance();
	}

	protected VectorFunctions getVectorFunctions() {
		return VectorFunctions.instance();
	}

	protected Control88meshCreator getControl88meshCreator() {
		return Control88meshCreator.instance();
	}

	protected DelaunayQuadEdgeHelpers getDelaunayQuadEdgeHelpers() {
		return DelaunayQuadEdgeHelpers.instance();
	}

	public double getDistanceBetweenSegmentAndPoint(Point2d pt1, Point2d pt2, Point2d pt) {
		return getDelaunayQuadEdgeHelpers().getDistanceBetweenSegmentAndPoint(pt1, pt2, pt);
	}

	protected GeometricRectangle calculateBounds(GeometricRectangle rectangle )
	{
		return( calculateBounds( rectangle.width, rectangle.height ) );
	}

	protected GeometricRectangle calculateBounds(BufferedImage image )
	{
		return( calculateBounds( image.getWidth(), image.getHeight() ) );
	}

	public GeometricRectangle createDelaunayRectangle(Rectangle rectangle )
	{
		return( GeometryCommonFactory.instance().createDelaunayRectangle(rectangle) );
	}

	protected GeometricRectangle calculateBounds(double width, double height )
	{
		GeometricRectangle masterBounds = getMasterBounds();
		Rectangle rect = ViewFunctions.instance().calculateFittingBounds(
																masterBounds.width,
																masterBounds.height,
																width, height );

		return( createDelaunayRectangle( rect ) );
	}

	public BufferedImage createFrame(double fadingFraction, double triangulationFraction,
									 AtomicBoolean hasToCancel,
									 Supplier<GeometricRectangle> frameSizeSupplier,
									 Consumer<Exception> exceptionConsumer,
									 int index)
	{
		BufferedImage result = null;
		try {
			if (hasToCancel.get())
				return (null);

			GeometricRectangle imageSize = frameSizeSupplier.get();
			GeometricRectangle imageBounds = calculateBounds(imageSize);

			result = createFrame(fadingFraction, triangulationFraction,
					hasToCancel, imageSize, imageBounds, null, exceptionConsumer, index);
		} catch(Exception ex) {
			LOGGER.error("Exception when creating frame", ex);
			if(exceptionConsumer != null) {
				exceptionConsumer.accept(ex);
			}
		}
		return result;
	}

	public BufferedImage createFrame(double fadingFraction, double triangulationFraction,
                                     AtomicBoolean hasToCancel, GeometricRectangle imageSize,
                                     GeometricRectangle bounds, GeometricRectangle clip,
                                     Consumer<Exception> exceptionConsumer,
                                     int index)
	{
		BufferedImage result = null;
		try {
			BufferedImage tmpResult = createEmptyImage(imageSize);

			TriangleMapperContext ctx = createCtx();

			browseTriangles(getMasterBounds(),
					triangle -> processTriangle(triangle, tmpResult, bounds,
							ctx, fadingFraction, triangulationFraction,
							hasToCancel, clip), index);

			result = tmpResult;
			if (hasToCancel.get()) {
				result = null;
			}
		} catch(Exception ex) {
			LOGGER.error("Exception when creating frame:", ex);
			if(exceptionConsumer != null) {
				exceptionConsumer.accept(ex);
			}
		}

		return (result);
	}

	protected TriangleMapperContext createCtx() {
		TriangleMapperContext result = new TriangleMapperContext();
		result.setImage1(getImageModel1().getImage());
		result.setImage2(getImageModel2().getImage());

		return result;
	}

	protected abstract void browseTriangles(GeometricRectangle bounds,
                                            Predicate<MT> trianglePredicate,
                                            int index);

	protected GeometricTriangle createTriangle(Point2d v1, Point2d v2, Point2d v3 )
	{
		return( GeometryCommonFactory.instance().createDelaunayTriangle( v1, v2, v3, false ) );
	}

	protected abstract Point2d mapPoint( MP masterPoint, String imageName );

	protected abstract MP getMasterVertex1(MT masterTriangle);
	protected abstract MP getMasterVertex2(MT masterTriangle);
	protected abstract MP getMasterVertex3(MT masterTriangle);

	protected GeometricTriangle createImageModelTriangle(String imageName,
														 MT masterTriangle )
	{
		return( createTriangle( mapPoint( getMasterVertex1(masterTriangle), imageName ),
					mapPoint( getMasterVertex2(masterTriangle), imageName ),
					mapPoint( getMasterVertex3(masterTriangle), imageName ) ) );
	}

	protected double weight( double value1, double value2, double fraction )
	{
		return( value1 * ( 1 - fraction ) + value2 * fraction );
	}

	protected abstract DynamicMultiMapElementBase<VertexModel> getMultiVertex(MP masterPoint);

	protected Point2d weightPoint(MP masterPoint, String imageName1, String imageName2,
                                  GeometricRectangle imageBounds, double fraction )
	{
		DynamicMultiMapElementBase<VertexModel> multiVertex = getMultiVertex(masterPoint);

		Point2d normalizedPoint1 = multiVertex.get(imageName1).getNormalizedPoint();
		Point2d normalizedPoint2 = multiVertex.get(imageName2).getNormalizedPoint();

		Point2d normalizedResultPoint = normalizedPoint1.createCopy();
		normalizedResultPoint.x = weight( normalizedPoint1.x, normalizedPoint2.x, fraction );
		normalizedResultPoint.y = weight( normalizedPoint1.y, normalizedPoint2.y, fraction );

		return( round( fromNormalizedPt( normalizedResultPoint, imageBounds ) ) );
	}

	protected Point2d round( Point2d pt )
	{
		return( getMorphingHelper().toIntCoordinates( pt ) );
	}

	protected Point2d fromNormalizedPt( Point2d pt, GeometricRectangle bounds )
	{
		return( getMorphingHelper().fromNormalized(pt, bounds ) );
	}

	protected GeometricTriangle createResultImageTriangle(MT masterTriangle,
														  String imageName1, String imageName2, GeometricRectangle imageBounds,
														  double fraction)
	{
		return( createTriangle( weightPoint( getMasterVertex1(masterTriangle), imageName1, imageName2, imageBounds, fraction ),
								weightPoint( getMasterVertex2(masterTriangle), imageName1, imageName2, imageBounds, fraction ),
								weightPoint( getMasterVertex3(masterTriangle), imageName1, imageName2, imageBounds, fraction ) ) );
	}

	protected boolean processTriangle(MT masterTriangle,
									  BufferedImage result, GeometricRectangle resultImageBounds,
									  TriangleMapperContext context,
									  double fadingFraction, double triangulationFraction,
									  AtomicBoolean hasToCancel, GeometricRectangle clip)
	{
		boolean hasToContinue = !hasToCancel.get();
		if(hasToContinue) {
			context.setFraction(fadingFraction);
			GeometricTriangle triangle1 = createImageModelTriangle(getImageModel1().getName(), masterTriangle);
			context.setTriangle1(triangle1);
			GeometricTriangle triangle2 = createImageModelTriangle(getImageModel2().getName(), masterTriangle);
			context.setTriangle2(triangle2);

			context.setResultImage(result);
			GeometricTriangle resultTriangle = createResultImageTriangle(masterTriangle,
					getImageModel1().getName(), getImageModel2().getName(),
					resultImageBounds, triangulationFraction);
			context.setResultTriangle(resultTriangle);

			context.setMasterTriangle(masterTriangle);

			// no protection increment, as it has been added when necessary at the resultTriangle
			int protectionIncrement = 0;
			if (!canDiscard(resultTriangle, clip))
				GeometryHelper.instance().browseTriangle(resultTriangle.getVertex1(),
						resultTriangle.getVertex2(), resultTriangle.getVertex3(),
						protectionIncrement, clip,
						context::processPoint);
		}
		return hasToContinue;
	}

	protected boolean canDiscard(GeometricTriangle triangle, GeometricRectangle clip )
	{
		return( Point2dDiscarder.instance().canDiscard(clip,
			triangle.getVertex1(), triangle.getVertex2(), triangle.getVertex3() ) );
	}

	protected class TriangleMapperContext
	{
		protected double[] _resultArgbArray = new double[4];

		protected double _fraction;

		protected BufferedImage _image1;
		protected GeometricTriangle _triangle1;

		protected BufferedImage _image2;
		protected GeometricTriangle _triangle2;

		protected BufferedImage _resultImage;
		protected GeometricRectangle _resultImageBounds;
		protected GeometricTriangle _resultTriangle;

		protected MT masterTriangle;

		public void processPoint( Point2d resultImagePoint )
		{
			Point2d pt1 = mapTrianglePt( resultImagePoint, _resultTriangle, _triangle1 );
			Point2d pt2 = mapTrianglePt( resultImagePoint, _resultTriangle, _triangle2 );

			int xx = resultImagePoint.getIntX();
			int yy = resultImagePoint.getIntY();

			if( ( xx >= 0 ) && ( xx < _resultImage.getWidth() ) &&
				( yy >= 0 ) && ( yy < _resultImage.getHeight() ) )
			{
				_resultImage.setRGB( xx, yy, calculateColor(pt1, pt2) );
			}
		}

		public MT getMasterTriangle() {
			return masterTriangle;
		}

		public void setMasterTriangle(MT masterTriangle) {
			this.masterTriangle = masterTriangle;
		}

		protected Point2d mapTrianglePt(Point2d pt, GeometricTriangle originTriangle,
										GeometricTriangle destinationTriangle )
		{
			return( mapTrianglePt(pt,
						originTriangle.getVertex1(), originTriangle.getVertex2(), originTriangle.getVertex3(),
						destinationTriangle.getVertex1(), destinationTriangle.getVertex2(), destinationTriangle.getVertex3() )
			);
		}

		protected void resetArgbArray()
		{
			_resultArgbArray[0] = 0;
			_resultArgbArray[1] = 0;
			_resultArgbArray[2] = 0;
			_resultArgbArray[3] = 0;
		}

		protected double limit( double value, double lower, double upper )
		{
			return( DoubleFunctions.instance().limit(value, lower, upper) );
		}

		protected int getRgb( BufferedImage image, Point2d pt )
		{
			int xx = (int) limit( pt.x, 0, image.getWidth() - 1 );
			int yy = (int) limit( pt.y, 0, image.getHeight() - 1 );

			return( image.getRGB( xx, yy ) );
		}

		protected void calculateFadedComponents(Point2d pt1, Point2d pt2, double[] resultArgbArray) {
			resetArgbArray();

			ImageFunctions.instance().addARGB(resultArgbArray,
					getRgb( _image1, pt1 ),
					1 - _fraction);
			ImageFunctions.instance().addARGB(resultArgbArray,
					getRgb( _image2, pt2 ),
					_fraction);
		}

		protected void calculateColorComponents(Point2d pt1, Point2d pt2, double[] resultArgbArray) {
			calculateFadedComponents(pt1, pt2, resultArgbArray);
		}

		protected int calculateColor(Point2d pt1, Point2d pt2)
		{
			calculateColorComponents(pt1, pt2, _resultArgbArray);

			return( ImageFunctions.instance().getARGB(_resultArgbArray) );
		}

		public Point2d mapTrianglePt( Point2d pt,
			Point2d pt11, Point2d pt12, Point2d pt13,
			Point2d pt21, Point2d pt22, Point2d pt23 )
		{
			return( GeometryHelper.instance().mapTrianglePt(pt, pt11, pt12, pt13, pt21, pt22, pt23) );
		}

		public double[] getResultArgbArray() {
			return _resultArgbArray;
		}

		public void setResultArgbArray(double[] _resultArgbArray) {
			this._resultArgbArray = _resultArgbArray;
		}

		public GeometricRectangle getResultImageBounds() {
			return _resultImageBounds;
		}

		public void setResultImageBounds(GeometricRectangle _resultImageBounds) {
			this._resultImageBounds = _resultImageBounds;
		}

		public double getFraction() {
			return _fraction;
		}

		public void setFraction(double _fraction) {
			this._fraction = _fraction;
		}

		public BufferedImage getImage1() {
			return _image1;
		}

		public void setImage1(BufferedImage _image1) {
			this._image1 = _image1;
		}

		public BufferedImage getImage2() {
			return _image2;
		}

		public void setImage2(BufferedImage _image2) {
			this._image2 = _image2;
		}

		public BufferedImage getResultImage() {
			return _resultImage;
		}

		public void setResultImage(BufferedImage _resultImage) {
			this._resultImage = _resultImage;
		}

		public GeometricTriangle getTriangle1() {
			return _triangle1;
		}

		public void setTriangle1(GeometricTriangle _triangle1) {
			this._triangle1 = _triangle1;
		}

		public GeometricTriangle getTriangle2() {
			return _triangle2;
		}

		public void setTriangle2(GeometricTriangle _triangle2) {
			this._triangle2 = _triangle2;
		}

		public GeometricTriangle getResultTriangle() {
			return _resultTriangle;
		}

		public void setResultTriangle(GeometricTriangle _resultTriangle) {
			this._resultTriangle = _resultTriangle;
		}


		public void multiplyColorComponentsByScalar(double[] resultArgbArray, double factor) {
			ImageFunctions.instance().multiplyByScalar(resultArgbArray, factor);
		}
	}
}
