/*
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.applications.libmorphing.frames.framebuilder.impl;

import com.frojasg1.applications.libmorphing.api.facedetector.constants.Landmark68constants;
import com.frojasg1.applications.libmorphing.frames.framebuilder.MorphingFrameCreatorBase;
import com.frojasg1.applications.libmorphing.model.images.image.MorphingImageModel;
import com.frojasg1.applications.libmorphing.model.images.vertex.VertexModel;
import com.frojasg1.applications.libmorphing.model.images.vertex.map.impl.control.IndexedVerticesMap;
import com.frojasg1.libgeometry.types.GeometricRectangle;
import com.frojasg1.libgeometry.types.GeometricTriangle;
import com.frojasg1.libgeometry.types.Point2d;
import com.frojasg1.general.map.multimap.dynamic.DynamicMultiMapElementBase;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.awt.image.BufferedImage;
import java.util.Arrays;
import java.util.Objects;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.function.Consumer;
import java.util.function.Predicate;
import java.util.function.Supplier;

/**
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 */
public class FaceDetectionTransitionMorphingFrameCreator extends MorphingFrameCreatorBase<int[], Integer> {
    private static final Logger LOGGER = LoggerFactory.getLogger(FaceDetectionTransitionMorphingFrameCreator.class);

    // 68_landmark_points_mesh.png
    protected static final int[] boundaryVertices = {20, 21, 22, 23};

	protected FadingToBlack fadingToBlack;
    protected IndexedVerticesMap verticesMap;
    protected int[][] masterTriangles;

    public FaceDetectionTransitionMorphingFrameCreator(MorphingImageModel imageModel1, MorphingImageModel imageModel2,
                                                       GeometricRectangle masterBounds, IndexedVerticesMap verticesMap,
                                                       int[][] masterTriangles, FadingToBlack fadingToBlack) {
        super(imageModel1, imageModel2, masterBounds);
        this.verticesMap = verticesMap;
        this.masterTriangles = masterTriangles;
		this.fadingToBlack = fadingToBlack;
    }

	protected FadingToBlack getFadingToBlack() {
		return fadingToBlack;
	}

	protected IndexedVerticesMap getVerticesMap() {
        return verticesMap;
    }

    protected int[][] getMasterTriangles() {
        return masterTriangles;
    }

    @Override
    protected Point2d mapPoint(Integer masterPoint, String imageName) {
        return (getMultiVertex(masterPoint).get(imageName).getPoint());
    }

    @Override
    public BufferedImage createFrame(double fadingFraction, double triangulationFraction, AtomicBoolean hasToCancel,
                                     Supplier<GeometricRectangle> frameSizeSupplier, Consumer<Exception> exceptionConsumer,
                                     int index) {
        return super.createFrame(fadingFraction, triangulationFraction, hasToCancel, frameSizeSupplier, exceptionConsumer, index);
    }

    @Override
    public BufferedImage createFrame(double fadingFraction, double triangulationFraction, AtomicBoolean hasToCancel,
                                     GeometricRectangle imageSize, GeometricRectangle bounds, GeometricRectangle clip,
                                     Consumer<Exception> exceptionConsumer, int index) {
        return super.createFrame(fadingFraction, triangulationFraction, hasToCancel, imageSize, bounds, clip,
                exceptionConsumer, index);
    }

    @Override
    protected void browseTriangles(GeometricRectangle bounds, Predicate<int[]> trianglePredicate,
                                   int index) {
        for (int[] triangle : getMasterTriangles()) {
            if (!trianglePredicate.test(triangle)) {
                break;
            }
        }
    }

//    protected Integer getFirstNonFrameIndex(int[] masterTriangle) {
//        Integer result = null;
//        int index = 0;
//        for(int vertexIndex: masterTriangle) {
//            if(!getControl88meshCreator().isBoundsVertex(vertexIndex)) {
//                result = index;
//                break;
//            }
//            index++;
//        }
//        return result;
//    }
//
//    protected double getProtectionFactor(Point2d p1, Point2d p2, Point2d p3, Point2d inCenter,
//                                         double normalBroadeningInPixels) {
//        Point2d normNt = getVectorFunctions().getUniraryVector(getVectorFunctions().getOrthogonalVector(p2.subtract(p1)));
//        int sgn = getDoubleFunctions().sgn(getVectorFunctions().scalarProduct(p3.subtract(p1), normNt));
//        Point2d normN = getVectorFunctions().multiplyByScalar(normNt, -sgn,  null);
//        Point2d d1 = p1.subtract(inCenter);
//        Point2d normD1 = getVectorFunctions().getUniraryVector(d1);
//
//        double dP1 = normalBroadeningInPixels / getVectorFunctions().scalarProduct(normD1, normN);
//
//        return 1.0d + dP1 / d1.norm();
//    }
//
//
//    protected boolean isReallyATriangle(List<Point2d> vertices, Point2d inCenter) {
//        Iterator<Point2d> it = vertices.iterator();
//        Point2d p1 = it.next();
//        Point2d p2 = it.next();
//        Point2d p3 = it.next();
//
//        return getDoubleFunctions().min(
//                getDoubleFunctions().min(getDistanceBetweenSegmentAndPoint(p1, p2, p3),
//                        getDistanceBetweenSegmentAndPoint(p3, p1, p2)),
//                getDistanceBetweenSegmentAndPoint(p2, p3, p1)) > 1.0d;
//    }
//
//    protected DelaunayTriangle applyProtectionIncrementFactor(int[] masterTriangle, DelaunayTriangle triangle,
//                                                              double normalBroadeningInPixels, double maxFactor) {
//        DelaunayTriangle result = triangle;
//
//        // see Calculation.of.factor.for.non.matched.edge.triangles.png
//        List<Point2d> myList = new ArrayList<>(Arrays.asList(triangle.getVertex1(), triangle.getVertex2(), triangle.getVertex3()));
//
//        Point2d inCenter = triangle.getInCenter();
//        if(isReallyATriangle(myList, inCenter)) {
//            Integer nonFrameVertexIndex = getFirstNonFrameIndex(masterTriangle);
//            Point2d p3 = myList.remove((int) nonFrameVertexIndex);
//            Point2d p1 = myList.get(0);
//            Point2d p2 = myList.get(1);
//
//            double factor1 = getProtectionFactor(p1, p2, p3, inCenter, normalBroadeningInPixels);
//            double factor2 = getProtectionFactor(p2, p1, p3, inCenter, normalBroadeningInPixels);
//
//            double factor = getDoubleFunctions().min(factor1, factor2);
//            factor = getDoubleFunctions().min(factor, maxFactor);
//
//            result = triangle.zoom(factor);
//        }
//
//        return result;
//    }

    @Override
    protected GeometricTriangle createResultImageTriangle(int[] masterTriangle,
                                                          String imageName1, String imageName2, GeometricRectangle imageBounds,
                                                          double fraction)
    {
        GeometricTriangle triangle = super.createResultImageTriangle(masterTriangle, imageName1, imageName2, imageBounds, fraction);
//
//        // broadening the triangles not matched (that have an edge on the external side, matching three edges on the internal side)
//        // see 88-control.point.mesh.with.68-landmark.jpg
//        if(isNotMatchedTriangle(masterTriangle)) {
//            // see Calculation.of.factor.for.non.matched.edge.triangles.png
//            triangle = applyProtectionIncrementFactor(masterTriangle, triangle, 3.0, 7.0);
//        }

        return triangle;
    }

    // it returns if the triangle is one having an edge matching equals to one of the edges of the face detection frame
    // which means that two of the triangle vertex indices match a face detection frame corner index
    protected boolean isNotMatchedTriangle(int[] triangle) {
        return Arrays.stream(triangle).filter(getControl88meshCreator()::isBoundsVertex).count() == 2;
    }

	@Override
	protected FaceDetectionTriangleMapperContext createCtx() {
		FaceDetectionTriangleMapperContext result = new FaceDetectionTriangleMapperContext();
		result.setImage1(getImageModel1().getImage());
		result.setImage2(getImageModel2().getImage());

		return result;
	}

    @Override
    protected Integer getMasterVertex1(int[] masterTriangle) {
        return masterTriangle[0];
    }

    @Override
    protected Integer getMasterVertex2(int[] masterTriangle) {
        return masterTriangle[1];
    }

    @Override
    protected Integer getMasterVertex3(int[] masterTriangle) {
        return masterTriangle[2];
    }

    @Override
    protected DynamicMultiMapElementBase<VertexModel> getMultiVertex(Integer masterPoint) {
        return getVerticesMap().get(masterPoint);
    }

	protected class FaceDetectionTriangleMapperContext extends TriangleMapperContext {
		@Override
		public void processPoint(Point2d resultImagePoint) {
			super.processPoint(resultImagePoint);
		}

        protected boolean isExternalTriangle() {
            int[] masterTriangle1 = getMasterTriangle();

            // 68_landmark_points_mesh.png
            boolean result = Arrays.stream(masterTriangle1) // if any is external
                    .anyMatch(ind -> ind >= Landmark68constants.NB_POINTS_LANDMARK_68);

            if(!result) { // if all are at the boundary
                result = Arrays.stream(masterTriangle1)
                        .allMatch(ind -> Arrays.stream(boundaryVertices).anyMatch(ind2 -> Objects.equals(ind2, ind)));
            }

            return result;
        }

        protected boolean hasToFadeFromOrToBlack() {
            return isExternalTriangle();
        }

		protected double getFadeToOrFromBlackFactor() {
            double result = 1d;
            if(hasToFadeFromOrToBlack()) {
                result = getFadingToBlack().getFactor(getFraction());
            }
			return result;
		}

		@Override
		protected void calculateColorComponents(Point2d pt1, Point2d pt2, double[] resultArgbArray) {
			double factor = getFadeToOrFromBlackFactor();

			if(factor == 0d) {
				resetArgbArray();
			} else {
				calculateFadedComponents(pt1, pt2, resultArgbArray);

				multiplyColorComponentsByScalar(resultArgbArray, factor);
			}
		}
	}
}
