/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.applications.libmorphing.frames.framebuilder.impl;

import com.frojasg1.applications.libmorphing.frames.framebuilder.MorphingFrameCreatorBase;
import com.frojasg1.applications.libmorphing.model.images.image.MorphingImageModel;
import com.frojasg1.applications.libmorphing.model.images.vertex.VertexModel;
import com.frojasg1.applications.libmorphing.model.images.vertex.map.impl.normal.MorphingVerticesMap;
import com.frojasg1.delaunay.triangulation.quadedge.DelaunaySubdivision;
import com.frojasg1.delaunay.triangulation.quadedge.helpers.DelaunayTriangleBrowser;
import com.frojasg1.general.getters.LazyHolder;
import com.frojasg1.libgeometry.types.GeometricRectangle;
import com.frojasg1.libgeometry.types.GeometricTriangle;
import com.frojasg1.libgeometry.types.Point2d;
import com.frojasg1.general.map.multimap.dynamic.DynamicMultiMapElementBase;
import com.frojasg1.general.map.multimap.dynamic.deductible.DeductibleDynamicMultiMapElementBase;

import java.awt.image.BufferedImage;
import java.util.List;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.function.Consumer;
import java.util.function.Predicate;
import java.util.function.Supplier;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 *
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 */
public class NormalTransitionMorphingFrameCreator extends MorphingFrameCreatorBase<GeometricTriangle, Point2d> {
	private static final Logger LOGGER = LoggerFactory.getLogger(NormalTransitionMorphingFrameCreator.class);

//	protected MorphingVerticesMapManager _morphingVerticesManager;

	protected Supplier<List<GeometricTriangle>> masterTrianglesGetter;
	protected MorphingVerticesMap verticesMap;
	protected MorphingImageModel masterImageModel;
	protected DelaunaySubdivision delaunaySubdivision;

	public NormalTransitionMorphingFrameCreator(Supplier<List<GeometricTriangle>> masterTrianglesGetter,
												MorphingImageModel imageModel1, MorphingImageModel imageModel2,
                                                GeometricRectangle masterBounds, MorphingVerticesMap verticesMap,
                                                MorphingImageModel masterImageModel, DelaunaySubdivision delaunaySubdivision) {
		super(imageModel1, imageModel2, masterBounds);
		this.masterTrianglesGetter = masterTrianglesGetter;
		this.verticesMap = verticesMap;
		this.masterImageModel = masterImageModel;
		this.delaunaySubdivision = delaunaySubdivision;
	}

	protected MorphingImageModel getMasterImageModel()
	{
		return masterImageModel;
	}

	protected MorphingVerticesMap getVerticesMap() {
		return verticesMap;
	}

	protected DelaunaySubdivision getDelaunaySubdivision()
	{
		return delaunaySubdivision;
	}

	protected String getMasterName()
	{
		return getMasterImageModel().getName();
	}

	@Override
	protected Point2d mapPoint( Point2d masterPoint, String imageName )
	{
		DeductibleDynamicMultiMapElementBase<VertexModel> multiVertex =
			getVerticesMap().get(getMasterName(), masterPoint);

		return( multiVertex.get(imageName).getPoint() );
	}

	@Override
	public BufferedImage createFrame(double fadingFraction, double triangulationFraction, AtomicBoolean hasToCancel,
                                     Supplier<GeometricRectangle> frameSizeSupplier, Consumer<Exception> exceptionConsumer,
                                     int index) {
		return super.createFrame(fadingFraction, triangulationFraction, hasToCancel,
				frameSizeSupplier, exceptionConsumer, index);
	}

	@Override
	public BufferedImage createFrame(double fadingFraction, double triangulationFraction, AtomicBoolean hasToCancel,
                                     GeometricRectangle imageSize, GeometricRectangle bounds, GeometricRectangle clip,
                                     Consumer<Exception> exceptionConsumer, int index) {
		return super.createFrame(fadingFraction, triangulationFraction, hasToCancel, imageSize, bounds, clip,
				exceptionConsumer, index);
	}

	@Override
	protected void browseTriangles(GeometricRectangle bounds, Predicate<GeometricTriangle> trianglePredicate,
                                   int index) {
//		DelaunayTriangleBrowser.instance().browseTriangles(getDelaunaySubdivision(),
//				getMasterBounds(),
//				trianglePredicate );
		for(GeometricTriangle masterTriangle: masterTrianglesGetter.get()) {
			if(!trianglePredicate.test(masterTriangle)) {
				break;
			}
		}
	}

	@Override
	protected Point2d getMasterVertex1(GeometricTriangle masterTriangle) {
		return masterTriangle.getVertex1();
	}

	@Override
	protected Point2d getMasterVertex2(GeometricTriangle masterTriangle) {
		return masterTriangle.getVertex2();
	}

	@Override
	protected Point2d getMasterVertex3(GeometricTriangle masterTriangle) {
		return masterTriangle.getVertex3();
	}

	@Override
	protected DynamicMultiMapElementBase<VertexModel> getMultiVertex(Point2d masterPoint) {
		return getVerticesMap().get(getMasterName(), masterPoint);
	}
}
