/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.applications.libmorphing.frames.runnable;

import com.frojasg1.applications.libmorphing.frames.MorphingFrame;
import com.frojasg1.applications.libmorphing.frames.context.FrameSequenceContext;
import com.frojasg1.general.threads.consumer.ConsumerRunnable;

import java.awt.image.BufferedImage;
import java.util.Iterator;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 *
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 */
public class MorphingFramesConsumtionPace extends ConsumerRunnable<MorphingFrame> {
	private static final Logger LOGGER = LoggerFactory.getLogger(MorphingFramesConsumtionPace.class);

//	protected static final int MAX_NUM_IMAGES_IN_A_ROW = 4;

	protected int _numFrames;
	protected int _currentIndex;
	protected FrameSequenceContext frameSequenceContext;

	protected int numMaxThreads;

	protected BufferedImage latestImage;

	public MorphingFramesConsumtionPace(FrameSequenceContext frameSequenceContext,
										int numFrames,
										int numMaxThreads) {
		this(frameSequenceContext, numFrames, numMaxThreads, true);
	}

	protected MorphingFramesConsumtionPace(FrameSequenceContext frameSequenceContext,
										   int numFrames,
										   int numMaxThreads,
										   boolean init) {
		super(0, false);

		_numFrames = numFrames;
		this.frameSequenceContext = frameSequenceContext;
		this.numMaxThreads = numMaxThreads;
		if( init )
			init();
	}

	protected int getNumMaxThreads() {
		return numMaxThreads;
	}

	protected boolean hasToCancel()
	{
		return( frameSequenceContext.hasToCancel() );
	}

	@Override
	protected boolean getHasToStop()
	{
//		boolean hasToStop = (_numFrames <= _currentIndex) || hasToCancel();
//		LOGGER.info( "_numFrames ({}) <= _currentIndex ({}) = {}",
//			_numFrames, _currentIndex, hasToStop);
//		return( super.getHasToStop() || hasToStop );
		return( super.getHasToStop() || hasToCancel());
	}

	protected int getNumMaxImagesInARow() {
		return getNumMaxThreads();
	}

	@Override
	protected boolean canAddItem()
	{
//		return( getItemsToBeConsumed().size() < 25 );
//		return( !tooManyImagesInARow( MAX_NUM_IMAGES_IN_A_ROW ) );
		return( !tooManyImagesInARow( getNumMaxImagesInARow() ) );
	}

	@Override
	protected void addItemToBeConsumedInternal( MorphingFrame item ) {
		Integer index = item != null ? item.getIndex(): null;
		LOGGER.info("Trying to add frame to be consumed, with index: {}", index);
		super.addItemToBeConsumedInternal(item);

		LOGGER.info("Added item (with index = {}) to be consumed, and signalled waiting for item to be consumed if necessary.",
				index);
	}

	protected boolean tooManyImagesInARow( int maxNumImages )
	{
		return( this.lockFunction( () -> tooManyImagesInARowInternal(maxNumImages) ) );
	}

	protected boolean tooManyImagesInARowInternal( int maxNumImages )
	{
		boolean result = true;
		for( int ii = 0; ii < maxNumImages; ii++ )
			if( !containsImageIndex( _currentIndex + ii ) )
			{
				result = false;
				break;
			}

		return( result );
	}

	protected boolean containsImageIndex( int imageIndex )
	{
		boolean result = false;

		MorphingFrame item;
		Iterator<MorphingFrame> it = getItemsToBeConsumed().iterator();
		for( MorphingFrame mf: getItemsToBeConsumed() )
		{
			if( mf.getIndex() == imageIndex )
			{
				result = true;
				break;
			}
		}

		return( result );
	}

	@Override
	protected void consume(MorphingFrame itemToBeConsumed)
	{
		lockProcedure(() -> consumeInternal( itemToBeConsumed ) );
	}

	public int getDelayMsBetweenFrames() {
		return frameSequenceContext.getDelayMsBetweenFrames();
	}

	protected BufferedImage getLatestImage() {
		return latestImage;
	}

	protected void setLatestImage(BufferedImage latestImage) {
		this.latestImage = latestImage;
	}

	// idempotent function
	protected BufferedImage getImage(MorphingFrame itemToBeConsumed) {
		BufferedImage result = itemToBeConsumed.getImage();

		// if it was null (probably due to thrown exception) we take the latest one
		if(result == null) {
			result = getLatestImage();
		}

		setLatestImage(result);

		return result;
	}

	protected void consumeInternal(MorphingFrame itemToBeConsumed)
	{
		int index = itemToBeConsumed.getIndex();
		LOGGER.info("Just consuming frame: {}", index);
		frameSequenceContext.processBufferedImage(getImage(itemToBeConsumed));
		LOGGER.info("Just consumed frame: {}", index);
		doSleep();
		LOGGER.info("Just after sleep, after having consumed frame: {}", itemToBeConsumed.getIndex());
	}

	protected void doSleep()
	{
		int delay = getDelayMsBetweenFrames();
		if( delay > 0 )
			sleep(delay);
	}

	protected boolean currentExists() {
		boolean result = false;

		MorphingFrame item;
		Iterator<MorphingFrame> it = getItemsToBeConsumed().iterator();
		while( it.hasNext() )
		{
			item = it.next();
			if( item.getIndex() == _currentIndex )
			{
				result = true;
				break;
			}
		}
		return result;
	}

	@Override
	protected void waitForItems()
	{
		boolean logged = false;
		while( !isInterrupted() && !currentExists() ) {
			if(!logged) {
				LOGGER.info("waiting to consume frame: {}", _currentIndex);
				logged = true;
			}
			awaitInterrupted(WAITING_FOR_ITEM_TO_BE_CONSUMED_CONDITION_INDEX, 500);
		}
	}

	@Override
	protected MorphingFrame getItemToBeConsumedInternalInternal()
	{
		MorphingFrame result = null;

		MorphingFrame item;
		Iterator<MorphingFrame> it = getItemsToBeConsumed().iterator();
		while( it.hasNext() )
		{
			item = it.next();
			if( item.getIndex() == _currentIndex )
			{
				_currentIndex++;
				result = item;
				it.remove();
				break;
			}
		}

		if(result != null) {
			LOGGER.info("Just about to consume frame: {}", result.getIndex());
		}

		return( result );
	}

	@Override
	public void doEnd() {
		frameSequenceContext.releaseResources();

		frameSequenceContext = null;
	}
}
