/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.applications.libmorphing.helpers;

import com.frojasg1.libgeometry.algebra.helpers.Point2dDiscarder;
import com.frojasg1.applications.libmorphing.constants.MorphingConstants;
import com.frojasg1.applications.libmorphing.model.images.color.ColorModel;
import com.frojasg1.applications.libmorphing.model.images.vertex.VertexModel;
import com.frojasg1.libgeometry.types.GeometricRectangle;
import com.frojasg1.libgeometry.types.Point2d;
import com.frojasg1.libgeometry.types.factory.GeometryCommonFactory;
import com.frojasg1.general.list.ListWrapper;
import com.frojasg1.general.number.DoubleFunctions;
import com.frojasg1.general.number.IntegerFunctions;
import java.awt.image.BufferedImage;
import java.util.concurrent.atomic.AtomicReference;
import java.util.function.BiFunction;

/**
 *
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 */
public class MorphingHelper {

	protected static MorphingHelper INSTANCE = new MorphingHelper();

	protected static GeometricRectangle _NORMALIZED_RECTANGLE = new GeometricRectangle( 0, 0, 1, 1 );

	protected GeometryCommonFactory _factory;

	public static MorphingHelper instance()
	{
		return( INSTANCE );
	}

	protected MorphingHelper()
	{
		this( GeometryCommonFactory.instance() );
	}

	protected MorphingHelper(GeometryCommonFactory factory)
	{
		_factory = factory;
	}

	public GeometryCommonFactory getFactory() {
		return _factory;
	}

	public Point2d fromNormalized( Point2d percentagesPoint, GeometricRectangle rect )
	{
		Point2d result = null;
		if( ( percentagesPoint != null ) && ( rect != null ) )
		{
			double xx = percentagesPoint.x * rect.width + rect.x;
			double yy = percentagesPoint.y * rect.height + rect.y;

			result = ( getFactory().createPoint2d(xx, yy) );
		}
		return( result );
	}

	public Point2d fromNormalizedInt( Point2d percentagesPoint, GeometricRectangle rect )
	{
		return( toIntCoordinates( fromNormalized( percentagesPoint, rect ) ) );
	}

	public Point2d toIntCoordinates( Point2d pt )
	{
		if( pt != null )
		{
			pt.x = toInt( pt.x );
			pt.y = toInt( pt.y );
		}

		return( pt );
	}

	public int toInt( double value )
	{
		return( (int) Math.round( value ) );
	}

	public Point2d toNormalized( Point2d pt, GeometricRectangle rect )
	{
		Point2d result = null;
		if( ( pt != null ) && ( rect != null ) )
		{
			double xx = ( pt.x - rect.x ) / rect.width ;
			double yy = ( pt.y - rect.y ) / rect.height ;

			result = getFactory().createPoint2d(xx, yy);
		}

		return( result );
	}

	public boolean checkNormalizedPoint( Point2d normPt )
	{
		return( _NORMALIZED_RECTANGLE.isInside(normPt) );
	}

	protected boolean isInt( double value )
	{
		return( DoubleFunctions.instance().isInt(value) );
	}

	public boolean coordinatesAreInt( Point2d pt )
	{
		boolean result = false;
		if( pt != null )
			result = isInt(pt.x) && isInt(pt.y);

		return( result );
	}

	public VertexModel createVertexModel( Point2d pt, GeometricRectangle bounds,
											AtomicReference<ListWrapper<ColorModel>> colors )
	{
		return( createVertexModel( pt, toNormalized( pt, bounds ), colors ) );
	}

	protected VertexModel createVertexModel( Point2d pt, Point2d normalizedPt,
											AtomicReference<ListWrapper<ColorModel>> colors )
	{
		VertexModel result = (colors == null) ?
							new VertexModel() :
							new VertexModel( colors );

		result.setPoint(pt);
		result.setNormalizedPoint(normalizedPt);

		return( result );
	}

	protected int zoomToInt( double value, double zoomFactor )
	{
		return( IntegerFunctions.zoomValueRound( value, zoomFactor ) );
	}

	protected double zoom( double value, double zoomFactor )
	{
		return( value * zoomFactor );
	}

	public GeometricRectangle unzoomtoInt(GeometricRectangle rect, double zoomFactor )
	{
		return( zoomToInt( rect, 1.0d / zoomFactor ) );
	}

	public GeometricRectangle zoomToInt(GeometricRectangle rect, double zoomFactor )
	{
		GeometricRectangle result = rect;
		if( ( rect != null ) && ( zoomFactor != 1.0d ) )
		{
			result = rect.createCopy();

			result.x = MorphingHelper.this.zoomToInt( rect.x, zoomFactor );
			result.y = MorphingHelper.this.zoomToInt( rect.y, zoomFactor );

			double right = MorphingHelper.this.zoomToInt( rect.x + rect.width, zoomFactor );
			double bottom = MorphingHelper.this.zoomToInt( rect.y + rect.height, zoomFactor );

			result.width = right - result.x;
			result.height = bottom - result.y;
		}

		return( result );
	}

	public Point2d unzoom( Point2d pt, double zoomFactor )
	{
		return( MorphingHelper.this.zoomToInt( pt, 1.0d / zoomFactor ) );
	}

	public Point2d zoom( Point2d pt, double zoomFactor )
	{
		return( zoom(pt, zoomFactor, this::zoom ) );
	}

	public Point2d zoomToInt( Point2d pt, double zoomFactor )
	{
		return( zoom(pt, zoomFactor, this::zoomToInt ) );
	}

	protected Point2d zoom( Point2d pt, double zoomFactor,
							BiFunction<Double, Double, Number> zoomFunction )
	{
		Point2d result = pt;
		if( ( pt != null ) && ( zoomFactor != 1.0d ) )
		{
			result = pt.createCopy();

			result.x = zoomFunction.apply( result.x, zoomFactor ).doubleValue();
			result.y = zoomFunction.apply( result.y, zoomFactor ).doubleValue();
		}

		return( result );
	}

	public GeometricRectangle intersection(GeometricRectangle rect1,
                                           GeometricRectangle rect2 )
	{
		GeometricRectangle result = null;
		
		if( rect1 == null )
			result = rect2;
		else if( rect2 == null )
			result = rect1;
		else
		{
			result = rect1.createCopy();

			double xx = max( rect1.x, rect2.x );
			double yy = max( rect1.y, rect2.y );
			double right = min( rect1.x + rect1.width, rect2.x + rect2.width );
			double bottom = min( rect2.y + rect2.height, rect2.y + rect2.height );

			result.x = xx;
			result.y = yy;
			result.width = max( 0, right - xx );
			result.height = max( 0, bottom - yy );
		}

		return( result );
	}

	protected double max( double v1, double v2 )
	{
		return( DoubleFunctions.instance().max(v1, v2) );
	}

	protected double min( double v1, double v2 )
	{
		return( DoubleFunctions.instance().min(v1, v2) );
	}

	public boolean canDiscard(GeometricRectangle bounds, Point2d ... pts)
	{
		return( Point2dDiscarder.instance().canDiscard(bounds, pts) );
	}

	public double getAspectRatio( double width, double height )
	{
		return( ((double) width) / height );
	}

	public double getAspectRatio( BufferedImage image )
	{
		return( getAspectRatio(image.getWidth(), image.getHeight() ) );
	}

	public double getAspectRatio(GeometricRectangle originalBounds)
	{
		return( getAspectRatio( originalBounds.width, originalBounds.height ) );
	}

	public int calculateHeightKeepingAspectRatio( double width,
							GeometricRectangle originalBounds)
	{
		return( MorphingHelper.this.zoomToInt( width, 1 / getAspectRatio(originalBounds) ) );
	}

	public int calculateWidthKeepingAspectRatio( double height,
							GeometricRectangle originalBounds )
	{
		return( MorphingHelper.this.zoomToInt( height, getAspectRatio(originalBounds) ) );
	}

	public void limitBoundsSize(GeometricRectangle result, BufferedImage image,
                                GeometricRectangle originalBounds )
	{
		double imageAspectRatio = getAspectRatio( image );
		double originalBoundsAspectRatio = getAspectRatio( originalBounds );
		if( imageAspectRatio < originalBoundsAspectRatio )
		{
			limitHeightSize( result, image, originalBounds );
			limitWidthSize( result, image, originalBounds );
		}
		else
		{
			limitWidthSize( result, image, originalBounds );
			limitHeightSize( result, image, originalBounds );
		}
	}

	protected void limitBoundsPosition(GeometricRectangle result, BufferedImage image )
	{
		result.x = limit( result.x, 0, image.getWidth() - result.width);

		result.y = limit( result.y, 0, image.getHeight() - result.height);
	}

	public void limitNewBounds(GeometricRectangle result, BufferedImage image,
                               GeometricRectangle originalBounds )
	{
		limitBoundsSize(result, image, originalBounds);
		limitBoundsPosition(result, image);
	}

	protected double limit( double value, double min, double max )
	{
		return( DoubleFunctions.instance().limit( value, min, max ) );
	}

	protected void limitWidthSize(GeometricRectangle result, BufferedImage image,
                                  GeometricRectangle originalBounds )
	{
		result.width = limit( result.width, MorphingConstants.MINIMUM_SIZE_FOR_BOUNDS,
							image.getWidth());
		result.height = calculateHeightKeepingAspectRatio(result.width, originalBounds);
	}

	protected void limitHeightSize(GeometricRectangle result, BufferedImage image,
                                   GeometricRectangle originalBounds )
	{
		result.height = limit( result.height, MorphingConstants.MINIMUM_SIZE_FOR_BOUNDS,
							image.getHeight());
		result.width = calculateWidthKeepingAspectRatio(result.height, originalBounds);
	}
}
