/*
 * Copyright (c) 2023. Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 3.0 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You may obtain a copy of the License at
 *
 *       http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */

package com.frojasg1.applications.libmorphing.mapper;

import com.frojasg1.libgeometry.algebra.helpers.GeometryHelper;
import com.frojasg1.libgeometry.algebra.helpers.VectorFunctions;
import com.frojasg1.applications.libmorphing.model.images.vertex.VertexModel;
import com.frojasg1.applications.libmorphing.model.images.vertex.map.impl.control.DynamicMultimapVertexWithIndex;
import com.frojasg1.applications.libmorphing.model.images.vertex.map.impl.control.IndexedVerticesMap;
import com.frojasg1.libgeometry.types.GeometricRectangle;
import com.frojasg1.libgeometry.types.Point2d;
import com.frojasg1.libgeometry.types.factory.GeometryCommonFactory;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.function.UnaryOperator;

public class ControlMeshMapper {
    private static final Logger LOGGER = LoggerFactory.getLogger(ControlMeshMapper.class);

    protected static ControlMeshMapper INSTANCE = new ControlMeshMapper();

    public static ControlMeshMapper instance() {
        return INSTANCE;
    }

    protected GeometryHelper getGeometryHelper() {
        return GeometryHelper.instance();
    }

    protected VectorFunctions getVectorFunctions() {
        return VectorFunctions.instance();
    }

    protected boolean isInside(Point2d pt, Point2d a, Point2d b, Point2d c) {
        return getVectorFunctions().isApproximatelyInside(pt, a, b, c, 0.5d);
    }

    protected GeometryCommonFactory getFactory() {
        return GeometryCommonFactory.instance();
    }

    public Point2d createPoint2d(double xx, double yy) {
        return getFactory().createPoint2d(xx, yy);
    }

    protected Point2d mapTrianglePt(Point2d pt, Point2d pt11, Point2d pt12, Point2d pt13, Point2d pt21, Point2d pt22, Point2d pt23) {
        return getGeometryHelper().mapTrianglePt(pt, pt11, pt12, pt13, pt21, pt22, pt23);
    }

    protected Point2d getVertex(IndexedVerticesMap vertices, String name, int index) {
        return vertices.get(index).get(name).getPoint();
    }

    protected Point2d[] fillTriangle(IndexedVerticesMap vertices, String name, int[] triangleIndices, Point2d[] result) {
        int ind = 0;
        for(int triangleIndex: triangleIndices) {
            result[ind++] = getVertex(vertices, name, triangleIndex);
        }
        return result;
    }

    protected Integer getIndex(IndexedVerticesMap vertices, String mapName, Point2d pt) {
        DynamicMultimapVertexWithIndex multiVertex = vertices.get(mapName, pt);
        Integer result = null;
        if(multiVertex != null) {
            result = multiVertex.getIndex();
        }
        return result;
    }

    public Point2d map(Point2d originPt, UnaryOperator<Point2d> pointLimiter,
                       IndexedVerticesMap vertices, String originName, String destinationName,
                       int[][] trianglesIndices, Point2d[] tmpOriginTriangle) {
        Point2d result = null;
        Point2d limitedPt = limit(originPt, pointLimiter);;
        for(int[] triangleIndices: trianglesIndices) {
            tmpOriginTriangle = fillTriangle(vertices, originName, triangleIndices, tmpOriginTriangle);
            if(isInside(originPt, tmpOriginTriangle[0], tmpOriginTriangle[1], tmpOriginTriangle[2])) {
                result = mapTrianglePt(limitedPt,
                        tmpOriginTriangle[0], tmpOriginTriangle[1], tmpOriginTriangle[2],
                        getVertex(vertices, destinationName, triangleIndices[0]),
                        getVertex(vertices, destinationName, triangleIndices[1]),
                        getVertex(vertices, destinationName, triangleIndices[2])
                        );
                break;
            }
        }
        if(result == null) {
            LOGGER.warn("Point[{}][{}] ({}) could not be mapped in: {}",
                    originName,
                    getIndex(vertices, originName, originPt),
                    originPt,
                    vertices.getMapByName(originName, Object.class));
        }
        return result;
    }

    protected Point2d limit(Point2d pt, UnaryOperator<Point2d> limiter) {
        Point2d result = pt;
        if(limiter != null) {
            result = limiter.apply(result);
        }
        return result;
    }

    public Point2d map(Point2d originPt, UnaryOperator<Point2d> limiter,
                       IndexedVerticesMap vertices, String originName, String destinationName,
                       int[][] triangles) {
        return map(originPt, limiter, vertices, originName, destinationName, triangles, new Point2d[3]);
    }

    public void createIdealNormalizedIndexedVerticesMap(IndexedVerticesMap idealControl80MeshVertices,
                                                        String idealName, String normalizedName,
                                                        int[][] trianglesIndicesResult,
                                                        IndexedVerticesMap result, GeometricRectangle bounds) {
        Point2d ulid = getVertex(idealControl80MeshVertices, idealName, 68);
        Point2d dlid = getVertex(idealControl80MeshVertices, idealName, 71);
        Point2d drid = getVertex(idealControl80MeshVertices, idealName, 74);
        Point2d urid = getVertex(idealControl80MeshVertices, idealName, 77);
        double width = ulid.distance(urid);
        double height = ulid.distance(dlid);

        Point2d ulnorm = createPoint2d(0, 0);
        Point2d dlnorm = createPoint2d(0, height).createIntCopyIfNotInt();
        Point2d urnorm = createPoint2d(width, 0).createIntCopyIfNotInt();
        Point2d drnorm = createPoint2d(width, height).createIntCopyIfNotInt();

        result.addMapByName(idealName);
        result.addMapByName(normalizedName);

        result.set(idealName, 0, buildVertexModel(ulid, bounds));
        result.set(idealName, 1, buildVertexModel(dlid, bounds));
        result.set(idealName, 2, buildVertexModel(drid, bounds));
        result.set(idealName, 3, buildVertexModel(urid, bounds));

        result.set(normalizedName, 0, buildVertexModel(ulnorm, null));
        result.set(normalizedName, 1, buildVertexModel(dlnorm, null));
        result.set(normalizedName, 2, buildVertexModel(drnorm, null));
        result.set(normalizedName, 3, buildVertexModel(urnorm, null));

        trianglesIndicesResult[0] = new int[]{0, 1, 2};
        trianglesIndicesResult[1] = new int[]{2, 3, 0};
    }

    protected VertexModel buildVertexModel(Point2d pt, GeometricRectangle bounds) {
        return VertexModel.build(pt, bounds);
    }
}
