/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.applications.libmorphing.model.images.image;

import com.frojasg1.applications.libmorphing.api.facedetector.pojo.FaceDetection;
import com.frojasg1.applications.libmorphing.frames.framebuilder.impl.MorphingTransition;
import com.frojasg1.applications.libmorphing.helpers.MorphingHelper;
import com.frojasg1.applications.libmorphing.model.facedetector.ImageFaceDetectionPersistenceModel;
import com.frojasg1.applications.libmorphing.model.images.configuration.MorphingProjectConfigurationModel;
import com.frojasg1.applications.libmorphing.model.images.vertex.VertexModel;
import com.frojasg1.libgeometry.types.GeometricRectangle;
import com.frojasg1.libgeometry.types.Point2d;
import com.frojasg1.general.ExecutionFunctions;
import com.frojasg1.general.copy.CopyFacilities;
import com.frojasg1.general.xml.model.NameKeyWithFileNameModelBase;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.stream.Collectors;
import javax.imageio.ImageIO;

/**
 *
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 */
public class MorphingImageModel
	extends NameKeyWithFileNameModelBase
	implements CopyFacilities<MorphingImageModel>
{
	protected boolean _isMaster;
	protected GeometricRectangle _bounds;
	protected MorphingProjectConfigurationModel _configurationModel;
	protected List<VertexModel> _vertices;

	protected MorphingHelper _morphingHelper;

	protected BufferedImage _image;

	protected FaceDetection faceDetection;

	protected int imageFaceDetectionIndex = -1;

	protected List<Point2d> faceDetection80Points = new ArrayList<>();

	protected ImageFaceDetectionPersistenceModel imageFaceDetectionPersistenceModel;

	protected MorphingTransition followingTransition = MorphingTransition.NORMAL;


	public MorphingImageModel()
	{
		this(MorphingHelper.instance(), true);
	}

	protected MorphingImageModel(MorphingHelper morphingHelper, boolean init)
	{
		super(null);
		_morphingHelper = morphingHelper;
		if( init )
			init();
	}

	protected void init()
	{
		_vertices = createList();
	}

	@Override
	public MorphingImageModel createCopy()
	{
		MorphingImageModel result = new MorphingImageModel( null, false );
		result.copy(this);

		return( result );
	}

	@Override
	public void copy(MorphingImageModel that)
	{
		_isMaster = that._isMaster;
		_bounds = that._bounds.createCopy();
		_configurationModel = copy( _configurationModel, that._configurationModel );
		_vertices = copy( that._vertices );
		_morphingHelper = that._morphingHelper;
		faceDetection = that.faceDetection;
		imageFaceDetectionPersistenceModel = that.imageFaceDetectionPersistenceModel;
		followingTransition = that.followingTransition;
		imageFaceDetectionIndex = that.imageFaceDetectionIndex;
		setName( that.getName() );
		setFileName( that.getFileName() );
	}

	protected <C> List<C> copy( List<C> that )
	{
		List<C> result = null;
		if( that != null )
		{
			result = createList();
			result.addAll(that);
		}

		return( result );
	}

	public MorphingTransition getFollowingTransition() {
		return followingTransition;
	}

	public void setFollowingTransition(MorphingTransition followingTransition) {
		this.followingTransition = followingTransition;
	}

	protected <C> List<C> createList()
	{
		return( new ArrayList<>() );
	}

	protected void setImage( BufferedImage image )
	{
		_image = image;
	}

	protected void assignImage( String fileName ) throws IOException
	{
		setImage( ImageIO.read( new File( fileName ) ) );
	}

	public BufferedImage getImage() {
		return _image;
	}

	@Override
	public void setFileName( String fileName )
	{
		if( getFileName() == null )
			ExecutionFunctions.instance().safeMethodExecution( () -> assignImage( fileName ) );
		else
			ExecutionFunctions.instance().runtimeExceptionMethodExecution( () -> assignImage( fileName ) );

		super.setFileName(fileName);
	}

	public MorphingHelper getMorphingHelper() {
		return _morphingHelper;
	}

	public boolean isMaster() {
		return _isMaster;
	}

	public void setIsMaster(boolean _isMaster) {
		this._isMaster = _isMaster;
	}

	public GeometricRectangle getBounds() {
		return _bounds;
	}

	protected VertexModel createVertexModel(Point2d pt, GeometricRectangle bounds) {
		return VertexModel.build(pt, bounds);
	}

	protected void addCorners(GeometricRectangle bounds) {
		_vertices.addAll(
				bounds.getCorners().stream()
						.map(pt -> createVertexModel(pt, bounds))
						.collect(Collectors.toList()));
	}

	protected void removeCorners(GeometricRectangle bounds) {
		Iterator<VertexModel> it = _vertices.iterator();
		VertexModel elem;
		while(it.hasNext()) {
			elem = it.next();
			if(bounds.isCorner(elem.getPoint())) {
				it.remove();
			}
		}
	}

	public void setBounds(GeometricRectangle _bounds) {
		if(this._bounds != null) {
			removeCorners(this._bounds);
			addCorners(_bounds);
		}

		this._bounds = _bounds;
	}

	public MorphingProjectConfigurationModel getConfigurationModel() {
		return _configurationModel;
	}

	public void setConfigurationModel(MorphingProjectConfigurationModel _configurationModel) {
		this._configurationModel = _configurationModel;
	}

	public List<VertexModel> getVertices() {
		return _vertices;
	}

	protected Point2d createNormalizedPoint( Point2d pt )
	{
		return( getMorphingHelper().toNormalized(pt, getBounds() ) );
	}

	protected boolean checkNormalizedPoint( Point2d normPt )
	{
		return( getMorphingHelper().checkNormalizedPoint(normPt) );
	}

	public VertexModel createEmptyVertexModel( Point2d imagePoint )
	{
		VertexModel result = new VertexModel();
		result.setPoint(imagePoint);

		Point2d normalizedPoint = createNormalizedPoint( imagePoint );
		if( !checkNormalizedPoint(normalizedPoint) )
			result = null;
		else
			result.setNormalizedPoint(normalizedPoint);

		return( result );
	}

	public synchronized void addVertex( VertexModel vertexModel )
	{
		getVertices().add(vertexModel);
	}

	public synchronized boolean removeVertexModel( VertexModel vertexModel )
	{
		return( getVertices().remove(vertexModel) );
	}

	public FaceDetection getFaceDetection() {
		return faceDetection;
	}

	public void setFaceDetection(FaceDetection faceDetection) {
		this.faceDetection = faceDetection;
	}

	public int getImageFaceDetectionIndex() {
		return imageFaceDetectionIndex;
	}

	public void setImageFaceDetectionIndex(int imageFaceDetectionIndex) {
		this.imageFaceDetectionIndex = imageFaceDetectionIndex;
	}

	public List<Point2d> getFaceDetection88Points() {
		return faceDetection80Points;
	}

	public void clearFaceDetectionPoints() {
		getFaceDetection88Points().clear();
	}

	public void addFaceDetectionPoint(Point2d pt) {
		getFaceDetection88Points().add(pt);
	}

	public ImageFaceDetectionPersistenceModel getImageFaceDetectionPersistenceModel() {
		return imageFaceDetectionPersistenceModel;
	}

	public void setImageFaceDetectionPersistenceModel(ImageFaceDetectionPersistenceModel imageFaceDetectionPersistenceModel) {
		this.imageFaceDetectionPersistenceModel = imageFaceDetectionPersistenceModel;
	}
}
