/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.applications.libmorphing.model.images.vertex;

import com.frojasg1.applications.libmorphing.helpers.MorphingHelper;
import com.frojasg1.applications.libmorphing.model.images.color.ColorModel;
import com.frojasg1.libgeometry.types.GeometricRectangle;
import com.frojasg1.libgeometry.types.Point2d;
import com.frojasg1.general.list.ConcurrentListWrapper;
import com.frojasg1.general.list.ListWrapper;
import java.util.Objects;
import java.util.concurrent.atomic.AtomicReference;

/**
 *
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 */
public class VertexModel
{
	protected Point2d _normalizedPoint;
	protected Point2d _point;
	protected AtomicReference<ListWrapper<ColorModel>> _colors;

	public VertexModel()
	{
		this( true );
	}

	protected VertexModel( boolean init )
	{
		if( init )
			init();
	}

	public VertexModel( AtomicReference<ListWrapper<ColorModel>> colors )
	{
		_colors = colors;
	}

	protected void init()
	{
		initColors();
	}

	public VertexModel createCopy()
	{
		VertexModel result = new VertexModel();
		result.copy(this);

		return( result );
	}

	public void copy( VertexModel that )
	{
		_normalizedPoint = that._normalizedPoint.createCopy();
		_point = that._point.createCopy();
	}

	protected void initColors()
	{
		_colors = new AtomicReference<>(createList());
	}

	protected <C> ListWrapper<ColorModel> createList()
	{
		return( new ConcurrentListWrapper<>() );
	}

	public void add( ColorModel cm )
	{
		if( !getColors().contains(cm) )
			getColors().add( cm );
	}

	public Point2d getPoint() {
		return _point;
	}

	public Point2d getNormalizedPoint() {
		return _normalizedPoint;
	}

	public void setNormalizedPoint(Point2d _normalizedPoint) {
		this._normalizedPoint = _normalizedPoint;
	}

	public void setPoint(Point2d _point) {
		this._point = _point;
	}

	public ListWrapper<ColorModel> getColors() {
		return( getColorsRef().get() );
	}

	public AtomicReference<ListWrapper<ColorModel>> getColorsRef() {
		return _colors;
	}

	@Override
	public String toString() {
		return "VertexModel{" + "_normalizedPoint=" + _normalizedPoint + ", _point=" + _point + ", _colors=" + _colors + '}';
	}

	@Override
	public int hashCode() {
		int hash = 7;
		hash = 29 * hash + Objects.hashCode(this._point);
		return hash;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj) {
			return true;
		}
		if (obj == null) {
			return false;
		}
		if (getClass() != obj.getClass()) {
			return false;
		}
		final VertexModel other = (VertexModel) obj;
		if (!Objects.equals(this._point, other._point)) {
			return false;
		}
		return true;
	}

	protected static Point2d toNormalized( Point2d pt, GeometricRectangle rect )
	{
		return( MorphingHelper.instance().toNormalized(pt, rect) );
	}

	public static VertexModel build(Point2d pt, GeometricRectangle bounds) {
		VertexModel result = new VertexModel();
		result.setPoint(pt);
		result.setNormalizedPoint(toNormalized(pt, bounds));

		return result;
	}
}
