/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.applications.libmorphing.view;

import java.awt.*;
import javax.swing.*;
import javax.swing.border.*;
import com.frojasg1.applications.common.components.internationalization.window.InternationalizationInitializationEndCallback;
import com.frojasg1.applications.common.components.internationalization.window.InternationalizedJFrame;
import com.frojasg1.applications.common.components.resizecomp.MapResizeRelocateComponentItem;
import com.frojasg1.applications.common.components.resizecomp.ResizeRelocateItem;
import com.frojasg1.applications.common.configuration.application.BaseApplicationConfigurationInterface;
import com.frojasg1.applications.libmorphing.contexts.MorphingEditorApplicationContext;
import com.frojasg1.applications.libmorphing.api.facedetector.pojo.FaceDetectionMeshData;
import com.frojasg1.applications.libmorphing.configuration.MorphingApplicationConfiguration;
import com.frojasg1.applications.libmorphing.controller.MorphingVerticesMapManager;
import com.frojasg1.applications.libmorphing.language.LibMorphingLanguageResources;
import com.frojasg1.applications.libmorphing.model.images.color.types.MultiTriangle;
import com.frojasg1.applications.libmorphing.model.images.image.MorphingImageModel;
import com.frojasg1.applications.libmorphing.mvc.DelaunayTriangulatorWindowView;
import com.frojasg1.applications.libmorphing.view.painters.TriangulationPainterLayerTypeEnum;
import com.frojasg1.applications.libmorphing.view.panels.DelaunayContentPanel;
import com.frojasg1.applications.libmorphing.view.panels.DelaunayViewerControlPanel;
import com.frojasg1.applications.libmorphing.view.panels.evt.DelaunayPanelsEventTypes;
import com.frojasg1.applications.libmorphing.view.panels.evt.ZoomEvt;
import com.frojasg1.libgeometry.types.GeometricRectangle;
import com.frojasg1.libgeometry.types.Point2d;
import com.frojasg1.general.desktop.view.ContainerFunctions;
import com.frojasg1.general.view.evt.wrapper.impl.GenComponentWrapperEvent;
import com.frojasg1.general.desktop.view.pdf.ImageJPanel;
import com.frojasg1.general.desktop.view.pdf.ImageJPanelControllerInterface;
import com.frojasg1.general.desktop.view.zoom.mapper.ComponentMapper;
import com.frojasg1.general.desktop.view.zoom.mapper.ComposedComponent;
import com.frojasg1.general.number.DoubleReference;
import com.frojasg1.general.view.evt.wrapper.GenericWrapperEvent;
import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Point;
import java.util.List;
import java.util.Vector;
import java.util.function.Consumer;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.SwingUtilities;

// Minimum Size: [570, 780]

/**
 *
 * @author Usuario
 */
public class DelaunayTriangulationViewerWindowBase< CC extends MorphingEditorApplicationContext>
	extends InternationalizedJFrame<CC>
	implements DelaunayTriangulatorWindowView<CC>, ImageJPanelControllerInterface
{
	public static final String sa_configurationBaseFileName = "DelaunayTriangulationViewerWindow";

	protected DelaunayViewerControlPanel _viewerControlView = null;
	protected DelaunayContentPanel _contentPanel = null;

	/**
	 * Creates new form MainWindow
	 */
	public DelaunayTriangulationViewerWindowBase(CC applicationContext )
	{
		super( applicationContext.getAppliConf(), applicationContext, true );
		init();
	}

	public DelaunayTriangulationViewerWindowBase( BaseApplicationConfigurationInterface appliConf,
								CC applicationContext,
								Consumer<InternationalizationInitializationEndCallback> initializationEndCallBack )
	{
		super( appliConf, applicationContext, initializationEndCallBack, true );
		init();
	}

	protected void init()
	{
		initComponents();

		initPanels();

		SwingUtilities.invokeLater( () -> setWindowConfiguration( ) );
	}

	protected DelaunayContentPanel createContentPanel()
	{
		DelaunayContentPanel result = new DelaunayContentPanel( getApplicationContext() );
		result.getImagePanel().setController( this );

		return( result );
	}

	public DelaunayViewerControlPanel getViewerControlView() {
		return _viewerControlView;
	}

	protected DelaunayViewerControlPanel createViewerControlView()
	{
		return( new DelaunayViewerControlPanel(getAppliConf(),
												getContentPanel().getListenersList(),
												getContentPanel().getEventBuilder()) );
	}

	protected Component getControlViewComponent()
	{
		return( (Component) _viewerControlView );
	}

	protected ComposedComponent getControlViewComposedComponent()
	{
		return( (ComposedComponent) _viewerControlView );
	}

	protected DelaunayViewerControlPanel getControlView()
	{
		return( _viewerControlView );
	}

	protected DelaunayContentPanel getContentPanel()
	{
		return( _contentPanel );
	}

	protected ImageJPanel getContentImageJPanel()
	{
		ImageJPanel result = null;
		if( getContentPanel() != null )
			result = getContentPanel().getImagePanel();

		return( result );
	}

	protected void initPanels()
	{
		try
		{
			int yy = 0;
			Container contentPane = getContentPane();
			_contentPanel = createContentPanel();
			_viewerControlView = createViewerControlView();

			contentPane.add( getControlViewComponent() );
			Dimension size = getControlViewComposedComponent().getInternalSize();
			getControlViewComponent().setBounds( 0, yy, size.width, size.height );
			yy += size.height;

			jPanel2.setBounds( 0, yy, size.width, jPanel2.getHeight() );
			ContainerFunctions.instance().addComponentToCompletelyFillParent(jPanel2, _contentPanel);

			pack();
		}
		catch( Throwable th )
		{
			th.printStackTrace();
		}
	}

	protected MapResizeRelocateComponentItem getMapResizeRelocateComponentItem()
	{
		MapResizeRelocateComponentItem mapRRCI = new MapResizeRelocateComponentItem();
		try
		{
			mapRRCI.putResizeRelocateComponentItem(_contentPanel, ResizeRelocateItem.FILL_WHOLE_PARENT );
			mapRRCI.putAll( getControlViewComposedComponent().getResizeRelocateInfo() );

			mapRRCI.putResizeRelocateComponentItem( jPanel2, ResizeRelocateItem.RESIZE_TO_RIGHT + ResizeRelocateItem.RESIZE_TO_BOTTOM );
			mapRRCI.putResizeRelocateComponentItem(_contentPanel, ResizeRelocateItem.FILL_WHOLE_PARENT );
			mapRRCI.putAll(_contentPanel.getResizeRelocateInfo() );
		}
		catch( Throwable th )
		{
			th.printStackTrace();
		}

		return( mapRRCI );
	}

	protected Vector<JPopupMenu> getJPopupMenuVector()
	{
		return( null );
	}

	protected void setWindowConfiguration( )
	{
		MapResizeRelocateComponentItem mapRRCI = getMapResizeRelocateComponentItem();
		Vector<JPopupMenu> popupMenuVector = getJPopupMenuVector();

		createInternationalization(	getAppliConf().getConfigurationMainFolder(),
									getAppliConf().getApplicationNameFolder(),
									getAppliConf().getApplicationGroup(),
									LibMorphingLanguageResources.sa_PROPERTIES_PATH_IN_JAR,
									sa_configurationBaseFileName,
									this,
									null,
									popupMenuVector,
									true,
									mapRRCI );
	}

	@Override
	public void internationalizationInitializationEndCallback()
	{
		super.internationalizationInitializationEndCallback();

		updateZoomComboBoxValues();

		addListeners();
	}

	/**
	 * This method is called from within the constructor to initialize the form.
	 * WARNING: Do NOT modify this code. The content of this method is always
	 * regenerated by the Form Editor.
	 */
	@SuppressWarnings("unchecked")
    // <editor-fold defaultstate="collapsed" desc="Generated Code">//GEN-BEGIN:initComponents
    // Generated using JFormDesigner non-commercial license
    private void initComponents() {
        jPanel2 = new JPanel();

        //======== this ========
        setDefaultCloseOperation(WindowConstants.DO_NOTHING_ON_CLOSE);
        setMinimumSize(new Dimension(570, 340));
        setName("");
        Container contentPane = getContentPane();
        contentPane.setLayout(null);

        //======== jPanel2 ========
        {
            jPanel2.setBorder(LineBorder.createBlackLineBorder());
            jPanel2.setName("jPanel2");
            jPanel2.setPreferredSize(new Dimension(510, 250));
            jPanel2.setLayout(null);

            {
                // compute preferred size
                Dimension preferredSize = new Dimension();
                for(int i = 0; i < jPanel2.getComponentCount(); i++) {
                    Rectangle bounds = jPanel2.getComponent(i).getBounds();
                    preferredSize.width = Math.max(bounds.x + bounds.width, preferredSize.width);
                    preferredSize.height = Math.max(bounds.y + bounds.height, preferredSize.height);
                }
                Insets insets = jPanel2.getInsets();
                preferredSize.width += insets.right;
                preferredSize.height += insets.bottom;
                jPanel2.setMinimumSize(preferredSize);
                jPanel2.setPreferredSize(preferredSize);
            }
        }
        contentPane.add(jPanel2);
        jPanel2.setBounds(0, 0, 550, 670);

        {
            // compute preferred size
            Dimension preferredSize = new Dimension();
            for(int i = 0; i < contentPane.getComponentCount(); i++) {
                Rectangle bounds = contentPane.getComponent(i).getBounds();
                preferredSize.width = Math.max(bounds.x + bounds.width, preferredSize.width);
                preferredSize.height = Math.max(bounds.y + bounds.height, preferredSize.height);
            }
            Insets insets = contentPane.getInsets();
            preferredSize.width += insets.right;
            preferredSize.height += insets.bottom;
            contentPane.setMinimumSize(preferredSize);
            contentPane.setPreferredSize(preferredSize);
        }
        pack();
        setLocationRelativeTo(getOwner());
    }// </editor-fold>//GEN-END:initComponents

    // Variables declaration - do not modify//GEN-BEGIN:variables
    // Generated using JFormDesigner non-commercial license
    private JPanel jPanel2;
    // End of variables declaration//GEN-END:variables

	protected void addListeners()
	{
		getControlView().addWrapperListener( this::processZoomFactorChange, DelaunayPanelsEventTypes.ZOOM_EVT );
	}

	protected void processZoomFactorChange( GenComponentWrapperEvent evt)
	{
		ZoomEvt<JPanel> zevt = (ZoomEvt) evt;
		setImageZoomFactor( zevt.getEventData().getNewFactor() );
	}

	protected DoubleReference getSelectedZoomFactor()
	{
		return( _viewerControlView.getSelectedZoomFactor() );
	}

	protected double getSelectedZoomFactorDouble()
	{
		double factor = 1.0D;
		
		DoubleReference dr = getSelectedZoomFactor();
		if( dr != null )
			factor = dr._value;
		
		return( factor );
	}

	protected void updateZoomComboBoxValues()
	{
		List<DoubleReference> listOfZoomFactors = _contentPanel.getStepsForFactorList();
		_viewerControlView.updateZoomComboBoxValues(listOfZoomFactors);
	}

	public void newImageSet()
	{
//		_pdfContentPanel.setPage( image, factor, pageIndex );
		updateZoomComboBoxValues();

		setZoomSelectedItem( getContentImageJPanel().getZoomFactor() );
		updateZoomFactor();
	}

	public void setZoomSelectedItem(DoubleReference selectedItem) {
		getControlView().setZoomSelectedItem( selectedItem );
	}

	protected void updateZoomFactor()
	{
		setImageZoomFactor( getSelectedZoomFactor() );
	}

	public void setImageZoomFactor( DoubleReference factor )
	{
		setZoomSelectedItem(factor);
		getContentPanel().setImageZoomFactor(factor);
	}

	public void block()
	{
		changeToWaitCursor();
	}

	public void unblock()
	{
		revertChangeToWaitCursor();
	}

	@Override
	public MorphingApplicationConfiguration getAppliConf()
	{
		return( (MorphingApplicationConfiguration) super.getAppliConf() );
	}

	@Override
	public void closeWindow()
	{
		boolean closeWindow = false;
		formWindowClosing( closeWindow );
		setVisible(false);
	}

	@Override
	public void changeZoomFactor( double zoomFactor )
	{
		double previousZoomFactor = _previousZoomFactor;
		
		super.changeZoomFactor( zoomFactor );

		DoubleReference newFactor = new DoubleReference( getSelectedZoomFactor()._value * zoomFactor / previousZoomFactor );
		setImageZoomFactor( newFactor );
	}

	@Override
	public void formWindowClosingEvent( )
	{
		closeWindow();
	}

	@Override
	protected void translateMappedComponents(ComponentMapper compMapper)
	{
		jPanel2 = compMapper.mapComponent( jPanel2 );
	}

	public void newZoomFactorSelected(DoubleReference newFactor)
	{
		setImageZoomFactor( newFactor );
	}

	protected void removeListeners()
	{
		getControlView().removeWrapperListener( this::processZoomFactorChange );
	}

	@Override
	public void releaseResources()
	{
		removeListeners();

		super.releaseResources();
	}

	@Override
	protected void changeZoomFactorPreventingToPaint(double zoomFactor, Point center)
	{
		a_intern.changeZoomFactor( zoomFactor, center );
	}

	@Override
	public void addDelaunayTriangulatorViewListener(Consumer<? super GenericWrapperEvent> listener, Long... eventTypes) {
		getContentPanel().addDelaunayTriangulatorViewListener(listener, eventTypes);
	}

	@Override
	public void removeDelaunayTriangulatorViewListener(Consumer<? super GenericWrapperEvent> listener) {
		getContentPanel().removeDelaunayTriangulatorViewListener(listener);
	}

	@Override
	public void setPointInProgress(String imageName, Point2d pt) {
		getContentPanel().setPointInProgress( imageName, pt );
	}

	@Override
	public void setClosestPoint(String imageName, Point2d pt) {
		getContentPanel().setClosestPoint( imageName, pt );
		repaint();
	}

	@Override
	public void setControlMeshClosestPoint(String imageName, Point2d pt) {
		getContentPanel().setControlMeshClosestPoint( imageName, pt );
	}

	@Override
	public void updateTriangulation(String imageName,
//									MorphingVerticesMapManager morphingVerticesManager,
									Point2d normalizedCenterPoint,
									boolean madatoryUpdate,
									Double newImageZoomFactor)
	{
		if( imageName != null )
			setTitle(imageName);

		getContentPanel().updateTriangulation(imageName,
//				getContentPanel().updateTriangulation(imageName, morphingVerticesManager,
											normalizedCenterPoint, madatoryUpdate,
											newImageZoomFactor);

		getControlView().setCurrentImageData(imageName, isMaster(//morphingVerticesManager,
				imageName) );
		getControlView().setMorphingTransition(getApplicationContext().getImageModel(imageName).getFollowingTransition(),
				getApplicationContext().allHaveFaceDetection());
		newImageSet();
	}

	@Override
	public double getImageZoomFactor()
	{
		return( getContentPanel().getImageZoomFactor() );
	}

	protected MorphingVerticesMapManager getVerticesManager() {
		return getApplicationContext().getVerticesManager();
	}

	protected boolean isMaster( //MorphingVerticesMapManager morphingVerticesManager,
								String imageName )
	{
		boolean result = false;
		MorphingVerticesMapManager morphingVerticesManager = getVerticesManager();
		if( morphingVerticesManager != null )
		{
			MorphingImageModel imageModel = morphingVerticesManager.getImageModel( imageName );
			if( imageModel != null )
				result = imageModel.isMaster();
		}

		return( result );
	}

	@Override
	public void previousPage() {
		// intentionally left blank
	}

	@Override
	public void nextPage() {
		// intentionally left blank
	}

	@Override
	public void setNewPdfZoomFactor(DoubleReference factor) {
		setImageZoomFactor( factor );
	}

	@Override
	public void setNewOnTheFlyBounds(String imageName, GeometricRectangle newOnTheFlyBounds) {
		getContentPanel().setNewOnTheFlyBounds(imageName, newOnTheFlyBounds);
	}

	@Override
	public double getFadingFraction()
	{
		return( getControlView().getFadingFraction() );
	}

	@Override
	public void fadeToMaster(double fadingFraction)
	{
		getContentPanel().fadeToMaster(fadingFraction);
	}

	@Override
	public void resetFadedToMasterImage()
	{
		getContentPanel().resetFadedToMasterImage();
	}

	@Override
	public Point2d getNormalizedPointImageCenterShown()
	{
		return( getContentPanel().getNormalizedPointImageCenterShown() );
	}	

	@Override
	public void setSelectedMultiTriangle( MultiTriangle multiTriangle )
	{
		getContentPanel().setSelectedMultiTriangle(multiTriangle);
	}

	@Override
	public MultiTriangle getSelectedMultiTriangle()
	{
		return( getContentPanel().getSelectedMultiTriangle() );
	}

	@Override
	public void resetSelectedMultiTriangle()
	{
		getContentPanel().resetSelectedMultiTriangle();
	}

	@Override
	public void setAllFaces(String imageName, List<FaceDetectionMeshData> allFaces) {
		getContentPanel().setAllFaces(imageName, allFaces);
	}

	@Override
	public void setLayerType(String imageName, TriangulationPainterLayerTypeEnum layerType) {
		getContentPanel().setLayerType(imageName, layerType);
	}

	@Override
	public void setFaceDetectionMeshData(String imageName, FaceDetectionMeshData faceDetectionMeshData) {
		getContentPanel().setFaceDetectionMeshData(imageName, faceDetectionMeshData);
	}
}
