/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.applications.libmorphing.view.newboundschooser;

import com.frojasg1.applications.common.configuration.application.BaseApplicationConfigurationInterface;
import com.frojasg1.libgeometry.types.GeometricRectangle;
import com.frojasg1.general.NullFunctions;
import com.frojasg1.general.desktop.image.ImageFunctions;
import com.frojasg1.general.desktop.view.color.ColorInversor;
import com.frojasg1.general.desktop.view.colors.Colors;
import com.frojasg1.general.desktop.view.painter.ExternalPainterBase;
import com.frojasg1.general.number.IntegerFunctions;
import com.frojasg1.general.view.ReleaseResourcesable;
import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.Stroke;
import java.awt.image.BufferedImage;
import java.util.function.Function;

/**
 *
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 */
public class BoundaryPainter extends ExternalPainterBase
								implements ReleaseResourcesable {
	protected static final double EDGE_WIDTH_100_PERCENT = 2d;
	protected static final int FONT_SIZE_100_PERCENT = 16;

	protected static final Color BOUNDS_COLOR = Color.WHITE;
	protected static final Color INNER_BOUNDS_COLOR = Color.BLACK;
	protected static final Color INSIDE_BOUNDS_COLOR = new Color(255, 255, 255, 150);
	protected static final Color TEXT_COLOR = Colors.GREY;

	protected Point _pointForCoordinates;

	protected NewBoundsChooserContext _context;

	protected BaseApplicationConfigurationInterface _appliConf;

	protected Font _font;

	/**
	 * Creates new form GraphicalEqualizerJPanel
	 */
	public BoundaryPainter(BaseApplicationConfigurationInterface appliConf,
							NewBoundsChooserContext context) {
		_appliConf = appliConf;
		_context = context;
	}

	protected <CC, RR> RR getIfNotNull( CC obj, Function<CC,RR> getter )
	{
		return( NullFunctions.instance().getIfNotNull(obj, getter) );
	}

	protected NewBoundsChooserContext getContext()
	{
		return( _context );
	}

	protected GeometricRectangle getBounds() {
		return getContext().getBounds();
	}

	protected Rectangle getBoundsOfImageInsidePanel() {
		return getContext().getBoundsOfImageInsidePanel();
	}

	public BufferedImage getRealImage() {
		return getContext().getRealImage();
	}

	public Point getPoint() {
		return getContext().getPoint();
	}

	@Override
	public void paintComponent(Component comp, Graphics grp)
	{
		super.paintComponent( comp, grp );
	}

	public Font getFont() {
		if( _font == null )
			_font = createFont();

		return _font;
	}

	@Override
	protected void paintComponentInternal(Component comp, Graphics grp)
	{
//		blackScreen( comp, grp );

		GeometricRectangle bounds = getBounds();
		if( bounds != null )
		{
			Rectangle viewBounds = boundsToView(bounds);
			if( viewBounds != null )
			{
				Graphics2D g2d = (Graphics2D) grp.create();

				paintInside(g2d, viewBounds);
				paintFrame(g2d, viewBounds);

				paintCoordinates( g2d, getPoint() );

				g2d.dispose();
			}
		}
	}

	protected void zoom( Rectangle result, double factor )
	{
		result.x = zoomValueCeil(result.x, factor);
		result.y = zoomValueCeil(result.y, factor);
		result.width = zoomValueCeil(result.width, factor);
		result.height = zoomValueCeil(result.height, factor);
	}

	protected double getViewFactor()
	{
		return( getContext().getViewFactor() );
	}

	protected Rectangle boundsToView(GeometricRectangle bounds )
	{
		Rectangle result = null;

		if( bounds != null )
		{
			result = new Rectangle((int)bounds.x, (int)bounds.y,
										(int)bounds.width, (int) bounds.height);
			double viewFactor = getViewFactor();
			zoom(result, viewFactor);

			Rectangle boundsOfImageInsidePanel = getBoundsOfImageInsidePanel();

			result.x += boundsOfImageInsidePanel.x;
			result.y += boundsOfImageInsidePanel.y;
		}

		return( result );
	}

	public void paintInside( Graphics grp, Rectangle viewBounds )
	{
		grp.setColor( INSIDE_BOUNDS_COLOR );
		grp.fillRect(viewBounds.x, viewBounds.y,
					viewBounds.width, viewBounds.height);
	}

	protected int getFontSize()
	{
		return( getZoomedValue(FONT_SIZE_100_PERCENT));
	}

	protected Font createFont()
	{
		return( new Font( "Lucida", Font.BOLD, getFontSize() ) );
	}

	protected void paintStringCentered( Graphics grp, String string, Color color, Point point )
	{
		ImageFunctions.instance().paintStringCentered(grp, getFont(),
			string, color, point );
	}

	public void paintCoordinates( Graphics2D g2d, Point point )
	{
		Point pointForCoordinates = getPointForCoordinates();
		if( ( point != null ) && ( pointForCoordinates != null ) )
		{
			paintStringCentered( g2d, getString( point ), TEXT_COLOR,
				pointForCoordinates );
		}
	}

	protected String getString( Point point )
	{
		return( String.format( "( %d, %d )", point.x, point.y ) );
	}

	protected Point getPointForCoordinates()
	{
		if( _pointForCoordinates == null )
			_pointForCoordinates = calculatePointForCoordinates();

		return( _pointForCoordinates );
	}

	protected Dimension getPanelSize()
	{
		return( getContext().getPanelSize() );
	}

	protected Point calculatePointForCoordinates()
	{
		Point result = null;
		Dimension panelSize = getPanelSize();
		if( panelSize != null )
		{
			result = new Point();

			result.x = zoomValueCeil( panelSize.width, 0.5d );
			result.y = zoomValueCeil( panelSize.height, 0.95d );
		}

		return( result );
	}

	public void paintFrame( Graphics grp, Rectangle viewBounds )
	{
		int width = getZoomedValue( EDGE_WIDTH_100_PERCENT );
		Stroke solid = new BasicStroke(width);
		((Graphics2D)grp).setStroke(solid);

		grp.setColor( BOUNDS_COLOR );
		grp.drawRect(viewBounds.x, viewBounds.y,
					viewBounds.width, viewBounds.height);

		grp.setColor( INNER_BOUNDS_COLOR );
		grp.drawRect(viewBounds.x + width, viewBounds.y + width,
					viewBounds.width - 2 * width, viewBounds.height - 2 * width);
	}

	public int zoomValueCeil( double value, double zoomFactor )
	{
		return( IntegerFunctions.zoomValueCeil(value, zoomFactor) );
	}

	protected int getZoomedValue( double value )
	{
		// application zoom factor may be different from _imageZoomFactor (the level of zoom the the image).
		return( zoomValueCeil(value, getAppliConf().getZoomFactor() ) );
	}

	@Override
	public void releaseResources()
	{
	}

	protected BaseApplicationConfigurationInterface getAppliConf()
	{
		return( _appliConf );
	}

	@Override
	public void invertColors( ColorInversor colorInversor)
	{
/*
		if( ! _doNotInvertColors )
		{
			invertSinglePanelColors(colorInversor);
			invertColorsChild(colorInversor);
		}
*/
	}
}
