/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.applications.libmorphing.view.newboundschooser;

import com.frojasg1.applications.libmorphing.constants.MorphingConstants;
import com.frojasg1.libgeometry.types.GeometricRectangle;
import com.frojasg1.libgeometry.types.Point2d;
import com.frojasg1.libgeometry.types.factory.GeometryCommonFactory;
import com.frojasg1.general.desktop.view.zoom.mapper.ComponentMapper;
import com.frojasg1.general.desktop.view.zoom.mapper.InternallyMappedComponentBase;
import com.frojasg1.general.number.DoubleFunctions;
import com.frojasg1.general.number.IntegerFunctions;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.image.BufferedImage;
import java.util.function.Predicate;
import javax.swing.JSpinner;
import javax.swing.SpinnerModel;

/**
 *
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 */
public class NewBoundsChooserContext extends InternallyMappedComponentBase
{

	protected Component _parent;

	protected GeometricRectangle _masterBounds;
	protected BufferedImage _realImage;
	protected Rectangle _boundsOfImageInsidePanel;
	protected GeometricRectangle _bounds;

	protected GeometricRectangle _originalBounds;

	protected Point _point;

	protected boolean _keepAspectRatio;

	protected Runnable _boundsChangedFunction;

	protected JSpinner _xSpinner;
	protected JSpinner _ySpinner;
	protected JSpinner _widthSpinner;
	protected JSpinner _heightSpinner;

	protected boolean _insideSpinnerChange = false;

	public NewBoundsChooserContext( Component jframe )
	{
		super( jframe );
	}

	public void setParent(Component _parent) {
		this._parent = _parent;
	}

	public Component getParent() {
		return _parent;
	}

	public BufferedImage getRealImage() {
		return _realImage;
	}

	public void setRealImage(BufferedImage _realImage) {
		this._realImage = _realImage;
	}

	public Rectangle getBoundsOfImageInsidePanel() {
		return _boundsOfImageInsidePanel;
	}

	public void setBoundsOfImageInsidePanel(Rectangle _boundsOfImageInsidePanel) {
		this._boundsOfImageInsidePanel = _boundsOfImageInsidePanel;
	}

	public GeometricRectangle getBounds() {
		return _bounds;
	}

	public void setBoundsUpdatingSpinners(GeometricRectangle _bounds) {
		setBounds( _bounds );

		updateSpinners();
	}

	public void setBounds(GeometricRectangle _bounds) {
		this._bounds = _bounds;
		invokeBoundsChangedFunction();
	}

	public GeometricRectangle getOriginalBounds() {
		return _originalBounds;
	}

	public void setOriginalBounds(GeometricRectangle _originalBounds) {
		this._originalBounds = _originalBounds;
	}

	public Point getPoint() {
		return _point;
	}

	public void setPoint(Point _point) {
		this._point = _point;
		invokeBoundsChangedFunction();
	}

	public Dimension getPanelSize()
	{
		Dimension result = null;
		Component parent = getParent();
		if( parent != null )
			result = parent.getSize();

		return( result );
	}

	public boolean isKeepAspectRatio() {
		return _keepAspectRatio;
	}

	public void setKeepAspectRatio(boolean _keepZoomFactor) {
		this._keepAspectRatio = _keepZoomFactor;
	}

	protected void updateSpinners()
	{
		GeometricRectangle bounds = getBounds();

		int xx = (int) bounds.x;
		int yy = (int) bounds.y;
		int width = (int) bounds.width;
		int height = (int) bounds.height;

		updateSpinnerValue( _xSpinner, xx );
		updateSpinnerValue( _ySpinner, yy );
		updateSpinnerValue( _widthSpinner, width );
		updateSpinnerValue( _heightSpinner, height );
	}

	protected void updateSpinnerValue( JSpinner spinner, int value )
	{
		if( spinner != null )
			spinner.setValue(value);
	}

	public double getViewFactor()
	{
		double result = 1.0d;
		BufferedImage realImage = getRealImage();
		Rectangle imageBoundsInPanel = getBoundsOfImageInsidePanel();

		if( ( realImage != null ) && ( imageBoundsInPanel != null ) )
			result = imageBoundsInPanel.width / ( (double) realImage.getWidth() );

		return( result );
	}

	public Runnable getBoundsChangedFunction() {
		return _boundsChangedFunction;
	}

	public void setBoundsChangedFunction(Runnable _boundsChangedFunction) {
		this._boundsChangedFunction = _boundsChangedFunction;
	}

	public void invokeBoundsChangedFunction()
	{
		Runnable function = getBoundsChangedFunction();
		if( function != null )
			function.run();
	}

	protected int zoom( double value, double factor )
	{
		return( IntegerFunctions.zoomValueRound(value, factor) );
	}

	protected int unzoom( double value, double factor )
	{
		return( zoom(value, 1/factor) );
	}

	public void setPointFromViewPoint( Point viewPoint )
	{
		setPoint( viewToModel( viewPoint ) );
	}

	public boolean isModelPointInsideImage( Point imagePoint )
	{
		return( areBoundsValid( imagePoint.x, imagePoint.y, 0, 0 ) );
	}

	public Point viewToModel( Point viewPoint )
	{
		Rectangle imageBoundsInPanel = getBoundsOfImageInsidePanel();

		Point result = null;
		if( ( imageBoundsInPanel != null ) && ( viewPoint != null ) )
		{
			double viewFactor = getViewFactor();
			result = new Point( viewPoint );

			result.x = unzoom( result.x - imageBoundsInPanel.x, viewFactor );
			result.y = unzoom( result.y - imageBoundsInPanel.y, viewFactor );
		}

		return( result );
	}

	public void setxSpinner(JSpinner _xSpinner) {
		this._xSpinner = _xSpinner;
	}

	public void setySpinner(JSpinner _ySpinner) {
		this._ySpinner = _ySpinner;
	}

	public void setWidthSpinner(JSpinner _widthSpinner) {
		this._widthSpinner = _widthSpinner;
	}

	public void setHeightSpinner(JSpinner _heightSpinner) {
		this._heightSpinner = _heightSpinner;
	}

	public GeometricRectangle getMasterBounds() {
		return _masterBounds;
	}

	public void setMasterBounds(GeometricRectangle _masterBounds) {
		this._masterBounds = _masterBounds;
	}

	protected void changeSpinnerModel(JSpinner spinner, int value,
		int min, int max, int step, boolean setIsValidFunction)
	{
		Predicate<Integer> isValidFunction = null;
		if( setIsValidFunction )
			isValidFunction = val -> isSpinnerValueChangeValid(spinner, val);

		SpinnerModel sm = createSpinnerModel( value, min, max, step, isValidFunction );
		spinner.setModel( sm );
	}

	public void changeSpinnerModels()
	{
		changeSpinnerModels( true );
	}

	protected void changeSpinnerModels(boolean setIsValidFunction)
	{
		int width = getRealImage().getWidth();
		int height = getRealImage().getHeight();
		GeometricRectangle bounds = getBounds();

		changeSpinnerModel( _xSpinner, (int) bounds.x, 0, width, 1, setIsValidFunction );
		changeSpinnerModel( _ySpinner, (int) bounds.y, 0, height, 1, setIsValidFunction );
		changeSpinnerModel( _widthSpinner, (int) bounds.width, 0, width, 1, setIsValidFunction );
		changeSpinnerModel( _heightSpinner, (int) bounds.height, 0, height, 1, setIsValidFunction );
	}

	public void removeSpinnerModels()
	{
		// in order to avoid cyclic reference
		changeSpinnerModels( false );

		_xSpinner = null;
		_ySpinner = null;
		_widthSpinner = null;
		_heightSpinner = null;
	}

	protected boolean isSpinnerValueChangeValid( JSpinner spinner, Integer value )
	{
		boolean result = true;
		if(!_insideSpinnerChange)
		{
			try
			{
				_insideSpinnerChange = true;
				result = false;
				GeometricRectangle bounds = getBounds();

				int xx = (int) bounds.x;
				int yy = (int) bounds.y;
				int width = (int) bounds.width;
				int height = (int) bounds.height;

				if( spinner == _xSpinner )
					xx = value;
				else if( spinner == _ySpinner )
					yy = value;
				else if( spinner == _widthSpinner )
				{
					width = value;
					height = calculateNewHeight( width, height );
				}
				else if( spinner == _heightSpinner )
				{
					height = value;
					width = calculateNewWidth( height, width );
				}

				result = areBoundsValid( xx, yy, width, height );

				if( result )
				{
					setNewValue( _xSpinner, spinner, xx );
					setNewValue( _ySpinner, spinner, yy );
					setNewValue( _widthSpinner, spinner, width );
					setNewValue( _heightSpinner, spinner, height );

					bounds.x = xx;
					bounds.y = yy;
					bounds.width = width;
					bounds.height = height;
					setBounds( bounds );
				}
			}
			finally
			{
				_insideSpinnerChange = false;
			}
		}

		return( result );
	}

	protected boolean areBoundsValid( int xx, int yy, int width, int height )
	{
		int imWidth = getRealImage().getWidth();
		int imHeight = getRealImage().getHeight();

		return( check( xx, imWidth ) && check( xx + width, imWidth ) &&
				check( yy, imHeight ) && check( yy + height, imHeight ) );
	}

	protected boolean check( double value, double max )
	{
		return( ( value >= 0 ) && ( value <= max ) );
	}

	protected double getAspectRatio()
	{
		return( ((double) getMasterBounds().width ) / getMasterBounds().height );
	}

	protected int calculateNewHeight( int width, int height )
	{
		int result = height;
		if( isKeepAspectRatio() )
			result = zoom( width, 1 / getAspectRatio() );

		return( result );
	}

	protected int calculateNewWidth( int height, int width )
	{
		int result = width;
		if( isKeepAspectRatio() )
			result = zoom( height, getAspectRatio() );

		return( result );
	}

	protected void setNewValue( JSpinner target, JSpinner changedSpinner, int value )
	{
		if( target != changedSpinner )
			target.setValue(value);
	}

	protected SpinnerNumberModelWithGenericLimits createSpinnerModel(
		int value, int min, int max, int step, Predicate<Integer> isValueValid )
	{
		return( new SpinnerNumberModelWithGenericLimits( value, min, max, step, isValueValid ) );
	}

	public GeometricRectangle limitByPosition(GeometricRectangle bounds )
	{
		int imWidth = getRealImage().getWidth();
		int imHeight = getRealImage().getHeight();

		GeometricRectangle result = bounds;

		if( ( result != null ) && check( result.width, imWidth ) &&
			check( result.height, imHeight ) )
		{
			result.x = limit( result.x, 0, imWidth - result.width );
			result.y = limit( result.y, 0, imHeight - result.height );
		}

		return( result );
	}

	protected double max( double v1, double v2 )
	{
		return( DoubleFunctions.instance().max(v1, v2) );
	}

	protected GeometricRectangle limitPosition(GeometricRectangle bounds,
                                               boolean fixedXSmallest,
                                               boolean fixedYSmallest )
	{
		double newX = max( 0, bounds.x );
		if( !fixedXSmallest )
			bounds.width += bounds.x - newX;
		bounds.x = newX;

		double newY = max( 0, bounds.y );
		if( !fixedYSmallest )
			bounds.height += bounds.y - newY;
		bounds.y = newY;

		return( bounds );
	}

	protected void updateBounds( Point firstPoint, Point secondPoint )
	{
		setBoundsUpdatingSpinners( limitBounds( firstPoint, secondPoint) );
	}

	protected Point2d createPoint2d( double xx, double yy )
	{
		return( GeometryCommonFactory.instance().createPoint2d(xx, yy) );
	}

	protected Point2d limitPoint( Point notLimitedPoint )
	{
		return( createPoint2d( limit( notLimitedPoint.x, 0, getRealImage().getWidth() ),
								limit( notLimitedPoint.y, 0, getRealImage().getHeight() ) )
			);
	}

	protected GeometricRectangle limitBounds(Point firstPoint, Point secondPoint )
	{
		GeometricRectangle bounds = getBounds().createCopy();

		Point2d limitedFirst = limitPoint( firstPoint );
		Point2d limitedSecond = limitPoint(secondPoint);

		bounds.x = min( limitedFirst.x, limitedSecond.x );
		bounds.y = min( limitedFirst.y, limitedSecond.y );
		boolean fixedXSmallest = (limitedFirst.x <= limitedSecond.x);
		boolean fixedYSmallest = (limitedFirst.y <= limitedSecond.y);

		bounds.width = abs( limitedFirst.x - limitedSecond.x );
		bounds.height = abs( limitedFirst.y - limitedSecond.y );

		bounds = limitBounds( bounds, fixedXSmallest, fixedYSmallest );

		return(bounds);
	}

	protected double min( double v1, double v2 )
	{
		return( DoubleFunctions.instance().min(v1, v2) );
	}

	protected double abs( double val )
	{
		return( DoubleFunctions.instance().abs(val) );
	}

	protected double sgn( double val )
	{
		return( DoubleFunctions.instance().sgn(val) );
	}

	public GeometricRectangle limitBounds(GeometricRectangle bounds,
                                          boolean fixedXSmallest,
                                          boolean fixedYSmallest)
	{
		limitPosition( bounds, fixedXSmallest, fixedYSmallest );
		return( limitBySize( bounds, fixedXSmallest, fixedYSmallest ) );
	}

	public GeometricRectangle limitBySize(GeometricRectangle bounds,
                                          boolean fixedXSmallest,
                                          boolean fixedYSmallest )
	{
		int imWidth = getRealImage().getWidth();
		int imHeight = getRealImage().getHeight();

		GeometricRectangle result = bounds;

		if( ( result != null ) && check( result.x, imWidth ) &&
			check( result.y, imHeight ) )
		{
			keepAspectRatioMax(result, fixedXSmallest, fixedYSmallest);
			limitWidth(result, imWidth, fixedXSmallest, fixedYSmallest);
			limitHeight(result, imHeight, fixedXSmallest, fixedYSmallest);
		}

		return( result );
	}

	protected void keepAspectRatioMax(GeometricRectangle result,
                                      boolean fixedXSmallest,
                                      boolean fixedYSmallest)
	{
		boolean toGreatest = true;
		keepAspectRatio(result, fixedXSmallest, fixedYSmallest, toGreatest);
	}

	protected void keepAspectRatio( GeometricRectangle bounds,
									boolean fixedXSmallest,
									boolean fixedYSmallest,
									boolean toGreatest )
	{
		if( isKeepAspectRatio() )
		{
			double aspectRatio = getAspectRatio();
			double widthForHeight = bounds.height * aspectRatio;
			if( ( ( widthForHeight > bounds.width ) && toGreatest ) ||
				( ( widthForHeight < bounds.width ) && !toGreatest ) )
			{
				double newWidth = zoom( bounds.height, aspectRatio );
				newWidth( bounds, newWidth, !fixedXSmallest );
			}
			else
			{
				double newHeight = unzoom( bounds.width, aspectRatio );
				newHeight( bounds, newHeight, !fixedYSmallest );
			}
		}
	}

	protected void newWidth(GeometricRectangle bounds, double width, boolean modifyPosition )
	{
		if( modifyPosition )
			bounds.x -= width - bounds.width;

		bounds.width = width;
	}

	protected void newHeight(GeometricRectangle bounds, double height, boolean modifyPosition )
	{
		if( modifyPosition )
			bounds.y -= height - bounds.height;

		bounds.height = height;
	}

	protected void limitWidth(GeometricRectangle bounds, int imWidth,
                              boolean fixedXSmallest,
                              boolean fixedYSmallest )
	{
		double newWidth = limit( bounds.width, MorphingConstants.MINIMUM_SIZE_FOR_BOUNDS,
							bounds.width + bounds.x );
		newWidth = limit( newWidth, MorphingConstants.MINIMUM_SIZE_FOR_BOUNDS,
							imWidth - bounds.x );

		newWidth( bounds, newWidth, !fixedXSmallest );

		double newHeight = calculateNewHeight((int) bounds.width, (int) bounds.height);
		newHeight( bounds, newHeight, !fixedYSmallest );
//		keepAspectRatioMin(bounds, fixedXSmallest, fixedYSmallest);
	}

	protected void limitHeight(GeometricRectangle bounds, int imHeight,
                               boolean fixedXSmallest,
                               boolean fixedYSmallest )
	{
		double newHeight = limit( bounds.height, MorphingConstants.MINIMUM_SIZE_FOR_BOUNDS,
								bounds.height + bounds.y );
		newHeight = limit( newHeight, MorphingConstants.MINIMUM_SIZE_FOR_BOUNDS,
								imHeight - bounds.y );

		newHeight( bounds, newHeight, !fixedYSmallest );

		double newWidth = calculateNewWidth((int) bounds.height, (int) bounds.width);
		newWidth( bounds, newWidth, !fixedXSmallest );
//		keepAspectRatioMin(bounds, fixedXSmallest, fixedYSmallest);
	}

	protected double limit( double value, double min, double max )
	{
		return( DoubleFunctions.instance().limit( value, min, max ) );
	}

	@Override
	public void setComponentMapper(ComponentMapper mapper) {
		super.setComponentMapper(mapper);

		_xSpinner = mapper.mapComponent( _xSpinner );
		_ySpinner = mapper.mapComponent( _ySpinner );
		_widthSpinner = mapper.mapComponent( _widthSpinner );
		_heightSpinner = mapper.mapComponent( _heightSpinner );
	}
}
