/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.applications.libmorphing.view.newboundschooser;

import java.awt.*;
import javax.swing.*;
import javax.swing.border.*;
import com.frojasg1.applications.common.components.internationalization.window.InternationalizationInitializationEndCallback;
import com.frojasg1.applications.common.components.internationalization.window.InternationalizedJDialog;
import com.frojasg1.applications.common.components.internationalization.window.exceptions.ValidationException;
import com.frojasg1.applications.common.components.resizecomp.MapResizeRelocateComponentItem;
import com.frojasg1.applications.common.components.resizecomp.ResizeRelocateItem;
import com.frojasg1.applications.common.configuration.application.BaseApplicationConfigurationInterface;
import com.frojasg1.applications.libmorphing.language.LibMorphingLanguageResources;
import com.frojasg1.applications.libmorphing.model.images.image.MorphingImageModel;
import com.frojasg1.libgeometry.types.GeometricRectangle;
import com.frojasg1.general.desktop.view.panels.AcceptCancelRevertControllerInterface;
import com.frojasg1.general.desktop.view.panels.AcceptCancelRevertPanel;
import com.frojasg1.general.desktop.view.panels.InformerInterface;
import com.frojasg1.general.desktop.view.pdf.ImageJPanel;
import com.frojasg1.general.desktop.view.zoom.mapper.ComponentMapper;
import java.awt.Component;
import java.awt.image.BufferedImage;
import java.util.function.Consumer;
import javax.swing.JFormattedTextField;
import javax.swing.JFrame;

/**
 *
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 */
public class NewBoundsChooserJDialog  extends InternationalizedJDialog
	implements AcceptCancelRevertControllerInterface
{
	private final static String a_configurationBaseFileName = "NewBoundsChooserJDialog";

	protected AcceptCancelRevertPanel _acceptPanel = null;

	protected ImageJPanel _imagePanel = null;

	protected MorphingImageModel _imageModel;
	protected GeometricRectangle _masterBounds;

	protected NewBoundsChooserContext _context;

	protected NewBoundsChooserController _panelController;

	public NewBoundsChooserJDialog( JFrame parent, boolean modal,
							BaseApplicationConfigurationInterface appliConf,
							Consumer<InternationalizationInitializationEndCallback> initializationEndCallBack,
							MorphingImageModel imageModel,
							GeometricRectangle masterBounds)
	{
		this( parent, modal, appliConf, initializationEndCallBack, imageModel,
			masterBounds, true );
	}

	/**
	 * Creates new form RegexConfJDialog
	 */
	protected NewBoundsChooserJDialog( JFrame parent, boolean modal,
							BaseApplicationConfigurationInterface appliConf,
							Consumer<InternationalizationInitializationEndCallback> initializationEndCallBack,
							MorphingImageModel imageModel,
							GeometricRectangle masterBounds,
							boolean init)
	{
		super(parent, modal, appliConf, null,
			initializationEndCallBack, true );

		_imageModel = imageModel;
		_masterBounds = masterBounds;

		if( init )
			init();
	}

	protected void init()
	{
		initComponents();

		initOwnComponents();

		setWindowConfiguration();

		addListeners();
	}

	protected BufferedImage getRealImage()
	{
		return( _imageModel.getImage() );
	}

	protected MorphingImageModel getImageModel()
	{
		return( _imageModel );
	}

	protected GeometricRectangle getOriginalBounds()
	{
		return( getImageModel().getBounds() );
	}

	protected ImageJPanel createImageJPanel()
	{
		ImageJPanel result = new ImageJPanel( getRealImage() );
		result.setCanInvertImageColors(false);

		return( result );
	}

	protected AcceptCancelRevertPanel createAcceptCancelRevertPanel()
	{
		return( new AcceptCancelRevertPanel( this ) );
	}

	protected NewBoundsChooserContext createNewBoundsChooserContext()
	{
		return( new NewBoundsChooserContext(this) );
	}

	protected void initOwnComponents()
	{
		_acceptPanel = createAcceptCancelRevertPanel();

		jPanel4.add( _acceptPanel );
		addComponentToCompletelyFillParent( jPanel4, _acceptPanel );

		_imagePanel = createImageJPanel();

		_context = createNewBoundsChooserContext();
//		updateNewBoundsChooserContext();

		BoundaryPainter painter = createPainter();
		_imagePanel.setExternalPainter(painter);

		_panelController = createPanelController();

		addComponentToCompletelyFillParent(jPanel5, _imagePanel);
	}

	protected void updateNewBoundsChooserContext()
	{
		_context.setParent(_imagePanel);
		_context.setBoundsChangedFunction(_imagePanel::repaint);
		_context.setRealImage( getRealImage() );
		_context.setBounds( getOriginalBounds().createCopy() );
		_context.setOriginalBounds( getOriginalBounds() );
		_context.setMasterBounds(_masterBounds);

		_context.setxSpinner(jSp_x);
		_context.setySpinner(jSp_y);
		_context.setWidthSpinner(jSp_width);
		_context.setHeightSpinner(jSp_height);
		_context.changeSpinnerModels();

		if( isDarkMode() )
			invertJFormattedTextFields( jSp_x, jSp_y, jSp_width, jSp_height );
	}

	protected void invertJFormattedTextFields( Component ... spinners )
	{
		for( Component spinner: spinners )
			browseComponentHierarchy(spinner, this::invertJFormattedTextField );
	}

	protected Component invertJFormattedTextField( Component comp )
	{
		if( comp instanceof JFormattedTextField )
			this.getColorInversor().invertSingleComponentColors(comp);

		return( null );
	}

	protected NewBoundsChooserController createPanelController()
	{
		return( new NewBoundsChooserController( _context ) );
	}

	protected BoundaryPainter createPainter()
	{
		return( new BoundaryPainter( getAppliConf(), _context ) );
	}

	protected void setWindowConfiguration( )
	{
		MapResizeRelocateComponentItem mapRRCI = new MapResizeRelocateComponentItem();
		try
		{
			mapRRCI.putResizeRelocateComponentItem( jPanel1, ResizeRelocateItem.FILL_WHOLE_WIDTH );
			mapRRCI.putResizeRelocateComponentItem( jPanel2, ResizeRelocateItem.FILL_WHOLE_WIDTH );
			mapRRCI.putResizeRelocateComponentItem( jPanel5, ResizeRelocateItem.FILL_WHOLE_WIDTH );
			mapRRCI.putResizeRelocateComponentItem( jPanel3, ResizeRelocateItem.FILL_WHOLE_WIDTH );
			mapRRCI.putResizeRelocateComponentItem( jPanel4, ResizeRelocateItem.MOVE_MID_HORIZONTAL_SIDE_PROPORTIONAL );
			mapRRCI.putResizeRelocateComponentItem( _acceptPanel, ResizeRelocateItem.RESIZE_TO_RIGHT + ResizeRelocateItem.RESIZE_TO_BOTTOM );
			mapRRCI.putResizeRelocateComponentItem( _imagePanel, ResizeRelocateItem.FILL_WHOLE_PARENT );
			mapRRCI.putAll( _acceptPanel.getResizeRelocateInfo() );
		}
		catch( Throwable th )
		{
			th.printStackTrace();
		}

		createInternationalization(	getAppliConf().getConfigurationMainFolder(),
									getAppliConf().getApplicationNameFolder(),
									getAppliConf().getApplicationGroup(),
									LibMorphingLanguageResources.sa_PROPERTIES_PATH_IN_JAR,
									getConfigurationBaseFileName(),
									this,
									getParent(),
									null,
									true,
									mapRRCI );

		getInternationalization().setMaxWindowHeightNoLimit(false);
	}

	@Override
	public void internationalizationInitializationEndCallback()
	{
		super.internationalizationInitializationEndCallback();

		updateNewBoundsChooserContext();
		updateImagePanelInContext();

		setMaximumSize( getSize() );

		revert(null);
	}

	public void updateImagePanelInContext()
	{
		_context.setBoundsOfImageInsidePanel( getImagePanel().getPanelImageBounds() );
		_context.setKeepAspectRatio(jCB_keepAspectRatio.isSelected());
	}

	protected void addListeners()
	{
		getImagePanel().addMouseListener(_panelController);
		getImagePanel().addMouseMotionListener(_panelController);
	}

	protected void removeListeners()
	{
		getImagePanel().removeMouseListener(_panelController);
		getImagePanel().removeMouseMotionListener(_panelController);
	}

	/**
	 * This method is called from within the constructor to initialize the form.
	 * WARNING: Do NOT modify this code. The content of this method is always
	 * regenerated by the Form Editor.
	 */
	@SuppressWarnings("unchecked")
    // <editor-fold defaultstate="collapsed" desc="Generated Code">//GEN-BEGIN:initComponents
    // Generated using JFormDesigner non-commercial license
    private void initComponents() {
        jPanel1 = new JPanel();
        jPanel2 = new JPanel();
        jSp_width = new JSpinner();
        jSp_x = new JSpinner();
        jSp_y = new JSpinner();
        jSp_height = new JSpinner();
        jL_width = new JLabel();
        jL_x = new JLabel();
        jL_y = new JLabel();
        jL_height = new JLabel();
        jCB_keepAspectRatio = new JCheckBox();
        jPanel5 = new JPanel();
        jPanel3 = new JPanel();
        jPanel4 = new JPanel();

        //======== this ========
        setDefaultCloseOperation(WindowConstants.DO_NOTHING_ON_CLOSE);
        setName("");
        Container contentPane = getContentPane();
        contentPane.setLayout(null);

        //======== jPanel1 ========
        {
            jPanel1.setBorder(new LineBorder(new Color(102, 102, 255)));
            jPanel1.setLayout(null);

            //======== jPanel2 ========
            {
                jPanel2.setBorder(LineBorder.createBlackLineBorder());
                jPanel2.setLayout(null);

                //---- jSp_width ----
                jSp_width.setMinimumSize(new Dimension(30, 20));
                jSp_width.setPreferredSize(new Dimension(37, 20));
                jPanel2.add(jSp_width);
                jSp_width.setBounds(355, 5, 60, jSp_width.getPreferredSize().height);

                //---- jSp_x ----
                jSp_x.setMinimumSize(new Dimension(30, 20));
                jSp_x.setPreferredSize(new Dimension(37, 20));
                jPanel2.add(jSp_x);
                jSp_x.setBounds(230, 5, 60, jSp_x.getPreferredSize().height);

                //---- jSp_y ----
                jSp_y.setMinimumSize(new Dimension(30, 20));
                jSp_y.setPreferredSize(new Dimension(37, 20));
                jPanel2.add(jSp_y);
                jSp_y.setBounds(230, 30, 60, jSp_y.getPreferredSize().height);

                //---- jSp_height ----
                jSp_height.setMinimumSize(new Dimension(30, 20));
                jSp_height.setPreferredSize(new Dimension(37, 20));
                jPanel2.add(jSp_height);
                jSp_height.setBounds(355, 30, 60, jSp_height.getPreferredSize().height);

                //---- jL_width ----
                jL_width.setHorizontalAlignment(SwingConstants.TRAILING);
                jL_width.setText("width");
                jL_width.setName("jL_width");
                jPanel2.add(jL_width);
                jL_width.setBounds(295, 10, 55, jL_width.getPreferredSize().height);

                //---- jL_x ----
                jL_x.setHorizontalAlignment(SwingConstants.TRAILING);
                jL_x.setText("x");
                jL_x.setName("jL_x");
                jPanel2.add(jL_x);
                jL_x.setBounds(195, 10, 30, jL_x.getPreferredSize().height);

                //---- jL_y ----
                jL_y.setHorizontalAlignment(SwingConstants.TRAILING);
                jL_y.setText("y");
                jL_y.setName("jL_y");
                jPanel2.add(jL_y);
                jL_y.setBounds(195, 35, 30, jL_y.getPreferredSize().height);

                //---- jL_height ----
                jL_height.setHorizontalAlignment(SwingConstants.TRAILING);
                jL_height.setText("height");
                jL_height.setName("jL_height");
                jPanel2.add(jL_height);
                jL_height.setBounds(295, 35, 55, jL_height.getPreferredSize().height);

                //---- jCB_keepAspectRatio ----
                jCB_keepAspectRatio.setSelected(true);
                jCB_keepAspectRatio.setText("Keep aspect ratio");
                jCB_keepAspectRatio.setName("jCB_keepAspectRatio");
                jCB_keepAspectRatio.addActionListener(e -> jCB_keepAspectRatioActionPerformed(e));
                jPanel2.add(jCB_keepAspectRatio);
                jCB_keepAspectRatio.setBounds(5, 20, 200, jCB_keepAspectRatio.getPreferredSize().height);

                {
                    // compute preferred size
                    Dimension preferredSize = new Dimension();
                    for(int i = 0; i < jPanel2.getComponentCount(); i++) {
                        Rectangle bounds = jPanel2.getComponent(i).getBounds();
                        preferredSize.width = Math.max(bounds.x + bounds.width, preferredSize.width);
                        preferredSize.height = Math.max(bounds.y + bounds.height, preferredSize.height);
                    }
                    Insets insets = jPanel2.getInsets();
                    preferredSize.width += insets.right;
                    preferredSize.height += insets.bottom;
                    jPanel2.setMinimumSize(preferredSize);
                    jPanel2.setPreferredSize(preferredSize);
                }
            }
            jPanel1.add(jPanel2);
            jPanel2.setBounds(0, 0, 420, 65);

            //======== jPanel5 ========
            {
                jPanel5.setLayout(null);

                {
                    // compute preferred size
                    Dimension preferredSize = new Dimension();
                    for(int i = 0; i < jPanel5.getComponentCount(); i++) {
                        Rectangle bounds = jPanel5.getComponent(i).getBounds();
                        preferredSize.width = Math.max(bounds.x + bounds.width, preferredSize.width);
                        preferredSize.height = Math.max(bounds.y + bounds.height, preferredSize.height);
                    }
                    Insets insets = jPanel5.getInsets();
                    preferredSize.width += insets.right;
                    preferredSize.height += insets.bottom;
                    jPanel5.setMinimumSize(preferredSize);
                    jPanel5.setPreferredSize(preferredSize);
                }
            }
            jPanel1.add(jPanel5);
            jPanel5.setBounds(0, 65, 420, 355);

            {
                // compute preferred size
                Dimension preferredSize = new Dimension();
                for(int i = 0; i < jPanel1.getComponentCount(); i++) {
                    Rectangle bounds = jPanel1.getComponent(i).getBounds();
                    preferredSize.width = Math.max(bounds.x + bounds.width, preferredSize.width);
                    preferredSize.height = Math.max(bounds.y + bounds.height, preferredSize.height);
                }
                Insets insets = jPanel1.getInsets();
                preferredSize.width += insets.right;
                preferredSize.height += insets.bottom;
                jPanel1.setMinimumSize(preferredSize);
                jPanel1.setPreferredSize(preferredSize);
            }
        }
        contentPane.add(jPanel1);
        jPanel1.setBounds(0, 0, 420, 420);

        //======== jPanel3 ========
        {
            jPanel3.setLayout(null);

            //======== jPanel4 ========
            {
                jPanel4.setBorder(new LineBorder(new Color(255, 204, 204)));
                jPanel4.setMinimumSize(new Dimension(130, 50));
                jPanel4.setPreferredSize(new Dimension(130, 50));
                jPanel4.setLayout(null);

                {
                    // compute preferred size
                    Dimension preferredSize = new Dimension();
                    for(int i = 0; i < jPanel4.getComponentCount(); i++) {
                        Rectangle bounds = jPanel4.getComponent(i).getBounds();
                        preferredSize.width = Math.max(bounds.x + bounds.width, preferredSize.width);
                        preferredSize.height = Math.max(bounds.y + bounds.height, preferredSize.height);
                    }
                    Insets insets = jPanel4.getInsets();
                    preferredSize.width += insets.right;
                    preferredSize.height += insets.bottom;
                    jPanel4.setMinimumSize(preferredSize);
                    jPanel4.setPreferredSize(preferredSize);
                }
            }
            jPanel3.add(jPanel4);
            jPanel4.setBounds(145, 10, 130, 50);

            {
                // compute preferred size
                Dimension preferredSize = new Dimension();
                for(int i = 0; i < jPanel3.getComponentCount(); i++) {
                    Rectangle bounds = jPanel3.getComponent(i).getBounds();
                    preferredSize.width = Math.max(bounds.x + bounds.width, preferredSize.width);
                    preferredSize.height = Math.max(bounds.y + bounds.height, preferredSize.height);
                }
                Insets insets = jPanel3.getInsets();
                preferredSize.width += insets.right;
                preferredSize.height += insets.bottom;
                jPanel3.setMinimumSize(preferredSize);
                jPanel3.setPreferredSize(preferredSize);
            }
        }
        contentPane.add(jPanel3);
        jPanel3.setBounds(0, 420, 420, 70);

        {
            // compute preferred size
            Dimension preferredSize = new Dimension();
            for(int i = 0; i < contentPane.getComponentCount(); i++) {
                Rectangle bounds = contentPane.getComponent(i).getBounds();
                preferredSize.width = Math.max(bounds.x + bounds.width, preferredSize.width);
                preferredSize.height = Math.max(bounds.y + bounds.height, preferredSize.height);
            }
            Insets insets = contentPane.getInsets();
            preferredSize.width += insets.right;
            preferredSize.height += insets.bottom;
            contentPane.setMinimumSize(preferredSize);
            contentPane.setPreferredSize(preferredSize);
        }
        pack();
        setLocationRelativeTo(getOwner());
    }// </editor-fold>//GEN-END:initComponents

    private void jCB_keepAspectRatioActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_jCB_keepAspectRatioActionPerformed
        // TODO add your handling code here:

		getContext().setKeepAspectRatio( jCB_keepAspectRatio.isSelected() );
    }//GEN-LAST:event_jCB_keepAspectRatioActionPerformed

    // Variables declaration - do not modify//GEN-BEGIN:variables
    // Generated using JFormDesigner non-commercial license
    private JPanel jPanel1;
    private JPanel jPanel2;
    private JSpinner jSp_width;
    private JSpinner jSp_x;
    private JSpinner jSp_y;
    private JSpinner jSp_height;
    private JLabel jL_width;
    private JLabel jL_x;
    private JLabel jL_y;
    private JLabel jL_height;
    private JCheckBox jCB_keepAspectRatio;
    private JPanel jPanel5;
    private JPanel jPanel3;
    private JPanel jPanel4;
    // End of variables declaration//GEN-END:variables

	@Override
	protected void translateMappedComponents(ComponentMapper compMapper)
	{
		jCB_keepAspectRatio = compMapper.mapComponent( jCB_keepAspectRatio );
		jL_height = compMapper.mapComponent( jL_height );
		jL_width = compMapper.mapComponent( jL_width );
		jL_x = compMapper.mapComponent( jL_x );
		jL_y = compMapper.mapComponent( jL_y );
		jPanel1 = compMapper.mapComponent( jPanel1 );
		jPanel2 = compMapper.mapComponent( jPanel2 );
		jPanel3 = compMapper.mapComponent( jPanel3 );
		jPanel4 = compMapper.mapComponent( jPanel4 );
		jPanel5 = compMapper.mapComponent( jPanel5 );
		jSp_height = compMapper.mapComponent( jSp_height );
		jSp_width = compMapper.mapComponent( jSp_width );
		jSp_x = compMapper.mapComponent( jSp_x );
		jSp_y = compMapper.mapComponent( jSp_y );

		_acceptPanel = compMapper.mapComponent( _acceptPanel );
	}

	@Override
	public void revert(InformerInterface panel)
	{
		getContext().setBoundsUpdatingSpinners( getOriginalBounds().createCopy() );
	}

	public ImageJPanel getImagePanel() {
		return _imagePanel;
	}

	@Override
	protected void validateFormChild() throws ValidationException
	{
	}

	@Override
	public void accept(InformerInterface panel)
	{
		validateForm();
		if( wasSuccessful() )
		{
			formWindowClosing(true);
		}
	}

	@Override
	public void cancel(InformerInterface panel)
	{
		setWasSuccessful( false );

		formWindowClosing(true);
	}

	protected String getConfigurationBaseFileName()
	{
		return( a_configurationBaseFileName );
	}

	@Override
	public void releaseResources()
	{
		super.releaseResources();

		getContext().removeSpinnerModels();
		removeListeners();
	}

	public NewBoundsChooserContext getContext() {
		return _context;
	}

	public GeometricRectangle getNewBounds()
	{
		return( getContext().getBounds() );
	}
}
