/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.applications.libmorphing.view.painters;

import com.frojasg1.applications.common.configuration.InternationalizedStringConf;
import com.frojasg1.applications.libmorphing.api.facedetector.helpers.Control88meshCreator;
import com.frojasg1.applications.libmorphing.api.facedetector.pojo.FaceDetection;
import com.frojasg1.applications.libmorphing.api.facedetector.pojo.FaceDetectionMeshData;
import com.frojasg1.applications.libmorphing.controller.ControlMeshManager;
import com.frojasg1.applications.libmorphing.controller.MorphingVerticesMapManager;
import com.frojasg1.applications.libmorphing.controller.MorphingVerticesMapManagerBase;
import com.frojasg1.applications.libmorphing.helpers.MorphingHelper;
import com.frojasg1.applications.libmorphing.model.images.image.MorphingImageModel;
import com.frojasg1.applications.libmorphing.model.images.vertex.VertexModel;
import com.frojasg1.applications.libmorphing.mvc.DelaunayTriangulatorView;
import com.frojasg1.applications.libmorphing.view.panels.evt.enums.RecalculateTriangulationType;
import com.frojasg1.applications.libmorphing.view.panels.evt.enums.VertexEventType;
import com.frojasg1.applications.libmorphing.view.panels.evt.*;
import com.frojasg1.delaunay.triangulation.quadedge.builder.QuadEdgeDelaunayBuilder;
import com.frojasg1.libgeometry.types.GeometricRectangle;
import com.frojasg1.libgeometry.types.Point2d;
import com.frojasg1.general.NullFunctions;
import com.frojasg1.general.desktop.mouse.MouseFunctions;
import com.frojasg1.general.desktop.view.icons.IconResources;
import com.frojasg1.general.desktop.view.menus.BaseJPopupMenu;
import com.frojasg1.general.number.IntegerFunctions;
import com.frojasg1.general.view.evt.wrapper.ListOfGenComponentWrapperEventContextListenersBase;
import com.frojasg1.general.view.evt.wrapper.impl.GenComponentWrapperEvent;
import com.frojasg1.generic.GenericFunctions;

import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.image.BufferedImage;
import java.util.List;
import java.util.Objects;
import java.util.function.Consumer;
import java.util.function.Function;

/**
 *
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 */
public abstract class DelaunayTriangulationJPanelControllerBase
	extends MouseAdapter
{
	public static final String CONF_DELAUNAY_TRIANGULATION = TriangulationPainterLayerTypeEnum.DELAUNAY_TRIANGULATION.name();
	public static final String CONF_FACE_DETECTION_SELECTION = TriangulationPainterLayerTypeEnum.FACE_DETECTION_SELECTION.name();
	public static final String CONF_FACE_DETECTION_VERTEX_MODIFICATION = TriangulationPainterLayerTypeEnum.FACE_DETECTION_VERTEX_MODIFICATION.name();
	public static final String CONF_RUN_FACE_DETECTION = "RUN_FACE_DETECTION";
	public static final String CONF_REAUTOCALCULATE_TRIANGULATION = "REAUTOCALCULATE_TRIANGULATION";
	public static final String CONF_REAUTOCALCULATE_ALL_TRIANGULATIONS = "REAUTOCALCULATE_ALL_TRIANGULATIONS";


	protected MorphingVerticesMapManager morphingVerticesManager;
	protected String imageName;

	protected Component _parent;

	protected QuadEdgeDelaunayBuilder _builder;

	protected DelaunayJPanelsEventBuilderBase _evtFactory;
	protected ListOfGenComponentWrapperEventContextListenersBase<JPanel, GenComponentWrapperEvent> _listenersList;

	protected double _imageZoomFactor;
//	protected MorphingImageModel _imageModel;

	protected Point2d _closestVertex;
	protected boolean _isMovingPoint;

	protected boolean isActive;

	protected int proximity100percentRadius;

	protected InternationalizedStringConf internationalStrings;

	protected BaseJPopupMenu myPopupMenu;

	protected TriangulationPainterLayerTypeEnum layerType;

	protected boolean rightClickPressed = false;

	public DelaunayTriangulationJPanelControllerBase(Component comp, TriangulationPainterLayerTypeEnum layerType,
													 String imageName, int proximity100percentRadius,
													 DelaunayJPanelsEventBuilderBase evtFactory,
                                                     ListOfGenComponentWrapperEventContextListenersBase<JPanel, GenComponentWrapperEvent> listenersList,
                                                     QuadEdgeDelaunayBuilder builder, InternationalizedStringConf internationalStrings )
	{
		_evtFactory = evtFactory;
		this.layerType = layerType;
		_listenersList = listenersList;
		_parent = comp;
		_builder = builder;
		this.imageName = imageName;
		this.proximity100percentRadius = proximity100percentRadius;
		this.internationalStrings = internationalStrings;
	}

	public DelaunayTriangulationJPanelControllerBase init() {
		myPopupMenu = createBasePopupMenu();

		return this;
	}

	public TriangulationPainterLayerTypeEnum getLayerType() {
		return layerType;
	}

	protected void doPopup(MouseEvent evt )
	{
		BaseJPopupMenu popupMenu = getPopupMenu();
		if( popupMenu != null )
			popupMenu.doPopup(evt);
	}

	protected abstract BaseJPopupMenu createBasePopupMenu();

	public BaseJPopupMenu getPopupMenu() {
		return myPopupMenu;
	}

	protected InternationalizedStringConf getInternationalStrings() {
		return internationalStrings;
	}

	public boolean isActive() {
		return isActive;
	}

	public void setActive(boolean active) {
		isActive = active;
	}

	protected MorphingVerticesMapManager getMorphingVerticesMapManager() {
		return morphingVerticesManager;
	}

	public ControlMeshManager getControlMeshManager() {
		return getIfNotNull(getMorphingVerticesMapManager(), MorphingVerticesMapManager::getControlMeshManager);
	}

	protected int getNumImages() {
		return getIfNotNull(getControlMeshManager(), ControlMeshManager::getNumImages, 0);
	}

	protected void setMorphingVerticesMapManager(MorphingVerticesMapManagerBase morphingVerticesMapManager) {
		this.morphingVerticesManager = morphingVerticesMapManager;
	}

	public Component getParent() {
		return _parent;
	}

	public void setParent(Component _parent) {
		this._parent = _parent;
	}

	protected double getApplicationZoomFactor()
	{
		return( GenericFunctions.instance().getAppliConf().getZoomFactor() );
	}

	protected int getProximity100percentRadius() {
		return proximity100percentRadius;
	}

	protected abstract void browseVertices(Consumer<VertexModel> vertexProcessor);


	protected Point2d getClosest( Point2d imagePoint )
	{
		Point2d result = null;
		
		if( imagePoint != null )
		{
			NearestVertexContext context = new NearestVertexContext(imagePoint);
			browseVertices(context::process);

			result = context.getResult();
		}

		return( result );
	}

	protected abstract boolean hasToDiscardForClosest(Point2d pt, Point2d imagePt);

	protected GeometricRectangle getBounds()
	{
		GeometricRectangle result = null;
		MorphingImageModel imageModel = getImageModel();
		if( imageModel != null )
			result = imageModel.getBounds();
		
		return( result );
	}

	protected GeometricRectangle getLimitingBounds()
	{
		return getBounds();
	}

	protected boolean isCorner( Point2d pt )
	{
		boolean result = false;

		if( pt != null )
		{
			GeometricRectangle bounds = getImageModel().getBounds();
			if( bounds != null )
				result = bounds.isCorner(pt);
		}

		return( result );
	}

	protected Point2d updateClosest( Point2d imagePoint )
	{
		return( updateClosestInternal( getClosest(imagePoint) ) );
	}

	protected void resetClosest()
	{
		updateClosestInternal( (Point2d) null );
	}

	protected Point2d updateClosestInternal( Point2d closest )
	{
		if( !Objects.equals( closest, _closestVertex ) )
			notifyClosest(closest);

		return( closest );
	}

	protected boolean isCtrlPressed(MouseEvent e)
	{
		return( MouseFunctions.isCtrlPressed(e) );
	}

	protected boolean isShiftPressed(MouseEvent e)
	{
		return( MouseFunctions.isShiftPressed(e) );
	}

	protected Point getImagePoint( MouseEvent evt )
	{
		return( getImagePoint( evt.getLocationOnScreen() ) );
	}

	protected Point getImagePointOfPointOnScreen()
	{
		return( getImagePoint( MouseFunctions.getMouseLocation()) );
	}

	protected Point getImagePoint( Point ptOnScreen )
	{
		Point result = null;
		if( ptOnScreen != null )
		{
			Point imageLocationOnScreen = getParent().getLocationOnScreen();
			result = new Point( ptOnScreen.x - imageLocationOnScreen.x,
						ptOnScreen.y - imageLocationOnScreen.y );
		}
		return( result );
	}

	public MorphingHelper getMorphingHelper() {
		return MorphingHelper.instance();
	}

	protected Point2d toNormalized( Point2d pt, GeometricRectangle rect )
	{
		return( getMorphingHelper().toNormalized(pt, rect) );
	}

	protected Point2d toNormalized( Point2d imagePoint )
	{
		return( toNormalized( imagePoint, getBounds() ) );
	}

	protected boolean isMaster()
	{
		boolean result = false;
		MorphingImageModel imageModel = getImageModel();
		if( imageModel != null )
			result = imageModel.isMaster();

		return( result );
	}

	protected BufferedImage getImage()
	{
		return( getImageModel().getImage() );
	}


	protected int zoom( double value, double factor )
	{
		return( IntegerFunctions.zoomValueRound(value, factor) );
	}

	protected void startMoving( Point2d imagePoint )
	{
//		Point2d imagePoint = createPoint2d( pt );
		if( imagePoint != null )
		{
			Point2d closest = updateClosest( imagePoint );
			if( closest != null )
			{
//			if( isMaster() )
//			{
//				notifyVertexDeletion(closest);
//				notifyNewPoint(null, pt, false, VertexEventType.INSERT);
//			}
//			else
				notifyNewMovingPoint(closest, imagePoint);

				_isMovingPoint = true;
			}
		}
	}

	protected DelaunayJPanelsEventBuilderBase getEvtFactory()
	{
		return( _evtFactory );
	}

	protected void notifyEvt( GenComponentWrapperEvent evt )
	{
		getListenersList().notifyEvt(evt);
	}

	protected Point2d createPoint2dInternal( double x, double y )
	{
		return( unzoom( getBuilder().createPoint( x, y ) ) );
	}

	protected int zoom( double value )
	{
		return( zoom( value, getImageZoomFactor() ) );
	}

	protected double zoomApplication( double value )
	{
		return( value * getApplicationZoomFactor() );
	}

	protected int unzoom( double value )
	{
		return( zoom(value, 1/getImageZoomFactor()) );
	}

	protected Point2d unzoom( Point2d pt )
	{
		return( MorphingHelper.instance().unzoom(pt, getImageZoomFactor() ) );
	}

	protected boolean isInside( Point2d pt )
	{
		boolean result = false;
		GeometricRectangle bounds = getLimitingBounds();
		if( bounds != null )
			result = bounds.isInside(pt);

		return( result );
	}

	protected Point2d createPoint2d( Point pt )
	{
		Point2d result = null;
		if( pt != null )
		{
			result = createPoint2dInternal( pt.x, pt.y );
			if( !isInside(result) )
				result = null;
		}

		return( result );
	}

	protected DelaunayTriangulatorView getView()
	{
		return( (DelaunayTriangulatorView) getParent() );
	}

	protected VertexEvent createVertexEvent( Point2d originalVertex, Point2d vertex,
											boolean confirmed, VertexEventType type )
	{
		return( getEvtFactory().createVertexEvent(getImageName(), originalVertex,
												vertex, confirmed, type) );
	}

	protected RunFaceDetectionEvent createRunFaceDetectionEvent() {
		return (getEvtFactory().createRunFaceDetectionEvent(getImageName()));
	}

	protected LayerSelectionEvent createLayerSelectionEvent(TriangulationPainterLayerTypeEnum layerType)
	{
		return( getEvtFactory().createLayerSelectionEvent(getImageName(), layerType) );
	}

	protected void notifyNewSelectedLayer(TriangulationPainterLayerTypeEnum layerType)
	{
		notifyEvt(createLayerSelectionEvent(layerType));
	}

	protected void notifyRunFaceDetection()
	{
		notifyEvt(createRunFaceDetectionEvent());
	}

	protected abstract void notifyClosest( Point2d pt );

	protected abstract void notifyNewMovingPoint( Point2d originalPoint, Point2d newPoint );

	protected abstract void endMovingPoint();

	protected void notifyNewPoint( Point2d imagePoint, boolean confirmed, VertexEventType type )
	{
		notifyNewPoint( null, imagePoint, confirmed, type );
	}

////	protected void notifyNewPoint( Point2d originalPoint, Point pt,
////									boolean confirmed, VertexEventType type )
////	{
////		notifyNewPoint( originalPoint, createPoint2d( pt ), confirmed, type );
////	}

	protected void notifyNewPoint( Point2d originalVertex, Point2d pt,
									boolean confirmed, VertexEventType type )
	{
		notifyEvt( createVertexEvent( originalVertex, pt, confirmed, type ) );
	}

	protected QuadEdgeDelaunayBuilder getBuilder()
	{
		return( _builder );
	}

	public double getImageZoomFactor() {
		return _imageZoomFactor;
	}

	public void setImageZoomFactor(double _imageZoomFactor) {
		this._imageZoomFactor = _imageZoomFactor;
	}

	public String getImageName() {
		return imageName;
	}

	public void setImageName(String imageName) {
		this.imageName = imageName;
	}

	protected NullFunctions getNullFunctions() {
		return NullFunctions.instance();
	}

	protected <CC, RR> RR getIfNotNull(CC obj, Function<CC,RR> getter )
	{
		return( getNullFunctions().getIfNotNull(obj, getter) );
	}

	public <CC, RR> RR getIfNotNull(CC obj, Function<CC, RR> getter, RR defaultValue) {
		return getNullFunctions().getIfNotNull(obj, getter, defaultValue);
	}

	protected MorphingImageModel getImageModel(MorphingVerticesMapManager morphingVerticesManager,
											   String imageName )
	{
		return getIfNotNull(morphingVerticesManager, manager -> manager.getImageModel(imageName));
	}

	public MorphingImageModel getImageModel()
	{
		return getImageModel(getMorphingVerticesMapManager(), getImageName());
	}

	protected void updateImageModel(MorphingImageModel imageModel ) {
		MorphingImageModel prevImageModel = getImageModel();
		if( prevImageModel != imageModel )
		{
			resetMovingPoint();
//			_imageModel = imageModel;
		}
	}

	protected void resetMovingPoint()
	{
//		MorphingImageModel prevImageModel = _imageModel;
//		if( _isMovingPoint && ( prevImageModel != null ) && prevImageModel.isMaster() )
//			notifyOnTheFlyVertexDeletion();

		resetClosest();
		if( _isMovingPoint )
			endMovingPoint();
		
		_isMovingPoint = false;
	}

	public ListOfGenComponentWrapperEventContextListenersBase<JPanel, GenComponentWrapperEvent> getListenersList() {
		return _listenersList;
	}

	@Override
	 public void mouseClicked(MouseEvent e) {
		if(isActive()) {
			mouseClickedInternal(e);
		}
	}

	/**
	 * {@inheritDoc}
	 */
	 public void mousePressed(MouseEvent e) {
		if(isActive()) {
			mousePressedInternal(e);
		}
	}

	/**
	 * {@inheritDoc}
	 */
	 public void mouseReleased(MouseEvent e) {
		if(isActive()) {
			mouseReleasedInternal(e);
		}
	}

	/**
	 * {@inheritDoc}
	 */
	 public void mouseEntered(MouseEvent e) {
		if(isActive()) {
			mouseEnteredInternal(e);
		}
	}

	/**
	 * {@inheritDoc}
	 */
	 public void mouseExited(MouseEvent e) {
		if(isActive()) {
			mouseExitedInternal(e);
		}
	}

	/**
	 * {@inheritDoc}
	 * @since 1.6
	 */
	 public void mouseWheelMoved(MouseWheelEvent e) {
		if(isActive()) {
			mouseWheelMovedInternal(e);
		}
	}

	/**
	 * {@inheritDoc}
	 * @since 1.6
	 */
	 public void mouseDragged(MouseEvent e) {
		if(isActive()) {
			mouseDraggedInternal(e);
		}
	}

	/**
	 * {@inheritDoc}
	 * @since 1.6
	 */
	 public void mouseMoved(MouseEvent e) {
		if(isActive()) {
			mouseMovedInternal(e);
		}
	}

	protected void mouseClickedInternal(MouseEvent e) {}

	/**
	 * {@inheritDoc}
	 */
	protected void mousePressedInternal(MouseEvent e) {}

	/**
	 * {@inheritDoc}
	 */
	protected void mouseReleasedInternal(MouseEvent e) {}

	/**
	 * {@inheritDoc}
	 */
	protected void mouseEnteredInternal(MouseEvent e) {}

	/**
	 * {@inheritDoc}
	 */
	protected void mouseExitedInternal(MouseEvent e) {}

	/**
	 * {@inheritDoc}
	 * @since 1.6
	 */
	protected void mouseWheelMovedInternal(MouseWheelEvent e){}

	/**
	 * {@inheritDoc}
	 * @since 1.6
	 */
	protected void mouseDraggedInternal(MouseEvent e){}

	/**
	 * {@inheritDoc}
	 * @since 1.6
	 */
	protected void mouseMovedInternal(MouseEvent e){}

	public void setTriangulation(String imageName,
								 MorphingVerticesMapManager morphingVerticesManager) {
		updateImageModel(getImageModel(morphingVerticesManager, imageName));

		setImageName(imageName);
		this.morphingVerticesManager = morphingVerticesManager;
	}

	protected void repaint() {
		_parent.repaint();
	}

	protected FaceDetection getFaceDetection() {
		return getIfNotNull(getImageModel(), MorphingImageModel::getFaceDetection);
	}

	protected boolean isFaceSelected() {
		return getFaceDetection() != null;
	}

	protected List<FaceDetectionMeshData> getAllFaces() {
		return getIfNotNull(getMorphingVerticesMapManager(), manager -> manager.getAllFaces(getImageName()));
	}

	protected boolean faceDetectionApiHasBeenRun() {
		return getAllFaces() != null;
	}

	protected RecalculateTriangulationEvent createRecalculateTriangulationEvent(RecalculateTriangulationType type) {
		return (getEvtFactory().createRecalculateTriangulationEvent(type, getImageName()));
	}

	protected void notifyRecalculateTriangulationEvent(RecalculateTriangulationType type)
	{
		notifyEvt(createRecalculateTriangulationEvent(type));
	}

	protected void notifyReautocalculateTriangulation() {
		notifyRecalculateTriangulationEvent(RecalculateTriangulationType.CURRENT);
	}

	protected void notifyReautocalculateAllTriangulations() {
		notifyRecalculateTriangulationEvent(RecalculateTriangulationType.ALL);
	}

	protected Control88meshCreator getControl88meshCreator() {
		return Control88meshCreator.instance();
	}


	protected class NearestVertexContext {
		protected int xx;
		protected int yy;
		protected double squaredRadius;
		protected double min;
		protected Point2d imagePoint;
		protected Point2d result;
		double sqDist;
		double dx;
		double dy;
		Point2d pt;

		public NearestVertexContext(Point2d imagePoint) {
			this.imagePoint = imagePoint;
			xx = (int) imagePoint.x;
			yy = (int) imagePoint.y;
			squaredRadius = zoomApplication( unzoom( getProximity100percentRadius() ) );
			squaredRadius = squaredRadius * squaredRadius;
			min = squaredRadius + 1;
		}

		public void process(VertexModel vm) {
			pt = vm.getPoint();

			dx = xx - pt.x;
			dy = yy - pt.y;
			sqDist = dx * dx + dy * dy;
			if(min > sqDist)
			{
				if( ! hasToDiscardForClosest(pt, imagePoint) )
				{
					min = sqDist;
					result = pt;
				}
			}
		}

		public Point2d getResult() {
			return result;
		}
	}


	protected class MorphingPanelControllerPopupMenuBase extends BaseJPopupMenu<String>
	{
		public MorphingPanelControllerPopupMenuBase()
		{
			super(_parent, String.class );
		}

		@Override
		public void init()
		{
			super.init();
		}

		protected void createCommonPart()
		{
			addLayerOption( CONF_FACE_DETECTION_SELECTION, null);
			addLayerOption( CONF_FACE_DETECTION_VERTEX_MODIFICATION, null);
			addLayerOption( CONF_DELAUNAY_TRIANGULATION, IconResources.TRIANGLES_ICON_RESOURCE);

			add( new Separator() );

			addMenuComponent( CONF_RUN_FACE_DETECTION, new JMenuItem(),
					() -> getInternationalStrings().getInternationalString(CONF_RUN_FACE_DETECTION),
					() -> notifyRunFaceDetection(),
					IconResources.API_CALL_ICON_RESOURCE);

			add( new Separator() );

			addMenuComponent( CONF_REAUTOCALCULATE_TRIANGULATION, new JMenuItem(),
					() -> getInternationalStrings().getInternationalString(CONF_REAUTOCALCULATE_TRIANGULATION),
					() -> notifyReautocalculateTriangulation(),
					null);

			addMenuComponent( CONF_REAUTOCALCULATE_ALL_TRIANGULATIONS, new JMenuItem(),
					() -> getInternationalStrings().getInternationalString(CONF_REAUTOCALCULATE_ALL_TRIANGULATIONS),
					() -> notifyReautocalculateAllTriangulations(),
					null);
		}

		protected void addLayerOption(String label, String pathToIconResource) {
			addMenuComponent( label, new JMenuItem(),
					() -> getInternationalStrings().getInternationalString(label),
					() -> notifyNewSelectedLayer(TriangulationPainterLayerTypeEnum.valueOf(label)),
					pathToIconResource);
		}

		protected void prepareCommon() {
			enableComponents(true, CONF_FACE_DETECTION_SELECTION);
			enableComponents(isFaceSelected(), CONF_FACE_DETECTION_VERTEX_MODIFICATION,
					CONF_DELAUNAY_TRIANGULATION);
			enableComponents(false, getLayerType().name());
			enableComponents(!faceDetectionApiHasBeenRun(), CONF_RUN_FACE_DETECTION);

			enableComponents(!isMaster(), CONF_REAUTOCALCULATE_TRIANGULATION);
			enableComponents(hasToEnableReautocalculateAllTriangulations(), CONF_REAUTOCALCULATE_ALL_TRIANGULATIONS);
		}

		protected boolean hasToEnableReautocalculateAllTriangulations() {
			return (getNumImages() > 1);
		}

		@Override
		protected void preparePopupMenuItems() {
			prepareCommon();
			setCurrentLanguageTexts();
		}
	}
}
