/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.applications.libmorphing.view.painters;

import com.frojasg1.applications.libmorphing.contexts.MorphingEditorApplicationContext;
import com.frojasg1.applications.libmorphing.api.facedetector.helpers.Control88meshCreator;
import com.frojasg1.applications.libmorphing.configuration.MorphingApplicationConfiguration;
import com.frojasg1.applications.libmorphing.controller.ControlMeshManager;
import com.frojasg1.applications.libmorphing.controller.MorphingVerticesMapManager;
import com.frojasg1.applications.libmorphing.helpers.MorphingHelper;
import com.frojasg1.applications.libmorphing.model.images.image.MorphingImageModel;
import com.frojasg1.applications.libmorphing.model.images.vertex.map.impl.normal.MorphingVerticesMap;
import com.frojasg1.delaunay.triangulation.quadedge.builder.QuadEdgeDelaunayBuilder;
import com.frojasg1.libgeometry.types.GeometricRectangle;
import com.frojasg1.libgeometry.types.Point2d;
import com.frojasg1.libgeometry.types.factory.GeometryCommonFactory;
import com.frojasg1.general.ExecutionFunctions;
import com.frojasg1.general.NullFunctions;
import com.frojasg1.general.desktop.image.ImageFunctions;
import com.frojasg1.general.desktop.view.ComponentFunctions;
import com.frojasg1.general.desktop.view.FrameworkComponentFunctions;
import com.frojasg1.general.desktop.view.ViewFunctions;
import com.frojasg1.general.desktop.view.color.ColorInversor;
import com.frojasg1.general.desktop.view.painter.ExternalPainterBase;
import com.frojasg1.general.desktop.view.pdf.BackgroundImageProcessing;
import com.frojasg1.general.number.IntegerFunctions;
import com.frojasg1.general.view.ReleaseResourcesable;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.swing.*;
import java.awt.*;
import java.awt.image.BufferedImage;
import java.lang.ref.WeakReference;
import java.util.Objects;
import java.util.concurrent.atomic.AtomicReference;
import java.util.function.BiConsumer;
import java.util.function.Consumer;
import java.util.function.Function;

/**
 *
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 */
public abstract class DelaunayTriangulationPainterBase extends ExternalPainterBase
								implements ReleaseResourcesable {
//	protected static final int TEXT_SIZE_100_PERCENT = 16;
	private static final Logger LOGGER = LoggerFactory.getLogger(DelaunayTriangulationPainterBase.class);

	protected static final double SQRT_2 = Math.sqrt(2d);

	protected AtomicReference<Font> _font = new AtomicReference<>();

	protected int textSize100Percent;
	protected double edgeWidth100percent;
	protected double vertexRadio100percent;
	protected Color vertexColor;
	protected Color VertexNumberColor;
	protected Color closestMovableVertexColor;
	protected Color edgeColor;

	protected String _currentImageName;
	protected Point2d _pointInProgress;
	protected Point2d _closestPoint;

//	protected MorphingApplicationConfiguration _appliConf;
	protected QuadEdgeDelaunayBuilder _delaunayBuilder;

//	protected MorphingVerticesMapManager _morphingVerticesManager;

	// level of zoomToInt of the image
	protected double _imageZoomFactor = 1.0d;

	protected WeakReference<Component> _parentRef;

	protected MorphingEditorApplicationContext appliCtx;

	/**
	 * Creates new form GraphicalEqualizerJPanel
	 */
	public DelaunayTriangulationPainterBase(MorphingEditorApplicationContext appliCtx,
                                            Component parent,
                                            QuadEdgeDelaunayBuilder delaunayBuilder, int textSize100Percent,
											double edgeWidth100percent, double vertexRadio100percent,
											Color vertexColor, Color vertexNumberColor, Color closestMovableVertexColor,
											Color edgeColor) {
		this.appliCtx = appliCtx;
		_delaunayBuilder = delaunayBuilder;
		_parentRef = new WeakReference<>(parent);

		this.textSize100Percent = textSize100Percent;
		this.edgeWidth100percent = edgeWidth100percent;
		this.vertexRadio100percent = vertexRadio100percent;
		this.vertexColor = vertexColor;
		VertexNumberColor = vertexNumberColor;
		this.closestMovableVertexColor = closestMovableVertexColor;
		this.edgeColor = edgeColor;
	}

	protected double getEdgeWidth100percent() {
		return edgeWidth100percent;
	}

	protected double getVertexRadio100percent() {
		return vertexRadio100percent;
	}

	protected Color getVertexColor() {
		return vertexColor;
	}

	protected Color getVertexNumberColor() {
		return VertexNumberColor;
	}

	protected Color getClosestMovableVertexColor() {
		return closestMovableVertexColor;
	}

	protected BackgroundImageProcessing createBackgroundImageProcessing()
	{
		BackgroundImageProcessing result = new BackgroundImageProcessing();
		result.setExecuteBackgroundFunction( this::addPendingExecutor );

		return( result );
	}

	protected Control88meshCreator getControl88meshCreator() {
		return Control88meshCreator.instance();
	}

	public MorphingVerticesMapManager getMorphingVerticesManager() {
		return getAppliCtx().getVerticesManager();
	}

	public ControlMeshManager getControlMeshManager() {
		return getIfNotNull(getMorphingVerticesManager(), MorphingVerticesMapManager::getControlMeshManager);
	}

	protected NullFunctions getNullFunctions() {
		return NullFunctions.instance();
	}

	protected <CC, RR> RR getIfNotNull(CC obj, Function<CC, RR> getter, RR defaultValue) {
		return getNullFunctions().getIfNotNull(obj, getter, defaultValue);
	}

	protected <CC, RR> RR getIfNotNull(CC obj, Function<CC,RR> getter )
	{
		return( getNullFunctions().getIfNotNull(obj, getter) );
	}

	protected QuadEdgeDelaunayBuilder getBuilder()
	{
		return( _delaunayBuilder );
	}

	protected Point2d getPointInProgress() {
		return _pointInProgress;
	}

	@Override
	public void paintComponent(Component comp, Graphics grp)
	{
		super.paintComponent( comp, grp );
	}

	protected abstract boolean hasToPaint();

	@Override
	protected void paintComponentInternal(Component comp, Graphics grp)
	{
////		blackScreen( comp, grp );

		if(hasToPaint()) {
			prePaint(grp);

			GeometricRectangle hundredPercentPaintLimits = getHundredPercentPaintLimits(grp);

			Graphics2D g2d = (Graphics2D) grp.create();
			Stroke solid = createSolidStroke();

			layerPrePaint(g2d, hundredPercentPaintLimits);

			paintEdges( g2d, hundredPercentPaintLimits, solid );

			paintVertices(g2d, hundredPercentPaintLimits);

			layerPostPaint(g2d, hundredPercentPaintLimits);

			g2d.dispose();
		}
	}

	protected abstract void prePaint(Graphics grp);
	protected abstract void layerPrePaint(Graphics2D g2d, GeometricRectangle hundredPercentPaintLimits);
	protected abstract void layerPostPaint(Graphics2D g2d, GeometricRectangle hundredPercentPaintLimits);

	protected GeometricRectangle getZoomedBounds()
	{
		return( zoom( getBounds() ) );
	}

	protected BufferedImage getCurrentOriginalImage()
	{
		return( getMorphingVerticesManager().getCurrentImageModel().getImage() );
	}

	protected double getFactor( GeometricRectangle newImageSize )
	{
		return( newImageSize.width / getCurrentOriginalImage().getWidth() );
	}

	protected GeometricRectangle getZoomedBounds(GeometricRectangle newImageSize )
	{
		return( zoom( getBounds(), getFactor( newImageSize ) ) );
	}

	protected GeometricRectangle getPaintLimits(Graphics grp )
	{
		GeometricRectangle result = null;
		Rectangle clipBounds = grp.getClipBounds();
		if( clipBounds != null )
			result = getBuilder().createRectangle(clipBounds.x, clipBounds.y,
												clipBounds.width, clipBounds.height);

		result = intersection( result, getZoomedBounds() );

		return( result );
	}

	protected GeometricRectangle getHundredPercentPaintLimits(Graphics grp )
	{
		GeometricRectangle result = null;
		Rectangle clipBounds = grp.getClipBounds();
		if( clipBounds != null )
		{
			double xx = unzoomImageValue( clipBounds.x );
			double yy = unzoomImageValue( clipBounds.y );
			double right = unzoomImageValue( clipBounds.x + clipBounds.width - 1 );
			double bottom = unzoomImageValue( clipBounds.y + clipBounds.height - 1 );

			result = getBuilder().createRectangle(xx, yy, right - xx + 1, right - yy + 1);
		}

		result = intersection( result, getEffectiveBounds() );

		return( result );
	}

	protected GeometricRectangle intersection(GeometricRectangle rect1,
                                              GeometricRectangle rect2 )
	{
		return( getMorphingHelper().intersection( rect1, rect2 ) );
	}

	protected void blackScreen( Component comp, Graphics gc )
	{
		int width = comp.getWidth();
		int height = comp.getHeight();

		gc.setColor( Color.BLACK );
		gc.fillRect( 0, 0, width, height );
	}

	public String getCurrentImageName() {
		return _currentImageName;
	}

	protected GeometricRectangle getBounds()
	{
		return( getBounds( getCurrentImageName() ) );
	}

	protected String getMasterImageName()
	{
		return( getMorphingVerticesManager().getMasterImageName() );
	}

	protected GeometricRectangle getMasterBounds()
	{
		return( getBounds( getMasterImageName() ) );
	}

	protected GeometricRectangle getBounds(String imageName )
	{
		return( getIfNotNull( 
			getIfNotNull( getMorphingVerticesManager(),
						vm -> vm.getImageModel( imageName ) ),
			MorphingImageModel::getBounds ) );
	}

	protected int getVertexRadius()
	{
		return getAppliZoomedValue(getVertexRadio100percent());
	}

	protected abstract void browseEdges(GeometricRectangle bounds, BiConsumer<Point2d, Point2d> edgeConsumer );

	protected abstract GeometricRectangle getBoundsForBrowsingEdges();

	protected void paintVertices(Graphics gc, GeometricRectangle hundredPercentPaintLimits)
	{
		int radius = getVertexRadius();

		browseEdges(getBoundsForBrowsingEdges(),
			(pt1, pt2) -> {
				paintVertex( gc, hundredPercentPaintLimits, pt1, radius );
				paintVertex( gc, hundredPercentPaintLimits, pt2, radius ); } );
	}

	protected boolean sureIsNotVisible( Rectangle rect, Point2d pt1, Point2d pt2 )
	{
		return( ( rect != null ) &&
			ViewFunctions.instance()
				.sureIsNotVisible( rect.x, rect.y, rect.width, rect.height,
									pt1.x, pt1.y, pt2.x, pt2.y ) );
	}

	protected Color getVertexColor( Point2d pt )
	{
		Color result;
		if( !Objects.equals(pt, getClosestPoint() ) )
			result = getVertexColor();
		else
			result = getClosestMovableVertexColor();

		return( result );
	}

	public boolean paintVertex(Graphics gc, GeometricRectangle hundredPercentPaintLimits,
							   Point2d pt, int radius )
	{
		Point2d mappedPt = mapPoint( pt );
		Color color = getVertexColor( mappedPt );
		return paintVertex(gc, hundredPercentPaintLimits, pt, radius, color);
	}

	public boolean paintVertex(Graphics gc, GeometricRectangle hundredPercentPaintLimits,
							   Point2d pt, int radius, Color color)
	{
		boolean result = false;
		Point2d mappedPt = mapPoint( pt );

		gc.setColor( color );
		if( hundredPercentPaintLimits.isInside(mappedPt) )
		{
			result = true;
			Point2d zpt = zoom( mappedPt );
			gc.fillOval( zpt.getIntX() - radius, zpt.getIntY() - radius,
					2 * radius, 2 * radius );
		}
		return result;
	}

	protected Color getEdgeColor()
	{
		return edgeColor;
	}

	protected void drawVertexNumberString(Graphics2D g2d, Point pt, String text )
	{
		ImageFunctions.instance().paintStringCentered(g2d, getFont(), text, getVertexNumberColor(), pt);
	}

	protected int getTextSize100Percent() {
		return textSize100Percent;
	}

	protected double getTextFontSize()
	{
		return( getAppliZoomedValue(getTextSize100Percent()) );
	}

	protected Font getFont()
	{
		return( createFontGeneric( _font, getTextFontSize(), 8, 60 ) );
	}

	protected Font createFontGeneric( AtomicReference<Font> arFont,
									double fontSize,
									int minimum, int maximum )
	{
		int fontSizeI = (int) Math.floor( fontSize );
		fontSizeI = IntegerFunctions.max( fontSizeI, minimum );
		fontSizeI = IntegerFunctions.min( fontSizeI, maximum );

		if( hasToCreateFont(arFont, fontSizeI) )
			arFont.set( new Font( "Lucida", Font.BOLD, fontSizeI ) );

		return( arFont.get() );
	}

	protected boolean hasToCreateFont( AtomicReference<Font> arFont, int fontSize )
	{
		return( (arFont.get()==null) || (arFont.get().getSize()!=fontSize) );
	}

	protected Stroke createSolidStroke()
	{
		int width = getAppliZoomedValue( getEdgeWidth100percent() );
		return( new BasicStroke(width) );
	}

	protected void paintEdges(Graphics2D g2d, GeometricRectangle hundredPercentPaintLimits, Stroke solid )
	{

		int width = getAppliZoomedValue( getEdgeWidth100percent() );

		// Set the stroke of the copy, not the original 
		Stroke dashed = new BasicStroke(width, BasicStroke.CAP_BUTT, BasicStroke.JOIN_BEVEL,
										0, new float[]{9}, 0);

		g2d.setColor( getEdgeColor() );

		browseEdges(getBoundsForBrowsingEdges(),
			(pt1, pt2) -> drawEdge( g2d, hundredPercentPaintLimits, pt1, pt2, dashed, solid ) );
	}

	protected boolean isPointInProgress( Point2d pt )
	{
		return( Objects.equals( pt, getPointInProgress() ) );
	}

	protected boolean isPointInProgress( Point2d pt1, Point2d pt2 )
	{
		return( isPointInProgress( pt1 ) ||	isPointInProgress( pt2 ) );
	}

	protected MorphingHelper getMorphingHelper()
	{
		return( MorphingHelper.instance() );
	}

	protected Point2d zoom( Point2d pt )
	{
		return( getMorphingHelper().zoomToInt(pt, getImageZoomFactor() ) );
	}

	protected GeometricRectangle zoom(GeometricRectangle rect )
	{
		return( zoom(rect, getImageZoomFactor() ) );
	}

	protected GeometricRectangle zoom(GeometricRectangle rect, double factor )
	{
		return( getMorphingHelper().zoomToInt(rect, factor ) );
	}

	protected MorphingVerticesMap getVerticesMap()
	{
		return( getMorphingVerticesManager().getVerticesMap() );
	}

	protected Point2d fromNormalizedInt( Point2d percentagesPoint, GeometricRectangle rect )
	{
		return( getMorphingHelper().fromNormalizedInt(percentagesPoint, rect) );
	}

	protected Point2d toNormalized( Point2d percentagesPoint, GeometricRectangle rect )
	{
		return( getMorphingHelper().toNormalized(percentagesPoint, rect) );
	}

	protected GeometricRectangle getEffectiveBounds() {
		return getBounds();
	}

	protected abstract Point2d mapPoint( Point2d masterPt );

	protected boolean noneIsNull(Object ... objs) {
		return getNullFunctions().noneIsNull(objs);
	}

	protected void drawEdge( Graphics2D g2d, GeometricRectangle hundredPercentPaintLimits,
							Point2d pt1, Point2d pt2, Stroke dashed, Stroke solid )
	{
		Stroke stroke = solid;

		// from masterImage point to image Point
		Point2d mappedPt1 = mapPoint( pt1 );
		Point2d mappedPt2 = mapPoint( pt2 );

		if( noneIsNull(mappedPt1, mappedPt2) &&
				!getMorphingHelper().canDiscard(hundredPercentPaintLimits, mappedPt1, mappedPt2) )
		{
			if( isPointInProgress( mappedPt1, mappedPt2 ) )
				stroke = dashed;

			// zooming with image factor
			Point2d zpt1 = zoom(mappedPt1);
			Point2d zpt2 = zoom(mappedPt2);

			g2d.setStroke(stroke);
			if( ( zpt1 == null ) || ( zpt2 == null ) )
			{
				int ii = 0;
			}

			g2d.drawLine( zpt1.getIntX(), zpt1.getIntY(),
							zpt2.getIntX(), zpt2.getIntY() );
		}
	}

	protected int unzoomImageValue( double value )
	{
		return( IntegerFunctions.zoomValueRound(value, 1.0d / getImageZoomFactor() ) );
	}

	protected int zoomImageValue( double value )
	{
		return( IntegerFunctions.zoomValueRound(value, getImageZoomFactor() ) );
	}

	protected double getAppZoom()
	{
		return( getAppliConf().getZoomFactor() );
	}

	protected int getAppliZoomedValue( double value )
	{
		// application zoomToInt factor may be different from _imageZoomFactor (the level of zoomToInt the the image).
		return( IntegerFunctions.zoomValueCeil(value, getAppZoom() ) );
	}

	@Override
	public void releaseResources()
	{
	}

	protected MorphingEditorApplicationContext getAppliCtx() {
		return appliCtx;
	}

	protected MorphingApplicationConfiguration getAppliConf()
	{
		return( getAppliCtx().getAppliConf() );
	}

	@Override
	public void invertColors( ColorInversor colorInversor)
	{
/*
		if( ! _doNotInvertColors )
		{
			invertSinglePanelColors(colorInversor);
			invertColorsChild(colorInversor);
		}
*/
	}

	public void setPointInProgress(String imageName, Point2d _pointInProgress) {
		this._currentImageName = imageName;
		this._pointInProgress = _pointInProgress;

		if( _pointInProgress == null )
		{
			int kk = 0;
		}
	}

	public double getImageZoomFactor() {
		return _imageZoomFactor;
	}

	public void setImageZoomFactor(double _zoomFactor) {
		this._imageZoomFactor = _zoomFactor;
	}

	public Point2d getClosestPoint() {
		return _closestPoint;
	}

	public void setClosestPoint(String imageName, Point2d _closestPoint) {
		this._currentImageName = imageName;
		this._closestPoint = _closestPoint;
		LOGGER.info("closestPoint: {}", _closestPoint);
	}

	protected Component getParent()
	{
		return( _parentRef.get() );
	}

	protected void repaint()
	{
		SwingUtilities.invokeLater( getParent()::repaint );
	}

	protected GeometricRectangle createDelaunayRectangle(Dimension size )
	{
		return( GeometryCommonFactory.instance().createDelaunayRectangle( size ) );
	}

	protected GeometricRectangle createDelaunayRectangle(Rectangle rectangle )
	{
		return( GeometryCommonFactory.instance().createDelaunayRectangle( rectangle ) );
	}

	protected GeometricRectangle getImageSize()
	{
		return( createDelaunayRectangle( getParent().getPreferredSize() ) );
	}

	protected GeometricRectangle getClip()
	{
		return( createDelaunayRectangle( ComponentFunctions.instance().getClip( getParent() ) ) );
	}

	protected <RR> boolean addPendingExecutor( ExecutionFunctions.UnsafeFunction<RR> function, Consumer<RR> callback )
	{
		return( FrameworkComponentFunctions.instance().addPendingExecutor(getParent(), function, callback) );
	}

	public Point2d createPoint( Point2d that ) {
		return( GeometryCommonFactory.instance().createPoint2d( that ) );
	}

	protected Point2d getZoomedCenterPointShown()
	{
		return( getIfNotNull( getClip(), GeometricRectangle::center ) );
	}

	public Point2d getNormalizedPointImageCenterShown()
	{
		return( toNormalized(getZoomedCenterPointShown(), getZoomedBounds()) );
	}

//	protected abstract void resetWhenImageChanges();

	public void updateTriangulation(String imageName) {//,
//								 MorphingVerticesMapManager morphingVerticesManager) {
		boolean changed = !Objects.equals(imageName, getCurrentImageName());
		_currentImageName = imageName;

//		_morphingVerticesManager = morphingVerticesManager;

//		if(changed) {
//			resetWhenImageChanges();
//		}
	}
}
