/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.applications.libmorphing.view.painters;

import com.frojasg1.applications.libmorphing.configuration.MorphingApplicationConfiguration;
import com.frojasg1.applications.libmorphing.controller.MorphingVerticesMapManagerBase;
import com.frojasg1.applications.libmorphing.helpers.MorphingHelper;
import com.frojasg1.applications.libmorphing.model.images.image.MorphingImageModel;
import com.frojasg1.applications.libmorphing.view.panels.evt.DelaunayJPanelsEventBuilderBase;
import com.frojasg1.delaunay.triangulation.quadedge.builder.QuadEdgeDelaunayBuilder;
import com.frojasg1.libgeometry.types.GeometricRectangle;
import com.frojasg1.libgeometry.types.Point2d;
import com.frojasg1.libgeometry.types.factory.GeometryCommonFactory;
import com.frojasg1.general.desktop.view.ComponentFunctions;
import com.frojasg1.general.desktop.view.ContainerFunctions;
import com.frojasg1.general.desktop.view.FrameworkComponentFunctions;
import com.frojasg1.general.desktop.view.pdf.ImageJPanel;
import com.frojasg1.general.number.DoubleReference;
import com.frojasg1.general.view.ReleaseResourcesable;
import com.frojasg1.general.view.evt.wrapper.ListOfGenComponentWrapperEventContextListenersBase;
import com.frojasg1.general.view.evt.wrapper.impl.GenComponentWrapperEvent;

import javax.swing.*;
import java.awt.*;
import java.awt.event.KeyListener;
import java.awt.image.BufferedImage;
import java.util.function.BiConsumer;

/**
 *
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 */
public abstract class DelaunayTriangulationPanelWrapper<CO extends DelaunayTriangulationJPanelControllerBase, PA extends DelaunayTriangulationPainterBase>
		implements ReleaseResourcesable {

	protected CO _delaunayListeners;

	protected ImageJPanel _imageJPanel;
	protected QuadEdgeDelaunayBuilder _delaunayBuilder;

	protected MorphingApplicationConfiguration _appliConf;

	protected PA _painter;

	protected ListOfGenComponentWrapperEventContextListenersBase<JPanel, GenComponentWrapperEvent> _listenersList;
	protected DelaunayJPanelsEventBuilderBase _eventBuilder;

	public DelaunayTriangulationPanelWrapper(ImageJPanel imageJPanel,
											 MorphingApplicationConfiguration appliConf,
                                             QuadEdgeDelaunayBuilder delaunayBuilder,
                                             ListOfGenComponentWrapperEventContextListenersBase<JPanel, GenComponentWrapperEvent> listenersList,
                                             DelaunayJPanelsEventBuilderBase eventBuilder) {
		_imageJPanel = imageJPanel;
		_delaunayBuilder = delaunayBuilder;
		_appliConf = appliConf;
		_listenersList = listenersList;
		_eventBuilder = eventBuilder;
	}

	public void init()
	{
		_delaunayListeners = createDelaunayListeners();
		createSetPainter( createPainter() );

		addListeners();
	}

	protected void createSetPainter( PA painter )
	{
		_painter = painter;
		this.getImageJPanel().setExternalPainter( _painter );
	}

	public ListOfGenComponentWrapperEventContextListenersBase<JPanel, GenComponentWrapperEvent> getListenersList() {
		return _listenersList;
	}

	public DelaunayJPanelsEventBuilderBase getEventBuilder() {
		return _eventBuilder;
	}

	protected PA getPainter() {
		return _painter;
	}

	protected abstract PA createPainter();
//	protected DelaunayTriangulationPainter createPainter()
//	{
//		return( new DelaunayTriangulationPainter( getAppliConf(), _imageJPanel, getBuilder()) );
//	}

	protected MorphingApplicationConfiguration getAppliConf() {
		return _appliConf;
	}

	protected JPanel getContentPane()
	{
		return( (JPanel) ComponentFunctions.instance().getFirstParentOf(
			ContainerFunctions.instance()::isContentPane, _imageJPanel ) );
	}

	protected Component getWindow()
	{
		return( (Component) FrameworkComponentFunctions.instance().getInternationalizedWindow(_imageJPanel) );
	}

	protected void addKeyListeners( KeyListener listener )
	{
		SwingUtilities.invokeLater( () ->
			operateKeyListeners( listener, Component::addKeyListener ) );
	}

	protected void removeKeyListeners( KeyListener listener )
	{
		operateKeyListeners( listener, Component::removeKeyListener );
	}

	protected void operateKeyListeners( KeyListener listener,
										BiConsumer<Component, KeyListener> keyListenerFunction )
	{
		Component window = getWindow();
		ComponentFunctions.instance().browseComponentHierarchy(window,
			comp -> {
				keyListenerFunction.accept( comp, listener );
				return( null );
		});
	}

	protected Component getViewPort()
	{
		return( ComponentFunctions.instance().getViewport( getImageJPanel() ) );
	}

	protected abstract void addListeners();
//	{
//		addKeyListeners(_delaunayListeners);
//		getImageJPanel().addMouseListener(_delaunayListeners);
//		getImageJPanel().addMouseMotionListener(_delaunayListeners);
//		getImageJPanel().addMouseWheelListener(_delaunayListeners);
//		getViewPort().addMouseMotionListener(_delaunayListeners);
//	}

	protected abstract void removeListeners();
//	{
//		removeKeyListeners(_delaunayListeners);
//		getImageJPanel().removeMouseListener(_delaunayListeners);
//		getImageJPanel().removeMouseMotionListener(_delaunayListeners);
//		getImageJPanel().removeMouseWheelListener(_delaunayListeners);
//		getViewPort().removeMouseMotionListener(_delaunayListeners);
//	}

	public CO getDelaunayListeners() {
		return _delaunayListeners;
	}

	protected ImageJPanel getImageJPanel() {
		return _imageJPanel;
	}

	protected abstract CO createDelaunayListeners();
//	protected DelaunayTriangulationJPanelController createDelaunayListeners()
//	{
//		return( new DelaunayTriangulationJPanelController( getImageJPanel(),
//															getEventBuilder(),
//															getListenersList(),
//															getBuilder() ) );
//	}

	protected QuadEdgeDelaunayBuilder getBuilder()
	{
		return( _delaunayBuilder );
	}

	@Override
	public void releaseResources()
	{
		removeListeners();
	}

	public void setImageZoomFactor( DoubleReference zoomFactorRef )
	{
		double zoomFactor = zoomFactorRef._value;
		getPainter().setImageZoomFactor(zoomFactor);
		getDelaunayListeners().setImageZoomFactor(zoomFactor);
		getImageJPanel().setZoomFactor(zoomFactorRef);
	}

	public MorphingImageModel getImageModel( MorphingVerticesMapManagerBase morphingVerticesManager,
											String imageName )
	{
		MorphingImageModel result = null;
		if( morphingVerticesManager != null )
			result = morphingVerticesManager.getImageModel(imageName);

		return( result );
	}

	protected BufferedImage getImage( MorphingVerticesMapManagerBase morphingVerticesManager,
									String imageName )
	{
		MorphingImageModel mim = getImageModel( morphingVerticesManager, imageName );
		BufferedImage result = null;
		if( mim != null )
			result = mim.getImage();

		return( result );
	}

//	public void setTriangulation(String imageName,
//								MorphingVerticesMapManager morphingVerticesManager,
//								Point2d normalizedCenterPoint,
//								Double newImageZoomFactor)
//	{
//		DoubleReference zoomFactorRef = getImageJPanel().getZoomFactor();
//		MorphingImageModel imageModel = getImageModel( morphingVerticesManager, imageName );
//		BufferedImage image = ( imageModel == null ) ? null : imageModel.getImage();
////		getImageJPanel().setNotZoomedImage(image, zoomFactorRef,
////										InitialVerticalPosition.NOTHING);
//		getImageJPanel().setImageToBePrezoomed(image,
//				previousCenter( normalizedCenterPoint, imageModel ),
//				newImageZoomFactor );
//
//		getPainter().setTriangulation(imageName, morphingVerticesManager);
//		getDelaunayListeners().setImageModel(imageModel);
//	}

	protected Point previousCenter(Point2d normalizedCenterPoint, MorphingImageModel imageModel )
	{
		return( toPoint( fromNormalizedInt(normalizedCenterPoint, imageModel.getBounds() ) ) );
	}

	protected Point toPoint( Point2d pt )
	{
		return( GeometryCommonFactory.instance().toPoint( pt ) );
	}

	protected Point2d fromNormalizedInt( Point2d percentagesPoint, GeometricRectangle rect )
	{
		return( MorphingHelper.instance().fromNormalizedInt(percentagesPoint, rect) );
	}

//	public void setNewOnTheFlyBounds( String imageName, DelaunayRectangle bounds )
//	{
//		getPainter().setNewOnTheFlyBounds( imageName, bounds );
//	}

	public void setPointInProgress(String imageName, Point2d pt) {
		getPainter().setPointInProgress( imageName, pt );
	}

	public void setClosestPoint(String imageName, Point2d pt) {
		getPainter().setClosestPoint( imageName, pt );
	}

//	public void fadeToMaster(double fadingFraction)
//	{
//		getPainter().fadeToMaster(fadingFraction);
//	}
//
//	public void resetFadedToMasterImage()
//	{
//		getPainter().resetFadedToMasterImage();
//	}
//
//	public Point2d getNormalizedPointImageCenterShown()
//	{
//		return( getPainter().getNormalizedPointImageCenterShown() );
//	}
//
//	public void setSelectedMultiTriangle( MultiTriangle multiTriangle )
//	{
//		getPainter().setSelectedMultiTriangle(multiTriangle);
//
//		getImageJPanel().repaint();
//	}
//
//	public MultiTriangle getSelectedMultiTriangle()
//	{
//		return( getPainter().getSelectedMultiTriangle() );
//	}
//
//	public void resetSelectedMultiTriangle()
//	{
//		getPainter().resetSelectedMultiTriangle();
//	}

	public double getImageZoomFactor()
	{
		return( getImageJPanel().getZoomFactorDouble() );
	}
}
