/*
 * Copyright (c) 2023. Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 3.0 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You may obtain a copy of the License at
 *
 *       http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */

package com.frojasg1.applications.libmorphing.view.painters;

import com.frojasg1.applications.common.configuration.InternationalizedStringConf;
import com.frojasg1.applications.libmorphing.contexts.MorphingEditorApplicationContext;
import com.frojasg1.applications.libmorphing.api.facedetector.pojo.FaceDetectionMeshData;
import com.frojasg1.applications.libmorphing.configuration.MorphingApplicationConfiguration;
import com.frojasg1.applications.libmorphing.controller.MorphingVerticesMapManager;
import com.frojasg1.applications.libmorphing.helpers.MorphingHelper;
import com.frojasg1.applications.libmorphing.language.LibMorphingInternationalTexts;
import com.frojasg1.applications.libmorphing.model.images.color.types.MultiTriangle;
import com.frojasg1.applications.libmorphing.model.images.image.MorphingImageModel;
import com.frojasg1.applications.libmorphing.view.painters.control.ControlDelaunayTriangulationPanelWrapperLayer;
import com.frojasg1.applications.libmorphing.view.painters.delaunay.DelaunayTriangulationPanelWrapperLayer;
import com.frojasg1.applications.libmorphing.view.painters.faceselector.FaceDetectionSelectorPanelWrapperLayer;
import com.frojasg1.applications.libmorphing.view.panels.evt.DelaunayJPanelsEventBuilderBase;
import com.frojasg1.delaunay.triangulation.quadedge.builder.QuadEdgeDelaunayBuilder;
import com.frojasg1.libgeometry.types.GeometricRectangle;
import com.frojasg1.libgeometry.types.Point2d;
import com.frojasg1.libgeometry.types.factory.GeometryCommonFactory;
import com.frojasg1.general.NullFunctions;
import com.frojasg1.general.desktop.view.painter.ExternalPainter;
import com.frojasg1.general.desktop.view.painter.impl.ComposedExternalPainter;
import com.frojasg1.general.desktop.view.pdf.ImageJPanel;
import com.frojasg1.general.number.DoubleReference;
import com.frojasg1.general.view.evt.wrapper.ListOfGenComponentWrapperEventContextListenersBase;
import com.frojasg1.general.view.evt.wrapper.impl.GenComponentWrapperEvent;

import javax.swing.*;
import java.awt.*;
import java.awt.image.BufferedImage;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.function.Function;

public class WholeDelaunayTriangulationPanelWrapper {

    protected DelaunayTriangulationPanelWrapperLayer delaunayTriangulationPanelWrapperLayer;
    protected ControlDelaunayTriangulationPanelWrapperLayer controlDelaunayTriangulationPanelWrapperLayer;
    protected FaceDetectionSelectorPanelWrapperLayer faceDetectionSelectorPanelWrapperLayer;

    protected ExternalPainter painter;

    protected ImageJPanel imageJPanel;
    protected MorphingApplicationConfiguration appliConf;
    protected QuadEdgeDelaunayBuilder delaunayBuilder;
    protected ListOfGenComponentWrapperEventContextListenersBase<JPanel, GenComponentWrapperEvent> listenersList;
    protected DelaunayJPanelsEventBuilderBase eventBuilder;
    protected InternationalizedStringConf internationalStrings;

    protected Map<TriangulationPainterLayerTypeEnum, DelaunayTriangulationPanelWrapperLayerBase<?, ?>> panelWrapperLayersMap
            = new ConcurrentHashMap<>();

    protected MorphingEditorApplicationContext appliCtx;

    public WholeDelaunayTriangulationPanelWrapper(MorphingEditorApplicationContext appliCtx,
                                                  ImageJPanel imageJPanel,// MorphingApplicationConfiguration appliConf,
                                                  QuadEdgeDelaunayBuilder delaunayBuilder,
                                                  ListOfGenComponentWrapperEventContextListenersBase<JPanel, GenComponentWrapperEvent> listenersList,
                                                  DelaunayJPanelsEventBuilderBase eventBuilder) {
        this(appliCtx, imageJPanel, delaunayBuilder, listenersList, eventBuilder, LibMorphingInternationalTexts.instance());
    }

    public WholeDelaunayTriangulationPanelWrapper(MorphingEditorApplicationContext appliCtx,
                                                  ImageJPanel imageJPanel,// MorphingApplicationConfiguration appliConf,
                                                  QuadEdgeDelaunayBuilder delaunayBuilder,
                                                  ListOfGenComponentWrapperEventContextListenersBase<JPanel, GenComponentWrapperEvent> listenersList,
                                                  DelaunayJPanelsEventBuilderBase eventBuilder,
                                                  InternationalizedStringConf internationalStrings) {
        this.appliCtx = appliCtx;
        this.imageJPanel = imageJPanel;
//        this.appliConf = appliConf;
        this.delaunayBuilder = delaunayBuilder;
        this.listenersList = listenersList;
        this.eventBuilder = eventBuilder;
        this.internationalStrings = internationalStrings;
    }

    public WholeDelaunayTriangulationPanelWrapper init() {
        createLayerWrappers();

        createSetPainter(createComposedExternalPainter());

        return this;
    }

    protected MorphingEditorApplicationContext getAppliCtx() {
        return appliCtx;
    }

    protected void createLayerWrappers() {
        delaunayTriangulationPanelWrapperLayer = createDelaunayTriangulationPanelWrapperLayerBase();
        controlDelaunayTriangulationPanelWrapperLayer = createControlDelaunayTriangulationPanelWrapperLayer();
        faceDetectionSelectorPanelWrapperLayer = createFaceDetectionSelectorPanelWrapperLayer();

        put(delaunayTriangulationPanelWrapperLayer);
        put(controlDelaunayTriangulationPanelWrapperLayer);
        put(faceDetectionSelectorPanelWrapperLayer);

        deactivateAll();
    }

    protected void put(DelaunayTriangulationPanelWrapperLayerBase<?, ?> layerWrapper) {
        panelWrapperLayersMap.put(layerWrapper.getLayerType(), layerWrapper);
    }

    protected DelaunayTriangulationPanelWrapperLayerBase<?, ?> get(TriangulationPainterLayerTypeEnum key) {
        return panelWrapperLayersMap.get(key);
    }

    public void activateLayer(TriangulationPainterLayerTypeEnum key, boolean value) {
        DelaunayTriangulationPanelWrapperLayerBase<?, ?> layerWrapper = get(key);
        if(layerWrapper != null) {
            layerWrapper.setActive(value);
        }
    }

    public void activateOnlyLayer(TriangulationPainterLayerTypeEnum key) {
        deactivateAll();
        activateLayer(key, true);
    }

    public void deactivateAll() {
        for(DelaunayTriangulationPanelWrapperLayerBase<?, ?> layerWrapper: panelWrapperLayersMap.values()) {
            layerWrapper.setActive(false);
        }
    }

    protected InternationalizedStringConf getInternationalStrings() {
        return internationalStrings;
    }

    protected DelaunayTriangulationPanelWrapperLayer getDelaunayTriangulationPanelWrapperLayer() {
        return delaunayTriangulationPanelWrapperLayer;
    }

    protected ControlDelaunayTriangulationPanelWrapperLayer getControlDelaunayTriangulationPanelWrapperLayer() {
        return controlDelaunayTriangulationPanelWrapperLayer;
    }

    protected FaceDetectionSelectorPanelWrapperLayer getFaceDetectionSelectorPanelWrapperLayer() {
        return faceDetectionSelectorPanelWrapperLayer;
    }

    protected ExternalPainter getPainter() {
        return painter;
    }

    protected ImageJPanel getImageJPanel() {
        return imageJPanel;
    }

    protected MorphingApplicationConfiguration getAppliConf() {
        return appliConf;
    }

    protected QuadEdgeDelaunayBuilder getDelaunayBuilder() {
        return delaunayBuilder;
    }

    protected ListOfGenComponentWrapperEventContextListenersBase<JPanel, GenComponentWrapperEvent> getListenersList() {
        return listenersList;
    }

    protected DelaunayJPanelsEventBuilderBase getEventBuilder() {
        return eventBuilder;
    }

    protected void createSetPainter(ExternalPainter painter )
    {
        this.painter = painter;
        this.getImageJPanel().setExternalPainter( painter );
    }

    protected void putPainter(ComposedExternalPainter<TriangulationPainterLayerTypeEnum> result, TriangulationPainterLayerTypeEnum layerType) {
        result.put(layerType, get(layerType).getPainter());
    }

    protected ExternalPainter createComposedExternalPainter() {
        ComposedExternalPainter<TriangulationPainterLayerTypeEnum> result = new ComposedExternalPainter<>();

        putPainter(result, TriangulationPainterLayerTypeEnum.DELAUNAY_TRIANGULATION);
        putPainter(result, TriangulationPainterLayerTypeEnum.FACE_DETECTION_VERTEX_MODIFICATION);
        putPainter(result, TriangulationPainterLayerTypeEnum.FACE_DETECTION_SELECTION);

        return result;
    }

    protected DelaunayTriangulationPanelWrapperLayer createDelaunayTriangulationPanelWrapperLayerBase() {
        return new DelaunayTriangulationPanelWrapperLayer(getAppliCtx(), getImageJPanel(),
                TriangulationPainterLayerTypeEnum.DELAUNAY_TRIANGULATION,// getAppliConf(),
                getDelaunayBuilder(),
                getInternationalStrings(), getListenersList(), getEventBuilder())
                .init();
    }

    protected ControlDelaunayTriangulationPanelWrapperLayer createControlDelaunayTriangulationPanelWrapperLayer() {
        return new ControlDelaunayTriangulationPanelWrapperLayer(getAppliCtx(), getImageJPanel(),
                TriangulationPainterLayerTypeEnum.FACE_DETECTION_VERTEX_MODIFICATION,// getAppliConf(),
                getDelaunayBuilder(),
                getInternationalStrings(), getListenersList(), getEventBuilder())
                .init();
    }

    protected FaceDetectionSelectorPanelWrapperLayer createFaceDetectionSelectorPanelWrapperLayer() {
        return new FaceDetectionSelectorPanelWrapperLayer(getAppliCtx(), getImageJPanel(),
                TriangulationPainterLayerTypeEnum.FACE_DETECTION_SELECTION,// getAppliConf(),
                getDelaunayBuilder(),
                getInternationalStrings(), getListenersList(), getEventBuilder())
                .init();
    }

    protected Point previousCenter(Point2d normalizedCenterPoint, MorphingImageModel imageModel )
    {
        return toPoint(fromNormalizedInt(normalizedCenterPoint, getIfNotNull(imageModel, MorphingImageModel::getBounds)));
    }

    protected Point toPoint( Point2d pt )
    {
        return( GeometryCommonFactory.instance().toPoint( pt ) );
    }

    protected Point2d fromNormalizedInt( Point2d percentagesPoint, GeometricRectangle rect )
    {
        return( MorphingHelper.instance().fromNormalizedInt(percentagesPoint, rect) );
    }

    protected NullFunctions getNullFunctions() {
        return NullFunctions.instance();
    }

    protected <CC, RR> RR getIfNotNull(CC obj, Function<CC, RR> getter) {
        return getNullFunctions().getIfNotNull(obj, getter);
    }

    public MorphingImageModel getImageModel(MorphingVerticesMapManager morphingVerticesManager,
                                            String imageName )
    {
        return getIfNotNull(morphingVerticesManager, manager -> manager.getImageModel(imageName));
    }

    protected MorphingVerticesMapManager getVerticesManager() {
        return getAppliCtx().getVerticesManager();
    }

    protected void updateJImagePanel(String imageName,
                                     Point2d normalizedCenterPoint,
                                     Double newImageZoomFactor) {
        MorphingVerticesMapManager morphingVerticesManager = getVerticesManager();
        MorphingImageModel imageModel = getImageModel(morphingVerticesManager, imageName);
        BufferedImage image = (imageModel == null) ? null : imageModel.getImage();
//		getImageJPanel().setNotZoomedImage(image, zoomFactorRef,
//										InitialVerticalPosition.NOTHING);
        if(image != null) {
            getImageJPanel().setImageToBePrezoomed(image,
                    previousCenter(normalizedCenterPoint, imageModel),
                    newImageZoomFactor);
        }
    }

    public void setTriangulation(String imageName,
//                                 MorphingVerticesMapManager morphingVerticesManager,
                                 Point2d normalizedCenterPoint,
                                 Double newImageZoomFactor) {
        updateJImagePanel(imageName, normalizedCenterPoint, newImageZoomFactor);

        getDelaunayTriangulationLayerWrapper().setTriangulation(imageName,// morphingVerticesManager,
                normalizedCenterPoint, newImageZoomFactor);
        getControlDelaunayTriangulationPanelWrapperLayer().setTriangulation(imageName,// morphingVerticesManager,
                normalizedCenterPoint, newImageZoomFactor);
        getFaceDetectionSelectorPanelWrapperLayer().setTriangulation(imageName,// morphingVerticesManager,
                normalizedCenterPoint, newImageZoomFactor);
    }

    public void setPointInProgress(String imageName, Point2d pt) {
        getDelaunayTriangulationLayerWrapper().setPointInProgress( imageName, pt );
    }

    public void setClosestPoint(String imageName, Point2d pt) {
        getDelaunayTriangulationLayerWrapper().setClosestPoint( imageName, pt );
    }

    public void setControlMeshClosestPoint(String imageName, Point2d pt) {
        getControlDelaunayTriangulationPanelWrapperLayer().setClosestPoint( imageName, pt );
    }

    protected DelaunayTriangulationPanelWrapperLayerBase<?, ?> getDelaunayTriangulationLayerWrapper() {
        return get(TriangulationPainterLayerTypeEnum.DELAUNAY_TRIANGULATION);
    }

    public double getImageZoomFactor()
    {
        return( getDelaunayTriangulationLayerWrapper().getImageZoomFactor() );
    }

    public Point2d getNormalizedPointImageCenterShown()
    {
        return( getDelaunayTriangulationPanelWrapperLayer().getNormalizedPointImageCenterShown() );
    }

    public void setSelectedMultiTriangle( MultiTriangle multiTriangle )
    {
        getDelaunayTriangulationPanelWrapperLayer().setSelectedMultiTriangle(multiTriangle);
    }

    public MultiTriangle getSelectedMultiTriangle()
    {
        return( getDelaunayTriangulationPanelWrapperLayer().getSelectedMultiTriangle() );
    }

    public void resetSelectedMultiTriangle()
    {
        getDelaunayTriangulationPanelWrapperLayer().resetSelectedMultiTriangle();
    }

    public void setNewOnTheFlyBounds( String imageName, GeometricRectangle bounds )
    {
        getDelaunayTriangulationPanelWrapperLayer().setNewOnTheFlyBounds( imageName, bounds );
    }

    public void fadeToMaster(double fadingFraction)
    {
        getDelaunayTriangulationPanelWrapperLayer().fadeToMaster(fadingFraction);
    }

    public void resetFadedToMasterImage()
    {
        getDelaunayTriangulationPanelWrapperLayer().resetFadedToMasterImage();
    }

    public void setImageZoomFactor( DoubleReference pdfZoomFactor )
    {
        for(DelaunayTriangulationPanelWrapperLayerBase<?, ?> layerWrapper: panelWrapperLayersMap.values()) {
            layerWrapper.setImageZoomFactor( pdfZoomFactor );
        }
        getDelaunayTriangulationPanelWrapperLayer().setImageJPanelZoomFactor(pdfZoomFactor);
    }

    public void setAllFaces(String imageName, List<FaceDetectionMeshData> allFaces) {
        getFaceDetectionSelectorPanelWrapperLayer().setAllFaces(imageName, allFaces);
    }

    public void setFaceDetectionMeshData(String imageName, FaceDetectionMeshData faceDetectionMeshData) {
        getFaceDetectionSelectorPanelWrapperLayer().setFaceDetectionMeshData(imageName, faceDetectionMeshData);
    }
}
