/*
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.applications.libmorphing.view.painters.control;

import com.frojasg1.applications.common.configuration.InternationalizedStringConf;
import com.frojasg1.applications.libmorphing.configuration.MorphingApplicationConfiguration;
import com.frojasg1.applications.libmorphing.controller.ControlMeshManager;
import com.frojasg1.applications.libmorphing.model.images.vertex.VertexModel;
import com.frojasg1.applications.libmorphing.model.images.vertex.map.impl.control.DynamicMultimapVertexWithIndex;
import com.frojasg1.applications.libmorphing.model.images.vertex.map.impl.control.IndexedVerticesMap;
import com.frojasg1.applications.libmorphing.view.painters.TriangulationPainterLayerTypeEnum;
import com.frojasg1.applications.libmorphing.view.panels.evt.enums.VertexEventType;
import com.frojasg1.applications.libmorphing.view.painters.DelaunayTriangulationJPanelControllerBase;
import com.frojasg1.applications.libmorphing.view.panels.evt.*;
import com.frojasg1.delaunay.triangulation.quadedge.builder.QuadEdgeDelaunayBuilder;
import com.frojasg1.libgeometry.types.Point2d;
import com.frojasg1.general.desktop.mouse.MouseFunctions;
import com.frojasg1.general.view.evt.wrapper.ListOfGenComponentWrapperEventContextListenersBase;
import com.frojasg1.general.view.evt.wrapper.impl.GenComponentWrapperEvent;

import javax.swing.*;
import java.awt.*;
import java.awt.event.MouseEvent;
import java.awt.event.MouseWheelEvent;
import java.util.List;
import java.util.function.Consumer;

/**
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 */
public class ControlDelaunayTriangulationJPanelController
        extends DelaunayTriangulationJPanelControllerBase {
    public static final String CONF_RESET_MODIFIED_DETECTED_MESH = "RESET_MODIFIED_DETECTED_MESH";
    public static final String CONF_RESET_FACE_MESH_FRAME = "RESET_FACE_MESH_FRAME";

    protected static final int PROXIMITY_100_PERCENT_RADIUS = 12;

    protected List<VertexModel> currentImageVertexList;

    public ControlDelaunayTriangulationJPanelController(Component comp, TriangulationPainterLayerTypeEnum layerType,
                                                        String imageName,
                                                        DelaunayJPanelsEventBuilderBase evtFactory,
                                                        ListOfGenComponentWrapperEventContextListenersBase<JPanel, GenComponentWrapperEvent> listenersList,
                                                        QuadEdgeDelaunayBuilder builder, InternationalizedStringConf internationalStrings) {
        super(comp, layerType,
                imageName, PROXIMITY_100_PERCENT_RADIUS, evtFactory, listenersList, builder, internationalStrings);
    }

    public ControlDelaunayTriangulationJPanelController init() {
        super.init();

        return this;
    }

    protected MyPopupMenu createBasePopupMenu()
    {
        MyPopupMenu result = new MyPopupMenu();
        result.init();

        return( result );
    }

    protected IndexedVerticesMap get88Mesh() {
        return getIfNotNull(getControlMeshManager(), ControlMeshManager::getControlMeshVerticesMap);
    }

    @Override
    protected void browseVertices(Consumer<VertexModel> vertexProcessor) {
        IndexedVerticesMap my88Mesh = get88Mesh();
        String imageName = getImageName();
        if((my88Mesh != null) && (my88Mesh.mapExists(imageName))) {
            int index = 0;
            DynamicMultimapVertexWithIndex multiVertex;
            while((multiVertex = my88Mesh.get(index++)) != null) {
                vertexProcessor.accept(multiVertex.get(imageName));
            }
        }
    }

    @Override
    protected boolean hasToDiscardForClosest(Point2d pt, Point2d imagePt) {
        return !getControlMeshManager().canMoveControlPoint(pt, getImageName());
    }

    @Override
    public void mouseMovedInternal(MouseEvent e) {
        boolean closestSet = false;
        Point2d imagePoint = createPoint2d(getImagePoint(e));

        if (imagePoint != null) {
            if (isShiftPressed(e)) {
                closestSet = true;
                updateClosest(imagePoint);
            }
        }

        if (!closestSet)
            resetMovingPoint();
    }

    @Override
    public void mouseExitedInternal(MouseEvent e) {
        resetMovingPoint();
    }

    @Override
    public void mousePressedInternal(MouseEvent e) {
        rightClickPressed = false;
        boolean hasToResetMovingPoint = _isMovingPoint;
        Point pt = getImagePoint(e);
        Point2d imagePoint = createPoint2d(pt);

        if (MouseFunctions.isMainButtonPressed(e)) {
            if (isShiftPressed(e)) {
                if (imagePoint != null) {
                    startMoving(imagePoint);
                    hasToResetMovingPoint = false;
                }
            }
        } else if(MouseFunctions.isRightButtonPressed(e)) {
            rightClickPressed = true;
        }

        if (hasToResetMovingPoint)
            resetMovingPoint();

        resetClosest();
    }

    @Override
    public void mouseReleasedInternal(MouseEvent e) {
        if (_isMovingPoint) {
            if (isShiftPressed(e))
                endMovingPoint();
            else
                resetMovingPoint();
        } else if(rightClickPressed) {
            doPopup(e);
        }

        resetClosest();
        _isMovingPoint = false;
        rightClickPressed = false;
    }

    @Override
    public void mouseDraggedInternal(MouseEvent e) {
        Point2d imagePoint = createPoint2d(getImagePoint(e));
        if (_isMovingPoint) {
            _isMovingPoint = MouseFunctions.isMainButtonPressed(e) && isShiftPressed(e);

            if (!_isMovingPoint || (imagePoint == null))
                resetMovingPoint();
            else {
                // we sendn as original vertex the previous coordinates of the closest vertex
                notifyNewMovingPoint(null, imagePoint);
            }
        }
    }

    protected void notifyClosest( Point2d pt )
    {
        _closestVertex = pt;
        notifyNewPoint( null, pt, false, VertexEventType.CONTROL_MESH_VERTEX_APPROACHING );
    }

    @Override
    protected void notifyNewMovingPoint( Point2d originalPoint, Point2d newPoint )
    {
//		Point2d newPoint = createPoint2d( pt );
        notifyNewPoint( originalPoint, newPoint, false, VertexEventType.CONTROL_MESH_VERTEX_MOVING );
    }

    @Override
    protected void endMovingPoint()
    {
        notifyNewPoint(null, true, VertexEventType.CONTROL_MESH_VERTEX_MOVING);
    }

    @Override
    public void mouseWheelMovedInternal(MouseWheelEvent e) {
    }

    protected MorphingApplicationConfiguration getAppliConf() {
        return getControlMeshManager().getAppliConf();
    }

    protected ResetDetectedFaceMeshFrameEvent createResetDetectedFaceMeshFrameEvent()
    {
        return(getEvtFactory().createResetDetectedFaceMeshFrameEvent(getImageName()));
    }

    protected ResetModifiedDetectedFaceMeshEvent createResetModifiedDetectedFaceMeshEvent()
    {
        return(getEvtFactory().createResetModifiedDetectedFaceMeshEvent(getImageName()));
    }

    protected void resetModifiedDetectedFaceMesh() {
        notifyEvt(createResetModifiedDetectedFaceMeshEvent());
    }

    protected void resetDetectedFaceMeshFrame() {
        notifyEvt(createResetDetectedFaceMeshFrameEvent());
    }

    protected class MyPopupMenu extends MorphingPanelControllerPopupMenuBase
    {
        public void init()
        {
            super.init();

            createFixedPart();
            createCommonPart();
        }

        protected void createFixedPart()
        {
            addMenuComponent(CONF_RESET_MODIFIED_DETECTED_MESH, new JMenuItem(),
                    () -> getInternationalStrings().getInternationalString(CONF_RESET_MODIFIED_DETECTED_MESH),
                    ControlDelaunayTriangulationJPanelController.this::resetModifiedDetectedFaceMesh,
                    null);

            addMenuComponent(CONF_RESET_FACE_MESH_FRAME, new JMenuItem(),
                    () -> getInternationalStrings().getInternationalString(CONF_RESET_FACE_MESH_FRAME),
                    ControlDelaunayTriangulationJPanelController.this::resetDetectedFaceMeshFrame,
                    null);

            add( new Separator() );
        }

        @Override
        protected void preparePopupMenuItems() {
            super.preparePopupMenuItems();
        }
    }
}
