/*
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.applications.libmorphing.view.painters.control;

import com.frojasg1.applications.libmorphing.contexts.MorphingEditorApplicationContext;
import com.frojasg1.applications.libmorphing.api.facedetector.helpers.Control88meshCreator;
import com.frojasg1.applications.libmorphing.api.facedetector.helpers.ControlMeshBrowser;
import com.frojasg1.applications.libmorphing.controller.ControlMeshManager;
import com.frojasg1.applications.libmorphing.model.images.vertex.map.impl.control.DynamicMultimapVertexWithIndex;
import com.frojasg1.applications.libmorphing.model.images.vertex.map.impl.control.IndexedVerticesMap;
import com.frojasg1.applications.libmorphing.view.painters.DelaunayTriangulationPainterBase;
import com.frojasg1.delaunay.triangulation.quadedge.builder.QuadEdgeDelaunayBuilder;
import com.frojasg1.libgeometry.types.GeometricRectangle;
import com.frojasg1.libgeometry.types.Point2d;
import com.frojasg1.general.desktop.view.colors.Colors;

import java.awt.*;
import java.util.Objects;
import java.util.function.BiConsumer;
import java.util.function.Consumer;

/**
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 */
public class ControlDelaunayTriangulationPainter extends DelaunayTriangulationPainterBase {
    protected static final int TEXT_SIZE_100_PERCENT = 10;

    protected static final double EDGE_WIDTH_100_PERCENT = 2d;
    protected static final double VERTEX_RADIO_100_PERCENT = 12;

    protected static final Color VERTEX_COLOR = Colors.BUTAN;
    protected static final Color VERTEX_NUMBER_COLOR = Color.BLACK;
    protected static final Color CLOSEST_VERTEX_COLOR = Colors.PURPLE;
    protected static final Color EDGE_COLOR = Colors.BROWN_ROSY_BROWN;
    protected static final Color INTERNAL_FRAME_FOR_AUTOCALCULATION_CORNER_VERTEX_COLOR = Color.RED;
    protected static final Color NOT_MOVABLE_VERTEX_COLOR = Colors.COOL_BLUE;
    protected static final Color INTERNAL_FRAME_FOR_AUTOCALCULATION_CLOSEST_CORNER_VERTEX_COLOR = Colors.LILAC;

    protected Color internalFrameForAutocalculationCornerVertexColor;
    protected Color notMovableVertexColor;
    protected Color internalFrameForAutocalculationClosestCornerVertexColor;

    /**
     * Creates new form GraphicalEqualizerJPanel
     */
    public ControlDelaunayTriangulationPainter(MorphingEditorApplicationContext appliCtx,
                                               Component parent,
                                               QuadEdgeDelaunayBuilder delaunayBuilder) {
        super(appliCtx, parent, delaunayBuilder, TEXT_SIZE_100_PERCENT, EDGE_WIDTH_100_PERCENT, VERTEX_RADIO_100_PERCENT,
                VERTEX_COLOR, VERTEX_NUMBER_COLOR, CLOSEST_VERTEX_COLOR, EDGE_COLOR);

        internalFrameForAutocalculationCornerVertexColor = INTERNAL_FRAME_FOR_AUTOCALCULATION_CORNER_VERTEX_COLOR;
        notMovableVertexColor = NOT_MOVABLE_VERTEX_COLOR;
        internalFrameForAutocalculationClosestCornerVertexColor = INTERNAL_FRAME_FOR_AUTOCALCULATION_CLOSEST_CORNER_VERTEX_COLOR;
    }

    protected Color getInternalFrameForAutocalculationCornerVertexColor() {
        return internalFrameForAutocalculationCornerVertexColor;
    }

    protected Color getNotMovableVertexColor() {
        return notMovableVertexColor;
    }

    protected Color getInternalFrameForAutocalculationClosestCornerVertexColor() {
        return internalFrameForAutocalculationClosestCornerVertexColor;
    }

    @Override
    public void paintComponent(Component comp, Graphics grp) {
        super.paintComponent(comp, grp);
    }

    @Override
    protected boolean hasToPaint() {
        return true;
//        return canMapControlMesh();
    }

    protected boolean canMapControlMesh() {
        return getIfNotNull(getControlMeshManager(), manager -> manager.canMap(getCurrentImageName()), false);
    }

    @Override
    protected void paintComponentInternal(Component comp, Graphics grp) {
        super.paintComponentInternal(comp, grp);
    }

    protected IndexedVerticesMap get88Mesh() {
        return getIfNotNull(getControlMeshManager(), ControlMeshManager::getControlMeshVerticesMap);
    }

    @Override
    protected void prePaint(Graphics grp) {
        // intentionally left blank
    }

    @Override
    protected void layerPrePaint(Graphics2D g2d, GeometricRectangle hundredPercentPaintLimits) {

        paintFrame(g2d);

        int radius = getAppliZoomedValue(VERTEX_RADIO_100_PERCENT);
        browseVertices(get88Mesh(), getCurrentImageName(),
                pt -> paintVertex(g2d, hundredPercentPaintLimits, pt, radius));
    }

    protected Control88meshCreator getControl88meshCreator() {
        return Control88meshCreator.instance();
    }

    public Point2d getVertex(int index) {
        return getControl88meshCreator().getVertex(get88Mesh(), getCurrentImageName(), index);
    }

    protected void paintFrame(Graphics2D g2d) {
        if(get88Mesh() != null) {
            // see 88-control.point.mesh.with.68-landmark.jpg
            Point2d ul = getVertex(80);
            Point2d dl = getVertex(82);
            Point2d dr = getVertex(84);
            Point2d ur = getVertex(86);

            Stroke solid = createSolidStroke();
            drawEdge(g2d, ul, dl, solid);
            drawEdge(g2d, dr, dl, solid);
            drawEdge(g2d, dr, ur, solid);
            drawEdge(g2d, ul, ur, solid);
        }
    }

    protected void drawEdge(Graphics2D g2d, Point2d pt1, Point2d pt2, Stroke stroke) {
        // zooming with image factor
        Point2d zpt1 = zoom(pt1);
        Point2d zpt2 = zoom(pt2);

        g2d.setStroke(stroke);

        g2d.drawLine( zpt1.getIntX(), zpt1.getIntY(),
                zpt2.getIntX(), zpt2.getIntY() );
    }

    protected void browseVertices(IndexedVerticesMap my88Mesh, String imageName, Consumer<Point2d> pointProcessor) {
        if((my88Mesh != null) && (my88Mesh.mapExists(imageName))) {
            int index = 0;
            DynamicMultimapVertexWithIndex multiVertex;
            while((multiVertex = my88Mesh.get(index++)) != null) {
                pointProcessor.accept(multiVertex.get(imageName).getPoint());
            }
        }
    }


    @Override
    protected void layerPostPaint(Graphics2D g2d, GeometricRectangle hundredPercentPaintLimits) {
        // intentionally left blank
    }

    protected ControlMeshBrowser getControlMeshBrowser() {
        return ControlMeshBrowser.instance();
    }

    protected void browseEdges(int[][] triangles, IndexedVerticesMap controlMeshVerticesMap, String imageName,
                               GeometricRectangle bounds, BiConsumer<Point2d, Point2d> edgeConsumer) {
        getControlMeshBrowser().browseEdges(triangles, controlMeshVerticesMap, imageName, bounds, edgeConsumer);
    }

    protected IndexedVerticesMap getControlMeshVerticesMap() {
        return getIfNotNull(getControlMeshManager(), ControlMeshManager::getControlMeshVerticesMap);
    }

    protected int[][] getControlMeshTriangles() {
//        return getIfNotNull(getControlMeshManager(), ControlMeshManager::getControlMeshTriangles);
        return null;
    }

    @Override
    protected void browseEdges(GeometricRectangle bounds, BiConsumer<Point2d, Point2d> edgeConsumer) {
        browseEdges(getControlMeshTriangles(), getControlMeshVerticesMap(),
                getCurrentImageName(), getBoundsForBrowsingEdges(), edgeConsumer);
    }

//    @Override
//    protected void resetWhenImageChanges() {
//        // intentionally left blank
//    }

    @Override
    protected GeometricRectangle getBoundsForBrowsingEdges() {
        return getBounds();
    }

    public int getIndex(Point2d pt, IndexedVerticesMap vertices, String normalizedName) {
        return getControl88meshCreator().getIndex(pt, vertices, normalizedName);
    }

    public boolean paintVertex(Graphics gc, GeometricRectangle hundredPercentPaintLimits,
                               Point2d pt, int radius ) {
        boolean result = super.paintVertex(gc, hundredPercentPaintLimits, pt, radius);
        if(result) {
            drawVertexNumberString((Graphics2D) gc, calculateVertexNumberStringPosition(pt),
                    String.valueOf(getIndex(pt, getControlMeshVerticesMap(), getCurrentImageName())));
        }

        return result;
    }

    protected boolean isFaceDetectionFrameCorner(Point2d pt) {
        return getControlMeshManager().isAnyOfTheControlFrameCorners(getCurrentImageName(), pt);
    }

    protected boolean differentColorForFaceDetectionFrameCorner(Point2d pt) {
        return isFaceDetectionFrameCorner(pt);
    }

    protected boolean isMovable(Point2d pt) {
        return getControlMeshManager().canMoveControlPoint(pt, getCurrentImageName());
    }

    protected Color getVertexColor( Point2d pt )
    {
        Color result;
        if( !Objects.equals(pt, getClosestPoint() ) )
            if(differentColorForFaceDetectionFrameCorner(pt)) {
                result = getInternalFrameForAutocalculationCornerVertexColor();
            } else if(!isMovable(pt)) {
                result = getNotMovableVertexColor();
            } else {
                result = getVertexColor();
            }
        else {
            if(differentColorForFaceDetectionFrameCorner(pt)) {
                result = getInternalFrameForAutocalculationClosestCornerVertexColor();
            } else if(!isMovable(pt)) {
                result = getNotMovableVertexColor();
            } else {
                result = getClosestMovableVertexColor();
            }
        }

        return( result );
    }

    protected Point calculateVertexNumberStringPosition(Point2d pt) {
        Point result = new Point();
        result.x = (int) (pt.x * getImageZoomFactor());
        result.y = (int) (pt.y * getImageZoomFactor());

        return (result);
    }

    // with that painter, we do not do any mapping, the point is directly the one of the image.
    @Override
    protected Point2d mapPoint(Point2d imagePt) {
        return imagePt;
    }
}
