/*
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.applications.libmorphing.view.painters.delaunay;

import com.frojasg1.applications.common.configuration.InternationalizedStringConf;
import com.frojasg1.applications.libmorphing.model.images.image.MorphingImageModel;
import com.frojasg1.applications.libmorphing.model.images.vertex.VertexModel;
import com.frojasg1.applications.libmorphing.view.painters.TriangulationPainterLayerTypeEnum;
import com.frojasg1.applications.libmorphing.view.panels.evt.enums.VertexEventType;
import com.frojasg1.applications.libmorphing.view.painters.DelaunayTriangulationJPanelControllerBase;
import com.frojasg1.applications.libmorphing.view.panels.evt.*;
import com.frojasg1.delaunay.triangulation.quadedge.builder.QuadEdgeDelaunayBuilder;
import com.frojasg1.libgeometry.types.GeometricRectangle;
import com.frojasg1.libgeometry.types.Point2d;
import com.frojasg1.general.desktop.mouse.MouseFunctions;
import com.frojasg1.general.number.IntegerFunctions;
import com.frojasg1.general.view.evt.wrapper.ListOfGenComponentWrapperEventContextListenersBase;
import com.frojasg1.general.view.evt.wrapper.impl.GenComponentWrapperEvent;

import javax.swing.*;
import java.awt.*;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseWheelEvent;
import java.awt.image.BufferedImage;
import java.util.List;
import java.util.function.Consumer;

/**
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 */
public class DelaunayTriangulationJPanelController
        extends DelaunayTriangulationJPanelControllerBase
        implements KeyListener {
    protected static final int PROXIMITY_100_PERCENT_RADIUS = 12;
    public static final String CONF_INSERT_68_DETECTED_POINTS = "INSERT_68_DETECTED_POINTS";

    protected boolean _isOnTheFlyPoint;

    protected GeometricRectangle _zoomedAndMovedBounds;
    protected Point2d _normalizedLocationForZoomedAndMovedBounds;

    public DelaunayTriangulationJPanelController(Component comp, TriangulationPainterLayerTypeEnum layerType,
                                                 String imageName,
                                                 DelaunayJPanelsEventBuilderBase evtFactory,
                                                 ListOfGenComponentWrapperEventContextListenersBase<JPanel, GenComponentWrapperEvent> listenersList,
                                                 QuadEdgeDelaunayBuilder builder, InternationalizedStringConf internationalStrings) {
        super(comp, layerType,
                imageName, PROXIMITY_100_PERCENT_RADIUS, evtFactory, listenersList, builder, internationalStrings);
    }

    public DelaunayTriangulationJPanelController init() {
        super.init();

        return this;
    }

    protected MyPopupMenu createBasePopupMenu()
    {
        MyPopupMenu result = new MyPopupMenu();
        result.init();

        return( result );
    }

    @Override
    public void keyTyped(KeyEvent e) {
        if(isActive()) {

        }
    }

    @Override
    public void keyPressed(KeyEvent e) {
        if(isActive()) {
            if ((e.getKeyCode() == KeyEvent.VK_CONTROL) &&
                    (MouseFunctions.isMouseInsideComponent(_parent))) {
                _isOnTheFlyPoint = true;
                Point2d imagePoint = createPoint2d(getImagePointOfPointOnScreen());
                if (imagePoint != null)
                    notifyNewPoint(imagePoint, false, VertexEventType.INSERT);
            }
        }
    }

    @Override
    public void keyReleased(KeyEvent e) {
        if(isActive()) {
            if ((e.getKeyCode() == KeyEvent.VK_CONTROL) &&
                    (MouseFunctions.isMouseInsideComponent(_parent))) {
//			notifyOnTheFlyVertexDeletion();
                resetOnTheFlyPoint();
            } else if (e.getKeyCode() == KeyEvent.VK_ESCAPE)
                resetFadedMasterImage();
        }
    }

    protected void resetFadedMasterImage() {
        notifyFadedMasterImage(0d);
    }

    protected List<VertexModel> getVertices() {
        List<VertexModel> result = null;
        MorphingImageModel imageModel = getImageModel();
        if (imageModel != null)
            result = imageModel.getVertices();

        return (result);
    }

    @Override
    protected void browseVertices(Consumer<VertexModel> vertexProcessor) {
        List<VertexModel> list = getVertices();
        if(list != null) {
            for(VertexModel vm: list) {
                vertexProcessor.accept(vm);
            }
        }
    }

    @Override
    protected boolean hasToDiscardForClosest(Point2d pt, Point2d imagePt) {
        return (isCorner(pt));
    }

    protected GeometricRectangle getLimitingBounds() {
        GeometricRectangle result = _zoomedAndMovedBounds;
        if (result == null)
            result = super.getLimitingBounds();

        return (result);
    }

    @Override
    public void mouseMovedInternal(MouseEvent e) {
        boolean onTheFlySet = false;
        boolean closestSet = false;
        Point2d imagePoint = createPoint2d(getImagePoint(e));

        if (imagePoint != null) {
            if (isCtrlPressed(e)) {
                onTheFlySet = true;
                _isOnTheFlyPoint = true;
                notifyOnTheFlyVertexInsertion(imagePoint);
            } else if (isShiftPressed(e)) {
                closestSet = true;
                updateClosest(imagePoint);
            }
        }

        if (!closestSet)
            resetMovingPoint();

        if (!onTheFlySet)
            resetOnTheFlyPoint();

        resetZoomedAndMovedBounds();
    }

    @Override
    public void mouseExitedInternal(MouseEvent e) {
//		if (isCtrlPressed(e))
//			notifyOnTheFlyVertexDeletion();
        resetOnTheFlyPoint();

        resetMovingPoint();

        resetZoomedAndMovedBounds();
    }

    @Override
    public void mousePressedInternal(MouseEvent e) {
        rightClickPressed = false;
        boolean hasToResetMovingPoint = _isMovingPoint;
        Point pt = getImagePoint(e);
        Point2d imagePoint = createPoint2d(pt);

        if (MouseFunctions.isMainButtonPressed(e)) {
            if (isCtrlPressed(e)) {
                if (imagePoint != null) {
                    notifyNewPoint(imagePoint, true, VertexEventType.INSERT);
                    _isOnTheFlyPoint = false;
                } else
                    notifyOnTheFlyVertexDeletion();
            } else if (isShiftPressed(e)) {
                if (imagePoint != null) {
                    startMoving(imagePoint);
                    hasToResetMovingPoint = false;
                }
            } else {
                if (imagePoint != null)
                    notifyTriangleSelectedEvent(imagePoint);
            }
        } else if (MouseFunctions.isMiddleButtonPressed(e)) {
            if (isMovingOrZoomingBounds()) {
                if (MouseFunctions.isMiddleButtonPressed(e))
                    notifyNewConfirmedBoundsEvt();
            }
        } else if (MouseFunctions.isRightButtonPressed(e)) {
            if (isShiftPressed(e)) {
                if (_closestVertex != null)
                    notifyVertexDeletion(_closestVertex);
                else if (imagePoint != null) {
                    if (!isMovingOrZoomingBounds())
                        _normalizedLocationForZoomedAndMovedBounds = toNormalized(imagePoint);
                }
            } else {
                rightClickPressed = true;
            }
//			else
//				notifyClearTriangulation();
        }

        if (hasToResetMovingPoint)
            resetMovingPoint();

        resetClosest();
    }

    @Override
    protected void notifyClosest( Point2d pt )
    {
        _closestVertex = pt;
        notifyNewPoint( null, pt, false, VertexEventType.APPROACHING );
    }

    @Override
    protected void notifyNewMovingPoint( Point2d originalPoint, Point2d newPoint )
    {
//		Point2d newPoint = createPoint2d( pt );
        notifyNewPoint( originalPoint, newPoint, false, VertexEventType.MOVING );
    }

    @Override
    protected void endMovingPoint()
    {
        notifyNewPoint(null, true, VertexEventType.MOVING);
    }

    protected void notifyTriangleSelectedEvent(Point2d point) {
        notifyEvt(createTriangleSelectedEvent(point, getImageName()));
    }

    protected void resetZoomedAndMovedBounds() {
        notifyNewOnTheFlyBounds(null);
    }

    protected void notifyNewOnTheFlyBounds(GeometricRectangle bounds) {
        if (bounds == null)
            _normalizedLocationForZoomedAndMovedBounds = null;

        if ((_zoomedAndMovedBounds != null) || (bounds != null))
            notifyNewOnTheFlyBoundsEvtInternal(bounds);

        _zoomedAndMovedBounds = bounds;
    }

    protected void notifyNewOnTheFlyBoundsEvtInternal(GeometricRectangle bounds) {
        notifyEvt(createNewOnTheFlyBoundsEvt(bounds));
    }

    protected void notifyNewConfirmedBoundsEvt() {
        if (_zoomedAndMovedBounds != null)
            notifyNewConfirmedBoundsEvtInternal(_zoomedAndMovedBounds);

        resetZoomedAndMovedBounds();
    }

    protected void notifyNewConfirmedBoundsEvtInternal(GeometricRectangle bounds) {
        notifyEvt(createNewConfirmedBoundsEvt(bounds));
    }

    @Override
    public void mouseReleasedInternal(MouseEvent e) {
        if (_isMovingPoint) {
            if (isShiftPressed(e))
                endMovingPoint();
            else
                resetMovingPoint();
        } else if(rightClickPressed) {
            doPopup(e);
        }

        resetClosest();
        _isMovingPoint = false;
        rightClickPressed = false;

        resetZoomedAndMovedBounds();
    }

    @Override
    public void mouseDraggedInternal(MouseEvent e) {
        boolean movingOrZoomingBoundsSet = false;
        Point2d imagePoint = createPoint2d(getImagePoint(e));
        if (_isMovingPoint) {
            _isMovingPoint = MouseFunctions.isMainButtonPressed(e) && isShiftPressed(e);

            if (!_isMovingPoint || (imagePoint == null))
                resetMovingPoint();
            else {
//				if( isMaster() )
//					notifyNewPoint( null, e.getPoint(), false, VertexEventType.INSERT);
//				else
                notifyNewMovingPoint(null, imagePoint);
            }
        } else {
            if (isMovingOrZoomingBounds()) {
                if (MouseFunctions.isRightButtonPressed(e))
                    movingOrZoomingBoundsSet = updateAndNotifyOnTheFlyBounds(imagePoint, 0);
            }
        }

        if (!movingOrZoomingBoundsSet)
            resetZoomedAndMovedBounds();
    }

    @Override
    public void mouseWheelMovedInternal(MouseWheelEvent e) {
        Point2d imagePoint = createPoint2d(getImagePointOfPointOnScreen());
        if (isMovingOrZoomingBounds()) {
            if (MouseFunctions.isRightButtonPressed(e)) {
                int sign = -IntegerFunctions.sgn(e.getWheelRotation());
                updateAndNotifyOnTheFlyBounds(imagePoint, sign * 10);
            }
        }
    }

    protected GeometricRectangle calculateNewSizeRoughly(GeometricRectangle prevBounds,
                                                         int sizeIncrement, GeometricRectangle originalBounds) {
        GeometricRectangle result = prevBounds.createCopy();

        if (originalBounds.width > originalBounds.height) {
            result.width += sizeIncrement;
            result.height = calculateHeightKeepingAspectRatio(
                    result.width, originalBounds);
        } else {
            result.height += sizeIncrement;
            result.width = calculateWidthKeepingAspectRatio(
                    result.height, originalBounds);
        }

        return (result);
    }

    protected double getAspectRatio(GeometricRectangle originalBounds) {
        return (originalBounds.width / originalBounds.height);
    }

    protected int calculateHeightKeepingAspectRatio(double width,
                                                    GeometricRectangle originalBounds) {
        return (zoom(width, 1 / getAspectRatio(originalBounds)));
    }

    protected int calculateWidthKeepingAspectRatio(double height,
                                                   GeometricRectangle originalBounds) {
        return (zoom(height, getAspectRatio(originalBounds)));
    }

    protected GeometricRectangle calculateNewBoundsRoughly(GeometricRectangle prevBounds,
                                                           Point2d normalizedInitialPoint,
                                                           int sizeIncrement, Point2d imagePoint, GeometricRectangle originalBounds) {
        GeometricRectangle result = calculateNewSizeRoughly(prevBounds, sizeIncrement,
                originalBounds);

        result.x = imagePoint.x - zoom(result.width, normalizedInitialPoint.x);
        result.y = imagePoint.y - zoom(result.height, normalizedInitialPoint.y);

        return (result);
    }

    public void limitNewBounds(GeometricRectangle result, BufferedImage image,
                               GeometricRectangle originalBounds) {
        getMorphingHelper().limitNewBounds(result, image, originalBounds);
    }

    protected GeometricRectangle calculateNewBounds(GeometricRectangle prevBounds,
                                                    Point2d normalizedInitialPoint,
                                                    int sizeIncrement, Point2d imagePoint) {
        GeometricRectangle result = null;
        GeometricRectangle originalBounds = getBounds();
        if (prevBounds == null)
            prevBounds = originalBounds;

        if ((prevBounds != null) && (imagePoint != null) &&
                (normalizedInitialPoint != null) && (originalBounds != null)) {
            result = calculateNewBoundsRoughly(prevBounds, normalizedInitialPoint,
                    sizeIncrement, imagePoint, originalBounds);
            limitNewBounds(result, getImage(), originalBounds);
        }

        return (result);
    }

    protected void limitBoundsSize(GeometricRectangle result, BufferedImage image,
                                   GeometricRectangle originalBounds) {
        getMorphingHelper().limitBoundsSize(result, image, originalBounds);
    }

    protected boolean updateAndNotifyOnTheFlyBounds(Point2d imagePoint, int sizeIncrement) {
        boolean result = false;
        GeometricRectangle newBounds = calculateNewBounds(_zoomedAndMovedBounds,
                _normalizedLocationForZoomedAndMovedBounds,
                sizeIncrement, imagePoint);

        if (newBounds != null) {
            notifyNewOnTheFlyBounds(newBounds);
            result = true;
        }

        return (result);
    }

    protected boolean isMovingOrZoomingBounds() {
        return (_normalizedLocationForZoomedAndMovedBounds != null);
    }

    protected void notifyOnTheFlyVertexInsertion(Point2d imagePoint) {
        notifyNewPoint(imagePoint, false, VertexEventType.INSERT);
    }

    protected void notifyVertexDeletion(Point2d pt) {
        notifyNewPoint(null, pt, true, VertexEventType.DELETE);
    }

    protected void notifyOnTheFlyVertexDeletion() {
        notifyNewPoint((Point2d) null, false, VertexEventType.DELETE);
    }

    protected void notifyFadedMasterImage(double fadingPercentage) {
        notifyEvt(createFadedMasterImageButtonPressedEvt(fadingPercentage));
    }

    protected void notifyInsert68DetectedPointsEvt() {
        notifyEvt(createInsert68DetectedPointsEvt());
    }

    protected Insert68DetectedPointsEvt createInsert68DetectedPointsEvt() {
        return (getEvtFactory().createInsert68DetectedPointsEvt(getImageName()));
    }

    protected FadedMasterImageButtonPressedEvt createFadedMasterImageButtonPressedEvt(double fadingPercentage) {
        return (getEvtFactory().createFadedMasterImageButtonPressedEvt(fadingPercentage));
    }

    public TriangleSelectedEvent createTriangleSelectedEvent(Point2d pt, String imageName) {
        return (getEvtFactory().createTriangleSelectedEvent(pt, imageName));
    }

    public NewOnTheFlyBoundsEvt createNewOnTheFlyBoundsEvt(GeometricRectangle newBounds) {
        return (getEvtFactory().createNewOnTheFlyBoundsEvt(getImageName(), newBounds));
    }

    public NewConfirmedBoundsEvt createNewConfirmedBoundsEvt(GeometricRectangle newBounds) {
        return (getEvtFactory().createNewConfirmedBoundsEvt(getImageName(), newBounds));
    }

//	protected void notifyNewPoint( Point2d originalPoint, Point pt,
//									boolean confirmed, VertexEventType type )
//	{
//		notifyNewPoint( originalPoint, createPoint2d( pt ), confirmed, type );
//	}

    protected void resetOnTheFlyPoint() {
        if (_isOnTheFlyPoint)
            notifyOnTheFlyVertexDeletion();

        _isOnTheFlyPoint = false;
    }

//    protected void resetMovingPoint() {
////        MorphingImageModel prevImageModel = _imageModel;
//////		if( _isMovingPoint && ( prevImageModel != null ) && prevImageModel.isMaster() )
//////			notifyOnTheFlyVertexDeletion();
//
//        resetClosest();
//        if (_isMovingPoint)
//            endMovingPoint();
//
//        _isMovingPoint = false;
//    }

    protected class MyPopupMenu extends MorphingPanelControllerPopupMenuBase
    {
        public MyPopupMenu()
        {
            super();
        }

        public void init()
        {
            super.init();

            createFixedPart();
            createCommonPart();
        }

        protected void createFixedPart()
        {
            addMenuComponent( CONF_INSERT_68_DETECTED_POINTS, new JMenuItem(),
                    () -> getInternationalStrings().getInternationalString(CONF_INSERT_68_DETECTED_POINTS),
                    DelaunayTriangulationJPanelController.this::notifyInsert68DetectedPointsEvt,
                    null);

            add( new Separator() );
        }

        protected boolean hasToEnable68DetectedPointsOption() {
            return isMaster() && getControlMeshManager().hasControlMesh(getImageName())
                    && getMorphingVerticesMapManager().isDelaunayTriangulationEmpty();
        }

        protected void prepareFixedPart() {
            enableComponents(hasToEnable68DetectedPointsOption(), CONF_INSERT_68_DETECTED_POINTS);
        }

        @Override
        protected void preparePopupMenuItems() {
            super.preparePopupMenuItems();
            prepareFixedPart();
        }
    }
}
