/*
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.applications.libmorphing.view.painters.delaunay;

import com.frojasg1.applications.libmorphing.contexts.MorphingEditorApplicationContext;
import com.frojasg1.applications.libmorphing.frames.framebuilder.impl.NormalTransitionMorphingFrameCreator;
import com.frojasg1.applications.libmorphing.controller.DelaunayTriangulatorController;
import com.frojasg1.applications.libmorphing.controller.MorphingVerticesMapManager;
import com.frojasg1.applications.libmorphing.model.images.color.types.MultiTriangle;
import com.frojasg1.applications.libmorphing.model.images.image.MorphingImageModel;
import com.frojasg1.applications.libmorphing.model.images.vertex.VertexModel;
import com.frojasg1.applications.libmorphing.model.images.vertex.map.impl.normal.MorphingVerticesMap;
import com.frojasg1.applications.libmorphing.view.painters.DelaunayTriangulationPainterBase;
import com.frojasg1.delaunay.triangulation.quadedge.DelaunaySubdivision;
import com.frojasg1.delaunay.triangulation.quadedge.builder.QuadEdgeDelaunayBuilder;
import com.frojasg1.delaunay.triangulation.quadedge.helpers.DelaunayTriangleBrowser;
import com.frojasg1.libgeometry.types.GeometricRectangle;
import com.frojasg1.libgeometry.types.GeometricTriangle;
import com.frojasg1.libgeometry.types.Point2d;
import com.frojasg1.general.desktop.view.colors.Colors;
import com.frojasg1.general.desktop.view.pdf.BackgroundImageProcessing;
import com.frojasg1.general.map.multimap.dynamic.deductible.DeductibleDynamicMultiMapElementBase;
import com.frojasg1.general.view.ReleaseResourcesable;

import java.awt.*;
import java.awt.image.BufferedImage;
import java.util.ArrayList;
import java.util.List;
import java.util.Objects;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.function.BiConsumer;

/**
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 */
public class DelaunayTriangulationPainter extends DelaunayTriangulationPainterBase
        implements ReleaseResourcesable {
    protected static final int TEXT_SIZE_100_PERCENT = 16;

    protected static final double EDGE_WIDTH_100_PERCENT = 2d;
    protected static final double VERTEX_RADIO_100_PERCENT = 5;

    protected static final Color VERTEX_COLOR = Colors.BRIGHT_GREEN;
    protected static final Color VERTEX_NUMBER_COLOR = Color.RED;
    protected static final Color CLOSEST_VERTEX_COLOR = Colors.BROWN_ROSY_BROWN;
    protected static final Color EDGE_COLOR = Colors.PURPLE;
    protected static final Color BOUNDS_ON_THE_FLY_EDGE_COLOR = Colors.OLIVE_DRAB;
    protected static final Color SELECTED_TRIANGLE_EDGE_COLOR = Colors.OLIVE_DRAB;

    protected GeometricRectangle _onTheFlyBounds;

    protected BackgroundImageProcessing _fadedToMasterImage;

    protected MultiTriangle _selectedMultiTriangle;

    protected Color boundsOnTheFlyEdgeColor;
    protected Color selectedTriangleEdgeColor;

    protected MorphingVerticesMapManager morphingVerticesManager;

    /**
     * Creates new form GraphicalEqualizerJPanel
     */
    public DelaunayTriangulationPainter(MorphingEditorApplicationContext appliCtx,
                                        Component parent,
                                        QuadEdgeDelaunayBuilder delaunayBuilder) {
        super(appliCtx, parent, delaunayBuilder, TEXT_SIZE_100_PERCENT, EDGE_WIDTH_100_PERCENT, VERTEX_RADIO_100_PERCENT,
                VERTEX_COLOR, VERTEX_NUMBER_COLOR, CLOSEST_VERTEX_COLOR, EDGE_COLOR);

        _fadedToMasterImage = createBackgroundImageProcessing();
        boundsOnTheFlyEdgeColor = BOUNDS_ON_THE_FLY_EDGE_COLOR;
        selectedTriangleEdgeColor = SELECTED_TRIANGLE_EDGE_COLOR;
    }

    protected Color getBoundsOnTheFlyEdgeColor() {
        return boundsOnTheFlyEdgeColor;
    }

    protected Color getSelectedTriangleEdgeColor() {
        return selectedTriangleEdgeColor;
    }

    protected BackgroundImageProcessing getFadedToMasterImage() {
        return _fadedToMasterImage;
    }

    protected DelaunaySubdivision getTriangulation() {
        return getIfNotNull(
                getIfNotNull(getMorphingVerticesManager(),
                        MorphingVerticesMapManager::getDelaunayController),
                DelaunayTriangulatorController::getTriangulation);
    }

    @Override
    public void paintComponent(Component comp, Graphics grp) {
        super.paintComponent(comp, grp);
    }

    @Override
    protected boolean hasToPaint() {
        return getTriangulation() != null;
    }

    @Override
    protected void paintComponentInternal(Component comp, Graphics grp) {
        super.paintComponentInternal(comp, grp);
    }

    @Override
    protected void prePaint(Graphics grp) {
        paintFadedMasterImage(grp, getPaintLimits(grp));
    }

    @Override
    protected void layerPrePaint(Graphics2D g2d, GeometricRectangle hundredPercentPaintLimits) {
    }

    @Override
    protected void layerPostPaint(Graphics2D g2d, GeometricRectangle hundredPercentPaintLimits) {
        Stroke solid = createSolidStroke();
        MultiTriangle selectedMultiTriangle = getSelectedMultiTriangle();

        paintSelectedTriangle(g2d, hundredPercentPaintLimits, selectedMultiTriangle, solid);
        paintSelectedTriangleVertexNumbers(g2d, selectedMultiTriangle);
    }

    protected void paintFadedMasterImage(Graphics grp, GeometricRectangle limits) {
        BufferedImage fadedMasterImage = getFadedMasterImage();

        if (fadedMasterImage != null)
            grp.drawImage(fadedMasterImage,
                    (int) limits.x, (int) limits.y,
                    (int) limits.x + (int) limits.width,
                    (int) limits.y + (int) limits.height,
                    (int) limits.x, (int) limits.y,
                    (int) limits.x + (int) limits.width,
                    (int) limits.y + (int) limits.height,
                    null);
    }

    protected BufferedImage getFadedMasterImage() {
        return (getFadedToMasterImage().getImage());
    }

    @Override
    protected void browseEdges(GeometricRectangle bounds, BiConsumer<Point2d, Point2d> edgeConsumer) {
        getBuilder().browseEdges(getTriangulation(), getBoundsForBrowsingEdges(), edgeConsumer);
    }

//    @Override
//    protected void resetWhenImageChanges() {
//        // intentionally left blank
//    }

    @Override
    protected GeometricRectangle getBoundsForBrowsingEdges() {
        return getMasterBounds();
    }

    @Override
    protected Color getEdgeColor() {
        return ((_onTheFlyBounds == null) ? super.getEdgeColor() : getBoundsOnTheFlyEdgeColor());
    }

    protected void paintSelectedTriangle(Graphics2D g2d, GeometricRectangle hundredPercentPaintLimits,
                                         MultiTriangle multiTriangle, Stroke solid) {
        if (multiTriangle != null) {
            String masterName = this.getMasterImageName();
            Point2d p1 = multiTriangle.getVertex1(masterName);
            Point2d p2 = multiTriangle.getVertex2(masterName);
            Point2d p3 = multiTriangle.getVertex3(masterName);

            if(noneIsNull(p1, p2, p3)) {
                g2d.setColor(SELECTED_TRIANGLE_EDGE_COLOR);
                drawEdge(g2d, hundredPercentPaintLimits, p1, p2, solid, solid);
                drawEdge(g2d, hundredPercentPaintLimits, p2, p3, solid, solid);
                drawEdge(g2d, hundredPercentPaintLimits, p3, p1, solid, solid);
            }
        }
    }

    protected void paintSelectedTriangleVertexNumbers(Graphics2D g2d, MultiTriangle multiTriangle) {
        if (multiTriangle != null) {
            String imageName = getCurrentImageName();
            Point2d p1 = multiTriangle.getVertex1(imageName);
            Point2d p2 = multiTriangle.getVertex2(imageName);
            Point2d p3 = multiTriangle.getVertex3(imageName);

            if(noneIsNull(p1, p2, p3)) {
                drawVertexNumberString(g2d, calculateVertexNumberStringPosition(p1, multiTriangle.getNormalizedBisector1(imageName)), "1");
                drawVertexNumberString(g2d, calculateVertexNumberStringPosition(p2, multiTriangle.getNormalizedBisector2(imageName)), "2");
                drawVertexNumberString(g2d, calculateVertexNumberStringPosition(p3, multiTriangle.getNormalizedBisector3(imageName)), "3");
            }
        }
    }

    protected Point calculateVertexNumberStringPosition(Point2d pt, Point2d normalBisector) {
        Point result = new Point();
        double factor = SQRT_2 * getTextFontSize() / 2;
        int vertexRadius = getVertexRadius();
        result.x = (int) (pt.x * getImageZoomFactor() - normalBisector.x * factor + vertexRadius);
        result.y = (int) (pt.y * getImageZoomFactor() - normalBisector.y * factor + vertexRadius);

        return (result);
    }

    @Override
    protected GeometricRectangle getEffectiveBounds() {
        return ((_onTheFlyBounds != null) ? _onTheFlyBounds : super.getEffectiveBounds());
    }

    @Override
    protected Point2d mapPoint(Point2d masterPt) {
        MorphingVerticesMap verticesMap = getVerticesMap();
        DeductibleDynamicMultiMapElementBase<VertexModel> multiPt =
                verticesMap.get(verticesMap.getMasterName(), masterPt);
        return ((multiPt == null) ?
                getPointInProgress() :  // if not found, it must be the point in progress, so we will return it
                getOrCalculateOnTheFlyBoundsPoint(multiPt.get(getCurrentImageName())));
    }


    protected Point2d getOrCalculateOnTheFlyBoundsPoint(VertexModel vm) {
        Point2d result = vm.getPoint();

        GeometricRectangle onTheFlyBounds = _onTheFlyBounds;
        if (onTheFlyBounds != null)
            result = fromNormalizedInt(vm.getNormalizedPoint(), onTheFlyBounds);

        return (result);
    }

    @Override
    public void updateTriangulation(String imageName) {//,
//                                 MorphingVerticesMapManager morphingVerticesManager) {
        _currentImageName = imageName;
        MorphingVerticesMapManager morphingVerticesManager = getMorphingVerticesManager();
        if (this.morphingVerticesManager != morphingVerticesManager)
            resetSelectedMultiTriangle();

        this.morphingVerticesManager = morphingVerticesManager;
        _onTheFlyBounds = null;
    }

    public void setNewOnTheFlyBounds(String imageName, GeometricRectangle bounds) {
        _onTheFlyBounds = bounds;
        if (bounds != null) {
            resetSelectedMultiTriangle();
            resetFadedToMasterImage();
        }
    }

    @Override
    public void setImageZoomFactor(double _zoomFactor) {
        if (_zoomFactor != getImageZoomFactor())
            resetFadedToMasterImage();

        super.setImageZoomFactor(_zoomFactor);
    }

    protected void setFadedToMasterImage(String imageName, BufferedImage image) {
        if (Objects.equals(getCurrentImageName(), imageName)) {
//			ExecutionFunctions.instance().safeMethodExecution( () -> ImageIO.write(image, "png", new File( "J:\\image.png" ) ) );
            repaint();
        } else
            resetFadedToMasterImage();
    }

    protected MorphingImageModel getImageModel(String imageName) {
        return getAppliCtx().getImageModel(imageName);
    }

    protected BufferedImage createFadedMasterImage(double fadingFraction, AtomicBoolean hasToCancel) {
        GeometricRectangle imageSize = getImageSize();
        MorphingImageModel masterImageModel = getImageModel(getMasterImageName());
        NormalTransitionMorphingFrameCreator morphingFrameCreator = new NormalTransitionMorphingFrameCreator(
                this::getMasterTriangles,
                getImageModel(getCurrentImageName()), masterImageModel, imageSize,
                getVerticesMap(), masterImageModel, getTriangulation());

        return (morphingFrameCreator.createFrame(fadingFraction, 0d,
                hasToCancel, imageSize, getZoomedBounds(imageSize), null, null, -1));//getClip() ) );
    }

    protected List<GeometricTriangle> getMasterTriangles() {
        List<GeometricTriangle> result = new ArrayList<>();
        DelaunayTriangleBrowser.instance().browseTriangles(getTriangulation(),
                getMasterBounds(),
                triangle -> addTriangle(result, triangle) );

        return result;
    }

    protected boolean addTriangle(List<GeometricTriangle> result, GeometricTriangle masterTriangle) {
        result.add(masterTriangle);
        return true;
    }

    public void resetFadedToMasterImage() {
        getFadedToMasterImage().reset();
    }

    protected void cancelFadedToMasterImage() {
        getFadedToMasterImage().cancel();
    }

    public void fadeToMaster(double fadingFraction) {
        if (fadingFraction > 0d) {
            String imageName = getCurrentImageName();
            getFadedToMasterImage().process(hasToCancel -> createFadedMasterImage(fadingFraction, hasToCancel),
                    image -> setFadedToMasterImage(imageName, image));
        } else {
            resetFadedToMasterImage();
            repaint();
        }
    }

    public void setSelectedMultiTriangle(MultiTriangle multiTriangle) {
        _selectedMultiTriangle = multiTriangle;
    }

    public MultiTriangle getSelectedMultiTriangle() {
        return _selectedMultiTriangle;
    }

    public void resetSelectedMultiTriangle() {
        setSelectedMultiTriangle(null);
    }
}
