/*
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.applications.libmorphing.view.painters.faceselector;

import com.frojasg1.applications.common.configuration.InternationalizedStringConf;
import com.frojasg1.applications.libmorphing.contexts.MorphingEditorApplicationContext;
import com.frojasg1.applications.libmorphing.api.facedetector.pojo.FaceDetectionFrame;
import com.frojasg1.applications.libmorphing.api.facedetector.pojo.FaceDetectionMeshData;
import com.frojasg1.applications.libmorphing.controller.MorphingVerticesMapManager;
import com.frojasg1.applications.libmorphing.model.facedetector.ImageFaceDetectionPersistenceModel;
import com.frojasg1.applications.libmorphing.model.images.image.MorphingImageModel;
import com.frojasg1.applications.libmorphing.model.images.vertex.VertexModel;
import com.frojasg1.applications.libmorphing.view.painters.DelaunayTriangulationJPanelControllerBase;
import com.frojasg1.applications.libmorphing.view.painters.TriangulationPainterLayerTypeEnum;
import com.frojasg1.applications.libmorphing.view.panels.evt.DelaunayJPanelsEventBuilderBase;
import com.frojasg1.applications.libmorphing.view.panels.evt.FaceDetectionSelectionEvent;
import com.frojasg1.applications.libmorphing.view.panels.evt.enums.FaceDetectionSelectionType;
import com.frojasg1.delaunay.triangulation.quadedge.builder.QuadEdgeDelaunayBuilder;
import com.frojasg1.libgeometry.types.Point2d;
import com.frojasg1.general.desktop.mouse.MouseFunctions;
import com.frojasg1.general.desktop.view.icons.IconResources;
import com.frojasg1.general.desktop.view.menus.BaseJPopupMenu;
import com.frojasg1.general.exceptions.IllegalCallException;
import com.frojasg1.general.view.evt.wrapper.ListOfGenComponentWrapperEventContextListenersBase;
import com.frojasg1.general.view.evt.wrapper.impl.GenComponentWrapperEvent;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.swing.*;
import java.awt.*;
import java.awt.event.MouseEvent;
import java.awt.event.MouseWheelEvent;
import java.util.List;
import java.util.Objects;
import java.util.function.Consumer;
import java.util.stream.Collectors;

/**
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 */
public class FaceDetectionSelectorJPanelController
        extends DelaunayTriangulationJPanelControllerBase {
    private static final Logger LOGGER = LoggerFactory.getLogger(FaceDetectionSelectorJPanelController.class);
    public static final String CONF_CANCEL_FACE_DETECTION_SELECTION = "CANCEL_FACE_DETECTION_SELECTION";
    protected static final int PROXIMITY_100_PERCENT_RADIUS = 500;

    protected FaceDetectionSelectorPainter painter;

    protected FaceDetectionMeshData faceDetectionMeshData;

    protected List<VertexModel> faceDetectionFrameCenterList;

    protected MorphingEditorApplicationContext appliCtx;

    public FaceDetectionSelectorJPanelController(MorphingEditorApplicationContext appliCtx,
                                                 Component comp, TriangulationPainterLayerTypeEnum layerType,
                                                 String imageName,
                                                 DelaunayJPanelsEventBuilderBase evtFactory,
                                                 ListOfGenComponentWrapperEventContextListenersBase<JPanel, GenComponentWrapperEvent> listenersList,
                                                 QuadEdgeDelaunayBuilder builder, InternationalizedStringConf internationalStrings,
                                                 FaceDetectionSelectorPainter painter) {
        super(comp, layerType,
                imageName, PROXIMITY_100_PERCENT_RADIUS, evtFactory, listenersList, builder, internationalStrings);
        this.appliCtx = appliCtx;
        this.painter = painter;
    }

    public FaceDetectionSelectorJPanelController init() {
        super.init();

        return this;
    }

    protected MorphingEditorApplicationContext getAppliCtx() {
        return appliCtx;
    }

    protected InternationalizedStringConf getInternationalStrings() {
        return internationalStrings;
    }

    protected FaceDetectionMeshData getFaceDetectionMeshData() {
        return faceDetectionMeshData;
    }

    protected void setFaceDetectionMeshData(FaceDetectionMeshData faceDetectionMeshData) {
//        getPainter().setFaceDetectionMeshData(faceDetectionMeshData);
        this.faceDetectionMeshData = faceDetectionMeshData;
    }

    public FaceDetectionSelectorPainter getPainter() {
        return painter;
    }

    protected String getImageFileName(String imageName) {
        return getIfNotNull(getImageModel(getMorphingVerticesMapManager(), imageName), MorphingImageModel::getFileName);
    }

    protected List<FaceDetectionMeshData> getFaceDetectionMeshes() {
        return getFaceDetectionMeshes(getImageName());
    }

    protected List<FaceDetectionMeshData> getFaceDetectionMeshes(String imageName) {
        return getIfNotNull(
                getIfNotNull(
                        getAppliCtx().getImageFaceDetectionPersistenceModelMap(),
                        map -> map.get(getImageFileName(imageName))),
                ImageFaceDetectionPersistenceModel::getDetectedFaceMeshes);
    }

    protected Point2d createFrameCenter(FaceDetectionMeshData faceDetectionMeshData) {
        return getIfNotNull(
                getIfNotNull(faceDetectionMeshData, FaceDetectionMeshData::getFaceDetectionFrame),
                FaceDetectionFrame::getCenter);
    }

    protected VertexModel createVertexModel(Point2d pt) {
        return getIfNotNull(
                getIfNotNull(getMorphingVerticesMapManager(), MorphingVerticesMapManager::getVerticesMap),
                map -> map.createElementValue(pt));
    }

    protected List<VertexModel> createVertexList(String imageName) {
        List<VertexModel> result = null;
        List<FaceDetectionMeshData> faceDetectionMeshDataList = getFaceDetectionMeshes(imageName);
        if (faceDetectionMeshDataList != null) {
            result = faceDetectionMeshDataList.stream()
                    .map(this::createFrameCenter)
                    .map(this::createVertexModel)
                    .collect(Collectors.toList());
        }
        return result;
    }

    protected List<VertexModel> getFaceDetectionFrameCenterList() {
        return faceDetectionFrameCenterList;
    }

    protected void setFaceDetectionFrameCenterList(List<VertexModel> faceDetectionFrameCenterList) {
        this.faceDetectionFrameCenterList = faceDetectionFrameCenterList;
    }

    protected List<VertexModel> getVertices() {
        if (getFaceDetectionFrameCenterList() == null) {
            setFaceDetectionFrameCenterList(createVertexList(getImageName()));
        }
        return getFaceDetectionFrameCenterList();
    }

    @Override
    protected void browseVertices(Consumer<VertexModel> vertexProcessor) {
        List<VertexModel> list = getVertices();
        if(list != null) {
            for(VertexModel vm: list) {
                vertexProcessor.accept(vm);
            }
        }
    }

    protected FaceDetectionMeshData getFaceDetectionMeshData(Point2d faceDetectionFrameCenter) {
        return getControl88meshCreator().getFaceDetectionMeshData(
                faceDetectionFrameCenter,
                getFaceDetectionMeshes());
    }

    protected void updateFaceDetectionMeshData(Point2d imagePoint, Point2d faceDetectionFrameCenter) {
        FaceDetectionMeshData faceDetectionMeshData1 = getFaceDetectionMeshData(faceDetectionFrameCenter);
        if (!isInside(imagePoint, faceDetectionMeshData1)) {
            faceDetectionMeshData1 = null;
        } else {
            LOGGER.debug("mouse ({}) inside face detection frame {}", imagePoint, faceDetectionMeshData1);
        }
        if(!Objects.equals(getFaceDetectionMeshData(), faceDetectionMeshData1)) {
            notifyFaceDetection(FaceDetectionSelectionType.APPROACHING, faceDetectionMeshData1);
            setFaceDetectionMeshData(faceDetectionMeshData1);
        }
    }

    protected FaceDetectionFrame getFaceDetectionFrame(FaceDetectionMeshData faceDetectionMeshData1) {
        return getIfNotNull(faceDetectionMeshData1, FaceDetectionMeshData::getFaceDetectionFrame);
    }

    protected boolean isInside(Point2d imagePoint, FaceDetectionMeshData faceDetectionMeshData1) {
        boolean result = false;
        FaceDetectionFrame frame = getFaceDetectionFrame(faceDetectionMeshData1);
        if (frame != null) {
            result = frame.isStrictlyInside(imagePoint);
        }
        return result;
    }

    @Override
    protected boolean hasToDiscardForClosest(Point2d frameCenterPt, Point2d imagePt) {
        return false;
//        return !getControlMeshManager().canMoveControlPoint(pt, getImageName());
    }

    @Override
    public void mouseMovedInternal(MouseEvent e) {
        Point2d imagePoint = createPoint2d(getImagePoint(e));
        Point2d result = getClosest(imagePoint);

//        if (result != null) {
            updateFaceDetectionMeshData(imagePoint, result);
//        }
    }

    @Override
    public void mouseExitedInternal(MouseEvent e) {
    }

    @Override
    public void mousePressedInternal(MouseEvent e) {
        rightClickPressed = false;
        if(MouseFunctions.isMainButtonPressed(e)) {
            notifyFaceDetection(FaceDetectionSelectionType.SELECTED, getFaceDetectionMeshData());
        } else if(MouseFunctions.isRightButtonPressed(e)) {
            rightClickPressed = true;
        }
    }

    @Override
    public void mouseReleasedInternal(MouseEvent e) {
        if(rightClickPressed) {
            doPopup(e);
        }
        rightClickPressed = false;
    }

    @Override
    public void mouseDraggedInternal(MouseEvent e) {
    }

    @Override
    public void mouseWheelMovedInternal(MouseWheelEvent e) {
    }

    @Override
    protected void notifyClosest( Point2d pt )
    {
        throw new IllegalCallException("Face detections cannot move points");
//        _closestVertex = pt;
//        notifyNewPoint( null, pt, false, VertexEventType.APPROACHING );
    }

    @Override
    protected void notifyNewMovingPoint( Point2d originalPoint, Point2d newPoint )
    {
        throw new IllegalCallException("Face detections cannot move points");
////		Point2d newPoint = createPoint2d( pt );
//        notifyNewPoint( originalPoint, newPoint, false, VertexEventType.MOVING );
    }

    @Override
    protected void endMovingPoint()
    {
        throw new IllegalCallException("Face detections cannot move points");
//        notifyNewPoint(null, true, VertexEventType.MOVING);
    }

    protected BaseJPopupMenu createBasePopupMenu()
    {
        MyPopupMenu result = new MyPopupMenu();
        result.init();

        return( result );
    }

    @Override
    public void setTriangulation(String imageName,
                                 MorphingVerticesMapManager morphingVerticesManager) {
        this.morphingVerticesManager = morphingVerticesManager;

        setImageName(imageName);
        setFaceDetectionMeshData(null);
        setFaceDetectionFrameCenterList(null);
    }

    protected FaceDetectionSelectionEvent createFaceDetectionSelectionEvent(FaceDetectionSelectionType type,
                                                                            FaceDetectionMeshData faceDetectionMeshData) {
        return (getEvtFactory().createFaceDetectionSelectionEvent(type, getImageName(), faceDetectionMeshData));
    }

    protected void notifyFaceDetection(FaceDetectionSelectionType type, FaceDetectionMeshData faceDetectionMeshData) {
        notifyEvt(createFaceDetectionSelectionEvent(type, faceDetectionMeshData));
    }

    protected class MyPopupMenu extends MorphingPanelControllerPopupMenuBase
    {
        public MyPopupMenu()
        {
            super();
        }

        public void init()
        {
            super.init();

            createFixedPart();
            createCommonPart();
        }

        protected void createFixedPart()
        {
            addMenuComponent( CONF_CANCEL_FACE_DETECTION_SELECTION, new JMenuItem(),
                    () -> getInternationalStrings().getInternationalString(CONF_CANCEL_FACE_DETECTION_SELECTION),
                    () -> notifyFaceDetection(FaceDetectionSelectionType.SELECTED, null),
                    IconResources.CANCEL_ICON_RESOURCE);

            add( new Separator() );
        }

        @Override
        protected void preparePopupMenuItems() {
            super.preparePopupMenuItems();
        }
    }
}
