/*
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.applications.libmorphing.view.painters.faceselector;

import com.frojasg1.applications.libmorphing.contexts.MorphingEditorApplicationContext;
import com.frojasg1.applications.libmorphing.api.facedetector.helpers.ControlMeshBrowser;
import com.frojasg1.applications.libmorphing.api.facedetector.pojo.FaceDetectionMeshData;
import com.frojasg1.applications.libmorphing.model.images.vertex.map.impl.control.IndexedVerticesMap;
import com.frojasg1.applications.libmorphing.view.painters.DelaunayTriangulationPainterBase;
import com.frojasg1.delaunay.triangulation.quadedge.builder.QuadEdgeDelaunayBuilder;
import com.frojasg1.libgeometry.types.GeometricRectangle;
import com.frojasg1.libgeometry.types.Point2d;
import com.frojasg1.general.desktop.view.colors.Colors;

import java.awt.*;
import java.util.Objects;
import java.util.function.BiConsumer;
import java.util.List;

/**
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 */
public class FaceDetectionSelectorPainter extends DelaunayTriangulationPainterBase {
    protected static final int TEXT_SIZE_100_PERCENT = 16;

    protected static final double EDGE_WIDTH_100_PERCENT = 2d;
    protected static final int VERTEX_RADIO_100_PERCENT = 5;

    protected static final Color ALL_VERTEX_COLOR = Colors.COOL_BLUE;
    protected static final Color APPROACHING_VERTEX_COLOR = Colors.BUTAN;
    protected static final Color EDGE_COLOR = Colors.BROWN_ROSY_BROWN;

    protected FaceDetectionMeshData faceDetectionMeshData;
//    protected List<FaceDetectionMeshData> allFaces;

    /**
     * Creates new form GraphicalEqualizerJPanel
     */
    public FaceDetectionSelectorPainter(MorphingEditorApplicationContext appliCtx,
                                        Component parent,
                                        QuadEdgeDelaunayBuilder delaunayBuilder) {
        super(appliCtx, parent, delaunayBuilder, TEXT_SIZE_100_PERCENT, EDGE_WIDTH_100_PERCENT, VERTEX_RADIO_100_PERCENT,
                APPROACHING_VERTEX_COLOR, null, null, EDGE_COLOR);
    }

    protected boolean imageNameChanges(String imageName) {
        return Objects.equals(getCurrentImageName(), imageName);
    }

    public void setFaceDetectionMeshData(String imageName, FaceDetectionMeshData faceDetectionMeshData) {
        this.faceDetectionMeshData = faceDetectionMeshData;
        if(imageNameChanges(imageName)) {
            updateAllFaces(imageName);
        }
        _currentImageName = imageName;
    }

    protected int[][] getTriangles() {
        return getIfNotNull(faceDetectionMeshData, FaceDetectionMeshData::getTriangles);
    }

    protected IndexedVerticesMap getIdealControlMeshVerticesMap() {
        return getIfNotNull(faceDetectionMeshData, FaceDetectionMeshData::getLandmark80points);
    }

    protected String getName() {
        return getIfNotNull(faceDetectionMeshData, FaceDetectionMeshData::getName);
    }

    @Override
    public void paintComponent(Component comp, Graphics grp) {
        super.paintComponent(comp, grp);
    }

    @Override
    protected boolean hasToPaint() {
        return true;
//        return thereIsMesh();
    }

    protected boolean thereIsMesh() {
        return getTriangles() != null;
    }

    @Override
    protected void paintComponentInternal(Component comp, Graphics grp) {
        super.paintComponentInternal(comp, grp);
    }

    @Override
    protected void prePaint(Graphics grp) {

    }

    @Override
    protected void layerPrePaint(Graphics2D g2d, GeometricRectangle hundredPercentPaintLimits) {
        List<FaceDetectionMeshData> meshes = getAllFaces();
        if(meshes != null) {
            for(FaceDetectionMeshData mesh: meshes) {
                paintMesh(g2d, mesh, ALL_VERTEX_COLOR, hundredPercentPaintLimits);
            }
        }
    }

    protected void paintMesh(Graphics2D g2d, FaceDetectionMeshData mesh, Color color,
                             GeometricRectangle hundredPercentPaintLimits) {
        for(Point2d pt: mesh.getModifiedFaceDetection().getPoints()) {
            paintVertex(g2d, hundredPercentPaintLimits, pt, getAppliZoomedValue(VERTEX_RADIO_100_PERCENT), color);
        }
    }

    @Override
    protected void layerPostPaint(Graphics2D g2d, GeometricRectangle hundredPercentPaintLimits) {
    }

    protected ControlMeshBrowser getControlMeshBrowser() {
        return ControlMeshBrowser.instance();
    }

    protected void browseEdges(int[][] triangles, IndexedVerticesMap controlMeshVerticesMap, String name,
                               GeometricRectangle bounds, BiConsumer<Point2d, Point2d> edgeConsumer) {
        getControlMeshBrowser().browseEdges(triangles, controlMeshVerticesMap, name, bounds, edgeConsumer);
    }

    @Override
    protected void browseEdges(GeometricRectangle bounds, BiConsumer<Point2d, Point2d> edgeConsumer) {
        browseEdges(getTriangles(), getIdealControlMeshVerticesMap(), getName(), getBoundsForBrowsingEdges(), edgeConsumer);
    }

//    @Override
//    protected void resetWhenImageChanges() {
//        setFaceDetectionMeshData(null, null);
//        setAllFaces(null, null);
//    }

    @Override
    protected GeometricRectangle getBoundsForBrowsingEdges() {
        return getBounds();
    }

    // with that painter, we do not do any mapping, the point is directly the one of the image.
    @Override
    protected Point2d mapPoint(Point2d imagePt) {
        return imagePt;
    }

    public void setAllFaces(String imageName, List<FaceDetectionMeshData> allFaces) {
        this._currentImageName = imageName;
//        this.allFaces = allFaces;
    }

//    public List<FaceDetectionMeshData> getAllFaces() {
//        return allFaces;
//    }
    public List<FaceDetectionMeshData> getAllFaces() {
        return getAppliCtx().getVerticesManager().getAllFaces(getCurrentImageName());
    }

    protected List<FaceDetectionMeshData> getAllFaces(String imageName) {
        return getIfNotNull(getMorphingVerticesManager(), manager -> manager.getAllFaces(imageName));
    }

    protected void updateAllFaces(String imageName) {
        setAllFaces(imageName, getAllFaces(imageName));
    }

    @Override
    public void updateTriangulation(String imageName){//,
//                                 MorphingVerticesMapManagerBase morphingVerticesManager) {
        super.updateTriangulation(imageName);//, morphingVerticesManager);

        updateAllFaces(imageName);
    }
}
