/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.applications.libmorphing.view.panels;

import java.awt.*;
import com.frojasg1.applications.common.components.resizecomp.MapResizeRelocateComponentItem;
import com.frojasg1.applications.common.components.resizecomp.ResizeRelocateItem;
import com.frojasg1.applications.libmorphing.contexts.MorphingEditorApplicationContext;
import com.frojasg1.applications.libmorphing.api.facedetector.pojo.FaceDetectionMeshData;
import com.frojasg1.applications.libmorphing.configuration.MorphingApplicationConfiguration;
import com.frojasg1.applications.libmorphing.model.images.color.types.MultiTriangle;
import com.frojasg1.applications.libmorphing.mvc.DelaunayTriangulatorView;
import com.frojasg1.applications.libmorphing.view.painters.TriangulationPainterLayerTypeEnum;
import com.frojasg1.applications.libmorphing.view.painters.WholeDelaunayTriangulationPanelWrapper;
import com.frojasg1.applications.libmorphing.view.panels.evt.DelaunayJPanelsEventBuilderBase;
import com.frojasg1.delaunay.triangulation.quadedge.builder.QuadEdgeDelaunayBuilder;
import com.frojasg1.delaunay.triangulation.quadedge.builder.impl.QuadEdgeDelaunayBuilderImpl;
import com.frojasg1.libgeometry.types.GeometricRectangle;
import com.frojasg1.libgeometry.types.Point2d;
import com.frojasg1.general.desktop.view.FrameworkComponentFunctions;
import com.frojasg1.general.desktop.view.color.ColorInversor;
import com.frojasg1.general.view.evt.wrapper.ListOfGenComponentWrapperEventContextListenersBase;
import com.frojasg1.general.view.evt.wrapper.impl.GenComponentWrapperEvent;
import com.frojasg1.general.desktop.view.panels.evt.CustomJPanelBaseWithJComponentEvt;
import com.frojasg1.general.desktop.view.pdf.ImageJPanel;
import com.frojasg1.general.desktop.view.zoom.mapper.ComponentMapper;
import com.frojasg1.general.desktop.view.zoom.mapper.ComposedComponent;
import com.frojasg1.general.number.DoubleReference;
import com.frojasg1.general.view.evt.wrapper.GenericWrapperEvent;
import java.awt.Dimension;
import java.awt.Rectangle;
import java.awt.image.BufferedImage;
import java.io.File;
import java.util.LinkedList;
import java.util.List;
import java.util.function.Consumer;
import javax.imageio.ImageIO;
import javax.swing.JPanel;
import javax.swing.JScrollPane;

/**
 *
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 */
public class DelaunayContentPanel extends CustomJPanelBaseWithJComponentEvt
	implements ComposedComponent, DelaunayTriangulatorView//,	InternationalizedStringConf
{
//	public static final String GLOBAL_CONF_FILE_NAME = "PdfContentPanel.properties";

//	protected static final String CONF_ERROR_WHEN_GETTING_PDF_PAGE = "ERROR_WHEN_GETTING_PDF_PAGE";


//	protected InternationalizedStringConfImp _internationalizedStringConf = null;

	protected MorphingEditorApplicationContext appliCtx;

	protected ImageJPanel _imagePanel = null;

	protected int _totalNumberOfPages = 0;
	protected int _currentPageIndex = 0;

	protected DoubleReference _pdfZoomFactor = null;

	protected MapResizeRelocateComponentItem _resizeRelocateInfo = null;

	protected WholeDelaunayTriangulationPanelWrapper _imagePanelWrapper;

	public DelaunayContentPanel(MorphingEditorApplicationContext appliCtx)
	{
		this(appliCtx, true);
	}

	/**
	 * Creates new form PdfContentPanel
	 */
	protected DelaunayContentPanel( MorphingEditorApplicationContext appliCtx,
									boolean init )
	{
		super( appliCtx.getAppliConf(), false, false );
//		_internationalizedStringConf = new InternationalizedStringConfImp( GLOBAL_CONF_FILE_NAME,
//								LibMorphingLanguageResources.sa_PROPERTIES_PATH_IN_JAR );

//		registerInternationalizedStrings();
		this.appliCtx = appliCtx;

		if( init )
			init();
	}

	@Override
	protected void init()
	{
		super.init();

		initComponents();

		initZoomFactor();
		initPanel();

		setWindowConfiguration();
	}

	protected MorphingEditorApplicationContext getAppliCtx() {
		return appliCtx;
	}

	@Override
	protected DelaunayJPanelsEventBuilderBase createEventBuilder()
	{
		// can be overriden by derived classes.
		return( new DelaunayJPanelsEventBuilderBase(this) );
	}

	@Override
	public DelaunayJPanelsEventBuilderBase getEventBuilder() {
		return (DelaunayJPanelsEventBuilderBase) super.getEventBuilder();
	}

	@Override
	public ListOfGenComponentWrapperEventContextListenersBase<JPanel, GenComponentWrapperEvent>
		getListenersList()
	{
		return( super.getListenersList() );
	}

	protected void initZoomFactor()
	{
		_pdfZoomFactor = new DoubleReference( 1.0D );
	}

	protected JScrollPane getImageScrollPane()
	{
		return( jSP_imageScrollPane );
	}

	protected ImageJPanel createImageJPanel()
	{
		ImageJPanel result = new ImageJPanel( getImageScrollPane(), null );
		result.setCanInvertImageColors(false);
		result.setAllowShiftForMouseWheel(false);

		return( result );
	}

	protected void initPanel()
	{
		try
		{
			_imagePanel = createImageJPanel();
			jSP_imageScrollPane.setViewportView(_imagePanel );

			_imagePanelWrapper = createImagePanelWrapper(_imagePanel);
		}
		catch( Throwable th )
		{
			th.printStackTrace();
		}
	}

	protected QuadEdgeDelaunayBuilder getBuilder()
	{
		return( QuadEdgeDelaunayBuilderImpl.instance() );
	}

	protected WholeDelaunayTriangulationPanelWrapper createImagePanelWrapper(ImageJPanel imagePanel)
	{
		return( new WholeDelaunayTriangulationPanelWrapper(getAppliCtx(), imagePanel,// getAppliConf(),
													getBuilder(),
													getListenersList(),
													getEventBuilder() )
				.init() );
	}

	protected MorphingApplicationConfiguration getAppliConf()
	{
		return (MorphingApplicationConfiguration) super.getAppliConf();
	}

	public WholeDelaunayTriangulationPanelWrapper getImagePanelWrapper() {
		return _imagePanelWrapper;
	}

	public ImageJPanel getImagePanel()
	{
		return( _imagePanel );
	}

	protected void setWindowConfiguration( )
	{
		_resizeRelocateInfo = new MapResizeRelocateComponentItem();
		MapResizeRelocateComponentItem mapRRCI = _resizeRelocateInfo;
		try
		{
			mapRRCI.putResizeRelocateComponentItem( jSP_imageScrollPane, ResizeRelocateItem.FILL_WHOLE_PARENT );
			mapRRCI.putAll( getImagePanel().getResizeRelocateInfo() );
		}
		catch( Throwable th )
		{
			th.printStackTrace();
		}

//		ExecutionFunctions.instance().safeMethodExecution( () -> registerInternationalizedStrings() );
	}

	@Override
	public MapResizeRelocateComponentItem getResizeRelocateInfo()
	{
		return( _resizeRelocateInfo );
	}

	public void setImage( String fileName )
	{
		try
		{
			File file = new File( fileName );
			setPage( ImageIO.read( file ) );
		}
		catch( Throwable th )
		{
			th.printStackTrace();
		}
	}

	/**
	 * This method is called from within the constructor to initialize the form.
	 * WARNING: Do NOT modify this code. The content of this method is always
	 * regenerated by the Form Editor.
	 */
	@SuppressWarnings("unchecked")
    // <editor-fold defaultstate="collapsed" desc="Generated Code">//GEN-BEGIN:initComponents
    // Generated using JFormDesigner non-commercial license
    private void initComponents() {
        jSP_imageScrollPane = new JScrollPane();

        //======== this ========
        setLayout(null);
        add(jSP_imageScrollPane);
        jSP_imageScrollPane.setBounds(0, 0, 550, 670);

        {
            // compute preferred size
            Dimension preferredSize = new Dimension();
            for(int i = 0; i < getComponentCount(); i++) {
                Rectangle bounds = getComponent(i).getBounds();
                preferredSize.width = Math.max(bounds.x + bounds.width, preferredSize.width);
                preferredSize.height = Math.max(bounds.y + bounds.height, preferredSize.height);
            }
            Insets insets = getInsets();
            preferredSize.width += insets.right;
            preferredSize.height += insets.bottom;
            setMinimumSize(preferredSize);
            setPreferredSize(preferredSize);
        }
    }// </editor-fold>//GEN-END:initComponents


    // Variables declaration - do not modify//GEN-BEGIN:variables
    // Generated using JFormDesigner non-commercial license
    private JScrollPane jSP_imageScrollPane;
    // End of variables declaration//GEN-END:variables

	public LinkedList<DoubleReference> getStepsForFactorList()
	{
		return( _imagePanel.getStepsForFactorList() );
	}
/*
	protected DoubleReference getSelectedZoomFactor()
	{
        DoubleReference newFactor = null;
        Object obj = jCBZoomFactor.getSelectedItem();
        if( obj instanceof DoubleReference )
        {
            newFactor = (DoubleReference) obj;
        }
		return( newFactor );
	}

	protected double getSelectedZoomFactorDouble()
	{
		double factor = 1.0D;
		
		DoubleReference dr = getSelectedZoomFactor();
		if( dr != null )
			factor = dr._value;
		
		return( factor );
	}
*/	
/*
	public void setNumberOfPages( int numberOfPages )
	{
		_totalNumberOfPages = numberOfPages;
		_currentPageIndex = 0;

		updatePage( 0, _pdfZoomFactor );
	}

	public void setNewPDF( PdfDocumentWrapper document )
	{
		setDocument( document );

		setNumberOfPages( _pdfDocument.getNumberOfPages() );
	}
*/
	protected boolean isDarkMode()
	{
		return( FrameworkComponentFunctions.instance().isDarkMode(this) );
	}
/*
	protected BufferedImage invertPageColors( BufferedImage image, double zoomFactor )
	{
		return( _pdfPageColorInversor.invertPdfPageColors( getColorInversor(), image,
					zoomFactor,	getImagesOfPage(), getGlyphsOfPage() )	);
	}
*/

	protected void setPage( BufferedImage image, DoubleReference factor, int pageIndex )
	{
		ImageJPanel.InitialVerticalPosition ivp = null;
/*
		if( isDarkMode() )
			image = invertPageColors( image, factor._value );


		if( _currentPageIndex < pageIndex )
			ivp = ImageJPanel.InitialVerticalPosition.TOP;
		else if( _currentPageIndex > pageIndex )
			ivp = ImageJPanel.InitialVerticalPosition.BOTTOM;
		else
			ivp = ImageJPanel.InitialVerticalPosition.NOTHING;
*/
		_imagePanel.setImage( image, factor, ivp );
/*
		_currentPageIndex = pageIndex;

		_controller.newPageSet(image, factor, pageIndex);
*/
	}

	public DoubleReference getMostSuitableFactor( DoubleReference factor )
	{
		return( _imagePanel.getMostSuitableFactor( factor ) );
	}

	protected double getSelectedZoomFactorDouble(DoubleReference dr)
	{
		double factor = 1.0D;
		
		if( dr != null )
			factor = dr._value;
		
		return( factor );
	}

	protected void setPage( BufferedImage image )
	{
		DoubleReference factor = new DoubleReference( 1.0D );
		setPage(image, factor, _currentPageIndex );
	}
/*
	public void setPageIndex( int pageIndex )
	{
		_currentPageIndex = pageIndex;
	}

	protected void setPage( int pageIndex, DoubleReference zoomFactor )
	{
		BufferedImage image = _pdfOwner.getPage( pageIndex, getSelectedZoomFactorDouble(zoomFactor) );
		if( image != null )
			setPage( image, zoomFactor, pageIndex );		
	}

	public void updatePage( int pageIndex, DoubleReference zoomFactor )
	{
////		if( isVisible() )
		{
			_currentPageIndex = pageIndex;
			setPage(_currentPageIndex, zoomFactor );
////			updateCurrentPageTexts();
		}
	}
*/
	public void setImageZoomFactor( DoubleReference pdfZoomFactor )
	{
//		updatePage( _currentPageIndex, _pdfZoomFactor );
		this.getImagePanelWrapper().setImageZoomFactor( pdfZoomFactor );
	}

	public DoubleReference getZoomFactor()
	{
		return( _imagePanel.getZoomFactor() );
	}

	public int getStartPage()
	{
		return( 0 );
	}

	public int getEndPage()
	{
		return( _totalNumberOfPages - 1 );
	}

	public void setNewImageZoomFactor( DoubleReference factor )
	{
		if( factor != null )
			_pdfZoomFactor = getMostSuitableFactor( factor );

		getImagePanelWrapper().setImageZoomFactor( factor );
	}

	@Override
	public void setComponentMapper(ComponentMapper mapper)
	{
		jSP_imageScrollPane = mapper.mapComponent( jSP_imageScrollPane );
		_imagePanel = mapper.mapComponent(_imagePanel );

		super.setComponentMapper(mapper);
	}

	@Override
	public Dimension getInternalSize()
	{
		return( jSP_imageScrollPane.getSize() );
	}
/*
	@Override
	public void changeLanguage(String language) throws Exception {
		throw new UnsupportedOperationException("Not supported yet."); //To change body of generated methods, choose Tools | Templates.
	}

	@Override
	public void registerInternationalString(String label, String value)
	{
		_internationalizedStringConf.registerInternationalString(label, value);
	}

	@Override
	public String getInternationalString(String label)
	{
		return( _internationalizedStringConf.getInternationalString(label ) );
	}

	@Override
	public String createCustomInternationalString(String label, Object... args) {
		return( _internationalizedStringConf.createCustomInternationalString( label, args ) );
	}

	protected void registerInternationalizedStrings()
	{
		registerInternationalString( CONF_ERROR_WHEN_GETTING_PDF_PAGE, "Error when getting pdf page" );
	}
*/
	@Override
	public Rectangle getInternalBounds()
	{
		return( jSP_imageScrollPane.getBounds() );
	}

	@Override
	protected void invertColorsChild(ColorInversor colorInversor)
	{
//		updatePage( getCurrentPage() );
	}

	@Override
	public void setPointInProgress(String imageName, Point2d pt) {
		getImagePanelWrapper().setPointInProgress( imageName, pt );
	}

	@Override
	public void setClosestPoint(String imageName, Point2d pt) {
		getImagePanelWrapper().setClosestPoint( imageName, pt );
	}

	@Override
	public void setControlMeshClosestPoint(String imageName, Point2d pt) {
		getImagePanelWrapper().setControlMeshClosestPoint( imageName, pt );
	}

	@Override
	public void updateTriangulation(String imageName,
//									MorphingVerticesMapManager morphingVerticesManager,
									Point2d normalizedCenterPoint,
									boolean madatoryUpdate,
									Double newImageZoomFactor)
	{
		getImagePanelWrapper().setTriangulation(imageName,// morphingVerticesManager,
												normalizedCenterPoint,
												newImageZoomFactor);
	}

	@Override
	public double getImageZoomFactor()
	{
		return( getImagePanelWrapper().getImageZoomFactor() );
	}

	@Override
	public Point2d getNormalizedPointImageCenterShown()
	{
		return( getImagePanelWrapper().getNormalizedPointImageCenterShown() );
	}	

	@Override
	public void setSelectedMultiTriangle( MultiTriangle multiTriangle )
	{
		getImagePanelWrapper().setSelectedMultiTriangle(multiTriangle);
	}

	@Override
	public MultiTriangle getSelectedMultiTriangle()
	{
		return( getImagePanelWrapper().getSelectedMultiTriangle() );
	}

	@Override
	public void resetSelectedMultiTriangle()
	{
		getImagePanelWrapper().resetSelectedMultiTriangle();
	}

	@Override
	public void setNewOnTheFlyBounds( String imageName, GeometricRectangle bounds )
	{
		getImagePanelWrapper().setNewOnTheFlyBounds( imageName, bounds );
	}

	@Override
	public void addDelaunayTriangulatorViewListener(Consumer<? super GenericWrapperEvent> listener, Long... eventTypes) {
		addWrapperListener(listener, eventTypes);
	}

	@Override
	public void removeDelaunayTriangulatorViewListener(Consumer<? super GenericWrapperEvent> listener) {
		removeWrapperListener(listener);
	}

	@Override
	public void fadeToMaster(double fadingFraction)
	{
		getImagePanelWrapper().fadeToMaster(fadingFraction);
	}

	@Override
	public void resetFadedToMasterImage()
	{
		getImagePanelWrapper().resetFadedToMasterImage();
	}

	@Override
	public double getFadingFraction() {
		throw new UnsupportedOperationException("Not supported yet."); //To change body of generated methods, choose Tools | Templates.
	}

	@Override
	public void setAllFaces(String imageName, List<FaceDetectionMeshData> allFaces) {
		getImagePanelWrapper().setAllFaces(imageName, allFaces);
	}

	@Override
	public void setLayerType(String imageName, TriangulationPainterLayerTypeEnum layerType) {
		getImagePanelWrapper().activateOnlyLayer(layerType);
	}

	@Override
	public void setFaceDetectionMeshData(String imageName, FaceDetectionMeshData faceDetectionMeshData) {
		getImagePanelWrapper().setFaceDetectionMeshData(imageName, faceDetectionMeshData);
	}
}
