/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.applications.libmorphing.view.timers;

import com.frojasg1.general.threads.LockAndConditionOwner;
import java.awt.image.BufferedImage;
import java.util.LinkedList;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.function.Consumer;
import javax.swing.Timer;

/**
 *
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 */
public class SwingImageSequenceShower extends LockAndConditionOwner {
	protected static int MAX_NUM_OF_STORED_ELEMS = 25;

	protected static final int NUM_CONDITIONS = 1;
	protected static final int WAITING_FOR_ROOM = 0;


	protected Timer _timer;
	protected LinkedList<BufferedImage> _list;
	protected Consumer<BufferedImage> _consumer;

	protected int _msBetweenFrames;


	protected boolean _hasToStop = false;

	protected AtomicInteger count = new AtomicInteger(0);

	public SwingImageSequenceShower(Consumer<BufferedImage> consumer,
									int msBetweenFrames)
	{
		_list = new LinkedList<>();
		_msBetweenFrames = msBetweenFrames;
		_consumer = consumer;
		_timer = new Timer( msBetweenFrames, evt -> showImage() );
		_timer.start();

		init();
	}

	protected void showImage()
	{
		this.lockProcedure( this::showImageInternal );
	}

	protected void showImageInternal()
	{
		if( !_hasToStop && _list.isEmpty() )
		{
			_timer.setDelay( 15 );
			_timer.start();
		}
		else if( ! _hasToStop && ! _list.isEmpty() )
		{
			_consumer.accept( _list.removeFirst() );
			_timer.setDelay( _msBetweenFrames );
			_timer.start();

			ifHasWaitersSignal(WAITING_FOR_ROOM);
		}
	}

	protected boolean canAdd()
	{
		return( _list.size() < MAX_NUM_OF_STORED_ELEMS );
	}

	public void addImage( BufferedImage image )
	{
		this.lockProcedure( () -> addImageInternal( image ) );
	}

	protected void addImageInternal( BufferedImage image )
	{
		while( ! _hasToStop && ! canAdd() )
		{
			try
			{
				awaitInterrupted( WAITING_FOR_ROOM, 100 );
			}
			catch( Exception ex )
			{
				_hasToStop = true;
				break;
			}
		}
		if( ! _hasToStop && canAdd() )
			_list.add( image );
	}

	public void hasToStop()
	{
		this.lockProcedure( this::hasToStopInternal );
	}

	protected void hasToStopInternal()
	{
		_hasToStop = true;

		ifHasWaitersSignal(WAITING_FOR_ROOM);

		_timer.stop();
	}

	@Override
	protected int getNumConditionsForLock() {
		return( NUM_CONDITIONS );
	}
}
