/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.applications.libmorphing.xml.facedetector.loader;

import com.frojasg1.applications.libmorphing.api.facedetector.pojo.FaceDetectionMeshData;
import com.frojasg1.applications.libmorphing.model.facedetector.ImageFaceDetectionPersistenceModel;
import com.frojasg1.applications.libmorphing.model.facedetector.ImageFaceDetectionPersistenceModelKey;
import com.frojasg1.applications.libmorphing.model.facedetector.ImageFaceDetectionPersistenceModelMap;
import com.frojasg1.applications.libmorphing.xml.facedetector.FaceDetectionWholeFileModel;
import com.frojasg1.libgeometry.types.GeometricRectangle;
import com.frojasg1.libgeometry.types.Point2d;
import com.frojasg1.general.xml.XmlElement;
import com.frojasg1.general.xml.persistency.loader.impl.ModelToXmlBase;

import java.util.List;
import java.util.Map;

/**
 *
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 */
public class FaceDetectionWholeFileModelToXml extends ModelToXmlBase<FaceDetectionWholeFileModel>
{
	protected static final String VERSION = "v1.1";

	@Override
	public XmlElement build( FaceDetectionWholeFileModel model )
	{
		String projectModelDirName = model.getDirName();

		XmlElement result = createElement( "image_face_detections" );
		XmlElement nameXe = createStringElement( "name", model.getName() );
		result.addChild( nameXe );
		result.addChild(createStringElement("version", VERSION));

		XmlElement faceDetector = createElement( "face_detector" );
		fillFaceDetections( faceDetector, model );
		result.addChild( faceDetector );

		return( result );
	}

	public void fillFaceDetections( XmlElement xe, FaceDetectionWholeFileModel model )
	{
		XmlElement faceDetectorImages = createElement( "images" );
		fillImageFaceDetectionsXe( faceDetectorImages, model.getImageFaceDetectionPersistenceModelMap() );
		xe.addChild( faceDetectorImages );
	}

	protected void fillImageFaceDetectionsXe(XmlElement imagesXe, ImageFaceDetectionPersistenceModelMap imageFaceDetectionPersistenceModelMap)
	{
		if( imageFaceDetectionPersistenceModelMap != null ) {
			for (Map.Entry<ImageFaceDetectionPersistenceModelKey, ImageFaceDetectionPersistenceModel> entry :
					imageFaceDetectionPersistenceModelMap.getMapByName(
							ImageFaceDetectionPersistenceModelMap.MAIN_MAP, ImageFaceDetectionPersistenceModelKey.class).entrySet()) {
				imagesXe.addChild(createImageFaceDetectionXe(entry.getKey(), entry.getValue()));
			}
		}
	}

	protected XmlElement createImageFaceDetectionXe(ImageFaceDetectionPersistenceModelKey key,
													ImageFaceDetectionPersistenceModel value) {
		XmlElement imageXe = createElement("image");

		addImageFaceDetectionKeyToXe(imageXe, key);
		imageXe.addChild(createFaceDetections(value.getDetectedFaceMeshes()));

		return imageXe;
	}

	protected void addImageFaceDetectionKeyToXe(XmlElement xe, ImageFaceDetectionPersistenceModelKey key) {
		xe.addChild(createStringElement("filename", key.getFileName()));
		xe.addChild(createStringElement("length", key.getFileLength()));
		xe.addChild(createStringElement("md5", key.getHexMd5()));
	}

	protected XmlElement createFaceDetections(List<FaceDetectionMeshData> list) {
		XmlElement result = createElement("face_detections");

		for(FaceDetectionMeshData faceDetectionMeshData: list) {
			result.addChild(createFaceDetectionXe(faceDetectionMeshData));
		}

		return result;
	}

	protected XmlElement createFaceDetectionXe(FaceDetectionMeshData faceDetectionMeshData) {
		XmlElement result = createElement("face_detection");

		result.addChild(createBoundingBoxXe(faceDetectionMeshData.getOriginalFaceDetection().getBoundingBox()));
		result.addChild(createFaceDetectionPointsXe(createElement("points"),
				faceDetectionMeshData.getOriginalFaceDetection().getPoints()));
		result.addChild(createFaceDetectionPointsXe(createElement("modified_points"),
				faceDetectionMeshData.getModifiedFaceDetection().getPoints()));

		return result;
	}

	protected XmlElement createBoundingBoxXe(GeometricRectangle delaunayRectangle) {
		return fillDelaunayRectangle(createElement("bbox"), delaunayRectangle);
	}

	protected XmlElement fillDelaunayRectangle(XmlElement result, GeometricRectangle delaunayRectangle) {
		result.addChild(createStringElement("x", delaunayRectangle.x));
		result.addChild(createStringElement("y", delaunayRectangle.y));
		result.addChild(createStringElement("width", delaunayRectangle.width));
		result.addChild(createStringElement("height", delaunayRectangle.height));

		return result;
	}

	protected XmlElement createFaceDetectionPointsXe(XmlElement result, List<Point2d> list) {
		for(Point2d point2d: list) {
			result.addChild(fillPoint2d(createElement("point"), point2d));
		}

		return result;
	}

	protected XmlElement fillPoint2d(XmlElement result, Point2d point2d) {
		result.addChild(createStringElement("x", point2d.x));
		result.addChild(createStringElement("y", point2d.y));

		return result;
	}
}
