/*
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.applications.libmorphing.xml.facedetector.loader;

import com.frojasg1.applications.libmorphing.api.facedetector.pojo.FaceDetection;
import com.frojasg1.applications.libmorphing.api.facedetector.pojo.FaceDetectionMeshData;
import com.frojasg1.applications.libmorphing.model.facedetector.ImageFaceDetectionPersistenceModel;
import com.frojasg1.applications.libmorphing.model.facedetector.ImageFaceDetectionPersistenceModelMap;
import com.frojasg1.applications.libmorphing.xml.facedetector.FaceDetectionWholeFileModel;
import com.frojasg1.applications.libmorphing.xml.facedetector.FaceDetectionWholeFileModelBase;
import com.frojasg1.libgeometry.types.GeometricRectangle;
import com.frojasg1.libgeometry.types.Point2d;
import com.frojasg1.libgeometry.types.factory.GeometryCommonFactory;
import com.frojasg1.general.xml.XmlElement;
import com.frojasg1.general.xml.XmlElementList;
import com.frojasg1.general.xml.persistency.loader.impl.XmlToModelBase;

import java.util.ArrayList;
import java.util.List;

/**
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 */
public class XmlToFaceDetectionWholeFileModel
        extends XmlToModelBase<FaceDetectionWholeFileModel> {
    @Override
    public FaceDetectionWholeFileModelBase build(XmlElement xmlElement) {
        FaceDetectionWholeFileModelBase result = createEmptyModel();

        if (xmlElement != null)
            fillWholeModel(result, xmlElement);

        return (result);
    }

    protected FaceDetectionWholeFileModelBase createEmptyModel() {
        return (new FaceDetectionWholeFileModelBase(null));
    }

    protected void fillWholeModel(FaceDetectionWholeFileModel result, XmlElement xmlElement) {
//		result.setName( getMandatoryChildString( xmlElement, "name" ) );

        String version = getOptionalChildString(xmlElement, "version");

        ImageFaceDetectionPersistenceModelMap map = loadImageFaceDetections(xmlElement.getChild("face_detector"));
        if (map != null) {
            result.setImageFaceDetectionPersistenceModelMap(map);
        }
    }

    protected GeometricRectangle createRectangle(double x, double y, double width, double height) {
        return (GeometryCommonFactory.instance().createDelaunayRectangle(x, y, width, height));
    }

    public ImageFaceDetectionPersistenceModelMap loadImageFaceDetections(XmlElement xe) {
        ImageFaceDetectionPersistenceModelMap result = null;
        if (xe != null) {
            result = new ImageFaceDetectionPersistenceModelMap();
            XmlElementList itemsXe = xe.getChild("images").getChildrenByName("image");

            for (XmlElement chXe : itemsXe.getCollection()) {
                ImageFaceDetectionPersistenceModel imageFaceDetectionPersistenceModel = createImageFaceDetectionPersistenceModel(chXe, result);
                result.put(imageFaceDetectionPersistenceModel);
            }
        }

        return (result);
    }

    protected ImageFaceDetectionPersistenceModel createImageFaceDetectionPersistenceModel(XmlElement xe,
                                                                                          ImageFaceDetectionPersistenceModelMap map) {
        ImageFaceDetectionPersistenceModel result = null;
        if (xe != null) {
            String fileName = getMandatoryChildString(xe, "filename");
            Long fileLength = getOptionalChildObject(xe, "length", Long.class);
            String hexMd5 = getMandatoryChildString(xe, "md5");
            result = new ImageFaceDetectionPersistenceModel(map.createKey(fileName, fileLength, hexMd5));

            XmlElementList itemsXe = xe.getChild("face_detections").getChildrenByName("face_detection");

            List<FaceDetectionMeshData> faceDetectionMeshDataList = new ArrayList<>();
            for (XmlElement chXe : itemsXe.getCollection()) {
                faceDetectionMeshDataList.add(loadFaceDetectionMeshData(chXe));
            }
            result.setDetectedFaceMeshes(faceDetectionMeshDataList);
        }
        return result;
    }

    protected FaceDetectionMeshData createEmptyFaceDetectionMeshData() {
        return new FaceDetectionMeshData();
    }

    protected FaceDetection createEmptyFaceDetection() {
        return new FaceDetection();
    }

    protected FaceDetectionMeshData loadFaceDetectionMeshData(XmlElement xe) {
        FaceDetectionMeshData result = createEmptyFaceDetectionMeshData();

        FaceDetection originalFaceDetection = fillFaceDetection(xe.getChild("points"));
        originalFaceDetection.setBoundingBox(loadDelaunayRectangle(xe.getChild("bbox")));

        FaceDetection modifiedFaceDetection = fillFaceDetection(xe.getChild("modified_points"));

        result.setOriginalFaceDetection(originalFaceDetection);
        result.setModifiedFaceDetection(modifiedFaceDetection);

        return result;
    }

    protected FaceDetection fillFaceDetection(XmlElement xe) {
		FaceDetection result = createEmptyFaceDetection();

        XmlElementList itemsXe = xe.getChildrenByName("point");
        for (XmlElement chXe : itemsXe.getCollection()) {
            result.addPoint(loadPoint2d(chXe));
        }

        return result;
    }

    protected GeometricRectangle loadDelaunayRectangle(XmlElement xe) {
        GeometricRectangle result = null;
        if (xe != null) {
            result = createRectangle(
                    getInt(xe, "x", "bbox"),
                    getInt(xe, "y", "bbox"),
                    getInt(xe, "width", "bbox"),
                    getInt(xe, "height", "bbox"));
        }

        return result;
    }

    protected Point2d loadPoint2d(XmlElement xe) {
        Point2d result = null;
        if (xe != null) {
            result = new Point2d(
                    getInt(xe, "x", "point"),
                    getInt(xe, "y", "point"));
        }

        return result;
    }
}
