/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.applications.libmorphing.xml.facedetector.manager;

import com.frojasg1.applications.libmorphing.controller.MorphingVerticesMapManagerBase;
import com.frojasg1.applications.libmorphing.xml.facedetector.FaceDetectionWholeFileModel;
import com.frojasg1.applications.libmorphing.xml.facedetector.FaceDetectionWholeFileModelBase;
import com.frojasg1.applications.libmorphing.xml.facedetector.loader.FaceDetectionWholeFilePersistency;
import com.frojasg1.general.copy.CopyFacilities;
import com.frojasg1.general.copy.CopyFacilitiesFunctions;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.xml.transform.TransformerException;
import java.io.IOException;

/**
 *
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 */
public class FaceDetectionWholeFileModelManagerBase {
	private static final Logger LOGGER = LoggerFactory.getLogger(FaceDetectionWholeFileModelManagerBase.class);

	protected MorphingVerticesMapManagerBase _verticesManager;

	protected FaceDetectionWholeFilePersistency faceDetectionXmlPersistency;
	protected FaceDetectionWholeFileModel faceDetectionModel;

	public FaceDetectionWholeFileModelManagerBase()
	{
		init();
	}

	protected FaceDetectionWholeFilePersistency getPersistency() {
		return faceDetectionXmlPersistency;
	}


	protected void init()
	{
		faceDetectionXmlPersistency = createFaceDetectionWholeFilePersistency();
		setFaceDetectionWholeModel( createEmptyFaceDetectionWholeModel() );
	}

	protected FaceDetectionWholeFileModelBase createEmptyFaceDetectionWholeModel()
	{
		return( createEmptyFaceDetectionWholeModel("Default") );
	}

	protected FaceDetectionWholeFileModelBase createEmptyFaceDetectionWholeModel(String projectName )
	{
		return( new FaceDetectionWholeFileModelBase( projectName ) );
	}

	protected FaceDetectionWholeFilePersistency createFaceDetectionWholeFilePersistency()
	{
		return( new FaceDetectionWholeFilePersistency() );
	}

	public FaceDetectionWholeFileModel loadFaceDetectionWholeModel(String fileName )
	{
		FaceDetectionWholeFileModel result = null;

		try
		{
			result = getPersistency().loadModelFromXml(fileName);
			setFaceDetectionWholeModel(result);
		}
		catch( Exception ex )
		{
			LOGGER.error( "Error loading face detections fileName: {}", fileName, ex );
			setFaceDetectionWholeModel(createEmptyFaceDetectionWholeModel());

			throw( new RuntimeException( "Error face detections fileName: " + fileName, ex ) );
		}

		return( result );
	}

	public int importFaceDetectionWholeModel(String fileName )
	{
		FaceDetectionWholeFileModel result = loadOnTheFlyFaceDetectionWholeModel(fileName);
		return getFaceDetectionWholeModel().addAll(result.getImageFaceDetectionPersistenceModelMap());
	}

	public FaceDetectionWholeFileModel loadOnTheFlyFaceDetectionWholeModel(String fileName )
	{
		FaceDetectionWholeFileModel result = null;

		try {
			result = createFaceDetectionWholeFilePersistency().loadModelFromXml(fileName);
		}
		catch( Exception ex ) {
			LOGGER.error( "Error loading face detections fileName: {}", fileName, ex );
			throw( new RuntimeException( "Error face detections fileName: " + fileName, ex ) );
		}

		return( result );
	}

	public void saveFaceDetectionWholeModel( ) throws IOException, TransformerException
	{
		saveFaceDetectionWholeModelAs( getFaceDetectionWholeModel().getFileName() );
	}

	public void saveFaceDetectionWholeModelAs(String fileName ) throws IOException, TransformerException
	{
		getPersistency().saveXmlConfigurationFile(getFaceDetectionWholeModel(), fileName);
	}

	public FaceDetectionWholeFileModel getFaceDetectionWholeModel() {
		return faceDetectionModel;
	}

	public void setFaceDetectionWholeModel(FaceDetectionWholeFileModel faceDetectionWholeModel) {
		this.faceDetectionModel = faceDetectionWholeModel;
	}

	protected <C extends CopyFacilities<C>> C copy( C obj, C that )
	{
		return( CopyFacilitiesFunctions.instance().copy(obj, that) );
	}
}
