/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.applications.libmorphing.xml.whole.loader;

import com.frojasg1.applications.libmorphing.contexts.MorphingEditorApplicationContext;
import com.frojasg1.applications.libmorphing.api.facedetector.pojo.FaceDetection;
import com.frojasg1.applications.libmorphing.model.images.color.ColorModel;
import com.frojasg1.applications.libmorphing.model.images.color.ColorsMap;
import com.frojasg1.applications.libmorphing.model.images.configuration.ColorsConfigurationModel;
import com.frojasg1.applications.libmorphing.model.images.configuration.MeshAspectConfigurationModel;
import com.frojasg1.applications.libmorphing.model.images.configuration.MorphingProjectConfigurationModel;
import com.frojasg1.applications.libmorphing.model.images.configuration.SizeMode;
import com.frojasg1.applications.libmorphing.model.facedetector.ImageFaceDetectionPersistenceModel;
import com.frojasg1.applications.libmorphing.model.facedetector.ImageFaceDetectionPersistenceModelKey;
import com.frojasg1.applications.libmorphing.model.facedetector.ImageFaceDetectionPersistenceModelMap;
import com.frojasg1.applications.libmorphing.model.images.image.MorphingImageModel;
import com.frojasg1.applications.libmorphing.model.images.image.MorphingImageModelList;
import com.frojasg1.applications.libmorphing.model.images.transitions.TransitionConfigurationModel;
import com.frojasg1.applications.libmorphing.model.images.transitions.TransitionConfigurationModelKey;
import com.frojasg1.applications.libmorphing.model.images.transitions.TransitionConfigurationModelMap;
import com.frojasg1.applications.libmorphing.model.images.vertex.VertexModel;
import com.frojasg1.applications.libmorphing.model.images.video.MorphingVideoConfigurationModel;
import com.frojasg1.applications.libmorphing.model.images.video.MorphingVideoConfigurationModelMapWrapper;
import com.frojasg1.applications.libmorphing.xml.facedetector.FaceDetectionWholeFileModel;
import com.frojasg1.applications.libmorphing.xml.facedetector.loader.FaceDetectionWholeFileModelToXml;
import com.frojasg1.applications.libmorphing.xml.whole.MorphingProjectWholeFileModel;
import com.frojasg1.libgeometry.types.GeometricRectangle;
import com.frojasg1.libgeometry.types.Point2d;
import com.frojasg1.libgeometry.types.Size2d;
import com.frojasg1.general.FileFunctions;
import com.frojasg1.general.xml.XmlElement;
import com.frojasg1.general.xml.persistency.loader.impl.ModelToXmlBase;
import java.util.List;
import java.util.Map;

/**
 *
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 */
public class MorphingProjectWholeFileModelToXml extends ModelToXmlBase<MorphingProjectWholeFileModel>
{
	protected static final String VERSION = "v1.1";

	protected MorphingEditorApplicationContext appliCtx;


	public MorphingProjectWholeFileModelToXml(MorphingEditorApplicationContext appliCtx)
	{
		this.appliCtx = appliCtx;
	}

	protected MorphingEditorApplicationContext getAppliCtx() {
		return appliCtx;
	}

	@Override
	public XmlElement build( MorphingProjectWholeFileModel model )
	{
		String projectModelDirName = model.getDirName();

		XmlElement result = createElement( "morphing_project" );
		XmlElement nameXe = createStringElement( "name", model.getName() );
		result.addChild( nameXe );
		result.addChild(createStringElement("version", VERSION));

		XmlElement colorsXe = createElement( "colors" );
		fillColorsXe( colorsXe, model.getColorsMap() );
		result.addChild( colorsXe );

		XmlElement defaultConfigurationXe = createElement( "defaultconfiguration" );
		fillConfigurationXe( defaultConfigurationXe, model.getDefaultConfiguration() );
		result.addChild( defaultConfigurationXe );

		XmlElement imagesXe = createElement( "images" );
		fillImagesXe( imagesXe, model.getImagesListWrapper(), projectModelDirName );
		result.addChild( imagesXe );

		XmlElement transitionsXe = createElement( "transitions" );
		fillTransitionsXe( transitionsXe, model.getTransitionConfigurationMap() );
		result.addChild( transitionsXe );

		XmlElement videoConfigurationsXe = createElement( "video_configurations" );
		fillVideoConfigurationsXe( videoConfigurationsXe, model.getVideoConfigurationMap() );
		result.addChild( videoConfigurationsXe );

//		XmlElement imageFaceDetectionConfigurationModelMapXe = createElement( "face_detector" );
//		fillImageFaceDetectionsXe( imageFaceDetectionConfigurationModelMapXe,
//				model.getImageFaceDetectionPersistenceModelMap() );
//		result.addChild( imageFaceDetectionConfigurationModelMapXe );

		XmlElement faceDetector = createElement("face_detector");
		fillFaceDetections(faceDetector, model.getFaceDetectionForXml());
		result.addChild(faceDetector);

		return( result );
	}

	protected void fillFaceDetections( XmlElement xe, FaceDetectionWholeFileModel model ) {
		createFaceDetectionWholeFileModelToXml().fillFaceDetections(xe, model);
	}

	protected FaceDetectionWholeFileModelToXml createFaceDetectionWholeFileModelToXml() {
		return new FaceDetectionWholeFileModelToXml();
	}

	protected void fillColorsXe( XmlElement xe, ColorsMap colorsMap )
	{
		for( String colorName: colorsMap.keySet() )
			xe.addChild( createColorXe( colorsMap.get( colorName ) ) );
	}

	protected XmlElement createColorXe( ColorModel colorModel )
	{
		XmlElement result = createElement( "color" );
		result.addChild( createStringElement( "name", colorModel.getName() ) );
		result.addChild( createStringElement( "red", colorModel.getColor().getRed() ) );
		result.addChild( createStringElement( "green", colorModel.getColor().getGreen() ) );
		result.addChild( createStringElement( "blue", colorModel.getColor().getBlue() ) );

		return( result );
	}

	protected String sizeModeToString( SizeMode sizeMode )
	{
		return( ( sizeMode != null ) ? sizeMode.name() : null );
	}

	protected void fillConfigurationXe( XmlElement xe, MorphingProjectConfigurationModel configurationModel )
	{
		if( configurationModel != null )
		{
			XmlElement sizeModeXe = createStringElement( "sizemode", sizeModeToString( configurationModel.getSizeMode() ) );
			xe.addChild( sizeModeXe );

			XmlElement transitionTimeMsXe = createStringElement( "transitiontime_ms", configurationModel.getTransitionTimeMs() );
			xe.addChild( transitionTimeMsXe );

			XmlElement colorsConfigurationModelXe = createElement( "colors" );
			fillConfigurationColorsXe( colorsConfigurationModelXe, configurationModel.getColorsConfigurationInternal() );
			xe.addChild( colorsConfigurationModelXe );

			XmlElement meshXe = createElement( "mesh" );
			fillMeshXe( meshXe, configurationModel.getMeshAspectConfigurationInternal() );
			xe.addChild( meshXe );
		}
	}

	protected void fillConfigurationColorsXe( XmlElement xe, ColorsConfigurationModel colorsConfiguration )
	{
		if( colorsConfiguration != null )
		{
			// name can be: edge, vertex, selectedvertex, erroneusvertex
			for( String name: colorsConfiguration.keySet() )
				xe.addChild( createStringElement(name, colorsConfiguration.get(name).getName() ) );
		}
	}

	protected void fillMeshXe( XmlElement xe, MeshAspectConfigurationModel meshConfiguration )
	{
		if( meshConfiguration != null )
		{
			xe.addChild( createStringElement( "vertexradius", meshConfiguration.getVertexRadiusInternal() ) );
			xe.addChild( createStringElement( "edgewidth", meshConfiguration.getEdgeWidthInternal() ) );

			xe.addChild( createStringElement( "showmesh", meshConfiguration.getShowMeshInternal() ) );
			xe.addChild( createStringElement( "showvertices", meshConfiguration.getShowVerticesInternal() ) );
			xe.addChild( createStringElement( "showcontents", meshConfiguration.getShowContentsInternal() ) );
		}
	}

	protected void fillImagesXe( XmlElement xe, MorphingImageModelList imagesListWrapper,
								String projectModelDirName)
	{
		int index = 1;
		for( MorphingImageModel imgModel: imagesListWrapper.getListCopy() )
			xe.addChild( createImageModelXe( imgModel, index++, projectModelDirName ) );
	}

	protected String getImageFileName( MorphingImageModel imgModel,
										String projectModelDirName )
	{
		String result = FileFunctions.instance().getRelativePath(imgModel.getFileName(), projectModelDirName);
/*
		String fileName = imgModel.getFileName();
		if( fileName.startsWith(projectModelDirName) ) {
			fileName = StringFunctions.instance().removeAtStart(fileName, projectModelDirName);
			fileName = StringFunctions.instance().removeCharsAtStart(fileName, "/\\");

			// in order for the same project file to work in other systems (from Windows to Linux or Mac)
			fileName = FileFunctions.instance().convertFolderSeparatorGen(fileName, "\\", "/" );
		}

		return( fileName );
*/
		return( result );
	}

	protected XmlElement createImageModelXe( MorphingImageModel imgModel, int index,
											String projectModelDirName)
	{
		XmlElement result = createElement( "image" );
		result.addChild( createStringElement( "name", imgModel.getName() ) );
		result.addChild( createStringElement( "file", getImageFileName( imgModel, projectModelDirName ) ) );
		result.addChild( createStringElement( "is_master", imgModel.isMaster() ) );
//		result.addChild( createStringElement( "index", index ) );
		result.addChild( createStringElement("following_transition", imgModel.getFollowingTransition().name()));

		XmlElement faceDetectionXe = createElement("face_detection");
		faceDetectionXe.addChild( createStringElement( "image_face_index", imgModel.getImageFaceDetectionIndex()));

		List<Point2d> controlPoints = imgModel.getFaceDetection88Points();
		if(!controlPoints.isEmpty()) {
			XmlElement controlPointsXe = fillControlPointsXe(createElement("points"), controlPoints);
			faceDetectionXe.addChild(controlPointsXe);
		}

		result.addChild(faceDetectionXe);

		XmlElement boundsXe = createElement( "bounds" );
		fillBoundsXe( boundsXe, imgModel.getBounds() );
		result.addChild( boundsXe );

		XmlElement configurationXe = createElement( "configuration" );
		fillConfigurationXe( configurationXe, imgModel.getConfigurationModel() );
		result.addChild( configurationXe );

		XmlElement verticesXe = createElement( "vertices" );
		fillVerticesXe( verticesXe, imgModel.getVertices() );
		result.addChild( verticesXe );

		return( result );
	}

	protected XmlElement fillControlPointsXe( XmlElement xe, List<Point2d> controlPointList )
	{
		for( Point2d pt: controlPointList )
			xe.addChild( fillPoint2d( createElement("point"), pt ) );

		return xe;
	}

	protected void fillBoundsXe( XmlElement xe, GeometricRectangle bounds )
	{
		xe.addChild( createStringElement( "x", bounds.x ) );
		xe.addChild( createStringElement( "y", bounds.y ) );
		xe.addChild( createStringElement( "width", bounds.width ) );
		xe.addChild( createStringElement( "height", bounds.height ) );
	}

	protected void fillVerticesXe( XmlElement xe, List<VertexModel> verticesList )
	{
		for( VertexModel vertexModel: verticesList )
			xe.addChild( createVertexModelXe( vertexModel ) );
	}

	protected XmlElement createVertexModelXe( VertexModel vertexModel )
	{
		XmlElement result = createElement( "vertex" );
		XmlElement coorXe = createElement( "coordinates" );
		fillVertexXe( coorXe, vertexModel.getPoint() );
		result.addChild( coorXe );

		XmlElement colorsXe = createElement( "colors" );
		fillList( colorsXe, "color", vertexModel.getColors().getListCopy() );
		result.addChild( colorsXe );

		return( result );
	}

	protected XmlElement fillVertexXe( XmlElement xe, Point2d point )
	{
		xe.addChild( createStringElement( "x", point.x ) );
		xe.addChild( createStringElement( "y", point.y ) );

		return xe;
	}

	protected void fillTransitionsXe( XmlElement xe, TransitionConfigurationModelMap transitionMap )
	{
		Map<TransitionConfigurationModelKey, TransitionConfigurationModel> map = 
			transitionMap.getMapByName(
					TransitionConfigurationModelMap.MAIN_MAP,
					TransitionConfigurationModelKey.class);
		for( TransitionConfigurationModelKey key: map.keySet() )
		{
			xe.addChild( createTransitionModelXe( map.get(key) ) );
		}
	}

	protected XmlElement createTransitionModelXe( TransitionConfigurationModel transitionModel )
	{
		XmlElement result = createElement( "transition" );
		result.addChild( createStringElement( "imagefromname", transitionModel.getKey().getFromImageName() ) );
		result.addChild( createStringElement( "imagetoname", transitionModel.getKey().getToImageName() ) );
		result.addChild( createStringElement( "transitiontime_ms", transitionModel.getTransitionTimeMs() ) );

		return( result );
	}

	protected void fillVideoConfigurationsXe( XmlElement xe, MorphingVideoConfigurationModelMapWrapper videoConfMap )
	{
		if( videoConfMap != null )
			for( String videoConfName: videoConfMap.keySet() )
				xe.addChild( createVideoConfigurationXe( videoConfMap.get( videoConfName ) ) );
	}

	protected void fillImageFaceDetectionsXe(XmlElement xe, ImageFaceDetectionPersistenceModelMap imageFaceDetectionPersistenceModelMap)
	{
		if( imageFaceDetectionPersistenceModelMap != null ) {
			XmlElement imagesXe = createElement("images");
			for (Map.Entry<ImageFaceDetectionPersistenceModelKey, ImageFaceDetectionPersistenceModel> entry :
					imageFaceDetectionPersistenceModelMap.getMapByName(
							ImageFaceDetectionPersistenceModelMap.MAIN_MAP, ImageFaceDetectionPersistenceModelKey.class).entrySet()) {
				imagesXe.addChild(createImageFaceDetectionXe(entry.getKey(), entry.getValue()));
			}
			xe.addChild(imagesXe);
		}
	}

	protected XmlElement createImageFaceDetectionXe(ImageFaceDetectionPersistenceModelKey key,
													ImageFaceDetectionPersistenceModel value) {
		XmlElement imageXe = createElement("image");

		addImageFaceDetectionKeyToXe(imageXe, key);
		imageXe.addChild(createFaceDetetions(value.getImageFaceDetection().getFaceDetections()));

		return imageXe;
	}

	protected void addImageFaceDetectionKeyToXe(XmlElement xe, ImageFaceDetectionPersistenceModelKey key) {
		xe.addChild(createStringElement("filename", key.getFileName()));
		xe.addChild(createStringElement("length", key.getFileLength()));
		xe.addChild(createStringElement("md5", key.getHexMd5()));
	}

	protected XmlElement createFaceDetetions(List<FaceDetection> list) {
		XmlElement result = createElement("face_detections");

		for(FaceDetection faceDetection: list) {
			result.addChild(createFaceDetectionXe(faceDetection));
		}

		return result;
	}

	protected XmlElement createFaceDetectionXe(FaceDetection faceDetection) {
		XmlElement result = createElement("face_detection");

		result.addChild(createBoundingBoxXe(faceDetection.getBoundingBox()));
		result.addChild(createFaceDetectionPointsXe(faceDetection.getPoints()));

		return result;
	}

	protected XmlElement createBoundingBoxXe(GeometricRectangle delaunayRectangle) {
		return fillDelaunayRectangle(createElement("bbox"), delaunayRectangle);
	}

	protected XmlElement fillDelaunayRectangle(XmlElement result, GeometricRectangle delaunayRectangle) {
		result.addChild(createStringElement("x", delaunayRectangle.x));
		result.addChild(createStringElement("y", delaunayRectangle.y));
		result.addChild(createStringElement("width", delaunayRectangle.width));
		result.addChild(createStringElement("height", delaunayRectangle.height));

		return result;
	}

	protected XmlElement createFaceDetectionPointsXe(List<Point2d> list) {
		XmlElement result = createElement("points");
		for(Point2d point2d: list) {
			result.addChild(fillPoint2d(createElement("point"), point2d));
		}

		return result;
	}

	protected XmlElement fillPoint2d(XmlElement result, Point2d point2d) {
		result.addChild(createStringElement("x", point2d.x));
		result.addChild(createStringElement("y", point2d.y));

		return result;
	}

	protected XmlElement createVideoConfigurationXe( MorphingVideoConfigurationModel videoConfModel )
	{
		XmlElement result = null;
		if( videoConfModel != null )
		{
			result = createElement( "video" );
			result.addChild( createStringElement( "name", videoConfModel.getName() ) );
			result.addChild( createStringElement( "format", videoConfModel.getFormatName() ) );
			result.addChild( createStringElement( "filename", videoConfModel.getFileName() ) );

			XmlElement sizeXe = createElement( "size" );
			fillSizeXe( sizeXe, videoConfModel.getResolution() );
			result.addChild( sizeXe );

			XmlElement confXe = createElement( "configuration" );
			fillConfigurationXe( confXe, videoConfModel.getMorphingConfiguration() );
			result.addChild( confXe );
		}

		return( result );
	}

	protected void fillSizeXe( XmlElement xe, Size2d size )
	{
		if( size != null )
		{
			xe.addChild( createStringElement( "width", size.width ) );
			xe.addChild( createStringElement( "height", size.height ) );
		}
	}
}
