/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.applications.libmorphing.xml.whole.loader;

import com.frojasg1.applications.libmorphing.contexts.MorphingEditorApplicationContext;
import com.frojasg1.applications.libmorphing.frames.framebuilder.impl.MorphingTransition;
import com.frojasg1.applications.libmorphing.helpers.MorphingHelper;
import com.frojasg1.applications.libmorphing.model.images.color.ColorModel;
import com.frojasg1.applications.libmorphing.model.images.color.ColorsMap;
import com.frojasg1.applications.libmorphing.model.images.configuration.ColorsConfigurationModel;
import com.frojasg1.applications.libmorphing.model.images.configuration.MeshAspectConfigurationModel;
import com.frojasg1.applications.libmorphing.model.images.configuration.MorphingProjectConfigurationModel;
import com.frojasg1.applications.libmorphing.model.images.configuration.SizeMode;
import com.frojasg1.applications.libmorphing.model.facedetector.ImageFaceDetectionPersistenceModelMap;
import com.frojasg1.applications.libmorphing.model.images.image.MorphingImageModel;
import com.frojasg1.applications.libmorphing.model.images.image.MorphingImageModelList;
import com.frojasg1.applications.libmorphing.model.images.transitions.TransitionConfigurationModel;
import com.frojasg1.applications.libmorphing.model.images.transitions.TransitionConfigurationModelKey;
import com.frojasg1.applications.libmorphing.model.images.transitions.TransitionConfigurationModelMap;
import com.frojasg1.applications.libmorphing.model.images.vertex.VertexModel;
import com.frojasg1.applications.libmorphing.model.images.video.MorphingVideoConfigurationModel;
import com.frojasg1.applications.libmorphing.model.images.video.MorphingVideoConfigurationModelMapWrapper;
import com.frojasg1.applications.libmorphing.xml.facedetector.FaceDetectionWholeFileModel;
import com.frojasg1.applications.libmorphing.xml.facedetector.FaceDetectionWholeFileModelBase;
import com.frojasg1.applications.libmorphing.xml.facedetector.loader.XmlToFaceDetectionWholeFileModel;
import com.frojasg1.applications.libmorphing.xml.whole.MorphingProjectWholeFileModel;
import com.frojasg1.applications.libmorphing.xml.whole.MorphingProjectWholeFileModelBase;
import com.frojasg1.libgeometry.types.GeometricRectangle;
import com.frojasg1.libgeometry.types.Point2d;
import com.frojasg1.libgeometry.types.Size2d;
import com.frojasg1.libgeometry.types.factory.GeometryCommonFactory;
import com.frojasg1.general.xml.XmlElement;
import com.frojasg1.general.xml.XmlElementList;
import com.frojasg1.general.xml.persistency.loader.impl.XmlToModelBase;
import java.awt.Color;

/**
 *
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 */
public class XmlToMorphingProjectWholeFileModel
		extends XmlToModelBase<MorphingProjectWholeFileModel>
{
	protected MorphingEditorApplicationContext appliCtx;

	public XmlToMorphingProjectWholeFileModel(MorphingEditorApplicationContext appliCtx) {
		this.appliCtx = appliCtx;
	}

	protected MorphingEditorApplicationContext getAppliCtx() {
		return appliCtx;
	}

	@Override
	public MorphingProjectWholeFileModelBase build(XmlElement xmlElement )
	{
		MorphingProjectWholeFileModelBase result = createEmptyModel();

		if( xmlElement != null )
			fillWholeModel( result, xmlElement );

		return( result );
	}

	protected MorphingProjectWholeFileModelBase createEmptyModel()
	{
		return( new MorphingProjectWholeFileModelBase(null, getAppliCtx()) );
	}

	protected void fillWholeModel(MorphingProjectWholeFileModelBase result, XmlElement xmlElement )
	{
		result.setName( getMandatoryChildString( xmlElement, "name" ) );

		String version = getOptionalChildString(xmlElement, "version");

		ColorsMap colorsMap = result.getColorsMap();
		fillColors( colorsMap, xmlElement.getChild( "colors" ) );
		fillMorphingProjectConfigurationModel( result.getDefaultConfiguration(), xmlElement.getChild( "defaultconfiguration" ),
							colorsMap );
		fillImages( result.getImagesListWrapper(), xmlElement.getChild( "images" ), colorsMap );
		result.setTransitionConfigurationMap( loadTransitions( xmlElement.getChild( "transitions" ) ) );
		result.setVideoConfigurationMap( loadVideoConfigurations( xmlElement.getChild( "video_configurations" ), colorsMap ) );
//		ImageFaceDetectionPersistenceModelMap map = loadImageFaceDetections(xmlElement.getChild("face_detector"));
//		if(map != null) {
//			result.setImageFaceDetectionPersistenceModelMap(map);
//		}

		result.setFaceDetectionForXml(loadFaceDetectionWholeFileModel(xmlElement.getChild("face_detector")));
	}

	protected FaceDetectionWholeFileModel loadFaceDetectionWholeFileModel(XmlElement xmlElement) {
		FaceDetectionWholeFileModel result = createEmptyFaceDetectionModel();
		ImageFaceDetectionPersistenceModelMap map = createXmlToFaceDetectionWholeFileModel().loadImageFaceDetections(xmlElement);;
		if (map != null) {
			result.setImageFaceDetectionPersistenceModelMap(map);
		}
		return result;
	}

	protected XmlToFaceDetectionWholeFileModel createXmlToFaceDetectionWholeFileModel() {
		return new XmlToFaceDetectionWholeFileModel();
	}

	protected FaceDetectionWholeFileModel createEmptyFaceDetectionModel() {
		return (new FaceDetectionWholeFileModelBase(null));
	}

	protected void fillColors( ColorsMap result, XmlElement xmlElement )
	{
//		XmlElement itemsXe = getMandatoryChild( xmlElement, "colors", "color" );
		XmlElementList itemsXe = xmlElement.getChildrenByName( "colors" );

		if( itemsXe != null )
			for( XmlElement cmXe: itemsXe.getCollection() )
				result.put( createColorModel( cmXe ) );
	}

	protected ColorModel createColorModel( XmlElement xe )
	{
		ColorModel result = new ColorModel();
		result.setName( getMandatoryChildString( xe, "name" ) );
		result.setColor( createColor( xe ) );

		return( result );
	}

	protected Color createColor( XmlElement xe )
	{
		return new Color(getMandatoryChildInt( xe, "red" ),
						getMandatoryChildInt( xe, "green" ),
						getMandatoryChildInt( xe, "blue" ) );
	}

	protected void fillMorphingProjectConfigurationModel( MorphingProjectConfigurationModel result,
														XmlElement xe, ColorsMap colorsMap )
	{
		String sm = getOptionalChildString( xe, "sizemode" );
		if( sm != null )
			result.setSizeMode( SizeMode.valueOf(sm) );

		result.setTransitionTimeMs( getOptionalChildObject( xe, "transitiontime_ms", Integer.class ) );
		fillColorConf( result.getColorsConfigurationInternal(), xe.getChild("colors"),
						colorsMap );
		result.setMeshAspectConfiguration( loadMeshConf( xe.getChild( "mesh" ) ) );
	}

	protected void fillColorConf( ColorsConfigurationModel result,
										XmlElement xe, ColorsMap colorsMap )
	{
		if( xe != null )
			for( XmlElement chXe: xe.getAllChildren() )
			{
				String colorName = chXe.getText();
				ColorModel cm = colorsMap.get( colorName );
				if( cm == null )
					throw( new RuntimeException( "Color model not defined: " + colorName ) );
				result.put( chXe.getName(), cm );
			}
	}

	protected MeshAspectConfigurationModel loadMeshConf( XmlElement xe )
	{
		MeshAspectConfigurationModel result = null;
		if( xe != null ) 
		{
			result = new MeshAspectConfigurationModel(null);

			result.setVertexRadius( getOptionalChildObject( xe, "vertexradius", Float.class ) );
			result.setEdgeWidth( getOptionalChildObject( xe, "edgewidth", Float.class ) );

			result.setShowMesh( getOptionalChildObject( xe, "showmesh", Boolean.class ) );
			result.setShowVertices( getOptionalChildObject( xe, "showvertices", Boolean.class ) );
			result.setShowContents( getOptionalChildObject( xe, "showcontents", Boolean.class ) );
		}

		return( result );
	}

	protected void fillImages(MorphingImageModelList result,
							  XmlElement xe, ColorsMap colorsMap )
	{
		XmlElementList itemsXe = xe.getChildrenByName("image" );

		for( XmlElement chXe: itemsXe.getCollection() )
			result.add( createImageModel( chXe, colorsMap ) );
	}

	protected MorphingImageModel createImageModel( XmlElement xe, ColorsMap colorsMap )
	{
		MorphingImageModel result = null;
		if( xe != null )
		{
			result = new MorphingImageModel();

			String imageName = getMandatoryChildStringErrorMessage( xe, "name" );
			result.setName( imageName );
			result.setFileName( getMandatoryChildStringErrorMessage( xe, "file" ) );
			result.setIsMaster(getMandatoryChildBoolean( xe, "is_master" ) );

			String morphingTransitionName = getOptionalChildString(xe, "following_transition");
			if(morphingTransitionName != null) {
				result.setFollowingTransition(MorphingTransition.valueOf(morphingTransitionName));
			}

			result.setBounds( createBounds( getMandatoryChildErrorMessage( xe, "bounds" ) ) );
			result.setConfigurationModel( createMorphingProjectConfigurationModel( xe.getChild( "configuration" ), colorsMap ) );
			addVertices( getMandatoryChild( xe, "vertices" ), result );

			// we let it be empty, for backwards compatibility
			XmlElement faceDetectionXe = xe.getChild("face_detection");
			if(faceDetectionXe != null) {
				int imageFaceIndex = getMandatoryChildInt(faceDetectionXe, "image_face_index");
				result.setImageFaceDetectionIndex(imageFaceIndex);

				XmlElement controlMeshXe = faceDetectionXe.getChild("points");
				if(controlMeshXe != null) {
					XmlElementList itemsXe = controlMeshXe.getChildrenByName("point");
					for (XmlElement chXe : itemsXe.getCollection()) {
						result.addFaceDetectionPoint(loadPoint2d(chXe));
					}
				} else {
					if(imageFaceIndex > -1) {
						throw new IllegalArgumentException(String.format("The image face detection index was set, but no vertices were found for image: %s", imageName));
					}
				}
			}
		}

		return( result );
	}

	protected Point2d loadPoint2d(XmlElement xe) {
		Point2d result = null;
		if (xe != null) {
			result = new Point2d(
					getInt(xe, "x", "point"),
					getInt(xe, "y", "point"));
		}

		return result;
	}

	protected void addVertices( XmlElement xe, MorphingImageModel imageModel )
	{
		XmlElementList itemsXe = xe.getChildrenByName("vertex" );

		for( XmlElement chXe: itemsXe.getCollection() )
			imageModel.addVertex( createVertexModel( chXe, imageModel.getBounds() ) );
	}

	protected Point2d toNormalized( Point2d pt, GeometricRectangle rect )
	{
		return( MorphingHelper.instance().toNormalized(pt, rect) );
	}

	protected VertexModel createVertexModel( XmlElement xe, GeometricRectangle bounds )
	{
		VertexModel result = new VertexModel();

		result.setPoint( createPoint(getMandatoryChildErrorMessage( xe, "coordinates" )) );
		result.setNormalizedPoint( toNormalized( result.getPoint(), bounds ) );

		return( result );
	}

	protected Point2d createPoint( double xx, double yy )
	{
		return( GeometryCommonFactory.instance().createPoint2d(xx, yy) );
	}

	protected Point2d createPoint( XmlElement xe )
	{
		return( createPoint( getMandatoryChildDouble(xe, "x"), getMandatoryChildDouble(xe, "y") ) );
	}

	public GeometricRectangle createRectangle(double x, double y, double width, double height )
	{
		return( GeometryCommonFactory.instance().createDelaunayRectangle( x, y, width, height ) );
	}

	protected GeometricRectangle createBounds(XmlElement xe )
	{
		GeometricRectangle result = null;
		if( xe != null ) 
		{
			result = createRectangle(
				getMandatoryChildDouble( xe, "x" ),
				getMandatoryChildDouble( xe, "y" ),
				getMandatoryChildDouble( xe, "width" ),
				getMandatoryChildDouble( xe, "height" )
			);
		}

		return( result );
	}

	protected TransitionConfigurationModelMap loadTransitions( XmlElement xe )
	{
		TransitionConfigurationModelMap result = null;
		if( xe != null )
		{
			result = new TransitionConfigurationModelMap();
			XmlElementList itemsXe = xe.getChildrenByName("transition" );

			for( XmlElement chXe: itemsXe.getCollection() )
				result.put( createTransitionConfigurationModel( chXe, result ) );
		}

		return( result );
	}

	protected TransitionConfigurationModel createTransitionConfigurationModel( XmlElement xe,
																				TransitionConfigurationModelMap map)
	{
		TransitionConfigurationModel result = null;
		if( xe != null ) 
		{
			String fromImageName = getMandatoryChildStringErrorMessage( xe, "imagefromname" );
			String toImageName = getMandatoryChildStringErrorMessage( xe, "imagetoname" );
			TransitionConfigurationModelKey key = map.createKey(fromImageName, toImageName);
			result = new TransitionConfigurationModel( key );

			result.setTransitionTimeMs( getMandatoryChildInt( xe, "transitiontime_ms" ) );
		}

		return( result );
	}

	protected MorphingVideoConfigurationModelMapWrapper loadVideoConfigurations( XmlElement xe, ColorsMap colorsMap )
	{
		MorphingVideoConfigurationModelMapWrapper result = null;
		if( xe != null )
		{
			result = new MorphingVideoConfigurationModelMapWrapper();
			XmlElementList itemsXe = xe.getChildrenByName( "video" );

			for( XmlElement chXe: itemsXe.getCollection() )
			{
				MorphingVideoConfigurationModel videoConf = createMorphingVideoConfigurationModel( chXe, colorsMap );
				result.put( videoConf.getName(), videoConf );
			}
		}

		return( result );
	}

//	protected ImageFaceDetectionPersistenceModelMap loadImageFaceDetections(XmlElement xe)
//	{
//		ImageFaceDetectionPersistenceModelMap result = null;
//		if( xe != null )
//		{
//			result = new ImageFaceDetectionPersistenceModelMap();
//			XmlElementList itemsXe = xe.getChild("images").getChildrenByName( "image" );
//
//			for( XmlElement chXe: itemsXe.getCollection() )
//			{
//				ImageFaceDetectionPersistenceModel imageFaceDetectionPersistenceModel = createImageFaceDetectionPersistenceModel(chXe, result);
//				result.put( imageFaceDetectionPersistenceModel );
//			}
//		}
//
//		return( result );
//	}
//
//	protected ImageFaceDetectionPersistenceModel createImageFaceDetectionPersistenceModel(XmlElement xe,
//																						  ImageFaceDetectionPersistenceModelMap map) {
//		ImageFaceDetectionPersistenceModel result = null;
//		if(xe != null) {
//			String fileName = getMandatoryChildString(xe, "filename");
//			Long fileLength = getOptionalChildObject(xe, "length", Long.class);
//			String hexMd5 = getMandatoryChildString(xe, "md5");
//			result = new ImageFaceDetectionPersistenceModel(map.createKey(fileName, fileLength, hexMd5));
//
//			ImageFaceDetection imageFaceDetection = new ImageFaceDetection();
//
//			XmlElementList itemsXe = xe.getChild("face_detections").getChildrenByName( "face_detection" );
//
//			for( XmlElement chXe: itemsXe.getCollection() )
//			{
//				imageFaceDetection.addFaceDetection(loadFaceDetection(chXe));
//			}
//			result.setImageFaceDetection(imageFaceDetection);
//		}
//		return result;
//	}
//
//	protected FaceDetection loadFaceDetection(XmlElement xe) {
//		FaceDetection result = new FaceDetection();
//		result.setBoundingBox(loadDelaunayRectangle(xe.getChild("bbox")));
//
//		XmlElementList itemsXe = xe.getChild("points").getChildrenByName("point");
//		for(XmlElement chXe: itemsXe.getCollection()) {
//			result.addPoint(loadPoint2d(chXe));
//		}
//
//		return result;
//	}
//
//	protected DelaunayRectangle loadDelaunayRectangle(XmlElement xe) {
//		DelaunayRectangle result = null;
//		if(xe != null) {
//			result = new DelaunayRectangle(
//					getInt(xe, "x", "bbox"),
//					getInt(xe, "y", "bbox"),
//					getInt(xe, "width", "bbox"),
//					getInt(xe, "height", "bbox"));
//		}
//
//		return result;
//	}
//
//	protected Point2d loadPoint2d(XmlElement xe) {
//		Point2d result = null;
//		if(xe != null) {
//			result = new Point2d(
//					getInt(xe, "x", "point"),
//					getInt(xe, "y", "point"));
//		}
//
//		return result;
//	}

	protected MorphingVideoConfigurationModel createMorphingVideoConfigurationModel( XmlElement xe, ColorsMap colorsMap )
	{
		MorphingVideoConfigurationModel result = null;
		if( xe != null ) 
		{
			String name = getMandatoryChildStringErrorMessage( xe, "name" );
			result = new MorphingVideoConfigurationModel( name );
			result.setFormatName(getMandatoryChildStringErrorMessage( xe, "format" ));
			result.setFileName(getMandatoryChildStringErrorMessage( xe, "file" ));
			result.setResolution( createSize2d( getMandatoryChildErrorMessage( xe, "size" ) ) );
			result.setMorphingConfiguration( createMorphingProjectConfigurationModel( xe.getChild("configuration" ), colorsMap ) );
		}

		return( result );
	}

	protected Size2d createSize2d( XmlElement xe )
	{
		Size2d result = null;
		if( xe != null ) 
			result = new Size2d( getMandatoryChildInt( xe, "x" ), getMandatoryChildInt( xe, "y" ) );

		return( result );
	}

	protected MorphingProjectConfigurationModel createMorphingProjectConfigurationModel( XmlElement xe, ColorsMap colorsMap )
	{
		MorphingProjectConfigurationModel result = null;
		if( xe != null ) 
		{
			result = new MorphingProjectConfigurationModel(null);
			fillMorphingProjectConfigurationModel( result, xe, colorsMap );
		}

		return( result );
	}
}
