/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.applications.libmorphing.xml.whole.manager;

import com.frojasg1.applications.libmorphing.contexts.MorphingEditorApplicationContext;
import com.frojasg1.applications.libmorphing.contexts.holder.ApplicationCtxHolder;
import com.frojasg1.applications.libmorphing.controller.MorphingVerticesMapManager;
import com.frojasg1.applications.libmorphing.controller.MorphingVerticesMapManagerBase;
import com.frojasg1.applications.libmorphing.model.facedetector.ImageFaceDetectionPersistenceModel;
import com.frojasg1.applications.libmorphing.model.images.configuration.MorphingProjectConfigurationModel;
import com.frojasg1.applications.libmorphing.model.images.image.MorphingImageModel;
import com.frojasg1.applications.libmorphing.xml.facedetector.FaceDetectionWholeFileModel;
import com.frojasg1.applications.libmorphing.xml.facedetector.FaceDetectionWholeFileModelBase;
import com.frojasg1.applications.libmorphing.xml.whole.MorphingProjectWholeFileModel;
import com.frojasg1.applications.libmorphing.xml.whole.MorphingProjectWholeFileModelBase;
import com.frojasg1.applications.libmorphing.xml.whole.loader.MorphingProjectWholeFilePersistency;
import com.frojasg1.libgeometry.types.GeometricRectangle;
import com.frojasg1.general.copy.CopyFacilities;
import com.frojasg1.general.copy.CopyFacilitiesFunctions;

import java.io.IOException;
import java.util.List;
import java.util.Objects;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.TransformerException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.xml.sax.SAXException;

/**
 *
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 */
public class MorphingProjectWholeFileModelManagerBase
		extends ApplicationCtxHolder
		implements MorphingProjectWholeFileModelManager {
	private static final Logger LOGGER = LoggerFactory.getLogger(MorphingProjectWholeFileModelManagerBase.class);

//	protected MorphingApplicationConfiguration applyConf;
	protected MorphingVerticesMapManager _verticesManager;

	protected MorphingProjectWholeFilePersistency _morphingProjectXmlPersistency;
	protected MorphingProjectWholeFileModel _morphingProjectModel;

	public MorphingProjectWholeFileModelManagerBase(MorphingEditorApplicationContext appliCtx)
	{
		super(appliCtx);
		init();
	}


	protected void init()
	{
		_morphingProjectXmlPersistency = createMorphingProjectWholeFilePersistency();
		setMorphingProjectModel( createEmptyMorphingProjectModel( "Default" ) );
	}

	protected MorphingProjectWholeFilePersistency getPersistency() {
		return _morphingProjectXmlPersistency;
	}

	protected MorphingVerticesMapManager createVerticesManager(MorphingProjectWholeFileModel morphingProjectModel)
	{
		MorphingVerticesMapManagerBase verticesManager = createMorphingVerticesMapManager();

//		verticesManager.setMorphingProjectWholeFileModel(getProjectModel());
		verticesManager.setImageModelList( morphingProjectModel.getImagesListWrapper(),
											getMasterImageName() );

		return( verticesManager );
	}

	protected MorphingProjectWholeFileModelBase createEmptyMorphingProjectModel(String projectName )
	{
		return( new MorphingProjectWholeFileModelBase( projectName, getAppliCtx() ) );
	}

	protected MorphingProjectWholeFilePersistency createMorphingProjectWholeFilePersistency()
	{
		return( new MorphingProjectWholeFilePersistency(getAppliCtx()) );
	}

	@Override
	public MorphingProjectWholeFileModel loadMorphingProjectModel(String fileName) throws IOException, SAXException, ParserConfigurationException
	{
		MorphingProjectWholeFileModel result = null;
		
		try {
			result = loadProject(fileName);
		}
		catch( Exception ex ) {
			LOGGER.error( "Error loading project fileName: {}", fileName, ex );
			result = createEmptyMorphingProjectModel(null);
			setMorphingProjectModel(result);

			throw( new RuntimeException( "Error loading project fileName: " + fileName, ex ) );
		}

		return( result );
	}

	protected void addFaceDetectionsToCommon(FaceDetectionWholeFileModel projectFaceDetectionWholeFileModel) {
		if(projectFaceDetectionWholeFileModel != null) {
			getAppliCtx().getFaceDetectionWholeFileModel().addAll(projectFaceDetectionWholeFileModel.getImageFaceDetectionPersistenceModelMap());
		}
	}

	@Override
	public void saveMorphingProjectModel() throws IOException, TransformerException
	{
		saveMorphingProjectModelAs( getMorphingProjectModel().getFileName() );
	}

	@Override
	public void saveMorphingProjectModelAs(String fileName) throws IOException, TransformerException
	{
		updateVerticesModel();
		updateImageModelFaceDetectionMeshes();
		setFaceDetections();
		getPersistency().saveXmlConfigurationFile(getMorphingProjectModel(), fileName);
	}

	protected FaceDetectionWholeFileModel createFaceDetectionForXml() {
		FaceDetectionWholeFileModel result = createEmptyFaceDetectionModel();

		for(MorphingImageModel imageModel: getProjectModel().getImagesListWrapper().getList()) {
			ImageFaceDetectionPersistenceModel imageFaceDetectionPersistenceModel =
					getAppliCtx().getFaceDetectionWholeFileModel().getImageFaceDetectionPersistenceModelMap().get(imageModel.getFileName());
			if(imageFaceDetectionPersistenceModel != null) {
				result.getImageFaceDetectionPersistenceModelMap().put(imageFaceDetectionPersistenceModel);
			}
		}

		return result;
	}

	protected void setFaceDetections() {
		getProjectModel().setFaceDetectionForXml(createFaceDetectionForXml());
	}

	protected FaceDetectionWholeFileModel createEmptyFaceDetectionModel() {
		return (new FaceDetectionWholeFileModelBase(null));
	}

	protected MorphingProjectWholeFileModel getProjectModel()
	{
		return( _morphingProjectModel );
	}

	protected String getMasterImageName()
	{
		String result = null;
		if( _morphingProjectModel != null )
			result = _morphingProjectModel.getMasterImageName();

		return( result );
	}

//	protected MorphingApplicationConfiguration getApplyConf() {
//		return applyConf;
//	}

	protected MorphingVerticesMapManagerBase createMorphingVerticesMapManager()
	{
		return( new MorphingVerticesMapManagerBase( getAppliCtx(), null ) );
	}

	@Override
	public MorphingImageModel addImage(String imageName, String imageFileName,
									   GeometricRectangle bounds,
									   MorphingProjectConfigurationModel configurationModel,
									   boolean isMaster,
									   int index)
	{
		MorphingImageModel result = createImageModel( imageName, imageFileName,
												bounds, configurationModel,
												isMaster );

		getVerticesManager().addImageModel(index, result);

		return( result );
	}

	protected MorphingImageModel createImageModel( String imageName, String imageFileName,
													GeometricRectangle bounds,
													MorphingProjectConfigurationModel configurationModel,
													boolean isMaster )
	{
		MorphingImageModel result = new MorphingImageModel();
		result.setIsMaster(isMaster);
		result.setFileName(imageFileName);
		result.setName(imageName);
		result.setConfigurationModel(getOrCreateConfigurationModel(configurationModel));

		return( result );
	}

	protected MorphingProjectConfigurationModel getOrCreateConfigurationModel(MorphingProjectConfigurationModel configurationModel)
	{
		MorphingProjectConfigurationModel result = configurationModel;
		MorphingProjectConfigurationModel parent = getProjectModel().getDefaultConfiguration();
		if( result == null )
			result = new MorphingProjectConfigurationModel(parent);
		else
			result.setParentToDelegateTo(parent);

		return( result );
	}

	@Override
	public MorphingProjectWholeFileModel loadProject(String fileName) throws IOException, ParserConfigurationException, SAXException
	{
		MorphingProjectWholeFileModel result = getPersistency().loadModelFromXml(fileName);
		addFaceDetectionsToCommon(result.getFaceDetectionForXml());
		setMorphingProjectModel( result );

		return( result );
	}

/*
	public boolean removeImageModel( int index )
	{
		return( getVerticesManager().removeImageModel( index ) );
	}

	public boolean removeImageModel( String imageName )
	{
		return( getVerticesManager().removeImageModel( imageName ) );
	}

	public void relocateImage( String imageName, int toIndex )
	{
		getVerticesManager().relocateImage(imageName, toIndex);
	}

	public void relocateImage( int fromIndex, int toIndex )
	{
		getVerticesManager().relocateImage(toIndex, toIndex);
	}
*/
	@Override
	public void relocateImages(List<MorphingImageModel> resultingList)
	{
		getVerticesManager().relocateImages( resultingList );
	}

	@Override
	public void deleteImages(int[] indicesToBeDeleted, List<MorphingImageModel> resultingList)
	{
		getVerticesManager().deleteImages( indicesToBeDeleted, resultingList );
	}

	@Override
	public synchronized MorphingVerticesMapManager getVerticesManager() {
		return _verticesManager;
	}

	@Override
	public synchronized void setVerticesManager(MorphingVerticesMapManager _verticesManager) {
		this._verticesManager = _verticesManager;
	}

	@Override
	public synchronized MorphingProjectWholeFileModel getMorphingProjectModel() {
		return _morphingProjectModel;
	}

	@Override
	public void setMorphingProjectModel(MorphingProjectWholeFileModel morphingProjectModel) {
		synchronized( this )
		{
			if( this._morphingProjectModel != morphingProjectModel )
			{
				this._morphingProjectModel = morphingProjectModel;
				MorphingVerticesMapManager verticesManager =  createVerticesManager(morphingProjectModel);

				setVerticesManager( verticesManager );
			}
		}
	}

	protected <C extends CopyFacilities<C>> C copy( C obj, C that )
	{
		return( CopyFacilitiesFunctions.instance().copy(obj, that) );
	}

	@Override
	public void changeImageModel(MorphingImageModel imageModel)
	{
		if( imageModel != null )
		{
			String imageName = imageModel.getName();
			MorphingImageModel existingImageModel = getImageModel( imageName );
			if( existingImageModel != null )
			{
				boolean changeMasterImage = ( imageModel.isMaster() && !existingImageModel.isMaster() );
				if( changeMasterImage )
					updateVerticesModel();

				copy(existingImageModel, imageModel);

				if( changeMasterImage )
					setMasterImageName( imageName, false );
			}
			else
				LOGGER.warn( "Image model to modify not found for name: {}", imageName );
		}
	}

	@Override
	public void setMasterImageName(String masterImageName)
	{
		setMasterImageName( masterImageName, true );
	}

	protected void setMasterImageName( String masterImageName, boolean update )
	{
		if( !Objects.equals( masterImageName, getMasterImageName() ) )
		{
			if( update )
				updateVerticesModel();

			getVerticesManager().setMasterImageName(masterImageName);
		}
	}

	protected void updateImageModelFaceDetectionMeshes()
	{
		getVerticesManager().updateImageModelFaceDetectionMeshes();
	}

	protected void updateVerticesModel()
	{
		getVerticesManager().updateVerticesModel();
	}

	protected MorphingImageModel getImageModel( String imageName )
	{
		return( getMorphingProjectModel().getImageModel(imageName) );
	}
}
