/* 
 * Copyright (C) 2021 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/gpl-3.0.txt
 *
 */
package com.frojasg1.general.jersey.queries;

import com.frojasg1.general.desktop.queries.InetQuery;
import com.frojasg1.general.desktop.queries.InetQueryException;
import com.frojasg1.general.desktop.queries.InetQueryResult;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.net.ssl.*;
import javax.ws.rs.client.ClientBuilder;
import javax.ws.rs.client.WebTarget;
import javax.ws.rs.core.MultivaluedMap;
import javax.ws.rs.core.Response;
import java.security.SecureRandom;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;
import java.util.List;
import java.util.Map;

/**
 *
 * @author fjavier.rojas
 */
public abstract class InetQueryBase2<CC extends InetQueryResult > extends JsonFunctionsBase<CC> implements InetQuery<CC>
{
	private static final Logger LOGGER = LoggerFactory.getLogger(InetQueryBase2.class);
	protected String _uri = null;
	protected String _path = null;

	@Override
	public void init( String uri, String path )
	{
		_uri = uri;
		_path = path;
	}

	public abstract void init( String uri );

	@Override
	public CC queryGen( String ... params ) throws InetQueryException
	{
		return( queryGen( createMultimap( params ) ) );
	}

	public String getUri() {
		return _uri;
	}

	public String getPath() {
		return _path;
	}

	// https://gist.github.com/outbounder/1069465
	protected ClientBuilder createClientBuilder() {
		TrustManager[] certs = new TrustManager[]{new X509TrustManager() {
			@Override
			public void checkClientTrusted(X509Certificate[] chain, String authType) throws CertificateException {

			}

			@Override
			public X509Certificate[] getAcceptedIssuers() {
				return null;
			}

			@Override
			public void checkServerTrusted(X509Certificate[] chain, String authType) throws CertificateException {
			}
		}};

		SSLContext ctx = null;
		try {
			ctx = SSLContext.getInstance("TLS");
			ctx.init(null, certs, new SecureRandom());
		} catch (java.security.GeneralSecurityException e) {
			LOGGER.error("", e);
			throw new RuntimeException(e);
		}

		HttpsURLConnection.setDefaultSSLSocketFactory(ctx.getSocketFactory());

		ClientBuilder clientBuilder = ClientBuilder.newBuilder();
		try {
			clientBuilder.sslContext(ctx);
			clientBuilder.hostnameVerifier(new HostnameVerifier() {
				@Override
				public boolean verify(String hostname, SSLSession session) {
					return true;
				}
			});
		} catch (Exception e) {
			LOGGER.error("", e);
			throw new RuntimeException(e);
		}

//        JERSEY_HTTPS_CLIENT = clientBuilder
//                .withConfig(new ClientConfig())
//                .register(JacksonFeature.class).build();
		return clientBuilder;
	}

	protected WebTarget addQueryParams(WebTarget result, MultivaluedMap<String, String> params) {
		for (Map.Entry<String, List<String>> entry : params.entrySet()) {
			result = result.queryParam(entry.getKey(), entry.getValue().toArray());
		}
		return result;
	}

	public CC queryGen( MultivaluedMap<String, String> params ) throws InetQueryException
	{
		CC result = null;

		javax.ws.rs.client.Client client = createClientBuilder().newClient();
		WebTarget server = client.target(getUri());

		Response response = null;
		try
		{
			response = addQueryParams(server.path(getPath()), params).request().get();
			result = convert( response );
		}
		catch( Exception ex )
		{
			ex.printStackTrace();
			result = createInetQueryResult( ex );
		}

		return( result );
	}

	protected boolean isOk(Response response)
	{
		return(response.getStatus() == Response.Status.OK.getStatusCode() );
	}

	protected CC convert( Response response ) throws InetQueryException
	{
		CC result = null;
		if( response != null )
		{
			if( isOk( response ) )
			{
				String jsonStr = getBody(response);
				result = convert( jsonStr );
			}
			else
			{
				result = createInetQueryResult( createInetQueryExceptionFromNotSuccessfulResponseCode( response ) );
			}
		}
/*
		if( result == null )
		{
			throwInetQueryException( clientResponse );
		}
*/
		return( result );
	}

	// https://stackoverflow.com/questions/18086621/read-response-body-in-jax-rs-client-from-a-post-request
	protected String getBody( Response response )
	{
		String result = null;

		// it works on Jersey 1.x
		String jsonStr = response.readEntity( String.class );

		// for Jersey 2.x
		// String jsonStr = clientResponse.readEntity( String.class );

		result = jsonStr;

		return( result );
	}

	protected InetQueryException createInetQueryExceptionFromNotSuccessfulResponseCode(Response response )
	{
		return( new InetQueryException( response.getStatus(),
										getBody( response ) ) );
	}

	protected void throwInetQueryException( Response response ) throws InetQueryException
	{
		throw( createInetQueryExceptionFromNotSuccessfulResponseCode( response ) );
	}
}
