package com.frojasg1.weka.clustering.model.gaussianmixture;

import com.frojasg1.general.FileFunctions;
import com.frojasg1.general.number.DoubleFunctions;
import com.frojasg1.libgenai.clustering.model.ClustererModel;
import com.frojasg1.libgenai.clustering.model.ClustererTrainer;
import com.frojasg1.libgenai.clustering.model.impl.gaussianmixture.GaussianMixtureConfiguration;
import com.frojasg1.weka.clustering.WekaEmTest;
import com.frojasg1.weka.clustering.factory.LibWekaFactoryImpl;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import org.junit.Before;
import org.junit.Rule;
import org.junit.Test;
import org.junit.rules.TemporaryFolder;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import weka.core.Instance;


import static org.junit.Assert.*;

public class WekaGaussianMixtureContextTest {
    private static final Logger LOGGER = LoggerFactory.getLogger(WekaGaussianMixtureContextTest.class);
    @Rule
    public TemporaryFolder folder= new TemporaryFolder();

    protected LibWekaFactoryImpl factory = LibWekaFactoryImpl.instance();
    protected WekaGaussianMixtureContext instance;
    
    @Before
    public void setUp() {
        instance = (WekaGaussianMixtureContext) factory.createContext();
    }

    @Test
    public void trainSaveLoadCompareTest() throws IOException {
        GaussianMixtureConfiguration config = factory.createConfiguration();
        
        config.setModelName("MyTestGaussianMixtureModel");
        config.setNumClusters(3);
        config.setSeed(0);
        config.setAttributeNames("field-1", "field-2", "field-3");
        config.setMinValueForSelectingCluster(1.0d);
        config.setMaxIterations(10000);

        ClustererTrainer<GaussianMixtureConfiguration> trainer = instance.createTrainer(config);

        List<double[]> dataset = new ArrayList<>();
        dataset.add(array(1.0d, 0d, 0d));
        dataset.add(array(0d, 1.0d, 0d));
        dataset.add(array(0d, 0d, 1.0d));
        dataset.add(array(0.99d, 0.01d, -0.01d));
        dataset.add(array(0.01d, 1.01d, 0.01d));
        dataset.add(array(-0.01d, -0.01d, 1.0d));
        dataset.add(array(1.01d, -0.01d, 0.01d));
        dataset.add(array(-0.01d, 0.99d, -0.01d));
        dataset.add(array(0.01d, 0.01d, 0.99d));

        ClustererModel<GaussianMixtureConfiguration> model = trainer.train(dataset);

        double[] i1 = array(1.0d, 1.0d, 0d);
        double[] i2 = array(0.97d, 0.02d, -0.02d);

        double[] w1 = model.getClusterProbabilities(i1);
        double[] w2 = model.getClusterProbabilities(i2);

        LOGGER.info("Model1");
        logFittingResult(i1, w1);
        logFittingResult(i2, w2);

        File createdFile= folder.newFile("myfile.xml");
        String fileName = createdFile.getAbsolutePath();

        instance.save(model, fileName);

        LOGGER.info("Saved xml: {} \n{}", fileName, FileFunctions.instance().loadTextFileContent(fileName));

        ClustererModel<GaussianMixtureConfiguration> model1 = instance.load(fileName);
        double[] w11 = model1.getClusterProbabilities(i1);
        double[] w12 = model1.getClusterProbabilities(i2);
        logFittingResult(i1, w11);
        logFittingResult(i2, w12);

        myAssertArrayEquals(w1, w11, 0.00003);
        myAssertArrayEquals(w2, w12, 0.00003);
    }

    @Test
    public void testUsingTempFolder() throws IOException {

        File createdFile= folder.newFile("myfile.txt");

        File createdFolder= folder.newFolder("subfolder");

        // ...
    }

    protected void myAssertArrayEquals(double[] array1, double[] array2, double relativeDelta) {
        assertEquals(array1.length, array2.length);
        double v1;
        double v2;
        for(int ii=0; ii<array1.length; ii++) {
            v1 = array1[ii];
            v2 = array2[ii];
            assertEquals(v1, v2, delta(v1, v2, relativeDelta));
        }
    }

    protected DoubleFunctions getDoubleFunctions() {
        return DoubleFunctions.instance();
    }

    public Double max(Double d1, Double d2) {
        return getDoubleFunctions().max(d1, d2);
    }

    public double abs(double dd) {
        return getDoubleFunctions().abs(dd);
    }

    protected double delta(double v1, double v2, double relativeDelta) {
        return max(abs(v1 * relativeDelta), abs(v2 * relativeDelta));
    }

    protected String toString(double[] array) {
        return Arrays.toString(array);
    }

    protected void logFittingResult(double[] instance, double[] logWeights) {
        LOGGER.info("{} -----> {}", toString(instance), toString(logWeights));
    }

    protected double[] array(double ... array) {
        return array;
    }
}