/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.libgenai.image;


import com.frojasg1.general.desktop.image.ImageFunctions;
import com.frojasg1.general.desktop.image.ResizeImageAccurateGen;
import com.frojasg1.general.number.DoubleFunctions;
import com.frojasg1.libgeometry.helper.GenericCropper;
import com.frojasg1.libgeometry.types.GeometricRectangle;
import java.awt.image.BufferedImage;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicReference;
import java.util.function.Function;
import java.util.function.Supplier;

/**
 *
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 */
public class CropAndResizeImageToNormalizedImage extends ResizeImageAccurateGen<Double>
{
	protected static CropAndResizeImageToNormalizedImage _instance = null;
	
	public static CropAndResizeImageToNormalizedImage instance()
	{
		if( _instance == null )
			_instance = new CropAndResizeImageToNormalizedImage();

		return( _instance );
	}

	public NormalizedImage resizeImage(BufferedImage originalImage, int newWidth,
									   int newHeight, int colorForCropping, int toleranceForCropping,
									   Function<Integer, Integer> colorTranslator )
		throws IllegalArgumentException
	{
		return( resizeImageManual( originalImage, newWidth, newHeight, colorForCropping, toleranceForCropping,
				colorTranslator, null ) );
	}

	public NormalizedImage resizeImage(BufferedImage originalImage,
									   int newWidth, int newHeight, int colorForCropping, int toleranceForCropping,
									   AtomicBoolean hasToCancel) throws IllegalArgumentException
	{
		return( resizeImageManual( originalImage, newWidth, newHeight, colorForCropping, toleranceForCropping,
				null, hasToCancel ) );
	}

	protected NormalizedImage resizeImage(BufferedImage originalImage,
										  int newWidth, int newHeight,
										  int colorForCropping, int toleranceForCropping,
										  Function<Integer, Integer> colorTranslator,
										  AtomicBoolean hasToCancel)
		throws IllegalArgumentException
	{
		return( resizeImageManual( originalImage, newWidth, newHeight, colorForCropping, toleranceForCropping,
									colorTranslator, hasToCancel ) );
	}

	protected NormalizedImage resizeImageManual(BufferedImage originalImage,
												int newWidth, int newHeight,
												int colorForCropping, int toleranceForCropping,
												Function<Integer, Integer> colorTranslator,
												AtomicBoolean hasToCancel)
		throws IllegalArgumentException
	{
		NormalizedImage result = null;
		if(originalImage != null) {
			if ((newWidth < 1) || (newHeight < 1))
				throw (new IllegalArgumentException("Bad size for image.   Width: " + newWidth + ". Height: " + newHeight));

			GeometricRectangle bounds = crop(originalImage, colorForCropping, toleranceForCropping);
			int[] pixels = cropAndGetPixels(originalImage, bounds);

			if(pixels.length > 0) {
				result = new NormalizedImage(newWidth, newHeight, null);
				NormalizedImage finalResult = result;

				AtomicReference<Double> pixelValueAccumulation = new AtomicReference<>(0.0d);

				Supplier<Double> pixelAccumulationValueGetter = () -> normalize(pixelValueAccumulation);

				resizeImageManual((int) bounds.width, (int) bounds.height, newWidth, newHeight, colorTranslator,
						pixels,
						() -> resetPixelValue(pixelValueAccumulation),
						(pixelValue, factor) -> addPixelToAccumulation(pixelValueAccumulation, pixelValue, factor),
						pixelAccumulationValueGetter, finalResult::set, hasToCancel);
			}
		}

		return( result );
	}

	protected int grayScale(int rgb) {
		return ImageFunctions.instance().getGrayScale(rgb);
	}

	protected boolean isBackground(BufferedImage originalImage, int xx, int yy,
								   int colorForCropping, int toleranceForCropping) {
		int grayScale = grayScale(originalImage.getRGB(xx, yy));

		return Math.abs(grayScale - colorForCropping) <= toleranceForCropping;
	}

	protected GeometricRectangle crop(BufferedImage originalImage,
									  int colorForCropping, int toleranceForCropping) {
		return GenericCropper.instance().crop(originalImage.getWidth(), originalImage.getHeight(),
				(x, y) -> isBackground(originalImage, x, y, colorForCropping, toleranceForCropping));
	}

	protected int[] cropAndGetPixels(BufferedImage originalImage,
									 GeometricRectangle bounds) {
		return ImageFunctions.instance().getBaW((int) bounds.x, (int) bounds.y, (int) bounds.width, (int) bounds.height, originalImage);
	}

	protected void resetPixelValue(AtomicReference<Double> pixelValueAccumulation) {
		pixelValueAccumulation.set(0.0d);
	}

	protected void addPixelToAccumulation(AtomicReference<Double> pixelValueAccumulation, int grayScale, double factor) {
		pixelValueAccumulation.set(pixelValueAccumulation.get() + grayScale * factor);
	}

	protected double normalize(AtomicReference<Double> pixelValueAccumulation)
	{
		return DoubleFunctions.instance().limit(pixelValueAccumulation.get(), 0.0d, 1.0d);
	}
}
