/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.libgenai.image;

import com.frojasg1.libgeometry.types.GeometricRectangle;
import com.frojasg1.libgeometry.types.factory.GeometryCommonFactory;
import com.frojasg1.general.ExecutionFunctions;
import com.frojasg1.general.number.DoubleFunctions;

import javax.imageio.ImageIO;
import java.awt.image.BufferedImage;
import java.awt.image.ColorModel;
import java.awt.image.WritableRaster;
import java.io.File;
import java.util.function.BinaryOperator;
import java.util.function.Function;

public class ImageHelper {

    protected static ImageHelper INSTANCE = new ImageHelper();

    public static ImageHelper instance() {
        return INSTANCE;
    }

    public int getRed(int rgb) {
        return (rgb >> 16) & 0xff;
    }

    public int getGreen(int rgb) {
        return (rgb >> 8) & 0xff;
    }

    public int getBlue(int rgb) {
        return rgb & 0xff;
    }

    // https://medium.com/javarevisited/converting-rgb-image-to-the-grayscale-image-in-java-9e1edc5bd6e7
    public int getGrey(int rgb) {
        return (int) limit(getRed(rgb) * 0.299d + getGreen(rgb) * 0.587d + getBlue(rgb) * 0.114d);
    }

    public int normalizedGreyToRgb(int previusRgb, double normalizedValue) {
        return normalizedToRgb(previusRgb, normalizedValue, this::interleaveGrey);
    }

    public int normalizedRedToRgb(int previusRgb, double normalizedValue) {
        return normalizedToRgb(previusRgb, normalizedValue, this::interleaveRed);
    }

    public int normalizedGreenToRgb(int previusRgb, double normalizedValue) {
        return normalizedToRgb(previusRgb, normalizedValue, this::interleaveGreen);
    }

    public int normalizedBlueToRgb(int previusRgb, double normalizedValue) {
        return normalizedToRgb(previusRgb, normalizedValue, this::interleaveBlue);
    }

    protected int denormalize(double normalizedValue) {
        return (int) (127.5d + (normalizedValue * 127.5d));
    }

    protected int normalizedToRgb(int previousRgb, double normalizedValue, BinaryOperator<Integer> interleaveChannelFunction) {
        int intValue = (int) limit(denormalize(normalizedValue));
        return interleaveChannelFunction.apply(previousRgb, intValue);
    }

    protected int interleaveRed(int previousRgb, int red) {
        return previousRgb & 0x00ffff | (red << 16) | 0xff000000;
    }

    protected int interleaveGreen(int previousRgb, int green) {
        return previousRgb & 0xff00ff | (green << 8) | 0xff000000;
    }

    protected int interleaveBlue(int previousRgb, int blue) {
        return previousRgb & 0xffff00 | blue | 0xff000000;
    }

    protected int interleaveGrey(int previousRgb, int grey) {
        int result = interleaveRed(previousRgb, grey);
        result = interleaveGreen(result, grey);
        result = interleaveBlue(result, grey);

        return result;
    }

    protected double limit(double byteValue) {
        return DoubleFunctions.instance().limit(byteValue, 0, 255);
    }

    protected double normalize(int rgb, Function<Integer, Integer> scaleExtractor) {
        return (limit(scaleExtractor.apply(rgb)) - 127.5d ) / 127.5d;
    }

    public double normalizeRedChannel(int rgb) {
        return normalize(rgb, this::getRed);
    }

    public double normalizeGreenChannel(int rgb) {
        return normalize(rgb, this::getGreen);
    }

    public double normalizeBlueChannel(int rgb) {
        return normalize(rgb, this::getBlue);
    }

    public double normalizeGrey(int rgb) {
        return normalize(rgb, this::getGrey);
    }

    // https://stackoverflow.com/questions/3514158/how-do-you-clone-a-bufferedimage
    public BufferedImage deepCopy(BufferedImage bi) {
        BufferedImage result = null;
        if(bi != null) {
            ColorModel cm = bi.getColorModel();
            boolean isAlphaPremultiplied = cm.isAlphaPremultiplied();
            WritableRaster raster = bi.copyData(null);
            result = new BufferedImage(cm, raster, isAlphaPremultiplied, null);
        }
        return result;
    }

    public BufferedImage createImage( int width, int height )
    {
        return( new BufferedImage( width, height, BufferedImage.TYPE_INT_ARGB ) );
    }

    protected GeometricRectangle createDelaunayRectangle(double xx, double yy, double width, double height ) {
        return GeometryCommonFactory.instance().createDelaunayRectangle(xx, yy, width, height);
    }

    public GeometricRectangle getBounds(BufferedImage image) {
        GeometricRectangle result = null;
        if(image != null) {
            result = createDelaunayRectangle(0, 0, image.getWidth(), image.getHeight());
        }
        return result;
    }

    public GeometricRectangle getBounds(NormalizedImage normalizedImage) {
        GeometricRectangle result = null;
        if(normalizedImage != null) {
            result = createDelaunayRectangle(0, 0, normalizedImage.getWidth(), normalizedImage.getHeight());
        }
        return result;
    }

    protected <CC> CC runtimeExceptionFunctionExecution( ExecutionFunctions.UnsafeFunction<CC> run ) {
        return ExecutionFunctions.instance().runtimeExceptionFunctionExecution(run);
    }

    public BufferedImage loadImage(String fileName) {
        return runtimeExceptionFunctionExecution(() -> ImageIO.read(new File(fileName)));
    }

    public boolean saveTiffImage( BufferedImage image, String longFileName )
    {
        return saveImage(image, longFileName, "TIFF");
    }

    public boolean saveImage(BufferedImage image, String fileName, String type) {
        return runtimeExceptionFunctionExecution(() -> ImageIO.write(image, type, new File(fileName)));
    }
}
