/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.libgenai.image;

import com.frojasg1.libgeometry.types.Point2d;
import com.frojasg1.general.number.DoubleFunctions;

import java.util.Arrays;

public class NormalizedImage {

    protected double[][] pixelChannels;

    protected NormalizedRange range;

    public NormalizedImage(int width, int height, NormalizedRange range) {
        pixelChannels = new double[width][height];
        this.range = range;
    }

    public NormalizedImage(int width, int height, double initialPixelValue, NormalizedRange range) {
        this(width, height, range);

        initWithValue(initialPixelValue);
    }

    public void initWithValue(double pixelValue) {
        int height;
        for(double[] line: pixelChannels) {
            height = line.length;
            for(int ii=0; ii<height; ii++) {
                line[ii] = pixelValue;
            }
        }
    }

    public NormalizedImage createCopy() {
        NormalizedImage result = new NormalizedImage(getWidth(), getHeight(), getRange());
        result.copy(this);

        return result;
    }

    public void copy(NormalizedImage that) {
        for(int ii=0; ii<getWidth(); ii++) {
            pixelChannels[ii] = Arrays.copyOf(that.pixelChannels[ii], that.pixelChannels[ii].length);
        }
    }

    public int getWidth() {
        return pixelChannels.length;
    }

    public int getHeight() {
        return pixelChannels[0].length;
    }

    public void set(double xx, double yy, double value) {
        pixelChannels[(int)xx][(int)yy] = limit(value);
    }

    public NormalizedRange getRange() {
        return range;
    }

    protected double limit(double value) {
        double result = value;
        NormalizedRange range = getRange();
        if(range != null) {
            result = range.limit(value);
        }
        return result;
    }

    protected double limit(double value, double lowerBound, double upperBound) {
        return DoubleFunctions.instance().limit(value, lowerBound, upperBound);
    }

    protected double limitX(double xx) {
        return limit(xx, 0, pixelChannels.length - 1);
    }

    protected double limitY(double yy) {
        return limit(yy, 0, pixelChannels[0].length - 1);
    }

    public double get(Point2d pt) {
        return pixelChannels[(int)limitX(pt.x)][(int)limitY(pt.y)];
    }

    public double get(int xx, int yy) {
        return pixelChannels[xx][yy];
    }

    public void checkCompatibility(NormalizedImage normalizedImage) {
        if(!isCompatible(normalizedImage)) {
            throw new RuntimeException(
                    String.format("NormalizedImages are not compatible: %s and %s",
                            this, normalizedImage));
        }
    }

    public boolean isCompatible(NormalizedImage normalizedImage) {
        return (normalizedImage != null) &&
                ( this.getWidth() == normalizedImage.getWidth() ) &&
                ( this.getHeight() == normalizedImage.getHeight() );
    }

    @Override
    public String toString() {
        return "NormalizedImage{" +
                String.format("pixelChannels=double[%d][%d]", getWidth(), getHeight()) +
                '}';
    }
}
