/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.libgenai.metrics;

import java.util.Iterator;
import java.util.List;
import java.util.function.BinaryOperator;

public class OneDMetrics {

    protected static OneDMetrics instance = new OneDMetrics();

    public static OneDMetrics instance() {
        return instance;
    }

    public double mean(List<Double> list) {
        double result = 0;

        for (Double value : list) {
            result += value;
        }

        return list.isEmpty() ? result : result / list.size();
    }


    public double variance(List<Double> list) {
        return variance(list, mean(list));
    }

    public double variance(List<Double> list, double mean) {
        double result = 0;

        double delta;
        for (Double value : list) {
            delta = value - mean;
            result += delta * delta;
        }

        return list.isEmpty() ? result : result / list.size();
    }

    public double variance(double[] array) {
        return variance(array, mean(array));
    }

    public double variance(double[] array, double mean) {
        double result = 0;

        double delta;
        for (double value : array) {
            delta = value - mean;
            result += delta * delta;
        }

        return array.length == 0 ? result : result / array.length;
    }

    public double max(double[] array) {
        double result = Double.MIN_VALUE;

        for(double value: array) {
            if(value > result) {
                result = value;
            }
        }

        return result;
    }

    public double mean(double[] array) {
        double result = 0;

        for (double value : array) {
            result += value;
        }

        return array.length == 0 ? result : result / array.length;
    }

    public double averageMetric(List<Double> list1, List<Double> list2, ScalarMetrics scalarMetrics) {
        if(list1.size() != list2.size()) {
            throw new IllegalArgumentException(String.format("Lists where not of the same size (List1[%d], List2[%d])",
                    list1.size(), list2.size()));
        }

        BinaryOperator<Double> termToAddCalculator = scalarMetrics.getTermToAddCalculator();
        int size = list1.size();
        double result = 0;
        Iterator<Double> it1 = list1.iterator();
        Iterator<Double> it2 = list2.iterator();
        while(it1.hasNext()) {
            result += termToAddCalculator.apply(it1.next(), it2.next());
        }
        result /= size;

        return result;
    }

    public double averageMetric(double[] arr1, double[] arr2, ScalarMetrics scalarMetrics) {
        if(arr1.length != arr2.length) {
            throw new IllegalArgumentException(String.format("Arrays where not of the same size (double[%d], double[%d])",
                    arr1.length, arr2.length));
        }

        BinaryOperator<Double> termToAddCalculator = scalarMetrics.getTermToAddCalculator();
        int size = arr1.length;
        double result = 0;
        for(int ii=0; ii<arr1.length; ii++) {
            result += termToAddCalculator.apply(arr1[ii], arr2[ii]);
        }
        result /= size;

        return result;
    }
}
