/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.general.arrays.impl;

import com.frojasg1.general.ArrayFunctions;
import com.frojasg1.general.arrays.ResizableArray;

/**
 *
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 */
public abstract class ResizableArrayImpl<EE> implements ResizableArray<EE>
{
	protected EE[] _array;
	protected int _numElems;
	protected int _step;
	protected int _initialNumElems;

	public ResizableArrayImpl()
	{
		this( 300, 1000 );
	}

	public ResizableArrayImpl( int initialNumElems, int step )
	{
		this( initialNumElems, step, true );
	}

	public ResizableArrayImpl( int initialNumElems, int step, boolean init )
	{
		_initialNumElems = initialNumElems;
		_step = step;
		
		if( init )
			init();
	}

	protected void init()
	{
		_array = createArray( _initialNumElems );
	}

	protected abstract EE[] createArray( int numElems );

	@Override
	public EE get(int index) {
		return( _array[index] );
	}

	@Override
	public int add(EE elem) {
		int ind = _numElems++;
		if( ind >= _array.length )
			resizeArray( ind );

		_array[ind] = elem;
		return( ind );
	}

	protected synchronized void resizeArray( int numElems )
	{
		int totalElems = calculateTotalElems( numElems );
		if( _array.length < totalElems )
			_array = createArray( totalElems );
	}

	protected int calculateTotalElems( int numElems )
	{
		return( (int) ( _initialNumElems +
			Math.ceil( ( numElems - _initialNumElems ) / _step ) * _step ) );
	}

	@Override
	public EE[] getArray()
	{
		return( _array );
	}

	@Override
	public void changeState( ResizableArray<EE> that )
	{
		if( !( that instanceof ResizableArrayImpl ) )
			throw( new IllegalArgumentException( "Cannot assign from: " +
					( ( that == null ) ? null : that.getClass() ) ) );

		copyInternal( (ResizableArrayImpl) that, this );
	}

	protected void copyInternal( ResizableArrayImpl<EE> from, ResizableArrayImpl<EE> to )
	{
		to._array = ArrayFunctions.instance().copy(from.getArray());
		to._numElems = from._numElems;
		to._step = from._step;
		to._initialNumElems = from._initialNumElems;
	}

	protected abstract ResizableArrayImpl<EE> createEmptyCopy();

	@Override
	public ResizableArrayImpl<EE> copyState()
	{
		ResizableArrayImpl<EE> result = createEmptyCopy();

		copyInternal(this, result);

		return( result );
	}
}
