/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.general.changestate.staterevertible;

import java.util.LinkedList;
import com.frojasg1.general.changestate.ChangeableState;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

/**
 *
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 */
public class AtomicReversionList {
	protected LinkedList<ReversionElement> _list;
	protected Set _set;

	public AtomicReversionList()
	{
		this( true );
	}

	protected AtomicReversionList( boolean init )
	{
		if( init )
			init();
	}

	protected void init()
	{
		_list = new LinkedList<>();
		_set = new HashSet<>();
	}

	public synchronized <C extends ChangeableState<C>> ReversionElement<C> add( C elem )
	{
		ReversionElement<C> result = null;

		if( !contains(elem) )
		{
			result = new ReversionElement<>( elem );
			_list.add(result);
		}

		return( result );
	}

	protected synchronized boolean contains( Object obj )
	{
		boolean result = _set.contains(obj);

		if( ! result )
			_set.add(obj);

		return( result );
	}

	public synchronized void revert()
	{
		Iterator<ReversionElement> it = _list.descendingIterator();
		while( it.hasNext() )
			it.next().revert();
	}

	public synchronized void clear()
	{
		_list.clear();
		_set.clear();
	}

	public synchronized boolean isEmpty()
	{
		return( _list.isEmpty() );
	}

	protected static class ReversionElement<CC extends ChangeableState<CC>>
	{
		protected CC _elem;
		protected CC _previousStateCopy;

		public ReversionElement( CC elem )
		{
			_elem = elem;
			_previousStateCopy = elem.copyState();
		}

		public void revert()
		{
			_elem.changeState( _previousStateCopy );
		}
	}
}
