/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.general.changestate.staterevertible;

import java.util.ArrayList;
import java.util.List;

/**
 *
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 */
public class MultiUndoRedoList
{
	protected List<AtomicUndoRedo> _list;

	protected Integer _index;

	public MultiUndoRedoList()
	{
		this( true );
	}

	protected MultiUndoRedoList( boolean init )
	{
		if( init )
			init();
	}

	protected void init()
	{
		_list = new ArrayList<>();
	}

	public synchronized boolean canUndo()
	{
		return( ! _list.isEmpty() &&
			( ( _index == null ) || ( _index > 0 ) ) );
	}

	protected synchronized int getIndexForUndo()
	{
		if( _index == null )
			_index = _list.size();

		_index--;
		if( _index < 0 )
		{
			if( _list.isEmpty() )
				_index = null;
			else
				_index = 0;
			throw( new IllegalStateException( "There are not more elements to undo" ) );
		}

		return( _index );
	}

	public synchronized void undo()
	{
		AtomicUndoRedo aur = _list.get( getIndexForUndo() );
		aur.undo();

		// if operation cannot be redone, we remove everything beyond this step
		if( !aur.canRedo() )
			removeRemaining();
	}

	public synchronized void redo()
	{
		if( _index == null )
			throw( new IllegalStateException( "Cannot redo if there is no undo to redo" ) );

		_list.get( _index ).redo();

		_index++;
		if( _index >= _list.size() )
			_index = null;
	}

	public void add( AtomicReversionList statesToRevert )
	{
		add( statesToRevert::revert, null );
	}

	protected AtomicUndoRedo createAtomicUndoRedo( Runnable undoTask, Runnable redoTask )
	{
		return( new AtomicUndoRedo( undoTask, redoTask ) );
	}

	public synchronized void add( Runnable undoTask, Runnable redoTask )
	{
		add( createAtomicUndoRedo( undoTask, redoTask ) );
	}

	protected synchronized void removeRemaining()
	{
		if( _index != null )
		{
			while( _list.size() > _index )
				_list.remove( (int) _index );

			_index = null;
		}
	}

	protected synchronized void add( AtomicUndoRedo aur )
	{
		removeRemaining();
		_list.add( aur );
	}

	protected class AtomicUndoRedo
	{
		protected Runnable _undoTask;
		protected Runnable _redoTask;

		public AtomicUndoRedo( Runnable undoTask, Runnable redoTask )
		{
			_undoTask = undoTask;
			_redoTask = redoTask;
		}

		public void undo()
		{
			_undoTask.run();
		}

		public void redo()
		{
			_redoTask.run();
		}

		public boolean canRedo()
		{
			return( _redoTask != null );
		}
	}
}
