/*
 * Copyright (c) 2023. Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 3.0 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You may obtain a copy of the License at
 *
 *       http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */

package com.frojasg1.general.functional;

import java.util.Arrays;
import java.util.List;
import java.util.function.Function;
import java.util.function.Supplier;
import java.util.function.UnaryOperator;

public class FunctionalInterfaceFunctions {

    protected static FunctionalInterfaceFunctions INSTANCE = new FunctionalInterfaceFunctions();

    public static FunctionalInterfaceFunctions instance() {
        return INSTANCE;
    }

    public <C> C getFirst(Supplier<C> ... suppliers) {
        C result = null;
        if(suppliers != null) {
            for(Supplier<C> supplier: suppliers) {
                result = supplier.get();
                if(result != null) {
                    break;
                }
            }
        }

        return result;
    }

    public <C> C apply(UnaryOperator<C> operator, C value) {
        C result = value;
        if(operator != null) {
            result = operator.apply(value);
        }
        return result;
    }

    public <C> C transformationChain(C value, UnaryOperator<C> ... transformationOperators) {
        C result = value;
        for(UnaryOperator<C> operator: transformationOperators) {
            if(operator != null) {
                result = operator.apply(result);
            }
        }

        return result;
    }

    public <C> C transformationChain(C value, List<UnaryOperator<C>> operators) {
        C result = value;
        if(operators != null) {
            for(UnaryOperator<C> operator: operators) {
                result = operator.apply(result);
            }
        }

        return result;
    }

    public <C> List<C> asList(C ... elems) {
        return Arrays.asList(elems);
    }

    public <I, O> O map(I value, List<UnaryOperator<I>> preprocessors, Function<I, O> mapper, UnaryOperator<O> ... postprocessors) {
        I processedInput = transformationChain(value, preprocessors);
        O result = mapper.apply(processedInput);
        result = transformationChain(result, postprocessors);

        return result;
    }
}
