/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.general.list;

import java.util.List;
import java.util.function.Supplier;

/**
 *
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 */
public class ConcurrentListWrapper<VV> extends ListWrapper<VV>
{
	public ConcurrentListWrapper()
	{
		this( true );
	}

	protected ConcurrentListWrapper( boolean init )
	{
		if( init )
			init();
	}

	@Override
	protected void init()
	{
		super.init();
	}

	@Override
	public void add( VV value )
	{
		executeSynchronized( () -> super.add(value) );
	}

	@Override
	public void add( int index, VV value )
	{
		executeSynchronized( () -> super.add(index, value) );
	}

	@Override
	public boolean remove( VV value )
	{
		return( executeSynchronized( () -> super.remove(value) ) );
	}

	@Override
	public void clear()
	{
		executeSynchronized( super::clear );
	}

	@Override
	public int size()
	{
		return( executeSynchronized( super::size ) );
	}

	@Override
	public List<VV> getListCopy()
	{
		return( executeSynchronized( super::getListCopy ) );
	}

	@Override
	public boolean contains( VV value )
	{
		return( executeSynchronized( () -> super.contains(value) ) );
	}

	@Override
	public void addAll( List<VV> that )
	{
		executeSynchronized( () -> super.addAll(that) );
	}

	@Override
	public void replace( List<VV> that )
	{
		executeSynchronized( () -> super.replace(that) );
	}

	@Override
	public VV get( int index )
	{
		return( executeSynchronized( () -> super.get(index) ) );
	}

	@Override
	public int indexOf( VV elem )
	{
		return( executeSynchronized( () -> super.indexOf(elem) ) );
	}

	protected void executeSynchronized( Runnable method )
	{
		executeSynchronized( () -> synchronizedMethod( method ) );
	}

	protected Object synchronizedMethod( Runnable method )
	{
		method.run();
		return null;
	}

	protected <R> R executeSynchronized( Supplier<R> function )
	{
		R result = null;
		List<VV> list = getList();
		synchronized( list )
		{
			result = function.get();
		}
		return result;
	}
}
