/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.general.map;

import com.frojasg1.general.copy.CopyFacilities;

import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.stream.Collectors;

/**
 *
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 */
public class MapWrapper<KK, VV> implements CopyFacilities<MapWrapper<KK, VV>>
{
	protected MapWrapper<KK, VV> _parent;
	protected Map<KK, VV> _map;

	public MapWrapper()
	{
		this( true );
	}

	protected MapWrapper( boolean init )
	{
		if( init )
			init();
	}

	protected void init()
	{
		_map = createMap();
	}

	@Override
	public MapWrapper<KK, VV> createCopy()
	{
		MapWrapper<KK, VV> result = new MapWrapper<>(false);
		result.copy(this);

		return( result );
	}

	@Override
	public void copy( MapWrapper<KK, VV> that )
	{
		_parent = that._parent;
		_map = createMapCopy( that._map );
	}

	protected Map<KK, VV> createMap()
	{
		return( new ConcurrentHashMap<>() );
	}

	protected Map<KK, VV> createMapCopy( Map<KK, VV> that)
	{
		Map<KK, VV> result = null;
		if( that != null )
		{
			result = createMap();
			result.putAll(that);
		}

		return( result );
	}

	public VV get( KK key )
	{
		VV result = _map.get(key);
		if( result == null )
			result = delegateGet(key);

		return( result );
	}

	protected VV delegateGet(KK key)
	{
		MapWrapper<KK, VV> parent = getParent();
		VV result = null;
		if( parent != null )
			result = parent.get(key);

		return( result );
	}

	public void put( KK key, VV value )
	{
		_map.put(key, value);
	}

	public Set<KK> keySet()
	{
		return( _map.keySet() );
	}

	public Set<Map.Entry<KK, VV>> entrySet()
	{
		return( _map.entrySet() );
	}

	public VV remove( KK key )
	{
		return( _map.remove(key) );
	}

	public void clear()
	{
		_map.clear();
	}

	public int size()
	{
		return( _map.size() );
	}

	public MapWrapper<KK, VV> getParent() {
		return _parent;
	}

	public void setParent(MapWrapper<KK, VV> _parent) {
		this._parent = _parent;
	}
//
//	protected boolean equals(Map map1, Map map2) {
//		boolean result = (map1 == map2);
//		if(!result && (map1 != null) && (map2 != null) && (map1.size() == map2.size())) {
//			for(Object key: map1.keySet()) {
//				result = map2.containsKey(key) && Objects.equals(map1.get(key), map2.get(key));
//				if(!result) {
//					break;
//				}
//			}
//		}
//		return result;
//	}
//
//	protected List<KK> sortedKeys(Map<KK, VV> map) {
//		return map.keySet().stream().collect(Collectors.toList())
//				.stream().sorted().collect(Collectors.toList());
//	}

	@Override
	public boolean equals(Object o) {
		if (this == o) return true;
		if (o == null || getClass() != o.getClass()) return false;
		MapWrapper<?, ?> that = (MapWrapper<?, ?>) o;
		return Objects.equals(_parent, that._parent) && Objects.equals(_map, that._map);
	}

//	protected int hash(Map<KK, VV> map) {
//		int result = 0;
//		int factor = 499;
//
//		for(KK key: sortedKeys(map)) {
//			result += key.hashCode() * factor;
//			result += map.get(key).hashCode() * factor;
//		}
//		return result;
//	}

	@Override
	public int hashCode() {
		return Objects.hash(_parent, _map);
	}
}
