/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.general.map.multimap;

import java.util.Map;
import java.util.Objects;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.function.Consumer;
import java.util.function.Function;
import org.slf4j.LoggerFactory;

/**
 *
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 * @param <VV>	- MultiValue
 * @param <EE>	- Map Entry
 */
public abstract class MultiMapGen<VV, EE extends MultiMapMapEntry<?, VV>> {
    private static final org.slf4j.Logger LOGGER = LoggerFactory.getLogger(MultiMapGen.class);

    protected Map<String, EE> mapByName;

	public MultiMapGen()
	{
		this(false);
	}

	public MultiMapGen(boolean init)
	{
		if( init )
			init();
	}
	
	protected void init() {
        mapByName = fillMultiMapMapEntryMap();
    }

    protected abstract Map<String, EE> fillMultiMapMapEntryMap();

    protected <K> EE addMultiMapMapEntry( Map<String, EE> map,
										String name, Class<K> keyClass,
										Function<VV, K> keyGetter)
	{
		EE result = createMultiMapMapEntry( name, keyClass, keyGetter );
		addMultiMapMapEntry( map, name, result );

		return( result );
	}

    protected void addMultiMapMapEntry( Map<String, EE> map,
										String name, EE mme)
	{
		map.put(name, mme );
	}

	public void clearAll()
	{
		getMapByName().clear();
	}

	public Set<String> getNames()
	{
		return( getMapByName().keySet() );
	}

	protected Map<String, EE> getMapByName() {
		return mapByName;
	}

	protected abstract <K> EE createEmptyMultiMapMapEntry(Class<K> keyClass,
														Function<VV, K> keyGetter);

	protected <K> EE createMultiMapMapEntry( String name,
											Class<K> keyClass,
											Function<VV, K> keyGetter )
	{
		EE result = createEmptyMultiMapMapEntry(keyClass, keyGetter);

		result.setName(name);

		return( result );
	}

	protected <K, V> Map<K, V> createMap() {
        return new ConcurrentHashMap<>();
    }

	protected EE getMultiMapElement(String mapName) {
		return mapByName.get(mapName);
	}

    public <K> Map<K, VV> getMapByName(String mapName, Class<K> keyClass) {
		Map<K, VV> result = null;
		EE multiMapElement = getMultiMapElement(mapName);
		if( multiMapElement != null )
			result = (Map<K, VV>) multiMapElement.getMap();

        return( result );
    }

	public boolean mapExists(String mapName) {
		return getMultiMapElement(mapName) != null;
	}

    public <K> VV remove(String mapName, K key) {
        EE me = getAndCheckMultiMapMapEntryForKey(mapName, key);
        VV result = remove((VV) me.getMap().get(key));

        return result;
    }

    public VV remove(VV value) {
        VV result = null;
        if (value != null) {
            for (Map.Entry<String, EE> entry : mapByName.entrySet()) {
                EE me = entry.getValue();
                Object key = me.getKeyGetter().apply(value);
                VV tmp = (VV) me.getMap().remove(me.getKeyClass().cast(key));
                if (tmp != null) {
                    if (!Objects.equals(tmp, value) || (result != null) && (result != value)) {
                        throw new RuntimeException(String.format("Error trying to remove value: %s. " +
                                        "Existing different element found: '%s' " +
                                        "for mapName='%s', cannot remove",
                                value, tmp, me.getName()));
                    }
                    result = value;
                }
            }
        }

        return result;
    }

    public <K> VV get(String mapName, K key) {
        EE me = getAndCheckMultiMapMapEntryForKey(mapName, key);
        VV result = (VV) me.getMap().get(key);

        return result;
    }

    protected <K> EE getAndCheckMultiMapMapEntryForKey(String mapName, K key) {
        EE result = null;
        if (key != null) {
            result = mapByName.get(mapName);

            if (result == null) {
                throw new IllegalArgumentException(String.format("Map with name: %s not found", mapName));
            } else if (!result.getKeyClass().isInstance(key)) {
                throw new IllegalArgumentException(String.format("key(%s) was not instance of %s",
                        key.getClass(), result.getKeyClass()));
            }
        } else {
            throw new IllegalArgumentException("key cannot be null");
        }

        return result;
    }

    public void put(VV value) {
        if (value != null) {
            for (Map.Entry<String, EE> entry : mapByName.entrySet())
                entry.getValue().put(value);
        } else {
            throw new IllegalArgumentException("value was null, cannot put");
        }
    }
	
	public void clear()
	{
		for( EE mmme : mapByName.values() )
			mmme.getMap().clear();
	}

	protected EE removeEntryByNameInternal( String name )
	{
		return( mapByName.remove(name) );
	}

	protected boolean canRemoveEntryNameMap( String name )
	{
		return( true );
	}

	public boolean removeEntryByName( String name )
	{
		boolean result = false;
		if( canRemoveEntryNameMap( name ) )
			result = hardModifyMapEntryName( name, mme -> mme.removeName( name ) );

		return( result );
	}

	protected void removeEntryByName( EE mme, String name )
	{
		mme.removeName( name );
	}

	protected boolean canRenameEntryMapName( String oldName, String newName )
	{
		return ( getMultiMapEntry(oldName) != null ) &&
				( getMultiMapEntry(newName) == null );
	}

	public boolean renameMapEntryName( String oldName, String newName )
	{
		boolean result = false;
		if( canRenameEntryMapName( oldName, newName ) )
			result = hardModifyMapEntryName( oldName, mme -> renameMapEntryName(mme, oldName, newName ) );

		return( result );
	}

	protected void renameMapEntryName( EE mme, String oldName, String newName )
	{
		mme.rename(newName);
		mapByName.put(newName, mme);
	}

	protected boolean hardModifyMapEntryName( String oldName,
												Consumer<EE> modificationFunction )
	{
		boolean result = false;
		EE mme = mapByName.remove(oldName);

		result = ( mme != null );
		if( result )
			modificationFunction.accept(mme);

		return( result );
	}

    protected EE getMultiMapEntry( String name )
	{
		return( getMapByName().get(name) );
	}
}
