/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.general.map.multimap;

import com.frojasg1.general.map.MapFunctions;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.function.Function;

/**
 *
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 */
public class MultiMapMapEntry<KK, VV> {
    protected String name;
    protected Class<KK> keyClass;
    protected volatile Map<KK, VV> map;
    protected Function<VV, KK> keyGetter;

	public MultiMapMapEntry()
	{
		
	}

	public MultiMapMapEntry( Class<KK> keyClass, Function<VV, KK> keyGetter )
	{
		this.keyClass = keyClass;
		this.keyGetter = keyGetter;
	}

	public MultiMapMapEntry<KK, VV> setName(String name) {
        this.name = name;
        return this;
    }

    public MultiMapMapEntry<KK, VV> setKeyClass(Class<KK> keyClass) {
        this.keyClass = keyClass;
        return this;
    }

    public MultiMapMapEntry<KK, VV> setKeyGetter(Function<VV, KK> keyGetter) {
        this.keyGetter = keyGetter;
        return this;
    }

    public String getName() {
        return name;
    }

    public Class<KK> getKeyClass() {
        return keyClass;
    }

    protected synchronized Map<KK, VV> createMapIfNull() {
        if (map == null) {
            Map<KK, VV> tmp = createMap();
            map = tmp;
        }
        return map;
    }

	public synchronized Map<KK, VV> getMapCopy()
	{
		return( MapFunctions.instance().copyMap( getMap() ) );
	}

	public Map<KK, VV> getMap() {
        if (map == null) {
            createMapIfNull();
        }
        return map;
    }

	public synchronized void setMap(Map<KK, VV> map) {
		this.map = map;
	}

    public Function<VV, KK> getKeyGetter() {
        return keyGetter;
    }

    protected <K, V> Map<K, V> createMap() {
        return new ConcurrentHashMap<>();
    }

	public void rename(String newName)
	{
		setName(newName);
	}

	public void removeName(String name)
	{
		// intentionally left blank
		// derived classes can implement it
	}

	protected KK getKeyAndCheckThatDoesNotExist( VV value )
	{
        KK key = getKeyGetter().apply(value);
		if( key != null )
		{
			VV existing = (VV) getMap().get(key);
			if ((existing != null) && !existing.equals(value)) {
				throw new IllegalArgumentException(String.format("Cannot insert value='%s', as key existed for mapName=%s " +
								"existing pair in that map: ( %s, %s )",
						value, getName(), key, existing));
			}
		}

		return( key );
	}

	public synchronized void put( VV value )
	{
		KK key = getKeyAndCheckThatDoesNotExist( value );

		getMap().put(key, value);
	}
}
