/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.general.map.multimap.dynamic.deductible;

import com.frojasg1.general.DateFunctions;
import com.frojasg1.general.map.MapFunctions;
import com.frojasg1.general.map.multimap.dynamic.DynamicMultiMapGen;
import com.frojasg1.general.map.multimap.dynamic.DynamicMultiMapMapEntry;
import java.util.Date;
import java.util.Map;
import java.util.Objects;
import java.util.concurrent.atomic.AtomicReference;
import java.util.function.Function;
import java.util.function.UnaryOperator;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 *
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 * @param <EV>	- element value (key)
 * @param <VV>	- value (multi-element value)
 * @param <CC>	- element Context
 */
public abstract class DeductibleDynamicMultiMapGen<EV,
												VV extends DeductibleDynamicMultiMapElementBase<EV>,
												CC extends DeductibleDynamicMultiMapMapEntryContext<EV>,
												EE extends DynamicMultiMapMapEntry<EV, VV, CC>>
	extends DynamicMultiMapGen<EV, VV, CC, EE>
{
	private static final Logger LOGGER = LoggerFactory.getLogger(DeductibleDynamicMultiMapGen.class);

	protected AtomicReference<String> _masterNameRef;

	public DeductibleDynamicMultiMapGen(Class<EV> keyClass, String masterName) {
		this(keyClass, masterName, false);
	}

	protected DeductibleDynamicMultiMapGen(Class<EV> keyClass, String masterName,
										boolean init) {
		super(keyClass);
		_masterNameRef = new AtomicReference<>(masterName);

		if( init )
			init();
	}

	public String getMasterName() {
		return _masterNameRef.get();
	}

	public void setMasterName(String masterName) {
		_masterNameRef.set(masterName);
	}

	protected EE getMasterMap()
	{
		EE result = null;
		String masterName = getMasterName();
		if( masterName != null )
			result = getMapByName().get( masterName );

		return( result );
	}

	public void addMapByName( String name, CC elementContext )
	{
		EE dmme = addMultiMapMapEntry( getMapByName(), name, elementContext );

		EE master = getMasterMap();

		if( master != null )
		{
			// the keys are of master
			// we will switch them when browsing the elements.
			dmme.setMap( master.getMapCopy() );

			Map<EV, VV> map = master.getMap();
			MapFunctions.instance().browseMap( map, (k, v) -> addElementToNewMapName(dmme, k, v) );
		}
	}

	@Override
	protected abstract <K> EE createEmptyMultiMapMapEntry( Class<K> keyClass, Function<VV, K> keyGetter );

	protected EE addMultiMapMapEntry( Map<String, EE> map, String name, CC elementContext)
	{
		if( mapByName.get(name) != null )
			throw( new IllegalArgumentException( "Cannot add " + name +
						" MultiMapMapElement. Already existed." ) );

		EE result = createEmptyMultiMapMapEntry(name);
		result.setElementContext(elementContext);

		addMultiMapMapEntry( map, name, result );

		return( result );
	}

	protected abstract EV initialTranslationFromMasterKey( EE dmme, EV masterElement, VV multiElementValue );

	protected EV addElementToNewMapName(EE dmme, EV masterElement, VV multiElementValue )
	{
		String name = dmme.getName();
		EV result = initialTranslationFromMasterKey( dmme, masterElement, multiElementValue );
		EV previous = multiElementValue.get(name);
		multiElementValue.put( name, result );

		VV masterMultiElem = get(getMasterName(), masterElement);
		VV existingMultiElem = get(name, result);
		if(masterMultiElem != multiElementValue) {
			LOGGER.warn("the multielem was not the master one");
		}
		if((existingMultiElem != null) && (existingMultiElem != multiElementValue)) {
			LOGGER.warn("The new element already existed at map('{}'), masterElem: {}, existing elem: {}, for point: {}",
					name, multiElementValue, existingMultiElem, result);
		}


		if(previous != null) {
			dmme.getMap().put( result, dmme.getMap().remove(previous) );
		} else {
			dmme.getMap().put(result, multiElementValue);
		}

		return( result );
	}

	@Override
	protected Map<String, EE> fillMultiMapMapEntryMap() {
		// intentionally left blank
		// we will add MultiMap entries on the fly, as the elements are added dynamically
		return( createMap() );
	}

	protected boolean isMaster( String name )
	{
		return( Objects.equals( getMasterName(), name ) );
	}

//	protected abstract UnaryOperator<EV> getFromMasterDirectTranslator( EE dmme );

	protected UnaryOperator<EV> getToMasterMapper( EE dmme )
	{
		return( dmme.getElementContext().getToMasterTranslator() );
	}

	protected UnaryOperator<EV> getFromMasterMapper( EE dmme )
	{
		return( dmme.getElementContext().getFromMasterTranslator() );
	}

	public VV addNewElement( String name, EV elementValue )
	{
		EE dmme = getMultiMapEntry( name );

		EV masterElement;
		if( isMaster( name ) )
			masterElement = elementValue;
		else
			masterElement = getToMasterMapper(dmme).apply(elementValue);

		return( addNewMasterElement( masterElement ) );
	}

	protected abstract VV createEmptyMultiValue();

	protected VV createMultiValue( EV masterElement )
	{
		VV result = createEmptyMultiValue();

		MapFunctions.instance().browseMap( getMapByName(),
			(key, dmme) ->
				populateMultiValue(key, dmme, masterElement, result ) );

		return( result );
	}

	protected boolean canRemoveEntryNameMap( String name )
	{
		return( ! isMaster(name) );
	}

	protected void populateMultiValue(String name, EE dmme, EV masterElement, VV result )
	{
		result.put(name, getFromMasterMapper(dmme).apply(masterElement));
	}

	protected String nowString()
	{
		return( DateFunctions.instance().formatDateTime(new Date()) );
	}

	protected VV addNewMasterElement( EV masterElement )
	{
		VV result = createMultiValue( masterElement );
		put( result );

		logInsertion( masterElement, result );

		return( result );
	}

	protected abstract void logInsertion( EV masterElement, VV result );

	public boolean renameMapEntryName( String oldName, String newName )
	{
		boolean result = super.renameMapEntryName(oldName, newName);
		if( result && isMaster(oldName) )
			setMasterName( newName );

		return( result );
	}
}
