/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.general.number;

import com.frojasg1.general.ExecutionFunctions;
import static com.frojasg1.general.number.IntegerFunctions.max;
import static com.frojasg1.general.number.IntegerFunctions.min;

import com.frojasg1.general.exceptions.IllegalCallException;
import com.frojasg1.general.exceptions.IllegalValueException;
import com.frojasg1.general.string.StringFunctions;
import java.text.DecimalFormat;
import java.text.NumberFormat;
import java.util.Locale;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;

public class DoubleFunctions
{
	
	protected static DoubleFunctions _instance;

	protected NumberFormat _formatter = null;

	protected Map<Locale, NumberFormat> _formatterMap = new ConcurrentHashMap<>();

	public static void changeInstance( DoubleFunctions inst )
	{
		_instance = inst;
	}

	public static DoubleFunctions instance()
	{
		if( _instance == null )
			_instance = new DoubleFunctions();
		return( _instance );
	}

	public DoubleFunctions()
	{
		_formatter = createNumberFormat();
	}

	protected NumberFormat createNumberFormat()
	{
		return( new DecimalFormat("#0.00000000") );
	}

	public String format( Double value )
	{
		String result = "null";
		if( value != null )
			result = _formatter.format(value);
		return( result );
	}

	public Double min( Double d1, Double d2 )
	{
		Double result = d1;
		if( ( d1 == null ) || ( d2 != null ) && ( d2 < d1 ) )
			result = d2;

		return( result );
	}

	public double aritmeticMean( double ... values )
	{
		double result = 0d;
		if( ( values != null ) && ( values.length > 0 ) )
		{
			for( double value: values )
				result += value;
			
			result /= values.length;
		}

		return( result );
	}

	public double relativeDifference( double d1, double d2 )
	{
		double greater = d1;
		double smaller = d2;
		if( Math.abs(d2) > Math.abs(d1) )
		{
			greater = d2;
			smaller = d1;
		}
			
		double result = 0;
		if( isInside( d2, -1e-6, 1e-6 ) )
			result = Math.abs(greater - smaller) / greater;

		return( result );
	}

	public Double max( Double d1, Double d2 )
	{
		Double result = d1;
		if( ( d1 == null ) || ( d2 != null ) && ( d2 > d1 ) )
			result = d2;

		return( result );
	}

	public int round( double value )
	{
		return( (int) Math.round(value) );
	}

	public int minInt( double d1, double d2 )
	{
		return( round( min(d1, d2) ) );
	}

	public int maxInt( double d1, double d2 )
	{
		return( round( max(d1, d2) ) );
	}

	public int sgn( double dd )
	{
		return( (dd==0) ? 0 : ( (dd>0) ? 1 : -1 ) );
	}

	public double abs( double dd )
	{
		return( (dd==0) ? 0 : ( (dd>0) ? dd : -dd ) );
	}

	public double limit( double value, double lowerBound, double upperBound )
	{
		return( max( lowerBound, min( upperBound, value ) ) );
	}

	public Double parseDouble( String str )
	{
		Double result = null;
		if( str != null )
		{
			String str2 = StringFunctions.instance().replaceSetOfChars(str, ",", "." );
			result = ExecutionFunctions.instance().safeFunctionExecution( () -> Double.parseDouble( str2 ) );
		}
		return( result );
	}

	protected NumberFormat getNumberFormat( Locale locale )
	{
		NumberFormat result = _formatterMap.get(locale);
		if( result == null )
		{
			result = createNumberFormat(locale);
			_formatterMap.put(locale, result);
		}
		return( result );
	}

	protected NumberFormat createNumberFormat( Locale locale )
	{
		NumberFormat result = NumberFormat.getNumberInstance(locale);
		DecimalFormat df = (DecimalFormat)result;
		df.applyPattern("###,###.00");

		return( result );
	}

	// https://stackoverflow.com/questions/10411414/how-to-format-double-value-for-a-given-locale-and-number-of-decimal-places/10416264
	public String formatDoubleLocale( double value, Locale locale )
	{
		NumberFormat nf = getNumberFormat(locale);
		DecimalFormat df = (DecimalFormat)nf;
		String result = df.format(value);

		return( result );
	}

	public boolean isInside( double value, double lowerBound, double upperBound )
	{
		return( ( value >= lowerBound ) && ( value <= upperBound ) );
	}

	public boolean isStrictlyInside( double value, double lowerBound, double upperBound )
	{
		return( ( value > lowerBound ) && ( value < upperBound ) );
	}

	public boolean isInt( double value )
	{
		return( ( value - Math.round(value) ) == 0 );
	}

	protected double calculateRelativeErrorInternal(double big, double small)
	{
		double result = abs(big - small);
		double absBig = abs(big);
		if( absBig > 1 )
			result = result / absBig;
		
		return( result );
	}

	public double calculateRelativeError(double value1, double value2)
	{
		double result = 0;
		if( abs(value1) > abs(value2) )
			result = calculateRelativeErrorInternal( value1, value2 );
		else
			result = calculateRelativeErrorInternal( value2, value1 );

		return( result );	
	}

	public boolean isClose(double value1, double value2, double tolerance) {
		if(tolerance < 0d) {
			throw new IllegalArgumentException("tolerance must be positive");
		}

		double diff = abs(value1 - value2);
		int sgn = sgn(diff);
		boolean result = (sgn == 0);
		if(!result) {
			if(sgn == 1) {
				result = (diff < tolerance);
			} else if(sgn == -1) {
				result = (-diff < tolerance);
			} else {
				throw new IllegalValueException(String.format("sgn should have been one of the following: (-1, 0, 1) but is %d", sgn));
			}
		}
		return result;
	}

	public double getSquareDistance(double[] arr1, double[] arr2) {
		double result = 0;
		double diff;
		for(int ii=0; ii<arr1.length; ii++) {
			diff = arr1[ii] - arr2[ii];
			result += diff * diff;
		}
		return result;
	}
}
