/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.general.process;

import com.frojasg1.general.ExecutionFunctions;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.concurrent.Semaphore;
import java.util.function.Consumer;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 *
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 */
public class ProcessFunctions
{
	private static final Logger LOGGER = LoggerFactory.getLogger(ProcessFunctions.class);

	protected static ProcessFunctions INSTANCE = new ProcessFunctions();

	public static ProcessFunctions instance()
	{
		return( INSTANCE );
	}

	// https://stackoverflow.com/questions/5711084/java-runtime-getruntime-getting-output-from-executing-a-command-line-program
	public ProcExecResult execute( String ... commandAndParams ) throws IOException
	{
		ProcExecResult result = new ProcExecResult();
		result.setCommand(commandAndParams);

		LOGGER.info( "Executing: {}", result.getCommand() );

		Runtime rt = Runtime.getRuntime();
		Process proc = rt.exec(commandAndParams);
 
		Semaphore sem = new Semaphore(0);
		StreamGobbler errorGobbler = new 
			StreamGobbler(proc.getErrorStream(), result::addStandardErrorLine, sem );

		// any output?
		StreamGobbler outputGobbler = new 
			StreamGobbler(proc.getInputStream(), result::addStandardOutputLine, sem );

		// kick them off
		errorGobbler.start();
		outputGobbler.start();

		ExecutionFunctions.instance().safeMethodExecution( () -> sem.acquire(2) );

		result.setExitValue( waitForExitValue( proc ) );

		return( result );
	}

	protected int waitForExitValue( Process proc )
	{
		Integer result = ExecutionFunctions.instance().safeFunctionExecution( () -> proc.waitFor() );
		if( result == null )
			result = -1;

		return( result );
	}

	public void logProcessResult( ProcExecResult procExecResult )
	{
		if( procExecResult != null )
			LOGGER.info( "Command: {}\\nProcess output:   \\nSuccess: {}\\nOutput: \\n{}\\nError: {}",
				procExecResult.getCommand(),
				procExecResult.wasSuccessful(),
				procExecResult.getAllOutputInAString(),
				procExecResult.getAllErrorInAString() );
		else
			LOGGER.info( "Process output:   null. SomethingFailed" );
	}

	public boolean wasASuccess( ProcExecResult procExecResult )
	{
		boolean result = false;
		if( procExecResult != null )
			result = procExecResult.wasSuccessful();

		return( result );
	}
}

// https://www.infoworld.com/article/2071275/when-runtime-exec---won-t.html?page=2
class StreamGobbler extends Thread
{
    protected InputStream is;
	protected Consumer<String> _lineAdder;
	protected Semaphore _sem;
//    String type;
    
//    StreamGobbler(InputStream is, String type)
    StreamGobbler(InputStream is, Consumer<String> lineAdder, Semaphore sem)
    {
        this.is = is;
		_lineAdder = lineAdder;
		_sem = sem;
//        this.type = type;
    }
    
    public void run()
    {
        try
        {
            InputStreamReader isr = new InputStreamReader(is);
            BufferedReader br = new BufferedReader(isr);
            String line=null;
            while ( (line = br.readLine()) != null)
				_lineAdder.accept(line);
		} catch (IOException ioe)
        {
			ioe.printStackTrace();  
        }
		finally
		{
			ExecutionFunctions.instance().safeMethodExecution( _sem::release );
		}
    }
}
