/*
 * Copyright (c) 2023. Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 3.0 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You may obtain a copy of the License at
 *
 *       http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */

package com.frojasg1.general.selectors;

import com.frojasg1.general.functional.FunctionalInterfaceFunctions;
import com.frojasg1.general.functional.interfaces.TriFunction;

import java.util.function.BiFunction;
import java.util.function.Supplier;
import java.util.function.UnaryOperator;

public class Nearest2DSelector {

    protected static Nearest2DSelector INSTANCE = new Nearest2DSelector();

    public static Nearest2DSelector instance() {
        return INSTANCE;
    }

    protected FunctionalInterfaceFunctions getFunctionalInterfaceFunctions() {
        return FunctionalInterfaceFunctions.instance();
    }

    protected <C> C getFirst(Supplier<C>... suppliers) {
        return getFunctionalInterfaceFunctions().getFirst(suppliers);
    }

    protected <C> C apply(UnaryOperator<C> operator, C value) {
        return getFunctionalInterfaceFunctions().apply(operator, value);
    }

    protected <C> C transformationChain(C value, UnaryOperator<C>... transformationOperators) {
        return getFunctionalInterfaceFunctions().transformationChain(value, transformationOperators);
    }

    public <C> C getNearestSuitable(C elem, UnaryOperator<C> selector, UnaryOperator<C> limiter,
                                    TriFunction<C, Integer, Integer, C> deltaXdeltaYfunction,
                                    int maxDelta) {
        elem = apply(limiter, elem);
        return getNearestSuitableInternal(elem, selector, limiter, deltaXdeltaYfunction, maxDelta);
    }

    protected <C> C getNearestSuitableInternal(C elem, UnaryOperator<C> selector, UnaryOperator<C> limiter,
                                               TriFunction<C, Integer, Integer, C> deltaXdeltaYfunction,
                                               int maxDelta) {
        C result = apply(selector, elem);
        if(result == null) {
            for (int dist = 0; (dist <= maxDelta) && (result == null); dist++) {
                result = getSuitableTotalSlice(elem, selector, limiter, deltaXdeltaYfunction, dist);
            }
        }
        if(result == null) {
            throw new IllegalStateException(String.format("Could find suitable element: %s, with maxDelta: %d", elem, maxDelta));
        }
        return result;
    }

    protected <C> C getSuitableTotalSlice(C elem, UnaryOperator<C> selector, UnaryOperator<C> limiter,
                                          TriFunction<C, Integer, Integer, C> deltaXdeltaYfunction, int sliceDistance) {
        C result = null;
        for(int iterDist = 0; (iterDist <= sliceDistance) && (result == null); iterDist++) {
            result = getSliceIteration(elem, selector, limiter, deltaXdeltaYfunction, sliceDistance, iterDist);
        }
        return result;
    }

    protected <C> C setSliceIterationParticularization(C elem, UnaryOperator<C> selector, UnaryOperator<C> limiter,
                                                       TriFunction<C, Integer, Integer, C> deltaXdeltaYfunction,
                                                       int deltax, int deltay) {
        C result = deltaXdeltaYfunction.apply(elem, deltax, deltay);
        return transformationChain(result, limiter, selector);
    }

    protected <C> C getSliceIteration(C elem, UnaryOperator<C> selector, UnaryOperator<C> limiter,
                                      TriFunction<C, Integer, Integer, C> deltaXdeltaYfunction,
                                      int sliceDistance, int iterDist) {
        C result = null;
        BiFunction<Integer, Integer, C> myFunction =
                (dx, dy) -> setSliceIterationParticularization(elem, selector, limiter, deltaXdeltaYfunction, dx, dy);

        if((iterDist == 0) && (sliceDistance == 0)) {
            result = myFunction.apply(0, 0);
        } else if(iterDist == 0) {
            result = getFirst(
                    () -> myFunction.apply(-sliceDistance, 0),
                    () -> myFunction.apply(sliceDistance, 0),
                    () -> myFunction.apply(0, -sliceDistance),
                    () -> myFunction.apply(0, sliceDistance)
            );
        } else if(iterDist == sliceDistance) {
            result = getFirst(
                    () -> myFunction.apply(-sliceDistance, -sliceDistance),
                    () -> myFunction.apply(-sliceDistance, sliceDistance),
                    () -> myFunction.apply(sliceDistance, -sliceDistance),
                    () -> myFunction.apply(sliceDistance, sliceDistance)
            );
        } else {
            result = getFirst(
                    () -> myFunction.apply(-sliceDistance, -iterDist),
                    () -> myFunction.apply(-sliceDistance, iterDist),
                    () -> myFunction.apply(sliceDistance, -iterDist),
                    () -> myFunction.apply(sliceDistance, iterDist),
                    () -> myFunction.apply(-iterDist, -sliceDistance),
                    () -> myFunction.apply(iterDist, -sliceDistance),
                    () -> myFunction.apply(-iterDist, +sliceDistance),
                    () -> myFunction.apply(iterDist, +sliceDistance)
            );
        }
        return result;
    }
}
