/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.general.threads;

import com.frojasg1.general.NullFunctions;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;
import java.util.function.Function;
import java.util.function.Supplier;

/**
 *
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 */
public class LockAndConditionBase {

	protected ReentrantLock _lock;
	protected Condition[] _conditions;

	protected int _numConditions;

	public LockAndConditionBase( int numConditions )
	{
		this( numConditions, true );
	}

	public LockAndConditionBase( LockAndConditionBase that )
	{
		this( that.getNumConditions() );
	}

	protected LockAndConditionBase( LockAndConditionBase that, boolean init )
	{
		this( that.getNumConditions(), init );
	}

	protected LockAndConditionBase( int numConditions, boolean init )
	{
		_numConditions = numConditions;
		if( init )
			init();
	}

	protected void init()
	{
		_lock = createLock();
		_conditions = createConditions();
	}

	protected ReentrantLock createLock()
	{
		return( new ReentrantLock(true) );
	}

	protected Condition[] createConditions()
	{
		Condition[] result = new Condition[getNumConditions()];

		for( int ii=0; ii<result.length; ii++ )
			result[ii] = createCondition();

		return( result );
	}

	protected Condition createCondition()
	{
		return( getLock().newCondition() );
	}

	public ReentrantLock getLock() {
		return _lock;
	}

	protected boolean conditionHasWaiters( int conditionIndex ) {
		return getLock().hasWaiters( getCondition( conditionIndex ) );
	}

	public Condition[] getConditions() {
		return _conditions;
	}

	public int getNumConditions() {
		return _numConditions;
	}

	public Condition getCondition( int index )
	{
		return( getConditions()[index] );
	}

	public void lockProcedure( Runnable procedure )
	{
		LockFunctions.instance().lockProcedure(getLock(), procedure);
	}

	public <CC> CC lockFunction( Supplier<CC> function )
	{
		return( LockFunctions.instance().lockFunction(getLock(), function) );
	}

	public boolean awaitInterrupted( int conditionIndex, Integer timeoutMs )
	{
		return( LockFunctions.instance().awaitInterrupted(
							getCondition(conditionIndex), timeoutMs) );
	}

	public void ifHasWaitersSignal( int conditionIndex )
	{
		if( hasWaiters( conditionIndex ) )
			signalCondition( conditionIndex );
	}

	public boolean hasWaiters( int conditionIndex )
	{
		return( getLock().hasWaiters( getCondition(conditionIndex) ) );
	}

	public void signalCondition( int conditionIndex )
	{
		getCondition(conditionIndex).signal();
	}

	public void signalAllCondition( int conditionIndex )
	{
		getCondition(conditionIndex).signalAll();
	}

	protected void runAndSignalInternal( Runnable runnable, int conditionIndex )
	{
		if( runnable != null )
			runnable.run();

		ifHasWaitersSignal( conditionIndex );
	}

	public void taskToSignalGen( Runnable runnable,
								int conditionIndex)
	{
		lockProcedure( () -> runAndSignalInternal( runnable, conditionIndex ) );
	}

	protected <CC, RR> RR getIfNotNull( CC obj, Function<CC, RR> getter )
	{
		return( NullFunctions.instance().getIfNotNull(obj, getter) );
	}

	protected void signalAllInternal()
	{
		for( int ii = 0; ii < getNumConditions(); ii++ )
			if( conditionHasWaiters(ii) )
				signalAllCondition(ii);
	}

	public void signalAll()
	{
		lockProcedure( () -> signalAllInternal() );
	}
}
