/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.general.threads.consumer;

import com.frojasg1.general.threads.RunnableBase;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.LinkedList;

/**
 *
 * @author fjavier.rojas
 */
public abstract class ConsumerRunnable<CC> extends RunnableBase
{
	private static final Logger LOGGER = LoggerFactory.getLogger(ConsumerRunnable.class);
	protected static final int NUM_ADDITIONAL_CONDITIONS = 2;
	protected static final int WAITING_FOR_ITEM_TO_BE_CONSUMED_CONDITION_INDEX = 1;
	protected static final int WAITING_FOR_INPUT_LIST_EMPTY_CONDITION_INDEX = 2;

	protected LinkedList<CC> _itemsToBeConsumed;

	protected ConsumerRunnable( int numAdditionalConditions,
								boolean init )
	{
		super(numAdditionalConditions + NUM_ADDITIONAL_CONDITIONS);

		if( init )
			init();
	}

	protected void init()
	{
		_itemsToBeConsumed = createLinkedList();
	}

	protected LinkedList<CC> createLinkedList()
	{
		return( new LinkedList<>() );
	}

	protected LinkedList<CC> copyLinkedList( LinkedList<CC> that )
	{
		return( new LinkedList<>(that) );
	}

	public void addItemToBeConsumed( CC item )
	{
		lockProcedure( () -> addItemToBeConsumedInternal( item ) );
	}

	protected boolean canAddItem()
	{
		return( true );
	}

	protected void addItemToBeConsumedInternal( CC item )
	{
		boolean logged = false;
		while( !isInterrupted() && !canAddItem() ) {
			if(!logged) {
				LOGGER.info("Waiting to add item to be consumed");
				logged = true;
			}
			awaitInterrupted(WAITING_FOR_INPUT_LIST_EMPTY_CONDITION_INDEX, 500);
		}

		_itemsToBeConsumed.add(item);
		ifHasWaitersSignal(WAITING_FOR_ITEM_TO_BE_CONSUMED_CONDITION_INDEX);
	}

	protected LinkedList<CC> getItemsToBeConsumed() {
		return _itemsToBeConsumed;
	}

	@Override
	protected boolean doInit()
	{
		return( true );
	}

	@Override
	protected boolean doRoutine()
	{
		while( !getHasToStop() && !getHasToPause() )
		{
			CC item = getItemToBeConsumed();
			if( item != null )
				consume( item );
		}

		return true;
	}

	protected abstract void consume( CC itemToBeConsumed );

	protected CC getItemToBeConsumed()
	{
		return( lockFunction( this::getItemToBeConsumedInternal ) );
	}

	protected void waitForItems()
	{
		while( !isInterrupted() && getItemsToBeConsumed().isEmpty() )
			awaitInterrupted( WAITING_FOR_ITEM_TO_BE_CONSUMED_CONDITION_INDEX, 500 );
	}

	protected CC getItemToBeConsumedInternal()
	{
		waitForItems();

		CC result = getItemToBeConsumedInternalInternal();

		if( result != null )
			ifHasWaitersSignal(WAITING_FOR_INPUT_LIST_EMPTY_CONDITION_INDEX);

		return( result );
	}

	protected CC getItemToBeConsumedInternalInternal()
	{
		CC result = null;
		if( getHasToStop() && !isInterrupted() && ! getItemsToBeConsumed().isEmpty() )
			result = getItemsToBeConsumed().removeFirst();

		return( result );
	}

	@Override
	protected void doBeforePause() {

	}

	@Override
	protected void doAfterPause() {

	}

	@Override
	public void doEnd() {

	}

	@Override
	protected void hasToStopAssociatedTasks() {

	}
}
