/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.general.view.gencomponentwrapper;

import com.frojasg1.general.ArrayFunctions;
import com.frojasg1.general.CallStackFunctions;
import com.frojasg1.general.ExecutionFunctions;
import com.frojasg1.general.number.IntegerFunctions;
import com.frojasg1.general.update.Updateable;
import com.frojasg1.general.view.ReleaseResourcesable;
import com.frojasg1.general.view.evt.wrapper.ListOfGenComponentWrapperEventContextListenersBase;
import com.frojasg1.general.view.evt.wrapper.impl.GenComponentWrapperEvent;
import com.frojasg1.general.view.evt.wrapper.impl.GenComponentWrapperEventBuilder;
import com.frojasg1.general.view.evt.wrapper.impl.GenComponentWrapperEventType;
import com.frojasg1.general.view.internationalization.window.GenInternationalizedWindow;

import java.util.ArrayList;
import java.util.List;
import java.util.function.Consumer;

/**
 *
 * @param <GC> - GenComponent
 * @param <EV> - Event
 */
public abstract class GenComponentWrapperBase<GC, EV extends GenComponentWrapperEvent>
	implements GenComponentWrapper<GC, EV>//, Blockable
{
	protected GC genComponent;

	protected ListOfGenComponentWrapperEventContextListenersBase<GC, EV> _listenersList;

	protected GenComponentWrapperEventBuilder<EV> _eventBuilder;

	protected GuiSystemDependingFunctions guiSystemDependingFunctions;

	public GenComponentWrapperBase(GC genComponent, GenComponentWrapperEventBuilder<EV> eventBuilder,
								   GuiSystemDependingFunctions guiSystemDependingFunctions)
	{
		_eventBuilder = eventBuilder;
		this.genComponent = genComponent;
		this.guiSystemDependingFunctions = guiSystemDependingFunctions;
	}

	protected void init()
	{
		_listenersList = createListenersList();

		addListeners();
	}

	protected void taskWithClock( Runnable runnable )
	{
		GenInternationalizedWindow window = getMainWindow();
		if(window != null) {
			window.changeToWaitCursor();
		}
		try
		{
			runnable.run();
		}
		finally
		{
			if(window != null) {
				window.revertChangeToWaitCursor();
			}
		}
	}

	protected abstract GenInternationalizedWindow getMainWindow();

	@Override
	public GC getGenComponent() {
		return genComponent;
	}

	@Override
	public void setGenComponent(GC genComponent) {
		this.genComponent = genComponent;
	}

	@Override
	public GenComponentWrapperEventBuilder<EV> getEventBuilder() {
		return _eventBuilder;
	}

	protected ListOfGenComponentWrapperEventContextListenersBase<GC, EV> createListenersList()
	{
		return( new ListOfGenComponentWrapperEventContextListenersBase<>() );
	}

	protected ListOfGenComponentWrapperEventContextListenersBase<GC, EV> getListenersList()
	{
		return( _listenersList );
	}

	@Override
	public void addWrapperListener(Consumer<? super EV> listener, Long... eventTypes)
	{
		getListenersList().addListenerContext(listener, eventTypes);
	}

	@Override
	public void removeWrapperListener(Consumer<? super EV> listener)
	{
		getListenersList().remove(listener);
	}

	@Override
	public void notifyEvt(EV evt)
	{
		getListenersList().notifyEvt(evt);
	}


	@Override
	public void releaseResources()
	{
		getListenersList().clear();

		removeListeners();
	}

	protected abstract void addListeners();
	protected abstract void removeListeners();

	protected void releaseResources( Object obj )
	{
		if( obj instanceof ReleaseResourcesable )
			ExecutionFunctions.instance().safeMethodExecution( () -> ( (ReleaseResourcesable) obj).releaseResources() );
	}


	protected int zoomValue( int value, double factor )
	{
		return( IntegerFunctions.zoomValueFloor( value, factor ) );
	}

	protected int unzoomValue( int value, double factor )
	{
		return( IntegerFunctions.zoomValueCeil( value, 1 / factor ) );
	}

	protected EV createEvt( long eventType )
	{
		return( getEventBuilder().buildEvent(eventType) );
	}

	@Override
	public <EVT extends EV> EVT createEvt(Class<EVT> clazz)
	{
		return( getEventBuilder().buildEvent(clazz) );
	}

	protected void notifyHasBeenModified()
	{
		notifyEvt( createEvt( GenComponentWrapperEventType.HAS_BEEN_MODIFIED ) );
	}


	protected List<Exception> createCallStacksList()
	{
		return( new ArrayList<>() );
	}

	protected CallStackFunctions getCallStackFun()
	{
		return( CallStackFunctions.instance() );
	}

	@Override
	public void addCallStack(String message, List<Exception> callStacks)
	{
		getCallStackFun().addCallStack( message, callStacks );
	}

	@Override
	public void logCallStacks(String message, List<Exception> callStacks)
	{
		getCallStackFun().logCallStacks( message, callStacks );
	}

	protected boolean isEmpty( int[] array )
	{
		return( ArrayFunctions.instance().isEmpty( array ) );
	}

	@Override
	public GuiSystemDependingFunctions getGuiSystemDependingFunctions() {
		return guiSystemDependingFunctions;
	}

	protected void invokeLaterEventDispatchThread(Runnable runnable) {
		getGuiSystemDependingFunctions().invokeLaterEventDispatchThread(runnable);
	}
}
