/*
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.general.view.gencomponentwrapper.jtable;

import com.frojasg1.general.ArrayFunctions;
import com.frojasg1.general.CollectionFunctions;
import com.frojasg1.general.ExecutionFunctions;
import com.frojasg1.general.collection.functions.ListStateContextBase;
import com.frojasg1.general.dialogs.DialogsWrapper;
import com.frojasg1.general.number.IntegerFunctions;
import com.frojasg1.general.update.Updateable;
import com.frojasg1.general.view.evt.wrapper.impl.GenComponentWrapperEventBuilder;
import com.frojasg1.general.view.evt.wrapper.impl.table.GenTableWrapperEventBase;
import com.frojasg1.general.view.evt.wrapper.impl.table.impl.*;
import com.frojasg1.general.view.gencomponentwrapper.GenComponentWrapperBase;
import com.frojasg1.general.view.gencomponentwrapper.GuiSystemDependingFunctions;
import com.frojasg1.generic.GenericFunctions;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Objects;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.function.BiConsumer;
import java.util.function.Function;
import java.util.stream.Collectors;

/**
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 */
public abstract class GenTableWrapper<GT, RR>
        extends GenComponentWrapperBase<GT, GenTableWrapperEventBase> {
    private static final Logger LOGGER = LoggerFactory.getLogger(GenTableWrapper.class);

    protected List<RR> _list;

    protected List<RR> _filteredList;

    protected ViewAndModelSelectionValue _currentRow;
    protected ViewAndModelSelectionValue _rowToHighlight;

    protected ViewAndModelSelectionValue _rightClickRow;
    protected ViewAndModelSelectionValue _rightClickColumn;
    protected ViewAndModelSelectionArray _rowSelectionBeforeClick;

    protected ViewAndModelSelectionArray _previousRowSelection;
    protected boolean _previousWasSelectionAdjusting = false;

    protected List<Updateable> _viewModelUpdateableList = new ArrayList<>();

    protected int y_coordinateOfInsertionPosition = -1;

//	protected int[] _previousViewSelection;

    protected boolean _isFiltered = false;

    protected Object[] _previousColumnIds;

    protected AtomicInteger _changedByProgram = new AtomicInteger(0);

    protected String _lastFilter;

    public GenTableWrapper(GT table, Class<RR> recordClass,
                           GuiSystemDependingFunctions guiSystemDependingFunctions) {
        this(table, new GenTableWrapperEventBuilderBase<>(table, recordClass, true), guiSystemDependingFunctions);
    }

    public GenTableWrapper(GT table, GenComponentWrapperEventBuilder<GenTableWrapperEventBase> eventBuilder,
                           GuiSystemDependingFunctions guiSystemDependingFunctions) {
        super(table, eventBuilder, guiSystemDependingFunctions);
    }

    @Override
    protected void init() {
        super.init();
    }

    public String getLastFilter() {
        return _lastFilter;
    }

    public List<RR> getFilteredList() {
        return _filteredList;
    }

    protected boolean isChangedByProgram() {
        return (_changedByProgram.get() > 0);
    }

    protected abstract boolean tableSelectionModelValueIsAdjusting();

    public void updatePreviousRowSelection() {
        setPreviousRowSelectionViewArray(getSelectedRows());
    }

    public void setPreviousRowSelectionViewArray(int[] viewArray) {
        setViewArray(_previousRowSelection, viewArray);
        notifyEvt(createSelectionChangedEvt());
    }

    public boolean isPreviousWasSelectionAdjusting() {
        return _previousWasSelectionAdjusting;
    }

    public int[] getPreviousViewRowSelection() {
        return getViewArray(_previousRowSelection);
    }

    protected void initViewAndModelAttribs() {
        _currentRow = createViewAndModelRowSelectionValue();
        _rowToHighlight = createViewAndModelRowSelectionValue();

        _rightClickRow = createViewAndModelRowSelectionValue();
        _rightClickColumn = createViewAndModelColumnSelectionValue();
        _rowSelectionBeforeClick = createViewAndModelRowSelectionArray(0);
    }

    @Override
    public GenTableWrapperEventBuilderBase<RR, GT> getEventBuilder() {
        return (GenTableWrapperEventBuilderBase<RR, GT>) super.getEventBuilder();
    }

    public ViewAndModelSelectionValue createViewAndModelRowSelectionValue() {
        ViewAndModelSelectionValue result = new ViewAndModelSelectionValue(
                this::convertViewRowIndexToModel,
                this::convertModelRowIndexToView);
        _viewModelUpdateableList.add(result);
        return (result);
    }

    public ViewAndModelSelectionValue createViewAndModelColumnSelectionValue() {
        ViewAndModelSelectionValue result = new ViewAndModelSelectionValue(
                this::convertViewColumnIndexToModel,
                this::convertModelColumnIndexToView);
        _viewModelUpdateableList.add(result);
        return (result);
    }

    public ViewAndModelSelectionArray createViewAndModelRowSelectionArray(int validityOfExpected) {
        ViewAndModelSelectionArray result = new ViewAndModelSelectionArray(
                this::convertViewRowIndexToModel,
                this::convertModelRowIndexToView,
                validityOfExpected);
        _viewModelUpdateableList.add(result);
        return (result);
    }

    public ViewAndModelSelectionArray createViewAndModelColumnSelectionArray(int validityOfExpected) {
        ViewAndModelSelectionArray result = new ViewAndModelSelectionArray(
                this::convertViewColumnIndexToModel,
                this::convertModelColumnIndexToView,
                validityOfExpected);
        _viewModelUpdateableList.add(result);
        return (result);
    }

    public boolean isEmpty() {
        List<RR> list = getList();

        return ((list == null) || list.isEmpty());
    }

    public List<RR> getList() {
        return _list;
    }

    protected abstract int getSelectedRow();

    protected abstract void setRowSelectionInterval(int start, int end);

    public Integer getCurrentViewRowIndex() {
        return (getViewIndex(_currentRow));
    }

    public Integer getCurrentModelRowIndex() {
        return (getModelIndex(_currentRow));
    }

    public Integer getViewIndex(ViewAndModelSelectionValue valueObj) {
        return ((valueObj == null) ? null : valueObj.getViewSelectedIndex());
    }

    public Integer getModelIndex(ViewAndModelSelectionValue valueObj) {
        return ((valueObj == null) ? null : valueObj.getModelSelectedIndex());
    }

    public void setViewIndex(ViewAndModelSelectionValue valueObj, Integer value) {
        if (valueObj != null) {
            valueObj.setViewSelectedIndex(value);
        }
    }

    public void setModelIndex(ViewAndModelSelectionValue valueObj, Integer value) {
        if (valueObj != null) {
            valueObj.setModelSelectedIndex(value);
        }
    }

    public int[] getViewArray(ViewAndModelSelectionArray arrayObj) {
        return ((arrayObj == null) ? null : arrayObj.getViewSelection());
    }

    public int[] getModelArray(ViewAndModelSelectionArray arrayObj) {
        return ((arrayObj == null) ? null : arrayObj.getModelSelection());
    }

    public void setViewArray(ViewAndModelSelectionArray arrayObj, int[] array) {
        if (arrayObj != null)
            arrayObj.setViewSelection(array);
    }

    public void setModelArray(ViewAndModelSelectionArray arrayObj, int[] array) {
        if (arrayObj != null)
            arrayObj.setModelSelection(array);
    }

    public boolean isEmpty(ViewAndModelSelectionArray arrayObj) {
        return (arrayObj.isEmpty());
    }

    protected RR getRecord(Integer index) {
        RR result = null;
        if((_list != null) && (index != null)) {
            result = _list.get(index);
        }
        return result;
    }

    public RR getCurrentRecord() {
        return (ExecutionFunctions.instance().safeFunctionExecution(() -> getRecord(getCurrentModelRowIndex())));
    }

    public void updateData() {
        int selectedRow = getSelectedRow();

        showData();

        if ((selectedRow >= 0) && (_list.size() > selectedRow))
            setRowSelectionInterval(selectedRow, selectedRow);
    }

    protected abstract Object[] getColumnIds();


    protected abstract int getColumnCount();


    public boolean isFiltered() {
        return (_isFiltered);
    }

    /*
        protected void storeAndResetSelection()
        {
            _previousViewSelection = this.getSelectedRows();
            resetSelection();
            createProxyForListSelectionModelAndSetItToTable( _previousViewSelection );
        }
    */
    protected void filterData(String filter) {
/*
		boolean isEmptyFilter = StringFunctions.instance().isEmpty(filter);
		if( !isEmptyFilter && !isFiltered() )
			storeAndResetSelection();
*/
        onlyFilterData(filter);
/*
		if( isEmptyFilter && isFiltered() )
			restoreSelection();
*/
    }

    protected void onlyFilterData(String filter) {
        showDataInternal(this::autoResizeAllColumns, filter);
    }

    /*
        protected void restoreSelection()
        {
            List<Exception> callStacks = this.createCallStacksList();
            addCallStack( "restoreSelection", callStacks );
            restablishListSelectionModel();

            SwingUtilities.invokeLater( () -> {
                setRowSelection( _previousViewSelection, callStacks );
                _previousViewSelection = null;
            });
        }
    */
    protected void showData() {
        showDataInternal(this::autoResizeAllColumns, getLastFilter());
    }

    protected void showDataInternal(Runnable adjustColumnsFunction, String filter) {
        this.taskWithClock(() -> showDataInternalFunction(adjustColumnsFunction, filter));
    }

    protected <CC> boolean equals(CC[] array1, CC[] array2) {
        return (ArrayFunctions.instance().equals(array1, array2));
    }

    protected abstract void increaseChangedByProgram();

    protected abstract void decreaseChangedByProgramInternal();

    protected void decreaseChangedByProgram() {
        invokeLaterEventDispatchThread(this::decreaseChangedByProgramInternal);
    }

    protected abstract void showDataInternalFunction(Runnable adjustColumnsFunction, String filter);

    protected void updateViewModel() {
        for (Updateable updateable : _viewModelUpdateableList)
            updateable.update();

        continueSelection();
    }

    protected void continueSelection() {
//		CallStackFunctions.instance().dumpCallStack( "setModelRowSelection" );
        List<Exception> callStacks = this.createCallStacksList();
        addCallStack("continueSelection", callStacks);
        int[] previousModelSelection = getModelArray(_previousRowSelection);
        int[] copy = ArrayFunctions.instance().copy(previousModelSelection);
        invokeLaterEventDispatchThread(() -> {
            setModelRowSelection(copy, callStacks);
//				getTable().repaint();
        });
    }


    protected abstract boolean matches(RR record, String filter);


    protected abstract void setColumnHeaders(Object[] columnIds);

    protected void addModelListeners() {

    }

    protected void removeModelListeners() {

    }

    protected abstract void tableFireTableDataChanged();

    protected void afterResizingColumns() {
//		unblock();
        continueSelection();

        increaseChangedByProgram();
        tableFireTableDataChanged();
//		SwingUtilities.invokeLater( () ->
//			SwingUtilities.invokeLater(
//				() -> _changedByProgram.decrementAndGet()
//			)
//		);
        decreaseChangedByProgram();

        updateViewport();
//		getTable().repaint();
    }

    protected abstract void updateViewport();

    protected abstract void autoResizeNoLimit(int columnIndex);

    protected abstract void autoResizeAllColumns();

    protected void showViewRowIndex(Integer viewRowIndex) {
        if ((viewRowIndex != null) && (viewRowIndex != -1)) {
            if (!isRowVisible(viewRowIndex))
                setRowVisible(viewRowIndex - 5);

            if (!isRowVisible(viewRowIndex))
                setRowVisible(viewRowIndex + 5);
        }
    }

    protected abstract boolean isRowVisible(int viewRowIndex);

    protected boolean isValidModelIndex(Integer modelIndex) {
        return ((modelIndex != null) && (modelIndex > -1) && (modelIndex < _list.size()));
    }

    protected Integer getFirstIndex() {
        return (((_list == null) || _list.isEmpty()) ? null : 0);
    }

    protected abstract void tableRepaint();

    protected void setCurrentIndexLight(Integer modelIndex) {
        if ((modelIndex != null) && !isValidModelIndex(modelIndex))
            modelIndex = getFirstIndex();

        if (modelIndex == null)
            setModelIndex(_currentRow, modelIndex);
        else {
            setModelIndex(_currentRow, modelIndex);
            //			setRowSelectionInterval( index, index );
            showCurrentRow();
        }

        tableRepaint();
    }

    protected void setCurrentIndex(Integer modelIndex) {
        if ((modelIndex != null) && !isValidModelIndex(modelIndex))
            modelIndex = getFirstIndex();

        if (modelIndex == null)
            setNewCurrentModelRow(modelIndex);
//            setModelIndex(_currentRow, modelIndex);
        else {
            setNewCurrentModelRow(modelIndex);
//            setModelIndex(_currentRow, modelIndex);
            //			setRowSelectionInterval( index, index );
            showCurrentRow();
        }

        tableRepaint();
    }

    protected void showCurrentRow() {
        showViewRowIndex(getCurrentRowViewIndex());
    }

    protected Integer getCurrentRowViewIndex() {
        return (getViewIndex(_currentRow));
    }

    protected abstract void makeRowBeVisible(int viewRowIndex);

    protected void setRowVisible(int index) {
        int size = getList().size();
        if (size > 0) {
            int index2 = IntegerFunctions.limit(index, 0, size - 1);
            makeRowBeVisible(index2);
        }
    }

    public void setRecord(RR record) {
        invokeLaterEventDispatchThread(() -> setCurrentIndex(_list.indexOf(record)));
    }

    public void setRecordLight(RR record) {
        invokeLaterEventDispatchThread(() -> setCurrentIndexLight(_list.indexOf(record)));
    }

    protected abstract void setRowSelection(int[] viewRowSelection, List<Exception> callStacks);

    public void setModelRowSelection(int[] modelRowSelection, List<Exception> callStacks) {
//		CallStackFunctions.instance().dumpCallStack( "setModelRowSelection" );
        addCallStack("setModelRowSelection", callStacks);
        invokeLaterEventDispatchThread(
                () -> setRowSelection(modelRowArrayToView(modelRowSelection), callStacks)
        );
    }

    protected int[] modelRowArrayToView(int[] modelRowSelection) {
        return (translateArray(modelRowSelection, this::convertModelRowIndexToView));
    }

    protected int[] translateArray(int[] viewIndexArray,
                                   Function<Integer, Integer> elementTranslatorFunction) {
        return (ArrayFunctions.instance().translateArray(viewIndexArray, elementTranslatorFunction));
    }

    public void updateCurrentList(ListStateContextBase<RR> modifiedListState) {
        setRecordList(modifiedListState.getList());
        setRecord(modifiedListState.getSelectedItem());
        ExecutionFunctions.instance().nestedInvokeLater(4, () ->
                setModelRowSelection(modifiedListState.getSelection(), null));
    }

    public void setRecordList(List<? extends RR> list) {
//		if( _list != list )
        {
//			System.out.println( "setChessGameList, _list != list ..." );

            List<RR> newList = list.stream().collect(Collectors.toList());

            if (!newList.equals(_list)) {
                _list = newList;
                invokeLaterEventDispatchThread(() -> {
                    updateData();

//                    if (_list.size() > 0) {
//                        setCurrentIndex(0);
//                        setRowSelectionInterval(0, 0);
//                    } else
                        setCurrentIndex(null);
                });
            }
        }
    }

    protected int[] calculateExpectedViewSelection() {
        return (_previousRowSelection.calculateExpectedViewSelection());
    }

    public List<RR> calculateExpectedViewSelectionRecords() {
        List<RR> result = null;
        if (getList() != null)
            result = CollectionFunctions.instance().translateArrayIntoList(
                    calculateExpectedModelSelection(), getList()::get);

        return (result);
    }

    protected int[] calculateExpectedModelSelection() {
        return (_previousRowSelection.viewArrayToModel(calculateExpectedViewSelection()));
    }

    protected GenTableWrapperSelectionHasChangedEvent createSelectionChangedEvt() {
        GenTableWrapperSelectionHasChangedEvent result = createEvt(GenTableWrapperSelectionHasChangedEvent.class);

        result.setNewWholeModelSelection(getModelArray(_previousRowSelection));
        result.setNewVisibleModelSelection(calculateExpectedViewSelection());

        return (result);
    }

    protected GenTableWrapperFilterHasChangedEvent<RR, GT> createFilterChangedEvt() {
        return (getEventBuilder()
                .createFilterChangedEvt(
                        CollectionFunctions.instance().copyList(_filteredList)
                )
        );
    }

    protected GenTableWrapperNewCurrentEvent createNewCurrentEvt(int newCurrentRow) {
        return (this.getEventBuilder().createNewCurrentEvt(newCurrentRow));
    }

    protected boolean setNewCurrentModelRow(Integer index) {
        return (setNewCurrentRow(index, this::setModelIndex));
    }

    protected boolean setNewCurrentViewRow(Integer index) {
        return (setNewCurrentRow(index, this::setViewIndex));
    }

    protected abstract void invokeLater(Runnable runnable);

    protected boolean setNewCurrentRow(Integer index,
                                       BiConsumer<ViewAndModelSelectionValue, Integer> indexSetter) {
        boolean success = false;

        if ((_list != null) && (index != null) &&
                (index < _list.size()) && (index >= 0)) {
//            Integer oldModelIndex = getModelIndex(_currentRow);
            indexSetter.accept(_currentRow, index);
//			_controller.newChessGameChosen( _list.get( index ), false );
            Integer newModelIndex = getModelIndex(_currentRow);
//            if(!Objects.equals(oldModelIndex, newModelIndex)) {
                notifyEvt(createNewCurrentEvt(newModelIndex));
//            }
            invokeLater(this::tableRepaint);
            success = true;
        }

        return (success);
    }

    public void insertRecordAfter(RR recordToBeInserted, RR afterThis) {
        int index = insertRecordAfterInternal(recordToBeInserted, afterThis);

        showData();
        tableChangeSelection(index, 0, false, false);
        setNewCurrentModelRow(index);

        notifyHasBeenModified();
    }

    public int insertRecordAfterInternal(RR recordToBeInserted, RR afterThis) {
        int index = IntegerFunctions.max(0, _list.indexOf(afterThis));
        _list.add(index, recordToBeInserted);

        return (index);
    }

    public GT getTable() {
        return (getGenComponent());
    }


    protected boolean isCurrentRow(int row, int column) {
        return (Objects.equals(row, getViewIndex(_currentRow)));
    }

    protected abstract Object[] getRow(int index, RR record);

    public int[] getRowModelSelectionBeforeRightClick() {
        return getModelArray(_rowSelectionBeforeClick);
    }

    public int[] getAppropriateRowModelSelection() {
        int[] result = getRowModelSelectionBeforeRightClick();
        if (isEmpty(result))
            result = getModelSelectionRows();

        return result;
    }

    protected List<RR> getListOfSelectedRecordsRemovingThem() {
        CollectionFunctions.instance();
        List<RR> result = new ArrayList<RR>();

        if (!isEmpty(_rowSelectionBeforeClick)) {
            int[] rowModelSelectionBeforeClick = getModelArray(_rowSelectionBeforeClick);

            for (int ii = rowModelSelectionBeforeClick.length - 1; ii >= 0; ii--)
                result.add(_list.remove(rowModelSelectionBeforeClick[ii]));
        }

        return (result);
    }

    protected int calculatePositionToInsertRecords(int destinationPositionPreviousToTheExtractionOfRecords) {
        int result = destinationPositionPreviousToTheExtractionOfRecords;

        if (!isEmpty(_rowSelectionBeforeClick)) {
            int[] rowModelSelectionBeforeClick = getModelArray(_rowSelectionBeforeClick);
            if (result >= rowModelSelectionBeforeClick[rowModelSelectionBeforeClick.length - 1]) {
                result = result - rowModelSelectionBeforeClick.length + 1;        // +1 because we want to insert it after the current row.
            }
        }

        return (result);
    }

    protected boolean isSelectionEmpty() {
        return (ArrayFunctions.instance().isEmpty(getSelectedRows()));
    }

    /*
        protected void eraseSelectedRecords()
        {
            int index = getModelIndex( _rightClickRow );
            getListOfSelectedRecordsRemovingThem();
            updateData();

            if( _list.size() == 0 )
                createRecord( 0 );
            else
            {
                if( index < 0 )
                    index = 0;
                else if( index >= _list.size() )
                    index = _list.size()-1;

                getTable().changeSelection(index, 0, false, false);
                setNewCurrentModelRow( index );

                notifyHasBeenModified();
            }
        }
    */
    public abstract void tableChangeSelection(int rowIndex, int columnIndex, boolean toggle, boolean extend);

    public void createRecord(int unitsToAddToIndex) {
        int index = getModelIndex(_rightClickRow) + unitsToAddToIndex;
        index = IntegerFunctions.max(0, IntegerFunctions.min(index, _list.size()));

        createRecordInternal(index);
        showData();
        tableChangeSelection(index, 0, false, false);
        setNewCurrentModelRow(index);

        notifyHasBeenModified();
    }

    protected abstract RR createEmptyRecord();

    protected RR createRecordInternal(int index) {
        RR record = null;

        try {
            record = createEmptyRecord();
        } catch (Throwable th) {
            th.printStackTrace();
        }

        if (record != null) {
            _list.add(index, record);
        }

        return (record);
    }

    protected RR getRecordAtRightClick() {
        return (getRecordAtModelRowIndex(getModelIndex(_rightClickRow)));
    }

    protected RR getFirstSelectedRecordAtRightClick() {
        RR result = null;

        if (!isEmpty(_rowSelectionBeforeClick)) {
            int modelRowIndex = getModelArray(_rowSelectionBeforeClick)[0];
            result = getRecordAtModelRowIndex(modelRowIndex);
        }
        if (result == null)
            result = getRecordAtRightClick();

        return (result);
    }

    protected RR getRecordAtModelRowIndex(Integer modelRowIndex) {
        RR result = null;
        if ((modelRowIndex != null) && (_list != null)) {
            int index = modelRowIndex;
            if ((index >= 0) && (index < _list.size()))
                result = _list.get(index);
        }

        return (result);
    }

    @Override
    protected abstract void addListeners();

    @Override
    protected abstract void removeListeners();


    public void filter(String filter) {
        filterData(filter);
    }

    public void updateInternal() {
        showData();
    }

    protected boolean listHasElements() {
        List<RR> list = getList();
        return ((list != null) && !list.isEmpty());
    }

    protected RR get(List<RR> list, int index) {
        return (CollectionFunctions.instance().get(list, index));
    }

    protected int translateIndex(List<RR> fromList, List<RR> toList, int fromIndex) {
        return (CollectionFunctions.instance().translateIndex(fromList, toList, fromIndex));
    }

    protected int getModelRowIndex(int filteredModelRowNum) {
        return (translateIndex(_filteredList, getList(), filteredModelRowNum));
    }

    protected int getFilteredModelRowIndex(int modelRowNum) {
        return (translateIndex(getList(), _filteredList, modelRowNum));
    }

    protected abstract int tableConvertRowIndexToModel(int viewRowIndex);

    protected int convertViewRowIndexToModel(int viewRownum) {
        int filteredModelRowNum = tableConvertRowIndexToModel(viewRownum);

        return (getModelRowIndex(filteredModelRowNum));
    }

    protected abstract int convertViewColumnIndexToModel(int viewColumnNum);

    protected abstract int tableConvertRowIndexToView(int modelRowIndex);

    protected int convertModelRowIndexToView(int modelRownum) {
        int filteredModelRownum = getFilteredModelRowIndex(modelRownum);

        return (tableConvertRowIndexToView(filteredModelRownum));
    }

    protected abstract int convertModelColumnIndexToView(int modelColumnNum);

    public boolean isRowSelected(int[] rowSelection, int viewRowNum) {
        boolean result = false;
        if (rowSelection != null)
            result = Arrays.stream(rowSelection).anyMatch(row -> row == viewRowNum);
        return (result);
    }

    protected abstract int[] getSelectedRows();

    protected int[] getModelSelectionRows() {
        return (ArrayFunctions.instance().translateArray(getSelectedRows(),
                this::convertViewRowIndexToModel));
    }

    @Override
    protected List<Exception> createCallStacksList() {
        return (null);
    }

    public void resetSelection() {
        List<Exception> callStacks = this.createCallStacksList();
        addCallStack("resetSelection", callStacks);
        setModelRowSelection(new int[0], callStacks);
    }

    protected abstract void notifyRecordsDeleted();

    public boolean confirmErasureOfSelectedRecords(GenTableWrapperRecordsModifiedEventBase<RR, ?, ? extends ListStateContextBase<RR>> evt) {
        boolean result = false;
        int[] selectedRows = evt.getEventData().getPrevious().getSelection();

        int numRowsToErase = getLength(selectedRows);
        if (numRowsToErase > 0)
            result = askToProceed(getTextForConfirmingErasureOfSelectedRows(numRowsToErase));

        return (result);
    }

    protected String getTextForConfirmingErasureOfSelectedRows(int numRowsToErase) {
        return (String.format("Are you sure you want to erase the (%d) selected rows ?", numRowsToErase));
    }

    protected int getLength(int[] array) {
        return (ArrayFunctions.instance().getLength(array));
    }

    protected boolean askToProceed(String text) {
        return (getDialogsWrapper().showYesNoDialog(getMainWindow(), text, null, 1) == 0);
    }

    public DialogsWrapper getDialogsWrapper() {
        return (GenericFunctions.instance().getDialogsWrapper());
    }
}
