/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.general.view.geometry;

import java.time.ZoneId;
import java.util.function.Function;


/**
 *
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 * @param <PT>	- Point class
 * @param <RE>	- Rectangle class
 */
public class PointDiscarderGen<PT, RE> {

	protected Function<PT, Double> _ptXgetter;
	protected Function<PT, Double> _ptYgetter;

	protected Function<RE, Double> _rectXgetter;
	protected Function<RE, Double> _rectYgetter;
	protected Function<RE, Double> _rectWidthgetter;
	protected Function<RE, Double> _rectHeightgetter;

	public PointDiscarderGen( Function<PT, Double> ptXgetter, Function<PT, Double> ptYgetter,
							Function<RE, Double> rectXgetter, Function<RE, Double> rectYgetter,
							Function<RE, Double> rectWidthgetter, Function<RE, Double> rectHeightgetter )
	{
		_ptXgetter = ptXgetter;
		_ptYgetter = ptYgetter;
		_rectXgetter = rectXgetter;
		_rectYgetter = rectYgetter;
		_rectWidthgetter = rectWidthgetter;
		_rectHeightgetter = rectHeightgetter;
	}

	protected double getPtX( PT pt )
	{
		return( _ptXgetter.apply(pt) );
	}

	protected double getPtY( PT pt )
	{
		return( _ptYgetter.apply(pt) );
	}

	protected double getRectX( RE re )
	{
		return( _rectXgetter.apply(re) );
	}

	protected double getRectY( RE re )
	{
		return( _rectYgetter.apply(re) );
	}

	protected double getRectWidth( RE re )
	{
		return( _rectWidthgetter.apply(re) );
	}

	protected double getRectHeight( RE re )
	{
		return( _rectHeightgetter.apply(re) );
	}

	// must be fast, so make it be fast
	protected ZoneId2d getZoneId( PT pt, RE bounds )
	{
		ZoneId2d result = null;
		if( ( pt != null ) && ( bounds != null ) )
		{
			if( getPtX(pt) < getRectX(bounds) )
			{
				if( getPtY(pt) < getRectY(bounds) )
					result = ZoneId2d.NORTH_WEST;
				else if( getPtY(pt) > (getRectY(bounds) + getRectHeight(bounds)) )
					result = ZoneId2d.NORTH_EAST;
				else
					result = ZoneId2d.NORTH;
			}
			else if( getPtX(pt) > getRectX(bounds) + getRectWidth(bounds) )
			{
				if( getPtY(pt) < getRectY(bounds) )
					result = ZoneId2d.SOUTH_WEST;
				else if( getPtY(pt) > (getRectY(bounds) + getRectHeight(bounds)) )
					result = ZoneId2d.SOUTH_EAST;
				else
					result = ZoneId2d.SOUTH;
			}
			else
			{
				if( getPtY(pt) < getRectY(bounds) )
					result = ZoneId2d.WEST;
				else if( getPtY(pt) > (getRectY(bounds) + getRectHeight(bounds)) )
					result = ZoneId2d.EAST;
				else
					result = ZoneId2d.INSIDE;
			}
		}

		return( result );
	}

	protected boolean zoneMatchesOutside( ZoneId zid, ZoneId zid2 )
	{
		return( ( zid == null ) || ( zid2 == null ) ||
				( zid == zid2 ) && !zid.isIsInside() );
	}

	protected boolean zoneMatchesOutside( ZoneId2d zi2d, ZoneId horizZi, ZoneId vertZi )
	{
		boolean result = true;
		if( zi2d != null )
			result = PointDiscarderGen.this.zoneMatchesOutside( zi2d.getHorizontalZoneId(), horizZi ) ||
				PointDiscarderGen.this.zoneMatchesOutside( zi2d.getVerticalZoneId(), vertZi );

		return( result );
	}

	protected ZoneId updateZoneId( ZoneId zid, ZoneId zid2 )
	{
		ZoneId result = null;
		if( zid == null )
			result = zid2;
		else if( zid2 == null )
			result = zid;
		else if( ( zid == zid2 ) && !zid.isIsInside() )
			result = zid;
		else
			result = ZoneId.INSIDE;

		return( result );
	}

	public boolean canDiscard( RE bounds, PT ... pts)
	{
		boolean result = false;

		if(bounds != null) {
			ZoneId horizZi = null;
			ZoneId vertZi = null;
			ZoneId2d zi2d;
			for (PT pt : pts) {
				zi2d = getZoneId(pt, bounds);
				if (zoneMatchesOutside(zi2d, horizZi, vertZi)) {
					if (zi2d != null) {
						result = true;
						horizZi = updateZoneId(horizZi, zi2d.getHorizontalZoneId());
						vertZi = updateZoneId(vertZi, zi2d.getVerticalZoneId());
					}

					if ((horizZi == vertZi) && (horizZi != null) && horizZi.isIsInside()) {
						result = false;
						break;
					}
				} else {
					result = false;
					break;
				}
			}
		}

		return( result );
	}

	protected static enum ZoneId2d
	{
		NORTH_WEST(ZoneId.LOWER, ZoneId.LOWER),
		NORTH(ZoneId.INSIDE, ZoneId.LOWER),
		NORTH_EAST(ZoneId.LOWER, ZoneId.LOWER),
		WEST(ZoneId.LOWER, ZoneId.INSIDE),
		INSIDE(ZoneId.INSIDE, ZoneId.INSIDE),
		EAST(ZoneId.HIGHER, ZoneId.INSIDE),
		SOUTH_WEST(ZoneId.LOWER, ZoneId.HIGHER),
		SOUTH(ZoneId.INSIDE, ZoneId.HIGHER),
		SOUTH_EAST(ZoneId.HIGHER, ZoneId.HIGHER);

		protected ZoneId _horizontalZoneId;
		protected ZoneId _verticalZoneId;

		private ZoneId2d( ZoneId horizontalZoneId, ZoneId verticalZoneId )
		{
			_horizontalZoneId = horizontalZoneId;
			_verticalZoneId = verticalZoneId;
		}

		public ZoneId getHorizontalZoneId() {
			return _horizontalZoneId;
		}

		public ZoneId getVerticalZoneId() {
			return _verticalZoneId;
		}
	}

	protected static enum ZoneId
	{
		LOWER(false),
		INSIDE(true),
		HIGHER(false);

		protected boolean _isInside;

		private ZoneId( boolean isInside )
		{
			_isInside = isInside;
		}

		public boolean isIsInside() {
			return _isInside;
		}
	}
}
