/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.general.xml.persistency;

import com.frojasg1.applications.common.configuration.InternationalizedStringConf;
import com.frojasg1.general.FileFunctions;
import com.frojasg1.general.exceptions.ConfigurationException;
import com.frojasg1.general.xml.XmlElement;
import com.frojasg1.general.xml.language.LanguageXmlConfiguration;
import com.frojasg1.general.xml.model.NameKeyWithFileNameModel;
import static com.frojasg1.general.xml.persistency.XmlFileListPersistency.CONF_ERROR_WHEN_SAVING_XML_CONFIGURATION_FILE;
import com.frojasg1.general.xml.persistency.loader.ModelToXml;
import com.frojasg1.general.xml.persistency.loader.XmlToModel;
import java.io.IOException;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.TransformerException;
import org.xml.sax.SAXException;

/**
 *
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 */
public abstract class XmlSimpleFilePersistency<MM extends NameKeyWithFileNameModel> extends LanguageXmlConfiguration
									implements //PersistentConfiguration,
									InternationalizedStringConf
{

	@Override
	public void init( String singleXmlFileName,
						String languageFolderResourceName,
						String languageFolderDiskFileName )
	{
		throw( new RuntimeException( "Do not init with this function. try calling: " +
									"init( BaseApplicationConfigurationInterface appliConf )" ) );
	}

	protected MM createFileModel( String item, XmlElement elem )
	{
		MM result = createModelObject( elem );

		return( result );
	}

	public MM loadModelFromXml( String fileName ) throws IOException, ParserConfigurationException, SAXException
	{
		return( loadAndAddDocumentCustomFileName(fileName) );
	}

	public void save(MM model) throws ConfigurationException, IOException
	{
		try {
			saveXmlConfigurationFile( model );
		} catch (TransformerException ex) {
			ex.printStackTrace();
			throw( new RuntimeException( this.createCustomInternationalString( CONF_ERROR_WHEN_SAVING_XML_CONFIGURATION_FILE, ex.getMessage(), ex ) ) );
		}
	}

	protected abstract ModelToXml<MM> createModelToXmlTranslator();

	public void saveXmlConfigurationFile( MM model ) throws IOException, TransformerException
	{
		saveXmlConfigurationFile( model, model.getFileName() );
	}

	public void saveXmlConfigurationFile( MM model, String fileName ) throws IOException, TransformerException
	{
		model.setFileName(fileName);
		ModelToXml<MM> builderToXml = createModelToXmlTranslator();
		XmlElement xmlElement = builderToXml.build( model );

		saveDocumentCustomFileName( xmlElement, fileName );
	}

	protected abstract XmlToModel<MM> createXmlToModelTranslator();

	protected MM createModelObject( XmlElement xmlElement )
	{
		XmlToModel<MM> builder = createXmlToModelTranslator();
		MM model = builder.build( xmlElement );

		return( model );
	}

	public void exportXmlFile( MM model, String xmlFileName ) throws IOException, TransformerException
	{
		ModelToXml<MM> builderToXml = createModelToXmlTranslator();
		XmlElement xmlElement = builderToXml.build( model );
		model.setFileName(xmlFileName);

		saveDocumentCustomFileName( xmlElement, xmlFileName );
	}

	protected MM loadAndAddDocumentCustomFileName( String longFileName ) throws IOException, ParserConfigurationException, SAXException
	{
		String singleFileName = FileFunctions.instance().getBaseName(longFileName);
		XmlElement elem = loadDocumentCustomFileName( longFileName );

		MM result = createFileModel( singleFileName, elem );
		result.setFileName( longFileName );

		postProcess( result );

		return( result );
	}

	protected void postProcess( MM result )
	{
		// intentionally left blank
	}
/*
	public MM importOriginalXmlFile( String xmlFileName ) throws IOException, TransformerException, ParserConfigurationException, SAXException
	{
		MM result = loadAndAddDocument( xmlFileName, xmlFileName );

//		saveXmlConfigurationFile( result );

		return( result );
	}
*/
	public MM importXmlFile( String xmlFileName ) throws IOException, TransformerException, ParserConfigurationException, SAXException
	{
		MM result = loadAndAddDocumentCustomFileName( xmlFileName );

//		saveXmlConfigurationFile( result );

		return( result );
	}
/*
	protected void eraseNonExisting( )
	{
		if( getPreviousModelContainer() != null )
		{
			getPreviousModelContainer().getComboBoxContent().getListOfItems().stream()
				.map( item -> getPreviousModelContainer().getRelativeFileNameFromItemList(item) )
				.filter( (fn) -> !getModelContainer().getComboBoxContent().contains( fn ) )
				.forEach( (fn) -> deleteFile( fn ) );
		}
	}
*/
	/*
	public RegexWholeFileModel getRegexWholeContainer()
	{
		return( _model );
	}

	public void setRegexWholeContainer( RegexWholeFileModel rewc )
	{
		_model = rewc;
	}


	protected void registerInternationalizedStrings()
	{
//		this.registerInternationalString(CONF_ERROR_WHEN_SAVING_XML_CONFIGURATION_FILE, "Error when saving xml configuration file name $1" );
	}

	@Override
	public void registerInternationalString(String label, String value)
	{
		_internationalizedStringConf.registerInternationalString(label, value);
	}

	@Override
	public String getInternationalString(String label)
	{
		return( _internationalizedStringConf.getInternationalString(label ) );
	}

	@Override
	public String createCustomInternationalString(String label, Object... args) {
		return( _internationalizedStringConf.createCustomInternationalString( label, args ) );
	}
*/
}
