/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.general.xml.persistency.loader.impl;

import com.frojasg1.general.ExecutionFunctions;
import com.frojasg1.general.string.translator.GenericStringTranslator;
import com.frojasg1.general.xml.XmlElement;
import com.frojasg1.general.xml.persistency.loader.XmlToModel;
import java.util.Objects;
import java.util.function.Supplier;

/**
 *
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 */
public abstract class XmlToModelBase<CC> implements XmlToModel<CC>
{
// Name of xml element: <$1> but was expected <$2>

	protected String getNameOfXmlNotExpectedErrorString( String name, String expectedName )
	{
		return( String.format( "Name of xml element: <%s> but was expected <%s>", name, expectedName ) );
	}

	protected void assertXmlElementName( XmlElement xe, String expectedName )
	{
		String name = ( xe == null ) ? null : xe.getName();
		if( !Objects.equals( name, expectedName ) )
			throw( new RuntimeException( getNameOfXmlNotExpectedErrorString( xe.getName(), expectedName ) ) );
	}

	

	protected XmlElement getMandatoryChild( XmlElement parent, String childName )
	{
		return( getMandatoryChild(parent, childName, () -> childNotPresentErrorMessage( parent.getName(), childName ) ) );
	}

	protected XmlElement getMandatoryChild( XmlElement parent, String childName,
											Supplier<String> errorMessageBuilder )
	{
		XmlElement result = parent.getChild( childName );
		if( result == null )
			throw( new RuntimeException( errorMessageBuilder.get() ) );

		return( result );
	}

	protected XmlElement getMandatoryChildErrorMessage( XmlElement parent,
														String childName )
	{
		return( getMandatoryChild( parent, childName,
				() -> childNotPresentErrorMessage( parent.getName(), childName ) ) );
	}

	protected XmlElement getMandatoryChild( XmlElement parent, String childName, String errorMessage )
	{
		return( getMandatoryChild(parent, childName, () -> errorMessage ) );
	}

	protected <C> C getMandatoryChildObject( XmlElement parent,
											String childName, Class<C> clazz )
	{
		return( getObject( getMandatoryChildErrorMessage( parent, childName ),
							clazz ) );
	}

	protected Boolean getMandatoryChildBoolean( XmlElement parent, String childName )
	{
		return( getMandatoryChildObject( parent, childName,	Boolean.class ) );
	}

	protected Integer getMandatoryChildInt( XmlElement parent, String childName )
	{
		return( getMandatoryChildObject( parent, childName,	Integer.class ) );
	}

	protected String getMandatoryChildString( XmlElement parent, String childName )
	{
		return( getMandatoryChildObject( parent, childName,	String.class ) );
	}

	protected Double getMandatoryChildDouble( XmlElement parent, String childName )
	{
		return( getMandatoryChildObject( parent, childName,	Double.class ) );
	}

	protected <C> C getMandatoryChildObject( XmlElement parent, String childName,
											Class<C> clazz, Supplier<String> errorMessageBuilder )
	{
		return( getObject( getMandatoryChild( parent, childName, errorMessageBuilder), clazz ) );
	}

	protected String getMandatoryChildStringErrorMessage( XmlElement parent, String childName )
	{
		return( getMandatoryChildObject( parent, childName, String.class ) );
	}

	protected String getMandatoryChildString( XmlElement parent, String childName, String errorMessage )
	{
		String result = null;
		XmlElement resultXe = getMandatoryChild( parent, childName, errorMessage );
		if( resultXe != null )
			result = resultXe.getText();

		return( result );
	}

	protected String getOptionalChildString( XmlElement parent, String childName )
	{
		String result = null;
		if( parent != null )
		{
			XmlElement resultXe = parent.getChild( childName );
			if( resultXe != null )
				result = resultXe.getText();
		}

		return( result );
	}

	protected <C> C getOptionalChildObject( XmlElement parent, String childName,
											Class<C> clazz )
	{
		return( translateObject( getOptionalChildString( parent, childName ), clazz ) );
	}

	protected String childNotPresentErrorMessage( String parentName, String childName )
	{
		return( String.format( "%s without %s in xml", parentName, childName ) );
	}

	protected int getInt( XmlElement xmlElement, String childName, String parentName )
	{
		return( (int) getLong( xmlElement, childName, parentName ) );
	}

	protected String getLongErrorText( String childName, String parentName )
	{
		return( String.format( "%s -> %s did not have a valid Long.", childName, parentName ) );
	}
		
	protected long getLong( XmlElement xmlElement, String childName, String parentName )
	{
		XmlElement xe = getMandatoryChild(xmlElement, childName, childNotPresentErrorMessage(parentName, childName ) );
//		Long result = IntegerFunctions.parseLong( xe.getText() );

		Double result = ExecutionFunctions.instance().safeFunctionExecution( () -> Double.parseDouble( xe.getText() ) );
		if( result == null )
			throw( new RuntimeException( getLongErrorText( parentName, childName ) ) );

		return( (long) ( (double) result ) );
	}

	protected <C> C getObject( XmlElement elem, Class<C> clazz )
	{
		C result = null;
		if( elem != null )
			result = translateObject( elem.getText(), clazz );

		return( result );
	}

	protected <C> C translateObject( String str, Class<C> clazz )
	{
		C result = null;
		if( str != null )
			result = GenericStringTranslator.instance().fromString( str, clazz );

		return( result );
	}
}
