/*
 * Copyright (c) 2023. Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 3.0 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You may obtain a copy of the License at
 *
 *       http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */

package com.frojasg1.general.desktop.color.colorfunctions.impl;

import com.frojasg1.general.desktop.color.colorfunctions.ColorFunctionBase;
import com.frojasg1.general.desktop.image.ImageFunctions;
import com.frojasg1.general.number.DoubleFunctions;
import org.apache.http.annotation.NotThreadSafe;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.Arrays;
import java.util.function.UnaryOperator;

@NotThreadSafe
public class KeyColorColorationFunction extends ColorFunctionBase {
    private static final Logger LOGGER = LoggerFactory.getLogger(KeyColorColorationFunction.class);

    protected int abscissa;
    protected Integer color;

    protected ComponentColorFunction[] componentColorFunctions;

    public KeyColorColorationFunction(Integer abscissa, Integer color) {
        this.abscissa = abscissa != null ? abscissa : 128;
        this.color = color;
    }

    public KeyColorColorationFunction(Integer color) {
        this(null, color);
    }

    public KeyColorColorationFunction(Integer color, boolean idempotentTransformation) {
        this((color != null) && idempotentTransformation ?
                        (int) ImageFunctions.instance().getGrayScale(color) : // (!!!!!)
                        null,
                color);
    }

    public int getAbscissa() {
        return abscissa;
    }

    public Integer getColor() {
        return color;
    }

    protected void validate() {
        if((getAbscissa() < 0) || (getAbscissa() > 255)) {
            throw new IllegalArgumentException(String.format("Color component out of range: %d", getAbscissa()));
        }
    }

    @Override
    public KeyColorColorationFunction init() {
        validate();

        super.init();

        if(getColor() != null) {
            double[] components = getComponents(getColor());

            if(LOGGER.isDebugEnabled()) {
                LOGGER.debug("components: {}", Arrays.toString(components));
            }
            componentColorFunctions = new ComponentColorFunction[3];
            for(int ii=0; ii<componentColorFunctions.length; ii++) {
                componentColorFunctions[ii] = createComponentColorFunction(abscissa, components[ii]);
            }
        }

        return this;
    }

    protected double complementary(double value) {
        return 255 - value;
    }

    @Override
    public Integer apply(Integer origColor) {
        Integer result = origColor;
        double components[] = applyInternal(origColor);
        if(components != null) {
            result = getARGB(components);
        }

        return result;
    }

    public double[] applyInternal(Integer origColor) {
        double[] result = null;
        if(hasToTranslate()) {
            double components[] = getComponents(origColor);

            int greyScale = getGrayScale(origColor);
            for (int ii = 0; ii < componentColorFunctions.length; ii++) {
                components[ii] = componentColorFunctions[ii].apply(greyScale);
            }
            result = components;
        }
        return result;
    }

    protected boolean hasToTranslate() {
        return getColor() != null;
    }

    protected ComponentColorFunction createComponentColorFunction(double abscissa, double colorComponent) {
        return new ComponentColorFunction(abscissa, colorComponent);
    }

    protected class ComponentColorFunction implements UnaryOperator<Integer>  {
        double abscissa;
        double color;
        double leftSlope = 1d;
        double rightSlope = 1d;

        public ComponentColorFunction(double abscissa, double color) {
            this.abscissa = DoubleFunctions.instance().limit(abscissa, 1d, 254d);
            this.color = color;
            this.leftSlope = color / abscissa;
            this.rightSlope = complementary(color) / complementary(abscissa);
        }

        protected boolean hasToApplyLeftLinearInterpolation(int componentValue) {
            return (componentValue < abscissa);
        }

        protected double applyLinearTransformation(double value, double slope) {
            return slope * value;
        }

        protected Integer applyLeftLinearTransformation(Integer value) {
            return (int) applyLinearTransformation(value, leftSlope);
        }

        protected Integer applyRightLinearTransformation(Integer value) {
            return (int) complementary(applyLinearTransformation(complementary(value), rightSlope));
        }

        @Override
        public Integer apply(Integer value) {
            return hasToApplyLeftLinearInterpolation(value) ?
                    applyLeftLinearTransformation(value) :
                    applyRightLinearTransformation(value);
        }
    }
}
