/*
 * Copyright (c) 2023. Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 3.0 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You may obtain a copy of the License at
 *
 *       http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */

package com.frojasg1.general.desktop.color.colorfunctions.impl;

import com.frojasg1.general.desktop.color.colorfunctions.ColorFunctionBase;
import com.frojasg1.general.number.DoubleFunctions;
import org.apache.http.annotation.NotThreadSafe;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

@NotThreadSafe
public class TransparencyColorFunction extends ColorFunctionBase {
    private static final Logger LOGGER = LoggerFactory.getLogger(TransparencyColorFunction.class);

    protected static final double DEFAULT_SQUARE_COLOR_TOLERANCE = 0d;


    protected double[] colorComponentsForTotalTransparency;

    protected Integer colorForTotalTransparency;
    protected Integer globalTransparency;
    protected double squareColorTolerance = 0d;

    protected long numTransparencies = 0;

    public TransparencyColorFunction(Integer colorForTotalTransparency, Integer globalTransparency, Double colorTolerance) {
        this.colorForTotalTransparency = colorForTotalTransparency;
        this.globalTransparency = globalTransparency;
        if(colorTolerance != null) {
            this.squareColorTolerance = 4 * colorTolerance * colorTolerance;
        }
    }

    public TransparencyColorFunction(Integer colorForTotalTransparency, Integer globalTransparency) {
        this(colorForTotalTransparency, globalTransparency, DEFAULT_SQUARE_COLOR_TOLERANCE);
    }

    public TransparencyColorFunction(Integer colorForTotalTransparency) {
        this(colorForTotalTransparency, null);
    }

    protected void validateGlobalTransparency(Integer globalTransparency) {
        if((globalTransparency != null) && ((globalTransparency > 255) || (globalTransparency < 0))) {
            throw new IllegalArgumentException(String.format("Global transparency not valid: %d", globalTransparency));
        }
    }

    protected void validate() {
        validateGlobalTransparency(getGlobalTransparency());
    }

    @Override
    public TransparencyColorFunction init() {
        validate();

        super.init();

        initColorComponentsForTotalTransparency();

        return this;
    }

    protected void initColorComponentsForTotalTransparency() {
        Integer colorForTotalTransparency = getColorForTotalTransparency();
        if(colorForTotalTransparency != null) {
            colorComponentsForTotalTransparency = createEmptyComponents();
            fillComponents(colorForTotalTransparency, colorComponentsForTotalTransparency);
        }
    }

    protected double[] getColorComponentsForTotalTransparency() {
        return colorComponentsForTotalTransparency;
    }

    protected void setColorComponentsForTotalTransparency(double[] colorComponentsForTotalTransparency) {
        this.colorComponentsForTotalTransparency = colorComponentsForTotalTransparency;
    }

    protected Integer getColorForTotalTransparency() {
        return colorForTotalTransparency;
    }

    protected Integer getGlobalTransparency() {
        return globalTransparency;
    }

    protected double getSquareColorTolerance() {
        return squareColorTolerance;
    }

    @Override
    public Integer apply(Integer origColor) {
        double components[] = getComponents(origColor);

        boolean changed = false;
        if(hasToCheckIfTotalTransparence()) {
            if(hasToSetTotalTransparency(components)) {
                components[3] = 0x00;
                numTransparencies++;
                changed = true;
            }
        }

        if(!changed && globalTransparencySet()) {
            components[3] = getGlobalTransparency().doubleValue();
        }

        return getImageFunctions().getARGB(components);
    }

    protected boolean hasToCheckIfTotalTransparence() {
        return getColorComponentsForTotalTransparency() != null;
    }

    protected double getSquareDistance(double[] arr1, double[] arr2) {
        return DoubleFunctions.instance().getSquareDistance(arr1, arr2);
    }

    protected boolean hasToSetTotalTransparency(double[] components) {
        double squareDistance = getSquareDistance(components, getColorComponentsForTotalTransparency());
        return(squareDistance <= getSquareColorTolerance());
    }

    protected boolean globalTransparencySet() {
        return getGlobalTransparency() != null;
    }

    public long getNumTransparencies() {
        return numTransparencies;
    }

    public void resetNumTransparencies() {
        numTransparencies = 0;
    }
}
