/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.general.desktop.edt;

import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import javax.swing.SwingUtilities;

/**
 *
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 */
public class EventDispatchThreadInvokeLaterPurge {

	protected Map<String, FunctionContext> _map;
	protected int _maxToSkip;

	public EventDispatchThreadInvokeLaterPurge(int maxToSkip)
	{
		this( maxToSkip, true );
	}
	
	public EventDispatchThreadInvokeLaterPurge( int maxToSkip, boolean init )
	{
		_maxToSkip = maxToSkip;
		if( init )
			init();
	}

	protected void init()
	{
		_map = createMap();
	}

	protected <KK, VV> Map<KK, VV> createMap()
	{
		return( new ConcurrentHashMap() );
	}

	protected FunctionContext createFunctionContext()
	{
		return( new FunctionContext( _maxToSkip ) );
	}

	public void invokeLater( String functionName, Runnable runnable )
	{
		FunctionContext ctxt = _map.computeIfAbsent(functionName, fn -> createFunctionContext() );

		ctxt.invokeLater(runnable);
	}

	protected class FunctionContext
	{
		protected Map<Long, Runnable> _contextMap;
		protected int _contextMaxToSkip;
		protected int _skipped;

		protected long _index;

		public FunctionContext( int maxToSkip )
		{
			_maxToSkip = maxToSkip;
			_contextMap = new ConcurrentHashMap<>();
		}

		protected synchronized long incKey()
		{
			return( _index++ );
		}

		public void invokeLater( Runnable runnable )
		{
			long key = incKey();
			_contextMap.put( key, runnable );

			SwingUtilities.invokeLater( () -> run( key ) );
		}

		protected void run( long key )
		{
			Runnable runnable = _contextMap.remove( key );
			if( ( runnable != null ) && hasToRun() )
				runnable.run();
		}

		protected synchronized boolean hasToRun()
		{
			boolean hasToRun = _contextMap.isEmpty();
			if( !hasToRun )
				hasToRun = ( ++_skipped > _contextMaxToSkip );

			if( hasToRun )
				_skipped = 0;

			return( hasToRun );
		}
	}
}
