/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.general.desktop.generic.dialogs.helper;

import com.frojasg1.applications.common.components.internationalization.window.InternationalizedWindow;
import com.frojasg1.applications.common.components.zoom.SwitchToZoomComponents;
import com.frojasg1.applications.common.configuration.application.BaseApplicationConfigurationInterface;
import com.frojasg1.general.CollectionFunctions;
import com.frojasg1.general.desktop.classes.Classes;
import com.frojasg1.general.desktop.generic.dialogs.helper.contexts.BrowseComponentHierarchyProcessingContext;
import com.frojasg1.general.desktop.generic.dialogs.helper.contexts.impl.BrowseComponentHierarchyResizeComponentContext;
import com.frojasg1.general.desktop.generic.dialogs.helper.contexts.impl.BrowseComponentHierarchyZoomFontContext;
import com.frojasg1.general.desktop.generic.dialogs.impl.DesktopFileChooserParameters;
import com.frojasg1.general.desktop.view.ComponentFunctions;
import com.frojasg1.general.desktop.view.FrameworkComponentFunctions;
import com.frojasg1.general.desktop.view.ViewFunctions;
import com.frojasg1.general.desktop.view.jtable.JTableFunctions;
import com.frojasg1.general.desktop.view.zoom.filechooser.ConvertDialogComponents;
import com.frojasg1.general.desktop.view.zoom.imp.ZoomIconImp;
import com.frojasg1.general.number.DoubleReference;
import com.frojasg1.general.number.IntegerFunctions;
import com.frojasg1.general.reflection.ReflectionFunctions;
import com.frojasg1.general.zoom.ZoomInterface;
import com.frojasg1.generic.GenericFunctions;
import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Insets;
import java.awt.LayoutManager;
import java.awt.Point;
import java.util.List;
import java.util.Locale;
import java.util.Set;
import java.util.function.Consumer;
import java.util.function.Predicate;
import javax.swing.Icon;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JPopupMenu;
import javax.swing.JScrollBar;
import javax.swing.JTable;
import javax.swing.plaf.ComponentUI;
import javax.swing.plaf.basic.BasicComboPopup;

/**
 *
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 */
public class DesktopDialogsWrapperHelper
{
	protected static DesktopDialogsWrapperHelper INSTANCE = new DesktopDialogsWrapperHelper();

	public static DesktopDialogsWrapperHelper instance()
	{
		return( INSTANCE );
	}

	public void zoomScrollBar( JScrollBar sb, double zoomFactor )
	{
		if( sb != null )
		{
			if( sb.getOrientation() == JScrollBar.HORIZONTAL )
				zoomHorizontalScrollBar( sb, zoomFactor );
			else if( sb.getOrientation() == JScrollBar.VERTICAL )
				zoomVerticalScrollBar( sb, zoomFactor );
		}
	}

	protected void zoomHorizontalScrollBar( JScrollBar hsb, double zoomFactor )
	{
		if( hsb != null )
		{
			Dimension size = hsb.getPreferredSize();
			hsb.setPreferredSize( new Dimension( size.width,
									IntegerFunctions.zoomValueCeil(hsb.getPreferredSize().height, zoomFactor)
													)
									);
		}
	}
	
	protected void zoomVerticalScrollBar( JScrollBar vsb, double zoomFactor )
	{
		if( vsb != null )
		{
			Dimension size = vsb.getPreferredSize();
			vsb.setPreferredSize( new Dimension(
										IntegerFunctions.zoomValueCeil( vsb.getPreferredSize().width, zoomFactor),
										size.width )
									);
		}
	}

	protected boolean hasSameReference(List<Object> list, Object obj) {
		return CollectionFunctions.instance().getIndexOfSame(list, obj) != null;
	}

	public void zoomFont(Component comp, double zoomFactor,
						 List<Object> processedObjs)
	{
		Font newFont = zoomFont(comp.getFont(), zoomFactor, processedObjs);
		if (newFont != null) {
			comp.setFont(newFont);
		}
	}

	public Font zoomFont(Font font, double zoomFactor,
						 List<Object> processedObjs)
	{
		Font result = null;
		if((processedObjs == null) || !hasSameReference(processedObjs, font)) {
			result = deriveFont(font, zoomFactor);
			if (result != null) {
				if(processedObjs != null) {
					processedObjs.add(result);
				}
			}
		}
		return result;
	}

	public Font deriveFont(Font originalFont, double factor)
	{
		Font result = null;

		if( originalFont != null )
		{
			if( factor > 0.0F )
			{
				result = originalFont.deriveFont( (float) ( factor * originalFont.getSize2D() ) );
			}
		}

		return( result );
	}

	protected void M_changeFontToApplicationFontSize(Component comp,
													 double zoomFactor,
													 int level,
													 int fromLevelToZoomFont)
	{
		processComponentHierarchy(comp, zoomFactor, level, fromLevelToZoomFont,
				createZoomFontContext(null, zoomFactor));
	}

	protected ViewFunctions getViewFunctions() {
		return ViewFunctions.instance();
	}

	public void zoomSizedSize(Component comp, double zoomFactor, boolean inconditionalSetPreferredSize) {
		if(comp.isPreferredSizeSet() || inconditionalSetPreferredSize) {
			comp.setPreferredSize(getViewFunctions().getNewDimension(comp.getPreferredSize(), zoomFactor));
		}
		if(comp.isMinimumSizeSet()) {
			comp.setMinimumSize(getViewFunctions().getNewDimension(comp.getMinimumSize(), zoomFactor));
		}
		if((comp.getParent() != null) && ComponentFunctions.instance().getLayout(comp.getParent()) == null) {
			comp.setBounds(getViewFunctions().calculateNewBounds(comp.getBounds(), null, ViewFunctions.ORIGIN, zoomFactor));
		}
	}

	protected void processComponentHierarchy(Component comp,
											 double zoomFactor,
											 int level,
											 int fromLevelToZoomFont,
											 BrowseComponentHierarchyProcessingContext processingContext)
//															 Set<Component> processedComps,
//															 List<Object> processedObjs,
//															 Predicate<Component> specialComponentProcessing)
	{
//		System.out.println( StringFunctions.instance().buildStringFromRepeatedChar( ' ', level * 4) +
//							comp.getClass().getName() );
		if(processingContext.hasAlreadyBeenInvoked(comp)) {
			return;
		}

		if(fromLevelToZoomFont < level )
			processingContext.processComponent(comp);

		Class<?> filePaneClass = Classes.getFilePaneClass();	// JRE 7, 8, ...

		if( ( comp instanceof JComboBox ) && !isZoomed( (JComponent) comp) )
		{
			JComboBox combo = (JComboBox) comp;
			BasicComboPopup popup = (BasicComboPopup) combo.getUI().getAccessibleChild(combo, 0);

			processComponentHierarchy( popup, zoomFactor, level + 1, level + 4,
					processingContext);
		}
//		else if( comp instanceof FilePane )
		else if( ( filePaneClass != null ) && filePaneClass.isInstance( comp ) )	// JRE 7, 8, ...
		{
//			JPopupMenu popupMenu = fp.getComponentPopupMenu();
			JPopupMenu popupMenu = (JPopupMenu) ReflectionFunctions.instance().invokeMethod( "getComponentPopupMenu",
																							comp.getClass(), comp );
			processComponentHierarchy( popupMenu, zoomFactor, level + 1, -1,
					processingContext);
/*
			JTable detailsTable = (JTable) ReflectionFunctions.instance().getAttribute("detailsTable", JTable.class, comp);
			if( detailsTable != null )
				M_changeFontToApplicationFontSize( detailsTable, zoomFactor, level + 1, -1 );

			JList list = (JList) ReflectionFunctions.instance().getAttribute("list", JList.class, comp);
			if( list != null )
				M_changeFontToApplicationFontSize( list, zoomFactor, level + 1, -1 );
*/
		}
		else if( ( comp instanceof JScrollBar ) && !isZoomed( (JComponent) comp) )
		{
			zoomScrollBar( (JScrollBar) comp, zoomFactor );
		}
		else if( comp instanceof JTable )
		{
//			JTableFunctions.instance().zoomCellRenderer( ( (JTable) comp ).getTableHeader().getDefaultRenderer(), zoomFactor );
			JTableFunctions.instance().switchCellEditors( (JTable) comp );

			processComponentHierarchy( ( (JTable) comp ).getTableHeader(), zoomFactor, level + 1,
					-1, processingContext);
		}
		else if( comp instanceof JLabel )
		{
			JLabel jlabel = (JLabel) comp;
			Icon icon = jlabel.getIcon();
			if( icon != null )
			{
				boolean centerSmaller = true;
				ZoomIconImp zi = new ZoomIconImp( icon, new DoubleReference( zoomFactor ), centerSmaller );
				jlabel.setIcon( zi );

				Dimension newDimension = ViewFunctions.instance().getNewDimension(jlabel.getSize(),
																				null, zoomFactor );
				jlabel.setSize(newDimension);
			}
		}

		if( comp instanceof Container	)
		{
			Container contnr = (Container) comp;
			Consumer<Component> childProcessingFunction = processingContext.getProcessChildFunction(contnr);
			Component child;
			for( int ii=0; ii<contnr.getComponentCount(); ii++ )
			{
				child  = contnr.getComponent(ii);
				processComponentHierarchy(child, zoomFactor, level + 1,
						fromLevelToZoomFont, processingContext);

				childProcessingFunction.accept(child);
			}

			if( contnr instanceof JMenu )
			{
				JMenu jmnu = (JMenu) contnr;
				for( int ii=0; ii<jmnu.getMenuComponentCount(); ii++ )
					processComponentHierarchy(jmnu.getMenuComponent( ii ), zoomFactor,
							level + 1, fromLevelToZoomFont, processingContext);
			}

			comp.doLayout();
		}
	}

	protected boolean isZoomed( JComponent comp )
	{
		boolean result = false;
		ComponentUI ui = ComponentFunctions.instance().getUI(comp);
		
		if( ( ui != null ) && ( ui.getClass().getName().contains("frojasg1") ) )
			result = true;

		return( result );
	}

	public Set<Component> M_changeFontToApplicationFontSizeAndApplyColorInversion_forComponent( double zoomFactor,
																								Component comp,
																								Component parent)
	{
		return( M_changeFontToApplicationFontSizeAndApplyColorInversion_forComponent( zoomFactor, comp,
																						parent, null ) );
	}

	public Set<Component> M_changeFontToApplicationFontSizeAndApplyColorInversion_forComponent( double zoomFactor,
																								Component comp,
																								Component parent,
																								Predicate<Component> compFilter)
	{
		boolean isDarkMode = isDarkMode();
//		double factor = 0.0D;
		double factor = zoomFactor;
		Set<Component> invertedComponentColors = null;

		if( factor > 0.0F )
		{
//			System.out.println( String.format( "%n%n%nAntes%n=====" ) +
//								ViewFunctions.instance().traceComponentTree(comp, (component) -> "preferredSize : " + component.getPreferredSize() + "     size : " + component.getSize() ));


//			System.out.println( ViewFunctions.instance().traceComponentTree(comp));

			SwitchToZoomComponents stzc = new SwitchToZoomComponents( null, null );
//			stzc.changeToColorInvertibleComponents( comp, isDarkMode );
			ConvertDialogComponents.instance().changeToColorInvertibleComponents( comp, isDarkMode );

			invertedComponentColors = invertColorsIfNecessary(comp, parent, compFilter, null);

			M_changeFontToApplicationFontSize( comp, factor );

			stzc.setZoomFactor(comp, zoomFactor, (comp_) -> {
				boolean result = ( comp_ != null ) &&
					(
						( comp_.getClass().getName().endsWith( "AlignedLabel" ) ) ||
						( comp_ instanceof ZoomInterface )
					);
				return( result );
			});

//			System.out.println( String.format( "%n%n%n" ) +
//								ViewFunctions.instance().traceComponentTree(comp, (component) -> stzc.getComponentResizingResult().get(component) ));
			System.out.println( String.format( "%n%n%n" ) +
					ViewFunctions.instance().traceComponentTree(comp,
							ViewFunctions.instance()::getInputMapToString) );

//			FileChooserUI ui;
//			if( ( comp instanceof JFileChooser ) ||
//				( comp instanceof JOptionPane ) )
			if( comp instanceof JFileChooser )
			{
				Dimension dim = comp.getPreferredSize();
				int width = (int) ( ( (float) dim.getWidth()) * factor );
				int height = (int) ( ( (float) dim.getHeight()) * factor );
				Dimension newDim = new Dimension( width, height);

//				comp.setPreferredSize( newDim );
			}

//			System.out.println( ViewFunctions.instance().traceComponentTree(comp, (comp_) -> getInfoString(comp_) ) );

//			System.out.println( String.format( "%n%n%nDespues%n=======" ) +
//								ViewFunctions.instance().traceComponentTree(comp, (component) -> "preferredSize : " + component.getPreferredSize() + "     size : " + component.getSize() ));
		}
		
		return( invertedComponentColors );
	}

	protected String getInfoString( Component comp )
	{
		String result = "";
		if( comp instanceof Container )
		{
			Container cont = (Container) comp;
			Insets insets = cont.getInsets();
			if( insets != null )
			{
				result = " insets - " + insets;
			}

			LayoutManager lm = cont.getLayout();
			if( lm != null )
			{
				result += "        " + lm.getClass().getName();
				if( lm instanceof BorderLayout )
				{
					result += " ( " + lm + " )";
				}
			}
		}
		return( result );
	}

	protected BaseApplicationConfigurationInterface getAppliConf()
	{
		return( GenericFunctions.instance().getAppliConf() );
	}

	public void setDefaultLocale( DesktopFileChooserParameters dfcp ) {
		Locale locale = dfcp.getLocale();
		if( locale == null )
		{
			BaseApplicationConfigurationInterface appliConf = getAppliConf();
			if( appliConf != null )
			{
				String language = appliConf.getLanguage();
				locale = GenericFunctions.instance().getObtainAvailableLanguages().getLocaleOfLanguage( language );
				dfcp.setLocale( locale );
			}
		}
	}

	public boolean isDarkMode()
	{
		boolean result = false;

		BaseApplicationConfigurationInterface appliConf = getAppliConf();
		if( appliConf != null )
			result = appliConf.isDarkModeActivated();

		return( result );
	}

	protected Set<Component> invertColorsIfNecessary( Component target, Component parent,
														Predicate<Component> compFilter,
														Set<Component> alreadyDone )
	{
		if( ( target != null ) && ( parent != null ) )
		{
			InternationalizedWindow iw = FrameworkComponentFunctions.instance().getInternationalizedWindow(parent);
			if( ( iw != null ) && ( iw.isDarkMode() ) )
				alreadyDone = iw.getColorInversor().setOneShotInconditionalDarkMode(target, compFilter, alreadyDone);
		}
		return( alreadyDone );
	}

	public Point getCenteredLocationForComponent( Component comp )
	{
		int width = java.awt.Toolkit.getDefaultToolkit().getScreenSize().width;
		int height = java.awt.Toolkit.getDefaultToolkit().getScreenSize().height;
		Point result = new Point( width/2 - comp.getWidth()/2, height/2 - comp.getHeight()/2 );

		return( result );
	}
//
//	protected boolean specialComponentZoomProcessingForJColorChooser(Component comp, double zoomFactor) {
//		boolean result = false;
//		if(comp.getClass().getName().equals("javax.swing.colorchooser.ColorChooserPanel")) {
//			result = true;
//			zoomSizedSize(comp, zoomFactor, true);
//		}
//		return result;
//	}
//
//	public void M_changeFontToApplicationFontSizeForJColorChooser( Component comp, double zoomFactor,
//																   Set<Component> processedComps )
//	{
//		M_changeFontToApplicationFontSize( comp, zoomFactor,
//				comp1 -> specialComponentZoomProcessingForJColorChooser(comp1, zoomFactor), processedComps);
//	}

	public void M_changeFontToApplicationFontSize( Component comp, double zoomFactor )
	{
		M_changeFontToApplicationFontSize( comp, zoomFactor, null);
	}

	public void zoomComponents(Component comp, double zoomFactor,
							   Predicate<Component> specialComponentProcessing )
	{
		processComponentHierarchy( comp, zoomFactor, 0, -1,
				createResizeCompsContext(specialComponentProcessing, zoomFactor));
	}

	public void M_changeFontToApplicationFontSize( Component comp, double zoomFactor,
												   Set<Component> processedComps)
	{
		processComponentHierarchy( comp, zoomFactor, 0, -1,
				createZoomFontContext(processedComps, zoomFactor));
	}

	protected BrowseComponentHierarchyProcessingContext createZoomFontContext(Set<Component> processedComps,
																			  double zoomFactor) {
		return new BrowseComponentHierarchyZoomFontContext(processedComps, zoomFactor);
	}

	protected BrowseComponentHierarchyProcessingContext createResizeCompsContext(Predicate<Component> specialComponentProcessing,
																				 double zoomFactor) {
		return new BrowseComponentHierarchyResizeComponentContext(specialComponentProcessing, zoomFactor);
	}
}
