/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.general.desktop.image;

import com.frojasg1.general.number.IntegerFunctions;
import java.awt.Graphics2D;
import java.awt.image.BufferedImage;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.function.Function;

/**
 *
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 */
public class ResizeImageAccurate extends ResizeImageAccurateGen<Integer> implements ResizeImageInterface
{
	protected static ResizeImageAccurate _instance = null;
	
	public static ResizeImageAccurate instance()
	{
		if( _instance == null )
			_instance = new ResizeImageAccurate();

		return( _instance );
	}

	protected BufferedImage resizeImageJava( BufferedImage originalImage, int newWidth, int newHeight )
	{
		BufferedImage resizedImage = new BufferedImage(newWidth , newHeight, originalImage.getType() );
		Graphics2D g = resizedImage.createGraphics();
		g.drawImage(originalImage, 0, 0, newWidth , newHeight , null);
		g.dispose();

		return( resizedImage );
	}

	public BufferedImage resizeImage( BufferedImage originalImage, int newWidth,
										int newHeight,
										Function<Integer, Integer> colorTranslator )
		throws IllegalArgumentException
	{
		return( resizeImageManual( originalImage, newWidth, newHeight, colorTranslator, null ) );
	}

	@Override
	public BufferedImage resizeImage( BufferedImage originalImage, int newWidth, int newHeight ) throws IllegalArgumentException
	{
		return( resizeImage( originalImage, newWidth, newHeight, (AtomicBoolean) null ) );
	}

	public BufferedImage resizeImage( BufferedImage originalImage,
										int newWidth, int newHeight,
										AtomicBoolean hasToCancel) throws IllegalArgumentException
	{
		return( resizeImageManual( originalImage, newWidth, newHeight, null, hasToCancel ) );
	}

	protected BufferedImage resizeImage( BufferedImage originalImage, int newWidth,
											int newHeight, Function<Integer, Integer> colorTranslator,
											AtomicBoolean hasToCancel)
		throws IllegalArgumentException
	{
		return( resizeImageManual( originalImage, newWidth, newHeight,
									colorTranslator, hasToCancel ) );
	}

	protected BufferedImage resizeImageManual( BufferedImage originalImage, int newWidth,
											int newHeight, Function<Integer, Integer> colorTranslator,
											AtomicBoolean hasToCancel)
		throws IllegalArgumentException
	{
		BufferedImage result = null;
		if(originalImage != null) {
			if ((newWidth < 1) || (newHeight < 1))
				throw (new IllegalArgumentException("Bad size for image.   Width: " + newWidth + ". Height: " + newHeight));

			result = new BufferedImage(newWidth, newHeight, BufferedImage.TYPE_INT_ARGB);
			BufferedImage finalResult = result;
			int[] pixels = ImageFunctions.instance().getRGB(0, 0, originalImage.getWidth(), originalImage.getHeight(), originalImage);

			DoublePixelComponents dpc = new DoublePixelComponents();

			resizeImageManual(originalImage.getWidth(), originalImage.getHeight(), newWidth, newHeight, colorTranslator,
					pixels, dpc::reset, dpc::addPixelToComponents, dpc::getRGB, finalResult::setRGB, hasToCancel);
		}

		return( result );
	}

	protected static class DoublePixelComponents
	{
		protected double[] _components = new double[4];
		protected double _factorAlpha0 = 0.0D;

		public DoublePixelComponents()
		{
			reset();
		}

		public void reset()
		{
			_factorAlpha0 = 0;
			for( int ii=0; ii< _components.length; ii++ )
				_components[ii] = 0;
		}

		public void addPixelToComponents( int argb, double factor )
		{
			double alpha = ( argb >>> 24 );
			_components[0] += alpha * factor;

			if( alpha == 0 )
				_factorAlpha0 += factor;
			else
			{
				double rr = ( ( argb & 0x00FF0000 ) >>> 16 );
				double gg = ( ( argb & 0x0000FF00 ) >>> 8 );
				double bb = ( argb & 0x000000FF );

				_components[1] += rr * factor;
				_components[2] += gg * factor;
				_components[3] += bb * factor;
			}
		}

		protected int normalize( double value )
		{
			return( IntegerFunctions.min( 255, IntegerFunctions.roundToInt(value) ) );
		}

		public int getRGB()
		{
			int result = ( normalize( _components[0] ) << 24 ) |
						( normalize( _components[1] / ( 1 - _factorAlpha0 ) ) << 16 ) |
						( normalize( _components[2] / ( 1 - _factorAlpha0 ) ) << 8 ) |
						normalize( _components[3] / ( 1 - _factorAlpha0 ) );

			return( result );
		}
	}
}
