/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.general.desktop.view.color.editors;

import com.frojasg1.applications.common.components.internationalization.JFrameInternationalization;
import com.frojasg1.applications.common.components.internationalization.window.InternationalizationOwner;
import com.frojasg1.applications.common.configuration.application.BaseApplicationConfigurationInterface;
import com.frojasg1.general.desktop.view.FontFunctions;
import com.frojasg1.general.desktop.view.FrameworkComponentFunctions;
import com.frojasg1.general.desktop.view.ViewFunctions;
import com.frojasg1.general.desktop.view.color.ColorInversor;
import com.frojasg1.general.desktop.view.zoom.componentcopier.GenericCompCopier;
import com.frojasg1.generic.GenericFunctions;
import java.awt.Component;
import java.lang.ref.WeakReference;
import java.util.EventObject;
import javax.swing.JTable;
import javax.swing.event.CellEditorListener;
import javax.swing.table.TableCellEditor;
import javax.swing.text.JTextComponent;
import org.slf4j.LoggerFactory;

/**
 *
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 */
public class TableCellEditorColorInversor implements TableCellEditor
{
    private static final org.slf4j.Logger LOGGER = LoggerFactory.getLogger(TableCellEditorColorInversor.class);

	protected WeakReference<JTable> _jTableRef;

	protected TableCellEditor _origEditor;
	protected ColorInversor _colorInversor;
	protected BaseApplicationConfigurationInterface _appliConf;

//	protected Component _latestCellEditorComponent;
//	protected float _originalEditorComponentFontSize;
//	protected double _latestZoomFactor = -1;
//	protected Boolean _latestModeWasDark;

	public TableCellEditorColorInversor( JTable table,
										TableCellEditor origEditor,
										ColorInversor colorInversor,
										BaseApplicationConfigurationInterface appliConf)
	{
		_jTableRef = new WeakReference(table);
		_origEditor = origEditor;
		_colorInversor = colorInversor;
		_appliConf = appliConf;
	}

	protected BaseApplicationConfigurationInterface getAppliConf()
	{
		return( GenericFunctions.instance().getAppliConf() );
	}

	protected <TC extends JTextComponent> JTextComponent cloneTextComponent( TC original )
	{
		TC result = original;
		
		try
		{
			LOGGER.info("InputMap antes: {}", ViewFunctions.instance().getInputMapToString(original));
			result = (TC) original.getClass().newInstance();
			GenericCompCopier.instance().copyToNew(original, result);
			LOGGER.info("InputMap después: {}", ViewFunctions.instance().getInputMapToString(result));
		}
		catch( Exception ex )
		{
			LOGGER.error( "Error clonning JTextComponent: {}", original.getClass() );
			result = original;
		}

		return( result );
	}

	@Override
	public Component getTableCellEditorComponent(JTable table, Object value, boolean isSelected, int row, int column) {
		Component result = _origEditor.getTableCellEditorComponent( table, value,
														isSelected, row, column );

		return( udpateEditorComponent(result, isSelected, row, column) );
	}

	protected Component udpateEditorComponent(Component result, boolean isSelected, int row, int column)
	{
//		if(result != _latestCellEditorComponent)
//		{
//			_latestCellEditorComponent = result;
//			_originalEditorComponentFontSize = result.getFont().getSize();
//			_latestZoomFactor = -1;
//			_latestModeWasDark = null;
//		}

		zoomEditor( result );

		updateDarkMode(result, isSelected, row, column);

		return( result );
	}

	public int getZoomedFontSize( int originalFontSize, double zoomFactor )
	{
		return( FontFunctions.instance().getZoomedFontSize( originalFontSize, zoomFactor) );
	}

	protected int getJTableFontSize()
	{
		return( getJTable().getFont().getSize() );
	}

	protected boolean isFontZoomed(Component result )
	{
		return( getJTableFontSize() == result.getFont().getSize() );
	}

	protected void zoomEditor( Component result )
	{
//		double zoomFactor = getAppliConf().getZoomFactor();
//		if( ( zoomFactor != 1.0d ) && ( _latestZoomFactor != zoomFactor ) )
		if( ! isFontZoomed( result ) )
		{
//			result.setFont(FontFunctions.instance().getZoomedFont(
//					result.getFont(), zoomFactor)
//			);
			result.setFont(FontFunctions.instance().
				getResizedFont(result.getFont(), getJTableFontSize()));
//			_latestZoomFactor = zoomFactor;
		}
	}

	protected void updateDarkMode(Component result, boolean isSelected, int row, int column)
	{
//		boolean isDarkMode = getAppliConf().isDarkModeActivated();
//		if((_latestModeWasDark == null) || (_latestModeWasDark != isDarkMode))
//		{
//			if( (_latestModeWasDark != null) || isDarkMode )
				invertColors( result, isSelected, row, column );

//			_latestModeWasDark = isDarkMode;
//		}
	}

	protected void invertColors(Component result, boolean isSelected,
							int row, int column) {
//		_colorInversor.invertSingleColorsGen(result);

//		_colorInversor.invertBackground(result);
//		_colorInversor.invertForeground(result);
//		if( isSelected )
//		{
//			_colorInversor.invertSelectionBackground(result);
//			_colorInversor.invertSelectionForeground(result);
//		}

//		_colorInversor.invertSingleComponentColors(result);
		boolean isDarkMode = getAppliConf().isDarkModeActivated();
		if( getInternationalization() != null )
			_colorInversor.setSingleDarkMode( result, isDarkMode,
				getInternationalization()::getColorThemeChangeableOnTheFly );
	}

	protected JTable getJTable()
	{
		return( _jTableRef.get() );
	}

	protected JFrameInternationalization getInternationalization()
	{
		return( FrameworkComponentFunctions.instance().getInternationalization( getJTable() ) );
	}

	@Override
	public Object getCellEditorValue() {
		return( _origEditor.getCellEditorValue() );
	}

	@Override
	public boolean isCellEditable(EventObject anEvent) {
		return( _origEditor.isCellEditable(anEvent) );
	}

	@Override
	public boolean shouldSelectCell(EventObject anEvent) {
		return( _origEditor.shouldSelectCell(anEvent) );
	}

	@Override
	public boolean stopCellEditing() {
		return( _origEditor.stopCellEditing() );
	}

	@Override
	public void cancelCellEditing() {
		_origEditor.cancelCellEditing();
	}

	@Override
	public void addCellEditorListener(CellEditorListener l) {
		_origEditor.addCellEditorListener(l);
	}

	@Override
	public void removeCellEditorListener(CellEditorListener l) {
		_origEditor.removeCellEditorListener(l);
	}
}
