/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.general.desktop.view.componentwrapper.jtable.files;

import com.frojasg1.general.collection.functions.ListOperationFunctions;
import com.frojasg1.general.collection.functions.ListRecordMoveContext;
import com.frojasg1.general.collection.functions.ListStateContextBase;
import com.frojasg1.general.desktop.view.componentwrapper.jtable.JTableFileDragAndDropEventContext;
import com.frojasg1.general.desktop.view.componentwrapper.jtable.JTableWrapper;
import com.frojasg1.general.view.evt.wrapper.impl.table.GenTableWrapperEventBase;
import com.frojasg1.general.view.evt.wrapper.impl.table.impl.GenTableWrapperEventBuilderBase;
import com.frojasg1.general.view.evt.wrapper.impl.table.impl.GenTableWrapperRecordsModifiedEventBase;
import com.frojasg1.general.view.evt.wrapper.impl.table.impl.deleterecords.GenTableWrapperRecordsDeletedEvent;
import com.frojasg1.general.view.evt.wrapper.impl.table.impl.moverecords.GenTableWrapperRecordsMovedEvent;
import com.frojasg1.general.view.evt.wrapper.impl.table.impl.sortedrecords.GenTableWrapperRecordsSortedEvent;
import com.frojasg1.general.desktop.view.transfer.file.FileTransferHandler;
import com.frojasg1.general.view.ReleaseResourcesable;
import com.frojasg1.general.view.gencomponentwrapper.jtable.ViewAndModelSelectionArray;

import java.awt.Point;
import java.lang.ref.WeakReference;
import java.util.Comparator;
import java.util.List;
import java.util.function.Function;
import java.util.function.Predicate;
import javax.swing.JTable;
import javax.swing.SwingUtilities;

/**
 *
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 * @param <RR>	- Record
 * @param <UI>	- JComponentUI derived class
 */
public class DragAndDropFilesJTableWrapperContextBase<RR, UI extends BaseFileListJTableUI>
		implements ReleaseResourcesable
{
	protected JTableWrapperDragAndDropFilesDecorator<UI> _dadFilesDecorator;

	protected WeakReference<JTableWrapper<RR, UI>> _tableWrapper;

	protected ViewAndModelSelectionArray _startDraggingRowSelection;

	public DragAndDropFilesJTableWrapperContextBase( JTableWrapper<RR, UI> tableWrapper )
	{
		_tableWrapper = new WeakReference<>(tableWrapper);
	}

	protected void init()
	{
		_dadFilesDecorator = createDragAndDropFilesDecorator();

		_startDraggingRowSelection = getTableWrapper().createViewAndModelRowSelectionArray(0);
	}

	protected JTableWrapper<RR, UI> getTableWrapper()
	{
		return( _tableWrapper.get() );
	}

	protected JTable getTable()
	{
		return( getTableWrapper().getTable() );
	}

	public int[] getViewArray( ViewAndModelSelectionArray arrayObj )
	{
		return( (arrayObj == null) ? null : arrayObj.getViewSelection() );
	}

	public void setViewArray( ViewAndModelSelectionArray arrayObj, int[] array )
	{
		if( arrayObj != null )
			arrayObj.setViewSelection(array);
	}

	public void setModelArray( ViewAndModelSelectionArray arrayObj, int[] array )
	{
		if( arrayObj != null )
			arrayObj.setModelSelection(array);
	}

	protected JTableWrapperDragAndDropFilesDecorator<UI> createDragAndDropFilesDecorator()
	{
		return( new JTableWrapperDragAndDropFilesDecorator<>(getTableWrapper()) );
	}

	public JTableWrapperDragAndDropFilesDecorator<UI> getDragAndDropFilesDecorator() {
		return _dadFilesDecorator;
	}

	protected FileTransferHandler<JTableFileDragAndDropEventContext> createFileTransferHandler()
	{
		return( getDragAndDropFilesDecorator().createFileTransferHandler() );
	}

	@Override
	public void releaseResources()
	{
		getDragAndDropFilesDecorator().releaseResources();
	}

	protected Predicate<Point> getCanStartDraggingFunction() {
		return( this::canStartDragging );
	}

	public int[] getPreviousViewRowSelection() {
		return( getTableWrapper().getPreviousViewRowSelection() );
	}

	protected boolean isFiltered()
	{
		return( getTableWrapper().isFiltered() );
	}

	public boolean canStartDragging( Point point )
	{
		boolean result = !isFiltered() && previousWasRowSelected(point); //&& !isPreviousWasSelectionAdjusting();
		if( result )
		{
			setViewArray( _startDraggingRowSelection, getPreviousViewRowSelection() );

			createProxyForListSelectionModelAndSetItToTable();
		}

		return( result );
	}

	protected boolean previousWasRowSelected( Point point )
	{
		return( getTableWrapper().isRowSelected(getPreviousViewRowSelection(), point ) );
	}

	protected void updateUI( UI result )
	{
		result.setCanStartDraggingFunction( getCanStartDraggingFunction() );
		result.setDropListener( this::notifyDrop );
		result.setDeleteListener( this::notifyRecordsDeleted );
		result.setStartDraggingListener( this::startDragging );
		result.setStopDraggingListener( this::stopDragging );
	}

	public void notifyEvt( GenTableWrapperEventBase evt )
	{
		getTableWrapper().notifyEvt(evt);
	}

	protected void notifyDrop()
	{
		notifyEvt( createRecordsMovedEvt() );
	}

	protected boolean isEmpty( int[] array )
	{
		return( ( array == null ) || ( array.length == 0 ) ||
				( array.length == 1 ) && ( array[0] == -1 ) );
	}

	public void notifyRecordsDeleted()
	{
		if( ! isEmpty( getTableWrapper().getTable().getSelectedRows() ) )
			notifyEvt( createRecordsDeletedEvt() );
	}

	public void notifyRecordsSorted( Comparator<RR> comparator )
	{
		notifyEvt( createRecordsSortedEvt(comparator) );
	}

	protected void createProxyForListSelectionModelAndSetItToTable()
	{
		int[] startDraggingRowSelection = getViewArray( _startDraggingRowSelection );

		getTableWrapper().createProxyForListSelectionModelAndSetItToTable( startDraggingRowSelection );
	}

	protected void restablishListSelectionModel()
	{
		getTableWrapper().restablishListSelectionModel();
	}

	protected void startDragging()
	{
//		createProxyForListSelectionModelAndSetItToTable();
	}

	protected void stopDragging()
	{
		SwingUtilities.invokeLater( this::restablishListSelectionModel );
	}

	protected ListOperationFunctions getListOperFunctions()
	{
		return( ListOperationFunctions.instance() );
	}

	public <RR> ListRecordMoveContext<RR> move( ListRecordMoveContext<RR> input )
	{
		return( getListOperFunctions().move(input) );
	}

	public <RR> ListStateContextBase<RR> delete( ListStateContextBase<RR> input )
	{
		return( getListOperFunctions().delete(input) );
	}

	public <RR> ListStateContextBase<RR> sort( ListStateContextBase<RR> input,
												Comparator<RR> comparator )
	{
		return( getListOperFunctions().sort(input, comparator) );
	}

	protected GenTableWrapperRecordsMovedEvent<RR, JTable> createRecordsMovedEvt()
	{
		Integer currentIndex = getTableWrapper().getCurrentModelRowIndex();
		List<RR> originalList = getTableWrapper().getList();
		int[] currentSelection = getTableWrapper().getAppropriateRowModelSelection();
		int indexWhereToMove = getDragAndDropFilesDecorator().getIndexWhereToDropFiles();

		ListRecordMoveContext<RR> previous = getEventBuilder().createListRecordMoveContext(
			currentIndex, originalList, currentSelection, indexWhereToMove );
		ListRecordMoveContext<RR> after = move( previous );

		GenTableWrapperRecordsMovedEvent<RR, JTable> result =
			getEventBuilder().createJTableWrapperRecordsMovedEvent(	previous, after );

		return( result );
	}

	protected GenTableWrapperRecordsDeletedEvent<RR, JTable> createRecordsDeletedEvt()
	{
		return( createRecordsModifiedGenEvt( GenTableWrapperRecordsDeletedEvent.class, this::delete ) );
	}

	protected GenTableWrapperRecordsSortedEvent<RR, JTable> createRecordsSortedEvt(Comparator<RR> comparator)
	{
		return( createRecordsModifiedGenEvt( GenTableWrapperRecordsSortedEvent.class,
											input -> sort( input, comparator ) ) );
	}

	public <CC extends GenTableWrapperRecordsModifiedEventBase<RR, JTable, ListStateContextBase<RR>>>
			CC createRecordsModifiedGenEvt( Class<CC> clazz,
							Function<ListStateContextBase<RR>, ListStateContextBase<RR>> modificationFunction )
	{
		Integer currentIndex = getTableWrapper().getCurrentModelRowIndex();
		List<RR> originalList = getTableWrapper().getList();
		int[] currentSelection = getTableWrapper().getAppropriateRowModelSelection();

		ListStateContextBase<RR> previous = getEventBuilder().createListStateContextBase(
			currentIndex, originalList, currentSelection );
		ListStateContextBase<RR> after = ( modificationFunction == null ) ?
											null :
											modificationFunction.apply( previous );

		CC result =	getEventBuilder().createJTableWrapperRecordsModifiedGenEvent(
															clazz, previous, after );

		return( result );
	}

	public GenTableWrapperEventBuilderBase<RR, JTable> getEventBuilder() {
		return getTableWrapper().getEventBuilder();
	}
}
