/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.general.desktop.view.componentwrapper.jtable.files;

import com.frojasg1.general.desktop.view.componentwrapper.jtable.JTableFileDragAndDropEventContext;
import com.frojasg1.general.desktop.view.componentwrapper.jtable.JTableWrapper;
import com.frojasg1.general.view.evt.wrapper.impl.table.GenTableWrapperEventBase;
import com.frojasg1.general.view.evt.wrapper.impl.table.impl.GenTableWrapperDropFilesEvent;
import com.frojasg1.general.desktop.view.transfer.DragAndDropEventType;
import com.frojasg1.general.desktop.view.transfer.file.FileDragAndDropEvent;
import com.frojasg1.general.desktop.view.transfer.file.FileTransferHandler;
import com.frojasg1.general.listeners.filterevents.EventListenerContextBase;
import com.frojasg1.general.view.ReleaseResourcesable;
import java.lang.ref.WeakReference;
import java.util.List;

/**
 *
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 */
public class JTableWrapperDragAndDropFilesDecorator<UI extends BaseFileListJTableUI>
	implements ReleaseResourcesable
{
	protected WeakReference<JTableWrapper<?, UI>> _tableWrapper;

	protected EventListenerContextBase<FileDragAndDropEvent<JTableFileDragAndDropEventContext>, DragAndDropEventType> _fileTransferListener;
	protected EventListenerContextBase<FileDragAndDropEvent<JTableFileDragAndDropEventContext>, DragAndDropEventType> _uiListener;

	public JTableWrapperDragAndDropFilesDecorator( JTableWrapper tableWrapper )
	{
		_tableWrapper = new WeakReference<>(tableWrapper);
	}

	protected UI getCompUI()
	{
		return( getTableWrapper().getCompUI() );
	}

	protected <EVT extends GenTableWrapperEventBase> EVT createEvt(Class<EVT> clazz )
	{
		return( getTableWrapper().createEvt(clazz) );
	}

	public FileTransferHandler<JTableFileDragAndDropEventContext> createFileTransferHandler()
	{
		removeListeners();

		FileTransferHandler<JTableFileDragAndDropEventContext> result = new JTableFileTransferHandler<>();
		_uiListener = result.addListener( getCompUI()::processDragAndDropEvent );
		_fileTransferListener = result.addListenerContext( this::filesDropped, DragAndDropEventType.DROPPED );

		return( result );
	}

	protected JTableWrapper<?, UI> getTableWrapper()
	{
		return( _tableWrapper.get() );
	}

	protected FileTransferHandler<JTableFileDragAndDropEventContext> getFileTransferHandler()
	{
		return( getTableWrapper().getFileTransferHandler() );
	}

	protected void removeListeners()
	{
		FileTransferHandler<JTableFileDragAndDropEventContext> fth = getFileTransferHandler();
		if( fth != null )
		{
			fth.removeListener(_uiListener);
			fth.removeListener(_fileTransferListener);
		}
	}

	protected void notifyEvt(GenTableWrapperEventBase evt)
	{
		getTableWrapper().notifyEvt(evt);
	}

	protected void filesDropped( FileDragAndDropEvent evt )
	{
		notifyEvt(createFilesDropped( evt ));
	}

	protected GenTableWrapperDropFilesEvent createFilesDropped(FileDragAndDropEvent evt )
	{
		GenTableWrapperDropFilesEvent result = null;
		if( evt != null )
			result = getTableWrapper().getEventBuilder().createFilesDropped(
				evt.getFileList(), getIndexWhereToDropFiles());

		return( result );
	}

	protected int getNumOfRecords()
	{
		int result = 0;
		List list = getTableWrapper().getList();
		if( list != null )
			result = list.size();

		return( result );
	}

	public int getIndexWhereToDropFiles()
	{
		return( Math.min( getNumOfRecords(), getCompUI().getRowIndexWhereToDrop() ) );
	}

	@Override
	public void releaseResources()
	{
		removeListeners();
		_fileTransferListener = null;
		_uiListener = null;
	}

	protected class JTableFileTransferHandler<CC> extends FileTransferHandler<CC>
	{
		@Override
		protected boolean canImportInternal(TransferSupport info) {
			return( !getTableWrapper().isFiltered() && super.canImportInternal(info) );
		}
	}
}
