/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.general.desktop.view.panels;

import com.frojasg1.applications.common.configuration.application.BaseApplicationConfigurationInterface;
import com.frojasg1.general.NullFunctions;
import com.frojasg1.general.desktop.generic.view.DesktopViewComponent;
import com.frojasg1.general.desktop.generic.view.imp.DesktopViewFacilitiesImp;
import com.frojasg1.general.desktop.view.FrameworkComponentFunctions;
import com.frojasg1.general.desktop.view.color.ColorInversor;

import java.awt.*;
import javax.swing.*;

import com.frojasg1.general.desktop.view.color.ColorThemeInvertible;
import com.frojasg1.general.desktop.view.zoom.mapper.ComponentMapper;
import com.frojasg1.general.desktop.view.zoom.mapper.InternallyMappedComponent;
import com.frojasg1.general.desktop.view.zoom.mapper.ContainerOfInternallyMappedComponent;
import com.frojasg1.general.desktop.view.zoom.mapper.ContainerOfInternallyMappedComponentBase;
import com.frojasg1.general.view.ReleaseResourcesable;
import java.util.List;
import java.util.function.Function;

/**
 *
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 */
public abstract class CustomJPanelBase extends JPanel
									implements ColorThemeInvertible,
												ReleaseResourcesable,
												InternallyMappedComponent,
												DesktopViewComponent,
												ContainerOfInternallyMappedComponent
{
	protected boolean _doNotInvertColors = false;
	protected boolean _alreadyMapped = false;

	protected ContainerOfInternallyMappedComponentBase _containerOfinternallyMappedComponent =
		new ContainerOfInternallyMappedComponentBase();

	protected BaseApplicationConfigurationInterface _appliConf;

	public CustomJPanelBase()
	{
		this(false);
	}

	public CustomJPanelBase(boolean doNotInvertColors )
	{
		this(null, doNotInvertColors);
	}

	public CustomJPanelBase( BaseApplicationConfigurationInterface appliConf, boolean doNotInvertColors )
	{
		super();

		_appliConf = appliConf;
		_doNotInvertColors = doNotInvertColors;
	}

	protected void init()
	{
	}

	protected BaseApplicationConfigurationInterface getAppliConf()
	{
		return( _appliConf );
	}

	protected void invertColorsChild(ColorInversor colorInversor)
	{
		// If children want to change the behaviour, they have to override
	}

	protected void invertSinglePanelColors(ColorInversor colorInversor)
	{
		colorInversor.invertSingleColorsGen(this);
	}

	@Override
	public void invertColors( ColorInversor colorInversor)
	{
		if( ! _doNotInvertColors )
		{
			invertSinglePanelColors(colorInversor);
			invertColorsChild(colorInversor);
		}
	}

	@Override
	public void setBackground( Color bc )
	{
		super.setBackground(bc);
	}

	protected ColorInversor getColorInversor()
	{
		return( FrameworkComponentFunctions.instance().getColorInversor(this) );
	}

	protected boolean isDarkMode()
	{
		return( FrameworkComponentFunctions.instance().isDarkMode(this) );
	}

	protected boolean wasLatestModeDark()
	{
		return( FrameworkComponentFunctions.instance().wasLatestModeDark(this) );
	}

	@Override
	public void setComponentMapper(ComponentMapper compMapper)
	{
		for( InternallyMappedComponent imc: getInternallyMappedComponentListCopy() )
			imc.setComponentMapper(compMapper);

		_alreadyMapped = true;
	}

	@Override
	public boolean hasBeenAlreadyMapped()
	{
		return( _alreadyMapped );
	}

	@Override
	public void releaseResources()
	{
		// intentionally left blank
	}

	@Override
	public CustomJPanelBase getComponent()
	{
		return( this );
	}

	@Override
	public void setVisible( boolean value )
	{
		super.setVisible(value);
	}

	@Override
	public void requestFocus()
	{
		super.requestFocus();
	}

	@Override
	public boolean isFocusable()
	{
		return( super.isFocusable() );
	}

	@Override
	public boolean hasFocus()
	{
		return( super.hasFocus() );
	}

	@Override
	public DesktopViewComponent getParentViewComponent()
	{
		return( DesktopViewFacilitiesImp.instance().getParentViewComponent( this ) );
	}

	protected FrameworkComponentFunctions getFrameworkComponentFunctions() {
		return FrameworkComponentFunctions.instance();
	}

	public <CC> CC getOwnOrFirstParentOfClass( Component comp, Class<CC> clazz ) {
		return getFrameworkComponentFunctions().getOwnOrFirstParentOfClass(comp, clazz);
	}

	protected JDialog getJDialog() {
		return getOwnOrFirstParentOfClass(this, JDialog.class);
	}

	protected JFrame getJFrame() {
		return getOwnOrFirstParentOfClass(this, JFrame.class);
	}

	public Component getWindow() {
		Component result = getJFrame();
		if(result == null) {
			result = getJDialog();
		}
		return result;
	}

	protected List<InternallyMappedComponent> getInternallyMappedComponentListCopy() {
		return( _containerOfinternallyMappedComponent.getInternallyMappedComponentListCopy() );
	}

	@Override
	public void addInternallyMappedComponent( InternallyMappedComponent im )
	{
		_containerOfinternallyMappedComponent.addInternallyMappedComponent(im);
	}

	@Override
	public void addPopupMenu( JPopupMenu jPopupMenu )
	{
		_containerOfinternallyMappedComponent.addPopupMenu(jPopupMenu);
	}

	protected List<JPopupMenu> getJPopupMenuListCopy()
	{
		return( _containerOfinternallyMappedComponent.getJPopupMenuListCopy() );
	}

	protected NullFunctions getNullFunctions() {
		return NullFunctions.instance();
	}

	protected <CC, RR> RR getIfNotNull(CC obj, Function<CC, RR> getter) {
		return getNullFunctions().getIfNotNull(obj, getter);
	}

	protected <CC, RR> RR getIfNotNull(CC obj, Function<CC, RR> getter, RR defaultValue) {
		return getNullFunctions().getIfNotNull(obj, getter, defaultValue);
	}
}
