/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.general.desktop.view.search;

import java.awt.*;
import javax.swing.*;
import javax.swing.border.*;
import com.frojasg1.applications.common.configuration.application.BaseApplicationConfigurationInterface;
import com.frojasg1.general.desktop.search.SearchWindowConfiguration;
import com.frojasg1.general.search.SearchReplaceContextInterface;
import com.frojasg1.general.search.SearchReplaceForWindowInterface;
import com.frojasg1.general.undoredo.text.TextUndoRedoInterface;
import com.frojasg1.applications.common.components.internationalization.window.InternationalizedJFrame;
import com.frojasg1.applications.common.components.resizecomp.MapResizeRelocateComponentItem;
import com.frojasg1.applications.common.components.resizecomp.ResizeRelocateItem;
import com.frojasg1.general.DateFunctions;
import com.frojasg1.general.ExecutionFunctions;
import com.frojasg1.general.desktop.GenericDesktopConstants;
import com.frojasg1.general.persistence.PersistentConfiguration;
import com.frojasg1.general.desktop.view.combobox.chained.ComboBoxGroupManager;
import com.frojasg1.general.desktop.view.combobox.chained.impl.ChainedParentChildComboBoxManagerAutomaticSave;
import com.frojasg1.general.desktop.view.zoom.mapper.ComponentMapper;
import com.frojasg1.general.search.RegExException;
import com.frojasg1.general.search.SearchReplaceTextInterface;
import com.frojasg1.general.search.SearchReplaceTextInterface.ReplaceResultInterface;
import com.frojasg1.general.search.SearchReplaceTextInterface.ReplaceSettingsInterface;
import com.frojasg1.general.search.SearchReplaceTextInterface.SearchResultInterface;
import com.frojasg1.general.search.SearchReplaceTextInterface.SearchSettingsInterface;
import com.frojasg1.general.search.imp.ReplaceSettings;
import com.frojasg1.general.search.imp.SearchSettings;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Insets;
import java.awt.Rectangle;
import java.util.Date;
import java.util.Vector;
import javax.swing.JComboBox;
import javax.swing.JPopupMenu;
import javax.swing.SwingUtilities;

/**
 *
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 */
public class DesktopSearchAndReplaceWindow extends InternationalizedJFrame
	implements SearchReplaceForWindowInterface
{
	protected static DesktopSearchAndReplaceWindow _instance = null;
	
	public static final String sa_configurationBaseFileName = "searchAndReplaceJFrame";
	protected static final String sa_configurationBaseFileNameForSearchCombo = "searchComboHistory";
	protected static final String sa_configurationBaseFileNameForReplaceCombo = "replaceComboHistory";
//	protected DesktopViewTextComponent _textComp = null;

	protected static final String sa_PROPERTIES_PATH_IN_JAR = GenericDesktopConstants.sa_PROPERTIES_PATH_IN_JAR;

//	protected SearchReplaceTextInterface _searchReplaceManager = null;
//	protected TextUndoRedoInterface _undoRedoManager = null;

	protected String _configurationBaseFileName = null;

//	protected boolean _newSearch = true;

	protected SearchWindowConfiguration _lastConfiguration = null;

//	protected TextComboBoxHistoryWithProperties _comboBoxHistoryForSearchCB = null;
//	protected TextComboBoxHistoryWithProperties _comboBoxHistoryForReplaceCB = null;

	protected ComboBoxGroupManager _comboGroupManagerForSearchCB = null;
	protected ComboBoxGroupManager _comboGroupManagerForReplaceCB = null;

	protected static final String CONF_SEARCHED_STRING_NOT_FOUND = "SEARCHED_STRING_NOT_FOUND";
	protected static final String CONF_SEARCHED_STRING_FOUND = "SEARCHED_STRING_FOUND";
	protected static final String CONF_STRING_REPLACED = "STRING_REPLACED";
	protected static final String CONF_COULD_NOT_REPLACE_STRING = "COULD_NOT_REPLACE_STRING";
	protected static final String CONF_NO_REPLACEMENTS_DONE = "NO_REPLACEMENTS_DONE";
	protected static final String CONF_NUMBER_OF_REPLACEMENTS_DONE = "NUMBER_OF_REPLACEMENTS_DONE";
	protected static final String CONF_ERROR_IN_REGEX = "ERROR_IN_REGEX";

	protected SearchReplaceContextInterface _src = null;

	public static DesktopSearchAndReplaceWindow createInstance( BaseApplicationConfigurationInterface appConf,
									String configurationBaseFileName )
	{
		new DesktopSearchAndReplaceWindow( appConf, configurationBaseFileName );
		return( _instance );
	}

	public static DesktopSearchAndReplaceWindow instance()
	{
		return( _instance );
	}

	/**
	 * Creates new form SearchAndReplaceWindow
	 */
	public DesktopSearchAndReplaceWindow( BaseApplicationConfigurationInterface appConf,
									String configurationBaseFileName )
	{
		super( appConf );

		if( configurationBaseFileName == null )
			_configurationBaseFileName = sa_configurationBaseFileName;
		else
			_configurationBaseFileName = configurationBaseFileName;

		openConfiguration();

		initComponents();

		setWindowConfiguration();

		_instance = this;
//		_searchReplaceManager = srt;
	}

	@Override
	public void internationalizationInitializationEndCallback()
	{
		super.internationalizationInitializationEndCallback();

//		initView( );
	}

	public TextUndoRedoInterface getUndoRedoManager()
	{
		return( _src.getUndoRedoManager() );
	}

	// https://stackoverflow.com/questions/9956533/select-all-text-in-editable-jcombobox-and-set-cursor-position
	public void openForSearch( String searchText )
	{
		boolean replace = false;
		openSearchReplace( searchText, replace );
	}

	public void openForReplace( String searchText )
	{
		boolean replace = true;
		openSearchReplace( searchText, replace );
	}

	protected void openSearchReplace( String searchText, boolean replace )
	{
		SwingUtilities.invokeLater( new Runnable() {
				@Override
				public void run()
				{
					if( ( searchText != null ) &&
						( searchText.length() > 0 ) )
					{
						jCB_search.getEditor().setItem( searchText );
					}

					jCB_search.getEditor().selectAll();

					jCB_replace.setSelected( replace );
					resizeWindowForReplace( replace );
					setVisible(true);

					deiconify();
					jCB_search.requestFocus();
				}
		});
	}

	protected ComboBoxGroupManager createComboBoxGroupManager( Integer maxItems, JComboBox combo, String confBaseFileName )
	{
		String language = null;
//		ParameterListConfiguration conf = new ParameterListConfiguration( getAppliConf().getConfigurationMainFolder(),
//															getAppliConf().getApplicationNameFolder(),
//															getAppliConf().getApplicationGroup(), language,
//															confBaseFileName + ".properties" );
//		ExecutionFunctions.instance().safeMethodExecution( () -> conf.M_openConfiguration() );
//		TextComboBoxHistoryWithProperties content = new TextComboBoxHistoryWithProperties( maxItems, conf );
//		content.init( (List<String>)null);
//		content.loadItems();

//		ChainedParentChildComboBoxManagerBase result = new ChainedParentChildComboBoxManagerBase( null, content, null ) {
//				public void newItemSelected( String newItem )
//				{
//					super.newItemSelected( newItem );
//					saveComboboxContents( this );
//				}
//		};
		ChainedParentChildComboBoxManagerAutomaticSave result =
			new ChainedParentChildComboBoxManagerAutomaticSave(null, getAppliConf(), confBaseFileName, maxItems, null );
		result.init();
		result.addComboComp( combo );

		return( result );
	}

	protected void initCombos( )
	{
		int maxItems = _lastConfiguration.M_getIntParamConfiguration( SearchWindowConfiguration.CONF_MAX_ITEMS_IN_HISTORY );
		_comboGroupManagerForSearchCB = createComboBoxGroupManager( maxItems, jCB_search, sa_configurationBaseFileNameForSearchCombo );
		_comboGroupManagerForReplaceCB = createComboBoxGroupManager( maxItems, jCB_replaceFor, sa_configurationBaseFileNameForReplaceCombo );
	}

	protected void initView( )
	{
		boolean regEx = _lastConfiguration.M_getBoolParamConfiguration( SearchWindowConfiguration.CONF_IS_REGEX );
		if( regEx )
			jRB_RegularExpression.setSelected( true );
		else
			jRB_NormalText.setSelected( true );
		radioButtonsRegExChanged();

		boolean wholeWords = _lastConfiguration.M_getBoolParamConfiguration( SearchWindowConfiguration.CONF_WHOLE_WORDS );
		jCB_wholeWords.setSelected( wholeWords );

		boolean matchCase = _lastConfiguration.M_getBoolParamConfiguration( SearchWindowConfiguration.CONF_MATCH_CASE );
		jCB_MatchCase.setSelected( matchCase );

		boolean replace = _lastConfiguration.M_getBoolParamConfiguration( SearchWindowConfiguration.CONF_REPLACE );
		jCB_replace.setSelected( replace );
		resizeWindowForReplace( replace );

		boolean alwaysOnTop = _lastConfiguration.M_getBoolParamConfiguration( SearchWindowConfiguration.CONF_ALWAYS_ON_TOP );
		jCB_alwaysOnTop.setSelected( alwaysOnTop );
		setAlwaysOnTop( alwaysOnTop );

		initCombos();

		setStatusText( "" );
	}

	protected void openConfiguration()
	{
		_lastConfiguration = new SearchWindowConfiguration( getAppliConf().getConfigurationMainFolder(),
															getAppliConf().getApplicationNameFolder(),
															getAppliConf().getApplicationGroup(),
															_configurationBaseFileName + ".properties" );

		try
		{
			_lastConfiguration.M_openConfiguration();
		}
		catch( Throwable th )
		{
			th.printStackTrace();
		}
	}
	
	protected SearchSettingsInterface getSearchSettings()
	{
		SearchSettings result = new SearchSettings( jRB_RegularExpression.isSelected(),
													jCB_wholeWords.isSelected(),
													jCB_MatchCase.isSelected(),
													(String) jCB_search.getEditor().getItem(),
													(String) jCB_search.getEditor().getItem() );

		return( result );
	}

	protected ReplaceSettingsInterface getReplaceSettings()
	{
		ReplaceSettings result = new ReplaceSettings( jRB_RegularExpression.isSelected(),
													jCB_wholeWords.isSelected(),
													jCB_MatchCase.isSelected(),
													(String) jCB_search.getEditor().getItem(),
													(String) jCB_search.getEditor().getItem(),
													(String) jCB_replaceFor.getEditor().getItem(),
													(String) jCB_replaceFor.getEditor().getItem() );

		return( result );
	}
	
/*
	public void setUndoRedoManager( TextUndoRedoInterface undoRedoManagerOfTextComp )
	{
		if( _undoRedoManager != null )
		{
			_undoRedoManager.unregisterListener(this);
		}

		_undoRedoManager = undoRedoManagerOfTextComp;
		if( _undoRedoManager != null )
		{
			setTextComponent( _undoRedoManager.getView() );

			_undoRedoManager.registerListener(this);
		}
		else
			_textComp = null;

		_searchReplaceManager.setUndoRedoManager(undoRedoManagerOfTextComp);
	}

	public void setTextComponent( ViewTextComponent vtc )
	{
		if( vtc instanceof DesktopViewTextComponent )
			_textComp = (DesktopViewTextComponent) vtc;
		else
			_textComp = null;
	}
*/
	protected void setWindowConfiguration( )
	{
		Vector<JPopupMenu> vectorJpopupMenus = null;

		MapResizeRelocateComponentItem mapRRCI = new MapResizeRelocateComponentItem( );
		try
		{
			mapRRCI.putResizeRelocateComponentItem(jPanel2, ResizeRelocateItem.RESIZE_TO_RIGHT +
															ResizeRelocateItem.RESIZE_TO_BOTTOM );
			mapRRCI.putResizeRelocateComponentItem(jPanel1, ResizeRelocateItem.RESIZE_TO_RIGHT +
															ResizeRelocateItem.RESIZE_TO_BOTTOM );
			mapRRCI.putResizeRelocateComponentItem(jP_textToFind, ResizeRelocateItem.RESIZE_TO_RIGHT );
			mapRRCI.putResizeRelocateComponentItem(jCB_search, ResizeRelocateItem.RESIZE_TO_RIGHT );
			mapRRCI.putResizeRelocateComponentItem(jP_replaceTextFor, ResizeRelocateItem.RESIZE_TO_RIGHT );
			mapRRCI.putResizeRelocateComponentItem(jP_status, ResizeRelocateItem.RESIZE_TO_RIGHT );
//			mapRRCI.putResizeRelocateComponentItem(jL_status, ResizeRelocateItem.RESIZE_TO_RIGHT );
			mapRRCI.putResizeRelocateComponentItem(jCB_replaceFor, ResizeRelocateItem.RESIZE_TO_RIGHT );
		}
		catch( Throwable th )
		{
			th.printStackTrace();
		}
/*
		a_intern = new JFrameInternationalization(	ApplicationConfiguration.sa_MAIN_FOLDER,
													ApplicationConfiguration.sa_APPLICATION_NAME,
													ApplicationConfiguration.sa_CONFIGURATION_GROUP,
													ApplicationConfiguration.sa_PATH_PROPERTIES_IN_JAR,
													a_configurationBaseFileName,
													this,
													parent,
													a_vectorJpopupMenus,
													false,
													mapRRCI );
*/
		boolean adjustSizeOfFrameToContents = true;
		boolean adjustMinSizeOfFrameToContents = true;
		boolean switchToZoomComponents = true;
		boolean internationalizeFont = true;
		Integer delayToInvokeCallback = null;
		createInternationalization(	getAppliConf().getConfigurationMainFolder(),
									getAppliConf().getApplicationNameFolder(),
									getAppliConf().getApplicationGroup(),
//									getAppliConf().getInternationalPropertiesPathInJar(),
									sa_PROPERTIES_PATH_IN_JAR,
									_configurationBaseFileName,
									this,
									null,
									vectorJpopupMenus,
									true,
									mapRRCI,
									adjustSizeOfFrameToContents,
									adjustMinSizeOfFrameToContents,
									getAppliConf().getZoomFactor(),
									getAppliConf().hasToEnableUndoRedoForTextComponents(),
									getAppliConf().hasToEnableTextCompPopupMenus(),
									switchToZoomComponents,
									internationalizeFont,
									delayToInvokeCallback);

		a_intern.setMaxWindowHeightNoLimit( false );

		registerInternationalString(CONF_SEARCHED_STRING_NOT_FOUND, "Searched string not found." );
		registerInternationalString(CONF_SEARCHED_STRING_FOUND, "Searched string found." );
		registerInternationalString(CONF_STRING_REPLACED, "String replaced." );
		registerInternationalString(CONF_COULD_NOT_REPLACE_STRING, "Could not replace string." );
		registerInternationalString(CONF_NO_REPLACEMENTS_DONE, "No replacemets have been done." );
		registerInternationalString(CONF_NUMBER_OF_REPLACEMENTS_DONE, "Number of replacements done." );
		registerInternationalString(CONF_ERROR_IN_REGEX, "Error in regular expression." );

//		setIcon( "com/frojasg1/chesspdfbrowser/resources/icons/App.icon.png" );
	}

	/**
	 * This method is called from within the constructor to initialize the form.
	 * WARNING: Do NOT modify this code. The content of this method is always
	 * regenerated by the Form Editor.
	 */
	@SuppressWarnings("unchecked")
    // <editor-fold defaultstate="collapsed" desc="Generated Code">//GEN-BEGIN:initComponents
    // Generated using JFormDesigner non-commercial license
    private void initComponents() {
        jPanel2 = new JPanel();
        jP_status = new JPanel();
        jL_status = new JLabel();
        jL_hour = new JLabel();
        jPanel1 = new JPanel();
        jP_RegExRadioGroupPanel = new JPanel();
        jRB_NormalText = new JRadioButton();
        jRB_RegularExpression = new JRadioButton();
        jP_textToFind = new JPanel();
        jCB_search = new JComboBox<>();
        jCB_alwaysOnTop = new JCheckBox();
        jP_buttons = new JPanel();
        jB_forward = new JButton();
        jB_replaceForward = new JButton();
        jB_replaceBackwards = new JButton();
        jB_backwards = new JButton();
        jB_replaceAll = new JButton();
        jB_replace = new JButton();
        jP_replaceTextFor = new JPanel();
        jCB_replaceFor = new JComboBox<>();
        jP_AdditionalSearchSettingsPanel = new JPanel();
        jCB_MatchCase = new JCheckBox();
        jCB_wholeWords = new JCheckBox();
        jCB_replace = new JCheckBox();
        bg_RegEx = new ButtonGroup();

        //======== this ========
        setDefaultCloseOperation(WindowConstants.DO_NOTHING_ON_CLOSE);
        setMinimumSize(new Dimension(606, 250));
        setName("SearchAndReplaceWindow");
        Container contentPane = getContentPane();
        contentPane.setLayout(null);

        //======== jPanel2 ========
        {
            jPanel2.setName("jPanel2");
            jPanel2.setLayout(null);

            //======== jP_status ========
            {
                jP_status.setBorder(LineBorder.createBlackLineBorder());
                jP_status.setName("jP_status");
                jP_status.setLayout(null);

                //---- jL_status ----
                jL_status.setFont(new Font("Tahoma", Font.BOLD, 11));
                jL_status.setForeground(new Color(1, 197, 1));
                jL_status.setText("jLabel1");
                jP_status.add(jL_status);
                jL_status.setBounds(70, 3, 510, jL_status.getPreferredSize().height);

                //---- jL_hour ----
                jL_hour.setFont(new Font("Tahoma", Font.BOLD, 11));
                jL_hour.setText("jLabel1");
                jP_status.add(jL_hour);
                jL_hour.setBounds(10, 3, 60, jL_hour.getPreferredSize().height);

                {
                    // compute preferred size
                    Dimension preferredSize = new Dimension();
                    for(int i = 0; i < jP_status.getComponentCount(); i++) {
                        Rectangle bounds = jP_status.getComponent(i).getBounds();
                        preferredSize.width = Math.max(bounds.x + bounds.width, preferredSize.width);
                        preferredSize.height = Math.max(bounds.y + bounds.height, preferredSize.height);
                    }
                    Insets insets = jP_status.getInsets();
                    preferredSize.width += insets.right;
                    preferredSize.height += insets.bottom;
                    jP_status.setMinimumSize(preferredSize);
                    jP_status.setPreferredSize(preferredSize);
                }
            }
            jPanel2.add(jP_status);
            jP_status.setBounds(0, 260, 590, 20);

            //======== jPanel1 ========
            {
                jPanel1.setPreferredSize(new Dimension(500, 260));
                jPanel1.setLayout(null);

                //======== jP_RegExRadioGroupPanel ========
                {
                    jP_RegExRadioGroupPanel.setBorder(new TitledBorder("Regular expression"));
                    jP_RegExRadioGroupPanel.setName("jP_RegExRadioGroupPanel");
                    jP_RegExRadioGroupPanel.setLayout(null);

                    //---- jRB_NormalText ----
                    jRB_NormalText.setSelected(true);
                    jRB_NormalText.setText("Normal text");
                    jRB_NormalText.setName("RB_NormalText");
                    jRB_NormalText.addActionListener(e -> jRB_NormalTextActionPerformed(e));
                    jP_RegExRadioGroupPanel.add(jRB_NormalText);
                    jRB_NormalText.setBounds(10, 20, 270, jRB_NormalText.getPreferredSize().height);

                    //---- jRB_RegularExpression ----
                    jRB_RegularExpression.setText("Regular expression");
                    jRB_RegularExpression.setName("RB_RegularExpression");
                    jRB_RegularExpression.addActionListener(e -> jRB_RegularExpressionActionPerformed(e));
                    jP_RegExRadioGroupPanel.add(jRB_RegularExpression);
                    jRB_RegularExpression.setBounds(10, 40, 270, jRB_RegularExpression.getPreferredSize().height);

                    {
                        // compute preferred size
                        Dimension preferredSize = new Dimension();
                        for(int i = 0; i < jP_RegExRadioGroupPanel.getComponentCount(); i++) {
                            Rectangle bounds = jP_RegExRadioGroupPanel.getComponent(i).getBounds();
                            preferredSize.width = Math.max(bounds.x + bounds.width, preferredSize.width);
                            preferredSize.height = Math.max(bounds.y + bounds.height, preferredSize.height);
                        }
                        Insets insets = jP_RegExRadioGroupPanel.getInsets();
                        preferredSize.width += insets.right;
                        preferredSize.height += insets.bottom;
                        jP_RegExRadioGroupPanel.setMinimumSize(preferredSize);
                        jP_RegExRadioGroupPanel.setPreferredSize(preferredSize);
                    }
                }
                jPanel1.add(jP_RegExRadioGroupPanel);
                jP_RegExRadioGroupPanel.setBounds(0, 20, 290, 70);

                //======== jP_textToFind ========
                {
                    jP_textToFind.setBorder(new TitledBorder("Text to find"));
                    jP_textToFind.setName("P_textToFind");
                    jP_textToFind.setLayout(null);

                    //---- jCB_search ----
                    jCB_search.setEditable(true);
                    jCB_search.setModel(new DefaultComboBoxModel<>(new String[] {

                    }));
                    jCB_search.addActionListener(e -> jCB_searchActionPerformed(e));
                    jP_textToFind.add(jCB_search);
                    jCB_search.setBounds(10, 20, 570, jCB_search.getPreferredSize().height);

                    {
                        // compute preferred size
                        Dimension preferredSize = new Dimension();
                        for(int i = 0; i < jP_textToFind.getComponentCount(); i++) {
                            Rectangle bounds = jP_textToFind.getComponent(i).getBounds();
                            preferredSize.width = Math.max(bounds.x + bounds.width, preferredSize.width);
                            preferredSize.height = Math.max(bounds.y + bounds.height, preferredSize.height);
                        }
                        Insets insets = jP_textToFind.getInsets();
                        preferredSize.width += insets.right;
                        preferredSize.height += insets.bottom;
                        jP_textToFind.setMinimumSize(preferredSize);
                        jP_textToFind.setPreferredSize(preferredSize);
                    }
                }
                jPanel1.add(jP_textToFind);
                jP_textToFind.setBounds(0, 160, 590, 50);

                //---- jCB_alwaysOnTop ----
                jCB_alwaysOnTop.setText("Always on top");
                jCB_alwaysOnTop.setName("CB_alwaysOnTop");
                jCB_alwaysOnTop.addActionListener(e -> jCB_alwaysOnTopActionPerformed(e));
                jPanel1.add(jCB_alwaysOnTop);
                jCB_alwaysOnTop.setBounds(10, 0, 260, jCB_alwaysOnTop.getPreferredSize().height);

                //======== jP_buttons ========
                {
                    jP_buttons.setLayout(null);

                    //---- jB_forward ----
                    jB_forward.setText("Search forward");
                    jB_forward.setName("B_forward");
                    jB_forward.addActionListener(e -> jB_forwardActionPerformed(e));
                    jP_buttons.add(jB_forward);
                    jB_forward.setBounds(10, 10, 160, jB_forward.getPreferredSize().height);

                    //---- jB_replaceForward ----
                    jB_replaceForward.setText("Replace and next search forward");
                    jB_replaceForward.setName("B_replaceForward");
                    jB_replaceForward.addActionListener(e -> jB_replaceForwardActionPerformed(e));
                    jP_buttons.add(jB_replaceForward);
                    jB_replaceForward.setBounds(180, 10, 270, jB_replaceForward.getPreferredSize().height);

                    //---- jB_replaceBackwards ----
                    jB_replaceBackwards.setText("Replace and next search backwards");
                    jB_replaceBackwards.setName("B_replaceBackwards");
                    jB_replaceBackwards.addActionListener(e -> jB_replaceBackwardsActionPerformed(e));
                    jP_buttons.add(jB_replaceBackwards);
                    jB_replaceBackwards.setBounds(180, 40, 270, jB_replaceBackwards.getPreferredSize().height);

                    //---- jB_backwards ----
                    jB_backwards.setText("Search backwards");
                    jB_backwards.setName("B_backwards");
                    jB_backwards.addActionListener(e -> jB_backwardsActionPerformed(e));
                    jP_buttons.add(jB_backwards);
                    jB_backwards.setBounds(10, 40, 160, jB_backwards.getPreferredSize().height);

                    //---- jB_replaceAll ----
                    jB_replaceAll.setText("Replace all");
                    jB_replaceAll.setName("B_replaceAll");
                    jB_replaceAll.addActionListener(e -> jB_replaceAllActionPerformed(e));
                    jP_buttons.add(jB_replaceAll);
                    jB_replaceAll.setBounds(460, 40, 120, jB_replaceAll.getPreferredSize().height);

                    //---- jB_replace ----
                    jB_replace.setText("Replace");
                    jB_replace.setName("B_replace");
                    jB_replace.addActionListener(e -> jB_replaceActionPerformed(e));
                    jP_buttons.add(jB_replace);
                    jB_replace.setBounds(460, 10, 120, jB_replace.getPreferredSize().height);

                    {
                        // compute preferred size
                        Dimension preferredSize = new Dimension();
                        for(int i = 0; i < jP_buttons.getComponentCount(); i++) {
                            Rectangle bounds = jP_buttons.getComponent(i).getBounds();
                            preferredSize.width = Math.max(bounds.x + bounds.width, preferredSize.width);
                            preferredSize.height = Math.max(bounds.y + bounds.height, preferredSize.height);
                        }
                        Insets insets = jP_buttons.getInsets();
                        preferredSize.width += insets.right;
                        preferredSize.height += insets.bottom;
                        jP_buttons.setMinimumSize(preferredSize);
                        jP_buttons.setPreferredSize(preferredSize);
                    }
                }
                jPanel1.add(jP_buttons);
                jP_buttons.setBounds(0, 90, 590, 70);

                //======== jP_replaceTextFor ========
                {
                    jP_replaceTextFor.setBorder(new TitledBorder("Text to replace for"));
                    jP_replaceTextFor.setName("P_replaceTextFor");
                    jP_replaceTextFor.setLayout(null);

                    //---- jCB_replaceFor ----
                    jCB_replaceFor.setEditable(true);
                    jCB_replaceFor.setModel(new DefaultComboBoxModel<>(new String[] {

                    }));
                    jP_replaceTextFor.add(jCB_replaceFor);
                    jCB_replaceFor.setBounds(10, 20, 570, jCB_replaceFor.getPreferredSize().height);

                    {
                        // compute preferred size
                        Dimension preferredSize = new Dimension();
                        for(int i = 0; i < jP_replaceTextFor.getComponentCount(); i++) {
                            Rectangle bounds = jP_replaceTextFor.getComponent(i).getBounds();
                            preferredSize.width = Math.max(bounds.x + bounds.width, preferredSize.width);
                            preferredSize.height = Math.max(bounds.y + bounds.height, preferredSize.height);
                        }
                        Insets insets = jP_replaceTextFor.getInsets();
                        preferredSize.width += insets.right;
                        preferredSize.height += insets.bottom;
                        jP_replaceTextFor.setMinimumSize(preferredSize);
                        jP_replaceTextFor.setPreferredSize(preferredSize);
                    }
                }
                jPanel1.add(jP_replaceTextFor);
                jP_replaceTextFor.setBounds(0, 210, 590, 50);

                //======== jP_AdditionalSearchSettingsPanel ========
                {
                    jP_AdditionalSearchSettingsPanel.setBorder(new TitledBorder("Additional searh settings"));
                    jP_AdditionalSearchSettingsPanel.setName("P_AdditionalSearchSettingsPanel");
                    jP_AdditionalSearchSettingsPanel.setLayout(null);

                    //---- jCB_MatchCase ----
                    jCB_MatchCase.setText("Match case");
                    jCB_MatchCase.setName("CB_MatchCase");
                    jP_AdditionalSearchSettingsPanel.add(jCB_MatchCase);
                    jCB_MatchCase.setBounds(10, 20, 280, jCB_MatchCase.getPreferredSize().height);

                    //---- jCB_wholeWords ----
                    jCB_wholeWords.setText("Whole words");
                    jCB_wholeWords.setName("CB_wholeWords");
                    jP_AdditionalSearchSettingsPanel.add(jCB_wholeWords);
                    jCB_wholeWords.setBounds(10, 40, 280, jCB_wholeWords.getPreferredSize().height);

                    //---- jCB_replace ----
                    jCB_replace.setText("Replace");
                    jCB_replace.setName("CB_replace");
                    jCB_replace.addActionListener(e -> jCB_replaceActionPerformed(e));
                    jP_AdditionalSearchSettingsPanel.add(jCB_replace);
                    jCB_replace.setBounds(10, 60, 280, jCB_replace.getPreferredSize().height);

                    {
                        // compute preferred size
                        Dimension preferredSize = new Dimension();
                        for(int i = 0; i < jP_AdditionalSearchSettingsPanel.getComponentCount(); i++) {
                            Rectangle bounds = jP_AdditionalSearchSettingsPanel.getComponent(i).getBounds();
                            preferredSize.width = Math.max(bounds.x + bounds.width, preferredSize.width);
                            preferredSize.height = Math.max(bounds.y + bounds.height, preferredSize.height);
                        }
                        Insets insets = jP_AdditionalSearchSettingsPanel.getInsets();
                        preferredSize.width += insets.right;
                        preferredSize.height += insets.bottom;
                        jP_AdditionalSearchSettingsPanel.setMinimumSize(preferredSize);
                        jP_AdditionalSearchSettingsPanel.setPreferredSize(preferredSize);
                    }
                }
                jPanel1.add(jP_AdditionalSearchSettingsPanel);
                jP_AdditionalSearchSettingsPanel.setBounds(290, 0, 300, 90);

                {
                    // compute preferred size
                    Dimension preferredSize = new Dimension();
                    for(int i = 0; i < jPanel1.getComponentCount(); i++) {
                        Rectangle bounds = jPanel1.getComponent(i).getBounds();
                        preferredSize.width = Math.max(bounds.x + bounds.width, preferredSize.width);
                        preferredSize.height = Math.max(bounds.y + bounds.height, preferredSize.height);
                    }
                    Insets insets = jPanel1.getInsets();
                    preferredSize.width += insets.right;
                    preferredSize.height += insets.bottom;
                    jPanel1.setMinimumSize(preferredSize);
                    jPanel1.setPreferredSize(preferredSize);
                }
            }
            jPanel2.add(jPanel1);
            jPanel1.setBounds(0, 0, 590, 260);

            {
                // compute preferred size
                Dimension preferredSize = new Dimension();
                for(int i = 0; i < jPanel2.getComponentCount(); i++) {
                    Rectangle bounds = jPanel2.getComponent(i).getBounds();
                    preferredSize.width = Math.max(bounds.x + bounds.width, preferredSize.width);
                    preferredSize.height = Math.max(bounds.y + bounds.height, preferredSize.height);
                }
                Insets insets = jPanel2.getInsets();
                preferredSize.width += insets.right;
                preferredSize.height += insets.bottom;
                jPanel2.setMinimumSize(preferredSize);
                jPanel2.setPreferredSize(preferredSize);
            }
        }
        contentPane.add(jPanel2);
        jPanel2.setBounds(0, 0, 590, 280);

        {
            // compute preferred size
            Dimension preferredSize = new Dimension();
            for(int i = 0; i < contentPane.getComponentCount(); i++) {
                Rectangle bounds = contentPane.getComponent(i).getBounds();
                preferredSize.width = Math.max(bounds.x + bounds.width, preferredSize.width);
                preferredSize.height = Math.max(bounds.y + bounds.height, preferredSize.height);
            }
            Insets insets = contentPane.getInsets();
            preferredSize.width += insets.right;
            preferredSize.height += insets.bottom;
            contentPane.setMinimumSize(preferredSize);
            contentPane.setPreferredSize(preferredSize);
        }
        pack();
        setLocationRelativeTo(getOwner());

        //---- bg_RegEx ----
        bg_RegEx.add(jRB_NormalText);
        bg_RegEx.add(jRB_RegularExpression);
    }// </editor-fold>//GEN-END:initComponents

    private void jB_forwardActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_jB_forwardActionPerformed
        // TODO add your handling code here:

		search( true );

    }//GEN-LAST:event_jB_forwardActionPerformed

    private void jB_backwardsActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_jB_backwardsActionPerformed
		// TODO add your handling code here:

		search( false );

    }//GEN-LAST:event_jB_backwardsActionPerformed

    private void jCB_alwaysOnTopActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_jCB_alwaysOnTopActionPerformed
        // TODO add your handling code here:

		this.setAlwaysOnTop( jCB_alwaysOnTop.isSelected() );

    }//GEN-LAST:event_jCB_alwaysOnTopActionPerformed

    private void jCB_replaceActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_jCB_replaceActionPerformed
		// TODO add your handling code here:

		boolean replace = jCB_replace.isSelected();

		resizeWindowForReplace( replace );

    }//GEN-LAST:event_jCB_replaceActionPerformed

    private void jRB_NormalTextActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_jRB_NormalTextActionPerformed
        // TODO add your handling code here:

		radioButtonsRegExChanged();

    }//GEN-LAST:event_jRB_NormalTextActionPerformed

    private void jRB_RegularExpressionActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_jRB_RegularExpressionActionPerformed
        // TODO add your handling code here:

		radioButtonsRegExChanged();

    }//GEN-LAST:event_jRB_RegularExpressionActionPerformed

    private void jB_replaceActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_jB_replaceActionPerformed
        // TODO add your handling code here:

		replace();

    }//GEN-LAST:event_jB_replaceActionPerformed

    private void jB_replaceForwardActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_jB_replaceForwardActionPerformed
        // TODO add your handling code here:

		replaceAndFindNext( true );

    }//GEN-LAST:event_jB_replaceForwardActionPerformed

    private void jB_replaceBackwardsActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_jB_replaceBackwardsActionPerformed
        // TODO add your handling code here:

		replaceAndFindNext( false );

    }//GEN-LAST:event_jB_replaceBackwardsActionPerformed

    private void jB_replaceAllActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_jB_replaceAllActionPerformed
        // TODO add your handling code here:

		replaceAll();

    }//GEN-LAST:event_jB_replaceAllActionPerformed

    private void jCB_searchActionPerformed(java.awt.event.ActionEvent evt) {//GEN-FIRST:event_jCB_searchActionPerformed
        // TODO add your handling code here:

		search( true );

//		Point locationOnScreen = ExecutionFunctions.instance().safeFunctionExecution( () -> jP_status.getLocationOnScreen() );
//		Component[] matchingComps = ExecutionFunctions.instance().safeFunctionExecution( () -> getComponentsMatchingPosition( locationOnScreen, 20 ) );
    }//GEN-LAST:event_jCB_searchActionPerformed


    // Variables declaration - do not modify//GEN-BEGIN:variables
    // Generated using JFormDesigner non-commercial license
    private JPanel jPanel2;
    private JPanel jP_status;
    private JLabel jL_status;
    private JLabel jL_hour;
    private JPanel jPanel1;
    private JPanel jP_RegExRadioGroupPanel;
    private JRadioButton jRB_NormalText;
    private JRadioButton jRB_RegularExpression;
    private JPanel jP_textToFind;
    private JComboBox<String> jCB_search;
    private JCheckBox jCB_alwaysOnTop;
    private JPanel jP_buttons;
    private JButton jB_forward;
    private JButton jB_replaceForward;
    private JButton jB_replaceBackwards;
    private JButton jB_backwards;
    private JButton jB_replaceAll;
    private JButton jB_replace;
    private JPanel jP_replaceTextFor;
    private JComboBox<String> jCB_replaceFor;
    private JPanel jP_AdditionalSearchSettingsPanel;
    private JCheckBox jCB_MatchCase;
    private JCheckBox jCB_wholeWords;
    private JCheckBox jCB_replace;
    private ButtonGroup bg_RegEx;
    // End of variables declaration//GEN-END:variables

	public void search( boolean forward )
	{
		SearchResultInterface result = null;

		SearchSettingsInterface ssi = getSearchSettings();

//		if( _comboGroupManagerForSearchCB != null )
//			_comboGroupManagerForSearchCB.saveCurrentItem();

		result = null;
		boolean regExException = false;

		try
		{
			if( _src != null )
				result = _src.search( ssi, forward );
		}
		catch( RegExException ree )
		{
			result = null;
			regExException = true;
		}

		if( ( result == null ) || !result.resultsDifferentFromNoResults() )
		{
			String status = null;
			if( regExException )
				status = getInternationalString( CONF_ERROR_IN_REGEX );
			else
				status = getInternationalString( CONF_SEARCHED_STRING_NOT_FOUND );
			setStatusFailure( status );
		}
		else
		{
			setStatusSuccess( getInternationalString( CONF_SEARCHED_STRING_FOUND ) );
		}
	}

	protected void replace()
	{
		ReplaceSettingsInterface settings = this.getReplaceSettings();

		if( _comboGroupManagerForReplaceCB != null )
			_comboGroupManagerForReplaceCB.saveCurrentItem();

		ReplaceResultInterface result = null;
		
		boolean regExException = false;
		
		try
		{
			result = _src.replace(settings);
		}
		catch( RegExException ree )
		{
			result = null;
			regExException = true;
		}
		
		if( result != null )
			setStatusSuccess( getInternationalString( CONF_STRING_REPLACED ) );
		else
		{
			String status = null;
			if( regExException )
				status = getInternationalString( CONF_ERROR_IN_REGEX );
			else
				status = getInternationalString( CONF_COULD_NOT_REPLACE_STRING );
			setStatusFailure( status );
		}
	}

	protected void updateConfigurationWithView()
	{
//		_lastConfiguration.setCollectionOfSearchItems(_comboBoxHistoryForSearchCB.getListOfItems() );
//		_lastConfiguration.setCollectionOfReplaceItems(_comboBoxHistoryForReplaceCB.getListOfItems() );

		_lastConfiguration.M_setBoolParamConfiguration( SearchWindowConfiguration.CONF_IS_REGEX, jRB_RegularExpression.isSelected() );
		_lastConfiguration.M_setBoolParamConfiguration( SearchWindowConfiguration.CONF_WHOLE_WORDS, jCB_wholeWords.isSelected() );
		_lastConfiguration.M_setBoolParamConfiguration( SearchWindowConfiguration.CONF_MATCH_CASE, jCB_MatchCase.isSelected() );
		_lastConfiguration.M_setBoolParamConfiguration( SearchWindowConfiguration.CONF_REPLACE, jCB_replace.isSelected() );
		_lastConfiguration.M_setBoolParamConfiguration( SearchWindowConfiguration.CONF_ALWAYS_ON_TOP, jCB_alwaysOnTop.isSelected() );
	}

	protected void saveConfiguration()
	{
		ExecutionFunctions.instance().safeMethodExecution( ()->_lastConfiguration.M_saveConfiguration() );
		ExecutionFunctions.instance().safeMethodExecution( ()-> ((PersistentConfiguration)_comboGroupManagerForSearchCB.getComboBoxContent()).save() );
		ExecutionFunctions.instance().safeMethodExecution( ()-> ((PersistentConfiguration)_comboGroupManagerForReplaceCB.getComboBoxContent()).save() );
	}

	@Override
	public void formWindowClosingEvent( )
	{
		formWindowClosing( false );
	}

	protected void removeListeners()
	{
		_comboGroupManagerForSearchCB.dispose();
		_comboGroupManagerForReplaceCB.dispose();
	}

	@Override
	public void formWindowClosing( boolean close )
	{
		setVisible( false );

		updateConfigurationWithView();
		saveConfiguration();

		if( close )
			removeListeners();

		super.formWindowClosing( close );
	}

	protected void resizeWindowForReplace( boolean replace )
	{
		if( ! SwingUtilities.isEventDispatchThread() )
		{
			SwingUtilities.invokeLater( new Runnable() {
					public void run()
					{
						resizeWindowForReplaceEDT( replace );
					}
			});
		}
		else
			resizeWindowForReplaceEDT( replace );
	}

	protected void resizeWindowForReplaceEDT( boolean replace )
	{
		jB_replaceForward.setVisible( replace );
		jB_replaceBackwards.setVisible( replace );
		jP_replaceTextFor.setVisible( replace );
		jB_replace.setVisible( replace );
		jB_replaceAll.setVisible( replace );

		int preferredHeight = (int) ( jP_AdditionalSearchSettingsPanel.getHeight() +
										jP_buttons.getHeight() +
										jP_textToFind.getHeight() );
		if( replace )
		{
			jP_replaceTextFor.setLocation( jP_replaceTextFor.getLocation().x, preferredHeight );
			preferredHeight += (int) jP_replaceTextFor.getHeight();
		}
		else
		{
			jP_replaceTextFor.setLocation( jP_replaceTextFor.getLocation().x,
				preferredHeight + jP_status.getSize().height );
		}

		jP_status.setBounds( 0, preferredHeight, jP_status.getWidth(), jP_status.getHeight() );
		jPanel1.setBounds( jPanel1.getX(), jPanel1.getY(), jPanel1.getWidth(), preferredHeight );
		jPanel2.setBounds( jPanel2.getX(), jPanel2.getY(), jPanel2.getWidth(), preferredHeight + jP_status.getHeight() );

		preferredHeight += jP_status.getHeight();
		Insets borders = a_intern.getFrameBorder();
		if( borders != null )
		{
			preferredHeight += borders.top + borders.bottom;
		}

		Rectangle bounds = getBounds();
		Dimension dimen = getPreferredSize();
		Dimension newDimen = new Dimension( (int) dimen.getWidth(), preferredHeight );
		setPreferredSize( newDimen );
		setMaximumSize( newDimen );
		setMinimumSize( new Dimension( (int) getMinimumSize().getWidth(), preferredHeight ) );

//		Dimension newSize = new Dimension( bounds.width, preferredHeight );
//		getContentPane().setSize( newSize );
		setSize( bounds.width, preferredHeight );
	}

	protected void radioButtonsRegExChanged()
	{
		boolean normal = jRB_NormalText.isSelected();
		
		jB_backwards.setEnabled( normal );
		jB_replaceBackwards.setEnabled( normal );
	}

	protected void replaceAndFindNext( boolean forward )
	{
		ReplaceSettingsInterface settings = this.getReplaceSettings();

		_comboGroupManagerForSearchCB.saveCurrentItem();
		_comboGroupManagerForReplaceCB.saveCurrentItem();

		SearchReplaceTextInterface.ReplaceAndFindNextResultInterface result = null;
		
		boolean regExException = false;

		try
		{
			result = _src.replaceAndFindNext(settings, forward);
		}
		catch( RegExException ree )
		{
			result = ree.getReplaceAndFindResult();
			regExException = true;
		}

		SearchResultInterface searchResult = null;
		ReplaceResultInterface replaceResult = null;

		if( result != null )
		{
			replaceResult = result.getReplaceResult();
			searchResult = result.getSearchResult();
		}

		String status = null;

		if( replaceResult != null )
		{
			status = getInternationalString( CONF_STRING_REPLACED );

			if( ( searchResult != null ) && ( searchResult.resultsDifferentFromNoResults() ) )
			{
				status = status + " " + getInternationalString( CONF_SEARCHED_STRING_FOUND );
				setStatusSuccess( status );
			}
			else
			{
				if( regExException )
					status = status + " " + getInternationalString( CONF_ERROR_IN_REGEX );
				else
					status = status + " " + getInternationalString( CONF_SEARCHED_STRING_NOT_FOUND );

				setStatusFailure( status );
			}
		}
		else
		{
			if( regExException )
				status = getInternationalString( CONF_ERROR_IN_REGEX );
			else
				status = getInternationalString( CONF_COULD_NOT_REPLACE_STRING );

			setStatusFailure( status );
		}
	}

	protected void replaceAll()
	{
		ReplaceSettingsInterface settings = this.getReplaceSettings();

		_comboGroupManagerForSearchCB.saveCurrentItem();
		_comboGroupManagerForReplaceCB.saveCurrentItem();

		SearchResultInterface[] result = null;
		
		boolean regExException = false;
		try
		{
			result = _src.replaceAll(settings);
		}
		catch( RegExException ree )
		{
			result = null;
			regExException = true;
		}

		String status = null;

		if( ( result != null ) && ( result.length > 0 ) )
		{
			status = getInternationalString( CONF_STRING_REPLACED );

			status = String.format( "%s : %d", getInternationalString( CONF_NUMBER_OF_REPLACEMENTS_DONE ), result.length );

			setStatusSuccess( status );
		}
		else
		{
			if( regExException )
				status = getInternationalString( CONF_ERROR_IN_REGEX );
			else
				status = getInternationalString( CONF_NO_REPLACEMENTS_DONE );
			
			setStatusFailure( status );
		}
	}

	protected void setStatusSuccess( String text )
	{
		setStatusText( text );
		jL_status.setForeground( new Color( 1, 191, 1 ) );
	}
	
	protected void setStatusFailure( String text )
	{
		setStatusText( text );
		jL_status.setForeground( Color.RED );
	}
	
	protected void setStatusText( String text )
	{
		Date date = new Date();

		jL_hour.setText( DateFunctions.instance().formatDate(date, "HH:mm:ss", null ) + " - " );
		
		if( ( text == null ) || ( text.length() == 0 ) )
			text = " ";		// so that the JLabel does not size width 0.

		jL_status.setText( text );
	}

	@Override
	public void changeZoomFactor(double zoomFactor)
	{
		boolean replace = jCB_replace.isSelected();
		resizeWindowForReplace( replace );

		super.changeZoomFactor( zoomFactor );
	}

	@Override
	public void setSearchReplaceContext(SearchReplaceContextInterface src)
	{
		_src = src;
	}

	@Override
	protected void translateMappedComponents(ComponentMapper compMapper) {
		jRB_RegularExpression = compMapper.mapComponent(jRB_RegularExpression);
		jRB_NormalText = compMapper.mapComponent(jRB_NormalText);
		jPanel2 = compMapper.mapComponent(jPanel2);
		jPanel1 = compMapper.mapComponent(jPanel1);
		jP_textToFind = compMapper.mapComponent(jP_textToFind);
		jP_status = compMapper.mapComponent(jP_status);
		jP_replaceTextFor = compMapper.mapComponent(jP_replaceTextFor);
		jP_buttons = compMapper.mapComponent(jP_buttons);
		jP_RegExRadioGroupPanel = compMapper.mapComponent(jP_RegExRadioGroupPanel);
		jP_AdditionalSearchSettingsPanel = compMapper.mapComponent(jP_AdditionalSearchSettingsPanel);
		jL_status = compMapper.mapComponent(jL_status);
		jL_hour = compMapper.mapComponent(jL_hour);
		jCB_wholeWords = compMapper.mapComponent(jCB_wholeWords);
		jCB_search = compMapper.mapComponent(jCB_search);
		jCB_replaceFor = compMapper.mapComponent(jCB_replaceFor);
		jCB_replace = compMapper.mapComponent(jCB_replace);
		jCB_alwaysOnTop = compMapper.mapComponent(jCB_alwaysOnTop);
		jCB_MatchCase = compMapper.mapComponent(jCB_MatchCase);
		jB_replaceForward = compMapper.mapComponent(jB_replaceForward);
		jB_replaceBackwards = compMapper.mapComponent(jB_replaceBackwards);
		jB_replaceAll = compMapper.mapComponent(jB_replaceAll);
		jB_replace = compMapper.mapComponent(jB_replace);
		jB_forward = compMapper.mapComponent(jB_forward);
		jB_backwards = compMapper.mapComponent(jB_backwards);
		bg_RegEx = compMapper.mapComponent(bg_RegEx);

		if( !hasBeenAlreadyMapped() )
			initView( );

		_comboGroupManagerForSearchCB.setComponentMapper(compMapper);
		_comboGroupManagerForReplaceCB.setComponentMapper(compMapper);
	}

	@Override
	public void setBounds( int xx, int yy, int width, int height )
	{
		super.setBounds( xx, yy, width, height );
	}

	@Override
	public void setBounds( Rectangle rect )
	{
		super.setBounds( rect );
	}

}
