/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.libgeometry.algebra.helpers;

import com.frojasg1.libgeometry.types.GeometricRectangle;
import com.frojasg1.libgeometry.types.GeometricTriangle;
import com.frojasg1.libgeometry.types.Point2d;
import com.frojasg1.libgeometry.types.factory.GeometryCommonFactory;
import com.frojasg1.general.ArrayFunctions;
import com.frojasg1.general.comparators.ComparatorFunctions;
import com.frojasg1.general.number.DoubleFunctions;

import java.awt.geom.Line2D;
import java.util.Arrays;
import java.util.Comparator;
import java.util.function.BiConsumer;
import java.util.function.BiFunction;
import java.util.function.Consumer;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 *
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 */
public class GeometryHelper {
	private static final Logger LOGGER = LoggerFactory.getLogger(GeometryHelper.class);

	protected static GeometryHelper INSTANCE = new GeometryHelper();

	protected static final double TOLERANCE = 1e-6d;

	public static final Point2d ORIGIN = new Point2d(0,0);

	protected Comparator<Point2d> TOP_LEFT_COMPARATOR;
	protected Comparator<Point2d> BOTTOM_RIGHT_COMPARATOR;

	public static GeometryHelper instance()
	{
		return( INSTANCE );
	}

	protected GeometryHelper()
	{
		init();
	}

	protected void init()
	{
		TOP_LEFT_COMPARATOR = notNullFirstComparator( this::topLeftComparator );
		BOTTOM_RIGHT_COMPARATOR = notNullFirstComparator( this::bottomRightComparator );
	}

	protected int topLeftComparator( Point2d pt1, Point2d pt2 )
	{
		int result = toInt( pt1.y - pt2.y );
		if( result == 0 )
			result = toInt( pt1.x - pt2.x );

		return( result );
	}

	protected int bottomRightComparator( Point2d pt1, Point2d pt2 )
	{
		return( - topLeftComparator( pt1, pt2 ) );
	}

	public boolean isNearZero( double value )
	{
		return( abs( value ) < 0.0001d );
	}

//  /            \   /     \   /           \
//  | pt2x - x21 |   | A B |   | pt1x - x11 |
//  | pt2y - y21 | = | C D | * | pt1y - y11 |
//  \            /   \     /   \           /
	public Point2d mapTrianglePt( Point2d pt,
		Point2d pt11, Point2d pt12, Point2d pt13,
		Point2d pt21, Point2d pt22, Point2d pt23 )
	{
		Point2d result = null;

		double dx12 = pt12.x - pt11.x;
		double dx13 = pt13.x - pt11.x;
		double dy12 = pt12.y - pt11.y;
		double dy13 = pt13.y - pt11.y;

		double dx22 = pt22.x - pt21.x;
		double dx23 = pt23.x - pt21.x;
		double dy22 = pt22.y - pt21.y;
		double dy23 = pt23.y - pt21.y;

		double det = dx12 * dy13 - dx13 * dy12;

		if( isNearZero( det ) )
		{
			if( ( !isNearZero(dx12) || !isNearZero(dy12) ) &&
				( !isNearZero(dx22) || !isNearZero(dy22) ) )
			{
				result = mapSegmentPt( pt, pt11, pt12, pt21, pt22 );
			}
			else if( ( !isNearZero(dx13) || !isNearZero(dy13) ) &&
				( !isNearZero(dx23) || !isNearZero(dy23) ) )
			{
				result = mapSegmentPt( pt, pt11, pt13, pt21, pt23 );
			}
			else
				result = pt21;
		}
		else
		{
			double A = (dx22 * dy13 - dx23 * dy12 ) / det;
			double B = (dx12 * dx23 - dx13 * dx22 ) / det;
			double C = (dy22 * dy13 - dy23 * dy12 ) / det;
			double D = (dx12 * dy23 - dx13 * dy22 ) / det;

			result = mapPoint(pt, pt11, pt21, A, B, C, D);
		}

		return( result );
	}

// PT11 = (x11, y11) = PT11
// PT12 = (x12, y12) = PT11 + (dx12, dy12)
// V1 = (dx12, dy12)

// PT21 = (x21, y21) = PT22
// PT22 = (x22, y22) = PT22 + (dx22, dy22)
// V2 = (dx22, dy22)

// res = PT21 + t * V2
	protected Point2d mapSegmentPt( Point2d pt, Point2d pt11, Point2d pt12,
									Point2d pt21, Point2d pt22 )
	{
		double dx12 = pt12.x - pt11.x;
		double dy12 = pt12.y - pt11.y;

		double dx22 = pt22.x - pt21.x;
		double dy22 = pt22.y - pt21.y;

		double t;
		if( isNearZero( dx12 ) && isNearZero( dy12 ) )
			t = 0;
		if( abs( dx12 ) > abs( dy12 ) )
			t = ( pt.x - pt11.x ) / dx12;
		else
			t = ( pt.y - pt11.y ) / dy12;

		return( mapPointOfSegment( pt21, t, dx22, dy22 ) );
	}

	protected Point2d mapPointOfSegment( Point2d pt0, double t,
										double dx, double dy )
	{
		Point2d result = createPoint(0, 0);
		mapPointOfSegment( pt0, t, dx, dy, result);

		return( result );
	}

	protected void mapPointOfSegment( Point2d pt0, double t,
										double dx, double dy,
										Point2d result)
	{
		result.x = toInt( pt0.x + t * dx );
		result.y = toInt( pt0.y + t * dy );
	}

	public void browseMappedTriangle( BiConsumer<Point2d, Point2d> function,
									  Point2d pt11, Point2d pt12, Point2d pt13,
									  Point2d pt21, Point2d pt22, Point2d pt23,
									  GeometricRectangle clip)
	{
		browseTriangle(pt21, pt22, pt23, clip,
				pt -> processMappedTrianglePoint(pt, function, pt21, pt22, pt23, pt11, pt12, pt13));
	}

	// pt is inside the triangle pt11, pt12, pt13
	protected void processMappedTrianglePoint(Point2d pt, BiConsumer<Point2d, Point2d> function,
											  Point2d pt11, Point2d pt12, Point2d pt13,
											  Point2d pt21, Point2d pt22, Point2d pt23) {
		Point2d mappedPt = mapTrianglePt(pt, pt11, pt12, pt13, pt21, pt22, pt23);
		function.accept(mappedPt, pt);
	}

// mapped triangle: pt21, pt22, pt23
//  /            \   /     \   /           \
//  | pt1x - x11 |   | A B |   | pt2x - x21 |
//  | pt1y - y11 | = | C D | * | pt2y - y21 |
//  \            /   \     /   \           /
	public void browseMappedTriangleBad( BiConsumer<Point2d, Point2d> function,
		Point2d pt11, Point2d pt12, Point2d pt13,
		Point2d pt21, Point2d pt22, Point2d pt23,
		GeometricRectangle clip)
	{
		double dx12 = pt12.x - pt11.x;
		double dx13 = pt13.x - pt11.x;
		double dy12 = pt12.y - pt11.y;
		double dy13 = pt13.y - pt11.y;

		double dx22 = pt22.x - pt21.x;
		double dx23 = pt23.x - pt21.x;
		double dy22 = pt22.y - pt21.y;
		double dy23 = pt23.y - pt21.y;

		double det = dx22 * dy23 - dx23 * dy22;

		if( isNearZero( det ) )
		{
			if( ( !isNearZero(dx12) || !isNearZero(dy12) ) &&
				( !isNearZero(dx22) || !isNearZero(dy22) ) )
			{
				browseMappedSegment( function, pt11, pt12, pt21, pt22 );
			}
			else if( ( !isNearZero(dx13) || !isNearZero(dy13) ) &&
				( !isNearZero(dx23) || !isNearZero(dy23) ) )
			{
				browseMappedSegment( function, pt11, pt13, pt21, pt23 );
			}
			else
				function.accept(pt11, pt21);
		}
		else
		{
			Point2d origPt = createPoint(0,0);

			// we are mapping the point in the mapped triangle, to a point on the original triangle
			// so the indices are reversed
			double A = (dx12 * dy23 - dx13 * dy22 ) / det;
			double B = (dx22 * dx13 - dx23 * dx12 ) / det;
			double C = (dy12 * dy23 - dy13 * dy22 ) / det;
			double D = (dx22 * dy13 - dx23 * dy12 ) / det;

			browseTriangle( pt21, pt22, pt23, clip,
				pt -> unmapAndInvoke( function, pt, pt21, pt11, A, B, C, D, origPt ) );
		}
	}

// PT11 = (x11, y11) = PT11
// PT12 = (x12, y12) = PT11 + (dx12, dy12)
// V1 = (dx12, dy12)

// PT21 = (x21, y21) = PT22
// PT22 = (x22, y22) = PT22 + (dx22, dy22)
// V2 = (dx22, dy22)

// res = PT21 + t * V2
	protected void browseMappedSegment( BiConsumer<Point2d, Point2d> function,
										Point2d pt11, Point2d pt12,
										Point2d pt21, Point2d pt22 )
	{
		double dx12 = pt12.x - pt11.x;
		double dy12 = pt12.y - pt11.y;

		double dx22 = pt22.x - pt21.x;
		double dy22 = pt22.y - pt21.y;

		Point2d mappedPt = createPoint(0,0);
		Point2d origPt = createPoint(0,0);
		if( isNearZero( dx22 ) && isNearZero( dy22 ) )
			function.accept( pt11, pt21 );
		if( dx12 > dy12 )
		{
			double dt = 1 / dx12;
			int steps = toInt( Math.abs( pt21.x - pt22.x ) );
			browseSegmentInternal( function, pt11, pt21, dt, dx12, dy12, dx22, dy22,
									steps, mappedPt, origPt );
		}
		else
		{
			double dt = 1 / dy12;
			int steps = toInt( Math.abs( pt21.y - pt22.y ) );
			browseSegmentInternal( function, pt11, pt21, dt, dx12, dy12, dx22, dy22,
									steps, mappedPt, origPt );
		}
	}

	protected void browseSegmentInternal( BiConsumer<Point2d, Point2d> function,
											Point2d pt11, Point2d pt21,
											double dt,
											double dx12, double dy12,
											double dx22, double dy22,
											int steps,
											Point2d mappedPt, Point2d origPt)
	{
		double t = 0;
		for( int ii = 0; ii <= steps; ii++ )
		{
			mapPointOfSegment( pt21, t, dx22, dy22, mappedPt );
			mapPointOfSegment( pt11, t, dx12, dy12, origPt );
			function.accept( origPt, mappedPt );
			t += dt;
		}
	}

	protected Point2d createPoint( double xx, double yy )
	{
		return( GeometryCommonFactory.instance().createPoint2d(xx, yy) );
	}

	protected int toInt( double value )
	{
		return( (int) Math.round(value) );
	}

	protected Point2d mapPoint( Point2d pt, Point2d pt11, Point2d pt21,
		double A, double B, double C, double D )
	{
		return( mapPoint( pt, pt11, pt21, A, B, C, D, createPoint( 0, 0 ) ) );
	}

//  /            \   /     \   /           \
//  | pt2x - x21 |   | A B |   | pt1x - x11 |
//  | pt2y - y21 | = | C D | * | pt1y - y11 |
//  \            /   \     /   \           /
	protected Point2d mapPoint( Point2d pt, Point2d pt11, Point2d pt21,
		double A, double B, double C, double D, Point2d result )
	{
		result.x = toInt( pt21.x + A * (pt.x - pt11.x ) + B * ( pt.y - pt11.y ) );
		result.y = toInt( pt21.y + C * (pt.x - pt11.x ) + D * ( pt.y - pt11.y ) );

		return( result );
	}

	protected void unmapAndInvoke( BiConsumer<Point2d, Point2d> function,
		Point2d mappedPt, Point2d pt21, Point2d pt11,
		double A, double B, double C, double D, Point2d origPt )
	{
		Point2d originalTrianglePt = mapPoint( mappedPt, pt21, pt11, A, B, C, D, origPt );
		function.accept( originalTrianglePt, mappedPt);
	}

	public void browseTriangle(Point2d pt1, Point2d pt2, Point2d pt3,
                               GeometricRectangle clip, Consumer<Point2d> function )
	{
		browseTriangle(pt1, pt2, pt3, 0, clip, function);
	}

	public void browseTriangle(Point2d pt1, Point2d pt2, Point2d pt3, int protectionIncrement,
                               GeometricRectangle clip, Consumer<Point2d> function )
	{
		browseTriangle( pt1, pt2, pt3, protectionIncrement, clip, function, createPoint(0, 0) );
	}

	protected void log( String controlStr, Object ... obj  )
	{
		System.out.println( String.format( controlStr, obj )  );
	}

	protected int round( double value )
	{
		return( DoubleFunctions.instance().round(value) );
	}

	protected GeometryCommonFactory getDelaunayCommonFactory() {
		return GeometryCommonFactory.instance();
	}

	public GeometricTriangle createDelaunayTriangle(Point2d pt1, Point2d pt2, Point2d pt3) {
		return getDelaunayCommonFactory().createDelaunayTriangle(pt1, pt2, pt3, false);
	}

	public GeometricTriangle applyProtectionIncrement(Point2d pt1, Point2d pt2, Point2d pt3,
													  int protectionIncrement) {
		return applyProtectionIncrement(createDelaunayTriangle(pt1, pt2, pt3), protectionIncrement);
	}

	public GeometricTriangle applyProtectionIncrement(GeometricTriangle result, int protectionIncrement) {
		Point2d inCenter = result.getInCenter();
		double longestSquaredDistance = inCenter.squaredDistance(result.getVertex1());
		double tmp = inCenter.squaredDistance(result.getVertex2());
		if(tmp > longestSquaredDistance) {
			longestSquaredDistance = tmp;
		}
		tmp = inCenter.squaredDistance(result.getVertex3());
		if(tmp > longestSquaredDistance) {
			longestSquaredDistance = tmp;
		}

		if(longestSquaredDistance > 1.0d) {
			double factorToApply = 1.0d + (protectionIncrement / Math.sqrt(longestSquaredDistance));
			result = result.zoom(factorToApply);
		}

		return result;
	}

	protected void browseTriangle( Point2d pt1, Point2d pt2, Point2d pt3,
								   int protectionIncrement, GeometricRectangle clip,
								   Consumer<Point2d> function, Point2d domainPt )
	{
		if(protectionIncrement != 0) {
			GeometricTriangle triangle = applyProtectionIncrement(pt1, pt2, pt3, protectionIncrement);
			pt1 = triangle.getVertex1();
			pt2 = triangle.getVertex2();
			pt3 = triangle.getVertex3();
		}

		Point2d topLeftPt = getTopLeftPoint( pt1, pt2, pt3 );
		Point2d bottomRightPt = getBottomRightPoint( pt1, pt2, pt3 );
		Point2d middlePt = getFirstNotRepeated( pt1, pt2, pt3, topLeftPt, bottomRightPt );

//		LOGGER.info( "p1 = {}", topLeftPt );
//		LOGGER.info( "p2 = {}", bottomRightPt );
//		LOGGER.info( "p3 = {}", middlePt );
//		log( "p1 = %s", topLeftPt );
//		log( "p2 = %s", bottomRightPt );
//		log( "p3 = %s", middlePt );

		double slopeInverse1 = getSlopeInverse( topLeftPt, bottomRightPt );
		double slopeInverse2 = getSlopeInverse( topLeftPt, middlePt );

		double x1 = topLeftPt.x;
		double x2 = ( topLeftPt.y - middlePt.y == 0 ) ? middlePt.x : topLeftPt.x;
//		AtomicReference<Double> x1Ref = new AtomicReference<>(topLeftPt.x);
//		AtomicReference<Double> x2Ref = new AtomicReference<>(
//			( topLeftPt.y - middlePt.y == 0 ) ? middlePt.x : topLeftPt.x );

		int fromY = (int) topLeftPt.y;
		int toY = (int) middlePt.y;
//		browseTwoEdgesInternal( function, fromY, toY, x1Ref, x2Ref,
//								slopeInverse1, slopeInverse2, domainPt );
		browseTwoEdgesInternal( function, fromY, toY, x1, x2,
								slopeInverse1, slopeInverse2, domainPt, clip );

//		x1Ref.set(middlePt.x);
//		x2Ref.set(
//			( middlePt.y - bottomRightPt.y == 0 ) ? bottomRightPt.x : middlePt.x );
		slopeInverse2 = getSlopeInverse( middlePt, bottomRightPt );
		fromY = (int) middlePt.y;
		toY = (int) bottomRightPt.y;
		x1 = topLeftPt.x + (fromY - topLeftPt.y) * slopeInverse1;
		x2 = middlePt.x;

		browseTwoEdgesInternal( function, fromY, toY, x1, x2,
								slopeInverse1, slopeInverse2, domainPt, clip );
	}

	public <CC> CC getFirstNotRepeated( CC ... array )
	{
		return( ArrayFunctions.instance().getFirstNotRepeated( array ) );
	}

	protected Point2d getTopLeftPoint( Point2d ... points )
	{
		return( chooseBest( TOP_LEFT_COMPARATOR, points ) );
	}

	protected Point2d getBottomRightPoint( Point2d ... points )
	{
		return( chooseBest( BOTTOM_RIGHT_COMPARATOR, points ) );
	}

	protected Point2d chooseBest( Comparator<Point2d> selector, Point2d ... points )
	{
		return( ArrayFunctions.instance().chooseBest( selector, points ) );
	}

	public <CC> Comparator<CC> notNullFirstComparator( Comparator<CC> comparator )
	{
		return( ComparatorFunctions.instance().notNullFirstComparator( comparator ) );
	}

	protected int sgn( double value )
	{
		return( DoubleFunctions.instance().sgn(value) );
	}

	protected double getSlopeInverse( Point2d pt1, Point2d pt2 )
	{
		double dy = pt2.y - pt1.y;
		double result = sgn(pt2.x - pt1.x);
		if( ! isNearZero( dy ) )
			result = (pt2.x - pt1.x ) / dy;

		return( result );
	}
/*
	protected void browseTwoEdgesInternal( Consumer<Point2d> function,
											int fromY, int toY,
											AtomicReference<Double> x1Ref,
											AtomicReference<Double> x2Ref,
											double slopeInverse1, double slopeInverse2,
											Point2d domainPt, DelaunayRectangle clip )
	{
		double x1 = x1Ref.get();
		double x2 = x2Ref.get();
		int fromX;
		int toX;
		for( int yy = fromY; yy<=toY; yy++ )
		{
			x1 += slopeInverse1;
			x2 += slopeInverse2;
			fromX = toInt( Math.min( x1, x2 ) );
			toX = toInt( Math.max( x1, x2 ) );
			for( int xx = fromX; xx <= toX; xx++ )
			{
				domainPt.x = xx;
				domainPt.y = yy;
				function.accept( domainPt );
			}
//			LOGGER.info( "from = ({}, {})", fromX, yy );
//			LOGGER.info( "to = ({}, {})", toX, yy );
//			log( "from = (%s, %s)", fromX, yy );
//			log( "to = (%s, %s)", toX, yy );

//			if( yy < toY )
//			{
//				x1 += slopeInverse1;
//				x2 += slopeInverse2;
//			}
		}

		x1Ref.set(x1);
		x2Ref.set(x2);
	}
*/
	protected int maxInt( double value1, double value2 )
	{
		return( DoubleFunctions.instance().maxInt(value1, value2) );
	}

	protected int minInt( double value1, double value2 )
	{
		return( DoubleFunctions.instance().minInt(value1, value2) );
	}

	protected void browseTwoEdgesInternal( Consumer<Point2d> function,
											int fromY, int toY,
											double iniX1,
											double iniX2,
											double slopeInverse1, double slopeInverse2,
											Point2d domainPt, GeometricRectangle clip )
	{
		double x1 = iniX1;
		double x2 = iniX2;
		int fromX;
		int toX;
		int clippedFromY = fromY;
		int clippedToY = toY;

		if( clip != null )
		{
			clippedFromY = maxInt( fromY, clip.y );
			clippedToY = minInt( toY, clip.y + clip.height );
		}

		for( int yy = clippedFromY; yy<=clippedToY; yy++ )
		{
			x1 = iniX1 + (yy - fromY) * slopeInverse1;
			x2 = iniX2 + (yy - fromY) * slopeInverse2;
			fromX = toInt( Math.min( x1, x2 ) );
			toX = toInt( Math.max( x1, x2 ) );
			if( clip != null )
			{
				fromX = maxInt( fromX, clip.x );
				toX = minInt( toX, clip.x + clip.width );
			}
			for( int xx = fromX; xx <= toX; xx++ )
			{
				domainPt.x = xx;
				domainPt.y = yy;
				function.accept( domainPt );
			}
		}
	}

	public double abs( double dd )
	{
		return( DoubleFunctions.instance().abs( dd ) );
	}
//
//	protected boolean triAreaMatches( Point2d a, Point2d b, Point2d c, double sgn )
//	{
//		// tolerance (if sgn == 0, then must be equal for matching)
//		// if not, a triarea of zero is also possible for matching.
//		return( abs( sgn( TriArea( a, b, c ) ) - sgn ) <= abs( sgn ) );
//	}
//
//	// returns if the point pt is strictly inside the half-space v1 = (b - a), v2 = (c - a),
//	// with the angle lesser than 180 degrees (origin point a)
//	public boolean isStrictlyInsideHalfSpace( Point2d pt, Point2d a, Point2d b, Point2d c )
//	{
//		boolean result = false;
//		double sgn = sgn(TriArea( a, b, c ));
//		if(!isClose(sgn, 0d)) {
//			double sgn1 = sgn(TriArea( a, b, pt ));
//			double sgn2 = sgn(TriArea( a, pt, c ));
//			result = (sgn1 == sgn) && (sgn2 == sgn);
//		}
//
//		return result;
//	}
//
//	public boolean isApproximatelyInsideHalfSpace( Point2d pt, Point2d a, Point2d b, Point2d c) {
//		return isApproximatelyInsideHalfSpace(pt, a, b, c, TOLERANCE);
//	}
//
//	// returns if the point pt is approximately inside the half-space v1 = (b - a), v2 = (c - a),
//	// with the angle lesser than 180 degrees (origin point a)
//	// it returns if it is strictly inside, or if it is approximately over any of the two semi-line
//	public boolean isApproximatelyInsideHalfSpace( Point2d pt, Point2d a, Point2d b, Point2d c, double tolerance )
//	{
//		boolean result = false;
//		double sgn = sgn(TriArea( a, b, c ));
//		if(!isClose(sgn, 0d)) {
//			double sgn1 = sgn(TriArea( a, b, pt ));
//			double sgn2 = sgn(TriArea( a, pt, c ));
//			result = ((sgn1 == sgn) || isClose(a, b, pt, tolerance)) &&
//					((sgn2 == sgn) || isClose(a, c, pt, tolerance));
//		}
//
//		return result;
//	}
//
//	// the function returns if pt is close to the line defined by the vector (a, b)
//	public boolean isClose(Point2d a, Point2d b, Point2d pt, double tolerance) {
//		return isClose(getSquaredDistanceBetweenLineAndPoint(a, b, pt), 0d, tolerance);
//	}
//
//
//	// returns if the point pt is inside the triangle a, b, c
//	public boolean isInside( Point2d pt, Point2d a, Point2d b, Point2d c )
//	{
//		double sgn = sgn( TriArea( a, b, c ) );
//
//		return triAreaMatches( a, pt, c, sgn ) &&
//				triAreaMatches( c, pt, b, sgn ) &&
//				triAreaMatches( b, pt, a, sgn );
//	}
//
//	public boolean isApproximatelyInside( Point2d pt, Point2d a, Point2d b, Point2d c, double tolerance ) {
//		return (isInside(pt, a, b, c) ||
//				isClose(a, b, pt, tolerance) ||
//				isClose(a, c, pt, tolerance) ||
//				isClose(b, c, pt, tolerance));
//
//	}
//
//	protected DelaunayQuadEdgeHelpers getDelaunayQuadEdgeHelpers() {
//		return DelaunayQuadEdgeHelpers.instance();
//	}
//
//	public double TriArea(Point2d a, Point2d b, Point2d c)
//	// Returns twice the area of the oriented triangle (a, b, c), i.e., the
//	// area is positive if the triangle is oriented counterclockwise.
//	{
//		return getDelaunayQuadEdgeHelpers().TriArea(a, b, c);
//	}
//
//	public double getSquaredDistanceBetweenLineAndPoint(Point2d pt1, Point2d pt2, Point2d pt) {
//		return getDelaunayQuadEdgeHelpers().getDistanceBetweenSegmentAndPoint(pt1, pt2, pt);
//	}
//
	public Point2d intAndBound(Point2d initialValue, GeometricRectangle bounds )
	{
		Point2d result = initialValue;
		if( !bounds.isInside(initialValue) )
			result = boundSame( result.createCopy(), bounds );

		return( result );
	}

	protected double limit( double value, double lower, double upper )
	{
		return( DoubleFunctions.instance().limit(value, lower, upper) );
	}

	public Point2d boundSame( Point2d result, GeometricRectangle bounds )
	{
		if((bounds != null) && (result != null)) {
			result.x = limit(result.x, bounds.x, bounds.x + bounds.width - 1);
			result.y = limit(result.y, bounds.y, bounds.y + bounds.height - 1);
		}

		return( result );
	}
//
//	public static void main( String[] args )
//	{
//		Point2d pt1 = INSTANCE.createPoint(100, 0);
//		Point2d pt2 = INSTANCE.createPoint(-25, 50);
//		Point2d pt3 = INSTANCE.createPoint(0, 100);
//
//		INSTANCE.browseTriangle(pt1, pt2, pt3, null, pt -> {} );
//	}
//
//	protected Point2d zoom( Point2d pt, double zoomFactor )
//	{
//		return( zoom(pt, zoomFactor, this::zoom ) );
//	}
//
//	protected double zoom( double value, double zoomFactor )
//	{
//		return( value * zoomFactor );
//	}
//
//	protected Point2d zoomToInt( Point2d pt, double zoomFactor )
//	{
//		return( zoom(pt, zoomFactor, this::zoomToInt ) );
//	}
//
//	protected int zoomToInt( double value, double zoomFactor )
//	{
//		return( IntegerFunctions.zoomValueRound( value, zoomFactor ) );
//	}
//
//	protected Point2d zoom( Point2d pt, double zoomFactor,
//							BiFunction<Double, Double, Number> zoomFunction )
//	{
//		Point2d result = pt;
//		if( ( pt != null ) && ( zoomFactor != 1.0d ) )
//		{
//			result = pt.createCopy();
//
//			result.x = zoomFunction.apply( result.x, zoomFactor ).doubleValue();
//			result.y = zoomFunction.apply( result.y, zoomFactor ).doubleValue();
//		}
//
//		return( result );
//	}
//
//	protected Point2d getNormalizedVector( Point2d p1, Point2d p2 )
//	{
//		Point2d result = p2.subtract(p1);
//		double dist = result.norm();
//		result = zoom( result, 1 / dist );
//
//		return( result );
//	}
//
//	public Point2d getNormalizedBisectorVector( Point2d p1, Point2d p2, Point2d p3 )
//	{
//		Point2d v1 = getNormalizedVector( p1, p2 );
//		Point2d v2 = getNormalizedVector( p1, p3 );
//		Point2d v3 = v1.createCopy();
//		v3.x = ( v1.x + v2.x ) / 2;
//		v3.y = ( v1.y + v2.y ) / 2;
//
//		return( getNormalizedVector( ORIGIN, v3 ) );
//	}
//
	public Point2d barycenter(GeometricTriangle triangle) {
		Point2d result = triangle.getVertex1().createCopy();

		result.x = toInt(triangle.getVertex1().x + triangle.getVertex2().x + triangle.getVertex3().x ) / 3;
		result.y = toInt(triangle.getVertex1().y + triangle.getVertex2().y + triangle.getVertex3().y ) / 3;

		return result;
	}

	public boolean isAnyInside(GeometricTriangle triangle, Point2d ... pts) {
		return Arrays.stream(pts)
				.anyMatch(pt -> isInside(pt, triangle.getVertex1(), triangle.getVertex2(), triangle.getVertex3()));
	}

	protected VectorFunctions getVectorFunctions() {
		return VectorFunctions.instance();
	}

	protected boolean isInside(Point2d pt, Point2d a, Point2d b, Point2d c) {
		return getVectorFunctions().isInside(pt, a, b, c);
	}

	/**
	 *
 	 * @param triangle
	 * @param visitor  - edge visitor. If visitor returns true, we continue with the next edge
	 * @return  boolean that says if all edge visits have returned true
	 */
	public boolean browseEdges(GeometricTriangle triangle, BiFunction<Point2d, Point2d, Boolean> visitor) {
		boolean result = false;
		if(triangle != null) {
			result = visitor.apply(triangle.getVertex1(), triangle.getVertex2()) &&
					visitor.apply(triangle.getVertex2(), triangle.getVertex3()) &&
					visitor.apply(triangle.getVertex3(), triangle.getVertex1());
		}
		return result;
	}

	public boolean overlap(GeometricTriangle t1, GeometricTriangle t2) {
		boolean result = isAnyInside(t1, t2.getVertex1(), t2.getVertex2(), t2.getVertex3()) ||
				isAnyInside(t2, t1.getVertex1(), t1.getVertex2(), t1.getVertex3());
		if(!result) {
			result = !browseEdges(t1, (pt1, pt2) -> !edgeIntersectsTriangleBoundary(t2, pt1, pt2));
		}
		return result;
	}

	public boolean edgeIntersectsTriangleBoundary(GeometricTriangle triangle, Point2d edgePt1, Point2d edgePt2) {
		return !browseEdges(triangle, (pt1, pt2) -> !edgesIntersect(pt1, pt2, edgePt1, edgePt2));
	}

	// https://stackoverflow.com/questions/16333650/how-to-check-whether-2-lines-segments-intersect
	public boolean edgesIntersect(Point2d pt11, Point2d pt12, Point2d pt21, Point2d pt22) {
		return Line2D.linesIntersect(pt11.x,pt11.y,pt12.x,pt12.y,pt21.x,pt21.y,pt22.x,pt22.y);
	}

	protected boolean areNear(Point2d pt1, Point2d pt2) {
		return isNearZero(pt1.x - pt2.x) && isNearZero(pt1.y - pt2.y);
	}

	public double getSquaredDistanceBetweenLineAndPoint(Point2d pt1, Point2d pt2, Point2d pt )
	{
		double result = 0;
		if(areNear(pt1, pt2)) {
			result = pt.squaredDistance(pt1);
		} else {
//			double A = pt2.y - pt1.y;
//			double B = pt1.x - pt2.x;
//			double C = -A * pt.x - B * pt.y;
//
//			return (getDistanceBetweenLineAndPoint(A, B, C, pt));

			// https://stackoverflow.com/questions/849211/shortest-distance-between-a-point-and-a-line-segment
			double A = pt.x - pt1.x;
			double B = pt.y - pt1.y;
			double C = pt2.x - pt1.x;
			double D = pt2.y - pt1.y;

			double dot = A * C + B * D;
			double len_sq = C * C + D * D;
			double param = -1;
			if (len_sq != 0) //in case of 0 length line
				param = dot / len_sq;

			double xx, yy;

			if (param < 0) {
				xx = pt1.x;
				yy = pt1.y;
			}
			else if (param > 1) {
				xx = pt2.x;
				yy = pt2.y;
			}
			else {
				xx = pt1.x + param * C;
				yy = pt1.y + param * D;
			}

			double dx = pt.x - xx;
			double dy = pt.y - yy;

			result = dx * dx + dy * dy;
		}

		return result;
	}


	public Point2d limit(Point2d pt, GeometricRectangle rectangle) {
		Point2d result = pt;
		if((rectangle != null) && !rectangle.isInside(result)) {
			result = pt.createCopy();
			result.x = limit(result.x, rectangle.x, rectangle.x + rectangle.width - 1);
			result.y = limit(result.y, rectangle.y, rectangle.y + rectangle.height - 1);
		}
		return result;
	}
//
//	public Point2d addVectors(Point2d ... points) {
//		Point2d result = null;
//		for(Point2d pt: points) {
//			if(result == null) {
//				result = pt.createCopy();
//			} else {
//				result.x += pt.x;
//				result.y += pt.y;
//			}
//		}
//
//		return result;
//	}
//
//	public Point2d multiplyByScalar(Point2d pt, double factor, Point2d result) {
//		if(result == null) {
//			result = pt.createCopy();
//		}
//		result.x = pt.x * factor;
//		result.y = pt.y * factor;
//
//		return result;
//	}
//
//	public Point2d arithmeticMean(Point2d ... points) {
//		Point2d result = addVectors(points);
//		if((result != null) && (points.length > 1)) {
//			double factor = 1d / points.length;
//			multiplyByScalar(result, factor, result);
//		}
//
//		return result;
//	}
//
//	public Point2d getUniraryVector(Point2d point) {
//		Point2d result = null;
//		if(point != null) {
//			result = point.createCopy();
//			double factor = 1.0d / result.norm();
//			multiplyByScalar(result, factor, result);
//		}
//		return result;
//	}
//
//	public double scalarProduct(Point2d vector1, Point2d vector2) {
//		return vector1.x * vector2.x + vector1.y * vector2.y;
//	}
//
//	public Point2d linearCombination(double factor1, Point2d vector1, double factor2, Point2d vector2) {
//		Point2d result = vector1.createCopy();
//		result.x = factor1 * vector1.x + factor2 * vector2.x;
//		result.y = factor1 * vector1.y + factor2 * vector2.y;
//
//		return result;
//	}
//
//	// https://matthew-brett.github.io/teaching/rotation_2d.html
//	public Point2d rotate(Point2d vector, double angle) {
//		Point2d result = vector.createCopy();
//
//		double cos = Math.cos(angle);
//		double sin = Math.sin(angle);
//
//		result.x = cos * vector.x - sin * vector.y;
//		result.y = sin * vector.x + cos * vector.y;
//
//		return result;
//	}
//
//	public Point2d rotate(Point2d originPt, Point2d destPt, double angle) {
//		return addVectors(originPt, rotate(destPt.subtract(originPt), angle));
//	}
//
//	public Point2d getOrthogonalVector(Point2d vector) {
//		Point2d result = vector.createCopy();
//
//		result.x = vector.y;
//		result.y = -vector.x;
//
//		return result;
//	}
//
//	/*
//		P1 = (x1, y1)
//		V1 = (v1x, v1y)
//
//		P2 = (x2, y2)
//		V2 = (v2x, v2y)
//
//		Intersection of
//
//		Line1 = P1 + t1 * V1
//		Line2 = P2 + t2 * V2
//
//        /             \  /    \    /         \
//        | v1x    -v1y |  | t1 |    | x2 - x1 |
//        |             |  |    | =  |         |
//        | v2x    -v2y |  | t2 |    | y2 - y1 |
//        \             /  \    /    \         /
//	 */
//	public Point2d intersection(Point2d point1, Point2d directorVector1, Point2d point2, Point2d directorVector2) {
//		double m11 = directorVector1.x;
//		double m21 = directorVector2.x;
//		double m12 = -directorVector1.y;
//		double m22 = -directorVector2.y;
//
//		double i1 = directorVector2.x - directorVector1.x;
//		double i2 = directorVector2.y - directorVector1.y;
//
//		Double[] t = solve(m11, m12, m21, m22, i1, i2);
//
//		double t1 = t[0];
//		Point2d result = linearCombination(1.0d, point1, t1, directorVector1);
//
//		return result;
//	}
//
//	/*
//		Liniar system solver using Cramer rule.
//	 */
//	public Double[] solve(double m11, double m12, double m21, double m22, double i1, double i2) {
//		double determinant = m11 * m22 - m21 * m12;
//
//		if(isClose(determinant, 0d)) {
//			throw new IllegalArgumentException(String.format("Trying to solve a linear system with determinant close to 0." +
//					" m11: %f, m21: %f, m12: %f, m22: %f", m11, m21, m12, m22));
//		}
//
//		double s1 = (i1 * m22 - i1 * m12) / determinant;
//		double s2 = (m11 * i2 - m21 * i1) / determinant;
//
//		return array(s1, s2);
//	}
//
//	protected DoubleFunctions getDoubleFunctions() {
//		return DoubleFunctions.instance();
//	}
//
//	public boolean isClose(double value1, double value2, double tolerance) {
//		return getDoubleFunctions().isClose(value1, value2, tolerance);
//	}
//
//	protected boolean isClose(double value1, double value2) {
//		return isClose(value1, value2, TOLERANCE);
//	}
//
//	protected <C> C[] array(C ... elements) {
//		return elements;
//	}
//
//	/**
//	 *
//	 * @param originalDirectionVector the vector with the original direction
//	 * @param secondVector the vector of which we will calculate the orthogonal component to the original direction
//	 * @return  is the projection of secondVector to the orthogonal direction of firstDirectionVector
//	 */
//	public Point2d getOrthogonalProjection(Point2d originalDirectionVector, Point2d secondVector) {
//		Point2d uniraryOriginalDirectionVector = getUniraryVector(originalDirectionVector);
//		double scalarProduct = scalarProduct(uniraryOriginalDirectionVector, secondVector);
//
//		return linearCombination(-scalarProduct, uniraryOriginalDirectionVector, 1.0d, secondVector);
//	}
//
//	public Point2d zoom(Point2d origin, Point2d destination, double factor) {
//		return linearCombination(1, origin, factor, destination.subtract(origin));
//	}
}
