/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.libgeometry.algebra.helpers;

import com.frojasg1.libgeometry.types.Point2d;
import com.frojasg1.general.number.DoubleFunctions;
import com.frojasg1.general.number.IntegerFunctions;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.function.BiFunction;

/**
 *
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 */
public class VectorFunctions {
	private static final Logger LOGGER = LoggerFactory.getLogger(VectorFunctions.class);

	protected static VectorFunctions INSTANCE = new VectorFunctions();
	protected static final double TOLERANCE = 1e-6d;


	public static final Point2d ORIGIN = new Point2d(0,0);


	public static VectorFunctions instance()
	{
		return( INSTANCE );
	}


	protected GeometryHelper getGeometryHelper() {
		return GeometryHelper.instance();
	}

	protected boolean isNearZero( double value )
	{
		return( getGeometryHelper().isNearZero(value) );
	}

//
//	protected Point2d createPoint( double xx, double yy )
//	{
//		return( GeometryCommonFactory.instance().createPoint2d(xx, yy) );
//	}

	protected int toInt( double value )
	{
		return( (int) Math.round(value) );
	}


	protected int round( double value )
	{
		return( DoubleFunctions.instance().round(value) );
	}


	protected int sgn( double value )
	{
		return( DoubleFunctions.instance().sgn(value) );
	}


	public double abs( double dd )
	{
		return( DoubleFunctions.instance().abs( dd ) );
	}

	protected boolean triAreaMatches( Point2d a, Point2d b, Point2d c, double sgn )
	{
		// tolerance (if sgn == 0, then must be equal for matching)
		// if not, a triarea of zero is also possible for matching.
		return( abs( sgn( TriArea( a, b, c ) ) - sgn ) <= abs( sgn ) );
	}

	// returns if the point pt is strictly inside the half-space v1 = (b - a), v2 = (c - a),
	// with the angle lesser than 180 degrees (origin point a)
	public boolean isStrictlyInsideHalfSpace( Point2d pt, Point2d a, Point2d b, Point2d c )
	{
		boolean result = false;
		double sgn = sgn(TriArea( a, b, c ));
		if(!isClose(sgn, 0d)) {
			double sgn1 = sgn(TriArea( a, b, pt ));
			double sgn2 = sgn(TriArea( a, pt, c ));
			result = (sgn1 == sgn) && (sgn2 == sgn);
		}

		return result;
	}

	public boolean isApproximatelyInsideHalfSpace( Point2d pt, Point2d a, Point2d b, Point2d c) {
		return isApproximatelyInsideHalfSpace(pt, a, b, c, TOLERANCE);
	}

	// returns if the point pt is approximately inside the half-space v1 = (b - a), v2 = (c - a),
	// with the angle lesser than 180 degrees (origin point a)
	// it returns if it is strictly inside, or if it is approximately over any of the two semi-line
	public boolean isApproximatelyInsideHalfSpace( Point2d pt, Point2d a, Point2d b, Point2d c, double tolerance )
	{
		boolean result = false;
		double sgn = sgn(TriArea( a, b, c ));
		if(!isClose(sgn, 0d)) {
			double sgn1 = sgn(TriArea( a, b, pt ));
			double sgn2 = sgn(TriArea( a, pt, c ));
			result = ((sgn1 == sgn) || isClose(a, b, pt, tolerance)) &&
					((sgn2 == sgn) || isClose(a, c, pt, tolerance));
		}

		return result;
	}

	// the function returns if pt is close to the line defined by the vector (a, b)
	public boolean isClose(Point2d a, Point2d b, Point2d pt, double tolerance) {
		return isClose(getSquaredDistanceBetweenLineAndPoint(a, b, pt), 0d, tolerance);
	}


	protected double getSquaredDistanceBetweenLineAndPoint(Point2d pt1, Point2d pt2, Point2d pt )
	{
		return getGeometryHelper().getSquaredDistanceBetweenLineAndPoint(pt1, pt2, pt);
	}

	// returns if the point pt is inside the triangle a, b, c
	public boolean isInside( Point2d pt, Point2d a, Point2d b, Point2d c )
	{
		double sgn = sgn( TriArea( a, b, c ) );

		return triAreaMatches( a, pt, c, sgn ) &&
				triAreaMatches( c, pt, b, sgn ) &&
				triAreaMatches( b, pt, a, sgn );
	}

	public boolean isApproximatelyInside( Point2d pt, Point2d a, Point2d b, Point2d c, double tolerance ) {
		return (isInside(pt, a, b, c) ||
				isClose(a, b, pt, tolerance) ||
				isClose(a, c, pt, tolerance) ||
				isClose(b, c, pt, tolerance));

	}
//
//	protected DelaunayQuadEdgeHelpers getDelaunayQuadEdgeHelpers() {
//		return DelaunayQuadEdgeHelpers.instance();
//	}

	public double TriArea(Point2d a, Point2d b, Point2d c)
	// Returns twice the area of the oriented triangle (a, b, c), i.e., the
	// area is positive if the triangle is oriented counterclockwise.
	{
		return (b.x - a.x)*(c.y - a.y) - (b.y - a.y)*(c.x - a.x);
	}

//	protected boolean isNearZero( double value )
//	{
//		return( Math.abs(value) < EPS );
//	}

//
//	public double getSquaredDistanceBetweenLineAndPoint(Point2d pt1, Point2d pt2, Point2d pt) {
//		return getDelaunayQuadEdgeHelpers().getDistanceBetweenSegmentAndPoint(pt1, pt2, pt);
//	}

//
//	public Point2d intAndBound(Point2d initialValue, GeometryRectangle bounds )
//	{
//		Point2d result = initialValue;
//		if( !bounds.isInside(initialValue) )
//			result = boundSame( result.createCopy(), bounds );
//
//		return( result );
//	}

	protected double limit( double value, double lower, double upper )
	{
		return( DoubleFunctions.instance().limit(value, lower, upper) );
	}
//
//	public Point2d boundSame( Point2d result, GeometryRectangle bounds )
//	{
//		if((bounds != null) && (result != null)) {
//			result.x = limit(result.x, bounds.x, bounds.x + bounds.width - 1);
//			result.y = limit(result.y, bounds.y, bounds.y + bounds.height - 1);
//		}
//
//		return( result );
//	}

	protected Point2d zoom( Point2d pt, double zoomFactor )
	{
		return( zoom(pt, zoomFactor, this::zoom ) );
	}

	protected double zoom( double value, double zoomFactor )
	{
		return( value * zoomFactor );
	}

	protected Point2d zoomToInt( Point2d pt, double zoomFactor )
	{
		return( zoom(pt, zoomFactor, this::zoomToInt ) );
	}

	protected int zoomToInt( double value, double zoomFactor )
	{
		return( IntegerFunctions.zoomValueRound( value, zoomFactor ) );
	}

	protected Point2d zoom( Point2d pt, double zoomFactor,
							BiFunction<Double, Double, Number> zoomFunction )
	{
		Point2d result = pt;
		if( ( pt != null ) && ( zoomFactor != 1.0d ) )
		{
			result = pt.createCopy();

			result.x = zoomFunction.apply( result.x, zoomFactor ).doubleValue();
			result.y = zoomFunction.apply( result.y, zoomFactor ).doubleValue();
		}

		return( result );
	}

	protected Point2d getNormalizedVector( Point2d p1, Point2d p2 )
	{
		Point2d result = p2.subtract(p1);
		double dist = result.norm();
		result = zoom( result, 1 / dist );

		return( result );
	}

	public Point2d getNormalizedBisectorVector( Point2d p1, Point2d p2, Point2d p3 )
	{
		Point2d v1 = getNormalizedVector( p1, p2 );
		Point2d v2 = getNormalizedVector( p1, p3 );
		Point2d v3 = v1.createCopy();
		v3.x = ( v1.x + v2.x ) / 2;
		v3.y = ( v1.y + v2.y ) / 2;

		return( getNormalizedVector( ORIGIN, v3 ) );
	}
//
//	public boolean isAnyInside(GeometryTriangle triangle, Point2d ... pts) {
//		return Arrays.stream(pts)
//				.anyMatch(pt -> isInside(pt, triangle.getVertex1(), triangle.getVertex2(), triangle.getVertex3()));
//	}

//	/**
//	 *
// 	 * @param triangle
//	 * @param visitor  - edge visitor. If visitor returns true, we continue with the next edge
//	 * @return  boolean that says if all edge visits have returned true
//	 */
//	public boolean browseEdges(GeometryTriangle triangle, BiFunction<Point2d, Point2d, Boolean> visitor) {
//		boolean result = false;
//		if(triangle != null) {
//			result = visitor.apply(triangle.getVertex1(), triangle.getVertex2()) &&
//					visitor.apply(triangle.getVertex2(), triangle.getVertex3()) &&
//					visitor.apply(triangle.getVertex3(), triangle.getVertex1());
//		}
//		return result;
//	}
//
//	public boolean overlap(GeometryTriangle t1, GeometryTriangle t2) {
//		boolean result = isAnyInside(t1, t2.getVertex1(), t2.getVertex2(), t2.getVertex3()) ||
//				isAnyInside(t2, t1.getVertex1(), t1.getVertex2(), t1.getVertex3());
//		if(!result) {
//			result = !browseEdges(t1, (pt1, pt2) -> !edgeIntersectsTriangleBoundary(t2, pt1, pt2));
//		}
//		return result;
//	}
//
//	public boolean edgeIntersectsTriangleBoundary(GeometryTriangle triangle, Point2d edgePt1, Point2d edgePt2) {
//		return !browseEdges(triangle, (pt1, pt2) -> !edgesIntersect(pt1, pt2, edgePt1, edgePt2));
//	}
//
//	// https://stackoverflow.com/questions/16333650/how-to-check-whether-2-lines-segments-intersect
//	public boolean edgesIntersect(Point2d pt11, Point2d pt12, Point2d pt21, Point2d pt22) {
//		return Line2D.linesIntersect(pt11.x,pt11.y,pt12.x,pt12.y,pt21.x,pt21.y,pt22.x,pt22.y);
//	}
//
//	public Point2d limit(Point2d pt, GeometryRectangle rectangle) {
//		Point2d result = pt;
//		if((rectangle != null) && !rectangle.isInside(result)) {
//			result = pt.createCopy();
//			result.x = limit(result.x, rectangle.x, rectangle.x + rectangle.width - 1);
//			result.y = limit(result.y, rectangle.y, rectangle.y + rectangle.height - 1);
//		}
//		return result;
//	}

	public Point2d addVectors(Point2d ... points) {
		Point2d result = null;
		for(Point2d pt: points) {
			if(result == null) {
				result = pt.createCopy();
			} else {
				result.x += pt.x;
				result.y += pt.y;
			}
		}

		return result;
	}

	public Point2d multiplyByScalar(Point2d pt, double factor, Point2d result) {
		if(result == null) {
			result = pt.createCopy();
		}
		result.x = pt.x * factor;
		result.y = pt.y * factor;

		return result;
	}

	public Point2d arithmeticMean(Point2d ... points) {
		Point2d result = addVectors(points);
		if((result != null) && (points.length > 1)) {
			double factor = 1d / points.length;
			multiplyByScalar(result, factor, result);
		}

		return result;
	}

	public Point2d getUniraryVector(Point2d point) {
		Point2d result = null;
		if(point != null) {
			result = point.createCopy();
			double factor = 1.0d / result.norm();
			multiplyByScalar(result, factor, result);
		}
		return result;
	}

	public double scalarProduct(Point2d vector1, Point2d vector2) {
		return vector1.x * vector2.x + vector1.y * vector2.y;
	}

	public Point2d linearCombination(double factor1, Point2d vector1, double factor2, Point2d vector2) {
		Point2d result = vector1.createCopy();
		result.x = factor1 * vector1.x + factor2 * vector2.x;
		result.y = factor1 * vector1.y + factor2 * vector2.y;

		return result;
	}

	// https://matthew-brett.github.io/teaching/rotation_2d.html
	public Point2d rotate(Point2d vector, double angle) {
		Point2d result = vector.createCopy();

		double cos = Math.cos(angle);
		double sin = Math.sin(angle);

		result.x = cos * vector.x - sin * vector.y;
		result.y = sin * vector.x + cos * vector.y;

		return result;
	}

	public Point2d rotate(Point2d originPt, Point2d destPt, double angle) {
		return addVectors(originPt, rotate(destPt.subtract(originPt), angle));
	}

	public Point2d getOrthogonalVector(Point2d vector) {
		Point2d result = vector.createCopy();

		result.x = vector.y;
		result.y = -vector.x;

		return result;
	}

	/*
		P1 = (x1, y1)
		V1 = (v1x, v1y)

		P2 = (x2, y2)
		V2 = (v2x, v2y)

		Intersection of

		Line1 = P1 + t1 * V1
		Line2 = P2 + t2 * V2

        /             \  /    \    /         \
        | v1x    -v1y |  | t1 |    | x2 - x1 |
        |             |  |    | =  |         |
        | v2x    -v2y |  | t2 |    | y2 - y1 |
        \             /  \    /    \         /
	 */
	public Point2d intersection(Point2d point1, Point2d directorVector1, Point2d point2, Point2d directorVector2) {
		double m11 = directorVector1.x;
		double m21 = directorVector2.x;
		double m12 = -directorVector1.y;
		double m22 = -directorVector2.y;

		double i1 = directorVector2.x - directorVector1.x;
		double i2 = directorVector2.y - directorVector1.y;

		Double[] t = solve(m11, m12, m21, m22, i1, i2);

		double t1 = t[0];
		Point2d result = linearCombination(1.0d, point1, t1, directorVector1);

		return result;
	}

	/*
		Liniar system solver using Cramer rule.
	 */
	public Double[] solve(double m11, double m12, double m21, double m22, double i1, double i2) {
		double determinant = m11 * m22 - m21 * m12;

		if(isClose(determinant, 0d)) {
			throw new IllegalArgumentException(String.format("Trying to solve a linear system with determinant close to 0." +
					" m11: %f, m21: %f, m12: %f, m22: %f", m11, m21, m12, m22));
		}

		double s1 = (i1 * m22 - i1 * m12) / determinant;
		double s2 = (m11 * i2 - m21 * i1) / determinant;

		return array(s1, s2);
	}

	protected DoubleFunctions getDoubleFunctions() {
		return DoubleFunctions.instance();
	}

	public boolean isClose(double value1, double value2, double tolerance) {
		return getDoubleFunctions().isClose(value1, value2, tolerance);
	}

	protected boolean isClose(double value1, double value2) {
		return isClose(value1, value2, TOLERANCE);
	}

	protected <C> C[] array(C ... elements) {
		return elements;
	}

	/**
	 *
	 * @param originalDirectionVector the vector with the original direction
	 * @param secondVector the vector of which we will calculate the orthogonal component to the original direction
	 * @return  is the projection of secondVector to the orthogonal direction of firstDirectionVector
	 */
	public Point2d getOrthogonalProjection(Point2d originalDirectionVector, Point2d secondVector) {
		Point2d uniraryOriginalDirectionVector = getUniraryVector(originalDirectionVector);
		double scalarProduct = scalarProduct(uniraryOriginalDirectionVector, secondVector);

		return linearCombination(-scalarProduct, uniraryOriginalDirectionVector, 1.0d, secondVector);
	}

	public Point2d zoom(Point2d origin, Point2d destination, double factor) {
		return linearCombination(1, origin, factor, destination.subtract(origin));
	}
}
