/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.libgeometry.types;

import com.frojasg1.general.functional.interfaces.TriFunction;
import com.frojasg1.general.number.DoubleFunctions;

import java.util.ArrayList;
import java.util.List;
import java.util.Objects;

/**
 *
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 */
public class GeometricRectangle {
	public double x;
	public double y;
	public double width;
	public double height;

	public GeometricRectangle(double x, double y, double width, double height )
	{
		this.x = x;
		this.y = y;
		this.width = width;
		this.height = height;
	}

	public GeometricRectangle createCopy()
	{
		return( new GeometricRectangle( x, y, width, height ) );
	}

	public boolean isInside( Point2d pt )
	{
		return( isInsideGen( pt, this::isInside ) );
	}

	public boolean isStrictlyInside( Point2d pt )
	{
		return( isInsideGen( pt, this::isStrictlyInside ) );
	}

	protected boolean isInsideGen( Point2d pt, TriFunction<Double, Double, Double, Boolean> isInsideFunction )
	{
		boolean result = false;
		if( pt != null )
			result = isInsideGen(pt.x, pt.y, isInsideFunction);

		return( result );
	}

	protected boolean isInsideGen( double x, double y, TriFunction<Double, Double, Double, Boolean> isInsideFunction )
	{
		return( isXinsideGen( x, isInsideFunction ) && isYinsideGen( y, isInsideFunction ) );
	}

	protected boolean isStrictlyInside( double value, double lowerBound, double upperBound )
	{
		return( DoubleFunctions.instance().isStrictlyInside( value, lowerBound, upperBound ) );
	}

	protected boolean isInside( double value, double lowerBound, double upperBound )
	{
		return( DoubleFunctions.instance().isInside( value, lowerBound, upperBound ) );
	}

	protected boolean isXinsideGen( double x, TriFunction<Double, Double, Double, Boolean> isInsideFunction )
	{
		return( isInsideFunction.apply(x, this.x, this.x + this.width - 1 ) );
	}

	protected boolean isYinsideGen( double y, TriFunction<Double, Double, Double, Boolean> isInsideFunction )
	{
		return( isInsideFunction.apply( y, this.y, this.y + this.height - 1 ) );
	}

	protected Point2d createPoint( double xx, double yy )
	{
		return( new Point2d( xx, yy ) );
	}

	public Point2d center()
	{
		return( createPoint( x + width / 2, y + height / 2 ) );
	}

	public boolean isBoundVertex(Point2d pt) {
		boolean result = false;
		if(pt != null) {
			result = isBoundVertex(pt.x, pt.y);
		}
		return result;
	}

	public boolean isBoundVertex(double xx, double yy) {
		return isBound(xx, x, x + width - 1) && isBound(yy, y, y + height - 1);
	}

	protected boolean isBound(double value, double lowerBound, double upperBound) {
		return (value == lowerBound) || (value == upperBound);
	}

	@Override
	public String toString() {
		return "DelaunayRectangle{" + "x=" + x + ", y=" + y + ", width=" + width + ", height=" + height + '}';
	}

	public boolean isCorner(Point2d pt) {
		return ( ( pt.x == x ) || ( pt.x == x + width - 1 ) ) &&
				( ( pt.y == y ) || ( pt.y == y + height - 1 ) );
	}

	public List<Point2d> getCorners() {
		List<Point2d> result = new ArrayList<>();
		double left = this.x;
		double right = this.x + this.width - 1;
		double up = this.y;
		double down = this.y + this.height - 1;

		result.add(createPoint(left, up));
		result.add(createPoint(left, down));
		result.add(createPoint(right, down));
		result.add(createPoint(right, up));

		return result;
	}

	@Override
	public boolean equals(Object o) {
		if (this == o) return true;
		if (o == null || getClass() != o.getClass()) return false;
		GeometricRectangle that = (GeometricRectangle) o;
		return Double.compare(that.x, x) == 0 && Double.compare(that.y, y) == 0 && Double.compare(that.width, width) == 0 && Double.compare(that.height, height) == 0;
	}

	@Override
	public int hashCode() {
		return Objects.hash(x, y, width, height);
	}
}
