/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.libgeometry.types;


import com.frojasg1.libgeometry.algebra.helpers.VectorFunctions;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.Objects;
import java.util.function.UnaryOperator;

/**
 *
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 */
public class GeometricTriangle implements Comparable<GeometricTriangle> {

	protected Point2d _vertex1;
	protected Point2d _vertex2;
	protected Point2d _vertex3;

	protected boolean sorted;

	protected Point2d incenter;

	public GeometricTriangle(Point2d p1, Point2d p2, Point2d p3 )
	{
		this( p1, p2, p3, true );
	}

	public GeometricTriangle(Point2d p1, Point2d p2, Point2d p3, boolean sort )
	{
		this.sorted = sort;
		List<Point2d> list = Arrays.asList( p1, p2, p3 );

		if( sort )
			Collections.sort(list);

		_vertex1 = list.get(0);
		_vertex2 = list.get(1);
		_vertex3 = list.get(2);
	}

	public Point2d getVertex1() {
		return _vertex1;
	}

	public Point2d getVertex2() {
		return _vertex2;
	}

	public Point2d getVertex3() {
		return _vertex3;
	}

	public boolean isSorted() {
		return sorted;
	}

	@Override
	public int hashCode() {
		int hash = 7;
		hash = 97 * hash + Objects.hashCode(this._vertex1);
		hash = 97 * hash + Objects.hashCode(this._vertex2);
		hash = 97 * hash + Objects.hashCode(this._vertex3);
		return hash;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj) {
			return true;
		}
		if (obj == null) {
			return false;
		}
		if (getClass() != obj.getClass()) {
			return false;
		}
		final GeometricTriangle other = (GeometricTriangle) obj;
		if (!Objects.equals(this._vertex1, other._vertex1)) {
			return false;
		}
		if (!Objects.equals(this._vertex2, other._vertex2)) {
			return false;
		}
		if (!Objects.equals(this._vertex3, other._vertex3)) {
			return false;
		}
		return true;
	}

	@Override
	public int compareTo(GeometricTriangle o) {
		int result = _vertex1.compareTo(o._vertex1);
		if(result == 0) {
			result = _vertex2.compareTo(o._vertex2);
		}
		if(result == 0) {
			result = _vertex3.compareTo(o._vertex3);
		}
		return result;
	}



	@Override
	public String toString() {
		return "DelaunayTriangle{" +
				"_vertex1=" + _vertex1 +
				", _vertex2=" + _vertex2 +
				", _vertex3=" + _vertex3 +
				'}';
	}

	public Point2d getCenter() {
		return VectorFunctions.instance().arithmeticMean(getVertex1(), getVertex2(), getVertex3());
	}

	protected VectorFunctions getVectorFunctions() {
		return VectorFunctions.instance();
	}

	protected Point2d arithmeticMean(Point2d... points) {
		return getVectorFunctions().arithmeticMean(points);
	}

	protected Point2d zoom(Point2d origin, Point2d destination, double factor) {
		return getVectorFunctions().zoom(origin, destination, factor);
	}

	public GeometricTriangle applyFunctionToVertices(UnaryOperator<Point2d> function) {
		return new GeometricTriangle(
				function.apply(getVertex1()),
				function.apply(getVertex2()),
				function.apply(getVertex3()),
				isSorted()
				);
	}

	public GeometricTriangle zoom(double factor) {
		Point2d incenter = getInCenter();
		return applyFunctionToVertices(pt -> zoom(incenter, pt, factor).toInt());
	}

	protected synchronized Point2d calculateIncenter() {
		if(incenter == null) {
			double d1 = getVertex2().distance(getVertex3());
			double d2 = getVertex1().distance(getVertex3());
			double d3 = getVertex1().distance(getVertex2());
			double sum = d1 + d2 + d3;
			Point2d result = getVertex1().createCopy();
			result.x = (getVertex1().x * d1 + getVertex2().x * d2 + getVertex3().x * d3) / sum;
			result.y = (getVertex1().y * d1 + getVertex2().y * d2 + getVertex3().y * d3) / sum;
			incenter = result;
		}
		return incenter;
	}

	public Point2d getInCenter() {
		if(incenter == null) {
			incenter = calculateIncenter();
		}
		return incenter;
	}
}
