/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.libgeometry.types;

import com.frojasg1.general.number.DoubleFunctions;

/**
 *
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 */
// http://karlchenofhell.org/cppswp/lischinski.pdf
public class Point2d implements Comparable<Point2d> {
	public double x;
	public double y;

	public Point2d ( double x, double y )
	{
		this.x = x;
		this.y = y;
	}

	public Point2d( Point2d that )
	{
		x = that.x;
		y = that.y;
	}

	public Point2d createCopy()
	{
		return( new Point2d( x, y ) );
	}

	protected boolean isInt( double value )
	{
		return( DoubleFunctions.instance().isInt(value) );
	}

	public boolean isInt() {
		return isInt(x) && isInt(y);
	}

	public Point2d createIntCopyIfNotInt()
	{
		return isInt() ? this : createCopy().toInt();
	}

	public Point2d toInt()
	{
		this.x = getIntX();
		this.y = getIntY();

		return this;
	}

	public double distance( Point2d pt )
	{
		return Math.sqrt(squaredDistance(pt));
	}

	public double squaredDistance( Point2d pt )
	{
		double dx = (x - pt.x);
		double dy = (y - pt.y);

		return( dx * dx + dy * dy );
	}

	public Point2d subtract( Point2d pt )
	{
		return( new Point2d( x-pt.x, y-pt.y ) );
	}

	public double norm()
	{
		return( Math.sqrt( squaredNorm() ) );
	}

	public double squaredNorm()
	{
		return( x * x + y * y );
	}

	public int getIntX()
	{
		return( (int) x );
	}

	public int getIntY()
	{
		return( (int) y );
	}

	public Point2d applyDeltas(double deltax, double deltay, Point2d result) {
		if(result == null) {
			result = createCopy();
		}
		result.x = this.x + deltax;
		result.y = this.y + deltay;
		return result;
	}

	@Override
	public int hashCode() {
		int hash = 7;
		return hash;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj) {
			return true;
		}
		if (obj == null) {
			return false;
		}
		if (getClass() != obj.getClass()) {
			return false;
		}
		final Point2d other = (Point2d) obj;
		if (Double.doubleToLongBits(this.x) != Double.doubleToLongBits(other.x)) {
			return false;
		}
		if (Double.doubleToLongBits(this.y) != Double.doubleToLongBits(other.y)) {
			return false;
		}
		return true;
	}

	@Override
	public String toString() {
		return "Point2d{" + "x=" + x + ", y=" + y + '}';
	}

	protected int sgn( double value )
	{
		int result = 0;
		if( value > 0 )
			result = 1;
		else if( value < 0 )
			result = -1;

		return( result );
	}

	@Override
	public int compareTo(Point2d o) {
		int result = sgn( x - o.x );
		if( result == 0 )
			result = sgn( y - o.y );

		return( result );
	}
}
