/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.libgeometry.types.factory;

import com.frojasg1.libgeometry.types.GeometricRectangle;
import com.frojasg1.libgeometry.types.GeometricTriangle;
import com.frojasg1.libgeometry.types.Point2d;
import com.frojasg1.general.UtilFunctions;
import com.frojasg1.general.number.IntegerFunctions;

import java.awt.Dimension;
import java.awt.Point;
import java.awt.Rectangle;
import java.util.function.Function;

/**
 *
 * @author Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 */
public class GeometryCommonFactory
{
	protected static GeometryCommonFactory INSTANCE = new GeometryCommonFactory();
	
	public static GeometryCommonFactory instance()
	{
		return( INSTANCE );
	}

	protected int round(double value) {
		return IntegerFunctions.roundToInt(value);
	}

	public Point2d createPoint2dInt( double xx, double yy )
	{
		return( new Point2d( round(xx), round(yy) ) );
	}

	public Point2d createPoint2d( double xx, double yy )
	{
		return( new Point2d( xx, yy ) );
	}

	public Point2d createPoint2d( Point2d that )
	{
		Point2d result = null;
		if( that != null )
			result = that.createCopy();

		return( result );
	}

	public GeometricRectangle createDelaunayRectangle(double xx, double yy, double width, double height )
	{
		return( new GeometricRectangle( xx, yy, width, height ) );
	}

	public GeometricTriangle createDelaunayTriangle(Point2d pt1, Point2d pt2, Point2d pt3,
													boolean sortPoints )
	{
		return( new GeometricTriangle( pt1, pt2, pt3, sortPoints ) );
	}

	protected <I, I2, O2, O> O functionWithCodec(Function<I2, O2> function, I input, Function<I, I2> encoder, Function<O2, O> decoder) {
		return UtilFunctions.instance().functionWithCodec(function, input, encoder, decoder);
	}

	public GeometricRectangle delaunayRectangleOfDimensionFunction(Function<Dimension, Dimension> function, GeometricRectangle delaunayRectangle) {
		return functionWithCodec(function, delaunayRectangle, this::createDimension, this::createDelaunayRectangle);
	}

	protected int toInt(double value) {
		return IntegerFunctions.roundToInt(value);
	}

	public Dimension createDimension( GeometricRectangle delaunayRectangle )
	{
		Dimension result = null;
		if( delaunayRectangle != null )
			result = new Dimension(toInt(delaunayRectangle.width), toInt(delaunayRectangle.height));

		return( result );
	}

	public GeometricRectangle createDelaunayRectangle(Dimension dimension )
	{
		GeometricRectangle result = null;
		if( dimension != null )
			result = createDelaunayRectangle( 0, 0, dimension.width, dimension.height );

		return( result );
	}

	public GeometricRectangle createDelaunayRectangle(Rectangle rectangle )
	{
		GeometricRectangle result = null;
		if( rectangle != null )
			result = createDelaunayRectangle( rectangle.x, rectangle.y,
											rectangle.width, rectangle.height );

		return( result );
	}

	public Point toPoint( Point2d pt )
	{
		Point result = null;
		if( pt != null )
			result = new Point( (int) pt.x, (int) pt.y );

		return( result );
	}
}
