/* 
 * Copyright (C) 2023 Francisco Javier Rojas Garrido <frojasg1@hotmail.com>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3.0 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You may obtain a copy of the License at
 *
 *      http://www.gnu.org/licenses/lgpl-3.0.txt
 *
 */
package com.frojasg1.general.desktop.queries.api.execution;


import com.frojasg1.applications.common.configuration.InternationalizedStringConf;
import com.frojasg1.general.ExecutionFunctions;
import com.frojasg1.general.NullFunctions;
import com.frojasg1.general.desktop.queries.InetQuery;
import com.frojasg1.general.desktop.queries.InetQueryResultWithResultBase;
import com.frojasg1.general.desktop.queries.language.LibInetQueryStringsConf;
import com.frojasg1.general.view.internationalization.window.GenInternationalizedWindow;
import com.frojasg1.generic.GenericFunctions;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.function.Consumer;
import java.util.function.Function;
import java.util.function.Supplier;

/**
 *
 * @author fjavier.rojas
 */
public abstract class GenericQueryWithResultExecution<KK, RR, QR extends InetQueryResultWithResultBase<RR>, QQ extends InetQuery<QR>> implements Runnable
{
	private static final Logger LOGGER = LoggerFactory.getLogger(GenericQueryWithResultExecution.class);

	public static final String INTERNAL_ERROR = "INTERNAL_ERROR";
	public static final String ERROR_EXECUTING_API_QUERY = "ERROR_EXECUTING_API_QUERY";
	public static final String SUCCESS_EXECUTING_API_QUERY = "SUCCESS_EXECUTING_API_QUERY";
	public static final String API_QUERY_RESULT_FOUND_IN_CACHE = "API_QUERY_RESULT_FOUND_IN_CACHE";

	protected Supplier<String> baseUrlGetter;

	protected GenInternationalizedWindow parentJFrame = null;

	protected Supplier<QQ> queryFactory;
	protected Consumer<RR> callback = null;

	protected InternationalizedStringConf internationalStringsConf;
	protected InternationalizedStringConf baseInternationalStringsConf;

	public GenericQueryWithResultExecution(GenInternationalizedWindow parentJFrame,
										   Supplier<String> baseUrlGetter,
										   Supplier<QQ> queryFactory,
										   InternationalizedStringConf internationalStringsConf,
										   Consumer<RR> callback ) {
		this(parentJFrame, baseUrlGetter, queryFactory, internationalStringsConf, LibInetQueryStringsConf.instance(),
				callback);
	}

	public GenericQueryWithResultExecution(GenInternationalizedWindow parentJFrame,
										   Supplier<String> baseUrlGetter,
										   Supplier<QQ> queryFactory,
										   InternationalizedStringConf internationalStringsConf,
										   InternationalizedStringConf baseInternationalStringsConf,
										   Consumer<RR> callback )
	{
		this.callback = callback;
		this.parentJFrame = parentJFrame;
		this.queryFactory = queryFactory;
		this.internationalStringsConf = internationalStringsConf;
		this.baseInternationalStringsConf = baseInternationalStringsConf;

		this.baseUrlGetter = baseUrlGetter;
	}

	protected abstract String getServiceName();

	protected Supplier<QQ> getQueryFactory() {
		return queryFactory;
	}

	protected Supplier<String> getBaseUrlGetter() {
		return baseUrlGetter;
	}

	protected String getBaseUrl() {
		return getBaseUrlGetter().get();
	}

	protected QQ createFaceDetectionQueryForApplications()
	{
		Supplier<QQ> factory = getQueryFactory();
		QQ result = null;

		if( factory != null )
		{
			result = factory.get();
			result.init( getBaseUrl() );
		}

		return( result );
	}

	protected abstract KK getKey();

	protected abstract RR getCacheResult(KK key);

	protected abstract QR getQueryResult(QQ query);

	protected ExecutionFunctions getExecutionFunctions() {
		return ExecutionFunctions.instance();
	}

	protected <CC> CC runtimeExceptionFunctionExecution(ExecutionFunctions.UnsafeFunction<CC> run) {
		return getExecutionFunctions().runtimeExceptionFunctionExecution(run);
	}

	@Override
	public void run()
	{
		try
		{
			KK key = getKey();
			RR result = getCacheResult(key);

			if(result != null) {
				processCacheResult(result);
			} else {
				QQ query = createFaceDetectionQueryForApplications();
				QR queryResult = (query != null) ?
						getQueryResult(query) :
						null;

				processQueryResult(queryResult);
			}
		}
		catch( Exception ex ) {
			LOGGER.error("Error executing query", ex);
			showError(ex.getMessage());
		}
	}

	protected void processCacheResult(RR result) {
		logAndShowInfo(getBaseInternationalStringsConf().createCustomInternationalString(API_QUERY_RESULT_FOUND_IN_CACHE, getServiceName()));
		invokeCallback(result);
	}

	protected void showError(String errorMessage) {
		String text = getBaseInternationalStringsConf().createCustomInternationalString(ERROR_EXECUTING_API_QUERY, getServiceName(), errorMessage);
		LOGGER.error("{}", text);
		errorDialog(text);
	}

	protected void showSuccess() {
		logAndShowInfo(getBaseInternationalStringsConf().createCustomInternationalString(SUCCESS_EXECUTING_API_QUERY, getServiceName()));
	}

	protected void logAndShowInfo(String text) {
		LOGGER.info("{}", text);
		infoDialog(text);
	}

	public GenInternationalizedWindow getParentJFrame() {
		return parentJFrame;
	}

	protected void errorDialog(String text) {
		GenericFunctions.instance().getDialogsWrapper().showError(getParentJFrame(), text);
	}

	protected void infoDialog(String text) {
		GenericFunctions.instance().getDialogsWrapper().showInformation(getParentJFrame(), text);
	}

	protected Consumer<RR> getCallback() {
		return callback;
	}

	protected NullFunctions getNullFunctions() {
		return NullFunctions.instance();
	}

	protected <CC, RR> RR getIfNotNull(CC obj, Function<CC, RR> getter) {
		return getNullFunctions().getIfNotNull(obj, getter);
	}

	protected RR getResult(QR queryResult) {
		return getIfNotNull(queryResult, InetQueryResultWithResultBase::getResult);
	}

	protected void processQueryResult(QR queryResult) {
		if(queryResult == null) {
			showError(getBaseInternationalStringsConf().getInternationalString(INTERNAL_ERROR));
		} else if(!queryResult.isSuccessful()) {
			showError(queryResult.getErrorMessage());
		} else {
			showSuccess();
			putInCache(queryResult.getResult());
		}

		invokeCallback(getResult(queryResult));
	}

	protected abstract void putInCache(RR result);

	protected InternationalizedStringConf getBaseInternationalStringsConf() {
		return baseInternationalStringsConf;
	}

	protected InternationalizedStringConf getInternationalStringsConf() {
		return internationalStringsConf;
	}

	public String getInternationalString(String label) {
		return getInternationalStringsConf().getInternationalString(label);
	}

	public String createCustomInternationalString(String label, Object... args) {
		return getInternationalStringsConf().createCustomInternationalString(label, args);
	}

	protected void invokeCallback(RR result)
	{
		if( callback != null )
			callback.accept(result);
	}
}
